
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.ajde;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;

import org.aspectj.ajde.internal.*;
import org.aspectj.ajde.ui.*;
import org.aspectj.asm.StructureModelManager;
import org.aspectj.compiler.Version;

/**
 * Singleton class responsible for AJDE initialization, and the main point of access to
 * Ajde functionality. 
 *
 * @author Mik Kersten
 */
public class Ajde {  

	private static final Ajde INSTANCE = new Ajde();
	private static final String NOT_INITIALIZED_MESSAGE = "Ajde is not initialized.";
	private static boolean isInitialized = false;
	private BuildManager buildManager;
	private EditorManager editorManager;
	private StructureViewManager structureViewManager;
	private StructureSearchManager structureSearchManager;
	private BuildConfigManager configurationManager ;
	private ProjectPropertiesAdapter projectProperties;
	private TaskListManager taskListManager;
	private IdeUIAdapter ideUIAdapter;
	private ErrorHandler errorHandler;
	private PrintStream logPrintStream = null;
	
	/**
	 * This class can only be constructured by itself (as a singleton) or by sub-classes. 
	 */ 
	protected Ajde() {
		configurationManager = new LstBuildConfigManager();
	}

	/**
	 * This method must be called before using Ajde.  A <CODE>RuntimeException</CODE> will
	 * be thrown if use is attempted before initialization.
	 */
	public static void init(
			EditorAdapter editorAdapter,
			TaskListManager taskListManager,
			BuildProgressMonitor compileProgressMonitor,
			ProjectPropertiesAdapter projectProperties,
			BuildOptionsAdapter buildOptionsAdapter,
			StructureViewNodeFactory structureViewNodeFactory,
			IdeUIAdapter ideUIAdapter,
			ErrorHandler errorHandler) {
		try {
			INSTANCE.projectProperties = projectProperties;
			INSTANCE.errorHandler = errorHandler;
			INSTANCE.taskListManager = taskListManager;
			INSTANCE.editorManager = new EditorManager(editorAdapter);
			INSTANCE.buildManager = new AspectJBuildManager(
				taskListManager, 
				compileProgressMonitor,
				buildOptionsAdapter);
 
			INSTANCE.buildManager.addListener(INSTANCE.BUILD_STATUS_LISTENER);
			INSTANCE.configurationManager.addListener(INSTANCE.STRUCTURE_UPDATE_CONFIG_LISTENER);
			INSTANCE.ideUIAdapter = ideUIAdapter;
			
			INSTANCE.structureSearchManager = new StructureSearchManager();	
			INSTANCE.structureViewManager = new StructureViewManager(structureViewNodeFactory);
			
			isInitialized = true;
		} catch (Throwable t) {
			System.err.println("AJDE ERROR: could not initialize Ajde.");
			t.printStackTrace();	
		}
	}   

	/**
	 * @return	the default singleton instance of <CODE>Ajde</CODE>
	 */
	public static Ajde getDefault() {
		if (!isInitialized) throw new RuntimeException(NOT_INITIALIZED_MESSAGE);
		return INSTANCE;	
	}

	/**
	 * Set a <CODE>ConfigurationManager</CODE> to use instead of the default one.
	 */ 
	public void setConfigurationManager(BuildConfigManager configurationManager) {
		this.configurationManager = configurationManager;	
	}

	public BuildManager getBuildManager() {
		return buildManager;
	}
	
	public EditorManager getEditorManager() {
		return editorManager;
	}	
	
	public StructureViewManager getStructureViewManager() {
		return structureViewManager;	
	}

	public StructureSearchManager getStructureSearchManager() {
		return structureSearchManager;
	}
	
	public BuildConfigManager getConfigurationManager() {
		return configurationManager;
	}
	
	public ProjectPropertiesAdapter getProjectProperties() {
		return projectProperties;
	}
	
	public TaskListManager getTaskListManager() {
		return taskListManager;
	}

	public IdeUIAdapter getIdeUIAdapter() {
		return ideUIAdapter;
	}

	public void setIdeUIAdapter(IdeUIAdapter ideUIAdapter) {
		this.ideUIAdapter = ideUIAdapter;
	}
	
	public ErrorHandler getErrorHandler() {
		return errorHandler;
	}
	
	public String getVersion() {
		return Version.text;
	}

	public void enableLogging(PrintStream logPrintStream) {
		this.logPrintStream = logPrintStream;
	}
	
	public void disableLogging() {
		this.logPrintStream = null;	
	}
	
	/**
	 * The structure manager is not a part of the public API and its
	 * use should be avoided.  Used <CODE>getStructureViewManager()</CODE>
	 * instead.
	 */
	public StructureModelManager getStructureModelManager() {
		return StructureModelManager.INSTANCE;	
	}
	
	public void logEvent(String message) {
		if (logPrintStream != null) {
			logPrintStream.println("<AJDE> " + message);	
		}	
	}

	private final BuildConfigListener STRUCTURE_UPDATE_CONFIG_LISTENER = new BuildConfigListener() {
		public void currConfigChanged(String configFilePath) {
			if (configFilePath != null) Ajde.getDefault().getStructureModelManager().readStructureModel(configFilePath);
		}
		
		public void configsListUpdated(List configsList) { }
	};
	
	private final BuildListener BUILD_STATUS_LISTENER = new BuildListener() {
    	
    	/**
    	 * Writes the default configuration file if it has been selected for compilation
    	 */
    	public void compileStarted(String buildConfig) { 
    		String configFilePath = projectProperties.getDefaultBuildConfigFile();
    		if (buildConfig.equals(configFilePath)) {
	    		configurationManager.writePaths(configFilePath, projectProperties.getProjectSourceFiles());	
	    		logEvent("wrote default build config: " + configFilePath);
    		}
    	} 
    	
    	/**
    	 * The strucutre model is annotated with error messages after an unsuccessful compile.
    	 */
        public void compileFinished(String buildConfig, int buildTime, boolean succeeded, boolean warnings) { 
        	String configFilePath = projectProperties.getDefaultBuildConfigFile();
        	if (!succeeded) {
	        	StructureModelManager.INSTANCE.fireModelUpdated();	
    	    }
        }
        
        /**
         * Ignored.
         */
        public void compileAborted(String buildConfigFile, String message) { }
    };
}


