
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
 
package org.aspectj.ajde.internal;

import java.io.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.asm.*;
import org.aspectj.ajde.*;

/**
 * @author Mik Kersten
 */
public class AjdeBuildErrorHandler extends org.aspectj.compiler.base.ErrorHandler {

	protected AjdeCompiler compiler;
	protected TaskListManager taskListManager;
    protected int errors = 0;
    protected int warnings = 0;

    public void cleanupErrorWarningCounters() {
        errors = 0;
        warnings = 0;
    }

    public AjdeBuildErrorHandler(TaskListManager taskListManager) {
        this.taskListManager = taskListManager;
    }

    public synchronized boolean showMessage(String message) {
        return showMessage(null, message, false, false);
    }

    public synchronized boolean showMessage(ASTObject where, String message) {
        return showMessage(where, message, false, false);
    }

    public synchronized boolean showMessage(File source, int line, int endLine, int column, String message, boolean isError) {
		return addMessage(source, line, column, message, isError);
    }

    public synchronized boolean showMessage(ASTObject where, String message, boolean showAll, boolean isError) {
        if (where == null) {
            return showMessage(null, -1, -1, -1, message, isError);
        }
        if (where.getSourceFile() == null || where.getBeginLine() == -1) {
            
            ASTObject newWhere = where.getSourceLocation().getSourceObject();
            if (newWhere != null) {
                return showMessage(newWhere, message, showAll, isError);
            }
        }
        return addMessage(where.getSourceFile(), where.getBeginLine(),
                    where.getBeginColumn(), message, isError);
    }

    public synchronized void showError(ASTObject where, String message) {
        if (showMessage(where, message, false, true)) errors += 1;
    }

    public synchronized void showError(File source, int line, int column, String message) {
        if (showMessage(source, line, -1, column, message, true)) errors += 1;
    }

    public synchronized void showWarning(ASTObject where, String message) {
        if (errors > 0) return;
        if (showMessage(where, message, false, false)) warnings += 1;
    }


	public synchronized void showWarning(ASTObject where, Message message) {
		if (compiler.getOptions().Xlint) {
			showWarning(where, message.getString());
		}
	}  

	protected boolean addMessage(
		File source,
		int line,
		int column,
		String message,
		boolean isError) {
		if (taskListManager == null) return false;
		StructureMessage.Kind kind; 
		if (isError) {
			kind = StructureMessage.Kind.ERROR;
		} else {
			kind = StructureMessage.Kind.WARNING;
		}
		
		if (source != null) {
			taskListManager.addSourcelineTask(
				message, 
				new org.aspectj.asm.SourceLocation(source.getPath(), line, column),
				kind);
			StructureNode node = Ajde.getDefault().getStructureModelManager().getStructureModel().findNodeForSourceLine(
				source.getPath(),
				line
			);
			
			if (node != null) node.setMessage(new StructureMessage(message, kind));	
		} else {
			taskListManager.addProjectTask(message, kind);			
		}  
		
		return true;
	}

    public synchronized boolean willExitWithErrors() {
        return errors > 0;
    }

    public synchronized void exitOnErrors() {
        if (errors > 0) {
            showWarnings();
            throw new CompilerErrors(errors);
        }
    }

	public int getErrorCount() {
		return errors;
	}

	public int getWarningCount() {
		return warnings;
	}

	public void setCompiler(AjdeCompiler compiler) {
		this.compiler = compiler;
	}
}
