
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
 
package org.aspectj.ajde.ui;

import java.util.*;
import org.aspectj.ajde.*;
import org.aspectj.ajde.ui.internal.*;
import org.aspectj.asm.*;
import org.aspectj.asm.associations.*;

/**
 * @author	Mik Kersten
 */
public class StructureViewManager {

	private TreeStructureViewBuilder treeViewBuilder;
	private String buildConfigFilePath = null;

	private NavigationHistoryModel historyModel = new NavigationHistoryModel();
	private ArrayList structureViews = new ArrayList();	
	private FileStructureView defaultFileView = null;
	
    private static final StructureViewProperties DEFAULT_VIEW_PROPERTIES; 
    private static final List AVAILABLE_RELATIONS;
	
    public final StructureModelListener VIEW_LISTENER = new StructureModelListener() {
        public void modelUpdated(StructureModel model) {        	
        	Ajde.getDefault().logEvent("updating structure views: " + structureViews);
        	
        	if (defaultFileView != null) {
        		defaultFileView.setSourceFile(Ajde.getDefault().getEditorManager().getCurrFile());
        	}
        	
        	for (Iterator it = structureViews.iterator(); it.hasNext(); ) {
        		treeViewBuilder.buildView((StructureView)it.next(), (StructureModel)model);
        	}
        }
    }; 
  
  	/**
  	 * @param nodeFactory			concrete factory for creating view nodes
  	 */
	public StructureViewManager(StructureViewNodeFactory nodeFactory) {
		treeViewBuilder = new TreeStructureViewBuilder(nodeFactory);
				
		StructureModelManager.INSTANCE.addListener(VIEW_LISTENER);			
	}
	
	public void fireNavigateBackAction(StructureView view) {
		ProgramElementNode backNode = historyModel.navigateBack();
		
		if (backNode == null) {
			Ajde.getDefault().getIdeUIAdapter().displayStatusInformation("No node to navigate back to in history");	
		} else {
			navigationAction(backNode, false);
		}
	}
  
	public void fireNavigateForwardAction(StructureView view) {
		ProgramElementNode forwardNode = historyModel.navigateForward();
		
		if (forwardNode == null) {
			Ajde.getDefault().getIdeUIAdapter().displayStatusInformation("No node to navigate forward to in history");	
		} else {
			navigationAction(forwardNode, false);
		}
	}

	/**
	 * Only navigations of the default view are registered with
	 * the history.
	 */
	public void fireNavigationAction(String newFilePath, int lineNumber) {				
		StructureNode currNode = Ajde.getDefault().getStructureModelManager().getStructureModel().findNodeForSourceLine(
			newFilePath,
			lineNumber);
		
		if (currNode instanceof ProgramElementNode) {
			navigationAction((ProgramElementNode)currNode, true);	
		}
	} 
		
	/**
	 * History is recorded for {@link LinkNode} navigations.
	 */
	public void fireNavigationAction(StructureNode structureNode) {
		ProgramElementNode node = null;
		boolean recordHistory = false;
		if (structureNode instanceof LinkNode) {
			node = ((LinkNode)structureNode).getProgramElementNode();
			recordHistory = true;
		} else if (structureNode instanceof ProgramElementNode) {
			node = (ProgramElementNode)structureNode;
		}
		if (node != null) navigationAction(node, recordHistory);
	}

	/**
	 * Highlights the given node in all structure views.  If the node represents code
	 * and as such is below the granularity visible in the view the parent is highlighted,
	 * along with the corresponding sourceline.
	 */ 
	private void navigationAction(ProgramElementNode node, boolean recordHistory) { 
		if (node == null 
			|| node == Ajde.getDefault().getStructureModelManager().getStructureModel().NO_STRUCTURE
			|| node.getSourceLocation().getSourceFilePath() == null) {
			Ajde.getDefault().getIdeUIAdapter().displayStatusInformation("Source not available for node: " + node.getName());
			return;    	
		}
		Ajde.getDefault().logEvent("navigating to node: " + node + ", recordHistory: " + recordHistory);
		if (recordHistory) historyModel.navigateToNode(node); 
    	if (defaultFileView != null && node.getSourceLocation() != null) {
    		String newFilePath = node.getSourceLocation().getSourceFilePath();
			if (defaultFileView.getSourceFile() != null
				&& !defaultFileView.getSourceFile().equals(newFilePath)) {
				defaultFileView.setSourceFile(newFilePath);
				treeViewBuilder.buildView(defaultFileView, StructureModelManager.INSTANCE.getStructureModel());
			}
		}
		   
	    for (Iterator it = structureViews.iterator(); it.hasNext(); ) {
    		StructureView view = (StructureView)it.next();
    		if (!(view instanceof GlobalStructureView) || !recordHistory || defaultFileView == null) {
	    		if (node.getProgramElementKind().equals(ProgramElementNode.Kind.CODE)) {
	    			ProgramElementNode parentNode = (ProgramElementNode)node.getParent();
	    			if (parentNode != null) {
		    			StructureViewNode currNode = view.findCorrespondingViewNode(parentNode);
		    			int lineOffset = node.getSourceLocation().getLineNumber() - parentNode.getSourceLocation().getLineNumber();
		    			if (currNode != null) view.setActiveNode(currNode, lineOffset);
	    			}
	    		} else {
	    			StructureViewNode currNode = view.findCorrespondingViewNode(node);
	    			if (currNode != null) view.setActiveNode(currNode);	
	    		}	
    		}
    	}
	}
	
	private ProgramElementNode getProgramElementNode(StructureViewNode node) {
		if (node.getStructureNode() instanceof ProgramElementNode) {
			return (ProgramElementNode)node.getStructureNode();	
		} else if (node.getStructureNode() instanceof LinkNode) {
			return ((LinkNode)node.getStructureNode()).getProgramElementNode();	
		} else {
			return null;
		}	
	}
	
	public void refreshView(StructureView view) {
		StructureViewNode activeNode = view.getActiveNode();
		treeViewBuilder.buildView(view, Ajde.getDefault().getStructureModelManager().getStructureModel());
		view.setActiveNode(activeNode);
	}		

	public StructureViewProperties getDefaultViewProperties() {
		return DEFAULT_VIEW_PROPERTIES;
	}

	/**
	 * Returns the list of all available relations.
	 */
	public List getAvailableRelations() {
		return AVAILABLE_RELATIONS;
	}
	
	/**
	 * @param	properties		can not be null
	 */
	public GlobalStructureView createGlobalView(GlobalViewProperties properties) {
		GlobalStructureView view = new GlobalStructureView(properties);
		structureViews.add(view);
		return view;
	}
	
	/**
	 * @param	sourceFilePath	full path to corresponding source file
	 * @param	properties		if null default properties will be used
	 * @return					always returns a view intance
	 */ 
	public FileStructureView createViewForSourceFile(String sourceFilePath, StructureViewProperties properties) {
		Ajde.getDefault().logEvent("creating view for file: " + sourceFilePath);
		if (properties == null) properties = DEFAULT_VIEW_PROPERTIES;
		FileStructureView view = new FileStructureView(properties);
		view.setSourceFile(sourceFilePath);
		treeViewBuilder.buildView(view, StructureModelManager.INSTANCE.getStructureModel()); 
		structureViews.add(view);
		return view; 
	}

	/**
	 * @return	true if the view was found and removed, false otherwise
	 */
	public boolean deleteView(StructureView view) {
		return structureViews.remove(view);
	}

	public void setDefaultFileView(FileStructureView defaultFileView) {
		this.defaultFileView = defaultFileView;
	}

	static {
		AVAILABLE_RELATIONS = new ArrayList();
        AVAILABLE_RELATIONS.add(Advice.METHOD_CALL_SITE_RELATION);
        AVAILABLE_RELATIONS.add(Advice.METHOD_RELATION);
        AVAILABLE_RELATIONS.add(Advice.CONSTRUCTOR_CALL_SITE_RELATION);
        AVAILABLE_RELATIONS.add(Advice.CONSTRUCTOR_RELATION);
        AVAILABLE_RELATIONS.add(Advice.FIELD_ACCESS_RELATION);
        AVAILABLE_RELATIONS.add(Advice.INITIALIZER_RELATION);
        AVAILABLE_RELATIONS.add(Advice.HANDLER_RELATION);
        AVAILABLE_RELATIONS.add(Advice.INTRODUCTION_RELATION);
        AVAILABLE_RELATIONS.add(Introduction.INTRODUCES_RELATION);
        AVAILABLE_RELATIONS.add(Inheritance.IMPLEMENTS_RELATION);
        AVAILABLE_RELATIONS.add(Inheritance.INHERITS_RELATION);
        AVAILABLE_RELATIONS.add(Inheritance.INHERITS_MEMBERS_RELATION);
        AVAILABLE_RELATIONS.add(Reference.USES_POINTCUT_RELATION);
        AVAILABLE_RELATIONS.add(Reference.IMPORTS_RELATION);
        
        DEFAULT_VIEW_PROPERTIES = new StructureViewProperties();
        DEFAULT_VIEW_PROPERTIES.setRelations(AVAILABLE_RELATIONS);
	}   
}

//		this.multiFileViewMode = multiFileViewMode;
//		if (!multiFileViewMode) {
//			structureViews.add(DEFAULT_FILE_VIEW);
//			structureViews.add(DECLARATION_VIEW);
//			structureViews.add(CROSSCUTTING_VIEW);
//			structureViews.add(INHERITANCE_VIEW);
//		}

//	public GlobalStructureView getGlobalStructureView(StructureViewProperties.Hierarchy hierarchy) {
//		if (hierarchy == StructureViewProperties.Hierarchy.CROSSCUTTING) {
//			return CROSSCUTTING_VIEW;
//		} else if (hierarchy == StructureViewProperties.Hierarchy.INHERITANCE) {
//			return INHERITANCE_VIEW;
//		} else {
//			return DECLARATION_VIEW;
//		} 		
//	}

//	public FileStructureView getDefaultFileStructureView() {
//		return DEFAULT_FILE_VIEW;
//	}


