
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.asm;

import java.util.*;
import java.io.*;
import org.aspectj.compiler.base.ast.*;


/**
 * @author Mik Kersten
 */
public class ProgramElementNode extends StructureNode {
		
    private List modifiers;
    private List relations;

	private Kind kind;
	private Accessibility accessibility;
    private String declaringType = "";
    private String formalComment = "";
    private String packageName = null;
    private boolean member = false;
    private boolean runnable = false;
    private boolean implementor = false; 
    private boolean overrider = false;
    
    private transient ASTObject astObject = null;

    /**
     * Used during de-externalization.
     */
    public ProgramElementNode() { }

	/**
	 * Use to create program element nodes that do not correspond to source locations.
	 */
	public ProgramElementNode(String signature, Kind kind, List children) {
		super(signature, kind.toString(), children);
		this.kind = kind;
	}
	
	/**
	 * Use to create program element nodes that correspond to source locations.
	 */
    public ProgramElementNode(String signature, Kind kind, List modifiers, Accessibility accessibility,
        String declaringType, String packageName, String formalComment, SourceLocation sourceLocation,
        List relations, List children, boolean member, ASTObject astObject) {

        super(signature, kind.toString(), children);
        super.sourceLocation = sourceLocation;
        this.kind = kind;
        this.modifiers = modifiers;
        this.accessibility = accessibility;
        this.declaringType = declaringType;
        this.packageName = packageName;
        this.formalComment = formalComment;
        this.relations = relations;
        this.member = member;
        this.astObject = astObject;
    }

	public Kind getProgramElementKind() {
		return kind;	
	}

    public List getModifiers() {
        return modifiers;
    }

    public Accessibility getAccessibility() {
        return accessibility;
    }

    public String getDeclaringType() {
        return declaringType;
    }

    public String getPackageName() {
        return packageName;
    }

    public ASTObject getAstObject() {
        return astObject;
    }
    
    public void clearAstObject() {
    	this.astObject = null;	
    }

    public String getKind() {
        return super.kind;
    }

    public String getSignature() {
        return super.name;
    }

    public boolean isCode() {
        return kind.equals(Kind.CODE);
    }

    public boolean isMemberKind() {
        return member;
    }

	public void setRunnable(boolean value) {
		this.runnable = value;	
	}

	public boolean isRunnable() {
		return runnable;	
	}

	public boolean isImplementor() {
		return implementor;	
	}

	public void setImplementor(boolean value) {
		this.implementor = value;	
	}
	
	public boolean isOverrider() {
		return overrider;		
	}

	public void setOverrider(boolean value) {
		this.overrider = value;	
	}

    public List getRelations() {
        return relations;
    }

    public void setRelations(List relations) {
        if (relations.size() > 0) {
            this.relations = relations;
        }
    }

    public String getFormalComment() {
        return formalComment;
    }

    public String toString() {
        return super.name;
    }

//	/**
//	 * Fix this to represent data-equality. 
//	 */
//    public boolean equals(ProgramElementNode node) {
//        if (node == null) {
//            return false;
//        } else {
//            if (sourceLocation != null && 
//            	name.equals(node.getName()) &&
//                sourceLocation.getSourceFilePath().equals(node.getSourceLocation().getSourceFilePath()) &&
//                sourceLocation.getLineNumber() == node.getSourceLocation().getLineNumber() &&
//                sourceLocation.getColumnNumber() == node.getSourceLocation().getColumnNumber()) {
//                return true;
//            } else {
//                return false;
//            }
//        }
//        //super.equals(node);
//    }
    
    /**
	 * Uses "typesafe enum" pattern.
	 */
	public static class Modifiers implements Serializable {
		
		public static final Modifiers STATIC = new Modifiers("static");
		public static final Modifiers FINAL = new Modifiers("final");
		public static final Modifiers ABSTRACT = new Modifiers("abstract");
		public static final Modifiers SYNCHRONIZED = new Modifiers("synchronized");
		public static final Modifiers VOLATILE = new Modifiers("volatile");
		public static final Modifiers STRICTFP = new Modifiers("strictfp");
		public static final Modifiers TRANSIENT = new Modifiers("transient");
		public static final Modifiers NATIVE = new Modifiers("native");
		public static final Modifiers[] ALL = { STATIC, FINAL, ABSTRACT, SYNCHRONIZED, TRANSIENT, VOLATILE, STRICTFP, NATIVE };
		private final String name;
		
		private Modifiers(String name) {
			this.name = name;
		}
		
		public String toString() {
			return name;
		}	

		// The 4 declarations below are necessary for serialization
		private static int nextOrdinal = 0;
		private final int ordinal = nextOrdinal++;
		private Object readResolve() throws ObjectStreamException {
			return ALL[ordinal];
		}
	}

	/**
	 * Uses "typesafe enum" pattern.
	 */
	public static class Accessibility implements Serializable {
		
		public static final Accessibility PUBLIC = new Accessibility("public");
		public static final Accessibility PACKAGE = new Accessibility("package");
		public static final Accessibility PROTECTED = new Accessibility("protected");
		public static final Accessibility PRIVATE = new Accessibility("private");
		public static final Accessibility PRIVILEGED = new Accessibility("privileged");
		public static final Accessibility[] ALL = { PUBLIC, PACKAGE, PROTECTED, PRIVATE, PRIVILEGED };
		private final String name;
		
		private Accessibility(String name) {
			this.name = name;
		}
		
		public String toString() {
			return name;
		}	
		
		// The 4 declarations below are necessary for serialization
		private static int nextOrdinal = 0;
		private final int ordinal = nextOrdinal++;
		private Object readResolve() throws ObjectStreamException {
			return ALL[ordinal];
		}
	}

	/**
	 * Uses "typesafe enum" pattern.
	 */
	public static class Kind implements Serializable {
		
		public static final Kind PROJECT = new Kind("project");
		public static final Kind PACKAGE = new Kind("package");
		public static final Kind FILE = new Kind("file");
		public static final Kind FILE_JAVA = new Kind("java source file");
		public static final Kind FILE_ASPECTJ = new Kind("aspect source file");
		public static final Kind FILE_LST = new Kind("build configuration file");
		public static final Kind CLASS = new Kind("class");
		public static final Kind INTERFACE = new Kind("interface");
		public static final Kind ASPECT = new Kind("aspect");
		public static final Kind INITIALIZER = new Kind("initializer");
		public static final Kind INTRODUCTION = new Kind("introduction");
		public static final Kind CONSTRUCTOR = new Kind("constructor");
		public static final Kind METHOD = new Kind("method");
		public static final Kind FIELD = new Kind("field");  
		public static final Kind POINTCUT = new Kind("pointcut");
		public static final Kind ADVICE = new Kind("advice");
		public static final Kind DECLARE_PARENTS = new Kind("declare parents");
		public static final Kind DECLARE_WARNING = new Kind("declare warning");
		public static final Kind DECLARE_ERROR = new Kind("declare error");
		public static final Kind DECLARE_SOFT = new Kind("declare soft");
		public static final Kind CODE = new Kind("decBodyElement");
		public static final Kind ERROR = new Kind("error");

		public static final Kind[] ALL = { PROJECT, PACKAGE, FILE, FILE_JAVA, 
			FILE_ASPECTJ, FILE_LST, CLASS, INTERFACE, ASPECT,
			INITIALIZER, INTRODUCTION, CONSTRUCTOR, METHOD, FIELD, POINTCUT, ADVICE, 
			DECLARE_PARENTS, DECLARE_WARNING, DECLARE_ERROR, DECLARE_SOFT, CODE, ERROR };
		
		public static Kind getKindForString(String kindString) {
			for (int i = 0; i < ALL.length; i++) {
				if (ALL[i].toString().equals(kindString)) return ALL[i];	
			}
			return ERROR;
		}
		
		private final String name;
		
		private Kind(String name) {
			this.name = name;
		}
		
		public String toString() {
			return name;
		}	
		
//		public boolean equals(Object o) {
//			return o.equals(name);	
//		}
		
		public static List getNonAJMemberKinds() {
			List list = new ArrayList();
			list.add(METHOD);
			list.add(FIELD);
			list.add(CONSTRUCTOR);
			return list;
		}
		
		public boolean isMemberKind() {
			return this == FIELD
				|| this == METHOD
				|| this == CONSTRUCTOR
				|| this == POINTCUT
				|| this == ADVICE;
		}
		
		public boolean isTypeKind() {
			return this == CLASS
				|| this == INTERFACE
				|| this == ASPECT;	
		}

		public boolean isSourceFileKind() {
			return this == FILE_ASPECTJ
				|| this == FILE_JAVA;
		}
		
		public boolean isDeclareKind() {
			return name.startsWith("declare");	
		} 

		// The 4 declarations below are necessary for serialization
		private static int nextOrdinal = 0;
		private final int ordinal = nextOrdinal++;
		private Object readResolve() throws ObjectStreamException {
			return ALL[ordinal];
		}
	}
}

//    /**
//     * Creates and returns a copy of this object.
//     */
//	public Object clone() {
//		List cloneModifiers = new ArrayList();
//		List cloneRelations = new ArrayList();
//		for (Iterator it = children.iterator(); it.hasNext(); ) {
//			cloneChildren.add(((StructureNode)it.next()).clone());	
//		}
//    	StructureNode cloneNode = new StructureNode(name, kind, cloneChildren);
//    	return cloneNode;
//	} 
