
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.asm;

import java.io.File;
import java.util.*;
import java.io.Serializable;

/**
 * @author Mik Kersten
 */
public class StructureModel implements Serializable {
	
    protected  StructureNode root = null;
    protected String configFile = null;
    private Map fileMap = null;
    public static final ProgramElementNode NO_STRUCTURE = new ProgramElementNode("<build to view structure>", ProgramElementNode.Kind.ERROR, null);

    public StructureNode getRoot() {
        return root;
    }

    public void setRoot(StructureNode root) {
        this.root = root;
    }

	public Map getFileMap() {
        return fileMap;
    }


	public void setFileMap(HashMap fileMap) {
        this.fileMap = fileMap;
    }


	public boolean isValid() {
        return root != null && fileMap != null;
    }

	/**
	 * @param		sourceFilePath	modified to '/' delimited path for consistency
	 * @return		a new structure node for the file if it was not found in the model
	 */
	public StructureNode findRootNodeForSourceFile(String sourceFilePath) {
       	if (!isValid() || sourceFilePath == null) {  
            return StructureModel.NO_STRUCTURE;
        } else {
            String correctedPath = sourceFilePath.replace('\\', '/');
            StructureNode node = (StructureNode)getFileMap().get(correctedPath);//findFileNode(filePath, model);
            if (node != null) {
                return node;
            } else {
                return createFileStructureNode(sourceFilePath);
            }
        }
    }

	/**
	 * Never returns null 
	 * 
	 * @param		sourceFilePath	modified to '/' delimited path for consistency
	 * @param 		lineNumber		if 0 or 1 the corresponding file node will be returned
	 * @return		a new structure node for the file if it was not found in the model
	 */
	public StructureNode findNodeForSourceLine(String sourceFilePath, int lineNumber) {
		String correctedPath = sourceFilePath.replace('\\', '/');
		StructureNode node = findNodeForSourceLineHelper(root, correctedPath, lineNumber);
		if (node != null) {
			return node;	
		} else {
			return createFileStructureNode(sourceFilePath);
		}
	}

	private StructureNode createFileStructureNode(String sourceFilePath) {
		String fileName = new File(sourceFilePath).getName();
		ProgramElementNode fileNode = new ProgramElementNode(fileName, ProgramElementNode.Kind.FILE_JAVA, null);
		fileNode.setSourceLocation(new SourceLocation(sourceFilePath, 1, 1));
		fileNode.addChild(NO_STRUCTURE);
		return fileNode;
	}


	private StructureNode findNodeForSourceLineHelper(StructureNode node, String sourceFilePath, int lineNumber) {
		if (matches(node, sourceFilePath, lineNumber) 
			&& !hasMoreSpecificChild(node, sourceFilePath, lineNumber)) {
			return node;	
		} 
		
		if (node != null && node.getChildren() != null) {
			for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
				StructureNode foundNode = findNodeForSourceLineHelper(
					(StructureNode)it.next(), 
					sourceFilePath, 
					lineNumber); 		
				if (foundNode != null) return foundNode;
			}
		}
		
		return null;		
	}

	private boolean matches(StructureNode node, String sourceFilePath, int lineNumber) {
		return node != null 
			&& node.getSourceLocation() != null
			&& node.getSourceLocation().getSourceFilePath().equals(sourceFilePath)
			&& ((node.getSourceLocation().getLineNumber() <= lineNumber
				&& node.getSourceLocation().getEndLineNumber() >= lineNumber)
			    ||
				(lineNumber <= 1
				 && node instanceof ProgramElementNode 
				 && ((ProgramElementNode)node).getProgramElementKind().isSourceFileKind())	
			);
	}
	
	private boolean hasMoreSpecificChild(StructureNode node, String sourceFilePath, int lineNumber) {
		for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
			ProgramElementNode child = (ProgramElementNode)it.next();
			if (matches(child, sourceFilePath, lineNumber)) return true;
		}
		return false;
	}

	public String getConfigFile() {
		return configFile;
	}

	public void setConfigFile(String configFile) {
		this.configFile = configFile;
	}

}

