/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;
import org.aspectj.compiler.base.ast.*;

import org.aspectj.tools.ide.SymbolManager;


import java.util.*;
import java.io.*;

public class CodeGenerator extends AbstractCompilerPass implements WorldPass, CompilationUnitPass {

    public CodeGenerator(JavaCompiler compiler) {
        super(compiler);
    }

    public String getDisplayName() {
        return "generating Java sources";
    }

    public void clearState() {
    }

    public void addFileToCompile(File f) {
        getCompiler().addFileToCompile(f);
    }

    public void transform(CompilationUnit cu) {
        File outputDir = getOptions().workingDir;
        try {
            unparseCompilationUnit(outputDir, cu);
        } catch (IOException ioe) {
            getCompiler().internalError(ioe, cu);
            return;
        }
    }


    public void unparseCompilationUnit(File outputdir, CompilationUnit cu)
      throws IOException
    {
        Decs decs = cu.getDecs();
        final int N = decs.size();
        for(int i=0; i<N; i++) {
            Dec dec = decs.get(i);
            unparseTypeDec(outputdir, (TypeDec)dec);
        }
    }

    final int FILE_SIZE_GUESS = 32*1024;

    private void unparseTypeDec(File outputdir, TypeDec typeDec) throws IOException {
        getCompiler().showMessage(this, typeDec.toShortString());

        File sourceFile = typeDec.getCompilationUnit().getSourceFile();
        File inputdir;
        if (sourceFile.getParent() == null) inputdir = new File(".");
        else inputdir = new File(sourceFile.getParent());
        String filename = typeDec.getId()+".java";
        String packagename = typeDec.getPackageName();

        Set topLevelTypes = getTopLevelTypes(typeDec);

        //ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        //Writer writer = new org.aspectj.util.UnicodeEscapeWriter(byteStream);

        File packagedir = typeDec.getPackageDir(outputdir);
        packagedir.mkdirs();

        File clsDir = packagename == null ? getOptions().outputDir
                                          : new File(getOptions().outputDir, packagename);
        File clsFile = new File(clsDir, typeDec.getId()+".class");
        //getCompiler().getProject().addBytecodeFile(sourceFile,clsFile);

        File outFile = new File(packagedir, filename);

        // maybe wrap this in a try/catch to print a partial stream on failure
//         LineMap lineMap = null;
//         if (!getOptions().nosymbols) {
//             lineMap = new LineMap(outFile, getWorld().getOutputToSourceLineMap(), 
//                                            getWorld().getSourceToOutputLineMap());
//         }

        CodeWriter unparser = new CodeWriter(getCompiler(), FILE_SIZE_GUESS, null); //lineMap);
        unparser.writeHeader();
        if (typeDec.getCompilationUnit().getPackageName() != null) {
            unparser.write("package ");
            unparser.write(typeDec.getCompilationUnit().getPackageName());
            unparser.writeln(";");
        }

        // write imports for all top-level types
        for(Iterator iter = topLevelTypes.iterator(); iter.hasNext(); ) {
            Type topType = (Type)iter.next();
            if (!topType.isInnerType() && 
                (topType.isClass() || topType.isInterface()) &&
                !(topType == getTypeManager().TYPE_NOT_FOUND) &&
                !(topType.getString().startsWith("<"))) {
                unparser.writeln("import "+topType.getString()+";");
            }
        }

        unparser.write(typeDec.getCompilationUnit().getImports());
        
        try {
            unparser.write(typeDec);
        } catch (RuntimeException e) {
            //writer.flush();
            e.printStackTrace();
            unparser.writeTo(System.out);
            throw e;
        }
        //writer.flush();

        if (outputdir == null) {
            // maybe do some chunking on this one to reduce memory overhead?
            unparser.writeTo(System.out);
        } else {
            // make package dir for sources
            if (outFile.exists()) {
                if (outFile.getCanonicalPath().equals(sourceFile.getCanonicalPath())) {
                    //XXX error -- writing result file over original
                }
            }
            FileOutputStream fileStream = new FileOutputStream(outFile);
            try {
                unparser.writeTo(fileStream);
            } finally {
                fileStream.close();
            }
            addFileToCompile(outFile);
        }
    }

    private Set getTopLevelTypes(ASTObject typeDec) {
        Set topTypesSet = new HashSet();
        // only generate these imports in non-top-level cu's
        if (typeDec.getCompilationUnit().getPackageName() == null) return topTypesSet;
        
        TopTypesFinder ttFinder = new TopTypesFinder(getCompiler(),topTypesSet);
        ttFinder.process(typeDec);
        return topTypesSet;
    }
}

class TopTypesFinder extends Walker {
    private Set topTypesSet;

    public TopTypesFinder(JavaCompiler compiler, Set topTypesSet) {
        super(compiler);
        this.topTypesSet = topTypesSet;
    }

    public void postProcess(ASTObject obj) {
        if (obj instanceof TypeD) {
            Type type = ((TypeD)obj).getType();
            if (type.getPackageName() == null) {
                topTypesSet.add(type);
            }
        } else if (obj instanceof FieldAccessExpr) {
            FieldAccessExpr fae = (FieldAccessExpr)obj;
            if (fae.getField() == null) return;
            if (fae.getField().isStatic()) {
                Type type = fae.getField().getDeclaringType();
                if (type != null && type.getPackageName() == null) {
                    topTypesSet.add(type);
                }
                return;
            }
        }
    }
}
