/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;

import java.util.*;

/** This pass signals errors for

    <ul>
      <li>used-before-assigned vars and blank-final fields</li>
      <li>assigned-twice blank finals</li>
      <li>constructors not filling in blank final fields</li>
      <li>unreachable stmts</li>
      <li>missing return stmts</li>
      <li>various illegal try/catch stmts</li>
    </ul>

    <p> It makes exactly one side-effect: In anonymous classes, it
    sets up the throws clause to include any thrown checked
    exceptions, as required by JLS 15.9.5.1.  </p>

    <p> Other than that one side-effect, it only signals errors.  </p>
*/

public final class FlowCheckerPass extends WalkerPass {
    public FlowCheckerPass(JavaCompiler jc) {
        super(jc);
    }

    public FlowCheckerPass(JavaCompiler jc, TypeDec currentTypeDec) {
        super(jc);
        currentType = currentTypeDec.getType();
    }

    public FlowCheckerPass(JavaCompiler jc, Set da, TypeDec currentTypeDec) {
        super(jc);
        setVars(Vars.makeSets(da, Set.getNone()));
        currentType = currentTypeDec.getType();
    }

    public FlowCheckerPass(JavaCompiler jc, Vars v, TypeDec currentTypeDec) {
        super(jc);
        setVars(v);
        currentType = currentTypeDec.getType();
    }

    private Type currentType = null;
    public boolean isCurrent(FieldDec fd) {
        return fd.getDeclaringType() == currentType;
    }

    public String getDisplayName() { return "analyzing control flow"; }

    // all process does is walk through types until we get to
    // something useful.  This is never called within exprs/stmts,
    // except for local classes.

    public ASTObject process(ASTObject object) {
        return this.process(object, true);
    }
    public ASTObject processBoolean(ASTObject object) {
        return this.process(object, false);
    }

    private ASTObject process(ASTObject object, boolean cleanUp) {
        if (object == null) return object;
        //printDebug(i++, object, true);
        if (! isLive() && object instanceof Stmt && ((Stmt)object).mustBeLive()) {
            object.showError("unreachable statement");
        } else {
            object.walkFlow(this);
        }
        if (cleanUp) {
            setVars(getVars().getTrue().join(getVars().getFalse()));
        }
        //printDebug(--i, object, false);

        return object;
    }

    int i = 0;
    private void printDebug(int i, ASTObject object, boolean in) {
        for (int j = 0; j < i; j++) System.err.print(" ");
        System.err.println("[" + i + "] da/dua:" + vars + " pa:" + possbilyAssigned
                           + " exns:" + getExns() + " "
                           + (liveFlag ? "live" : "dead") + " "
                           + (in ? " --> " : " <-- ") + object);
    }

    // ------------------------------
    // state

    private Vars vars = Vars.noVars;
    private Set possbilyAssigned = Set.empty;
    private ESet possibleExns = ESet.getEmpty();
    private boolean liveFlag = true;
    private ControlContext context = new ControlContext();

    // methods

    public Vars getVars() { return vars; }
    public void setVars(Vars v) { vars = v; }
    public void setVars(Vars v0, Vars v1) { vars = Vars.makePair(v0, v1); }

    //
    public void addPossiblyAssigned(VarDec dec) {
        possbilyAssigned = possbilyAssigned.add(dec);
    }
    public Set popPossiblyAssigned() {
        Set pa = possbilyAssigned;
        possbilyAssigned = Set.empty;
        return pa;
    }
    public void mergePossiblyAssigned(Set pa) {
        possbilyAssigned = possbilyAssigned.union(pa);
    }

    //
    public ESet popExns() {
        ESet e = possibleExns;
        possibleExns = ESet.getEmpty();
        return e;
    }
    public void setExns(ESet e) { possibleExns = e; }
    public ESet getExns() { return possibleExns; }

    public TypeDs getCheckedExns() {
        AST ast = getAST();
        TypeDs ds = ast.makeTypeDs();
        for (Iterator i = possibleExns.iterator(); i.hasNext(); ) {
            NameType ty = (NameType) i.next();
            if (ty.isSubtypeOf(getTypeManager().getRuntimeExceptionType())) continue;
            if (ty.isSubtypeOf(getTypeManager().getErrorType())) continue;
            else ds.add(ty.makeTypeD());
        }
        return ds;
    }

    //
    public void setLive(boolean b) { liveFlag = b; }
    public boolean isLive() { return liveFlag; }


    // ------------------------------
    // Useful vars methods

    public static Vars getAllVars() { return Vars.allVars; }
    public static Vars getNoVars() { return Vars.noVars; }

    // ------------------------------
    // error showing

    java.util.Set errorVars = new java.util.HashSet();
    public void showVarError(ASTObject o, VarDec v, String msg) {
        if (! errorVars.contains(v)) {
            o.showError(msg);
            errorVars.add(v);
        }
    }

    // ------------------------------
    // control context

    public ControlContext popControlContext() {
        ControlContext c = this.context;
        context = new ControlContext();
        return c;
    }

    public void enterContext(Stmt s) { context.enter(s); }
    public void leaveContext() { context.exit(); }

    public boolean isLabelUsed(String name) {
        return context.isLabelUsed(name);
    }

    public void doBreak(String label) {
        Stmt target = context.getBreakTarget(label);
        Vars v = getVars();
        annotateWindsUntil(target, v);
        annotate(breaks, target, v);
        addPendingTransfer(true, target);
    }

    public void doContinue(String label) {
        Stmt target = context.getContinueTarget(label);
        Vars v = getVars();
        annotateWindsUntil(target, v);
        annotate(continues, target, v);
        addPendingTransfer(false, target);
    }

    public void doReturn() {
        Vars v = getVars();
        annotateWindsUntil(null, v);
    }

    private void annotateWindsUntil(Stmt target, Vars v) {
        for (Iterator i = context.getWindsUntil(target); i.hasNext(); ) {
            Object o = i.next();
            annotate(tries, o, v);
        }
    }

    private Map breaks = new HashMap();
    private Map continues = new HashMap();
    private Map tries = new HashMap();
    private void annotate(Map m, Object s, Vars v) {
        Vars tab = (Vars) m.get(s);
        if (tab == null) {
            m.put(s, v);
        } else {
            m.put(s, tab.join(v));
        }
    }

    public Vars getBreakVars(Stmt s) { return getVars(breaks, s); }
    public Vars getContinueVars(Stmt s) { return getVars(continues, s); }
    public Vars getTryVars(Stmt s) { return getVars(tries, s); }
    private Vars getVars(Map m, Stmt s) {
        Vars tab = (Vars) m.get(s);
        if (tab == null) {
            return getAllVars();
        } else {
            return tab;
        }
    }

    public void checkLoopingFinals(Stmt stmt, Set pa, Vars v) {
        // forall x in pa, x must be in v.dua
        // in other words, pa.diff(v.dua) == empty
        Set s = pa.diff(v.getDua());
        while (! s.isEmpty()) {
            VarDec dec = s.first();
            s = s.rest();
            showVarError(stmt, dec, "variable " + dec.getId() +
                         " might be assigned in loop");
        }
    }

    public static Vars buildCatchVars(Vars daHolder, Vars duaHolder) {
        return Vars.makeSets(daHolder.getDa(), duaHolder.getDua());
    }

    // ------------------------------

    // for break and continue control, we need to annotate the next
    // try stmt up that it should 're-break' if it finds that it
    // completes normally.  This is because s: try { break s; }
    // finally { return; } is not broken.  Strangely enough, this is
    // _just_ for control, not for variables.  !!! Note that it is for
    // exceptions also, which we don't check either. So we may want to
    // change TryStmt to make sure that we signal an unreachable catch
    // clause error for try {try { throw new E(); } finally { return; } }
    // catch (E e) { ... }

    private java.util.Set broken = new HashSet();
    private java.util.Set continued = new HashSet();

    public boolean isBroken(Stmt s) { return broken.contains(s); }
    public boolean isContinued(Stmt s) { return continued.contains(s); }

    // SEE
    // http://developer.java.sun.com/developer/bugParade/bugs/4088988.html
    // for a description of why we have all this damn verbiage here
    // and we don't use any of it.  When we want to check this kind of
    // bug, we should comment out the below text.

    public void releasePendingTransfers(TryStmt s) {
        if (getOptions().strict) {
            Redo l = (Redo) pendingTransfers.get(s);
            if (l == null) return;
            l.redo();
        }
    }

    private Map/*<TryStmt, Redo>*/ pendingTransfers = new HashMap();

    private class Redo {
        boolean isBreak; Stmt target;
        Redo(boolean isBreak, Stmt target) {
            this.isBreak = isBreak; this.target = target;
        }
        void redo() {
            Iterator i = context.getWindsUntil(target);
            if (i.hasNext()) {
                pendingTransfers.put(i.next(), this);
            } else {
                (isBreak ? broken : continued).add(target);
            }
        }
    }

    private void addPendingTransfer(boolean isBreak, Stmt target) {
        if (! getOptions().strict) {
            (isBreak ? broken : continued).add(target);
        } else {
            Iterator i = context.getWindsUntil(target);
            if (i.hasNext()) {
                pendingTransfers.put(i.next(), new Redo(isBreak, target));
            } else {
                (isBreak ? broken : continued).add(target);
            }
        }
    }

    // ------------------------------

    public static abstract class Vars {
        public static Vars getEmpty() { return noVars; }
        public abstract Vars getTrue();
        public abstract Vars getFalse();
        public abstract Set getDa();
        public abstract Set getDua();
        public abstract Vars join(Vars other);
        public abstract Vars joinUnion(Vars other);
        public abstract Vars addAssigned(VarDec v);
        public abstract Vars addUnassigned(VarDec v);
        public abstract boolean isDefinitelyAssigned(VarDec dec);
        public abstract boolean isDefinitelyUnassigned(VarDec dec);

        public static Vars allVars = new Sets(Set.full, Set.full);
        public static Vars noVars = new Sets(Set.empty, Set.empty);
        public static Vars makePair(Vars v0, Vars v1) {
            return new Pair(v0, v1);
        }
        public static Vars makeSets(Set v0, Set v1) {
            return new Sets(v0, v1);
        }

        private static final class Pair extends Vars {
            final Vars ifTrue, ifFalse;
            Pair(Vars ifTrue, Vars ifFalse) {
                this.ifTrue = ifTrue; this.ifFalse = ifFalse;
            }
            public Vars getTrue() { return ifTrue; }
            public Vars getFalse() { return ifFalse; }
            public Set getDa() { return normalize().getDa(); }
            public Set getDua() { return normalize().getDua(); }
            public Vars join(Vars other) { return normalize().join(other); }
            public Vars joinUnion(Vars other) { return normalize().joinUnion(other); }
            public Vars addAssigned(VarDec v) { return normalize().addAssigned(v); }
            public Vars addUnassigned(VarDec v) { return normalize().addUnassigned(v); }
            public boolean isDefinitelyAssigned(VarDec dec) {
                return normalize().isDefinitelyAssigned(dec);
            }
            public boolean isDefinitelyUnassigned(VarDec dec) {
                return normalize().isDefinitelyUnassigned(dec);
            }
            private Vars normalForm = null;
            private Vars normalize() {
                if (normalForm == null)
                    normalForm = ifTrue.join(ifFalse);
                return normalForm;
            }
            public String toString() {
                return "<" + ifTrue + ", " + ifFalse + ">";
            }
        }

        private static final class Sets extends Vars {
            final Set da, dua;
            Sets(Set da, Set dua) {  this.da = da; this.dua = dua; }
            public Vars getTrue() { return this; }
            public Vars getFalse() { return this; }
            public Set getDa() { return da; }
            public Set getDua() { return dua; }

            public Vars join(Vars other) {
                return new Sets(da.inter(other.getDa()), dua.inter(other.getDua()));
            }
            public Vars joinUnion(Vars other) {
                return new Sets(da.union(other.getDa()), dua.union(other.getDua()));
            }
            public Vars addAssigned(VarDec v) {
                return new Sets(da.add(v), dua.remove(v));
            }
            public Vars addUnassigned(VarDec v) {
                // guaranteed to be new
                return new Sets(da, dua.add(v));
            }
            public boolean isDefinitelyAssigned(VarDec dec) {
                return da.contains(dec);
            }
            public boolean isDefinitelyUnassigned(VarDec dec) {
                return dua.contains(dec);
            }
            public String toString() {
                return "<" + da + ", " + dua + ">";
            }
        }
    }

    public static abstract class Set {
        private static Zero empty = new Zero();
        private static Neg full = new Neg(empty);
        public static Set getAll() { return full; }
        public static Set getNone() { return empty; }

        public abstract Set neg();
        public abstract Set add(VarDec v);
        public abstract Set remove(VarDec v);
        public abstract boolean contains(VarDec v);
        public abstract Set union(Set s);
        public abstract Set inter(Set s);
        public abstract Set diff(Set s);
        public abstract boolean isEmpty();
        public VarDec first() { throw new RuntimeException("can't take first"); }
        public NonNeg rest() { throw new RuntimeException("can't take first"); }
        public String toStringRest() {
            throw new RuntimeException("can't get rest of set");
        }
        public static abstract class NonNeg extends Set {
            protected abstract Set union1(Set s);
            protected abstract Set inter1(Set s);
            protected abstract Set diff1(Set s);
        }
        private static class Zero extends NonNeg {
            public Set neg() { return full; }
            public Set add(VarDec v) { return new One(v, this); }
            public Set remove(VarDec v) { return this; }
            public boolean contains(VarDec v) { return false; }
            public Set union(Set s) { return s; }
            public Set inter(Set s) { return this; }
            public Set diff(Set s) { return this; }
            protected Set union1(Set s) { return s; }
            protected Set inter1(Set s) { return this; }
            protected Set diff1(Set s) { return this; }
            public boolean isEmpty() { return true; }
            public String toStringRest() { return "}"; }
            public String toString() { return "{}"; }
        }
        private static class Neg extends Set {
            private NonNeg x;
            Neg(NonNeg x) { this.x = x; }
            public Set neg() { return x; }
            public Set add(VarDec v) { return x.remove(v).neg(); }
            public Set remove(VarDec v) { return x.add(v).neg(); }
            public boolean contains(VarDec v) { return ! x.contains(v); }
            public Set union(Set s) {
                if (s instanceof Neg) return x.inter(s.neg()).neg();
                else                  return x.diff(s).neg();
            }
            public Set inter(Set s) {
                if (s instanceof Neg) return x.union(s.neg()).neg();
                else                  return s.diff(x);
            }
            public Set diff(Set s) {
                if (s instanceof Neg) return s.neg().diff(x);
                else                  return x.union(s).neg();
            }
            public boolean isEmpty() { return false; }
            public String toString() { return "!" + x; }
        }
        private static class One extends NonNeg {
            VarDec v;
            NonNeg rest;
            // note, was private, but javac 1.4 didn't like
            One(VarDec v, NonNeg rest) { this.v = v; this.rest = rest; }
            public Set neg() { return new Neg(this); }
            public Set add(VarDec v) { return contains(v) ? this : new One(v, this); }
            public Set remove(VarDec v) {
                return (v == this.v) ? rest : rest.remove(v).add(this.v);
            }
            public boolean contains(VarDec v) {
                return (v == this.v) || rest.contains(v);
            }
            public Set union(Set s) {
                return (s instanceof Neg) ? s.union(this) : this.union1(s);
            }
            public Set inter(Set s) {
                return (s instanceof Neg) ? s.inter(this) : this.inter1(s);
            }
            public Set diff(Set s) {
                return (s instanceof Neg) ? this.inter(s.neg()) : this.diff1(s);
            }
            public boolean isEmpty() { return false; }
            public VarDec first() { return v; }
            public NonNeg rest() { return rest; }

            protected Set union1(Set s) {
                return s.contains(v) ? rest.union1(s) : rest.union1(s).add(v);
            }
            protected Set inter1(Set s) {
                return s.contains(v) ? rest.inter1(s).add(v) : rest.inter1(s);
            }
            protected Set diff1(Set s) {
                return s.contains(v) ? rest.diff1(s) : rest.diff1(s).add(v);
            }

            public String toString() { return "{" + v.getId() + rest.toStringRest(); }
            public String toStringRest() {
                return ", " + v.getId() + rest.toStringRest();
            }
        }
    }

    public static abstract class ESet {
        static ESet empty = new Null();
        public static ESet getEmpty() { return empty; }
        abstract ESet remove(NameType v);
        public abstract boolean containsSuperOrSub(NameType v);
        public abstract boolean containsSuper(NameType v);
        public abstract ESet union(ESet s);
        public abstract ESet diff(ESet s);
        public Iterator iterator() {
            return new Iterator() {
                    private ESet e = ESet.this;
                    public boolean hasNext() { return !(e instanceof Null); }
                    public Object next() {
                        One o = (One) e;
                        e = o.rest;
                        return o.v;
                    }
                    public void remove() {
                        throw new UnsupportedOperationException();
                    }
                };
        }
        public String toStringRest() {
            throw new RuntimeException("can't get rest of set");
        }
        public ESet add(NameType v) {
            return v.isUncheckedThrowable() ? this : this.add1(v);
        }
        abstract ESet add1(NameType v);

        private static class Null extends ESet {
            ESet add1(NameType v) { return new One(v, this); }
            ESet remove(NameType v) { return this; }
            public boolean containsSuperOrSub(NameType v) { return false; }
            public boolean containsSuper(NameType v) { return false; }
            public ESet union(ESet s) { return s; }
            public ESet diff(ESet s) { return this; }
            public boolean isEmpty() { return true; }
            public String toStringRest() { return "}"; }
            public String toString() { return "{}"; }
        }

        private static class One extends ESet {
            NameType v;
            ESet rest;
            // note, was private, but javac 1.4 didn't like it
            One(NameType v, ESet rest) { this.v = v; this.rest = rest; }
            ESet add1(NameType v) { return containsSuper(v) ? this : new One(v, this); }
            ESet remove(NameType v) {
                return (v == this.v) ? rest : rest.remove(v).add(this.v);
            }
            public boolean containsSuperOrSub(NameType v) {
                return v.isSubtypeOf(this.v) || this.v.isSubtypeOf(v) ||
                    rest.containsSuperOrSub(v); }
            public boolean containsSuper(NameType v) {
                return v.isSubtypeOf(this.v) || rest.containsSuper(v);
            }
            public ESet union(ESet s) {
                return s.containsSuper(v) ? rest.union(s) : rest.union(s).add(v);
            }
            public ESet diff(ESet s) {
                return s.containsSuper(v) ? rest.diff(s) : rest.diff(s).add(v);
            }
            public String toString() { return "{" + v.getId() + rest.toStringRest(); }
            public String toStringRest() {
                return ", " + v.getId() + rest.toStringRest();
            }
        }
    }
}

/*
There's a tricky bit in dealing with blank finals.  That tricky bit is
in the looping constructs.

The basic thing for:

   do S0 while E1
   while E0 S1

is this:

   If I _relied on_ some variable V being in dua to evaluate 0, and
   later I find that after the flow of 1 V is not in dua, show an
   error for V.

Thus, we return, conceptually TWO values from statements:

  da:    variables definitely assigned
  dua:   variables definitely unassigned

and save one in the state

  ba:    blank finals _possibly_ assigned

where ba: is _only_ checked in the checkers for loops, but it has to
be maintained for everyone.

A better job of this could be done by annotating loop statements with
bit vectors for every blank final closed over.

*/
