/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import java.io.File;

import java.util.*;
import java.io.OutputStream;
import java.lang.reflect.*;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.World;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CompilerObject;

public class JavaCWrapper extends CompilerObject {

    public JavaCWrapper(JavaCompiler compiler) {
        super(compiler);
    }

    private String addClassPath(File pathdir) {
        String classpath = getOptions().classpath; //System.getProperty("java.class.path");
        if (classpath != null) return pathdir.getPath()+File.pathSeparator+classpath;
        classpath = System.getProperty("java.class.path");
        if (classpath != null) return pathdir.getPath()+File.pathSeparator+classpath;
        return pathdir.getPath();
    }

    public boolean compile(File inputdir, File outputdir, File[] files) {
        List args = new ArrayList(files.length+6);

        args.add("-d");
        args.add(inputdir.getPath());
        args.add("-classpath");
        args.add(addClassPath(outputdir));
        if (getOptions().deprecation) {
            args.add("-deprecation");
        }
        if (getOptions().O) {
            args.add("-O");
        }
        if (getOptions().g) {
            args.add("-g");
        }
        if (getOptions().source.equals("1.4")) {
            args.add("-source");
            args.add("1.4");
        }
        if (getOptions().target != null) {
            args.add("-target");
            args.add(getOptions().target);
        }

        if (getOptions().bootclasspath != null) {
            args.add("-bootclasspath");
            args.add(getOptions().bootclasspath);
        }

        if (getOptions().extdirs != null) {
            args.add("-extdirs");
            args.add(getOptions().extdirs);
        }

        //add javafiles to args
        for(int i=0; i<files.length; i++) {
            args.add(files[i].getPath());
        }

        /*
        StringBuffer cmd = new StringBuffer();
        cmd.append("javac");
        for(int i=0; i<args.size(); i++) {
            cmd.append(" ");
            cmd.append(args.get(i));
        }
        */


        String[] argsArray = (String[])args.toArray(new String[0]);
        //System.out.println("javac "+args);

        Class compilerClass;

        try {
            compilerClass = Class.forName("com.sun.tools.javac.Main");
            return doModernCompile(compilerClass, argsArray);
        } catch (ClassNotFoundException ce1) {
            try {
                compilerClass = Class.forName("sun.tools.javac.Main");
                return doClassicCompile(compilerClass, argsArray);
            } catch (ClassNotFoundException ce2) {
                showNoJavacError();
                return false;
            }
        }
    }


    void showNoJavacError() {
        System.err.println("The -usejavac flag requires that com.sun.tools.javac.Main");
        System.err.println("is on your classpath.  Generally, this will involve editing");
        System.err.println("your launch script (ajc.bat or ajc) to include tools.jar");
        System.err.println("on the classpath that is passed to the jvm which runs ajc.");
        System.err.println("See README-TOOLS.html for suggestions on how to fix this.");
        throw new ExitRequestException(-1);
    }



    boolean doModernCompile(Class compiler, String[] args) {
        getCompiler().showMessage("  modern compile");
        // use reflection to work with both 1.3 and 1.4
        try {
            Method m = compiler.getMethod("compile", new Class[] {String[].class});
            Object inst = compiler.newInstance();
            Integer i = (Integer)m.invoke(inst, new Object[] {args});
            return i.intValue() == 0;
        } catch (Exception e) {
            System.out.println(e);
            return false;
        }
    }

    boolean doClassicCompile(Class compiler, String[] args) {
        getCompiler().showMessage("  classic compile");
        // use reflection so that we don't depend on the compiler sources to build
        //sun.tools.javac.Main main = new sun.tools.javac.Main(System.err, "ajc");
        //return main.compile(args);
        try {
            Constructor c = compiler.getConstructor(new Class[] {OutputStream.class, String.class});
            Object inst = c.newInstance(new Object[] {System.err, "ajc"});
            Method m = compiler.getMethod("compile", new Class[] {String[].class});
            Boolean b = (Boolean)m.invoke(inst, new Object[] {args});
            return b.booleanValue();
        } catch (Exception e) {
            System.out.println(e);
            return false;
        }
    }
}
