/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.World;
import org.aspectj.compiler.base.ast.AST;
import org.aspectj.compiler.base.ast.CompilationUnit;
import org.aspectj.compiler.base.ast.ASTObject;
import org.aspectj.compiler.base.ast.TypeDec;
import org.aspectj.compiler.base.parser.JavaParser;


import java.util.*;
import java.util.Stack;
import java.io.*;
import java.text.NumberFormat;
import java.lang.reflect.Constructor;

public class JavaCompiler {
    // Compiler Options
    private Options                 options = new Options();
    // The TypeManager linked to this compiler
    private TypeManager             typeManager;
    // The World linked to this compiler
    protected World                   world;
    // error handler
    protected ErrorHandler errorHandler;
    // interface scanning class
    protected Class                 scanner;
    // world-wide passes
    protected List/*WorldPass*/     passes;
    // List of idle parsers
    private Stack                   parsers = new Stack();
    // files to be compiled by javac
    protected List/*File*/          filesToCompile;


    /** Method to support CompilerObject delegation for Options */
    public final Options getOptions() { return options; }
    /** Method to support CompilerObject delegation for World */
    public final World getWorld() { return world; }
    /** Method to support CompilerObject delegation for TypeManager */
    public final TypeManager getTypeManager() { return typeManager; }


    public boolean willGenerateSourceCode() {
        return options.usejavac || options.preprocess;
    }

    public String getVersion() {
        return Options.version;
    }

    public long getBuildTime() {
        return Options.build;
    }

    public List getFilesToCompile() {
        if (filesToCompile == null) return new LinkedList();
        return filesToCompile;
    }

    public void addFileToCompile(File f) {
        if (filesToCompile == null) filesToCompile = new LinkedList();
        filesToCompile.add(f);
    }

    public CompilationUnit newCompilationUnit(File _file) {
        return new CompilationUnit(this,_file);
    }

    public JavaCompiler(ErrorHandler errHandler) {
        // if err_handler isn't null use it, otherwise construct a default one
        this.errorHandler = (errHandler != null) ?  errHandler : makeErrorHandler();
        this.errorHandler.setCompiler(this);
    }

    /**
     * Returns a default <code>ErrorHandler</code> writing to <code>System.err</code>.
     * @return A default <code>ErrorHandler</code> writing to <code>System.err</code>.
     */
    protected ErrorHandler makeErrorHandler() {
            return new ErrorHandler(new OutputStreamWriter(System.err));
    }


    public JavaParser makeJavaParser() {
        return new JavaParser(this);
    }

    public boolean isIncremental() {
        return options.incremental;
    }

    public boolean getOptionDumpStack() {
        return options.dumpstack;
    }

    public boolean getOptionIgnoreErrors() {
        return options.ignoreErrors;
    }

    public boolean getOptionVerbose() {
        return options.verbose;
    }

    public void clearState() {
        world.cleanup();
        filesToCompile = null;
        errorHandler.cleanupErrorWarningCounters();
    }

    public void compile(List filenames) {
        addPasses();

       if (getOptions().threads != 0) {
            System.out.println("warning: multi-threaded compilation not supported in this release");
            getOptions().threads = 0;
        }

        try {
            internalCompile(filenames);
        } catch (CompilerErrors compilerErrors) {
            throw compilerErrors;
        } catch (ExitRequestException exitRequest) {
            throw exitRequest;
        } catch (InternalCompilerError internalError) {
            // only throw an internal error if there's aren't any regular errors
            if (!getOptions().torture) errorHandler.exitOnErrors();
            throw internalError;
        } catch (Throwable throwable) {
            // only throw an internal error if there's aren't any regular errors
            //!!! think about how this might hide errors...
            if (!getOptions().torture) errorHandler.exitOnErrors();
            throw new InternalCompilerError(this, throwable, getCurrentNode());
        } finally {
            //XXXclearState();
        }
    }

    /* This section containce compilation pases implementations */
    protected void createWorld(List filenames) {
        typeManager = new TypeManager(this);
        world = new World(this);
        if (!typeManager.checkLoadable("java.lang", "Object")) {
            warnNoObject();
        }
        if (!typeManager.checkLoadable("org.aspectj.lang", "JoinPoint")) {
            warnNoRuntime(false);
        }
        if (!typeManager.checkLoadable("org.aspectj.runtime.internal", "AroundClosure")) {
            warnNoRuntime(true);
        }
        // add filenames to compile, the World will check for duplications
        for(Iterator i=filenames.iterator(); i.hasNext();) {
            Object o = i.next();
            String filename;
            File file;
            if (o instanceof String) {
                filename = (String)o;
                file = new File(filename);
            } else {
                file = (File)o;
                filename = file.getPath();
            }
            world.addFile(filename);
        }
    }

    private static final String noRuntimeWarning =
      "Can't find org.aspectj.lang.JoinPoint on your classpath anywhere.\n" +
      "You need to include aspectjrt.jar on your classpath when compiling or\n" +
      "running applications with ajc.  See README-TOOLS.html in the top directory of\n" +
      "the distribution for more details on how to configure this correctly.";

    private static final String oldRuntimeWarning =
      "You appear to have an out-of-date version of aspectjrt.jar on your classpath.\n" +
      "You need to include the latest verion of aspectjrt.jar on your classpath when\n" +
      "compiling or running applications with ajc.  See README-TOOLS.html in the top\n" +
      "directory of the distribution for more details on how to configure this\n" +
      "correctly.";

    /**
     * To be overridden by compilers that don't use console output.
     */
    protected void warnNoRuntime(boolean foundOld) {
        System.err.println(foundOld ? oldRuntimeWarning : noRuntimeWarning);
        System.err.println();
        System.err.println("I looked in: ");
        for (Iterator i = getTypeManager().getClassPathStrings().iterator(); i.hasNext(); ) {
            System.err.print("    ");
            System.err.println(i.next());
        }
        
        //System.err.println("sun.boot.class.path: " + System.getProperty("sun.boot.class.path"));
        
        throw new ExitRequestException(-1);
    }


    private static final String noObjectWarning =
        "Serious configuration problem: can't find java.lang.Object";
    private static final String modifiedBootpathWarning =
        "Check your custom bootclasspath for validity: ";
    private static final String jviewWarning =
        "You must run 'claspack -auto' to use ajc with MS's jview VM";


    private void warnNoObject() {
        System.err.println(noObjectWarning);
        if (options.bootclasspath != null) {
            System.err.println(modifiedBootpathWarning+options.bootclasspath);
        } else {
            String vendor = System.getProperty("java.vm.vendor");
            if (vendor != null && vendor.startsWith("Microsoft")) {
                System.err.println(jviewWarning);
            }
        }

        throw new ExitRequestException(-1);
    }

    protected void addPasses() {}


    public synchronized JavaParser allocateParser() {
        JavaParser parser;
        if (parsers.empty()) {
            parser = makeJavaParser();
        } else {
            parser = (JavaParser)parsers.pop();
        }
        return parser;
    }

    public synchronized void freeParser(JavaParser parser) {
        parser.cleanup(false);
        parsers.push(parser);
    }




    /* This section of code handles compiler state */
    private Hashtable timingInformation = new Hashtable();
    private ThreadLocal sectionLabel = new ThreadLocal();
    private ThreadLocal localNodes = new ThreadLocal();
    private long startTime;

    private Stack getNodes() {
        Object nodes = localNodes.get();
        if (nodes == null) {
            nodes = new Stack();
            localNodes.set(nodes);
        }
        return (Stack)nodes;
    }

    public String getCurrentSection() {
        return (String)sectionLabel.get();
    }

    public void setCurrentSection(String label) {
        sectionLabel.set(label);
    }

    private String formatTime(long time) {
        NumberFormat format = NumberFormat.getNumberInstance();
        format.setMinimumFractionDigits(2);
        format.setMaximumFractionDigits(2);
        String value = format.format(time/1000.0);
        while (value.length() < 3+2) value = " "+value;
        return value;
    }

    protected double totalWorkEstimate = 0.0;
    protected double percentComplete = 0.0;
    public double getPercentComplete() {
        return percentComplete;
    }

    protected void updateCompileState(String section, String subSection, double pct) {
        //System.out.println((int)(pct*100.0)+"%  " + section+":"+subSection);
    }


    public void completedFile(CompilerPass pass, CompilationUnit cu) {
        exitIfRequested();

        double pct = (pass.getWorkEstimate()/totalWorkEstimate)*
                        (1.0/getWorld().getCompilationUnits().size());
        percentComplete += pct;
        updateCompileState(pass.getDisplayName(), cu.getSourceCanonicalPath(),
                                percentComplete);
    }

    protected void initializeWorld(List filenames) {
        // create/initialize the world
        createWorld(filenames);
        errorHandler.exitOnErrors();

        Set doneFiles = new HashSet();
        for(Iterator i = world.getFiles().iterator(); i.hasNext(); ) {
            File file = new File((String)i.next());
            try {
                if (doneFiles.contains(file.getCanonicalPath())) {
                    showMessage(this, "ignoring duplicate file: "+file.getPath());
                    continue;
                } else {
                    doneFiles.add(file.getCanonicalPath());
                }
            } catch (IOException e) {
                this.showError(null,e.toString());
                continue;
            }
            world.addCompilationUnit(newCompilationUnit(file));
        }
    }


    private volatile boolean exitRequested = false;
    private void exitIfRequested() {
        if (exitRequested) {
            //!!! this value is depended on by ajde to know that this is user
            //!!! requested.  this should be done in a more principled way.
            throw new ExitRequestException(0);
        }
    }

    /**
     * This method can be called from a thread other than the thread the
     * compiler is running in to request that the compiler exit ASAP.
     * The compiler will then terminate when it can.
     */
    public void requestCompileExit() {
        exitRequested = true;
    }


    protected void internalCompile(List filenames) {
        exitRequested = false;
        initializeWorld(filenames);

        for (Iterator i = passes.iterator(); i.hasNext(); ) {
            CompilerPass pass = (CompilerPass)i.next();
            pass.transformWorld();
            exitIfRequested();
        }
        errorHandler.exitOnErrors();
        finish();
    }


    public void beginSection(String label) {
        beginSection(label, true);
    }

    public void beginSection(String label, boolean show) {
        if (!getNodes().empty()) {
            // internal error, non-empty stack after section
        }

        if (show) showMessage(label);
        if (options.timings && options.threads == 0) {
            long thisTime = System.currentTimeMillis();
            if (sectionLabel.get() != null) {
                Long t = (Long)timingInformation.get(sectionLabel.get());
                if (t == null) {
                    t = new Long(thisTime-startTime);
                } else {
                    t = new Long(t.longValue()+(thisTime-startTime));
                }
                timingInformation.put(sectionLabel.get(),t);
            }
            startTime = thisTime;
        }
        setCurrentSection(label);
    }

    public void finish() {
        beginSection("compilation complete");
        if (options.timings && options.threads == 0) {
            errorHandler.showMessage("Timing information:");
            Map.Entry[] a = (Map.Entry[])
                timingInformation.entrySet().toArray(new Map.Entry[0]);
            Arrays.sort(a, new Comparator() {
                    public int compare(Object x, Object y) {
                        return
                            ((Long) ((Map.Entry) x).getValue())
                            .compareTo(((Map.Entry) y).getValue());
                    }
                });
            long totalTime = 0;
            for (int i = 0, len = a.length; i < len; i++) {
                Map.Entry n = a[i];
                String sectionLabel = (String)n.getKey();
                Long tm = (Long)n.getValue();
                long thisTime = tm.longValue();
                String msg = "  "+formatTime(thisTime)+" seconds in "+sectionLabel;
                totalTime += thisTime;
                errorHandler.showMessage(msg);
            }
            errorHandler.showMessage("---------------------------------------------");
            String msg = "  "+formatTime(totalTime)+" total seconds";
            errorHandler.showMessage(msg);
        }
    }

    public void enterNode(ASTObject where) {
        getNodes().push(where);
    }

    public void exitNode(ASTObject where) {
        if (getNodes().pop() != where) {
            // internal error, mismatched push/pop;
        }
    }

    public ASTObject getCurrentNode() {
        return getNodes().empty() ? null : (ASTObject)getNodes().peek();
    }


    /* This code delegates error handling to ErrorHandler */
    public void internalError(Throwable uncaughtException, ASTObject where) {
        if (!errorHandler.willExitWithErrors()) {
            errorHandler.internalError(uncaughtException, where);
        }
    }

    public void internalError(String message) {
        internalError(null, message);
    }

    public void internalError(ASTObject where, String message) {
        if (!errorHandler.willExitWithErrors()) {
            errorHandler.showError(where, "INTERNAL: " + message);
        }
    }

    public void showError(String message) {
        this.showError(null, message);
    }

    public void showError(ASTObject where, String message) {
        errorHandler.showError(where, message);
    }

    public void showError(File source, int line, int column, String message) {
        errorHandler.showError(source, line, column, message);
    }

    public void showWarning(ASTObject where, String message) {
        errorHandler.showWarning(where, message);
    }
    
    public void showWarning(ASTObject where, ErrorHandler.Message message) {
    	errorHandler.showWarning(where, message);
    }

    public void showMessage(Object where, String message) {
        if (where != null && where instanceof ASTObject) {
            errorHandler.showMessage((ASTObject)where, message);
            return;
        }

        if (options.verbose) {
            errorHandler.showMessage("  "+message);
        }
    }

    public void showMessage(String message) {
        if (options.verbose) {
            errorHandler.showMessage(message);
        }
    }

    public void warnVersion(String oldVersion, ASTObject object, String message) {
        if (options.porting) {
            showWarning(object, "Deprecated since "+oldVersion+": "+message);
        } else {
            errorVersion(oldVersion, object, message);
        }
    }

    public void errorVersion(String oldVersion, ASTObject object, String message) {
        showError(object, "Removed in "+oldVersion+": "+message);
    }
}
