/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;
import java.util.*;

public class LocalClassPass extends AbstractCompilerPass {

    public LocalClassPass(JavaCompiler compiler) {
        super(compiler);
    }

    public String getDisplayName() {
        return "lifting local classes ";
    }

    public void transform(TypeDec typeDec) {
        //System.out.println("lifting: " + typeDec.unparse());
        
        // pass one:  analyze, set static, change names
        AnalysisWalker analysisWalker = new AnalysisWalker(getCompiler());
        analysisWalker.process(typeDec);

		//System.err.println("needsVars: " + analysisWalker.needsVarTable);
		//System.err.println("needsFields: " + analysisWalker.needsFieldTable);

        // pass two:  anonymous constructors, lift, initializersToConstructors
        new LiftWalker(getCompiler()).process(typeDec);

        // pass three: thread locals through constructors to where they want to go
        new ThreadingWalker(getCompiler(), analysisWalker).process(typeDec);
        
        //System.err.println("lifted: " + typeDec.unparse());
    }

    // intro on TypeDec, VarDec, NewInstanceExpr, ConstructorCallExpr,
    // ConstructorDec, MethodDec, ConstructorBody
    public static class ThreadingWalker extends Walker {
        HashMap/*<TypeDec, Set<VarDec>>*/ needsVarTable;
        HashMap/*<TypeDec, Set<VarDec>>*/ needsFieldTable;

        ThreadingWalker(JavaCompiler compiler, AnalysisWalker analysisWalker) {
            super(compiler);
            this.needsVarTable   = analysisWalker.needsVarTable;
            this.needsFieldTable = analysisWalker.needsFieldTable;
        }
        public ASTObject process(ASTObject object) {
            object.preThreading(this);
            object.walk(this);
            return object.postThreading(this);
        }

        Stack/*<TypeDec>*/                   typeContext = new Stack();
        Stack/*<HashMap<VarDec, FieldDec>>*/ fieldContext = new Stack();
        Stack/*<HashMap<VarDec, FieldDec + FormalDec>>*/ env = new Stack();

        {
            fieldContext.push(new HashMap());
        }

        public void pushConstructorEnv(Formals formals) {
            final AST ast = getAST();
            Map envMap = new HashMap();
            Set s = (Set) needsVarTable.get(typeContext.peek());
            if (s != null) {
                for (Iterator i = s.iterator(); i.hasNext(); ) {
                    VarDec varDec = (VarDec)i.next();
                    FormalDec formalDec =
                        ast.makeFormal(varDec.getTypeD().getType(),
                                       "val$" + varDec.getId());
                    formals.add(0, formalDec);
                    envMap.put(varDec, formalDec);
                }
            }
            env.push(envMap);
        }
        public void pushNonConstructorEnv() {
            env.push(fieldContext.peek());
        }
        public void popEnv() {
            env.pop();
        }

        public void pushTypeDec(TypeDec typeDec) {
            respectOurParentsWishes(typeDec);
            typeContext.push(typeDec);
            addFields(typeDec);
        }
        public void popTypeDec() {
            typeContext.pop();
            fieldContext.pop();
        }

        private void addFields(TypeDec typeDec) {
            final AST ast = getAST();
            Map envMap = new HashMap();
            Set s = (Set) needsFieldTable.get(typeContext.peek());
            if (s != null) {
                for (Iterator i = s.iterator(); i.hasNext(); ) {
                    VarDec varDec = (VarDec)i.next();
                    FieldDec fieldDec =
                        ast.makeField(ast.makeModifiers(Modifiers.FINAL),
                                      varDec.getTypeD().getType(),
                                      "val$" + varDec.getId());
                    typeDec.getBody().add(fieldDec);
                    envMap.put(varDec, fieldDec);
                }
            }
            fieldContext.push(envMap);
        }
		private void respectOurParentsWishes(TypeDec typeDec) {
			Set weNeed = (Set) needsVarTable.get(typeDec);
			Set parentNeeds = (Set) needsVarTable.get(typeDec.getSuperClassType().getTypeDec());
			Set enclosingHas;
			if (typeDec.isInner()) {
				enclosingHas = (Set) needsVarTable.get(typeDec.getEnclosingInstanceTypeDec());
			} else {
				enclosingHas = null;
			}

			//System.err.println("needs of " + typeDec);
			//System.err.println("we " + weNeed);
			//System.err.println("rent " + parentNeeds);
			//System.err.println("have " + enclosingHas);
			


			if (weNeed == null) {
				if (parentNeeds != null) {
					weNeed = new HashSet(parentNeeds);
					if (enclosingHas != null) {
						weNeed.removeAll(enclosingHas);
					}
					needsVarTable.put(typeDec, weNeed);
				}
			} else {
				if (parentNeeds != null) {
				    weNeed.addAll(parentNeeds);
				}
				if (enclosingHas != null) {
					weNeed.removeAll(enclosingHas);
				}
			}
			//System.err.println("so we need " + needsVarTable.get(typeDec));

		}

        public Object /*FieldDec + FormalDec*/ lookup(VarDec varDec) {
            for (int i = env.size() - 1; i >= 0; i--) {
                HashMap map = (HashMap) env.get(i);
                Object o = map.get(varDec);
                if (o != null) return o;
            }
            return null;
        }

        public void addArgs(Exprs args, TypeDec typeDec) {
            final AST ast = getAST();
            Set s = (Set) needsVarTable.get(typeDec);
            if (s == null) return;
            for (Iterator i = s.iterator(); i.hasNext(); ) {
                args.add(0, ast.makeVar((VarDec)i.next()));
            }
        }

        public void addFieldSets(Stmts stmts) {
            final AST ast = getAST();
            for (Iterator i = ((Map)fieldContext.peek()).entrySet().iterator();
                 i.hasNext(); ) {
                Map.Entry entry = (Map.Entry) i.next();
                VarDec varDec = (VarDec) entry.getKey();
                FieldDec fieldDec = (FieldDec) entry.getValue();
                stmts.add(0, ast.makeStmt(ast.makeSet(ast.makeDynamicGet(fieldDec),
                                                      ast.makeVar(varDec))));
            }
        }
    }

    // intro to NewInstanceExpr, TypeDec
    public static class LiftWalker extends Walker {
        LiftWalker(JavaCompiler compiler) {
            super(compiler);
        }
        public ASTObject process(ASTObject object) {
            object.preLift(this); // NewInstanceExpr
            object.walk(this);
            return object.postLift(this); // TypeDec
        }
        private Stack/*<Decs>*/ decsStack = new Stack();

        private Decs decs = getAST().makeDecs();

        public void pushPendingDecs() {
            decsStack.push(getAST().makeDecs());
        }
        public void addToPendingDecs(TypeDec dec) {
            ((Decs) decsStack.peek()).add(dec);
        }
        public void popPendingDecsInto(TypeDec dec) {
            dec.getBody().addAll((Decs) decsStack.pop()); // XXX may need to be reversed            
        }
    }

    // intro to VarExpr, TypeDec, CodeDec
    public static class AnalysisWalker extends Walker {
        AnalysisWalker(JavaCompiler compiler) {
            super(compiler);
        }
        public ASTObject process(ASTObject object) {
            final int saved = this.context;
            object.walkAnalysis(this);
            this.context = saved;
            return object;
        }
        private static final int NOWHERE = 0;
        private static final int DYNAMIC = 1;     // dynamic method
        private static final int SEMI_STATIC = 2; // constructor/dynamic init
        private static final int NIGH_STATIC = 3; // constructorCall
        private static final int STATIC = 4;      // static method/init

        /** context can be any of the 5 constants above */
        private int context = NOWHERE;
        public void inType()            { context = NOWHERE;     }
        public void inConstructor()     { context = SEMI_STATIC; }
        public void inConstructorCall() { context = NIGH_STATIC; }
        public void inCode(boolean isStatic) {
            context = isStatic ? STATIC : DYNAMIC;
        }

        HashMap/*<TypeDec, Set<VarDec>>*/ needsVarTable   = new HashMap();
        HashMap/*<TypeDec, Set<VarDec>>*/ needsFieldTable = new HashMap();

        /** XXX should be a stack of triples */
        private Stack/* <CodeDec, Integer, TypeDec>, ... */ contextStack = new Stack();
        public void enterTypeDec(TypeDec typeDec) {
            if (context == NIGH_STATIC || context == STATIC) {
                typeDec.getModifiers().setStatic(true);
            }
            contextStack.push(new Integer(context));
            contextStack.push(typeDec);
        }
        public void leaveTypeDec() {
            contextStack.pop();
            contextStack.pop();
        }
        public void enterCodeDec(CodeDec codeDec) {
            contextStack.push(codeDec);
        }
        public void leaveCodeDec() {
            contextStack.pop();
        }

        public void processVarExpr(VarExpr varExpr) {
            VarDec varDec = varExpr.getVarDec();
            CodeDec referenceContextDec = varDec.getEnclosingCodeDec();
            int referenceContext = context;

            // if the VarDec is truly local, then no more work is needed
            if (contextStack.peek() == referenceContextDec) return;

            if (!varDec.isFinal()) {
                varExpr.showError("local variable " + varDec.getId() +
                  " is referenced from inside inner class; " +
                  " needs to be declared final");
                return;
            }
            
            for (int i = contextStack.size() - 2; i >= 0; i -= 3) {
                if (referenceContext == STATIC) {
                    varExpr.showError("Not accessible from static context");
                    break;
                }
                TypeDec typeDec = (TypeDec) contextStack.get(i);
                int typeDecContext = ((Integer) contextStack.get(i - 1)).intValue();
                CodeDec maybeDeclarationContextDec = (CodeDec) contextStack.get(i - 2);

                if (maybeDeclarationContextDec == referenceContextDec) {
                	//System.err.println("HERE");
                	//System.err.println("adding " + varDec + " to " + typeDec);
                	//System.err.println("dynamic? " + (referenceContext == DYNAMIC));
                    // found it
                    addToTable(needsVarTable, typeDec, varDec);
                    //if (referenceContext == DYNAMIC) {
                        addToTable(needsFieldTable, typeDec, varDec);
                    //}
                    break;
                } else if (typeDecContext == NIGH_STATIC) {
                    addToTable(needsVarTable, typeDec, varDec);
                    if (referenceContext == DYNAMIC) {
                        addToTable(needsFieldTable, typeDec, varDec);
                    }
                    referenceContext = NIGH_STATIC;
                } else {
                    referenceContext = DYNAMIC;
                }
            }
        }

        private static void addToTable(HashMap/*<Object, Set>*/ map,
                                       Object key, Object elem) {
            //System.err.println(key + " " + elem);
            Object o = map.get(key);
            if (o == null) {
                Set s = new HashSet();
                s.add(elem);
                map.put(key, s);
            } else {
                ((Set) o).add(elem);
            }
        }

        private int localCounter = 0;
        //XXX JJH hacked these names for the sake of source generation
        //XXX JJH needs to fix them to be nice to bcg again
        public String makeNewId(TypeDec typeDec) {
            return typeDec.isAnonymous() ?
                "_" + ++localCounter + "" :
                "_" + ++localCounter + "_" + typeDec.getId();
        }
    }
}

/*
First, a pass that annotates each inner typedec with the local
variables it needs.  No replacements are done.  Purely an analysis
pass, and only cares about varExprs.

On the way down the tree, build up an environment associating

 * EnclosingCodeDec 
 * LocalTypeDec
 * 'context'

where context can be

 * DYNAMIC     -- in non-static method:  Will get closed over variables from 
                  fields of some enclosing type.

 * SEMI_STATIC -- in constructor or non-static initializer.  Will get
                  some closed over variables from constructor args, if they are
                  being passed in directly for some reason, and
                  others from the fields of some enclosing type.

 * NIGH_STATIC -- in a constructor call expr.  Will get all closed-over
                  variables from arguments to the constructor.

 * STATIC      -- in static method or initializer.  

Then, for each varExpr v we find, in some context, wander down the
environment annotating each typeDec that needs to know about the varExpr. 

A rule of thumb is that if the varExpr was found in dynamic context,
then some field will be built for it, either in the current class or
an enclosing class.  

If the varExpr was found in static context, it's not allowed. 

If the varExpr was found in semi or nigh static context, the
_enclosing_ class will not have a field added for the varExpr, but
some other enclosing class might.

End result is, for every typedec, both a set of 'needed' vars (NOT
INCLUDING those needed by the superclass), and a subset of those
'needed' vars that need fields.

------------------------------

Second, a pass that lifts typeDecs and integrates initializers.

------------------------------


Third, a pass that does everything else, including pushing needed
vars into subclasses, and making fields, and resetting varExprs, and
adding arguments to newInstance and ConstructorCall exprs.

typeDec T:

  look at superclass.  Get superclass's set of { v ... }.  We are
  guaranteed that superclass has already been processed.

  for each { v : needs field }, create field.
  Let {v ... } be union (our {v ... }, superclass' {v ...}):

  for each constructor, process:
     add {val$v ...} to constructor args
     make environment: { v : val$v }
     process body under said environment.
     add set expr for each { v : needs field }

  for each method, process:
     make environment: { v : T.this.val$v }
     process body

constructor call expr (both this and super), new instance expr:

  let T be target type:
  let { v ... } be T's needs
  add v ... to args
  process self... this will convert the v's, appropriately.

varExpr v:
  replace with env.apply.

*/
