/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.util.*;

import java.util.*;

public class NameHygienePass extends ScopeWalker implements CompilationUnitPass {     
    public NameHygienePass(JavaCompiler compiler) {
        super(compiler);
    }
    
    public double getWorkEstimate() { return 1.0; }
    
    public void transformWorld() {
        if (getDisplayName() != null) getCompiler().beginSection(getDisplayName());
        for (Iterator i = getWorld().getCompilationUnits().iterator(); i.hasNext(); ) {
            CompilationUnit cu = (CompilationUnit)i.next();
            transform(cu);
            getCompiler().completedFile(this, cu);
        }
    }
    
    public void enterCU(CompilationUnit cu) {
    	// don't print error messages for ambiguous imports,
    	// we will fix them later
        scope = new CUScope(getCompiler(), null, cu) {
            protected Type handleAmbiguousImport(String id, Type type1, Type type2, ASTObject showWhere) {
                return this.getTypeManager().TYPE_NOT_FOUND;
            }
        };
        methods = new HashMap();
    }
    
    private boolean varDecCollision(VarDec varDec) {
        Expr expr = scope.bindUnqualifiedName(varDec.getId(), varDec);
        if (expr == null) return false;
        if (! (expr instanceof VarExpr) ) return false;
        return ((VarExpr)expr).getVarDec() != varDec;
//        if (foundVarDec.isField()) {
//            return varDec.isField() &&
//              (foundVarDec.getBytecodeTypeDec() == varDec.getBytecodeTypeDec());
//        }
//        return true;
    }
    
    // used by mangler to validate name binding better
    static int globalId = 0;
    
    public void addVarDec(VarDec varDec) {
        int index = 0;
        String id = varDec.getId();
        if (getCompiler().getOptions().torture) {
            varDec.setId(id + "_" + globalId++);
        } else {
            while (varDecCollision(varDec)) {
                varDec.setId(id + index++);
            }
        }
        super.addVarDec(varDec);
    }
    
    public void addTypeDec(TypeDec typeDec) {
        //System.out.println("adding typeDec: " + typeDec);
        int index = 0;
        String id = typeDec.getId();
        
        // make sure that the name of the typeDec is legal (generated inners are often not)
        if (!Character.isJavaIdentifierStart(id.charAt(0))) {
            id = "_" + id;
            typeDec.setId(id);
        }
        
        if (typeDec.isLocallyDefined()) {
            if (getCompiler().getOptions().torture) {
                typeDec.setId(id + "_" + globalId++);
            } else {
                while (scope.findType(typeDec.getId(), typeDec) != null) {
                    //XXX check for a true collision
                    typeDec.setId(id + index++);
                }
            }
        }
        super.addTypeDec(typeDec);
    }
    
    Map methods = new HashMap();
    
    public void addMethodDec(MethodDec methodDec) {
        if (!methodDec.isPrivate() || methodDec.fromSource()) return;
        
        int index = 0;
        String id = methodDec.getId();
        while (scope.findMethodLookupType(methodDec.getId(), methodDec) != null ||
                methods.containsKey(methodDec.getId())) {
            methodDec.setId(id + index++);
        }
        methods.put(methodDec.getId(), methodDec);
    }
    
    public boolean canUseUnqualifiedName(Type type, ASTObject fromWhere) {
        String id = type.getId();
        
        //fromWhere.showWarning("can use unqualified name? " + id);
        //System.out.println("can use unqualified name? " + scope);
        fromWhere = fromWhere.getBytecodeTypeDec().getBody();
        
        
        if (scope.bindUnqualifiedName(id, fromWhere) != null) return false;

        Type foundType = scope.findType(id, fromWhere);
        if (foundType != type) return false;
        
        //XXX this is a hack to deal with some inner naming issues
        Scope searchScope = scope;
        
        // getStackParent() and getParent() should always return the same thing
        // because useJavaScopes() returns true
        while (searchScope.getStackParent() != null) {
            searchScope = searchScope.getStackParent();
            Type outerFoundType = searchScope.findType(id, fromWhere);
            if (outerFoundType == null) return true;
            if (outerFoundType != type) return false;
        }
        return true;
    }
    
    public boolean mustUseNullTrick(Type type, ASTObject fromWhere) {
        String id = type.getPackageName();
        if (id == null) {
            id = type.getId();
        } else {
            int dot = id.indexOf('.');
            if (dot != -1) id = id.substring(0, dot);
        }
        return scope.bindUnqualifiedName(id, fromWhere) != null;
    }

    public void transform(CompilationUnit cu) {
        process(cu);
    }
    
    public boolean useJavaScopes() { return true; }
    
    public String getDisplayName() {
        return "name hygiene";
    }
}
