/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CompilerObject;
import org.aspectj.compiler.base.TypeManager;
import org.aspectj.compiler.crosscuts.ast.AnonymousMethodExpr;

import org.aspectj.util.JavaStrings;

import java.lang.reflect.*;
import java.util.*;

public class AST extends CompilerObject {
    private SourceLocation sourceLocation;

    public AST(SourceLocation location) {
        super(location.getCompiler());
        this.sourceLocation = location;
    }

    public SourceLocation getSourceLocation() {
        return sourceLocation;
    }

    private static String makeLegalIdentifier(String s) {
        if (s.length() == 0)
            return "_";

        char[] chars = s.toCharArray();
        if (!Character.isJavaIdentifierStart(chars[0])) {
            chars[0] = '_';
        }

        for (int i = 1; i < chars.length; i++) {
            if (!Character.isJavaIdentifierPart(chars[i]))
                chars[i] = '_';
        }
        return new String(chars);
    }

    public static final String PREFIX = "aspectj$";

    public String makeGeneratedName(String s) {
        s = makeLegalIdentifier(s);
        if (getOptions().XaddSafePrefix) {
            if (s.startsWith(PREFIX))
                return s;
            else
                return PREFIX + s;
        } else {
            return s;
        }
    }

    public StringLiteralExpr makeString(String s) {
        return new StringLiteralExpr(getSourceLocation(), s);
    }

    public CallExpr makeStaticCall(MethodDec method, Exprs args) {
        return makeStaticCall(method.getMethod(), args);
    }

    public CallExpr makeStaticCall(MethodDec dec, Expr a1) {
        return makeStaticCall(dec.getMethod(), a1);
    }

    public CallExpr makeStaticCall(Method method, Exprs args) {
        return new CallExpr(
            getSourceLocation(),
            method,
            makeTypeExpr(method.getDeclaringType()),
            args);
    }

    public CallExpr makeStaticCall(Method dec, Expr a1) {
        return makeStaticCall(dec, makeExprs(a1));
    }

    public CallExpr makeStaticCall(Type type, String name, Exprs args) {
        return new CallExpr(
            getSourceLocation(),
            type.getMethod(name, null, args, true),
            makeTypeExpr(type),
            args);
    }

    public CallExpr makeStaticCall(Type type, String name) {
        return makeStaticCall(type, name, new Exprs(getSourceLocation()));
    }
    public CallExpr makeStaticCall(Type type, String name, Expr a1) {
        return makeStaticCall(type, name, new Exprs(getSourceLocation(), a1));
    }
    public CallExpr makeStaticCall(Type type, String name, Expr a1, Expr a2) {
        return makeStaticCall(type, name, new Exprs(getSourceLocation(), a1, a2));
    }
    public CallExpr makeStaticCall(
        Type type,
        String name,
        Expr a1,
        Expr a2,
        Expr a3) {
        return makeStaticCall(type, name, new Exprs(getSourceLocation(), a1, a2, a3));
    }
    public CallExpr makeStaticCall(
        Type type,
        String name,
        Expr a1,
        Expr a2,
        Expr a3,
        Expr a4) {
        return makeStaticCall(type, name, makeExprs(a1, a2, a3, a4));
    }
    public CallExpr makeStaticCall(
        Type type,
        String name,
        Expr a1,
        Expr a2,
        Expr a3,
        Expr a4,
        Expr a5) {
        return makeStaticCall(type, name, makeExprs(a1, a2, a3, a4, a5));
    }
    public CallExpr makeStaticCall(
        Type type,
        String name,
        Expr a1,
        Expr a2,
        Expr a3,
        Expr a4,
        Expr a5,
        Expr a6) {
        return makeStaticCall(type, name, makeExprs(a1, a2, a3, a4, a5, a6));
    }
    public CallExpr makeStaticCall(
        Type type,
        String name,
        Expr a1,
        Expr a2,
        Expr a3,
        Expr a4,
        Expr a5,
        Expr a6,
        Expr a7) {
        return makeStaticCall(type, name, makeExprs(a1, a2, a3, a4, a5, a6, a7));
    }

    public NewInstanceExpr makeNew(
        Type type,
        Exprs args,
        Constructor constructor) {
        return new NewInstanceExpr(
            getSourceLocation(),
            null,
            type.makeTypeD(),
            args,
            null,
            constructor);
    }

    public NewInstanceExpr makeNew(Type type, Exprs args) {
        return new NewInstanceExpr(
            getSourceLocation(),
            type.getConstructor(null, args, true),
            args);
    }

    public NewInstanceExpr makeNew(Type type) {
        return makeNew(type, new Exprs(getSourceLocation()));
    }
    public NewInstanceExpr makeNew(Type type, Expr a1) {
        return makeNew(type, new Exprs(getSourceLocation(), a1));
    }
    public NewInstanceExpr makeNew(Type type, Expr a1, Expr a2) {
        return makeNew(type, new Exprs(getSourceLocation(), a1, a2));
    }

    public NewArrayExpr makeNewArray(ArrayType type, ArrayInitializer init) {
        return new NewArrayExpr(getSourceLocation(), type.makeTypeD(), init, null);
    }

    public ConstructorCallExpr makeQualifiedConstructorCall(
        Exprs args,
        Constructor constructor,
        Expr primary) {
        return new ConstructorCallExpr(
            getSourceLocation(),
            primary,
            true,
            args,
            constructor);
    }

    public ConstructorCallExpr makeConstructorCall(
        boolean isSuper,
        Exprs args,
        Constructor constructor) {
        return new ConstructorCallExpr(getSourceLocation(), isSuper, args, constructor);
    }

    public ConstructorCallExpr makeSuperConstructorCall(
        Exprs args,
        Constructor constructor) {
        return makeConstructorCall(true, args, constructor);
    }

    public ConstructorCallExpr makeSuperConstructorCall(Constructor constructor) {
        return makeConstructorCall(true, makeExprs(), constructor);
    }

    public ConstructorBody makeConstructorBody(ConstructorCallExpr c, Stmts s) {
        return new ConstructorBody(getSourceLocation(), s, true, c);
    }

    public ConstructorDec makeConstructor(
        Modifiers m,
        Formals f,
        TypeDs ts,
        ConstructorCallExpr c,
        Stmts s) {
        return new ConstructorDec(
            getSourceLocation(),
            m,
            f,
            ts,
            makeConstructorBody(c, s));
    }

    public CallExpr makeCall(Expr instance, String name, Exprs args) {
        return new CallExpr(
            getSourceLocation(),
            instance.getType().getMethod(name, null, args, true),
            instance,
            args);
    }

    public CallExpr makeCall(Expr instance, String name) {
        return makeCall(instance, name, new Exprs(getSourceLocation()));
    }

    public CallExpr makeCall(Expr instance, String name, Expr a1) {
        return makeCall(instance, name, new Exprs(getSourceLocation(), a1));
    }

    public CallExpr makeCall(Expr instance, String name, Expr a1, Expr a2) {
        return makeCall(instance, name, new Exprs(getSourceLocation(), a1, a2));
    }

    public CallExpr makeCall(Method md, Expr instance, Exprs args) {
        if (instance == null) {
            md.getMethodDec().showError("INTERNAL: instance null in call to");
        }
        //??? check args match
        return new CallExpr(getSourceLocation(), md, instance, args);
    }

    public CallExpr makeCall(Method md, Expr instance) {
        return makeCall(md, instance, new Exprs(getSourceLocation()));
    }

    public CallExpr makeCall(Method md, Expr instance, Expr a1) {
        return makeCall(md, instance, new Exprs(getSourceLocation(), a1));
    }

    public CallExpr makeCall(Method md, Expr instance, Expr a1, Expr a2) {
        return makeCall(md, instance, new Exprs(getSourceLocation(), a1, a2));
    }

    public CallExpr makeCall(MethodDec md, Expr instance, Exprs args) {
        if (instance == null) {
            md.showError("INTERNAL: instance null in call to");
        }
        //??? check args match
        return new CallExpr(getSourceLocation(), md.getMethod(), instance, args);
    }

    public CallExpr makeCall(MethodDec md, Expr instance) {
        return makeCall(md, instance, new Exprs(getSourceLocation()));
    }

    public CallExpr makeCall(MethodDec md, Expr instance, Expr a1) {
        return makeCall(md, instance, new Exprs(getSourceLocation(), a1));
    }

    public CallExpr makeCall(MethodDec md, Expr instance, Expr a1, Expr a2) {
        return makeCall(md, instance, new Exprs(getSourceLocation(), a1, a2));
    }

    public CallExpr makeSuperCall(Method md, Exprs args) {
        //??? check args match
        return new CallExpr(
            getSourceLocation(),
            makeThis(md.getDeclaringType()),
            md.getName(),
            args,
            md,
            true);
    }

    public LogNotOpExpr makeLogNot(Expr arg) {
        return new LogNotOpExpr(getSourceLocation(), "!", arg);
    }

    public Exprs makeExprs() {
        Exprs exprs = new Exprs(getSourceLocation());
        return exprs;
    }

    public Exprs makeExprs(Expr a1) {
        Exprs exprs = new Exprs(getSourceLocation(), a1);
        return exprs;
    }

    public Exprs makeExprs(Expr a1, Expr a2) {
        Exprs exprs = new Exprs(getSourceLocation(), a1, a2);
        return exprs;
    }

    public Exprs makeExprs(Expr a1, Expr a2, Expr a3) {
        Exprs exprs = new Exprs(getSourceLocation(), a1, a2, a3);
        return exprs;
    }

    public Exprs makeExprs(Expr a1, Expr a2, Expr a3, Expr a4) {
        Exprs exprs = new Exprs(getSourceLocation());
        exprs.add(a1);
        exprs.add(a2);
        exprs.add(a3);
        exprs.add(a4);
        return exprs;
    }
    public Exprs makeExprs(Expr a1, Expr a2, Expr a3, Expr a4, Expr a5) {
        Exprs exprs = new Exprs(getSourceLocation());
        exprs.add(a1);
        exprs.add(a2);
        exprs.add(a3);
        exprs.add(a4);
        exprs.add(a5);
        return exprs;
    }
    public Exprs makeExprs(Expr a1, Expr a2, Expr a3, Expr a4, Expr a5, Expr a6) {
        Exprs exprs = new Exprs(getSourceLocation());
        exprs.add(a1);
        exprs.add(a2);
        exprs.add(a3);
        exprs.add(a4);
        exprs.add(a5);
        exprs.add(a6);
        return exprs;
    }
    public Exprs makeExprs(
        Expr a1,
        Expr a2,
        Expr a3,
        Expr a4,
        Expr a5,
        Expr a6,
        Expr a7) {
        Exprs exprs = new Exprs(getSourceLocation());
        exprs.add(a1);
        exprs.add(a2);
        exprs.add(a3);
        exprs.add(a4);
        exprs.add(a5);
        exprs.add(a6);
        exprs.add(a7);
        return exprs;
    }

    public TriTestExpr makeTriTest(Expr t, Expr c, Expr a) {
        return new TriTestExpr(getSourceLocation(), t, c, a);
    }

    public Stmt makeStmt(Expr expr) {
        return new ExprStmt(getSourceLocation(), expr);
    }

    public TriTestExpr makeIfExpr(Expr test, Expr trueValue, Expr falseValue) {
        return new TriTestExpr(getSourceLocation(), test, trueValue, falseValue);
    }

    public BinopExpr makeBinop(String op, Expr lhs, Expr rhs) {
        return BinopExpr.build(getSourceLocation(), op, lhs, rhs);
    }

    public UnopExpr makeUnop(String op, Expr lhs) {
        return UnopExpr.build(getSourceLocation(), op, lhs);
    }

    public FieldAccessExpr makeDynamicGet(FieldDec dec) {
        return makeGet(makeQualifiedThis(dec.getDeclaringType()), dec);
    }

    public FieldAccessExpr makeStaticGet(Type type, String name) {
        return new FieldAccessExpr(
            getSourceLocation(),
            makeTypeExpr(type),
            type.getField(name, null, true),
            false);
    }

    public Expr forceCast(Type toType, Expr expr) {
        CastExpr castExpr = new CastExpr(getSourceLocation(), toType.makeTypeD(), expr);
        if (!castExpr.isPossibleCast()) {
            expr.showError("can not be cast to " + toType.getString());
        }
        return makeParen(castExpr);
    }

    public Expr makeCast(Type toType, Expr expr) {
        Type exprType = expr.getType();
        if (exprType == expr.getTypeManager().anyType) {
            getCompiler().internalError(expr, "expr is anytype: " + expr);
        }
        if (exprType instanceof NullType) {
            return forceCast(toType, expr);
        }
        if (exprType.isEquivalent(toType) || toType.isAssignableFrom(exprType)) {
            if (expr instanceof CastExpr) {
                return makeParen(expr);
            } else {
                return expr;
            }
        }
        return forceCast(toType, expr);
    }

    public Expr makeNotInstanceofTest(Expr expr, Type ofType) {
        //??? test if possible
        if (expr.getType().isSubtypeOf(ofType))
            return makeLiteral(true);
        return makeUnop("!", makeParen(makeInstanceof(expr, ofType)));
    }

    public Expr makeInstanceof(Expr expr, Type ofType) {
        //??? test if possible
        return new InstanceofExpr(getSourceLocation(), expr, ofType.makeTypeD());
    }

    //public SuperExpr makeSuper() { return new SuperExpr(getSourceLocation()); }

    //public ThisExpr makeThis() { return new ThisExpr(getSourceLocation()); }
    public ThisExpr makeThis(Type thisType) {
        return new ThisExpr(getSourceLocation(), thisType);
    }

    public QualifiedThisExpr makeQualifiedThis(TypeD typeD) {
        return new QualifiedThisExpr(getSourceLocation(), typeD);
    }
    public QualifiedThisExpr makeQualifiedThis(Type type) {
        return makeQualifiedThis(type.makeTypeD());
    }

    public TypeExpr makeTypeExpr(TypeD typeD) {
        return new TypeExpr(getSourceLocation(), typeD);
    }
    public TypeExpr makeTypeExpr(Type type) {
        return makeTypeExpr(type.makeTypeD());
    }

    public Expr makeNonNullTest(Expr expr) {
        return makeBinop("!=", expr, makeNull());
    }
    public Expr makeNullTest(Expr expr) {
        return makeBinop("==", expr, makeNull());
    }

    public BooleanLiteralExpr makeLiteral(boolean b) {
        return new BooleanLiteralExpr(getSourceLocation(), b);
    }

    public IntLiteralExpr makeLiteral(int i) {
        return new IntLiteralExpr(getSourceLocation(), i);
    }

    public LongLiteralExpr makeLiteral(long l) {
        return new LongLiteralExpr(getSourceLocation(), l);
    }

    public FloatLiteralExpr makeLiteral(float f) {
        return new FloatLiteralExpr(getSourceLocation(), f);
    }

    public DoubleLiteralExpr makeLiteral(double d) {
        return new DoubleLiteralExpr(getSourceLocation(), d);
    }

    public StringLiteralExpr makeLiteral(String s) {
        return new StringLiteralExpr(getSourceLocation(), s);
    }

    public Expr makeLiteral(Type type) {
        return type.getClassExpr();
    }

    public NullExpr makeNull() {
        return new NullExpr(getSourceLocation());
    }

    public Exprs makeVars(Formals formals) {
        return makeVars(formals, 0);
    }

    public Exprs makeVars(Formals formals, int i) {
        Exprs es = makeExprs();
        for (int len = formals.size(); i < len; i++) {
            es.add(makeVar(formals.get(i)));
        }
        return es;
    }

    public Expr makeNotFoundExpr() {
        return makeVar(makeVarDec(getTypeManager().TYPE_NOT_FOUND, "not_found", null));
    }

    public VarExpr makeVar(VarDec dec) {
        return new VarExpr(getSourceLocation(), dec);
    }

    public Expr makeParen(Expr expr) {
        return new ParenExpr(getSourceLocation(), expr);
    }

    public AssignExpr makeSet(AssignableExpr lhs, Expr rhs) {
        return AssignExpr.build(getSourceLocation(), lhs, "", rhs);
    }

    public AssignExpr makeSet(AssignableExpr lhs, String op, Expr rhs) {
        return AssignExpr.build(getSourceLocation(), lhs, op, rhs);
    }

    public AssignExpr makeSet(VarDec vd, Expr rhs) {
        return AssignExpr.build(getSourceLocation(), makeVar(vd), "", rhs);
    }
    public AssignExpr makeSet(FieldDec fd, Expr rhs) {
        return makeSet(fd.getField(), rhs);
    }
    public AssignExpr makeSet(Field fd, Expr rhs) {
        return AssignExpr.build(getSourceLocation(), makeGet(fd), "", rhs);
    }

    public AssignExpr makeSet(Expr i, Field field, Expr rhs) {
        return AssignExpr.build(getSourceLocation(), makeGet(i, field), "", rhs);
    }

    public AssignExpr makeSet(Expr i, FieldDec field, Expr rhs) {
        return makeSet(i, field.getField(), rhs);
    }
    public FieldAccessExpr makeGet(Expr i, FieldDec field) {
        return makeGet(i, field.getField());
    }
    public FieldAccessExpr makeGet(FieldDec field) {
        return makeGet(field.getField());
    }
    public FieldAccessExpr makeGet(Expr i, Field field) {
        if (i == null) {
            field.getFieldDec().showError("INTERNAL: instance null in get");
            return makeGet(field);
        }
        return new FieldAccessExpr(getSourceLocation(), i, field, false);
    }

    public PrefixExpr makePrefix(Expr i, FieldDec field, String op) {
        return new PrefixExpr(getSourceLocation(), makeGet(i, field), op);
    }
    public PostfixExpr makePostfix(Expr i, FieldDec field, String op) {
        return new PostfixExpr(getSourceLocation(), makeGet(i, field), op);
    }

    public Expr makePrimary(SemanticObject o, NameType referenceType) {
        NameType declaringType = (NameType) o.getDeclaringType();
        return makePrimary(declaringType, referenceType, o.isStatic());
    }

    public Expr makePrimary(SemanticObject o, TypeDec referenceDec) {
        NameType declaringType = (NameType) o.getDeclaringType();
        NameType referenceType = (NameType) referenceDec.getType();
        return makePrimary(declaringType, referenceType, o.isStatic());
    }

    public Expr makePrimary(Dec dec, TypeDec referenceDec) {
        NameType declaringType = (NameType) dec.getDeclaringType();
        NameType referenceType = (NameType) referenceDec.getType();
        return makePrimary(declaringType, referenceType, dec.isStatic());
    }

    public Expr makePrimary(Dec dec, NameType referenceType) {
        NameType declaringType = (NameType) dec.getDeclaringType();
        return makePrimary(declaringType, referenceType, dec.isStatic());
    }

	public Expr makePrimary(NameType declaring, NameType reference, boolean isStatic) {
		return makePrimary(declaring, reference, isStatic, true);
	}

    public Expr makePrimary(
        NameType declaring,
        NameType reference,
        boolean isStatic,
        boolean isThisAvailable) {
        //System.out.println(declaring + " : " + reference);
		if (reference.isSubtypeOf(declaring) && isStatic) {
			return makeTypeExpr(reference);
		} 
		if (reference.isSubtypeOf(declaring) && isThisAvailable) {
			return makeThis(reference);
		}

//        if (reference.isSubtypeOf(declaring)) {
//            return isStatic ? makeTypeExpr(reference) : (Expr) makeThis(reference);
//        }
        NameType curr = reference;
        while (!curr.isPackageMember()) {
            curr = curr.getEnclosingType();
            if (curr.isSubtypeOf(declaring)) {
                return isStatic ? makeTypeExpr(curr) : (Expr) makeQualifiedThis(curr);
            }
        }
        throw new RuntimeException(
            "No enclosing type for " + declaring + " and " + reference);
    }

    public Expr makeEnclosingPrimary(NameType declaring, NameType reference) {
        NameType curr = reference;
        while (!curr.isPackageMember()) {
            curr = curr.getEnclosingType();
            if (curr.isSubtypeOf(declaring)) {
                return makeQualifiedThis(curr);
            }
        }
        throw new RuntimeException(
            "No enclosing type for " + declaring + " and " + reference);
    }

    public FieldAccessExpr makeGet(Field field) {
        // !!! DANGER, this should be filled with a search
        // debugging strategy:  uncomment this exception, and replace all places
        // that call this method that throw the exception with a variant
        // of makeDynamicGet.
        //if (! field.isStatic()) throw new RuntimeException("bad null");
        return makeGet(
            field.isStatic()
                ? (Expr) makeTypeExpr(field.getDeclaringType())
                : makeThis(null),
            field);
    }

    public Decs makeDecs() {
        return new Decs(getSourceLocation());
    }
    public TypeDs makeTypeDs() {
        return new TypeDs(getSourceLocation());
    }
    public TypeDs makeTypeDs(TypeD ty0) {
        return new TypeDs(getSourceLocation(), ty0);
    }

    public NewArrayExpr makeArray(Type type, Exprs values) {
        return new NewArrayExpr(
            getSourceLocation(),
            new ResolvedTypeD(getSourceLocation(), type.getArrayType()
        // new ArrayType(getSourceLocation(),type)
        ), new ArrayInitializer(getSourceLocation(), values), null);
    }

    public NewArrayExpr makeObjectArray(Exprs values) {
        Exprs objectExprs = new Exprs(getSourceLocation());

        for (int i = 0; i < values.size(); i++) {
            Expr value = values.get(i);
            objectExprs.add(value.getType().makeObject(value));
        }

        return makeArray(getTypeManager().getObjectType(), objectExprs);
    }

    public Modifiers makeModifiers(int value) {
        return new Modifiers(getSourceLocation(), value);
    }

    public FormalDec makeFormal(Type type, String id) {
        return makeFormal(type, id, false);
    }
    public FormalDec makeFinalFormal(Type type, String id) {
        return makeFormal(type, id, true);
    }

    public FormalDec makeFormal(Type type, String id, boolean isFinal) {
        return new FormalDec(
            getSourceLocation(),
            makeModifiers(isFinal ? Modifiers.FINAL : 0),
            type.makeTypeD(),
            id,
            null);
    }

    public Formals makeFormals() {
        return new Formals(getSourceLocation());
    }

    public Formals makeFormals(FormalDec f1) {
        return new Formals(getSourceLocation(), f1);
    }

    public Formals makeFormals(FormalDec f1, FormalDec f2) {
        return new Formals(getSourceLocation(), f1, f2);
    }
    public VarDec makeVarDec(Modifiers mods, Type type, String id, Expr init) {
        return new VarDec(getSourceLocation(), mods, type.makeTypeD(), id, init);
    }
    public VarDec makeVarDec(Type type, String id, Expr init, boolean isFinal) {
        return new VarDec(
            getSourceLocation(),
            makeModifiers(isFinal ? Modifiers.FINAL : 0),
            type.makeTypeD(),
            id,
            init);
    }

    public VarDec makeVarDec(Type type, String id, Expr init) {
        return new VarDec(getSourceLocation(), type.makeTypeD(), id, init);
    }
    public VarDec makeFinalVar(Type type, String id, Expr init) {
        return new VarDec(
            getSourceLocation(),
            makeModifiers(Modifiers.FINAL),
            type.makeTypeD(),
            id,
            init);
    }

    public ArrayExpr makeArrayRef(Expr arrayExpr, int index) {
       return new ArrayExpr(getSourceLocation(), arrayExpr, makeLiteral(index));
    }

    public FieldDec makeField(Modifiers mods, Type type, String id) {
        return new FieldDec(getSourceLocation(), mods, type.makeTypeD(), id, null);
    }

    public FieldDec makeField(Modifiers mods, Type type, String id, Expr init) {
        return new FieldDec(getSourceLocation(), mods, type.makeTypeD(), id, init);
    }

    public InitializerDec makeInitializer(Stmts body) {
        return new InitializerDec(
            getSourceLocation(),
            makeModifiers(0),
            new CodeBody(getSourceLocation(), body, true));
    }
    public InitializerDec makeStaticInitializer(Stmts body) {
        return new InitializerDec(
            getSourceLocation(),
            makeModifiers(Modifiers.STATIC),
            new CodeBody(getSourceLocation(), body, true));
    }

    public MethodDec makeMethod(
        Modifiers mods,
        Type retType,
        String name,
        Formals formals,
        BlockStmt body) {
        MethodDec ret =
            new MethodDec(
                getSourceLocation(),
                mods,
                retType.makeTypeD(),
                name,
                formals,
                null,
                body);
        ret.computeMinimalThrows();
        return ret;
    }

    public CodeBody makeCodeBody(Stmt s) {
        return new CodeBody(getSourceLocation(), makeStmts(s), true);
    }

    public CodeBody makeCodeBody(Stmt s0, Stmt s1) {
        return new CodeBody(getSourceLocation(), makeStmts(s0, s1), true);
    }

    public IfStmt makeIf(Expr test, Stmt trueStmt, Stmt falseStmt) {
        //          if (falseStmt == null) {
        //              throw new RuntimeException("BLEH");
        //          }
        return new IfStmt(getSourceLocation(), test, trueStmt, falseStmt);
    }

    public IfStmt makeIf(Expr test, Stmt trueStmt) {
        IfStmt s = new IfStmt(getSourceLocation(), test, trueStmt, null);
        s.setElse((Stmt) s.getAST().makeEmptyStmt());
        return s;
    }

    public TryFinallyStmt makeTryFinally(Stmt block, Stmt finallyBlock) {
        return new TryFinallyStmt(getSourceLocation(), block, finallyBlock);
    }

    public TryCatchStmt makeTryCatch(
        BlockStmt block,
        FormalDec formal,
        Stmt catchStmt) {
        return makeTryCatch(
            block,
            new CatchClause(getSourceLocation(), formal, catchStmt));
    }

    public TryCatchStmt makeTryCatch(BlockStmt block, CatchClause catchClause) {
        return new TryCatchStmt(
            getSourceLocation(),
            block,
            new CatchClauses(getSourceLocation(), catchClause));
    }

    public WhileStmt makeWhile(Expr test, Stmt body) {
        return new WhileStmt(getSourceLocation(), test, body);
    }

    public ReturnStmt makeReturn() {
        return new ReturnStmt(getSourceLocation());
    }

    public ReturnStmt makeReturn(Expr e) {
        return new ReturnStmt(getSourceLocation(), e);
    }

    public ThrowStmt makeThrow(Expr e) {
        return new ThrowStmt(getSourceLocation(), e);
    }

    public ThrowStmt makeThrow(Type t) {
        return new ThrowStmt(getSourceLocation(), makeNew(t));
    }

    public ThrowStmt makeThrow(Type t, String message) {
        return new ThrowStmt(getSourceLocation(), makeNew(t, makeLiteral(message)));
    }

    public CatchClause makeCatch(FormalDec formal, BlockStmt body) {
        return new CatchClause(getSourceLocation(), formal, body);
    }

    public LabeledStmt makeLabeled(String label, Stmt stmt) {
        return new LabeledStmt(getSourceLocation(), label, stmt);
    }
    
    public Stmt makeBreak(String label) {
        return new BreakStmt(getSourceLocation(), label);
    }

    public Stmt makeContinue(String label) {
        return new ContinueStmt(getSourceLocation(), label);
    }

    public BlockStmt makeBlock() {
        return new BlockStmt(getSourceLocation(), new Stmts(getSourceLocation()));
    }

    public BlockStmt makeBlock(ASTObject s1) {
        if (s1 instanceof Stmts)
            return new BlockStmt(getSourceLocation(), (Stmts) s1);
        else
            return new BlockStmt(getSourceLocation(), s1);
    }

    public BlockStmt makeBlock(ASTObject s1, ASTObject s2) {
        return new BlockStmt(getSourceLocation(), s1, s2);
    }

    public BlockStmt makeBlock(ASTObject s1, ASTObject s2, ASTObject s3) {
        return new BlockStmt(getSourceLocation(), s1, s2, s3);
    }

    public BlockStmt makeBlock(
        ASTObject s1,
        ASTObject s2,
        ASTObject s3,
        ASTObject s4) {
        return new BlockStmt(getSourceLocation(), s1, s2, s3, s4);
    }

    public BlockStmt makeBlock(Stmts stmts) {
        return new BlockStmt(getSourceLocation(), stmts);
    }

    public Stmts makeStmts() {
        return new Stmts(getSourceLocation());
    }

    public Stmts makeStmts(Stmt s1) {
        return new Stmts(getSourceLocation(), s1);
    }

    public Stmts makeStmts(Stmt s1, Stmt s2) {
        return new Stmts(getSourceLocation(), s1, s2);
    }

    public Stmts makeStmts(Stmt s1, Stmt s2, Stmt s3) {
        return new Stmts(getSourceLocation(), s1, s2, s3);
    }

    public EmptyStmt makeEmptyStmt() {
        return new EmptyStmt(getSourceLocation());
    }

    public InterfaceDec makeInnerInterface(
        TypeDec outerTypeDec,
        Modifiers modifiers,
        String name,
        TypeDs _implements) {
        InterfaceDec ret =
            new InterfaceDec(
                getSourceLocation(),
                modifiers,
                name,
                _implements,
                new Decs(getSourceLocation()));
        //ret.fullName = outerTypeDec.getFullName() + '$' + name;
        outerTypeDec.getBody().add(ret);
        return ret;
    }

    public InterfaceDec makeInterface(
        Modifiers modifiers,
        String name,
        TypeDs _implements) {
        InterfaceDec ret =
            new InterfaceDec(
                getSourceLocation(),
                modifiers,
                name,
                _implements,
                new Decs(getSourceLocation()));
        //ret.fullName = fullName;
        return ret;
    }

    /**
     * The contents of this class dec must all be name bound
     */
    public ClassDec makeClass(
        Modifiers modifiers,
        String name,
        TypeD superClass,
        TypeDs _implements) {
        ClassDec ret =
            new ClassDec(
                getSourceLocation(),
                modifiers,
                name,
                superClass,
                _implements,
                new Decs(getSourceLocation()));
        //ret.fullName = fullName;
        return ret;
    }

    public AnonymousMethodExpr makeAnonMethod(Type retType, Stmts stmts) {
        return new AnonymousMethodExpr(
            getSourceLocation(),
            makeExprs(),
            makeMethod(makeModifiers(0), retType, "anon", makeFormals(), makeBlock(stmts)));
    }
}