/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

// to get the passes for introductions
import org.aspectj.compiler.crosscuts.*;

import java.util.*;
import java.io.IOException;

public abstract class ASTObject extends CompilerObject {
    protected SourceLocation sourceLocation = null;
    protected ASTObject parent;

    public ASTObject(SourceLocation loc) {
        super(loc.getCompiler());
        this.sourceLocation = loc;
    }

    public boolean hasSourceLocation() {
        return sourceLocation != null;
    }

    public SourceLocation getSourceLocation() {
        return sourceLocation;
    }
    public void setSourceLocation(SourceLocation newLocation) {
        sourceLocation = newLocation;
    }

    public int getStartPosition() { return getSourceLocation().getStartPosition(); }
    public int getEndPosition() { return getSourceLocation().getEndPosition(); }
    public int getBeginLine() { return getSourceLocation().getBeginLine(); }
    public int getEndLine() { return getSourceLocation().getEndLine(); }
    public int getBeginColumn() { return getSourceLocation().getBeginColumn(); }
    public int getEndColumn() { return getSourceLocation().getBeginLine(); }

    public Comment getComment() {
        return getSourceLocation().getComment();
    }

    public void addComment(Comment _comment) {
        getSourceLocation().addComment(_comment);
    }

    public boolean isSynthetic() {
        return getSourceLocation().isSynthetic();
    }
    public boolean isLanguageVisible() {
        return !getSourceLocation().isSynthetic();
    }


    public void setFormalComment(String comment) {
        getSourceLocation().setFormalComment(comment);
    }

    public String getFormalComment() {
        return getSourceLocation().getFormalComment();
    }

    public void clearComment() {
        getSourceLocation().clearComment();
    }

    public CompilationUnit getCompilationUnit() {
        return getSourceLocation().getCompilationUnit();
    }

    public java.io.File getSourceFile() {
        return getSourceLocation().getSourceFile();
    }

    final public String getSourceFileName() {
        return getSourceLocation().getSourceFileName();
    }

    final public String getSourceDirectoryName() {
        return getSourceLocation().getSourceDirectoryName();
    }


    public boolean hasSource() {
        return getSourceLocation().hasSource();
    }

    public boolean fromSource() {
        return getSourceLocation().fromSource();
    }


    //!!! probably should remove
    public void setParent(ASTObject x) {
        //!!! this should probably lead to warnings at some point in time
        if (getOptions().strictTree && parent != null && parent != x) {
            this.showError("changing parent from " + parent + " to " + x);
        }
        parent = x;
    }


    // basically steal the API from javax.swing.Tree.TreeNode
    public ASTObject getParent() { return parent; }

    public int getChildCount() { return 0; }

    // sets a policy about how to handle out-of-bounds indices
    public ASTObject getChildAt(int childIndex) { return null; }
    public String getChildNameAt(int childIndex) { return null; }
    public void setChildAt(int childIndex, ASTObject child) {}
    public void removeChildAt(int childIndex) {
        remove(childIndex);
    }

    public void remove(int index) {
        //XXX
        setChildAt(index, null);
        //throw new RuntimeException("can't remove child from: " + this);
    }
    
    public void remove(ASTObject child) {
        int index = indexOf(child);
        if (index == -1) {
            throw new IllegalArgumentException("not a child");
        }
        remove(index);
    }
    

    public boolean isLeaf() {
        return getChildCount() == 0;
    }

    public int indexOf(ASTObject node) {
        int n = getChildCount();
        for(int i=0; i<n; i++) {
            //System.out.println(this);
            ASTObject child = getChildAt(i);

            if (child != null && child == node) return i; //.equals(node)) return i;
        }
        return -1;
    }

    public boolean contains(ASTObject node) {
        return indexOf(node) != -1;
    }

    // this is a freaky method to have here
    public boolean hasLegalProtectedAccess(Type fromType) {
        //XXX must be overriden appropriately to implement JLSv2 6.6.2
        return true;
    }
    
    public final boolean containsTypes() {
        if (this instanceof TypeDec) return true;
        
        for(int i=0, n=getChildCount(); i<n; i++) {
            ASTObject child = getChildAt(i);
            if (child != null && child.containsTypes()) return true;
        }
        return false;
    }

    // ------------------------------
    // Intro: ForwardReferenceChecker

    public void walkForwardReference(ForwardReferenceChecker w) {
        this.walk(w);
    }

    // ------------------------------
    // Intro: FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        this.walk(w);
    }

    // ------------------------------
    // Intro:  FrameLocPass

    public void walkFrameLoc(FrameLocPass walker) {
        this.walk(walker);
    }



    //INTRO from MovingWalker
    public void preMove(MovingWalker walker) {
    }

    public ASTObject postMove(MovingWalker walker) {
        return this;
    }

    //INTRO from AccessFixer
    public ASTObject fixAccessPost(org.aspectj.compiler.crosscuts.AccessFixer fixer) {
        return this;
    }

    //INTRO from CopyWalker
    public ASTObject copyWalk(CopyWalker walker) { return null; }
    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
        return this;
    }
    public void preCopy(CopyWalker walker, ASTObject oldObject) { }

    //INTRO from IntroductionPlannerPass
    public void preIntroduction(IntroductionPlannerPass pass) { }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) {  }
    public ASTObject postScope(ScopeWalker walker) { return this; }
    public void walkScope(ScopeWalker walker) { walk(walker); }


    public void walk(Walker walker) {
        //int N = getChildCount();
        for(int i=0; i<getChildCount(); i++) {
            /*
            if (this instanceof TypeDec) {
                System.out.println("walk: " + getChildNameAt(i));
                if (getChildNameAt(i).equals("implements")) {
                    if (getChildAt(i) == null) {
                        System.out.println("    null");
                    } else {
                        getChildAt(i).showWarning("implements clause: ");
                        getChildAt(i).display(2);
                    }
                }
            }
            */
            ASTObject child = getChildAt(i);
            if (child != null) {
                if (getOptions().strictTree && child.getParent() != this) {
                    child.showError("wrong parent: " + this.unparse() + " '" + child.unparse() + "'");
                    System.out.println(child.getParent().unparse());
                }
                ASTObject newChild = walker.process(child);
                if (newChild == child) continue;
                if (newChild == null) {
                    removeChildAt(i);
                    //System.out.println("removed: " + i);
                    i -= 1;
                    //N -= 1;
                } else {
                    setChildAt(i, newChild);
                }
            }
        }
    }

    public Type getOutermostLexicalType() {
        if (parent == null) return getTypeManager().TYPE_NOT_FOUND;
        return parent.getOutermostLexicalType();
    }

    public Type getOutermostBytecodeType() {
        if (parent == null) return getTypeManager().TYPE_NOT_FOUND;

        return parent.getOutermostBytecodeType();
    }

    /** returns the lexically enclosing Type for this object.
     *  For pure Java code this is always identical to getDeclaringType() and
     *  getBytecodeTypeDec().getType().
     *
     * For introduced methods, this can be different, i.e.
     * <code>
     * interface I { }
     * class C implements I { }
     * aspect A {
     *   void C.m() { ... }
     * }
     * </code>
     *
     * For the method I.m() implemented in C and for any ASTObjects inside of it,
     *   getLexicalType() == A
     *   getDeclaringType() == I
     *   getBytecodeTypeDec().getType() == C
     */


    public Type getLexicalType() {
        //XXX awful hack
        if (getParent() == null) {
            //showError("no parent: " + this);
            return null;
        }
        if (this instanceof Decs && getParent() instanceof TypeDec) return ((TypeDec)getParent()).getType();
        return getParent().getLexicalType();
    }

    /**
     * returns the declaring type of a method.  For non-static methods, this type
     * will always be the type of 'this' within the method.  For static methods, this
     * type will always be the type used to name the method for outside
     * of its scope, i.e. DeclaringType.m()
     *
     * For a comparision with getLexicalType() and getBytecodeTypeDec() see
     * getLexicalType().
     */

    public Type getDeclaringType() {
        if (getParent() == null) {
            //showError("no parent: " + this);
            return null;
        }
        if (getParent() instanceof TypeDec) return ((TypeDec)getParent()).getType();
        return getParent().getDeclaringType();
    }

    /**
     * not part of the language's semantic model, but part of the compiler's
     * generated code model.  This will return the type declaration
     * that will actually be
     * enclosing this point in the bytecodes produced by the compiler.
     */
    public TypeDec getBytecodeTypeDec() {
        if (getParent() == null) {
            showError("no parent: " + this + ": " + System.identityHashCode(this));
            System.out.println(unparse());
            return getTypeManager().TYPE_DEC_NOT_FOUND; //null;
        }
        if (getParent() instanceof TypeDec) return (TypeDec)getParent();
        return getParent().getBytecodeTypeDec();
    }


    public Type getBytecodeType() {
        TypeDec dec = getBytecodeTypeDec();
        if (dec == null) return null;
        return dec.getType();
    }

    public Dec getEnclosingDec() {
        if (getParent() == null) {
            //this.showWarning("missing parent");
            //this.display(2);
            return null;
        }
        return getParent().getEnclosingDec();
    }

    public CodeDec getEnclosingCodeDec() {
        if (getParent() == null) {
            //this.showWarning("missing parent");
            //this.display(2);
            return null;
        }
        return getParent().getEnclosingCodeDec();
    }
    


    public boolean inStaticContext() {
        if (getParent() == null) return true;
        return getParent().inStaticContext();
    }

    public void showTypeError(Type foundType, Type requiredType) {
        if (foundType.isMissing() || requiredType.isMissing()) return;
        showError("incompatible types\n" +
                  "found   : " + foundType.getString() + "\n" +
                  "required: " + requiredType.getString());
    }
    public void showError(String message) {
        getCompiler().showError(this,message);
    }
    public void showWarning(String message) {
        getCompiler().showWarning(this,message);
    }
    //XXX the Tester treats these like showWarning
    public void showMessage(String message) {
        getCompiler().showMessage(this,message);
    }

    public String toShortString() {
        return this.getClass().getName();
    }

    public ASTObject copy() { return CopyWalker.copy(this); }


    public void checkNoSharing(ASTObject other) {
        Set s1 = new HashSet();
        Set s2 = new HashSet();
        this.addContainedNodes(s1);
        other.addContainedNodes(s2);
        
        s1.retainAll(s2);
        if (!s1.isEmpty()) {
            getCompiler().showError(this, "shares nodes with " + other + "\n" +
                                        s1);
        }
    }
    
    /**
     * adds nodes I contain to a Set
     * 
     * if I'm already in the Set signals an error to catch circularities
     */
    private void addContainedNodes(Set ret) {
        if (ret.contains(this)) {
            getCompiler().showError(this, "duplicate node in " + ret);
        }
        ret.add(this);
        for(int i=0; i<getChildCount(); i++) {
            ASTObject o = getChildAt(i);
            if (o != null) o.addContainedNodes(ret);
        }
    }

    /* Sets the ASTObject that this one came from, not the source file.
     * Call this method whenever you transfroms one astobject into another one.
     *
     * @returns this in order to allow lazy method chaining
     */
    public ASTObject setSource(ASTObject sourceObject) {
        sourceLocation = sourceObject.sourceLocation;
        return this;
    }

    /** Sets the ASTObject that this one came from, not the source
        file, but from a synthetic location.
    */
    public ASTObject setSyntheticSource(ASTObject sourceObject) {
        sourceLocation = new SourceSourceLocation(sourceObject);
        return this;
    }

    protected void makeChild(ASTObject object) {
        if (object != null) object.parent = this;
    }

    protected void makeChild(ASTObject[] array) {
        if (array == null) return;
        int n = array.length;
        for(int i=0; i<n; i++) {
            makeChild(array[i]);
        }
    }

    public String getDefaultDisplayName() {
        return this.getClass().getName();
    }

    public void clearParent() { parent = null; }

    protected void setParents() {}

    public void replaceWith(ASTObject other) {
        ASTObject parent = getParent();
        if (parent == null) {
            showWarning("no parent to replace in");
            display(2);
            return;
        }
        int i = parent.indexOf(this);
        if (i == -1) {
            showError("invalid replacement attempt, not in: " + parent);
            parent.display(2);
            return;
        }
        //System.out.println("replacing: " + this + " with " + other + " at " + i + " in " + parent);
        parent.setChildAt(i, other);
    }

    public String toString() {
        return getDefaultDisplayName(); // + System.identityHashCode(this);
    }

    public void display(int indent) {
        for(int s=0; s<indent; s++) System.out.print(" ");
        System.out.println(getDefaultDisplayName() + System.identityHashCode(this)); //+": "+getContext()); //System.identityHashCode(this)); //getContext());

        final int N = getChildCount();
        for(int i=0; i<N; i++) {
            ASTObject child = getChildAt(i);
            if (child != null) {
                child.display(indent+1);
            }
        }
    }

    public AST getAST() {
        //XXX would like a more sophisticated source location here
        return new AST(new SourceSourceLocation(this));
    }

    public String unparse() {
        CodeWriter writer = new CodeWriter(getCompiler());
        writer.write(this);
        return writer.getString();
    }

    //XXX this should be abstract before all is said and done
    public void unparse(CodeWriter writer) throws IOException {
        writer.write("/* ");
        writer.write(getDefaultDisplayName());
        writer.write(" */");
        writer.newLine();
    }

    public void cleanup() {
    }

    //INTRO from ASTFixer
    public void fixAST(ASTFixerPass fixer) { this.walk(fixer); }
    public ASTObject postFixAST(ASTFixerPass fixer) { return this; }

    //INTRO from MixinImplementationPass
    public void implementMixin(MixinImplementationPass walker) { this.walk(walker); }
    public ASTObject postImplementMixin(MixinImplementationPass walker) { return this; }

    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass walker) {
        this.walk(walker);
    }
    public void postInnerInfo(InnerInfoPass walker) { }

    // ------------------------------
    // INTRO:  SpecCheckerPass

    public void checkSpec() { }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public void preAssignmentCheck(AssignmentCheckerPass walker) { }
    public void walkAssignmentCheck(AssignmentCheckerPass walker) {
        this.walk(walker);
    }
    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        return this;
    }

    // ------------------------------
    // INTRO from LocalClassPass

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        walk(walker);
    }
    public void preLift(LocalClassPass.LiftWalker walker) {}
    public ASTObject postLift(LocalClassPass.LiftWalker walker) {
        return this;
    }
    public void preThreading(LocalClassPass.ThreadingWalker walker) {}
    public ASTObject postThreading(LocalClassPass.ThreadingWalker walker) {
        return this;
    }

    // ------------------------------
    // INTRO from MemberClassMunger

    public ASTObject walkMemberMunger(MemberClassMunger w) {
        this.walk(w);
        return this;
    }

    // ------------------------------
    //INTRO from InnerAccessFixer

    /** Converts an expression accessing inaccessable context to one
        that is legal.  */
    public ASTObject postInnerAccess(InnerAccessFixer w) {
        return this;
    }

    /** Builds a method that would perform the actions of this
        expression.  The InnerAccessFixer object keeps a cache of such
        methods to provide to the InnerAccessFixer walk, but if no
        appropriate method has yet been built, the original expression
        is the one that knows how to build it.  This should only be
        overridden in {@link PrefixExpr}, {@link PostfixExpr}, {@link
        FieldAccessExpr}, {@link AssignExpr}, and {@link CallExpr}.
        If and when we do cookie methods for constructor, we may also
        need to override this in {@link NewInstanceExpr} and {@link
        ConstructorCallExpr}.
    */
    public MethodDec buildAccessMethod(InnerAccessFixer w) {
        throw new RuntimeException("No access method should be built for " + this);
    }

    // ------------------------------
    // bcg cleanups

    public void walkCleanup(ByteCodeCleanupPass walker) {
        this.walk(walker);
    }
    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        return this;
    }
}
