/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.io.IOException;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar rand1 + rand2
 */
public class AddOpExpr extends BinopExpr {

    void checkAssignmentOperatorType() {
        Type ty1 = getRand1().getType();
        Type ty2 = getRand2().getType();

        if (ty1.isVoid() || ty2.isVoid()) {
            showOperatorTypeError("+", ty1, ty2);
        } else if (ty1.isAnyType() || ty2.isAnyType() ||
                   ty1.isString() || ty2.isString() ||
                   (ty1.isNumeric() && ty2.isNumeric())) {
        } else {
            showOperatorTypeError("+", ty1, ty2);
        }
    }

    protected LiteralExpr halfFold(Type type, LiteralExpr lit1, LiteralExpr lit2) {
        return type.foldAddOp(lit1, lit2);
    }

    protected Type discoverType() {
        Expr rand1 = getRand1();
        Expr rand2 = getRand2();
        Type ty1 = rand1.getType();
        Type ty2 = rand2.getType();

        if        (ty1.isAnyType() || ty2.isAnyType()) {
            return getTypeManager().anyType;
        } else if (ty1.isVoid() || ty2.isVoid()) {
            showOperatorTypeError("+", ty1, ty2);
            return getTypeManager().anyType;
        } else if (ty1.isString() || ty2.isString()) {
            return getTypeManager().getStringType();
        } else if (ty1.isNumeric() && ty2.isNumeric()) {
            return getTypeManager().binaryNumericPromotion(ty1, ty2);
        } else {
            showOperatorTypeError("+", ty1, ty2);
            return getTypeManager().anyType;
        }
    }

    protected Type getLiftType() {
        return getType();
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        Type liftTy = getLiftType();
        if (liftTy.isString()) {
            NameType stringBufferType = getTypeManager().getStringBufferType();
            cb.emitNEW(stringBufferType);
            cb.emitDUP();
            cb.emitINVOKESPECIAL(stringBufferType, "<init>", "()V", -1);
            rand1.cgBuffer(cb);
            rand2.cgBuffer(cb);
            cb.emitINVOKEVIRTUAL(stringBufferType, "toString",
                                 "()Ljava/lang/String;", 0);
        } else {
            rand1.cgValue(cb, liftTy);
            rand2.cgValue(cb, liftTy);
            liftTy.emitAdd(cb);
        }
    }
    protected void cgBuffer(CodeBuilder cb) {
        if (getType().isString()) {
            rand1.cgBuffer(cb);
            rand2.cgBuffer(cb);
        } else {
            super.cgBuffer(cb);
        }
    }
    protected void cgOp(CodeBuilder cb, Type ty) {
        ty.emitAdd(cb);
    }

    //BEGIN: Generated from @child and @property

    public AddOpExpr(SourceLocation location, Expr _rand1, String _op, Expr _rand2) {
        super(location, _rand1, _op, _rand2);

    }
    protected AddOpExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        AddOpExpr ret = new AddOpExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (rand1 != null) ret.setRand1( (Expr)walker.process(rand1) );
        ret.op = op;
        if (rand2 != null) ret.setRand2( (Expr)walker.process(rand2) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "AddOpExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
