/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

/**
  * @grammar assert test [:message]
  * @child Expr test
  * @child Expr message
  */

public class AssertStmt extends TestStmt {

    public void unparse(CodeWriter writer) {
        writer.writeKeyword("assert");
        writer.requiredSpace();
        writer.write(test);
        if (getMessage() != null) {
            writer.optionalSpace();
            writer.write(":");
            writer.optionalSpace();
            writer.write(getMessage());
        }
        writer.closeStmt();
    }

    /* Should this only be used in bcg mode???
     * (no, because of the next point)
     * This doesn't work correctly with "moved" code???
     * (yes, it does, because we use the lexical type to get the field)
     */
    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (! getCompiler().willGenerateSourceCode()) {
            return deSugar();
        }

        if (getOutermostLexicalType() == getOutermostBytecodeType()) {
            return this;
        }

        if (getLexicalType().isEffectivelyStatic()) { // may be "effectively"
            return deSugar();
        } else {
            showWarning("compiler limitation: this assert will behave as if inside " + getOutermostBytecodeType().toShortString());
            return this;
        }
    }

    /** can't be called in non-bcg mode if my immediately enclosing
     * lexical type is non-static
     */
    private ASTObject deSugar() {
        final AST ast = getAST();
        Field adf =
            getLexicalType().getTypeDec().getAssertionsDisabledField();

        Type assertionErrorType =
            getTypeManager().getType("java.lang", "AssertionError");
        Exprs args = ast.makeExprs();
        if (message != null) {
            Expr e = message;
            if (e.getType().isReferenceType()) {
                e = ast.forceCast(getTypeManager().getObjectType(), e);
            }
            args.add(e);
        }
        Stmt ret = ast.makeIf(ast.makeUnop("!", ast.makeParen(test)),
                   ast.makeThrow(ast.makeNew(assertionErrorType, args)));

        ret = ast.makeIf(ast.makeUnop("!", ast.makeGet(adf)), ret);
        return ret;

    }


    // ------------------------------
    // flow
    public void walkFlow(FlowCheckerPass w) {
        FlowCheckerPass.Vars v0 = w.getVars();

        w.process(getTest());
        if (getMessage() != null) w.process(getMessage());

        FlowCheckerPass.Vars v1 = w.getVars();
        w.setVars(v0);
    }

    //BEGIN: Generated from @child and @property
    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }

    protected Expr message;
    public Expr getMessage() { return message; }
    public void setMessage(Expr _message) {
        if (_message != null) _message.setParent(this);
        message = _message;
    }

    public AssertStmt(SourceLocation location, Expr _test, Expr _message) {
        super(location);
        setTest(_test);
        setMessage(_message);
    }
    protected AssertStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        AssertStmt ret = new AssertStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        if (message != null) ret.setMessage( (Expr)walker.process(message) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return test;
        case 1: return message;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "test";
        case 1: return "message";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTest((Expr)child); return;
        case 1: setMessage((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "AssertStmt()";
    }

    //END: Generated from @child and @property
}


