/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;

/**
 *
 * @grammar lhs = rhs
 */
public class BasicAssignExpr extends AssignExpr {

    public Type discoverType() {
        Type lhsType = lhs.getType();
        Type rhsType = rhs.getType();
        return lhsType;
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        Type lhsType = lhs.getType();
        Type rhsType = rhs.getType();
        if (! rhs.isAssignableTo(lhsType)) {
            rhs.showTypeError(rhs.getType(), lhsType);
        }
        return this;
    }

    //INTRO from AccessFixer
    public ASTObject fixAccessToFieldSet(FieldAccessExpr expr) {
        return getAST().makeCall(expr.getField().getBackdoorSetterMethod(),
                                 expr.getExpr(), getRhs());
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    // this is big and dumb, but careful.
    public void walkFlow(FlowCheckerPass w) {
        if (getLhs() instanceof VarExpr) {
            w.process(getRhs());
            VarExpr lhs = (VarExpr)getLhs();
            VarDec dec = lhs.getVarDec();
            if (dec.isFinal()) {
                if (! dec.isBlank()) {
                    w.showVarError(this, dec, "Final variable " + dec.getId() +
                                   " already has a value");
                } else if (! w.getVars().isDefinitelyUnassigned(dec)) {
                    w.showVarError(this, dec, "Final variable " + dec.getId() +
                                   " may already have a value");
                } else {
                    w.addPossiblyAssigned(dec);
                }
            }
            w.setVars(w.getVars().addAssigned(dec));
        } else if (getLhs() instanceof FieldAccessExpr) {
            FieldAccessExpr lhs = (FieldAccessExpr) getLhs();
            FieldDec dec = lhs.getFieldDec();
            if (lhs.isBare()) {
                w.process(getRhs());
                if (dec.isFinal()) {
                    if (! dec.isBlank()) {
                        w.showVarError(this, dec, "Final field " + dec.getId() +
                                       " already has a value");
                    } else if (! w.getVars().isDefinitelyUnassigned(dec)) {
                        w.showVarError(this, dec, "Final field " + dec.getId() +
                                       " may already have a value");
                    } else {
                        w.addPossiblyAssigned(dec);
                        w.setVars(w.getVars().addAssigned(dec));
                    }
                }
            } else {
                w.process(lhs.getExpr());
                w.process(getRhs());
                if (dec.isFinal()
                    && ((! dec.isBlank())
                        || (! w.getVars().isDefinitelyUnassigned(dec)))) {
                    w.showVarError(this, dec, "Final field " + dec.getId() +
                                   " already has a value");
                }
            }
        } else {
            w.process(getLhs());
            w.process(getRhs());
        }
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        cgValueEffect(cb, true);
    }
    protected void cgEffect(CodeBuilder cb) {
        cgValueEffect(cb, false);
    }
    private void cgValueEffect(CodeBuilder cb, boolean needsValue) {
        AssignableExpr lhs = getLhs();
        Expr rhs = getRhs();
        lhs.cgLvalue(cb);
        rhs.cgValue(cb, lhs.getType());
        if (needsValue) lhs.cgDupRvalue(cb);
        lhs.cgAssignment(cb);
    }

    //BEGIN: Generated from @child and @property

    public BasicAssignExpr(SourceLocation location, AssignableExpr _lhs, String _op, Expr _rhs) {
        super(location, _lhs, _op, _rhs);

    }
    protected BasicAssignExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        BasicAssignExpr ret = new BasicAssignExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (lhs != null) ret.setLhs( (AssignableExpr)walker.process(lhs) );
        ret.op = op;
        if (rhs != null) ret.setRhs( (Expr)walker.process(rhs) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "BasicAssignExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
