/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.BlockScope;

import org.aspectj.compiler.base.bcg.CodeBuilder;

/**
  * @grammar { stmts* }
  * @child Stmts tmpStmts
  * @child Stmts stmts
  */
public class BlockStmt extends Stmt {

    public boolean isEmpty() {
        final int len = getStmts().size();
        for (int i = 0; i < len; i++) {
            if (! getStmts().get(i).isEmpty())
                return false;
        }
        return true;
    }
    
    public void requireBlockStmt() { ; }

    private static Stmt makeStmt(ASTObject node) {
        if (node instanceof Stmt) return (Stmt)node;
        if (node instanceof Expr) return new ExprStmt(node.getSourceLocation(),(Expr)node);
        throw new RuntimeException("node can not be made into a statement: "+node);
    }

    public BlockStmt(SourceLocation location, ASTObject node1) {
        this(location,node1 instanceof BlockStmt ? ((BlockStmt)node1).getStmts() :
             new Stmts(location,new Stmt[] {makeStmt(node1)}));
    }
    public BlockStmt(SourceLocation location, ASTObject node1, ASTObject node2) {
        this(location,new Stmts(location,new Stmt[] {makeStmt(node1), makeStmt(node2)}));
    }
    public BlockStmt(SourceLocation location, ASTObject node1, ASTObject node2, ASTObject node3) {
        this(location,new Stmts(location,new Stmt[] {makeStmt(node1), makeStmt(node2), makeStmt(node3)}));
    }
    public BlockStmt(SourceLocation location, ASTObject node1, ASTObject node2, ASTObject node3, ASTObject node4) {
        this(location,new Stmts(location,new Stmt[] {makeStmt(node1), makeStmt(node2), makeStmt(node3), makeStmt(node4)}));
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    /**
      * when the -lenient option is used,
      *   we declare that empty block don't have to be live.
      * This makes code like: return 2; {} legal.
      */
    public boolean mustBeLive() { return !getOptions().lenient; }

    public void walkFlow(FlowCheckerPass w) {
        Stmts s = getStmts();
        for (int i = 0, len = s.size(); i < len; i++) {
            Stmt st = s.get(i);
            if (! w.isLive() && st.mustBeLive()) {
                st.showError("unreachable statement");
                break;
            } else {
                w.process(s.get(i));
            }
        }
    }

    // ------------------------------
    // Intro: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        makeTempsExplicit();
        getStmts().walkCleanup(w);
    }

    public ASTObject postCleanup(ByteCodeCleanupPass w) {
        if (getStmts().size() == 0) {
            return getAST().makeEmptyStmt().setSource(this);
        } else {
            return this;
        }
    }

    protected void makeTempsExplicit() {
        if (tmpStmts != null) {
            for (int i = 0, len = tmpStmts.size(); i < len; i++) {
                getStmts().add(0, tmpStmts.get(i));
            }
            tmpStmts = null;
        }
    }

    // ------------------------------
    // Intro:  FrameLocPass

    public void walkFrameLoc(FrameLocPass walker) {
        int start = walker.getfs();
        super.walkFrameLoc(walker);
        walker.setfs(start);
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) { walker.pushBlock(); }
    public ASTObject postScope(ScopeWalker walker) { walker.popBlock(); return this; }

    //XXX this counter could be trouble
    //int temporaries = 0;
    //Stmts tmpStmts; // = new Stmts();
    public VarDec makeTemporary(Type type) {
//        String id = "ajc$tmp$" + ++temporaries;
//
//        if (getDeclaringType() != null && getDeclaringType().isAnonymous()) {
//            id += "x";
//        }

        VarDec tmp = getAST().makeVarDec(type, "ajc$tmp", null);
        addTemporary(tmp);
        tmpStmts.add(0, tmp);
        return tmp;
    }

    public void addTemporary(VarDec tmp) {
        if (getTmpStmts() == null) setTmpStmts(getAST().makeStmts()); //new Stmts(getCompiler());
        tmp.setId(tmp.getId() + tmpStmts.size());
        tmpStmts.add(0, tmp);
    }

    public void checkReturnType(Type expectedReturnType) {
        ReturnTypeChecker rtc = new ReturnTypeChecker(getCompiler(),expectedReturnType);
        rtc.process(stmts);
    }

    //!!! way too simple
    static class ReturnTypeChecker extends Walker {
        Type expectedReturnType;

        ReturnTypeChecker(JavaCompiler compiler, Type expectedReturnType) {
            super(compiler);
            this.expectedReturnType = expectedReturnType;
        }

        public boolean preProcess(ASTObject node) {
            if (node instanceof ReturnStmt) {
                Expr returnExpr = ((ReturnStmt)node).getExpr();
                if (returnExpr == null) {
                    if (!expectedReturnType.isEquivalent(getTypeManager().voidType)) {
                        node.showError(expectedReturnType.toShortString()+" return expected");
                    }
                } else {
                    if (!expectedReturnType.isAssignableFrom(returnExpr.getType())) {
                        node.showError(expectedReturnType.toShortString()+" return expected");
                    }
                }
                return true;
            }
            return !(node instanceof TypeDec);
        }
    }

    public void unparse(CodeWriter writer) {
        writer.openBlock();
        if (tmpStmts != null) writer.write(tmpStmts);
        writer.write(stmts);
        writer.closeBlock();
    }

    public void cleanup() {
        stmts = null;
        tmpStmts = null;
        super.cleanup();
    }

    // ------------------------------
    // bcg
    protected void cgStmt(CodeBuilder cb) {
        cb.enterBlock();
        stmts.cgStmts(cb);
        cb.exitBlock();
    }

    // ------------------------------
    // convenience constructors

    public BlockStmt(SourceLocation location, Stmts _stmts) {
        this(location, null, _stmts);
    }

    //BEGIN: Generated from @child and @property
    protected Stmts tmpStmts;
    public Stmts getTmpStmts() { return tmpStmts; }
    public void setTmpStmts(Stmts _tmpStmts) {
        if (_tmpStmts != null) _tmpStmts.setParent(this);
        tmpStmts = _tmpStmts;
    }

    protected Stmts stmts;
    public Stmts getStmts() { return stmts; }
    public void setStmts(Stmts _stmts) {
        if (_stmts != null) _stmts.setParent(this);
        stmts = _stmts;
    }

    public BlockStmt(SourceLocation location, Stmts _tmpStmts, Stmts _stmts) {
        super(location);
        setTmpStmts(_tmpStmts);
        setStmts(_stmts);
    }
    protected BlockStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        BlockStmt ret = new BlockStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (tmpStmts != null) ret.setTmpStmts( (Stmts)walker.process(tmpStmts) );
        if (stmts != null) ret.setStmts( (Stmts)walker.process(stmts) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return tmpStmts;
        case 1: return stmts;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "tmpStmts";
        case 1: return "stmts";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTmpStmts((Stmts)child); return;
        case 1: setStmts((Stmts)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "BlockStmt()";
    }

    //END: Generated from @child and @property
}


