/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;
import org.aspectj.compiler.crosscuts.joinpoints.CalleeSideCallJp;

import org.aspectj.compiler.base.cst.BlockScope;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

import java.util.*;


/**
  * @grammar expr.id(args)
  * @child Expr expr
  * @property String id
  * @child Exprs args
  * @property Method method
  * @property boolean isSuper
  */
public class CallExpr extends AnyCallExpr implements SOLink {

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        if (getExpr() != null) {
            w.process(getExpr());
        }
        w.process(getArgs());
        if (getMethod().getThrows() != null) {
            TypeDs ts = getMethod().getThrows();
            for (int i = 0, len = ts.size(); i < len; i++) {
                w.setExns(w.getExns().add((NameType) ts.get(i).getType()));
            }
        }
    }

    //INTRO from MovingWalker
    public ASTObject postMove(MovingWalker walker) {
        return walker.moveLinkExpr(this);
    }

    //INTRO from ScopeWalker
    public ASTObject postScope(ScopeWalker walker) {
        if (method != null)
            return this;

        Type lookupType;
        if (expr != null) {
            lookupType = expr.getType();
        } else {
            lookupType = walker.getScope().findMethodLookupType(id, this);
            if (lookupType == null) {
                this.showError("can't find any method with name: " + id);
                return this;
            }
        }

        method = lookupType.getMethod(id, this, args, true);
        //!!! do the fill a la typeScope
        if (method == null) {
            setType(getTypeManager().TYPE_NOT_FOUND);
            return this;
        }

        if (expr != null && expr instanceof SuperExpr) {
            if (method.isStatic()) {
            } else {
                setIsSuper(true);
            }
        }
          
        if (getExpr() == null) {
            setExpr(
                (method.isStatic())
                    ? getAST().makeTypeExpr(lookupType)
                    : (Expr) getAST().makeQualifiedThis(lookupType));
        }

        //super.bindNamesSelf(scope);
        return super.postScope(walker);
    }

    // ------------------------------
    // INTRO from InnerInfoPass -- should have a more generic name

    public void postInnerInfo(InnerInfoPass w) {
        if (expr == null && method != null) {
            w.checkStaticAccess(this, method);
            setExpr(getAST().makePrimary(getMethod(), w.currentType()));
        }
    }

    /**
     * Section 6.6.2.1 from JLSv2
     * E.Id requires that E is subtype of S
     */
    public boolean hasLegalProtectedAccess(Type fromType) {
        if (expr == null || getIsSuper() || expr instanceof SuperExpr)
            return true;
        return expr.getType().isSubtypeOf(fromType);
    }

    public void checkSpec() {
        if (method == null)
            return;
        if (expr instanceof TypeExpr && !method.isStatic()) {
            showError(
                "non-static method "
                    + method.toShortString()
                    + " cannot be accessed through a static reference");
        }
        if (getIsSuper()) {
            if (method.isAbstract()) {
                showError(
                    "abstract method " + method.toShortString() + " cannot be target of super");
            }
        }
    }

    public boolean isLegalStmt() {
        return true;
    }

    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (getIsSuper()) {
            MethodDec superMethodDec = getMethodDec();
            CalleeSideCallJp targetPoint =
                (CalleeSideCallJp) getWorld().calleeSideCallPoints.get(superMethodDec);
            if (targetPoint == null || !targetPoint.hasPlans()) {
                return this;
            }

            setMethod(targetPoint.getPostMethodDec().getMethod());
        }
        // handle introduced statics on interfaces
        setExpr(method.updateTargetExpr(getExpr()));
        return this;
    }

    //INTRO from AccessFixer
    public ASTObject fixAccessPost(AccessFixer fixer) {
        if (getMethod() == null)
            return this; //XXX should be error
        //TypeDec fromTypeDec = getBytecodeTypeDec(); //fixer.fromTypeDec;
        if (getMethod().isAccessible(this, true))
            return this;

        getCompiler().showMessage(
            "  fixing privileged call: " + getMethod().toShortString());
        //this.getParent().getParent().display(2);

        Method newMethod = getMethod().getBackdoorMethod();
        if (!fixer.apply)
            return this;

        CallExpr newCallExpr = getAST().makeCall(newMethod, getExpr(), getArgs());
        newCallExpr.setSource(this);
        return newCallExpr;
    }

    public MethodDec getMethodDec() {
        return (MethodDec) method.getCorrespondingDec();
    }
    public void setMethodDec(MethodDec md) {
        setMethod((Method) md.getCorrespondingSemanticObject());
    }
    public CodeDec getCodeDec() {
        return getMethodDec();
    }
    public SemanticObject getTarget() {
        return getMethod();
    }
    public void setTarget(SemanticObject newTarget) {
        setMethod((Method) newTarget);
    }

    public Type getReturnType() {
        return getMethod().getReturnType();
    }

    public Type getCalledType() {
        if (expr == null) {
            return getDeclaringType();
        } else {
            return expr.getType();
        }
    }

    public CallExpr(SourceLocation source, Expr expr, String id) {
        this(source, expr, id, new Exprs(source));
    }

    public Type discoverType() {
        if (getMethod() == null)
            return getTypeManager().anyType;
        else
            return getMethod().getReturnType();
    }

    static Expr makeParentExpr(Method md, Expr instance) {
        if (instance != null)
            return instance;
        if (md.isStatic()) {
            return md.getAST().makeTypeExpr(md.getDeclaringType());
        } else {
            return null; //return new ThisExpr(md.getDeclaringTypeDec().getNameTypeD());
        }
    }

    public Expr getExprOrThis() {
        if (method.isStatic())
            return null;
        //new TypeExpr(getCompiler(), getMethodDec().getDeclaringType().makeTypeD());
        if (expr != null)
            return expr;
        return new ThisExpr(getSourceLocation(), method.getDeclaringType());
    }
    
    private boolean isInlinableTarget(Expr e, Type bytecodeType) {
        if (e == null) return true;
        if (e instanceof TypeExpr) return true;
        if (e instanceof ThisExpr) {
            ThisExpr thisExpr = (ThisExpr)e;
            if (thisExpr.getType() == bytecodeType) return true;
        }
        return false;
    }
    
    public boolean tryToInline(Type bytecodeType) {
        //System.out.println(this.unparse() + ", " + this.getParent());
        //if (getArgs() != null && getArgs().size() != 0) return false;
        if (!isInlinableTarget(getExpr(), bytecodeType)) return false;
        
        
        
        MethodDec md = getMethodDec();
	if (md.containsTypes()) return false;

        final AST ast = getAST();
           
        try {
            if (getParent() instanceof ReturnStmt) {
                //System.err.println("GOT HERE " + getParent().unparse());
                RemapWalker rw = new RemapWalker(getCompiler(), args, md, null, null);
                
                getParent().replaceWith(rw.remap());
            } else if (getParent() instanceof ExprStmt) {
                //System.err.println("GOT HERE1 " + getParent().unparse());
                String label = getWorld().genLabel();
                RemapWalker rw = new RemapWalker(getCompiler(), args, md, label, null);
                getParent().replaceWith(rw.remap());           
            } else if (isEffectiveCast(getParent())) {
                ReturnStmt s = getEnclosingReturnStmt(this);
                RemapWalker rw = new RemapWalker(getCompiler(), args, md, null, s.getExpr().getType());
                s.replaceWith(rw.remap());
            } else {
                return false;
            }
        } catch (InvalidOptimizationException e) {
            return false;
        }
        
        md.getDeclaringType().getTypeDec().getBody().remove(md);
        return true; 
    }
    
    private ReturnStmt getEnclosingReturnStmt(ASTObject o) {
        if (o instanceof ReturnStmt) return (ReturnStmt) o;
        else return getEnclosingReturnStmt(o.getParent());
    }
    
    
    /** is o a possible cast inside a return statement */
    private boolean isEffectiveCast(ASTObject o) {
	if (o instanceof ParenExpr) return isEffectiveCast(o.getParent());
	if (o instanceof CastExpr) return isEffectiveCast1(o.getParent());
        else return false;
    }
    
    /** seen a cast */
    private boolean isEffectiveCast1(ASTObject o) {
        if (o instanceof ParenExpr) return isEffectiveCast1(o.getParent());
        if (o instanceof ReturnStmt) return true;
        else return false;
    }    
    
    static class InvalidOptimizationException extends RuntimeException {}
    
    class RemapWalker extends MovingWalker {
        private Exprs args;
        private MethodDec calledMethodDec;
        private String breakLabel;
        
        private Type desiredReturnType;
        
        // This is a list of Exprs that either are references or 
        // are ReferenceExprs that we will have squirrled away the appropriate
        // VarDec for
        private List/*Exprs*/ argRefs;
        private VarDec returnVarDec;
        
        public RemapWalker(JavaCompiler compiler, Exprs args, MethodDec calledMethodDec, 
        			String breakLabel,
        			Type desiredReturnType) {
            super(compiler);
            this.args = args;
            this.calledMethodDec = calledMethodDec;
            this.breakLabel = breakLabel;
            this.desiredReturnType = desiredReturnType;
        }
        
        
        public Stmt remap() {
            final AST ast = this.getAST();
            
//            if (!methodDec.getResultType().isVoid()) {
//                returnVarDec =
//                    ast.makeVarDec(methodDec.getResultType(), "returnValue", null, true);
//            }

            Stmts stmts = this.remapStmts();
            //if (returnVarDec != null) stmts.add(0, returnVarDec);
            Stmt ret = ast.makeBlock(stmts);
            //System.out.println(ret);
            //System.out.println(ret.unparse());
            if (breakLabel != null) ret = ast.makeLabeled(breakLabel, ret);
            return ret;
        }
        
        public Stmts remapStmts() {
            final AST ast = this.getAST();
            Stmts block = ast.makeStmts();
            //System.out.println(block + ": " + System.identityHashCode(block));
            this.argRefs = new ArrayList();           
            
            Set effectivelyFinalFormals = calledMethodDec.getEffectivelyFinalFormals();
            
            for (int i=0; i<args.size(); i++) {
                FormalDec formal = calledMethodDec.getFormals().get(i);
                Type type = formal.getType();
                
                Expr arg = (Expr)args.get(i).copy();
                
                if (effectivelyFinalFormals.contains(formal) && args.get(i).canBeCopied()) {
                    argRefs.add(ast.makeCast(type, arg));
                } else {
                    VarDec dec = ast.makeVarDec(type, formal.getId(), arg);
                    block.add(dec);
                    argRefs.add(ast.makeVar(dec));
                }
            }
            
            Stmts body = (Stmts)calledMethodDec.getBody().getStmts().copy();
            //body.checkNoSharing(methodDec);
            //System.out.println(body + ": " + System.identityHashCode(body));
            //System.out.println(methodDec.getBody().getStmts() + ": " + System.identityHashCode(methodDec.getBody().getStmts()));
            
            //try {
                body = (Stmts)process(body);
//            } finally {
//                body.checkNoSharing(methodDec);
//                block.checkNoSharing(methodDec);
//            }
            block.addAll(body); //methodDec.getBody().getStmts());
            return block;
        }
        
        protected Expr handleFreeVar(VarExpr var) {
            VarDec varDec = var.getVarDec();
            if (!(varDec instanceof FormalDec)) {
                return super.handleFreeVar(var);
            }
            
            FormalDec formalDec = (FormalDec)varDec;
            int index = calledMethodDec.getFormals().indexOf(formalDec);
            if (index == -1) return super.handleFreeVar(var);
            
            
            //note: makeReference copies for non-ReferenceExprs
            //      and makes the apprpriate VarExpr for ReferenceExprs
            return (Expr)((Expr)argRefs.get(index)).copy();
        }

        private ASTObject handleReturn(ReturnStmt object) {
            final AST ast = this.getAST();  
            // are we in a return context, not an exprStmt context  ?      
            if (breakLabel == null) { 
                // do we care about the return type? If yes, there's
 		// a cast involved (we came from a 
 		// return (TYPE) [here]; context.
 		if (desiredReturnType != null) {  
 		    // is the cast to our desired type is guaranteed to succeed?
 		    if (desiredReturnType.isAssignableFrom(object.getExpr().getType())) { 	        			  
 		    	object.setExpr(ast.makeCast(desiredReturnType, object.getExpr()));
 		    } else {
 		        // otherwise, we came from... 
 		        throw new InvalidOptimizationException();
 		    }
 		}
 		return object;
            }
        
            Stmt stmt = ast.makeBreak(breakLabel);
            
            Expr expr = object.getExpr();
            if (expr == null) return stmt;
            if (expr.canBeCopied()) return stmt;
            if (expr.isLegalStmt()) return ast.makeBlock(ast.makeStmt(expr), stmt);
            
            return ast.makeVarDec(calledMethodDec.getResultType(), "retValue", expr, true);
        }
//            
//            
//            if (returnVarDec != null) {
//                stmt = ast.makeBlock(
//                    ast.makeSet(returnVarDec, object.getExpr()),
//                    stmt);
//            }
//            
//            return stmt;
//        }
//        
        
        public Expr moveThisExpr(ThisExpr thisExpr, Type thisType) {
            return thisExpr;
        }
        
        public ASTObject process(ASTObject object) {
            //System.out.println(object);
            if (object instanceof TypeDec || object instanceof CodeDec) {
                return object;
            }
            
            object = super.process(object);
            if (object instanceof ReturnStmt) {
                return handleReturn((ReturnStmt)object);
            } else {
                return object;
            }
        }
	

    }
    
    public void unparse(CodeWriter writer) {
        if (expr != null) {
            writer.write(expr);
            writer.write('.');
        }

        if (method != null)
            id = method.getBytecodeId();
        writer.write(id);
        writer.openParen('(');
        writer.write(args);
        writer.closeParen(')');
    }

    public CallExpr(
        SourceLocation source,
        Method toMethod,
        Expr instance,
        Exprs args) {
        this(
            source,
            makeParentExpr(toMethod, instance),
            toMethod.getId(),
            args,
            toMethod,
            false);
    }

    public CallExpr(SourceLocation source, String _id, Exprs _args) {
        this(source, new ThisExpr(source), _id, _args);
    }

    public CallExpr(SourceLocation source, Expr expr, String _id, Exprs _args) {
        this(source, expr, _id, _args, null, false);
    }

    private boolean needsAccessMethod(InnerAccessFixer w) {
        Expr q = getExpr();
        MethodDec oldMethodDec = getMethodDec();

        // there are no QualifiedThisExprs by this point
        //XXX reformat later
        // we believe that CallExpr should have a property that encodes whether or
        // not it is a special super access call generated below
		if (!w.isAccessible(oldMethodDec, q)) {
			return true;
		} else {
			if (getIsSuper()) {
				//System.err.println(unparse());
				if (getExpr() instanceof ThisExpr || getExpr() instanceof SuperExpr) {
					return false;
				} else {
					return true;
				}
//				if (getIsSuperAccessMethodCall()) {
//					return false;
//				} else {
//					return true;
//				}
			} else {
				return false; 
			}
		}
    }

    // ------------------------------
    //INTRO from InnerAccessFixer

    public ASTObject postInnerAccess(InnerAccessFixer w) {
        // 		if (getIsSuper()) {
        // 			System.err.println("CEbefore: ");
        // 			this.display(0);
        // 		}

        if (!needsAccessMethod(w))
            return this;
        //        System.err.println("CE: " + this);

        Expr q = getExpr();
        MethodDec oldMethodDec = getMethodDec();

        Type qType = q.getType();
        MethodDec newMethodDec = w.getAccessMethod(qType, oldMethodDec, "", this);
        setMethod(newMethodDec.getMethod());
        if (!oldMethodDec.isStatic()) {
            getArgs().add(0, q);
            setExpr(getAST().makeTypeExpr(qType));
        }
        return this;
    }

    public MethodDec buildAccessMethod(InnerAccessFixer w) {
        final AST ast = getAST();
        Expr q = getExpr();
        Type qType = q.getType();

        Type returnType = getReturnType();
        Method oldMethod = getMethod();

        Formals newFormals = (Formals) oldMethod.getFormals().copy();
        Exprs newArgs = ast.makeVars(newFormals);
        Expr newExpr = w.makeInsidePrimary(oldMethod.isStatic(), newFormals, qType);

        CallExpr newCall = ast.makeCall(oldMethod, newExpr, newArgs);
        if (getIsSuper())
            newCall.setIsSuper(true);

        return w.makeAccessMethod(returnType, newFormals, newCall);
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        cb.enterLocation(getSourceLocation());
        MethodDec dec = getMethodDec();
        String name = dec.getInternalId();
        String descriptor = dec.getDescriptor();
        int delta = dec.getStackDelta();
        if (dec.isStatic())
            getExpr().cgEffect(cb);
        else
            getExpr().cgValue(cb);
        getArgs().cgValues(cb, dec.getFormals());
        RefType refTy = (RefType) getExpr().getType();
        NameType ty =
            (refTy instanceof ArrayType)
                ? getTypeManager().getObjectType()
                : (NameType) refTy;
        if (dec.getDeclaringType().isObject()) {
            // this special cases the line in JLS 13.1: "Given a
            // method invocation expression ... referencing a method
            // named m declared in a class or interface D, we define
            // the qualifying type of the method invocation as
            // follows: If D is Object then the qualifying type of the
            // expression is Object.  Otherwise..."
            ty = getTypeManager().getObjectType();
        }
        if (dec.isStatic()) {
            cb.emitINVOKESTATIC(ty, name, descriptor, delta);
        } else if (getIsSuper()) {
            cb.emitINVOKESPECIAL(ty, name, descriptor, delta);
        } else if (dec.isPrivate()) {
            cb.emitINVOKESPECIAL(ty, name, descriptor, delta);
        } else if (ty.isInterface()) {
            int arglen = -delta + dec.getResultType().getSlotCount();
            cb.emitINVOKEINTERFACE(ty, name, descriptor, delta, arglen);
        } else {
            cb.emitINVOKEVIRTUAL(ty, name, descriptor, delta);
        }
    }    
    
    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }
    
    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }
    
    protected Exprs args;
    public Exprs getArgs() { return args; }
    public void setArgs(Exprs _args) {
        if (_args != null) _args.setParent(this);
        args = _args;
    }
    
    protected Method method;
    public Method getMethod() { return method; }
    public void setMethod(Method _method) { method = _method; }
    
    protected boolean isSuper;
    public boolean getIsSuper() { return isSuper; }
    public void setIsSuper(boolean _isSuper) { isSuper = _isSuper; }
    
    public CallExpr(SourceLocation location, Expr _expr, String _id, Exprs _args, Method _method, boolean _isSuper) {
        super(location);
        setExpr(_expr);
        setId(_id);
        setArgs(_args);
        setMethod(_method);
        setIsSuper(_isSuper);
    }
    protected CallExpr(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        CallExpr ret = new CallExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        ret.id = id;
        if (args != null) ret.setArgs( (Exprs)walker.process(args) );
        ret.method = method;
        ret.isSuper = isSuper;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        case 1: return args;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        case 1: return "args";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        case 1: setArgs((Exprs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "CallExpr(id: "+id+", "+"method: "+method+", "+"isSuper: "+isSuper+")";
    }
    
    //END: Generated from @child and @property
}
