/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar (typeD)expr
  * @child TypeD typeD
  * @child Expr expr
  */

public class CastExpr extends Expr {

    public boolean canBeCopied() {
        return expr.isUltimatelyLiteral();
    }
    
    public boolean isUltimatelyLiteral() {
        return expr.isUltimatelyLiteral();
    }


    public boolean isPossibleCast() {
        if (expr.getType() == null || typeD.getType() == null) return true;
        return expr.getType().isCoercableTo(typeD.getType());
    }

    //INTRO from MovingWalker
    /*
    public ASTObject postMove(MovingWalker walker) {
        //XXX this might simplify too much
        //XXX this is needed because some casts can become illegal when moved
        //XXX (only illegal for Java source code, not bytecode)
        if (walker.hasToType() && !isPossibleCast()) {
            //XXX this should ensure that an exception is throw
            return getAST().forceCast(getType(), getAST().makeNull());
        }
        return this;
    }
    */

    public void checkSpec() {
        Type exprType = expr.getType();
        Type castType = typeD.getType();
        setType(castType);
        if (!isPossibleCast()) {
            showError("Can't convert from " + exprType.getString()
                      + " to " + castType.getString());
        }
    }

    public Type discoverType() {
        return typeD.getType();
    }

    public void unparse(CodeWriter writer) {
        writer.write('(');
        writer.write(typeD);
        writer.write(')');

        if (isSynthetic() && !isPossibleCast()) {
            //XXX this should ensure that an exception is throw
            writer.write("null");
            return;
        }

        //!!! this should be done in a cleaner and more general way
        if (isSynthetic() && !(expr instanceof ParenExpr)) {
            writer.write('(');
            writer.write(expr);
            writer.write(')');
        } else {
            writer.write(expr);
        }
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        expr.cgValue(cb);
        expr.getType().emitCast(cb, typeD.getType());
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        checkSpec();
        if (expr instanceof LiteralExpr && ! getTypeD().getType().isObject()) {
            return getTypeD().getType().foldCast((LiteralExpr) expr);
        } else {
            return this;
        }
    }

    //BEGIN: Generated from @child and @property
    protected TypeD typeD;
    public TypeD getTypeD() { return typeD; }
    public void setTypeD(TypeD _typeD) {
        if (_typeD != null) _typeD.setParent(this);
        typeD = _typeD;
    }

    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }

    public CastExpr(SourceLocation location, TypeD _typeD, Expr _expr) {
        super(location);
        setTypeD(_typeD);
        setExpr(_expr);
    }
    protected CastExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        CastExpr ret = new CastExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (typeD != null) ret.setTypeD( (TypeD)walker.process(typeD) );
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return typeD;
        case 1: return expr;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "typeD";
        case 1: return "expr";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTypeD((TypeD)child); return;
        case 1: setExpr((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "CastExpr()";
    }

    //END: Generated from @child and @property
}
