/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;


import java.util.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

/**
 * @grammar ...
 * @child TypeD superClass @before superInterfaces
 */

public class ClassDec extends TypeDec {
    public String getKind() { return "class"; }

    private boolean isAnonymousFlag = false;
    public boolean isAnonymous() { return isAnonymousFlag || getId().equals("ANONYMOUS"); }
    public void setIsAnonymous() { isAnonymousFlag = true; }

    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
        this.isAnonymousFlag = ((ClassDec)oldObject).isAnonymousFlag;
        this.isInnerFlag = ((ClassDec)oldObject).isInnerFlag;
        // !!!If moved to a static context, then this class can no
        // longer be inner.  These are ramifications to this that need to be explored
        if (walker.isStatic) this.isInnerFlag = false;
        return super.postCopy(walker, oldObject);
    }

    protected void walkExtendsAndImplements(ScopeWalker walker) {
        if (superClass != null) {
            setSuperClass((TypeD)walker.process(superClass));
        }
        //System.out.println(superClass);
        super.walkExtendsAndImplements(walker);
    }


    public Type getSuperClassType() {
        Type superType;
        if (superClass == null) superType = getTypeManager().getObjectType();
        else superType = superClass.getType();

        //XXX bad modularity putting this here
        if (!type.isAspect() && superType.isAspect()) {
            superClass.showError("class may not extend an aspect");
        }

        if (superType.isClass()) return superType;

        //XXX bad design
        if (isAnonymous()) {
            setSuperInterfaces(getAST().makeTypeDs(superClass));
            superType = getTypeManager().getObjectType();
            setSuperClass(superType.makeTypeD());
        } else {
            superClass.showError(toShortString() + " must extend a class, not " + superType.toShortString());
            return getTypeManager().TYPE_NOT_FOUND;
        }
        return superType;
    }


    public void unparse(CodeWriter writer) {
        writer.write(modifiers);
        writer.writeKeyword("class");
        writer.requiredSpace();
        writer.write(id);

        if (superClass != null) {
            writer.requiredSpace();
            writer.write("extends");
            writer.requiredSpace();
            writer.write(superClass);
        }

        writeNames(writer, "implements", superInterfaces);

        if (!writer.isOnlySignatures()) {
            writer.optionalSpace();
            writer.openBlock();
            writer.write(body);
            writer.newLine();
            writer.closeBlock();
        }
    }


    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
        setIsInner(! isStatic() && w.isInDynamicContext());
        super.walkInnerInfo(w);
    }

    // ------------------------------
    // Inner class stuff

    // Only classDecs can be inner (not interfaces, not aspects).

    private boolean isInnerFlag = false;
    public void setIsInner(boolean b) { isInnerFlag = b; }
    public boolean isInner() { return isInnerFlag; }

    //BEGIN: Generated from @child and @property
    protected TypeD superClass;
    public TypeD getSuperClass() { return superClass; }
    public void setSuperClass(TypeD _superClass) {
        if (_superClass != null) _superClass.setParent(this);
        superClass = _superClass;
    }

    public ClassDec(SourceLocation location, Modifiers _modifiers, String _id, TypeD _superClass, TypeDs _superInterfaces, Decs _body) {
        super(location, _modifiers, _id, _superInterfaces, _body);
        setSuperClass(_superClass);
    }
    protected ClassDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ClassDec ret = new ClassDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        ret.id = id;
        if (superClass != null) ret.setSuperClass( (TypeD)walker.process(superClass) );
        if (superInterfaces != null) ret.setSuperInterfaces( (TypeDs)walker.process(superInterfaces) );
        if (body != null) ret.setBody( (Decs)walker.process(body) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return superClass;
        case 2: return superInterfaces;
        case 3: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "superClass";
        case 2: return "superInterfaces";
        case 3: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setSuperClass((TypeD)child); return;
        case 2: setSuperInterfaces((TypeDs)child); return;
        case 3: setBody((Decs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 4;
    }

    public String getDefaultDisplayName() {
        return "ClassDec(id: "+id+")";
    }

    //END: Generated from @child and @property
}
