/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.parser.SourceInfo;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.CUScope;

import java.util.*;
import java.io.File;
import java.io.IOException;

/**
  * @grammar package imports decs
  * @property String packageName
  * @child Imports imports
  * @child Decs decs
  * @property SourceInfo sourceInfo
  * @property boolean scanned
  */
public class CompilationUnit extends ASTObject {

    private CUScope scope = new CUScope(getCompiler(), null, this);

    private ArrayList   ownDecs = new ArrayList();

//    private class CUSourceLocation extends SourceLocation {
//        private JavaCompiler compiler;
//        public CUSourceLocation(JavaCompiler compiler) {
//            super(null, 0, 0);
//            this.compiler = compiler;
//        }
//
//        public CompilationUnit getCompilationUnit() { return CompilationUnit.this; }
//        public JavaCompiler getCompiler() { return compiler; }
//    }

    // An empty compilation unit constructor
    public CompilationUnit(JavaCompiler compiler, File _file) {
        this(new DummySourceLocation(compiler), null, null, null, new SourceInfo(_file, null), false);
        sourceLocation = new TextSourceLocation(this, 0, 0);
        setDecs(new Decs(getSourceLocation()));
    }

    public List getDefinedTypes() {
        return ownDecs;
    }

    public void addDefinedType(TypeDec dec) {
        //System.out.println("adding " + dec + " to " + this);
        ownDecs.add(dec);
    }


    public Type getDeclaringType() {
        return null;
    }
    public Type getLexicalType() {
        return null;
    }

    public TypeDec getBytecodeTypeDec() {
        return null;
    }

    public CodeDec getEnclosingCodeDec() {
        return null;
    }

    public CUScope getScope() {
        return scope;
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) { walker.enterCU(this); }
    //???public ASTObject postScope(ScopeWalker walker) { walker.popBlock(); return this; }

//
//    class ImplementationCleanupWalker extends ASTWalker {
//        ImplementationCleanupWalker(JavaCompiler compiler) {
//            super(compiler);
//        }
//        public ASTObject walk0(ASTObject obj) {
//            if (obj == null)
//                return obj;
//            obj.clearComment();
//            if (obj instanceof CodeDec) {
//                obj.cleanup();
//            }
//            else if (obj instanceof CodeBody) {
//                obj.cleanup();
//                return obj;
//            }
//            else if (obj instanceof ArrayInitializer) {
//                obj.cleanup();
//                return obj;
//            }
//            super.walk1(obj);
//            return obj;
//        }
//    }
//
    public String toString() { return getSourceFile().toString(); }

    // cleanup implementations
    public void cleanup() {
//        ImplementationCleanupWalker w = new ImplementationCleanupWalker(getCompiler());
//        this.walk(w);
    }
    public void cleanupDefinedTypes() {
        ownDecs = new ArrayList();
    }

    private String sourceCanonicalPath = null;
    public String getSourceCanonicalPath() {
        if (sourceCanonicalPath != null) return sourceCanonicalPath;
        File file = getSourceFile();
        if (file == null) return null;

        try {
            sourceCanonicalPath = file.getCanonicalPath();
        } catch (IOException ioe) {}

        return sourceCanonicalPath;
    }

    private String sourceDirectory = null;
    public String getSourceDirectory() {
        if (sourceDirectory != null) return sourceDirectory;

        sourceDirectory = new File(getSourceCanonicalPath()).getParent();

        return sourceDirectory;
    }


    public File getSourceFile() { return sourceInfo.getFile(); }
    public int getLine(int position) { if (sourceInfo == null) return -1; return sourceInfo.getLine(position); }
    public int getColumn(int position) { return sourceInfo.getColumn(position); }

    public CompilationUnit getCompilationUnit() {
        return this;
    }

    public void addTypesToTypeGraph() {
        ScopeWalker walker = new ScopeWalker(getCompiler(), getScope());
        Decs decs = getDecs();
        for (int i=0, N=decs.size(); i<N; i++) {
            TypeDec typeDec = (TypeDec)decs.get(i);
            typeDec.addToTypeGraph(walker);
        }
    }

    public void buildSignatures() {
        ScopeWalker walker = new ScopeWalker(getCompiler(), getScope());
        Decs decs = getDecs();
        for (int i=0, N=decs.size(); i<N; i++) {
            TypeDec typeDec = (TypeDec)decs.get(i);
            typeDec.buildSignatures(walker);
        }
    }


    public void checkSpec() {
        Imports imports = getImports();
        Map importedTypeNames = new HashMap();
        for (int i=0,N=imports.size(); i<N; i++) {
            Import imp = imports.get(i);
            if (!imp.getStar()) {
                Type t = imp.getType();
                // check to see if there's a type in this cu
    	        Decs decs = getDecs();
                for (int j=0, M=decs.size(); j<M; j++) {
                    TypeDec typeDec = (TypeDec)decs.get(j);
                    String id = typeDec.getId();
                    if (t.getId().equals(id) && !(t == typeDec.getType())) {
                        imp.showError("import conflicts with definition of " +
                                       typeDec.getPrettyString());
                    }
                }                
                Type iType = (Type)importedTypeNames.get(t.getId());
                if (iType != null && iType != t) {
                    imp.showError("import conflicts with previous import of " +
                                  iType.getPrettyString());
                }
                importedTypeNames.put(t.getId(), t);
            }
        }
        
        
        // in strict mode, check that public types are in source files of same name
        if (getOptions().strict && sourceInfo != null) {
	    Decs decs = getDecs();
            for (int i=0, N=decs.size(); i<N; i++) {
                TypeDec typeDec = (TypeDec)decs.get(i);
                if (typeDec.isPublic()) {
                    String id = typeDec.getId();
                    String filename = sourceInfo.getFile().getName();
                    //XXX lazy string matching
                    if (!filename.startsWith(id)) {
                        this.showError("public class must be in file of same name, i.e " +
                                       id + ".java");
                    }
                }
            }
        }        
    }


//    public String getPackageName() {
//        return _package == null ? null : _package.getName();
//    }

    public void unparse(CodeWriter writer) {
        if (getPackageName() != null) {
            writer.writeKeyword("package");
            writer.requiredSpace();
            writer.write(getPackageName());
            writer.closeStmt();
        }

        writer.writeChildren(imports);
        writer.write(decs);
    }

    // ------------------------------
    // bcg
    public final void generateBytecode(File outputDir) throws IOException {
        for (Iterator i = getDefinedTypes().iterator(); i.hasNext(); ) {
            TypeDec typeDec = (TypeDec) i.next();
            typeDec.generateBytecode(outputDir);
        }
    }

    //BEGIN: Generated from @child and @property
    protected String packageName;
    public String getPackageName() { return packageName; }
    public void setPackageName(String _packageName) { packageName = _packageName; }

    protected Imports imports;
    public Imports getImports() { return imports; }
    public void setImports(Imports _imports) {
        if (_imports != null) _imports.setParent(this);
        imports = _imports;
    }

    protected Decs decs;
    public Decs getDecs() { return decs; }
    public void setDecs(Decs _decs) {
        if (_decs != null) _decs.setParent(this);
        decs = _decs;
    }

    protected SourceInfo sourceInfo;
    public SourceInfo getSourceInfo() { return sourceInfo; }
    public void setSourceInfo(SourceInfo _sourceInfo) { sourceInfo = _sourceInfo; }

    protected boolean scanned;
    public boolean getScanned() { return scanned; }
    public void setScanned(boolean _scanned) { scanned = _scanned; }

    public CompilationUnit(SourceLocation location, String _packageName, Imports _imports, Decs _decs, SourceInfo _sourceInfo, boolean _scanned) {
        super(location);
        setPackageName(_packageName);
        setImports(_imports);
        setDecs(_decs);
        setSourceInfo(_sourceInfo);
        setScanned(_scanned);
    }
    protected CompilationUnit(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        CompilationUnit ret = new CompilationUnit(getSourceLocation());
        ret.preCopy(walker, this);
        ret.packageName = packageName;
        if (imports != null) ret.setImports( (Imports)walker.process(imports) );
        if (decs != null) ret.setDecs( (Decs)walker.process(decs) );
        ret.sourceInfo = sourceInfo;
        ret.scanned = scanned;
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return imports;
        case 1: return decs;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "imports";
        case 1: return "decs";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setImports((Imports)child); return;
        case 1: setDecs((Decs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "CompilationUnit(packageName: "+packageName+", "+"sourceInfo: "+sourceInfo+", "+"scanned: "+scanned+")";
    }

    //END: Generated from @child and @property
}



