/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import org.aspectj.compiler.base.LocalClassPass;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;
import org.aspectj.compiler.base.bcg.MethodBuilder;
import org.aspectj.compiler.base.bcg.ClassfileBuilder;

/**
  * @grammar { constructorCall; stmts* }
  * @child ConstructorCallExpr constructorCall
  */

public class ConstructorBody extends CodeBody {

    protected void setupDefaultConstructorCall() {
        if (constructorCall == null) {
            Type superType = getDeclaringType().getTypeDec().getSuperClassType();
            Exprs args = getAST().makeExprs();
            Constructor superConstructor = superType.getConstructor(this, args, true);
            ConstructorCallExpr newCallExpr =
                getAST().makeSuperConstructorCall(superConstructor);
            setConstructorCall(newCallExpr);
        }
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) {
        if (stmts != null && walker.walkBodies()) setupDefaultConstructorCall();
        super.preScope(walker);
    }

    // for auto-added constructors
    public ConstructorBody(SourceLocation location, Stmts _stmts, ConstructorCallExpr _constructorCall, boolean _parsed) {
        super(location,_stmts,_parsed);
        setConstructorCall(_constructorCall);
    }

    public void unparse(CodeWriter writer) {
        writer.openBlock();
        writer.write(constructorCall);
        writer.newLine();
        if (tmpStmts != null) writer.write(tmpStmts);
        writer.write(stmts);
        writer.closeBlock();
    }

    public void cleanup() {
        constructorCall = null;
        super.cleanup();
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public void preThreading(LocalClassPass.ThreadingWalker walker) {
        if (getConstructorCall().getIsSuper()) {
            walker.addFieldSets(getStmts());
        }
    }

    // ------------------------------
    // bcg
    protected void cgStmt(CodeBuilder cb) {
        getConstructorCall().cgEffect(cb);
        super.cgStmt(cb);
    }

    // ------------------------------
    // convenience constructors

    public ConstructorBody(SourceLocation location, Stmts _stmts,
                           boolean _parsed, ConstructorCallExpr _constructorCall) {
        this(location, null, _stmts, _parsed, _constructorCall);
    }

    //BEGIN: Generated from @child and @property
    protected ConstructorCallExpr constructorCall;
    public ConstructorCallExpr getConstructorCall() { return constructorCall; }
    public void setConstructorCall(ConstructorCallExpr _constructorCall) {
        if (_constructorCall != null) _constructorCall.setParent(this);
        constructorCall = _constructorCall;
    }

    public ConstructorBody(SourceLocation location, Stmts _tmpStmts, Stmts _stmts, boolean _parsed, ConstructorCallExpr _constructorCall) {
        super(location, _tmpStmts, _stmts, _parsed);
        setConstructorCall(_constructorCall);
    }
    protected ConstructorBody(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ConstructorBody ret = new ConstructorBody(getSourceLocation());
        ret.preCopy(walker, this);
        if (tmpStmts != null) ret.setTmpStmts( (Stmts)walker.process(tmpStmts) );
        if (stmts != null) ret.setStmts( (Stmts)walker.process(stmts) );
        ret.parsed = parsed;
        if (constructorCall != null) ret.setConstructorCall( (ConstructorCallExpr)walker.process(constructorCall) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 2: return constructorCall;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 2: return "constructorCall";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 2: setConstructorCall((ConstructorCallExpr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 3;
    }

    public String getDefaultDisplayName() {
        return "ConstructorBody(parsed: "+parsed+")";
    }

    //END: Generated from @child and @property
}
