/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
    A representation of the control context, the lexical context of
    statements that affect control flow.  This is used to handle the
    control issues arising from labeled statements, breaks and
    continues, returns and throws, tries and continues, loops (for,
    while, do), and switches.

    @see LabeledStmt
    @see WhileStmt
    @see DoStmt
    @see ForStmt
    @see SwitchStmt

    @see WindingStmt
    @see TryStmt
    @see SynchronizedStmt

    @see BreakStmt
    @see ContinueStmt
    @see ReturnStmt
    @see ThrowStmt
*/

public final class ControlContext {

    private List context = new ArrayList();

    /** Enters the context of a control-effecting statement.

        @param the control-effecting statement entered.
     */
    public void enter(Stmt s) {
        context.add(s);
    }

    /** Exits the context of a previously-entered control-effecting statement.
     */
    public void exit() {
        context.remove(context.size() - 1);
    }

    /** Returns the target statement of a break statement.

        @param label label of break statement, or null
        @return the target of the break statement.
        @exception java.util.NoSuchElementException there is no target
     */
    public Stmt getBreakTarget(String label) {
        return getTarget(label, true);
    }

    /** Returns the target statement of a continue statement.

        @param label label of continue statement, or null
        @return the target of the continue statement.
        @exception java.util.NoSuchElementException there is no target
        @exception UnsupportedOperationException the target is not continuable
    */
    public Stmt getContinueTarget(String label) {
        return getTarget(label, false);
    }

    /** Returns true if a {@link LabeledStmt} with a
        label is in the context.

        @return true if label is used in the context, false otherwise.
        @param label label to check for.  */
    public boolean isLabelUsed(String label) {
        for (int i = context.size() - 1; i >= 0; i--) {
            Object s = context.get(i);
            if (s instanceof LabeledStmt &&
                label.equals(((LabeledStmt)s).getLabel()))
                return true;
        }
        return false;
    }

    private Stmt getTarget(String target, boolean isBreak) {
//          System.err.println("CC " + context);
//          System.err.println("-- " + target);
//          System.err.println("-- " + (isBreak ? "break" : "continue"));
        int i = context.size() - 1;
        if (target == null) {
            if (isBreak) {
                for (; i >= 0; i--) {
                    Stmt s = (Stmt) context.get(i);
                    if (s.isBreakable()) return s;
                }
            } else {
                for (; i >= 0; i--) {
                    Stmt s = (Stmt) context.get(i);
                    if (s.isContinuable()) return s;
                }
            }
        } else {
            for (; i >= 0; i--) {
                Stmt s = (Stmt) context.get(i);
                if (s instanceof LabeledStmt
                        && target.equals(((LabeledStmt)s).getLabel())) {
                    if (isBreak) return s;
                    s = ((LabeledStmt) s).getStmt();
//                    while (s instanceof LabeledStmt) {
//                        s = ((LabeledStmt) s).getStmt();
//                    }
                    if (s.isContinuable()) return s;
                    throw new UnsupportedOperationException();
                }
            }
        }
        throw new NoSuchElementException();
    }

    /** Returns an iterator view of the winding statements up to a
        control target.  If null is passed in, will return all winding
        statements.  This iterator is not only non-thread-safe, it
        cannot be stored anywhere either, since it shares its backing
        with the context.  So as soon as it is returned, it should be
        iterated over.

        @return an iterator view of the winding statements up to a control target.
        @param end the control-effecting statement enclosing the winding statements.
    */
    public Iterator getWindsUntil(final Stmt end) {
        return new Iterator() {
                private int i = context.size() - 1;
                private boolean live = true;
                private void search() {
                    for (; i >= 0; i--) {
                        Object o = context.get(i);
                        if (o == end) {
                            break;
                        } else if (o instanceof WindingStmt) {
                            return;
                        }
                    }
                    live = false;
                }
                {
                    search();
                }
                public boolean hasNext() { return live; }
                public Object next() {
                    if (! live) throw new NoSuchElementException();
                    Object o = context.get(i--);
                    search();
                    return o;
                }
                public void remove() {
                    throw new UnsupportedOperationException();
                }
            };
    }
}
