/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar do {body} while (test);
  * @child Stmt body
  * @child Expr test
  */
public class DoStmt extends TestStmt {

    public boolean isBreakable() { return true; }
    public boolean isContinuable() { return true; }

    public void unparse(CodeWriter writer) throws IOException {
        writer.writeKeyword("do");
        writer.requiredSpace();
        writer.write(body);
        writer.writeKeyword("while");
        writer.requiredSpace();
        writer.parenExpr(test);
        writer.closeStmt();
    }
    
    public void checkSpec() {
        body.requireStmt();
        super.checkSpec();
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        // w.setArgs(w.getArgs());
        FlowCheckerPass.Set enteringPossiblyAssigned = w.popPossiblyAssigned();
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();
        FlowCheckerPass.Vars v = w.getVars();
        FlowCheckerPass.Vars cv = w.getContinueVars(this);
        FlowCheckerPass.Vars bv = w.getBreakVars(this);
        boolean bodyLive = w.isLive();

        w.setVars(v.join(cv));
        w.processBoolean(getTest());
        FlowCheckerPass.Vars p = w.getVars();
        FlowCheckerPass.Vars tv = p.getTrue();
        FlowCheckerPass.Vars fv = p.getFalse();

        FlowCheckerPass.Set loopPossiblyAssigned = w.popPossiblyAssigned();

        w.checkLoopingFinals(this, loopPossiblyAssigned.inter(enteringPossiblyAssigned), tv);
        w.mergePossiblyAssigned(enteringPossiblyAssigned);
        w.mergePossiblyAssigned(loopPossiblyAssigned);

        w.setLive(w.isBroken(this) ||
                  (! getTest().isConstantTrue() &&
                   (bodyLive || w.isContinued(this))));
        w.setVars(fv.join(bv));
    }

    // ------------------------------
    // INTRO from ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();

        w.setLive(w.isBroken(this)
                  || (! getTest().isConstantTrue()
                      && (w.isLive() || w.isContinued(this))));
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label startLab = cb.genLabel();
        Label testLab = cb.genLabel();
        Label endLab = cb.genLabel();

        cb.emitLabel(startLab);
        cb.enterNonWindingContext(this, endLab, testLab);
        getBody().cgTop(cb);
        cb.leaveContext();

        cb.emitLabel(testLab);
        getTest().cgTest(cb, startLab, endLab);
        cb.emitLabel(endLab);
    }

    //BEGIN: Generated from @child and @property
    protected Stmt body;
    public Stmt getBody() { return body; }
    public void setBody(Stmt _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }

    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }

    public DoStmt(SourceLocation location, Stmt _body, Expr _test) {
        super(location);
        setBody(_body);
        setTest(_test);
    }
    protected DoStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        DoStmt ret = new DoStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (body != null) ret.setBody( (Stmt)walker.process(body) );
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return body;
        case 1: return test;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "body";
        case 1: return "test";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setBody((Stmt)child); return;
        case 1: setTest((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "DoStmt()";
    }

    //END: Generated from @child and @property
}


