/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

import java.util.*;

/**
  * @grammar (formal, )*;
  * @children FormalDec formal
  */

public class Formals extends ASTObject {

    // stolen from CodeBody
    public CodeDec getEnclosingCodeDec() {
        if (getParent() instanceof CodeDec) return (CodeDec)getParent();
        return super.getEnclosingCodeDec();
    }
    
    public Set makeSet() {
     	Set ret = new HashSet();
     	for (int i = 0, len = size(); i < len; i++) {
     	    ret.add(get(i));
     	}
     	return ret;
    }

    public Iterator iterator() {
        return new Iterator() {
                private int i = 0;
                public boolean hasNext() { return i < size; }
                public Object next() { return children[i++]; }
                public void remove() {
                    throw new UnsupportedOperationException();
                }
            };
    }

    public boolean canBeCalledWith(Exprs parameters) {
        if (this.size() != parameters.size()) return false;
        for(int i=0; i<parameters.size(); i++) {
            if (!parameters.get(i).getType().isMethodConvertableTo(this.get(i).getType())) {
                return false;
            }
        }
        return true;
    }

    public boolean matches(Formals otherFormals) {
//        FormalDec[] formals1 = children;
//        FormalDec[] formals2 = otherFormals.children;

        if (size() != otherFormals.size()) return false;
        for(int i=0; i<size(); i++) {
//            System.out.println(formals1[i] + ", " + formals2[i]);
            if (!get(i).getType().isEquivalent(otherFormals.get(i).getType())) return false;
        }
        return true;
    }

    public FormalDec findName(String name) {
        final int N = size();
        for(int i=0; i<N; i++) {
            if (get(i).getId().equals(name)) {
                return get(i);
            }
        }
        return null;
    }

//    protected static VarDec makeTransformVar(FormalDec formal1, FormalDec formal2) {
//        return new VarDec(formal1.getCompiler(),formal1.getTypeD(), formal1.getName(),
//                          new VarExpr(formal1.getCompiler(),formal2.getName()));
//    }

    public Map findTransform(Formals otherFormals) {
        FormalDec[] formals1 = children;
        FormalDec[] formals2 = otherFormals.children;
        Map ft = new HashMap();

        if (formals1.length != formals2.length) return null;
        for(int i=0; i<formals1.length; i++) {
            if (!formals1[i].getType().isEquivalent(formals2[i].getType())) return null;
            ft.put(formals1[i].getName(), formals2[i]);
        }
        return ft;
    }

    public String toShortString() {
        StringBuffer buf = new StringBuffer();
        buf.append("(");
        final int N = size();
        for(int i=0; i<N; i++) {
            if (i > 0) buf.append(", ");
            FormalDec child = get(i);
            if (child == null) {
                this.showWarning("null child: "+i+" : "+children.length);
                //System.out.println("1st child: "+children[1]+", "+children[0]);
                continue;
            }
            TypeD typeD = child.getTypeD();
            if (typeD == null) child.showError("null typeD on :"+child.getId());
            buf.append(typeD.toShortString());
        }
        buf.append(")");
        return buf.toString();
    }


    public void unparse(CodeWriter writer) {
        writer.openParen('(');
        writer.writeChildrenWithCommas(this);
        writer.closeParen(')');
    }

    public Exprs makeExprs() {
        final AST ast = getAST();

        final int N = size();
        Exprs args = ast.makeExprs();
        for(int i=0; i<N; i++) {
            args.add(ast.makeVar(get(i)));
        }
        return args;
    }

    //BEGIN: Generated from @child and @property
    protected int size;
    public FormalDec[] children;

    public Formals(SourceLocation location, FormalDec[] _children) {
        super(location);
        for(int i=0; i<_children.length; i++) {
            if (_children[i] != null) _children[i].setParent(this);
        }
        children = _children;
        size = _children.length;
    }

    public Formals(SourceLocation location) {
        this(location, new FormalDec[] {});
    }

    public Formals(SourceLocation location, FormalDec child1) {
        this(location, new FormalDec[] {child1});
    }

    public Formals(SourceLocation location, FormalDec child1, FormalDec child2) {
        this(location, new FormalDec[] {child1, child2});
    }

    public Formals(SourceLocation location, FormalDec child1, FormalDec child2, FormalDec child3) {
        this(location, new FormalDec[] {child1, child2, child3});
    }

    public ASTObject copyWalk(CopyWalker walker) {
        final int N = size;
        FormalDec[] copiedChildren = new FormalDec[N];
        int newIndex = 0;
        for(int oldIndex=0; oldIndex<N; oldIndex++) {
            FormalDec newChild = (FormalDec)walker.process(children[oldIndex]);
            if (newChild != null) copiedChildren[newIndex++] = newChild;
        }
        Formals ret = new Formals(getSourceLocation(),copiedChildren);
        ret.size = newIndex;
        ret.setSource(this);
        return ret;
    }

    public ASTObject getChildAt(int childIndex) { return get(childIndex); }
    public void setChildAt(int childIndex, ASTObject child) { set(childIndex, (FormalDec)child); }
    public String getChildNameAt(int childIndex) { return "formal"+childIndex; }

    public int getChildCount() { return size; }
    public int size() { return size; }

    public FormalDec get(int index) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        return children[index];
    }

    public void set(int index, FormalDec child) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        children[index] = child;
        child.setParent(this);
    }

    public void resize(int newSize) {
        if (newSize > children.length) {
            FormalDec[] newChildren = new FormalDec[children.length*2 + 1];
            System.arraycopy(children, 0, newChildren, 0, children.length);
            children = newChildren;
        }
        size = newSize;
    }

    public void addAll(Formals collection) {
        addAll(size, collection);
    }

    public void addAll(int index, Formals collection) {
        for(int i=0; i<collection.size(); i++) {
            add(index+i, collection.get(i));
        }
    }

    public void add(FormalDec child) {
        add(size, child);
    }

    public void add(int index, FormalDec child) {
        if (child == null) return;

        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        resize(size+1);

        for(int moveIndex = size-1; moveIndex > index; moveIndex--) {
            children[moveIndex] = children[moveIndex-1];
        }

        children[index] = child;
        child.setParent(this);
    }

    public void remove(int index) {
        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        size -= 1;

        for(int moveIndex = index; moveIndex < size; moveIndex++) {
            children[moveIndex] = children[moveIndex+1];
        }
    }

    public String getDefaultDisplayName() {
        return "Formals()";
    }

    //END: Generated from @child and @property
}
