/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar if (test) {then} else {else}
  * @child Expr test
  * @child Stmt then
  * @child Stmt else
  */

public class IfStmt extends TestStmt {

    public void unparse(CodeWriter writer) {
        writer.writeKeyword("if");
        writer.optionalSpace();
        writer.parenExpr(test);
        writer.optionalSpace();
        writer.write(then);
        if (! (_else instanceof EmptyStmt && _else.isSynthetic())) {
            writer.writeKeyword("else");
            writer.optionalSpace();
            writer.write(_else);
        }
    }
    
    public void checkSpec() {
        then.requireStmt();
        if (_else != null) _else.requireStmt();
        super.checkSpec();
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        // w.setArgs(w.getArgs());
        w.processBoolean(getTest());
        FlowCheckerPass.Vars p0 = w.getVars();

        w.setVars(p0.getTrue());
        w.process(getThen());
        FlowCheckerPass.Vars v0 = w.getVars();
        boolean trueLive = w.isLive();

        w.setLive(true);
        w.setVars(p0.getFalse());
        w.process(getElse());
        FlowCheckerPass.Vars v1 = w.getVars();
        boolean falseLive = w.isLive();

        w.setLive(falseLive || trueLive);
        w.setVars(v0.join(v1));
    }

    // ------------------------------
    // INTRO from ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantTrue()) {
            w.process(getThen());
        } else if (getTest().isConstantFalse()) {
            w.process(getElse());
        } else {
            w.setLive(true);
            setThen((Stmt) w.process(getThen()));
            boolean thenLive = w.isLive();
            w.setLive(true);
            setElse((Stmt) w.process(getElse()));
            w.setLive(thenLive || w.isLive());
        }
    }

    public ASTObject postCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantTrue()) {
            return getThen();
        } else if (getTest().isConstantFalse()) {
            return getElse();
        } else if (getThen() instanceof EmptyStmt
                   && getElse() instanceof EmptyStmt) {
            return getAST().makeStmt(getTest()).setSource(this);
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label trueLab = cb.genLabel();
        Label falseLab = cb.genLabel();
        Label endLab = cb.genLabel();
        getTest().cgTest(cb, trueLab, falseLab);
        cb.emitLabel(trueLab);
        getThen().cgTop(cb);
        if (getThen().completesNormally()) {
            cb.emitJump(endLab);
        }
        cb.emitLabel(falseLab);
        getElse().cgTop(cb);
        cb.emitLabel(endLab);
    }

    //BEGIN: Generated from @child and @property
    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }

    protected Stmt then;
    public Stmt getThen() { return then; }
    public void setThen(Stmt _then) {
        if (_then != null) _then.setParent(this);
        then = _then;
    }

    protected Stmt _else;
    public Stmt getElse() { return _else; }
    public void setElse(Stmt __else) {
        if (__else != null) __else.setParent(this);
        _else = __else;
    }

    public IfStmt(SourceLocation location, Expr _test, Stmt _then, Stmt __else) {
        super(location);
        setTest(_test);
        setThen(_then);
        setElse(__else);
    }
    protected IfStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        IfStmt ret = new IfStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        if (then != null) ret.setThen( (Stmt)walker.process(then) );
        if (_else != null) ret.setElse( (Stmt)walker.process(_else) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return test;
        case 1: return then;
        case 2: return _else;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "test";
        case 1: return "then";
        case 2: return "else";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTest((Expr)child); return;
        case 1: setThen((Stmt)child); return;
        case 2: setElse((Stmt)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 3;
    }

    public String getDefaultDisplayName() {
        return "IfStmt()";
    }

    //END: Generated from @child and @property
}


