/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar label: stmt;
  * @property String label
  * @child Stmt stmt
  */

public class LabeledStmt extends Stmt {

    public void unparse(CodeWriter writer) throws IOException {
        writer.write(label);
        writer.write(": ");
        writer.write(stmt);
    }
    
    public void requireStmt() {
        this.showError("labeled statement not allowed here");
    }
    
    public void checkSpec() {
        if (!(stmt instanceof LabeledStmt)) {
            stmt.requireStmt();
        }
        super.checkSpec();
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        if (w.isLabelUsed(getLabel())) {
            showError("duplicate definition of label " + getLabel());
        }
        w.enterContext(this);
        w.process(getStmt());
        w.leaveContext();
        w.setLive(w.isLive() || w.isBroken(this));
        w.setVars(w.getVars().join(w.getBreakVars(this)));
    }

    // ------------------------------
    // Intro: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        w.enterContext(this);
        w.process(getStmt());
        w.leaveContext();
        w.setLive(w.isLive() || w.isBroken(this));
    }

    // does this "optimization" even do anything?
    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        if (getStmt() instanceof EmptyStmt) {
            return getStmt();
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label endLab = cb.genLabel();
        cb.enterNonWindingContext(this, endLab, null);
        stmt.cgTop(cb);
        cb.leaveContext();
        cb.emitLabel(endLab);
    }

    // ------------------------------
    // flow

//      protected FlowEnv checkFlow(FlowEnv env) {
//          return getStmt().checkLabeledFlow(env, getLabel());
//      }

    //BEGIN: Generated from @child and @property
    protected String label;
    public String getLabel() { return label; }
    public void setLabel(String _label) { label = _label; }

    protected Stmt stmt;
    public Stmt getStmt() { return stmt; }
    public void setStmt(Stmt _stmt) {
        if (_stmt != null) _stmt.setParent(this);
        stmt = _stmt;
    }

    public LabeledStmt(SourceLocation location, String _label, Stmt _stmt) {
        super(location);
        setLabel(_label);
        setStmt(_stmt);
    }
    protected LabeledStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        LabeledStmt ret = new LabeledStmt(getSourceLocation());
        ret.preCopy(walker, this);
        ret.label = label;
        if (stmt != null) ret.setStmt( (Stmt)walker.process(stmt) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return stmt;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "stmt";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setStmt((Stmt)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }

    public String getDefaultDisplayName() {
        return "LabeledStmt(label: "+label+")";
    }

    //END: Generated from @child and @property
}


