/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

public final class LongType extends IntegralType {

    protected boolean isAssignableFromOtherType(Type other) {
        return (other instanceof IntishType);
    }

    public String getName()      { return "long"; }
    public int    getTypeIndex() { return Type.LONG; }
    public Type   getRefType()   { return getTypeManager().getType("java.lang", "Long"); }
    public Expr   getNullExpr()  { return getAST().makeLiteral(0l); }

    public LongType(JavaCompiler compiler) {
        super(compiler);
    }

    // vm internals
    public String getDescriptor() { return "J"; }
    public int getSlotCount() { return 2; }

    // folding
    private LiteralExpr makeLit(long l) {
        return getAST().makeLiteral(l);
    }
    private LiteralExpr makeBooleanLit(boolean b) {
        return getAST().makeLiteral(b);
    }

    LiteralExpr foldCast(LiteralExpr rand) {
        return makeLit(rand.getLongValue());
    }

    LiteralExpr foldPlusOp(LiteralExpr rand) {
        return makeLit(+ rand.getLongValue());
    }
    LiteralExpr foldMinusOp(LiteralExpr rand) {
        return makeLit(- rand.getLongValue());
    }
    LiteralExpr foldBitNotOp(LiteralExpr rand) {
        return makeLit(~ rand.getLongValue());
    }

    LiteralExpr foldAddOp(LiteralExpr rand1, LiteralExpr rand2) {
        return makeLit(rand1.getLongValue() + rand2.getLongValue());
    }

    LiteralExpr foldNumericOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        long a = lit1.getLongValue();
        long b = lit2.getLongValue();
        if      (op == "-") return ast.makeLiteral(a - b);
        else if (op == "%") return ast.makeLiteral(a % b);
        else if (op == "/") return ast.makeLiteral(a / b);
        else if (op == "*") return ast.makeLiteral(a * b);
        else throw new RuntimeException("bad numeric op " + op);
    }

    LiteralExpr foldBitwiseOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        long a = lit1.getLongValue();
        long b = lit2.getLongValue();
        if      (op == "&") return ast.makeLiteral(a & b);
        else if (op == "|") return ast.makeLiteral(a | b);
        else if (op == "^") return ast.makeLiteral(a ^ b);
        else throw new RuntimeException("bad bitwise op " + op);
    }

    LiteralExpr foldShiftOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        long a = lit1.getLongValue();
        int b = lit2.getIntValue();
        if      (op == "<<") return ast.makeLiteral(a << b);
        else if (op == ">>") return ast.makeLiteral(a >> b);
        else if (op == ">>>") return ast.makeLiteral(a >>> b);
        else throw new RuntimeException("bad bitwise op " + op);
    }

    LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        long a = lit1.getLongValue();
        long b = lit2.getLongValue();
        if      (op == "==") return ast.makeLiteral(a == b);
        else if (op == "!=") return ast.makeLiteral(a != b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        long a = lit1.getLongValue();
        long b = lit2.getLongValue();
        if      (op == "<")  return ast.makeLiteral(a < b);
        else if (op == "<=") return ast.makeLiteral(a <= b);
        else if (op == ">=") return ast.makeLiteral(a >= b);
        else if (op == ">")  return ast.makeLiteral(a > b);
        else throw new RuntimeException("bad op " + op);
    }

    // code gen

    void emitAdd(CodeBuilder cb) { cb.emitLADD(); }

    void emitNumericOp(CodeBuilder cb, String op) {
        if      (op == "-") cb.emitLSUB();
        else if (op == "%") cb.emitLREM();
        else if (op == "/") cb.emitLDIV();
        else if (op == "*") cb.emitLMUL();
    }
    void emitBitwiseOp(CodeBuilder cb, String op) {
        if      (op == "&") cb.emitLAND();
        else if (op == "|") cb.emitLOR();
        else if (op == "^") cb.emitLXOR();
    }
    void emitShiftOp(CodeBuilder cb, String op) {
        if      (op == "<<")  cb.emitLSHL();
        else if (op == ">>")  cb.emitLSHR();
        else if (op == ">>>") cb.emitLUSHR();
    }


    void emitNeg(CodeBuilder cb) { cb.emitLNEG(); }

    public void emitLoad(CodeBuilder cb, int loc) { cb.emitLLOAD(loc); }
    public void emitStore(CodeBuilder cb, int loc) { cb.emitLSTORE(loc); }
    public void emitReturn(CodeBuilder cb) { cb.emitLRETURN(); }

    void emitCast(CodeBuilder cb, Type castTo) {
        ((NumericType) castTo).emitCastFromLong(cb);
    }
    void emitCastFromInt(CodeBuilder cb)    { cb.emitI2L(); }
    void emitCastFromFloat(CodeBuilder cb)  { cb.emitF2L(); }
    void emitCastFromLong(CodeBuilder cb)   { }
    void emitCastFromDouble(CodeBuilder cb) { cb.emitD2L(); }

    void emitPop(CodeBuilder cb) { cb.emitPOP2(); }
    void emitDup(CodeBuilder cb) { cb.emitDUP2(); }
    void emitDupX1(CodeBuilder cb) { cb.emitDUP2_X1(); }
    void emitDupX2(CodeBuilder cb) { cb.emitDUP2_X2(); }

    void emitZero(CodeBuilder cb) { cb.emitLongConstant(0L); }
    void emitOne(CodeBuilder cb) { cb.emitLongConstant(1L); }
    void emitMinusOne(CodeBuilder cb) { cb.emitLongConstant(-1L); }

    void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f) {
        cb.emitLCMP();
        if      (op == "==") cb.emitIFEQ(t, f);
        else if (op == "!=") cb.emitIFNE(t, f);
    }

    void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f) {
        cb.emitLCMP();
        if      (op == "<")  cb.emitIFLT(t, f);
        else if (op == "<=") cb.emitIFLE(t, f);
        else if (op == ">=") cb.emitIFGE(t, f);
        else if (op == ">")  cb.emitIFGT(t, f);
    }

    void emitAload(CodeBuilder cb) { cb.emitLALOAD(); }
    void emitAstore(CodeBuilder cb) { cb.emitLASTORE(); }

}

