/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.*;
import java.io.IOException;

//ASPECT use introduction from crosscuts package
import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;


/**
 * A unique Type for each class, interface or aspect in the system,
 * either defined in source or loaded from bytecode.
 * There is a 1-to-1 mapping between NameTypes and TypeDecs
 *
 * Virtually all of the code for validating Types and for looking up decs
 * efficiently on them is handled by this class.
 */
public class NameType extends org.aspectj.compiler.base.ast.RefType {
    final SemanticMap fields = new SemanticMap(this, "field");

    final SemanticMap pointcuts = new SemanticMap(this, "pointcut");

    final SemanticMap innerTypes = new SemanticMap(this, "type");

    final SemanticMap methods = new SemanticMap(this, "method");

    //??? having a SemanticMap for constructors simplifies the design, but isn't terribly
    //??? elegant since constructors don't even really have names...
    final SemanticMap constructors = new SemanticMap(this, "constructor");

    public NameType(TypeDec typeDec) {
        super(typeDec.getCompiler());
        this.dec = typeDec;
    }

    public TypeDec getTypeDec() { return (TypeDec)dec; }

    public boolean isCoercableTo(Type other) {
        if (this.isAssignableFrom(other)) return true;
        if (other.isAssignableFrom(this)) return true;
        
        if (! (other instanceof NameType) ) return false;

        // if they are both classes, then a direct relationship is required
        if (!isInterface() && !other.isInterface()) return false;
        
        // if they are both interfaces, then check method compatibility
        if (isInterface() && other.isInterface()) {
            return methodsCompatibleWith(other);
        }
        
        // if one is a class and the other is an interface, then check that the
        // class is not final -- compatible with JLS 5.5
        // would be nice if spec included method compatibility check here,
        // but it doesn't
        if (isInterface()) return !((NameType)other).isFinal();
        if (other.isInterface()) return !this.isFinal();

        return false;
    }


    /**
     * Checks to see if all of my methods are compatible with all
     * of the methods of some other type.
     */
    private boolean methodsCompatibleWith(Type other) {
        for (Iterator i = getMethods().iterator(); i.hasNext(); ) {
            Method m = (Method)i.next();
            Method mOther = (Method)other.findMatchingSemanticObject(m);
            if (mOther != null) {
                if (!m.getReturnType().isEquivalent(mOther.getReturnType())) {
                    return false;
                }
            }
        }
        return true;
    }


    public boolean dominates(Type other) {
        TypeDec t1 = getTypeDec();
        TypeDec t2 = other.getTypeDec();
        if (t2 != null && t1 instanceof AspectDec && t2 instanceof AspectDec) {
            return ((AspectDec)t1).dominates((AspectDec)t2);
        } else {
            return super.dominates(other);
        }
    }

    public boolean isEffectivelyStatic() {
        return isStatic() || isPackageMember();
    }

    public Expr getClassExpr() {
        if (! getTypeDec().hasGlobalName()) return getAST().makeNull();
        return new ClassExpr(getAST().getSourceLocation(), this.makeTypeD());
    }

    public Expr makeObject(Expr expr) { return expr; }

    public Expr fromObject(Expr expr) {
        return getAST().makeCast(this, expr);
    }

    public Expr getNullExpr() {
        return getAST().makeNull();
    }

    public Type getRefType() { return this; }

    public Type getOutermostType() {
        return getTypeDec().getOutermostType();
    }


    public PointcutSO getPointcut(String id, ASTObject fromWhere, boolean showError) {
        ensureFinishedSignature();
        return (PointcutSO)pointcuts.get(id, fromWhere, null, showError);
    }

    public Field getField(String id, ASTObject fromWhere, boolean showError) {
        //System.out.println("get field on " + this + ", " + System.identityHashCode(this));
        ensureFinishedSignature();
        return (Field)fields.get(id, fromWhere, null, showError);
    }

    public Type getInnerType(String id, ASTObject fromWhere, boolean showError) {
//        System.out.println("get inner type on " + this + ", " + id);
//        System.out.println("  " + fromWhere + ", " + innerTypes.map.keySet());
        //??? including this makes some order issues much harder, can I skip it???
        ensureBuiltTypeGraph();
        //??? what about null returns from this???
        Type ret = (Type)innerTypes.get(id, fromWhere, null, showError);
        //System.out.println("  returning: " + ret);
        return ret;
    }

    public Method getMethod(String id, ASTObject fromWhere, Exprs params, boolean showError) {
        ensureFinishedSignature();
        //System.out.println(" " + id + " in " + this);
        Method ret = (Method)methods.get(id, fromWhere, params, showError);
//        if (ret == getWorld().METHOD_DEC_NOT_FOUND) {
//            getTypeDec().showError("can't find: " + id);
//        }
//        System.out.println("    " + ret);
        return ret;
    }

    // currently only used by ReceptionsPointcut and by TypeScope.findMethodLookupType
    public boolean hasMethodNamed(String id) {
        ensureFinishedSignature();
        return methods.hasName(id);
    }

    public Constructor getConstructor(ASTObject fromWhere, Exprs params, boolean showError) {
        ensureFinishedSignature();

        return (Constructor)constructors.get("new", fromWhere, params, showError);
    }

    public SemanticObject findMatchingSemanticObject(SemanticObject o) {
        if (o instanceof Method) {
            return methods.findMatchingSemanticObject(o);
        } else if (o instanceof Constructor) {
            return constructors.findMatchingSemanticObject(o);
        } else if (o instanceof Field) {
            return fields.findMatchingSemanticObject(o);
        } else {
            return null;
        }
    }

    private void addInheritedMembers(SemanticMap map, Collection bag) {
        for (Iterator i = map.getAllSemanticObjects().iterator(); i.hasNext(); ) {
            SemanticObject so = (SemanticObject)i.next();
            if (so.getDeclaringType() != this) { bag.add(so); }
        }
    }

    // currently only used by AdviceWeaver for dealing with synthetic reception points
    public Collection getInheritedMethods() {
        List ret = new LinkedList();
        addInheritedMembers(methods, ret);
        return ret;
    }

    public Collection getInheritedMembers() {
        List ret = new LinkedList();
        addInheritedMembers(methods, ret);
        addInheritedMembers(innerTypes, ret);
        addInheritedMembers(fields, ret);
        addInheritedMembers(pointcuts, ret);
        return ret;
    }

    public Collection getConstructors() {
        return constructors.getAllSemanticObjects();
    }

    /** Used by ajdoc. */
    public Collection getMethods() {
        return methods.getAllSemanticObjects();
    }
    /** Used by ajdoc. */
    public Collection getFields() {
        return fields.getAllSemanticObjects();
    }
    /** Used by ajdoc. */
    public Collection getPointcuts() {
        return pointcuts.getAllSemanticObjects();
    }


    public void addPointcut(PointcutDec pointcut) {
        pointcuts.addDeclared(pointcut.getCorrespondingSemanticObject());
    }

    public void addField(FieldDec field) {
        //System.out.println("adding " + field + " to " + System.identityHashCode(this));
        fields.addDeclared(field.getCorrespondingSemanticObject());
    }

    public void addMethod(MethodDec method) {
        methods.addDeclared(method.getCorrespondingSemanticObject());
    }

    public void addConstructor(ConstructorDec constructor) {
        constructors.addDeclared(constructor.getCorrespondingSemanticObject());
    }

    public NameType getSuperClassType() {
        return (NameType)getTypeDec().getSuperClassType();
    }

    public Type getDeclaringType() {
        return getTypeDec().getDeclaringType();
    }

    public void addInnerType(TypeDec type) {
        //System.out.println("adding inner " + type + " to " + this);
        innerTypes.addDeclared(type.getType());
    }

    public Collection getInheritedFromTypes() {
        //??? true for classes
        return directSuperTypes;
    }

    private void initializeInheritedTypeMap() {
        for (Iterator i = getInheritedFromTypes().iterator(); i.hasNext(); ) {
            NameType superType = (NameType)i.next();
            superType.ensureBuiltTypeGraph();
            innerTypes.addInherited(superType.innerTypes);
        }

 //       System.out.println(this.getString() + " inners " + 
//                           getInnerTypes() +
//                           " supers " + getInheritedFromTypes());
    }


    private void handleInheritedMembers(NameType superType) {
        superType.ensureFinishedSignature();

        pointcuts.addInherited(superType.pointcuts);
        fields.addInherited(superType.fields);
        methods.addInherited(superType.methods);
    }

    private void handleInheritedMembers() {
        NameType superClassType = getSuperClassType();
        //System.out.println(this + " super: " + superClassType);
        if (superClassType != null && getInheritedFromTypes().contains(superClassType)) {
            //System.out.println("   inherited");
            handleInheritedMembers(superClassType);
        }

        for (Iterator i = getInheritedFromTypes().iterator(); i.hasNext(); ) {
            NameType superType = (NameType)i.next();
            if (superType != superClassType) handleInheritedMembers(superType);
        }
    }

    //private String prettyString = null;
    public String getPrettyString() {
        String prettyString = null;
        if (prettyString != null) return prettyString;

//        if (! getTypeDec().hasGlobalName()) {
//            prettyString = getTypeDec().getDeclaringType().getPrettyString() + "$" +
//                    getTypeDec().getId();
//        } else {
            prettyString = getLegalString().replace('$', '.');
//        }

        return prettyString;
    }

    private String makeLegalString() {
        return getTypeDec().getFullName();
//        String legalName = getTypeDec().getFullName();
//        if (legalName != null) return legalName;
//        legalName = getTypeDec().getExtendedId();
//        String packageName = getTypeDec().getPackageName();
//        if (packageName != null) {
//            legalName = packageName+'.'+legalName;
//        }
//        return legalName;
    }

    //private String legalString = null;
    public String getLegalString() {
        return makeLegalString();
//        if (legalString != null) return legalString;
//        legalString = makeLegalString();
//        return legalString;
    }

    public String getSourceString() {
        TypeDec typeDec = getTypeDec();
//        if (! typeDec.hasGlobalName()) {
//            return typeDec.getId();
//        }
        return getPrettyString();
    }

    //!!! should be cleaned up in TypeD
    public String getString() {
        return getSourceString(); //+ "%" + System.identityHashCode(this);
    }


	/**
	 * Returns the name of this class minus the package name
	 * 
	 * uses '$' as a separator for inners
	 * 
	 * java.util.Map.Entry goes to Map$Entry
	 */
    public String getExtendedId() {
        return getTypeDec().getExtendedId();
    }


	/**
	 * Returns the simple name of this class, ignoring enclosing types
	 * 
	 * java.util.Map.Entry goes to Entry
	 */
    public String getId() {
        return getTypeDec().getId();
    }

    public String getExternalName() {
        return getClassName();
    }

    public String getClassName() {
        return getTypeDec().getFullName();
    }

    public String getPackageName() {
        return getTypeDec().getPackageName();
    }


    public void unparse(CodeWriter writer) throws IOException {
        writer.write(getSourceString());
    }

    //!!! name should change to getPrettyClassName
    /**
     * @return	e.g. qualified type name java.util.Map.Entry will be returned as Map.Entry	
     */
    public String toShortString() {
        String ret = getTypeDec().getSourceExtendedId();
        //ret += "%" + System.identityHashCode(this);
        return ret;
    }

    public String toString() {
        return "NameType("+getPrettyString()+")";
    }


    public boolean isInterface() { return getTypeDec() instanceof InterfaceDec; }
    public boolean isClass() { return getTypeDec() instanceof ClassDec; }
    public boolean isAspect() { return getTypeDec() instanceof AspectDec; }

    public boolean isInnerType() { return getTypeDec().isInnerType(); }

    public boolean isAnonymous() { return getTypeDec().isLocallyDefined(); }

    public boolean isAbstract() { return isInterface() || getTypeDec().isAbstract(); }
    public boolean isFinal() { return getTypeDec().isFinal(); }

    public void ensureNoAbstractDecs() {
        pointcuts.ensureNoAbstracts();
        methods.ensureNoAbstracts();
    }


    synchronized public void ensureBuiltTypeGraph() {
        if (builtTypeGraph) return;
        //if (getTypeDec().fromSource()) {
        getTypeDec().addToTypeGraph(); //getOutermostTypeDec().addToTypeGraph();
            //} else {
            //    addToTypeGraph();
            //}
    }

    synchronized public void ensureFinishedSignature() {
        if (builtSignatures) return;
			// !!! this was a good idea that didn't match up with the structure of the
			// in particular, NameTypes don't really finish their signatures until they
			// handle their inherited members.  But, they call this method before handling
			// their inherited members.
////        if (!getWorld().collectedTypeDecPlanners) {
////        	getCompiler().internalError("should have collected type dec planners by now");
////        }
        
        ensureBuiltTypeGraph();
        //if (getTypeDec().fromSource()) {
        getTypeDec().buildSignatures(); //getOutermostTypeDec().buildSignatures();
    }

    /*
      Here are the important stages in the lifecyle of a NameType

      parse -- produces concrete syntax tree
      bind names in extends and implements clauses
      build type graph -- requires parents type graphs built
      
      collect and bind introduction planners -- requires type graph
      implement declare parents -- requires all introduction planners collected
      add introduced members to type -- requires declare parents implemented
      bind names in declared members signatures -- requires declare parents implemented

      build type signature??? -- requires built type signature in parents
                              names bound in declared members
                              introduced members added to type

      bind names in bodies
      folding and checking assignments???

      collect advice planners

      ...
    */
    private boolean builtTypeGraph = false;
    public void addToTypeGraph() {
        if (builtTypeGraph) {
            getCompiler().internalError("already built type graph " + 
                                        getPrettyString());
            return;
        }
        builtTypeGraph = true;
        getCompiler().showMessage("  building type graph " + getPrettyString());
        
        if (this != getTypeManager().getObjectType()) {
            addDirectSuperType(getTypeDec().getSuperClassType());
        }

        for (Iterator i = getTypeDec().getSuperInterfaceTypes().iterator(); 
             i.hasNext(); )
        {
            NameType superInterfaceType = (NameType)i.next();
            addDirectSuperType(superInterfaceType);
        }

        initializeInheritedTypeMap();

        getTypeDec().addInnerTypes();
    }

    private boolean builtSignatures = false;
    public void buildSignatures() {
        if (builtSignatures) {
            getCompiler().internalError("already built signatures: " + 
                                        getPrettyString());
            return;
        }
        builtSignatures = true;
        getCompiler().showMessage("  building signatures " + getPrettyString());
        
        if (getTypeDec().fromSource()) checkForCircularities();

        getWorld().runTypeDecPlanners(getTypeDec(), TypeDecPlanner.TYPE);
        
        getTypeDec().addDecs(this);

        getWorld().runTypeDecPlanners(getTypeDec(), TypeDecPlanner.SIGNATURE);
        //System.out.println("almost finished: " + this);

        handleInheritedMembers();
    }

    public Collection getInnerTypes() {
        return innerTypes.getAllSemanticObjects();
    }

    public Type getOutermostLexicalType() {
        return getTypeDec().getOutermostLexicalType();
    }
    
    public Type getOutermostBytecodeType() {
        return getTypeDec().getOutermostBytecodeType();
    }
    
 


    //XXX I hate to do this, but make everything lazy for now
    private boolean isLazy = true;
    public void setLazy() { isLazy = true; }
    

    public void checkForCircularities() {
        checkForCircularities(new LinkedList());
    }

    private boolean checkForCircularities(LinkedList list) {
        if (list.contains(this)) {
            StringBuffer message = new StringBuffer("cyclic inheritance, ");
            boolean foundThis = false;
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                Type t = (Type)i.next();
                if (t == this) {
                    foundThis = true;
                }
                if (foundThis) {
                    message.append(t.getString());
                    message.append(" -> ");
                }
            }
            message.append(getString());
//            message.append(getString());
//            while (list.getLast() != this) {
//                message.append(" <- ");
//                message.append(((Type)list.removeLast()).getString());
//            }
//            message.append(" <- ");
//            message.append(getString());
            getTypeDec().showError(message.toString());
            //??? this makes error handling more robust...
            directSuperTypes.clear();
            return true;
        } else {
            list.add(this);
            if (getEnclosingType() != null) {
                if (getEnclosingType().checkForCircularities(list)) return true;
            }
            
            for (Iterator i = directSuperTypes.iterator(); i.hasNext(); ) {
                NameType superType = (NameType)i.next();
                if (superType.checkForCircularities(list)) return true;
            }
            list.removeLast();
            return false;
        }
    }

    // ------------------------------
    // vm internals
    private String internalName;
    private String descriptor;

    public synchronized String getInternalName() {
        if (internalName == null) {
            internalName = getClassName().replace('.', '/');
        }
        return internalName;
    }

    public synchronized String getDescriptor() {
        if (descriptor == null) {
            descriptor = "L" + getInternalName() + ";";
        }
        return descriptor;
    }

    // ------------------------------
    // ASPECT: Fold
    final LiteralExpr foldAddOp(LiteralExpr rand1, LiteralExpr rand2) {
        return getAST().makeLiteral(rand1.getStringValue() + rand2.getStringValue());
    }
    final LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        String a = lit1.getStringValue();
        String b = lit2.getStringValue();
        if      (op == "==") return ast.makeLiteral(a.equals(b));
        else if (op == "!=") return ast.makeLiteral(! a.equals(b));
        else throw new RuntimeException("bad op " + op);
    }
    final LiteralExpr foldCast(LiteralExpr lit) {
        return getAST().makeLiteral(lit.getStringValue());
    }

    // ------------------------------
    // Dealing with non-package-level types

    // !!! These dispatch to the typeDec.  It's probably wrong that
    // isInner and isPackageMember do so... rather, the typeDec may
    // want to dispatch here.  In particular, it seems more a property
    // of NameTypes that they're "inner", that is, that their
    // constructor takes an extra arg and that they have an enclosing
    // instance.  isLocal() is definitely a property of the dec, rather
    // than the type, though.

    public boolean isInner() {
        return getTypeDec().isInner();
    }
    public NameType getEnclosingInstanceType() {
        // only if isInner()
        return (NameType)getDeclaringType();
    }

    public boolean isLocal() {
        TypeDec typeDec = getTypeDec();
        if (typeDec == null) return false; // ???
        return typeDec.isLocal();
    }
    public CodeDec getEnclosingCodeDec() {
        // only if isLocal()
        return getTypeDec().getEnclosingCodeDec();
    }

    public boolean isPackageMember() {
        return getTypeDec().isPackageMember();
    }
    public NameType getEnclosingType() {
        // only if ! isPackageMember()
        return (NameType)getDeclaringType();
    }
}
