/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.base.bytecode.ClassManager;

import java.util.*;

/**
 * Represents a single package
 * Almost certainly should extends SemanticObject
 */

public class PackageSO extends CompilerObject {
    private String packageName;
    private Map typeCache = new HashMap();
    private Map childPackages = new HashMap();

    public ClassManager classManager;

    public PackageSO(JavaCompiler compiler, String packageName,
                   ClassManager classManager) {
        super(compiler);
        this.packageName = packageName;
        this.classManager = classManager;
    }

    public void addType(Type newType) {
        String className = newType.getTypeDec().getExtendedId();
        
        Type type = (Type)typeCache.get(className);
        if (type != null && type != newType) {
            type.getTypeDec().showError("duplicate type name not allowed: " + 
                                        type.getString());
            newType.getTypeDec().showError("duplicate type name not allowed: " + 
                                           type.getString());
        }
        if (childPackages.containsKey(className)) {
            newType.getTypeDec().showError(newType.getString() + 
                                           " clashes with package of the same name");
            return;
        }
        typeCache.put(className, newType);
    }
    
    private static final Type NOT_FOUND = new NullType(null);
            
    public Type findType(String className) {
        Type type = (Type)typeCache.get(className);
        if (type != null) {
            if (type == NOT_FOUND) return null;
            else return type;
        }
        
        type = findTypeOnClassPath(className); //makeName(packageName, className));
            
        if (type != null) {
            typeCache.put(className, type);
            ((NameType)type).setLazy();
            return type;
        } else {
            typeCache.put(className, NOT_FOUND);
            return null;
        }
    }

    public boolean containsSubPackage(String packageName) {
        PackageSO p = (PackageSO)childPackages.get(packageName);
        if (p == null) return false;
        return p.size() > 0;
    }

    public PackageSO findPackage(String packageName) {
        //???
        //if (findType(packageName) != null) return null;

        PackageSO p = (PackageSO)childPackages.get(packageName);
        if (p != null) return p;

//         getCompiler().showMessage("making sub package: " + 
//                                   makeName(this.packageName, packageName));
        p = new PackageSO(getCompiler(), makeName(this.packageName, packageName),
                        classManager.makeSubPackageManager(packageName));
        //getCompiler().showMessage("made sub package: " + p.packageName);
        //Thread.currentThread().dumpStack();
        childPackages.put(packageName, p);
        return p;
    }

    public int size() { return typeCache.size(); }
        
    private final String makeName(String n1, String n2) {
        if (n1 == null) return n2;
        return n1 + '.' + n2;
    }

    //!!! made public for use by TypeManager.checkLoadable
    public Type findTypeOnClassPath(String name) {
        //System.out.println("finding " + name);
        //System.out.println(packageName + ", " + name + ", " + classManager);
        TypeDec ret = classManager.findTypeDec(name);
        if (ret == null) return null;

        // make sure that the type's name matches the name that was searched for
        // this is to catch case errors on Windows            
        if (!(ret.getFullName().equals(makeName(packageName, name)))) {
            getCompiler().showMessage("    case mismatch, not loading: " + ret.getFullName());
            return null;
        }
        return ret.getType();
    }
}
