/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 *
 * @grammar
 * @property String op
 */
public class PrefixExpr extends BangExpr {

    //INTRO from AccessFixer
    public ASTObject fixAccessToFieldSet(FieldAccessExpr expr) {
        return makeInPlaceSet(expr, op.substring(1), getAST().makeLiteral(1));
    }

    public void unparse(CodeWriter writer) {
        writer.write(op);
        writer.write(lhs);
    }
    
     public void checkSpec() {
        Type ty = getType();
	if (ty.isVoid() ||
	    !(ty.isNumeric() || ty.isAnyType())) {
	        showOperatorTypeError(op, ty);
	    }
    }


    // ------------------------------
    //INTRO from InnerAccessFixer

    public ASTObject postInnerAccess(InnerAccessFixer w) {
        if (! (getLhs() instanceof FieldAccessExpr)) return this;

        FieldAccessExpr lhs = (FieldAccessExpr) getLhs();
        Expr q = lhs.getExpr();
        FieldDec dec = lhs.getFieldDec();
        if (w.isAccessible(dec, q)) return this;

        final AST ast = getAST();
        Type qType = q.getType();
        MethodDec newMethodDec = w.getAccessMethod(qType, dec, getOp() + "x", this);
        Exprs newArgs = ast.makeExprs();
        Expr newExpr = w.makeOutsidePrimary(dec.isStatic(), newArgs, q);
        return ast.makeCall(newMethodDec, newExpr, newArgs);
    }

    public MethodDec buildAccessMethod(InnerAccessFixer w) {
        final AST ast = getAST();
        FieldAccessExpr lhs = (FieldAccessExpr) getLhs();
        Expr q = lhs.getExpr();
        Type qType = q.getType();
        FieldDec dec = lhs.getFieldDec();
        Type fieldType = dec.getType();

        Formals newFormals = ast.makeFormals();
        Expr newExpr = w.makeInsidePrimary(dec.isStatic(), newFormals, qType);

        return
            w.makeAccessMethod(fieldType,
                               newFormals,
                               ast.makePrefix(newExpr, dec, getOp()));
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        cgValueEffect(cb, true);
    }
    protected void cgEffect(CodeBuilder cb) {
        cgValueEffect(cb, false);
    }
    private void cgValueEffect(CodeBuilder cb, boolean needsValue) {
        Type ty = getType();
        AssignableExpr lhs = getLhs();
        if (ty.hasFastIncOp(lhs, 1)) {
            ty.emitFastIncOp(cb, lhs, op.equals("++") ? 1 : -1);
            if (needsValue) lhs.cgValue(cb);
        } else {
            lhs.cgLvalue(cb);
            lhs.cgDupLvalue(cb);
            lhs.cgLtoRvalue(cb);
            ty.emitOne(cb);
            if (op.equals("++")) ty.emitAdd(cb);
            else                 ty.emitNumericOp(cb, "-");
            if (ty instanceof IntishType) {
                ((IntishType)ty).emitCastFromInt(cb);
            }
            if (needsValue) lhs.cgDupRvalue(cb); // pushes dup'd value down
            lhs.cgAssignment(cb);
        }
    }

    //BEGIN: Generated from @child and @property
    protected String op;
    public String getOp() { return op; }
    public void setOp(String _op) { op = _op; }

    public PrefixExpr(SourceLocation location, AssignableExpr _lhs, String _op) {
        super(location, _lhs);
        setOp(_op);
    }
    protected PrefixExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        PrefixExpr ret = new PrefixExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (lhs != null) ret.setLhs( (AssignableExpr)walker.process(lhs) );
        ret.op = op;
        return ret;
    }


    public String getDefaultDisplayName() {
        return "PrefixExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
