/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

abstract public class RefType extends Type {
    public RefType(JavaCompiler compiler) {
        super(compiler);
    }

    // constant folding
    final LiteralExpr foldPlusOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldMinusOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldBitNotOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldLogNotOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldNumericOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    final LiteralExpr foldBitwiseOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    final LiteralExpr foldShiftOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    final LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    
    // overridden in nameType, for String
    LiteralExpr foldCast(LiteralExpr lit) { return unsupportedFold(); }
    LiteralExpr foldAddOp(LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }

    // code gen
    void emitCast(CodeBuilder cb, Type castTo) {
        if (castTo.isAssignableFrom(this)) return;
        if (!(castTo instanceof RefType)) {
            System.err.println("can't emit cast for " + this + " to " + castTo);
            return;
        }
        cb.emitCHECKCAST((RefType)castTo);
    }

    void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "==") cb.emitIF_ACMPEQ(t, f);
        else if (op == "!=") cb.emitIF_ACMPNE(t, f);
    }

    void emitAload(CodeBuilder cb) { cb.emitAALOAD(); }
    void emitAstore(CodeBuilder cb) { cb.emitAASTORE(); }
    void emitNewarray(CodeBuilder cb) { cb.emitANEWARRAY(this); }
    void emitInstanceof(CodeBuilder cb) { cb.emitINSTANCEOF(this); }

    public void emitLoad(CodeBuilder cb, int loc) { cb.emitALOAD(loc); }
    public void emitStore(CodeBuilder cb, int loc) { cb.emitASTORE(loc); }
    public void emitReturn(CodeBuilder cb) { cb.emitARETURN(); }

    String getInternalId() {
        return getId();
    }

    // general type-based emitters
    final void emitBitNot(CodeBuilder cb)     { unsupportedEmit(); }
    final void emitLogNot(CodeBuilder cb)     { unsupportedEmit(); }
    final void emitZero(CodeBuilder cb)       { unsupportedEmit(); }
    final void emitOne(CodeBuilder cb)        { unsupportedEmit(); }
    final void emitMinusOne(CodeBuilder cb)   { unsupportedEmit(); }
    final void emitAdd(CodeBuilder cb)        { unsupportedEmit(); }
    final void emitNeg(CodeBuilder cb)        { unsupportedEmit(); }
    final void emitNumericOp(CodeBuilder cb, String op) { unsupportedEmit(); }
    final void emitBitwiseOp(CodeBuilder cb, String op) { unsupportedEmit(); }
    final void emitShiftOp(CodeBuilder cb, String op)   { unsupportedEmit(); }
    final void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f) {
        throw new RuntimeException("Type " + this + " doesn't support op " + op);
    }

    // ------------------------------
    // "fast" opcode handling

    final boolean hasFastEqualityTestOp() { return true; }
    final void emitFastEqualityTestOp(CodeBuilder cb, EqualityTestOpExpr e,
                                Label t, Label f) {
        Expr rand1 = e.getRand1();
        Expr rand2 = e.getRand2();
        String op = e.getOp();
        if        (rand1.isConstantZero()) {
            rand2.cgValue(cb, this);
            emitEqualityCompare0(cb, op, t, f);
        } else if (rand2.isConstantZero()) {
            rand1.cgValue(cb, this);
            emitEqualityCompare0(cb, op, t, f);
        } else {
            rand1.cgValue(cb, this);
            rand2.cgValue(cb, this);
            emitEqualityCompare(cb, op, t, f);
        }
    }
    private void emitEqualityCompare0(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "==") cb.emitIFNULL(t, f);
        else if (op == "!=") cb.emitIFNONNULL(t, f);
    }
}
