/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar return (expr)?;
 * @child Expr expr
 */

public class ReturnStmt extends Stmt {

    public void checkSpec() {
        //if (getOptions().bcg) {
            // need to find out why this doesn't work in general.  tests/new/AdviceOnPrivileged.java
            //this.getParent().display(0);
            //System.err.println("" + this + this.hashCode());
            CodeDec codeDec = getEnclosingCodeDec();
            //System.err.println("done");
            if (codeDec instanceof InitializerDec) {
                showError("A return statement may not appear in an initializer block");
                return;
            }
            Type resultTy = codeDec.getResultType();
            Expr expr = getExpr();
            if (expr == null) {
                if (! resultTy.isEquivalent(getTypeManager().voidType)) {
                    showError("missing return value");
                }
            } else if (resultTy instanceof VoidType) {
                showError("cannot return a value from method whose result type is void");
            }
        //}
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        if (getExpr() != null) {
            w.process(getExpr());
        }
        w.doReturn();
        w.setLive(false);
        w.setVars(w.getAllVars());
    }

    // ------------------------------
    // INTRO from ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        w.setLive(false);
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        Expr expr = getExpr();
        if (expr == null) return this;
        CodeDec codeDec = getEnclosingCodeDec();
        Type resultTy = codeDec.getResultType();
        //System.out.println(">>>" + codeDec + ", " + resultTy);
        if (! expr.isAssignableTo(resultTy)) {
            expr.showTypeError(expr.getType(), resultTy);
        } else if (resultTy.isPrimitive() && resultTy != expr.getType()) {
            setExpr((Expr)getAST().forceCast(resultTy, expr).setSource(this));
        } else {
            setExpr((Expr)getAST().makeCast(resultTy, expr).setSource(this));
        }
        return this;
    }

    public void unparse(CodeWriter writer) throws IOException {
        writer.writeKeyword("return");

        if (expr != null) {
            writer.requiredSpace();
            writer.write(expr);
        }

        writer.closeStmt();
    }

    /*
    public void parseStmt(Parser parser) {
        if (parser.peekToken(SEMICOLON)) {
            expr = parser.parseExpr();
        }
        parser.eatTopToken(SEMICOLON);
    }
    */

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        if (expr != null) {
            // assumes we've already cast the return expr to the right type.
            expr.cgValue(cb);
            cb.doReturn(expr.getType());
        } else {
            cb.doReturn(getTypeManager().voidType);
        }
    }

    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }

    public ReturnStmt(SourceLocation location, Expr _expr) {
        super(location);
        setExpr(_expr);
    }
    protected ReturnStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ReturnStmt ret = new ReturnStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }

    public String getDefaultDisplayName() {
        return "ReturnStmt()";
    }

    //END: Generated from @child and @property
}


