/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import java.util.*;
import org.aspectj.compiler.base.*;

public class SemanticMap extends CompilerObject {
    protected final Map map = new HashMap();
    protected final Type type;

    protected String kind;

    public SemanticMap(Type type, String kind) {
        super(type.getCompiler());
        this.type = type;
        this.kind = kind;
    }

    public SemanticObject get(String id, ASTObject fromWhere, Exprs params, boolean showError) {
        //System.out.println(map);

        SemanticGroup semanticGroup = (SemanticGroup)map.get(id);
        if (semanticGroup == null) {
            //System.out.println("no map for " + id);
            if (showError) {
                showNotFoundError(id, fromWhere);
            }
            return getNotFoundSemanticObject();
        }
        return semanticGroup.resolve(fromWhere, params, showError);
    }

    public void addDeclared(SemanticObject objects) {
        String id = objects.getId();
        //System.out.println("adding " + dec + " id: " + id);
        SemanticGroup semanticGroup = (SemanticGroup)map.get(id);
        if (semanticGroup == null) {
            map.put(id, makeSemanticGroup(objects));
        } else {
            semanticGroup.addDeclaredSO(objects);
        }
        //System.out.println(map);
    }

    public SemanticObject addIntroduced(SemanticObject objects) {
        String id = objects.getId();
        //System.out.println("adding " + dec + " id: " + id);
        SemanticGroup semanticGroup = (SemanticGroup)map.get(id);
        if (semanticGroup == null) {
            map.put(id, makeSemanticGroup(objects));
            return null;
        } else {
            return semanticGroup.addIntroducedSO(objects);
        }
        //System.out.println(map);
    }

    public boolean hasName(String id) {
//    	System.out.println("hasName? " + id + " in " + map);
//    	System.out.println("    for type " + type);
    	
        return map.get(id) != null;
    }

    public void addInherited(SemanticMap other) {
        for (Iterator i = other.map.values().iterator(); i.hasNext(); ) {
            SemanticGroup superGroup = (SemanticGroup)i.next();
            String name = superGroup.getName();
            SemanticGroup myGroup = (SemanticGroup)map.get(name);
            if (myGroup == null) {
            	myGroup = superGroup.makeInheritedCopy(this);
            } else {
                myGroup.inheritFrom(superGroup);
            }
            
            if (!myGroup.isEmpty()) map.put(name, myGroup);
        }
    }

    public Collection getAllSemanticObjects() {
        List ret = new LinkedList();
        for (Iterator i = map.values().iterator(); i.hasNext(); ) {
            SemanticGroup group = (SemanticGroup)i.next();
            ret.addAll(group.members);
        }
        return ret;
    }

    public void ensureNoAbstracts() {
        for (Iterator i = getAllSemanticObjects().iterator(); i.hasNext(); ) {
            SemanticObject object = (SemanticObject)i.next();
            if (object.isAbstract()) {
                if (object.getDeclaringType() == getType()) {
                    object.getCorrespondingDec().showError("abstract " +
                            getSemanticObjectTypeName() +
                            " not allowed in concrete type " + getType().toShortString());
                } else {
                    getType().getTypeDec().showError("concrete type " +
                        getType().toShortString() + " must implement " +
                        getSemanticObjectTypeName() + " " + object.toShortString());
                }
            }
        }
    }

    // onlys used by CodePattern
    public SemanticObject findMatchingSemanticObject(SemanticObject matchSemanticObject) {
        for (Iterator i = getAllSemanticObjects().iterator(); i.hasNext(); ) {
            SemanticObject object = (SemanticObject)i.next();
//            System.out.println("  matches? " + object + " == " + matchSemanticObject);
            if (object.conflictsWith(matchSemanticObject)) return object;
        }
        return null;
    }


    protected SemanticObject getNotFoundSemanticObject() { return null; }
    protected String getSemanticObjectTypeName() { return kind; }

    public Type getType() { return type; }

    protected SemanticGroup makeSemanticGroup(SemanticObject object) {
        return new SemanticGroup(this, object);
    }

    protected void showNotFoundError(String id, ASTObject fromWhere) {
        //System.out.println(map);
        getCompiler().showError(fromWhere, "no " + getSemanticObjectTypeName() + " named "
                        + id + " defined in " + type.toShortString());
    }
}
