/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar ...
 */

public abstract class Stmt extends ASTObject {

    /** checks whether control must be live to reach this statement.
        Block statements, for example, need not cause the "unreachable
        statement" error if they are empty.  This is used as part of
        {@link org.aspectj.compiler.base.FlowCheckerPass}. */
    public boolean mustBeLive() { return true; }

    public boolean isBreakable() { return false; }
    public boolean isContinuable() { return false; }

    public boolean isEmpty() { return false; }

    public void requireBlockStmt() {
        this.showError("block statement required");
    }
    
    public void requireStmt() {
        ;
    }

    // ------------------------------
    // Intro: FlowCheckerPass

    private boolean completesNormallyFlag;
    public void setCompletesNormally(boolean b) { completesNormallyFlag = b; }
    public boolean completesNormally() { return completesNormallyFlag; }

    // ------------------------------
    // bcg

    /** The purpose of this procedure is to dispatch to {@link
        #cg}, but to perform an invariant check beforehand to make
        sure the current stack size is zero. This method is final only
        because it seems unlikely that anybody would want to override
        it (they should rather provide a definition for {@link
        #cg}).  If there is a good reason to override this, the
        final modifier can be removed.  */
    final void cgTop(CodeBuilder cb) {
        registerLocation(cb);
        cgStmt(cb);
        if (cb.getStackSize() != 0) {
            showWarning("Sanity check: stack size is " + cb.getStackSize() +
                        " after stmt " + this);
            //display(0);
            cb.emitStackSize();
            cb.setStackSize(0);
        }
    }

    // overridden in break and continue
    protected void registerLocation(CodeBuilder cb) {
        cb.enterLocation(getSourceLocation());
    }

    /** This procedure does the work of generating bytecode.  Its main
        requirement is to leave the stack empty.  This method should
        only be called from {@link #cgTop}.  */
    protected void cgStmt(CodeBuilder cb) {
        throw new RuntimeException("Invalid statement " + this);
    }

    //BEGIN: Generated from @child and @property

    public Stmt(SourceLocation location) {
        super(location);

    }


    public String getDefaultDisplayName() {
        return "Stmt()";
    }

    //END: Generated from @child and @property
}






