/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar (stmt)*
 * @children Stmt stmt
 */

public class Stmts extends ASTObject {
    public Stmt getSingleStmt() {
        if (size() == 0) return new EmptyStmt(getSourceLocation());
        else if (size() == 1) return get(0);
        else return new BlockStmt(getSourceLocation(),this);
    }

    public BlockStmt makeBlockStmt() {
        if (size() == 1 && get(0) instanceof BlockStmt) {
            return ((BlockStmt)get(0));
        } else {
            return getAST().makeBlock(this);
        }
    }


    static class RemoveReturnWalker extends Walker {
        public boolean hasReturn = false;
        public String label;

        public RemoveReturnWalker(JavaCompiler compiler, String label) {
            super(compiler);
            this.label = label;
        }

        public Stmts removeReturns(Stmts stmts) {
            Stmts ret = (Stmts)process(stmts);
            if (hasReturn) {
                final AST ast = getAST();
                return ast.makeStmts(
                  new LabeledStmt(ast.getSourceLocation(), label, ast.makeBlock(stmts)));
            } else {
                return ret;
            }
        }

        public ASTObject process(ASTObject node) {
            if (node instanceof TypeDec) return node;
            if (node instanceof CodeDec) return node;

            if (node instanceof ReturnStmt) {
                hasReturn = true;
                return new BreakStmt(getAST().getSourceLocation(), label);
            }
            return super.process(node);
        }
    }


    public Stmts removeReturns() {
        RemoveReturnWalker retWalker =
                new RemoveReturnWalker(getCompiler(), "AJC$BLOCK"+size());
        return retWalker.removeReturns(this);
    }


    public void unparse(CodeWriter writer) {
        final int N = size;
        int lastBeginLine = -1;
        for (int i = 0; i < N; i++) {
            Stmt stmt = children[i];

            // see if this is separated from the statement above by any extra lines
            // if so add an extra line to the output
//            int thisBeginLine = stmt.getBeginLine();
//            if (lastBeginLine != -1 && thisBeginLine != -1) {
//                if ((thisBeginLine - lastBeginLine) > 1) {
//                    writer.newLine();
//                }
//            }
//            lastBeginLine = thisBeginLine;

            // now actually write the child
            writer.write(children[i]);
            writer.newLine();
        }
    }

    // ------------------------------
    // INTRO: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass walker) {
        final int N = size;
        int j = 0;
        for(int i = 0; i < N; i++) {
            if (! walker.isLive()) break;
            ASTObject s = walker.process(get(i));
            if (s != null && ! (s instanceof EmptyStmt)) {
                set(j++, (Stmt) s);
            }
        }
        for (int i = j; i < N; i++) {
            children[i] = null;
        }
        size = j;
    }

    // ------------------------------
    // bcg

    final void cgStmts(CodeBuilder cb) {
        for (int i = 0, len = size; i < len; i++) {
            children[i].cgTop(cb);
        }
    }

    //BEGIN: Generated from @child and @property
    protected int size;
    public Stmt[] children;

    public Stmts(SourceLocation location, Stmt[] _children) {
        super(location);
        for(int i=0; i<_children.length; i++) {
            if (_children[i] != null) _children[i].setParent(this);
        }
        children = _children;
        size = _children.length;
    }

    public Stmts(SourceLocation location) {
        this(location, new Stmt[] {});
    }

    public Stmts(SourceLocation location, Stmt child1) {
        this(location, new Stmt[] {child1});
    }

    public Stmts(SourceLocation location, Stmt child1, Stmt child2) {
        this(location, new Stmt[] {child1, child2});
    }

    public Stmts(SourceLocation location, Stmt child1, Stmt child2, Stmt child3) {
        this(location, new Stmt[] {child1, child2, child3});
    }

    public ASTObject copyWalk(CopyWalker walker) {
        final int N = size;
        Stmt[] copiedChildren = new Stmt[N];
        int newIndex = 0;
        for(int oldIndex=0; oldIndex<N; oldIndex++) {
            Stmt newChild = (Stmt)walker.process(children[oldIndex]);
            if (newChild != null) copiedChildren[newIndex++] = newChild;
        }
        Stmts ret = new Stmts(getSourceLocation(),copiedChildren);
        ret.size = newIndex;
        ret.setSource(this);
        return ret;
    }

    public ASTObject getChildAt(int childIndex) { return get(childIndex); }
    public void setChildAt(int childIndex, ASTObject child) { set(childIndex, (Stmt)child); }
    public String getChildNameAt(int childIndex) { return "stmt"+childIndex; }

    public int getChildCount() { return size; }
    public int size() { return size; }

    public Stmt get(int index) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        return children[index];
    }

    public void set(int index, Stmt child) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        children[index] = child;
        child.setParent(this);
    }

    public void resize(int newSize) {
        if (newSize > children.length) {
            Stmt[] newChildren = new Stmt[children.length*2 + 1];
            System.arraycopy(children, 0, newChildren, 0, children.length);
            children = newChildren;
        }
        size = newSize;
    }

    public void addAll(Stmts collection) {
        addAll(size, collection);
    }

    public void addAll(int index, Stmts collection) {
        for(int i=0; i<collection.size(); i++) {
            add(index+i, collection.get(i));
        }
    }

    public void add(Stmt child) {
        add(size, child);
    }

    public void add(int index, Stmt child) {
        if (child == null) return;

        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        resize(size+1);

        for(int moveIndex = size-1; moveIndex > index; moveIndex--) {
            children[moveIndex] = children[moveIndex-1];
        }

        children[index] = child;
        child.setParent(this);
    }

    public void remove(int index) {
        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        size -= 1;

        for(int moveIndex = index; moveIndex < size; moveIndex++) {
            children[moveIndex] = children[moveIndex+1];
        }
    }

    public String getDefaultDisplayName() {
        return "Stmts()";
    }

    //END: Generated from @child and @property
}
