/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

import java.util.HashSet;
import java.util.Arrays;

/**
 * @grammar switch (expr) (clauses)
 * @child Expr expr
 * @child SwitchClauses clauses
 */
public class SwitchStmt extends Stmt {

    public boolean isBreakable() { return true; }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        SwitchClauses c = getClauses();
        w.process(getExpr());
        FlowCheckerPass.Vars v = w.getVars();

        for (int i = 0, len = c.size(); i < len; i++) {
            v = findTopLevelVars(v, c.get(i).getStmts());
        }
        w.setVars(v);
        w.enterContext(this);
        boolean hasDefault = false;
        for (int i = 0, len = c.size(); i < len; i++) {
            if (c.get(i).getLabel() == null) { hasDefault = true; }
            w.setLive(true);
            w.setVars(v.join(w.getVars()));
            w.process(c.get(i));
            w.setVars(w.getVars());
        }
        w.leaveContext();
        if (!hasDefault) w.setVars(w.getVars().join(v));

        FlowCheckerPass.Vars bv = w.getBreakVars(this);

        w.setLive(w.isLive() || ! hasDefault || w.isBroken(this));
        w.setVars(w.getVars().join(bv));
    }

    private FlowCheckerPass.Vars findTopLevelVars(FlowCheckerPass.Vars v, Stmts ss) {
        for (int i = 0, len = ss.size(); i < len; i++) {
            Stmt s = ss.get(i);
            if (s instanceof VarDec) {
                v = v.addUnassigned((VarDec) s);
            } else if (s instanceof Decs) {
                v = findTopLevelVars(v, (Decs) s);
            }
        }
        return v;
    }

    private FlowCheckerPass.Vars findTopLevelVars(FlowCheckerPass.Vars v, Decs ds) {
        for (int i = 0, len = ds.size(); i < len; i++) {
            Dec d = ds.get(i);
            if (d instanceof VarDec) {
                v = v.addUnassigned((VarDec) d);
            }
        }
        return v;
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) { walker.pushBlock(); }
    public ASTObject postScope(ScopeWalker walker) { walker.popBlock(); return this; }

    public void checkSpec() {
        Type exprType = expr.getType();
        if (! (exprType instanceof IntishType)) {
            expr.showTypeError(exprType, getTypeManager().intType);
        }
        boolean hasDefault = false;
        HashSet seenLabels = new HashSet();

        for (int i = 0, len = clauses.size(); i < len; i++) {
            SwitchClause c = clauses.get(i);
            Expr label = c.getLabel();
            if (label == null) {
                if (hasDefault) {
                    c.showError("duplicate default label");
                } else {
                    hasDefault = true;
                }
            } else {
                if (label instanceof LiteralExpr) {
                    if (!label.isAssignableTo(exprType)) {
                        label.showTypeError(label.getType(), exprType);
                    } else {
                        if (label instanceof IntLiteralExpr) {
                            Integer intObj =
                                new Integer(((IntLiteralExpr)label).getIntValue());
                            if (seenLabels.contains(intObj)) {
                                label.showError("duplicate label");
                            } else {
                                seenLabels.add(intObj);
                            }
                        } else {
                            label.showError("expected int constant");
                        }
                    }
                } else {
                    label.showError("constant expression required");
                }
            }
        }
    }

    public void unparse(CodeWriter writer) {
        writer.writeKeyword("switch");
        writer.requiredSpace();
        writer.parenExpr(expr);
        writer.openBlock();
        writer.writeChildren(clauses);
        writer.closeBlock();
    }

    // ------------------------------
    // Intro: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        SwitchClauses c = getClauses();
        w.enterContext(this);
        boolean hasDefault = false;
        for (int i = 0, len = c.size(); i < len; i++) {
            if (c.get(i).getLabel() == null) { hasDefault = true; }
            w.setLive(true);
            c.get(i).getStmts().walkCleanup(w);
        }
        w.leaveContext();
        w.setLive(w.isLive() || ! hasDefault || w.isBroken(this));
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        class Pair implements Comparable {
            int i;
            Label l;
            Pair(int i, Label l) { this.i = i; this.l = l; }
            public int compareTo(Object o) {
                int other = ((Pair)o).i;
                
                return (i > other) ? 1 : (i == other) ? 0 : -1;
            }
            public String toString() {
                return "<" + i + ", " + "l" + ">";
            }
        }

        Label endLab = cb.genLabel();
        Label defaultLab = endLab;

        final int len = getClauses().size();

        // do some analysis to build the switch stmt
        boolean hasCases = false;
        Pair[] table = new Pair[len];
        Label[] clauseStarts = new Label[len];
        int ti = 0;
        for (int i = 0; i < len; i++ ) {
            SwitchClause clause = getClauses().get(i);
            IntLiteralExpr caseLabel = (IntLiteralExpr) clause.getLabel();
            Label caseCodeLabel = cb.genLabel();
            clauseStarts[i] = caseCodeLabel;
            if (caseLabel == null) {
                defaultLab = caseCodeLabel;
            } else {
                hasCases = true;
                table[ti++] = new Pair(caseLabel.getIntValue(), caseCodeLabel);
            }
        }

        //System.err.println("Pre-sort " + Arrays.asList(table));
        
        Arrays.sort(table, 0, ti);

        //System.err.println("Post-sort " + Arrays.asList(table));

        int[] keys = new int[ti];
        Label[] labels = new Label[ti];
        for (int i = 0; i < ti; i++) {
            keys[i] = table[i].i;
            labels[i] = table[i].l;
        }

        if (hasCases) {
            getExpr().cgValue(cb);
            cb.emitSwitch(keys, labels, defaultLab);
        } else {
            getExpr().cgEffect(cb);
        }

        // and do the body, finally
        cb.enterNonWindingContext(this, endLab, null);
        for (int i = 0; i < len; i++ ) {
            Label caseCodeLabel = clauseStarts[i];
            Stmts caseCode = getClauses().get(i).getStmts();
            cb.emitLabel(caseCodeLabel);
            caseCode.cgStmts(cb);
        }
        cb.leaveContext();
        cb.emitLabel(endLab);
    }

    // ------------------------------
    // Intro:  FrameLocPass

    // we might allocate something, so we must clean up.

    public void walkFrameLoc(FrameLocPass walker) {
        int start = walker.getfs();
        super.walkFrameLoc(walker);
        walker.setfs(start);
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    /*
    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        boolean seenDefault = false;
        final int len = clauses.size();
        SwitchClause[] clauses = getClauses().children;
        for (int i = 0; i < len; i++ ) {
            SwitchClause clause = clauses[i];
            Expr key = clause.getLabel();
            if (key == null) {
                if (seenDefault) showError("Duplicate default clause");
                else             seenDefault = true;
            } else if (! (key instanceof IntLiteralExpr)) {
                key.showError("Must be an integer constant");
            } else {
                Integer keyObj = new Integer(((IntLiteralExpr)key).getIntValue());
                if (seenKeys.contains(keyObj)) {
                    key.showError("Duplicate case clause");
                } else {
                    seenKeys.add(keyObj);
                }
            }
        }
        return this;
    }
    */

    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }

    protected SwitchClauses clauses;
    public SwitchClauses getClauses() { return clauses; }
    public void setClauses(SwitchClauses _clauses) {
        if (_clauses != null) _clauses.setParent(this);
        clauses = _clauses;
    }

    public SwitchStmt(SourceLocation location, Expr _expr, SwitchClauses _clauses) {
        super(location);
        setExpr(_expr);
        setClauses(_clauses);
    }
    protected SwitchStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        SwitchStmt ret = new SwitchStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        if (clauses != null) ret.setClauses( (SwitchClauses)walker.process(clauses) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        case 1: return clauses;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        case 1: return "clauses";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        case 1: setClauses((SwitchClauses)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "SwitchStmt()";
    }

    //END: Generated from @child and @property
}


