/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar synchronized (expr) {body}
 * @child Expr expr
 * @child Stmt body
 */
public class SynchronizedStmt extends Stmt implements WindingStmt {

    public void unparse(CodeWriter writer) {
        writer.writeKeyword("synchronized");
        writer.requiredSpace();
        writer.parenExpr(expr);
        writer.write(body);
    }

    public void checkSpec() {
        if (!expr.getType().isReferenceType()) {
            expr.showError("reference type required");
        }

        if (expr instanceof NullExpr) {
            expr.showError("null not allowed");
        }

        // should probably be handled by parser
        if (!(body instanceof BlockStmt)) {
            body.showError("block {} required");
        }

    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        w.process(getExpr());
        w.process(getBody());
    }

    // ------------------------------
    // Intro: ByteCodeCleanupPass

    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        if (getBody() instanceof EmptyStmt) {
            return getAST().makeStmt(getExpr()).setSource(getExpr());
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label startBody = cb.genAnchor();
        Label endBody = cb.genAnchor();
        Label startExn = cb.genAnchor();
        Label startFinally = cb.genAnchor();
        Label end = cb.genLabel();

        expr.cgValue(cb);           // put the lock on the stack
        cb.emitDUP();                    // dup it
        cb.emitASTORE(lockLoc);          // store the duplicate
        cb.emitMONITORENTER();           // and lock it.

        cb.emitLabel(startBody);
        cb.enterWindingContext(this, startFinally);
        body.cgTop(cb);
        cb.leaveContext();

        cb.emitLabel(endBody);
        if (body.completesNormally()) {
            cb.emitALOAD(lockLoc);       // load the lock onto the stack
            cb.emitMONITOREXIT();        // unlock the lock
            cb.emitJump(end);            // and we're done
        }

        cb.emitLabel(startExn);
        cb.pushStack(1);                 // there's an exn on the stack
        cb.emitASTORE(retValLoc);        // store exn
        cb.emitJSR(startFinally);        // exit the monitor
        cb.emitALOAD(retValLoc);         // reload the exn
        cb.emitATHROW();                 // rethrow

        cb.emitLabel(startFinally);
        cb.pushStack(1);                 // return addr on the stack
        cb.emitASTORE(retPtrLoc);        // store return addr, shouldn't be lockType
        cb.emitALOAD(lockLoc);           // retrieve lock object
        cb.emitMONITOREXIT();            // unlock
        cb.emitRET(retPtrLoc);           // return

        cb.emitLabel(end);
        cb.addHandler(startBody, endBody, startExn, null); // add default handler
    }
    // end bcg

    // ------------------------------
    // Intro:  FrameLocPass

    // We allocate space for the lock, and enter the block.

    // After analyzing the block, we need to remember to "pre-analyze"
    // the jsr block we're going to have to use to make this work.
    // This means allocating space for the retval or exn, costing 2
    // slots (actually, only (max 1 retvalSlotSize), but we'll allow
    // for 2), and the return pointer, costing one slot.  We save the
    // locations of each of these.

    private int lockLoc;
    private int retValLoc;  // same as exn loc
    private int retPtrLoc;

    public int getRetValLoc() { return retValLoc; }

    public void walkFrameLoc(FrameLocPass walker) {
        boolean inTryFinally = walker.inTryFinally;
        walker.inTryFinally = true;
        int start = walker.getfs();
        lockLoc = walker.allocate(1);
        walker.process(getBody());
        walker.setfs(start + 1);
        retValLoc = walker.allocate(2);
        retPtrLoc = walker.allocate(1);
        walker.inTryFinally = inTryFinally;
        walker.setfs(start);
    }

    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }

    protected Stmt body;
    public Stmt getBody() { return body; }
    public void setBody(Stmt _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }

    public SynchronizedStmt(SourceLocation location, Expr _expr, Stmt _body) {
        super(location);
        setExpr(_expr);
        setBody(_body);
    }
    protected SynchronizedStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        SynchronizedStmt ret = new SynchronizedStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        if (body != null) ret.setBody( (Stmt)walker.process(body) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        case 1: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        case 1: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        case 1: setBody((Stmt)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "SynchronizedStmt()";
    }

    //END: Generated from @child and @property
}


