/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.util.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

import java.util.Iterator;

/**
 * @grammar try {body} (catches)*;
 * @child Stmt body
 * @child CatchClauses catches
 */
public class TryCatchStmt extends TryStmt {

	public void checkSpec() {
		super.checkSpec();
		if (catches.size() == 0) {
       		showError("'try' without 'catch' or 'finally'");
		}
		
		body.requireBlockStmt();
	}
	
    public void unparse(CodeWriter writer) {
        if (catches.size() == 0) {
            writer.write(body);
            return;
        }

        writer.writeKeyword("try");
        writer.optionalSpace();
        writer.write(body);

        writer.writeChildren(catches);
    }

    // ------------------------------
    // INTRO from ASTFixerPass
    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (! isOptional()) return super.postFixAST(fixer);

        // remove any catch clauses that are impossible
        Set exceptions = ExceptionFinder.getPossibleExceptions(getBody(), true);
        
        //System.err.println("TCS: exceptions: " + exceptions);
        //this.display(0);
        
        exceptions.add(getTypeManager().getRuntimeExceptionType());
        exceptions.add(getTypeManager().getErrorType());
        //showError("body might throw: "+exceptions);

        filterCatchClauses(catches, exceptions);
        setIsOptional(false);
        return super.postFixAST(fixer);
    }    
    private void filterCatchClauses(CatchClauses ccs, Set exceptions) {
        //System.out.println("***********************" + exceptions);
        for (int i = 0; i < ccs.size(); i++) {
            CatchClause cc = ccs.get(i);
            //System.out.println(cc.getFormal().getType());
            if (!possibleCatchClause(cc, exceptions)) {
                //System.out.println("    can't throw");
                ccs.remove(i);
                i--;
            }
        }
    }
    private boolean possibleCatchClause(CatchClause cc, Set exceptions) {
        return canThrow(exceptions, cc.getFormal().getType());
    }
    //XXX cut-and-paste from JoinPoint
    private boolean canThrow(Set types, Type baseType) {
        //if (isUncheckedException(baseType)) {
        //    return true;
        //}

        for(Iterator iter = types.iterator(); iter.hasNext(); ) {
            Type checkType = (Type)iter.next();
            if (checkType != null && (checkType.isSubtypeOf(baseType) || 
                baseType.isSubtypeOf(checkType))) {
                return true;
            }
        }
        return false;
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        FlowCheckerPass.Vars beforeV = w.getVars();
        FlowCheckerPass.ESet beforeExns = w.popExns();
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();
        boolean bodyLive = w.isLive();
        FlowCheckerPass.ESet bodyExns = w.popExns();

        FlowCheckerPass.Vars v = w.getVars();  // da
        FlowCheckerPass.Vars tv = w.getTryVars(this); //dua

        FlowCheckerPass.Vars cv =
            FlowCheckerPass.Vars.makeSets(beforeV.getDa(),
                                          v.getDua().inter(tv.getDua()));

        FlowCheckerPass.ESet caughtExns = FlowCheckerPass.ESet.getEmpty();
        FlowCheckerPass.ESet thrownExns = beforeExns;
        FlowCheckerPass.Vars cvExit = v;
        boolean liveSoFar = bodyLive;
        CatchClauses c = getCatches();
        if (c != null) {
            for (int i = 0, len = c.size(); i < len; i++) {
                CatchClause cl = c.get(i);
                NameType exnType = (NameType) cl.getFormal().getType();
                if (! bodyExns.containsSuperOrSub(exnType) &&
                    ! exnType.isUncheckedThrowable() &&
                    exnType != getTypeManager().getExceptionType() &&
                    exnType != getTypeManager().getThrowableType()) {
                    cl.showError("exception " + exnType.getId() +
                                 " is never thrown in body of corresponding try statement");
                    continue;
                }
                if (caughtExns.containsSuper(exnType)) {
                    cl.showError("exception " + exnType.getId() + " has already been caught");
                    continue;
                }
                w.setLive(true);
                w.setVars(cv);
                // here we're just entering and leaving the context to
                // prevent breaks and continues to escape without
                // figuring out whether our (possible) finally
                // terminates normally.  We won't use any of the vars
                // associated with this stmt anymore, they were taken
                // out earlier.
                w.enterContext(this);
                w.process(cl);
                w.leaveContext();
                liveSoFar = liveSoFar | w.isLive();
                cvExit = cvExit.join(w.getVars());
                thrownExns = thrownExns.union(w.popExns());
                caughtExns = caughtExns.add(exnType);
            }
        }
        w.setExns(bodyExns.diff(caughtExns).union(thrownExns).union(beforeExns));
        w.setVars(cvExit);
        w.releasePendingTransfers(this);
        w.setLive(liveSoFar);
    }

    // ------------------------------
    // Intro: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        w.enterContext(this);
        setBody((Stmt) w.process(getBody()));
        w.leaveContext();
        boolean bodyLive = w.isLive();

        boolean liveSoFar = bodyLive;
        CatchClauses c = getCatches();
        for (int i = 0, len = c.size(); i < len; i++) {
            CatchClause cl = c.get(i);
            w.setLive(true);
            w.process(cl);
            liveSoFar = liveSoFar || w.isLive();
        }
        w.setLive(liveSoFar);
    }

    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        if (getBody() instanceof EmptyStmt) {
            return getAST().makeEmptyStmt().setSource(this);
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label startBody = cb.genAnchor();
        Label endBody = cb.genAnchor();
        Label end = cb.genLabel();
        cb.emitLabel(startBody);
        getBody().cgTop(cb);
        cb.emitLabel(endBody);
        if (body.completesNormally()) cb.emitJump(end);
        for (int i = 0, len = catches.size(); i < len; i++ ) {
            CatchClause clause = catches.get(i);
            Label exnStart = cb.genAnchor();
            FormalDec clauseFormal = clause.getFormal();
            Stmt clauseBody = clause.getBody();

            cb.enterBlock();
            cb.enterVar(clauseFormal);
            cb.emitLabel(exnStart);
            cb.pushStack(1); // there's an exn on the stack
            cb.emitASTORE(clauseFormal.getFrameLoc()); // store it
            clauseBody.cgTop(cb);
            if (clauseBody.completesNormally()) cb.emitJump(end);
            cb.addHandler(startBody, endBody, exnStart, (NameType) clauseFormal.getType());
            cb.exitBlock();
        }
        cb.emitLabel(end);
    }    
    
    //BEGIN: Generated from @child and @property
    protected Stmt body;
    public Stmt getBody() { return body; }
    public void setBody(Stmt _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }
    
    protected CatchClauses catches;
    public CatchClauses getCatches() { return catches; }
    public void setCatches(CatchClauses _catches) {
        if (_catches != null) _catches.setParent(this);
        catches = _catches;
    }
    
    public TryCatchStmt(SourceLocation location, Stmt _body, CatchClauses _catches) {
        super(location);
        setBody(_body);
        setCatches(_catches);
    }
    protected TryCatchStmt(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        TryCatchStmt ret = new TryCatchStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (body != null) ret.setBody( (Stmt)walker.process(body) );
        if (catches != null) ret.setCatches( (CatchClauses)walker.process(catches) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return body;
        case 1: return catches;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "body";
        case 1: return "catches";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setBody((Stmt)child); return;
        case 1: setCatches((CatchClauses)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "TryCatchStmt()";
    }
    
    //END: Generated from @child and @property
}


