/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.util.*;

//ASPECT use introduction from crosscuts package
import org.aspectj.compiler.crosscuts.ast.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * A single instance of this class will exist for each type in the world
 * This class provides 6 kinds of methods
 * 1. methods for comparing types for assignability, subtyping, ...
 * 2. methods for looking up methods, fields, and other decs in types
 * 3. methods to determine if a type is of a certain kind (isAspect, isClass, ...)
 * 4. methods for producing reasonable string representations of types
 * 5. generally static methods for manipulating sets of Types
 * 6. a few special purpose methods that need more thought
 */
public abstract class Type extends SemanticObject {
    private ArrayType arrayType = null;
    public synchronized ArrayType getArrayType() {
        if (arrayType == null) {
            arrayType = new ArrayType(getCompiler(), this);
        }
        return arrayType;
    }

    public int getArrayDimCount() { return 0; }

    public Type getBaseComponentType() { return this; }

    // code to handle the type graph
    // I THINK this is the right place for it...
    protected final Set directSubTypes = new HashSet();
    protected final Set directSuperTypes = new HashSet();
    public void addDirectSuperType(Type superType) {
        //XXX awful hack for minimal circularity checking
        if (superType == this) {
            getCompiler().showError(getTypeDec(), getString() + " tries to extend itself");
        }

        directSuperTypes.add(superType);
        superType.directSubTypes.add(this);
    }


    public Type(JavaCompiler compiler) {
        super(compiler);
    }

    protected void showNotFoundError(String id, ASTObject fromWhere, String kind) {
        getCompiler().showError(fromWhere, "no " + kind + "s on " + toShortString());
    }

    public boolean isEffectivelyStatic() {
        return true;
    }

    //XXX should be abstract
    public void ensureBuiltTypeGraph() {
    }

    public void ensureFinishedSignature() {
    }


    public void buildTypeGraph() { }
    public void validateTypeGraph() { }
    public void finishTypeIntroductions() { ensureBuiltTypeGraph(); }
    public void finish() { finishTypeIntroductions(); }
	

    public Collection getDirectSuperTypes() {
        buildTypeGraph();
        return directSuperTypes;
    }

    public Collection getDirectSubTypes() {
        buildTypeGraph();
        return directSubTypes;
    }

    public Type getDeclaringType() {
        return null;
    }

    //XXX too dangerous
    public TypeDec getTypeDec() { return (TypeDec)dec; }

    public PointcutSO getPointcut(String id, ASTObject fromWhere, boolean showError) {
        if (showError) showNotFoundError(id, fromWhere, "pointcut");
        return null; //getWorld().POINTCUT_DEC_NOT_FOUND;
    }

    public Field getField(String id, ASTObject fromWhere, boolean showError) {
        if (showError) showNotFoundError(id, fromWhere, "field");
        return null; //getWorld().FIELD_DEC_NOT_FOUND;
    }

    public Type getInnerType(String id, ASTObject fromWhere, boolean showError) {
        if (showError) showNotFoundError(id, fromWhere, "inner type");
        return getTypeManager().TYPE_NOT_FOUND;
    }

    public Method getMethod(String id, ASTObject fromWhere, Exprs params,
                               boolean showError)
    {
        if (showError) showNotFoundError(id, fromWhere, "method");
        return null; //getWorld().METHOD_DEC_NOT_FOUND;
    }

    public Constructor getConstructor(ASTObject fromWhere, Exprs params,
                                         boolean showError)
    {
        if (showError) showNotFoundError(getPrettyString(), fromWhere, "constructor");
        return null; //getWorld().CONSTRUCTOR_DEC_NOT_FOUND;
    }

    // currently only used by ReceptionsPointcut
    public boolean hasMethodNamed(String id) {
        return false;
    }

    // only used by CodePattern
    public SemanticObject findMatchingSemanticObject(SemanticObject o) {
        return null;
    }

    //XXX replace with above
    /**
     * Will return a declaration in this types signature which matches the signature
     * of dec.
     */
    public Dec findMatchingDec(Dec dec) {
        SemanticObject o = findMatchingSemanticObject(dec.getCorrespondingSemanticObject());
        if (o == null) return null;
        return o.getCorrespondingDec();
    }

    // currently only used by AdviceWeaver for dealing with synthetic reception points
    public Collection getInheritedMethods() {
        return Collections.EMPTY_LIST;
    }

    // used by StructureViewer
    public Collection getInheritedMembers() {
        return Collections.EMPTY_LIST;
    }

    public boolean isCloneable() { return false; }

    public boolean isInterface() { return false; }
    public boolean isClass() { return false; }
    public boolean isAspect() { return false; }

    //??? why isn't this just !isAbstract()???
    public boolean isConcrete() { return isClass(); }
    public boolean isAbstract() { return false; }

    public boolean isInnerType() { return false; }

    /** is this the TypeNotFound sentinel? */
    public boolean isMissing() { return false; }

    public boolean isAnonymous() { return false; }

    public boolean isAssignableFrom(Type other) {
        return other.isSubtypeOf(this);
    }

    public boolean isSubtypeOf(Type other) {
        ensureBuiltTypeGraph();
        //System.out.println(this.getString() + " subtype? " + other.getString());
        if (this.isEquivalent(other)) return true;

        for (Iterator i = directSuperTypes.iterator(); i.hasNext(); ) {
            Type superType = (Type)i.next();
            // don't return true based on missing super types
            if (superType.isMissing()) continue;
            //System.out.println("  supertype: " + superType);
            if (superType.isSubtypeOf(other)) return true;
        }
        return false;
    }

    public boolean isStrictSubtypeOf(Type other) {
        return (this != other) && this.isSubtypeOf(other);
    }

    //XXX these don't handle interfaces correctly at all
    public abstract boolean isCoercableTo(Type other);
//     {
//        return isAssignableFrom(other) || other.isAssignableFrom(this);
//    }

    //??? what extra do I need to worry about here ???
    public boolean isMethodConvertableTo(Type other) {
//        System.out.println("  " + this + " is convertable? " + other + " == " +
//                other.isAssignableFrom(this));
        return other.isAssignableFrom(this);
    }

    public boolean isEquivalent(Type other) {
        return isAnyType() || other.isAnyType() || this == other;
    }

    public boolean dominates(Type other) {
        return this.isStrictSubtypeOf(other);
    }

    public TypeD makeTypeD() {
        //XXXthrow new RuntimeException("shouldn't ever call Type.makeTypeD");
        return new ResolvedTypeD(getAST().getSourceLocation(), this);
    }

    //XXX all five should be abstract
    public Expr getClassExpr() {
        throw new RuntimeException("internal error: no class expr for " + this);
    }
    public Expr fromObject(Expr expr) { return null; }
    public Expr makeObject(Expr expr) { return expr; } //XXX ugly}
    public Expr getNullExpr() { return null; }
    public Type getRefType() { return this; }


    public Type getOutermostType() {
        return this;
    }

    public Type getPrivateCookieType() {
        return getOutermostType();
    }

    public Type getPackageCookieType() {
        return getOutermostType();
    }


    // The next six methods will almost certainly go by the wayside
    // when the type reorganization is done.
    public boolean isObject() {
        if (this == getTypeManager().getObjectType()) return true;
        return false;
    }

    public boolean isUncheckedThrowable() {
        if (this.isSubtypeOf(getTypeManager().getErrorType())) return true;
        if (this.isSubtypeOf(getTypeManager().getRuntimeExceptionType())) return true;
        return false;
    }

    public boolean isVoid() {
        return this == getTypeManager().voidType;
    }

    public boolean isReferenceType() {
        return this instanceof RefType;
    }

    public boolean isString() {
        return this == getTypeManager().getStringType();
    }
    public final boolean isPrimitive() {
        return isNumeric() || isBoolean() || isVoid();
    }
    public final boolean isNumeric() {
        return isIntegral() || this == getTypeManager().floatType || this == getTypeManager().doubleType;
    }
    public final boolean isIntegral() {
        return this == getTypeManager().byteType || this == getTypeManager().charType || this == getTypeManager().shortType ||
               this == getTypeManager().intType || this == getTypeManager().longType;
    }
    public final boolean isBoolean() {
        return this == getTypeManager().booleanType;
    }

    public boolean isAnyType() {
        return false;
    }


    public abstract String toShortString();
    public String getString() { return toShortString(); }
    public String getPrettyString() { return toShortString(); }
    public String getId() { return toShortString(); }
    public String getExtendedId() { return toShortString(); }
    
    /**
     * @return	e.g. qualified type name java.util.Map.Entry will be returned as Map.Entry	
     */
    public String getSourceExtendedId() { return toShortString(); }
    public String getPackageName() { return null; }

    public Type getOutermostLexicalType() { return this; }
    public Type getOutermostBytecodeType() { return getOutermostType(); }

    public boolean isAccessible(ASTObject fromWhere, boolean inBytecode) {
        //System.out.println(this + " : " + getArrayDimCount());
        if (isPrimitive()) return true;
        if (getArrayDimCount() > 0) return getBaseComponentType().isAccessible(fromWhere, inBytecode);
        return super.isAccessible(fromWhere, inBytecode);
    }

    // utility functions on sets
    private static boolean anyParentsInSet(Type type, Set set) {
        for(Iterator iter = set.iterator(); iter.hasNext(); ) {
            Type testType = (Type)iter.next();
            if (type.isStrictSubtypeOf(testType)) return true;
        }
        return false;
    }


    public static boolean hasMatchingType(Collection c, Type t) {
        for (Iterator j = c.iterator(); j.hasNext(); ) {
            Type testType = (Type)j.next();
            if (t.isSubtypeOf(testType)) return true;
        }
        return false;
    }

    public static Set intersect(Set t1, Set t2) {
        Set ret = new HashSet();
        for (Iterator i = t1.iterator(); i.hasNext(); ) {
            Type t = (Type)i.next();
            if (hasMatchingType(t2, t)) {
                ret.add(t);
            }
        }
        
        for (Iterator i = t2.iterator(); i.hasNext(); ) {
            Type t = (Type)i.next();
            if (hasMatchingType(t1, t)) {
                ret.add(t);
            }
        }
        return ret;
    }

    public static Set filterTopTypes(Set set) {
        Set topSet = new HashSet();
        for(Iterator iter = set.iterator(); iter.hasNext(); ) {
            Type testType = (Type)iter.next();
            if (!anyParentsInSet(testType, set)) {
                topSet.add(testType);
            }
        }
        return topSet;
    }

    public static Set filterSubTypes(Type type, Set set) {
        Set topSet = new HashSet();
        for(Iterator iter = set.iterator(); iter.hasNext(); ) {
            Type testType = (Type)iter.next();
            if (testType.isSubtypeOf(type)) {
                topSet.add(testType);
            } else if (type.isSubtypeOf(testType)) {
                topSet.add(type);
            }
        }
        return topSet;
    }

    private void addMySubTypes(Set types) {
        if (types.contains(this)) return;

        types.add(this);

        for (Iterator i = directSubTypes.iterator(); i.hasNext(); ) {
            Type addType = (Type)i.next();
            addType.addMySubTypes(types);
        }
    }


    public Set getSubTypes() {
        Set subTypes = new HashSet();
        addMySubTypes(subTypes);
        return subTypes;
    }


    public static Set addSubTypes(Set types) {
        Set newTypes = new HashSet();

        for(Iterator iter = types.iterator(); iter.hasNext(); ) {
            Type type = (Type)iter.next();
            newTypes.addAll(type.getSubTypes());
        }
        return newTypes;
    }


    private void addMySuperInterfaces(Set types) {
        if (types.contains(this)) return;

        if (this.isInterface()) types.add(this);

        for (Iterator i = directSuperTypes.iterator(); i.hasNext(); ) {
            Type addType = (Type)i.next();
            addType.addMySuperInterfaces(types);
        }
    }
    

    public Set getAllSuperInterfaces() {
        Set newTypes = new HashSet();
        addMySuperInterfaces(newTypes);
        return newTypes;
    }

    public static Set filterConcreteTypes(Set types) {
        Set newTypes = new HashSet();

        for(Iterator iter = types.iterator(); iter.hasNext(); ) {
            Type type = (Type)iter.next();
            if (type.isConcrete()) newTypes.add(type);
        }
        return newTypes;
    }

    public static Set makeTypeSet(Collection typeDecSet) {
        HashSet ret = new HashSet();
        for (Iterator i = typeDecSet.iterator(); i.hasNext(); ) {
            ret.add( ((TypeDec)i.next()).getType() );
        }
        return ret;
    }

    // ------------------------------
    // ASPECT: Constant folding
    
    final LiteralExpr unsupportedFold() {
        throw new RuntimeException("Unsupported fold on " + this);
    }
    abstract LiteralExpr foldCast(LiteralExpr lit);

    abstract LiteralExpr foldPlusOp(LiteralExpr rand);
    abstract LiteralExpr foldMinusOp(LiteralExpr rand);
    abstract LiteralExpr foldBitNotOp(LiteralExpr rand);
    abstract LiteralExpr foldLogNotOp(LiteralExpr rand);

    abstract LiteralExpr foldAddOp(LiteralExpr lit1, LiteralExpr lit2);
    abstract LiteralExpr foldNumericOp(String op, LiteralExpr lit1, LiteralExpr lit2);
    abstract LiteralExpr foldBitwiseOp(String op, LiteralExpr lit1, LiteralExpr lit2);
    abstract LiteralExpr foldShiftOp(String op, LiteralExpr lit1, LiteralExpr lit2);
    abstract LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2);
    abstract LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2);

    // ------------------------------
    // ASPECT: code generation

    // type numbers used only for the NEWARRAY instruction.
    public static final int REF = 0;
    public static final int VOID = 1;
    public static final int BOOLEAN = 4;
    public static final int CHAR = 5;
    public static final int FLOAT = 6;
    public static final int DOUBLE = 7;
    public static final int BYTE = 8;
    public static final int SHORT = 9;
    public static final int INT = 10;
    public static final int LONG = 11;

    public int getTypeIndex() { return REF; }

    // only overridden in ArrayType
    void emitMultiNewarray(CodeBuilder cb, int filledDims) {
        throw new RuntimeException("Unsupported emit on " + this);
    }

    // things overridden based on slot count (i.e., in DoubleType, LongType, VoidType)
    void emitPop(CodeBuilder cb)   { cb.emitPOP(); }
    void emitDup(CodeBuilder cb)   { cb.emitDUP(); }
    void emitDupX1(CodeBuilder cb) { cb.emitDUP_X1(); }
    void emitDupX2(CodeBuilder cb) { cb.emitDUP_X2(); }

    // general type-based emitters
    final void unsupportedEmit() {
        throw new RuntimeException("Unsupported emit on " + this);
    }
    abstract void emitAload(CodeBuilder cb);
    abstract void emitAstore(CodeBuilder cb);
    abstract void emitInstanceof(CodeBuilder cb);
    abstract void emitBitNot(CodeBuilder cb);
    abstract void emitLogNot(CodeBuilder cb);
    abstract void emitNewarray(CodeBuilder cb);
    abstract void emitZero(CodeBuilder cb);
    abstract void emitOne(CodeBuilder cb);
    abstract void emitMinusOne(CodeBuilder cb);
    abstract void emitAdd(CodeBuilder cb);
    abstract void emitNeg(CodeBuilder cb);

    abstract void emitNumericOp(CodeBuilder cb, String op);
    abstract void emitBitwiseOp(CodeBuilder cb, String op);
    abstract void emitShiftOp(CodeBuilder cb, String op); 

    abstract void emitCast(CodeBuilder cb, Type castTo);

    public abstract void emitLoad(CodeBuilder cb, int loc);
    public abstract void emitStore(CodeBuilder cb, int loc);
    public abstract void emitReturn(CodeBuilder cb);

    abstract void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f);
    abstract void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f);

    // ------------------------------
    // "fast" opcode handling

    boolean hasFastEqualityTestOp() { return false; }
    void emitFastEqualityTestOp(CodeBuilder cb, EqualityTestOpExpr e,
                                Label t, Label f) {
        throw new RuntimeException("no fast equality op available");
    }


    boolean hasFastNumericTestOp() { return false; }
    void emitFastNumericTestOp(CodeBuilder cb, NumericTestOpExpr e,
                               Label t, Label f) {
        throw new RuntimeException("no fast numeric op available");
    }

    boolean hasFastIncOp(Expr lhs, int rhs) { return false; }
    void emitFastIncOp(CodeBuilder cb, Expr lhs, int rhs) {
        throw new RuntimeException("no fast inc op available");
    }

    // ------------------------------
    // ASPECT: Virtual machine internal names

    /** The internal name is the name internal to the VM.  Notably,
        this uses / rather than . as a package separator.  This is
        only really defined on {@link NameType} and {@link ArrayType}.  */
    public String getInternalName() {
        throw new RuntimeException("No internal name for " + this);
    }

    /** What is the JVM external name for this type?  Strangely
        enough, this is the same as getString() for primitive and
        name-types, but is the same as getDescriptor() with '.'s
        instead of '/'s for array types.  Extremely wacky, and
        currently only used in desugaring {@link ClassExpr}
        expressions.
    */

    // XXX important:  this must generate the post-inner-munged name
    public String getExternalName() { // overridden in ArrayType and NameType
        return getString();
    }
    

    public List/*String*/ getNamePieces() {
        String s = getString();
        List ret = new ArrayList();
        int lastDot = 0;
        while (true) {
            int nextDot = s.indexOf('.', lastDot);
            if (nextDot == -1) break;
            ret.add(s.substring(lastDot, nextDot));
            lastDot = nextDot+1;
        }
        ret.add(s.substring(lastDot));
        return ret;
    }

    /** The descriptor for this type, as used by classfiles. Note that
        {@link FieldDec} and {@link MethodDec} objects also have
        descriptors, which are built up out of type descriptors. */
    public abstract String getDescriptor();

    /** The number of slots for elements of this type.  This is always
        one, but is overridden for two in {@link LongType} and {@link
        DoubleType}, and zero in {@link VoidType}. */
    public int getSlotCount() { return 1; }

}
