/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.BlockScope;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar modifiers typeD id initializer
  * @child Modifiers modifiers
  * @child TypeD typeD
  * @property String id
  * @child Expr initializer
  *
  */

public class VarDec extends Dec {
    public String getKind() { return "var"; }

    public void checkSpec() {
        if (typeD.getType().isVoid()) {
            typeD.showError("void type not allowed here");
        }
    }

    public boolean isBlank() {
        return (getInitializer() == null);
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        w.setVars(w.getVars().addUnassigned(this));
        // some duplicated code from AssignExpr
        if (getInitializer() != null) {
            w.process(getInitializer());
            FlowCheckerPass.Vars v = w.getVars();
            if (isFinal() && ! v.isDefinitelyUnassigned(this)) {
                w.showVarError(this, this,
                               "Final variable " + getId() +
                               " may already be assigned");
            }
            w.setVars(v.addAssigned(this));
        }
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public void preAssignmentCheck(AssignmentCheckerPass walker) {
        Expr rhs = getInitializer();
        if (rhs != null && rhs instanceof ArrayInitializer) {
            ArrayInitializer ai = (ArrayInitializer) rhs;
            Type ty = getTypeD().getType();
            if (! (ty instanceof ArrayType)) {
                this.showError("Not of array type");
                setInitializer(null);
            } else {
                ArrayType arrayType = (ArrayType) ty;
                setInitializer(getAST().makeNewArray(arrayType, ai));
            }
        }
    }

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        this.checkAssignmentType();
        return this;
    }

    protected void checkAssignmentType() {
        Type ty = getType();
        Expr init = getInitializer();
        if (init != null && ! init.isAssignableTo(ty)) {
            showTypeError(init.getType(), ty);
        }
    }

    //INTRO from ScopeWalker
    public void preScope(ScopeWalker walker) {
        walker.addVarDec(this);
    }

    //XXX not doing this is a great way to check parents
    public boolean isIntroduced() {
        if (isField()) return super.isIntroduced();
        else return false;
    }

    public boolean isField() {
        return false;
    }

    public boolean inStaticContext() { return getParent().inStaticContext(); }

    public Type getType() {
        return getTypeD().getType();
    }

    public Expr getExpr() {
        return null; //new VarExpr(id);
    }


    public String toShortString() { return getTypeD().toShortString() + " " + getId(); }

    public void unparse(CodeWriter writer, boolean isStmt, boolean showType) {
        if (showType) {
            writeModifiers(writer);
            writer.write(typeD);
            writer.requiredSpace();
            writer.write(getBytecodeId());
        } else {
            writer.write(getBytecodeId());
        }

        if (!writer.isOnlySignatures() && initializer != null) {
            writer.writeOp("=");
            if (initializer instanceof Exprs) writer.write("{ ");
            writer.write(initializer);
            if (initializer instanceof Exprs) writer.write(" }");
        }
        if (!writer.isOnlySignatures() && isStmt) writer.closeStmt();

    }

    public void unparse(CodeWriter writer) {
        unparse(writer, true, true);
    }

    // ------------------------------
    // Intro:  FrameLocPass

    public void walkFrameLoc(FrameLocPass walker) {
        setFrameLoc(walker.allocate(getType().getSlotCount()));
    }

    // ------------------------------
    // bcg

    public String getDescriptor() {
        return getType().getDescriptor();
    }
    protected void cgStmt(CodeBuilder cb) {
        cb.enterVar(this);
        if (initializer != null) {
            Type ty = getType();
            initializer.cgValue(cb, ty);
            ty.emitStore(cb, frameLoc);
        }
    }
    private int frameLoc;
    public int getFrameLoc() { return frameLoc; }
    void setFrameLoc(int frameLoc) { this.frameLoc = frameLoc; }

    // ------------------------------
    // convenience constructors

    public VarDec(SourceLocation location, TypeD typeD, String id, Expr initializer) {
        this(location,new Modifiers(location,0), typeD, id, initializer);
    }


    //BEGIN: Generated from @child and @property
    protected Modifiers modifiers;
    public Modifiers getModifiers() { return modifiers; }
    public void setModifiers(Modifiers _modifiers) {
        if (_modifiers != null) _modifiers.setParent(this);
        modifiers = _modifiers;
    }

    protected TypeD typeD;
    public TypeD getTypeD() { return typeD; }
    public void setTypeD(TypeD _typeD) {
        if (_typeD != null) _typeD.setParent(this);
        typeD = _typeD;
    }

    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }

    protected Expr initializer;
    public Expr getInitializer() { return initializer; }
    public void setInitializer(Expr _initializer) {
        if (_initializer != null) _initializer.setParent(this);
        initializer = _initializer;
    }

    public VarDec(SourceLocation location, Modifiers _modifiers, TypeD _typeD, String _id, Expr _initializer) {
        super(location);
        setModifiers(_modifiers);
        setTypeD(_typeD);
        setId(_id);
        setInitializer(_initializer);
    }
    protected VarDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        VarDec ret = new VarDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (typeD != null) ret.setTypeD( (TypeD)walker.process(typeD) );
        ret.id = id;
        if (initializer != null) ret.setInitializer( (Expr)walker.process(initializer) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return typeD;
        case 2: return initializer;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "typeD";
        case 2: return "initializer";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setTypeD((TypeD)child); return;
        case 2: setInitializer((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 3;
    }

    public String getDefaultDisplayName() {
        return "VarDec(id: "+id+")";
    }

    //END: Generated from @child and @property
}
