/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import java.io.*;
import java.util.Enumeration;
import org.aspectj.compiler.base.*;
import org.aspectj.util.CollectionUtil;
import org.aspectj.compiler.base.parser.SourceInfo;

//XXX this doesn't belong here
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.crosscuts.ast.*;


import java.util.*;
import java.lang.reflect.*;

public class World extends CompilerObject {
    private List compilationUnits = new ArrayList();
    private List files = new LinkedList();
    private ArrayList codeDecs = new ArrayList();

    //public final Modifiers EmptyModifiers = new Modifiers(source,0);

    public final SourceLocation source = new DummySourceLocation(getCompiler());

    public final CompilationUnit missingCompilationUnit =
        new CompilationUnit(source, null, null, new Decs(source), new SourceInfo(null, null),true);

    public final FieldDec FIELD_DEC_NOT_FOUND =
        new FieldDec(source,new Modifiers(source,1),
                getTypeManager().anyType.makeTypeD(), "not$found", null);
    public final FieldDec FIELD_NOT_FOUND = FIELD_DEC_NOT_FOUND;

    public final VarDec VAR_DEC_NOT_FOUND =
        new VarDec(source,getTypeManager().anyType.makeTypeD(), "not$found", null);

    public final MethodDec METHOD_DEC_NOT_FOUND =
        new MethodDec(source,new Modifiers(source, 0),getTypeManager().anyType.makeTypeD(),
            "not$found", new Formals(source),
                new TypeDs(source), null);

    public final ConstructorDec CONSTRUCTOR_DEC_NOT_FOUND =
        new ConstructorDec(source,new Modifiers(source, 0), new Formals(source),
                new TypeDs(source), null);

    public final PointcutDec POINTCUT_DEC_NOT_FOUND =
        new PointcutDec(source,new Modifiers(source, 0), "NOT_FOUND", null, null, new EmptyPcd(source));
    public final PointcutDec POINTCUT_NOT_FOUND = POINTCUT_DEC_NOT_FOUND;

    public List getFiles() {
      return files;
    }

    private class CompilationUnitEnumerator implements Enumeration {
        int idx = 0;
        public boolean hasMoreElements() { return idx < compilationUnits.size(); }
        public Object nextElement() { return compilationUnits.get(idx++); }
    }
    private class FileEnumerator implements Enumeration {
        int idx = 0;
        public boolean hasMoreElements() { return idx < files.size(); }
        public Object nextElement() { return files.get(idx++); }
    }

    //public static World getDefault() { return JavaCompiler.TheCompiler.getWorld(); }

    // this function will be used for cleaning
    // world before next compilation (in incremental mode)
    public void cleanup()  {
        files = new LinkedList();
        codeDecs = new ArrayList();
        compilationUnits = new ArrayList();
    }

    public World(JavaCompiler compiler) {
        super(compiler);
    }

//    public Collection getLoadedTypes() {
//        return getTypeManager().getLoadedTypes();
//    }

    public Enumeration enumerateCompilationUnits() {
        return new CompilationUnitEnumerator();
    }

    public List getCompilationUnits() {
        return compilationUnits;
    }

    public void addCompilationUnit(CompilationUnit cu) {
        compilationUnits.add(cu);
    }

    public Enumeration enumerateFiles() {
        return new FileEnumerator();
    }

    public void addFile(String filename) {
        if (files.contains(filename))
            return;
        files.add(filename);
    }
    
    /** returns the next break-label.
     */
    public String genLabel() {
    	return "aj$label" + labelCount++;   
    }
    private int labelCount = 0;
    
	/**
	 * returns all types passed to the compiler in compilation units
	 * 
	 * does not include types loaded from the classpath
	 */
    public List getTypes() {
        ArrayList types = new ArrayList();
        Enumeration en = enumerateCompilationUnits();
        while (en.hasMoreElements()) {
            CompilationUnit cu = (CompilationUnit)en.nextElement();
            types.addAll(cu.getDefinedTypes());
        }
        return types;
    }

//    public TypeDec findTypeDec(String packageName, String className) {
//        return getTypeManager().findTypeDec(packageName, className);
//    }


    private List sourceTypes = new LinkedList();

    public void addSourceType(Type type) {
        //System.out.println("adding source type: " + type.toShortString());
        sourceTypes.add(type);
        getTypeManager().addType(type);
    }

    public Iterator sourceTypesIterator() {
        return new Iterator() {
            int index = 0;
            public boolean hasNext() { return index < sourceTypes.size(); }
            public Object next() { return sourceTypes.get(index++); }
            public void remove() { throw new RuntimeException("can't remove from this"); }
        };
//        return sourceTypes.iterator();
    }


    private LinkedList/*TypeDec*/ buildingTypeGraph = new LinkedList();
    
    public void pushBuildingTypeGraph(TypeDec td) {
        buildingTypeGraph.addLast(td);
        //System.out.println(buildingTypeGraph);
    }
    
    public void popBuildingTypeGraph() {
        buildingTypeGraph.removeLast();
    }
    
    public LinkedList getBuildingTypeGraph() {
        return buildingTypeGraph;
    }


    //static List typeDecPlanners;


    // ASPECT from crosscut package
    //static List typeDecPlanners;
    private List jpPlanners = new LinkedList();

    public void addJpPlanner(JpPlanner planner) {
        if (planner.isStaticPlanner()) {
            staticJpPlanners.add(new TopJpPlanner(planner));
        } else {
            jpPlanners.add(new TopJpPlanner(planner));
        }
    }

    public List getJpPlanners() { return jpPlanners; }

    private List staticJpPlanners = new LinkedList();

    public List getStaticJpPlanners() { return staticJpPlanners; }

    //XXX use insertion order for these planners
    //XXX this is an approximation to the correct behavior
    private Collection typePlanners = new LinkedList(); //new HashSet();

    public void addTypeDecPlanner(TypeDecPlanner planner) {
        if (!typePlanners.contains(planner)) typePlanners.add(planner);
    }
    
    public boolean collectedTypeDecPlanners = false;

    public void runTypeDecPlanners(TypeDec typeDec, int phase) {
        //XXX don't like this restiction
        if (!typeDec.fromSource()) return;

        //XXX add error checking that each phase only runs once
        for (Iterator i = typePlanners.iterator(); i.hasNext(); ) {
            TypeDecPlanner planner = (TypeDecPlanner)i.next();

            //System.out.println("running planner: " + planner + " on " + typeDec);
            planner.plan(typeDec, phase);
        }
    }

    public Map/*MethodDec, CalleeSideCallPoint*/ calleeSideCallPoints = new HashMap();

    private Map sourceToOutput = new HashMap();
    private Map outputToSource = new HashMap();

    public Map getSourceToOutputLineMap() { return sourceToOutput; }
    public Map getOutputToSourceLineMap() { return sourceToOutput; }

    public void clearState() {
        sourceToOutput = null;
        outputToSource = null;
    }
}


