/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bcg;

import org.aspectj.compiler.base.bcg.pool.*;

import java.lang.reflect.Modifier;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;

import java.io.DataOutputStream;
import java.io.IOException;

import java.io.UnsupportedEncodingException;
import org.aspectj.compiler.base.ast.NameType;

/** Lifecycle: creation, addition of attributes (through the
    addXAttribute methods), resolution of constant pool (done in the
    constant pool, when all constant pool entries are set), output.
    It is an error for these steps to occur out of order, but this is
    not currently checked. */

public final class Attributes {
    ConstantPool pool;
    Attributes(ConstantPool pool) {
        this.pool = pool;
    }

    // ------------------------------
    // state
    private /* Attribute */ List attributes = new ArrayList();

    // ------------------------------
    // building methods

    void addSyntheticAttribute() {
        attributes.add(new EmptyAttribute(pool.addUtf8("Synthetic")));
    }
    void addDeprecatedAttribute() {
        attributes.add(new EmptyAttribute(pool.addUtf8("Deprecated")));
    }
    void addSourceFileAttribute(String sourceFile) {
        attributes.add(new PoolAttribute(pool.addUtf8("SourceFile"),
                                         pool.addUtf8(sourceFile)));
    }

	void addSouceDebugExtensionAttribute(String data) {
		attributes.add(new SourceDebugExtensionAttribute(pool.addUtf8("SourceDebugExtension"),
														 data));
	}

    void addConstantValueAttribute(Constant c) {
        attributes.add(new PoolAttribute(pool.addUtf8("ConstantValue"),
                                         c));

    }

    void addCodeAttribute(CodeBuilder cb) {
        attributes.add(new CodeAttribute(pool.addUtf8("Code"), cb));
    }

    void addLineNumberTableAttribute(Label[] pcs, int[] lines) {
        attributes.add(new LineNumberTableAttribute(pool.addUtf8("LineNumberTable"), pcs, lines));
    }

    void addLocalVariableTableAttribute(Label[] starts,
                                        Label[] ends,
                                        Constant[] names,
                                        Constant[] descriptors,
                                        int[] indices) {
        attributes.add(new LocalVariableTableAttribute(pool.addUtf8("LocalVariableTable"), starts, ends, names, descriptors, indices));
    }

    PoolsAttribute exceptionAttribute = null;
    void addToExceptionsAttribute(NameType nameType) {
        if (exceptionAttribute == null) {
            exceptionAttribute = new PoolsAttribute(pool.addUtf8("Exceptions"));
            attributes.add(exceptionAttribute);
        }
        exceptionAttribute.add(pool.addClass(nameType));
    }

    InnerClassesAttribute innerClassesAttribute = null;
    void addToInnerClassesAttribute(NameType nameType) {
        if (innerClassesAttribute == null) {
            innerClassesAttribute = new InnerClassesAttribute(pool.addUtf8("InnerClasses"));
            attributes.add(innerClassesAttribute);
        }
        if (nameType.isInterface()) {
            nameType.getTypeDec().getModifiers().setInterface(true);
        }
        innerClassesAttribute.add(pool.addClassNoInnerCheck(nameType),
                                  nameType.isLocal() ? null : pool.addClass(nameType.getEnclosingType()),
                                  nameType.isAnonymous() ? null : pool.addUtf8(nameType.getId()),
                                  nameType.getTypeDec().getModifiers().getAcceptableInnerClassValue());
    }

    // ------------------------------
    // output methods

    int getByteSize() {
        int s = 2;
        for (int i = 0, len = attributes.size(); i < len; i++) {
            s += ((Attribute) attributes.get(i)).getByteSize();
        }
        return s;
    }

    public void writeTo(DataOutputStream stream) throws IOException {
        stream.writeShort((short) attributes.size());
        for (Iterator i = attributes.iterator(); i.hasNext(); ) {
            ((Attribute) i.next()).writeTo(stream);
        }
    }

    public String toString() {
        String s = "(Attr";
        for (Iterator i = attributes.iterator(); i.hasNext(); ) {
            s += i.next().toString();
        }
        return s + ")";
    }

    // ------------------------------
    // debugging

    void display(int indent, boolean inline) {
        indent += 2;
        System.err.print("(attributes");
        display(attributes, indent, inline);
        System.err.print(")");
    }

    static void display(List l, int indent, boolean inline) {
        for (Iterator i = l.iterator(); i.hasNext(); ) {
            Attribute c = (Attribute) i.next();
            between(indent, inline);
            c.display(indent, inline);
        }
    }

    static void between(int indent, boolean inline) {
        if (inline) {
            System.err.print(" ");
        } else {
            System.err.println();
            for (int s=indent; s >= 0; s--) System.err.print(" ");
        }
    }

    // ------------------------------
    // classes

    private abstract class Attribute {
        Constant name;
        Attribute(Constant name) { this.name = name; }
        abstract void writeTo(DataOutputStream stream) throws IOException;
        abstract void display(int indent, boolean inline);
        abstract int getByteSize();
    }
    private class EmptyAttribute extends Attribute {
        EmptyAttribute(Constant name) { super(name); }
        int getByteSize() { return 6; }
        void writeTo(DataOutputStream stream) throws IOException {
            name.writeIndex(stream);
            stream.writeInt(0);
        }
        public String toString() { return " " + name; }

        // this is a start if we really want to do a disassembly
//          static EmptyAttribute readFrom(DataInputStream stream) throws IOException {
//              int len = stream.readInt();
//              if (len != 0) {
//                  System.err.println("warning: bad empty attribute");
//                  stream.skipBytes(len);
//              }
//          }
        void display(int indent, boolean inline) {
            System.err.print("(" + name + ")");
        }
    }

    private class PoolAttribute extends Attribute {
        Constant data;
        PoolAttribute(Constant name, Constant data) {
            super(name);
            this.data = data;
        }
        int getByteSize() { return 8; }
        void writeTo(DataOutputStream stream) throws IOException {
            name.writeIndex(stream);
            stream.writeInt(2);
            data.writeIndex(stream);
        }
        public String toString() {
            return " (" + name + " " + data + ")";
        }
        void display(int indent, boolean inline) {
            System.err.print("(" + name);
            between(0, true);
            data.display(indent + 2, inline);
            System.err.print(")");
        }
    }

    private class CodeAttribute extends Attribute {
        CodeBuilder cb;
        CodeAttribute(Constant name, CodeBuilder cb) {
            super(name);
            this.cb = cb;
        }
        int getByteSize() { return 2 + 4 + cb.getByteSize(); }
        void writeTo(DataOutputStream stream) throws IOException {
            name.writeIndex(stream);
            stream.writeInt(cb.getByteSize());
            cb.writeTo(stream);
        }
        public String toString() {
            return " (" + name + " " + cb + ")";
        }
        void display(int indent, boolean inline) {
            System.err.print("(" + name);
            between(indent + 2, inline);
            cb.display(indent + 2, inline);
            System.err.print(")");
        }
    }

    private class PoolsAttribute extends Attribute {
        /* Constant */ List data = new ArrayList();
        PoolsAttribute(Constant name) {
            super(name);
        }

        void add(Constant constant) {
            data.add(constant);
        }

        int getByteSize() { return 2 + 4 + 2 + (2 * data.size()); }
        void writeTo(DataOutputStream stream) throws IOException {
            name.writeIndex(stream);
            stream.writeInt(2 * (data.size() + 1));
            stream.writeShort(data.size());
            for (Iterator i = data.iterator(); i.hasNext(); ) {
                ((Constant) i.next()).writeIndex(stream);
            }
        }
        public String toString() {
            String s = " (" + name;
            for (Iterator i = data.iterator(); i.hasNext(); ) {
                s += " " + (Constant) i.next();
            }
            return s + ")";
        }
        void display(int indent, boolean inline) {
            System.err.print("(" + name);
            ConstantPool.display(data, indent + 2, inline);
            System.err.print(")");
        }
    }

    private class LineNumberTableAttribute extends Attribute {
        Label[] pcs;
        int[] lines;
        LineNumberTableAttribute(Constant name, Label[] pcs, int[] lines) {
            super(name);
            this.pcs = pcs;
            this.lines = lines;
        }
        int getByteSize() { return 2 + 4 + 2 + (4 * pcs.length); }
        void writeTo(DataOutputStream stream) throws IOException {
            final int len = pcs.length;
            name.writeIndex(stream);
            stream.writeInt((4 * len) + 2);
            stream.writeShort(len);
            for (int i = 0; i < len; i++ ) {
                stream.writeShort((short)pcs[i].getPc());
                stream.writeShort((short)lines[i]);
            }
        }
        public String toString() {
            String s = " (" + name;
            return s + ")";
        }
        void display(int indent, boolean inline) {
            System.err.print("(" + name);
            for (int i = 0; i < pcs.length; i++) {
                between(indent + 2, inline);
                System.err.print("((pc " + pcs[i].getPc() + ") (line " + lines[i] + "))");
            }
            System.err.print(")");
        }
    }

    private class LocalVariableTableAttribute extends Attribute {
        Label[] starts;
        Label[] ends;
        Constant[] names;
        Constant[] descriptors;
        int[] indices;
        LocalVariableTableAttribute(Constant name,
                                    Label[] starts,
                                    Label[] ends,
                                    Constant[] names,
                                    Constant[] descriptors,
                                    int[] indices) {
            super(name);
            this.starts = starts;
            this.ends = ends;
            this.names = names;
            this.descriptors = descriptors;
            this.indices = indices;
        }
        int getByteSize() { return 2 + 4 + 2 + (10 * starts.length); }
        void writeTo(DataOutputStream stream) throws IOException {
//              System.err.println("(ATTRIBUTES) " +
//                                 java.util.Arrays.asList(starts) + " " +
//                                 java.util.Arrays.asList(ends));
            final int len = starts.length;
            name.writeIndex(stream);
            stream.writeInt((10 * len) + 2);
            stream.writeShort(len);
            for (int i = 0; i < len; i++ ) {
                int start = starts[i].getPc();
                int end = ends[i].getPc();
                stream.writeShort((short)start);
                stream.writeShort((short)(end - start));
                names[i].writeIndex(stream);
                descriptors[i].writeIndex(stream);
                stream.writeShort((short)indices[i]);
            }
        }
        public String toString() {
            String s = " (" + name;
            return s + ")";
        }
        void display(int indent, boolean inline) {
            System.err.print("(" + name);
            for (int i = 0; i < starts.length; i++) {
                between(indent + 2, inline);
                System.err.print("(" + names[i]
                                 + " " + descriptors[i]
                                 + " at " + indices[i]
                                 + " from " + starts[i].getPc()
                                 + " to " + ends[i].getPc()
                                 + ")");
            }
            System.err.print(")");
        }
    }

    private class InnerClassesAttribute extends Attribute {
        List inners = new ArrayList();
        List outers = new ArrayList();
        List names = new ArrayList();
        List modifiers = new ArrayList();

        InnerClassesAttribute(Constant name) {
            super(name);
        }

        void add(ClassConstant inner,
                 ClassConstant outer,
                 Utf8Constant  name,
                 int           mods) {
            inners.add(inner); outers.add(outer);
            names.add(name); modifiers.add(new Integer(mods));
        }

        int getByteSize() { return 2 + 4 + 2 + (8 * inners.size()); }
        void writeTo(DataOutputStream stream) throws IOException {
            name.writeIndex(stream);
            final int len = inners.size();
            stream.writeInt((8 * len) + 2);
            stream.writeShort(len);
            for (int i = 0; i < len; i++) {
                ConstantPool.writeIndex((ClassConstant) inners.get(i), stream);
                ConstantPool.writeIndex((ClassConstant) outers.get(i), stream);
                ConstantPool.writeIndex((Utf8Constant) names.get(i), stream);
                stream.writeShort(((Integer) modifiers.get(i)).intValue());
            }
        }
        public String toString() {
            return "(innerclassattribute)";
        }
        void display(int indent, boolean inline) {
            System.err.print("(" + name);
            final int len = inners.size();
            for (int i = 0; i < len; i++) {
                between(indent + 2, inline);
                System.err.print("(class "); ((ClassConstant) inners.get(i)).display(0, true);
                between(indent + 4, inline);
                System.err.print("(inside "); ((ClassConstant) outers.get(i)).display(0, true);
                System.err.print(")");
                between(indent + 4, inline);
                System.err.print("(knownAs "); ((Utf8Constant) names.get(i)).display(0, true);
                System.err.print(")");
                between(indent + 4, inline);
                System.err.print("(modifiers " + Modifier.toString(((Integer) modifiers.get(i)).intValue()) + ")");
                System.err.print(")");
            }
            System.err.print(")");
        }
    }

    private class SourceDebugExtensionAttribute extends Attribute {
    	byte[] data;
    	SourceDebugExtensionAttribute(Constant name, String data) {
    		super(name);
    		try {
    		this.data = data.getBytes("UTF-8");
    		} catch (java.io.UnsupportedEncodingException e) {}
    	}
    	int getByteSize() { return 2 + 4 + data.length; }
        void writeTo(DataOutputStream stream) throws IOException {
            name.writeIndex(stream);
            final int len = data.length;
            stream.writeInt(len);
            stream.write(data);
        }
        public String toString() {
			try {
            	return "(SourceDebugExtensionAttribute " + "\"" + new String(data,"UTF-8") + "\")";
			} catch (UnsupportedEncodingException e) { return null; }
        }

		void display(int indent, boolean inline) {
			System.err.println(this);
		}

	}

    // ------------------------------
    // disassembly methods

    void readFrom(java.io.DataInputStream stream) throws java.io.IOException {
        int count = stream.readUnsignedShort();
        for (int i = count; i > 0; i--) {
            Utf8Constant name = (Utf8Constant) pool.get(stream.readUnsignedShort());
            int len = stream.readInt();
            String id = name.value;
            if ("ConstantValue".equals(id)) {
                attributes.add(new PoolAttribute(name,
                                                 pool.get(stream.readUnsignedShort())));
            } else if ("SourceFile".equals(id)) {
                attributes.add(new PoolAttribute(name,
                                                 pool.get(stream.readUnsignedShort())));
            } else if ("Code".equals(id)) {
                attributes.add(new EmptyAttribute(name));
                stream.skipBytes(len);
            } else if ("Exceptions".equals(id)) {
                PoolsAttribute a = new PoolsAttribute(name);
                int size = stream.readUnsignedShort();
                for (int s = size - 1; s >= 0; s--) {
                    a.add(pool.get(stream.readUnsignedShort()));
                }
                attributes.add(a);
            } else if ("InnerClasses".equals(id)) {
                InnerClassesAttribute a = new InnerClassesAttribute(name);
                int size = stream.readUnsignedShort();
                for (int s = size - 1; s >= 0; s--) {
                    a.add((ClassConstant) pool.get(stream.readUnsignedShort()),
                          (ClassConstant) pool.get(stream.readUnsignedShort()),
                          (Utf8Constant) pool.get(stream.readUnsignedShort()),
                          stream.readUnsignedShort());
                }
                attributes.add(a);
            } else {
                // Code
                // LineNumberTable
                // LocalVariableTable
                attributes.add(new EmptyAttribute(name));
                stream.skipBytes(len);
            }
        }
    }
}
