/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bytecode;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.*;


import java.io.*;
import java.util.zip.*;
import java.util.StringTokenizer;

import java.util.*;

abstract public class ClassManager extends CompilerObject {
    private ASTConnection astConnection = null;

    public ClassManager(JavaCompiler compiler) {
        super(compiler);
    }

    private ASTConnection getASTConnection() {
        if (astConnection == null) {
            astConnection = new ASTConnection(getCompiler());
        }
        return astConnection;
    }

    public abstract TypeDec findTypeDec(String name);
    public abstract ClassManager makeSubPackageManager(String subPackageName);
    
    /**
     * returns all possible type names visible to this class manager
     * 
     * they are encoded as a List of elements (broken on '.' and '$')
     */
    public abstract Collection/*List<String>*/ getAllPossibleTypeNames();

    public TypeDec makeTypeDec(String name, InputStream inputStream) {
        //InputStream inputStream = classPathManager.findClass(name);
        if (inputStream == null) return null;

        try {
            ClassFile classFile = new ClassFile();
            try {
                DataInputStream dataStream =
                        new DataInputStream(prereadInputStream(inputStream));
                classFile.readFrom(dataStream);
            } catch (IOException e) {
                System.err.println("ioexception for: "+name+", "+inputStream);
                e.printStackTrace();
                return null;
            } catch (InvalidClassFileException icfe) {
                System.err.println("bad class file for: "+name+", "+inputStream);
                icfe.printStackTrace();
                return null;
            }

            if (getCompiler().getOptionVerbose()) {
                getCompiler().showMessage("loading: "+name);
            }

            TypeDec typeDec = getASTConnection().makeTypeDec(classFile);
            
            return typeDec;
        } finally {
            try {
                inputStream.close();
            } catch (IOException e) {}
        }
    }
    
    public Collection/*String*/ getPathStrings() {
        Collection ret = new LinkedList();
        ret.add(getPathString());
        return ret;
    }
    
    protected String getPathString() {
        throw new UnsupportedOperationException("no simple string");
    }
        
    
    //!!! assume that classfiles are never larger than 500K here
    //!!! should degrade gracefully if this is violated
    private static byte[] bytes = new byte[500*1024];
    private InputStream prereadInputStream(InputStream stream) throws IOException {
        int index = 0;
        while (true) {
            int nread = stream.read(bytes, index, bytes.length-index);
            if (nread == -1) break;
            index = index+nread;
        }
        stream.close();

        return new ByteArrayInputStream(bytes, 0, index);
    }
    
    public List filenameToClassName(String filename) {
		int last = 0;
		List ret = new ArrayList();
		for (int i=0, len = filename.length(); i < len; i++) {
			char ch = filename.charAt(i);
			if (ch == '.' || ch == '/' || ch == '\\' || ch == '$') {
				ret.add(filename.substring(last, i));
				last = i+1;
			}
		}
		// The last .class doesn't get added to the list, and this is good
		
		return ret;
	}  
}
