/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.parser;

import org.aspectj.compiler.base.ast.*;

import org.aspectj.compiler.base.*;

import java.util.*;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.io.*;

class Comment extends Token {
    public Comment(String text) {
        super(MULTI_LINE_COMMENT, text, false);
        //System.out.println("comment: " + text);
    }
}
/**
class FormalComment extends Token {
    public FormalComment(String text) {
        super(FORMAL_COMMENT, text, false);
    }
}
**/

class Literal extends Token {
    public Literal(String image, int kind) {
        super(kind, image, false);
    }
}


class LongLiteral extends Literal {
    public int radix;
    public LongLiteral(String image, int radix) {
        super(image, INTEGER_LITERAL);
        this.radix = radix;
    }

    public String toString() {
        return "LongLiteral("+image+")";
    }

    public long parseLong(SourceLocation source, boolean negative) {
        long ret = 0;
        String s = image;
        if (radix == 10) {
            try {
                if (negative) s = "-"+s;
                ret = Long.parseLong(s, radix);
            } catch (NumberFormatException nfe) {
                //!!! handle MIN_VALUE
                //System.out.println(nfe);
                source.showError("invalid long decimal literal");
            }
        } else {
            final int N = s.length();
            if (N == 0 && radix == 16) {
                source.showError("0x must be followed by digits");
            }
            final int shift = radix == 16 ? 4 : 3;
            final long limit = ~(0xffffffffffffffffL >>> shift);
            for (int i=0; i < N; i++) {
                if ((ret & limit) != 0) {
                    source.showError("long literals must be no more than 64-bit");
                    break;
                }
                ret <<= shift;
                int digit = Character.digit(s.charAt(i), radix);
                ret = ret | digit;
            }
            if (negative) ret = -ret;
        }
        return ret;
    }


    public Expr getExpr(SourceLocation source) {
        return new LongLiteralExpr(source, parseLong(source, false));
    }
    public Expr getNegativeExpr(SourceLocation source) {
        return new LongLiteralExpr(source, parseLong(source, true));
    }
}

class IntegerLiteral extends Literal {
    public int radix;
    public IntegerLiteral(String image, int radix) {
        super(image, INTEGER_LITERAL);
        this.radix = radix;
        //value = (int)LongLiteral.parse(image, radix);
    }

    public String toString() {
        return "IntegerLiteral("+image+")";
    }

    public int parseInt(SourceLocation source, boolean negative) {
        int ret = 0;
        String s = image;
        if (radix == 10) {
            if (s.length() > 1 && s.charAt(0) == '0') {
                source.showError("invalid octal literal");
                return 0;
            }
            
            try {
                if (negative) s = "-"+s;
                ret = Integer.parseInt(s, radix);
            } catch (NumberFormatException nfe) {
                //!!! handle MIN_VALUE
                //System.out.println(nfe);
                source.showError("invalid int decimal literal");
            }
        } else {
            final int N = s.length();
            if (N == 0 && radix == 16) {
                source.showError("0x must be followed by digits");
            }
                    
            final int shift = radix == 16 ? 4 : 3;
            final int limit = ~(0xffffffff >>> shift);
            for (int i=0; i < N; i++) {
                if ((ret & limit) != 0) {
                    source.showError("integer literal must be no more than 32-bit");
                    break;
                }
                ret <<= shift;
                int digit = Character.digit(s.charAt(i), radix);
                ret = ret | digit;
            }
            if (negative) ret = -ret;
        }
        return ret;
    }


    public Expr getExpr(SourceLocation source) {
        return new IntLiteralExpr(source, parseInt(source, false));
    }
    public Expr getNegativeExpr(SourceLocation source) {
        return new IntLiteralExpr(source, parseInt(source, true));
    }
}

class FloatLiteral extends Literal {
    public float value;
    public FloatLiteral(String image) {
        super(image, FLOATING_POINT_LITERAL);           
    }

    public String getValueImage() {
        if (image.endsWith("F") || image.endsWith("f")) {
            return image.substring(0, image.length()-1);
        } else {
            return image;
        }
    }

    public Expr getExpr(SourceLocation source) {
    	Expr ret;
        try {
        	value = Float.valueOf(getValueImage()).floatValue();
        	ret = new FloatLiteralExpr(source,value);
	        if (value == 0.0f) {
	            BigDecimal bd = new BigDecimal(getValueImage());
	            if (bd.compareTo(new BigDecimal("0")) != 0) {
	                ret.showError("non-zero literal rounds to 0");
	            }
	        } else if (value == Float.POSITIVE_INFINITY || 
	                   value == Float.NEGATIVE_INFINITY)
	        {
	            ret.showError("floating point literal rounds to IEEE infinity");
	        }
		} catch (NumberFormatException nfe) {
			source.showError("illegal floating point literal format");
			ret = new FloatLiteralExpr(source,0.0f);
        }        
        return ret;
    }
}

class DoubleLiteral extends Literal {
    public double value;
    public DoubleLiteral(String image) {
        super(image, FLOATING_POINT_LITERAL);      	
    }
    public String getValueImage() {
        if (image.endsWith("D") || image.endsWith("d")) {
            return image.substring(0, image.length()-1);
        } else {
            return image;
        }
    }

    public Expr getExpr(SourceLocation source) {
        Expr ret;
        
        try {
        	value = Double.valueOf(getValueImage()).doubleValue();
        	ret = new DoubleLiteralExpr(source,value);
	        if (value == 0.0d) {
	            BigDecimal bd = new BigDecimal(getValueImage());
	            if (bd.compareTo(new BigDecimal("0")) != 0) {
	                ret.showError("non-zero literal rounds to 0");
	            }
	        } else if (value == Double.POSITIVE_INFINITY || 
	                   value == Double.NEGATIVE_INFINITY)
	        {
	            ret.showError("floating point literal rounds to IEEE infinity");
	        }
		} catch (NumberFormatException nfe) {
			source.showError("illegal floating point literal format");
			ret = new DoubleLiteralExpr(source,0.0);
        } 
        return ret;
    }
}

class CharacterLiteral extends Literal {
    public char value;
    public CharacterLiteral(char value) {
        super("\""+value+"\"", CHARACTER_LITERAL);
        this.value = value;
    }

    public static void quoteCharacter(char ch, StringBuffer buf) {
        //if (ch >= 32 && ch <= 127) {
        //    buf.append(ch); return;
        //}

        switch (ch) {
        case '\b': buf.append("\\b"); return;
        case '\t': buf.append("\\t"); return;
        case '\n': buf.append("\\n"); return;
        case '\f': buf.append("\\f"); return;
        case '\r': buf.append("\\r"); return;
        case '\"': buf.append("\\\""); return;
        case '\'': buf.append("\\\'"); return;
        case '\\': buf.append("\\\\"); return;
        }

        buf.append(ch);

        //buf.append("\\u");
        //String hex = Integer.toHexString(ch);
        //for(int i=0; i<(4-hex.length()); i++) buf.append('0');
        //buf.append(hex);
    }

    public String toJavaSource() {
        StringBuffer buf = new StringBuffer("'");
        quoteCharacter(value, buf);
        buf.append("'");

        //System.out.println("char: "+buf+", "+(int)value);

        return buf.toString();
    }

    public Expr getExpr(SourceLocation source) {
        return new IntLiteralExpr(source,source.getCompiler().getTypeManager().charType, toJavaSource(), (int)value);
    }
}

class StringLiteral extends Literal {
    public String value;
    public StringLiteral(String value) {
        super(value, STRING_LITERAL);
        this.value = value;
    }

    public String toString() {
        return "StringLiteral("+value+")";
    }

    public Expr getExpr(SourceLocation source) {
        return new StringLiteralExpr(source,value);
    }
}


public class JavaTokenizer extends CompilerObject implements JavaConstants {
    void cleanup(boolean fully) {
        specialToken = null;
        text = null;
    }

    char[] text;
    int index = 0;

    int startIndex = 0;
    //int startLine = 0;
    //int startLineStart = 0;

    //int line = 0;
    //int lineStart = 0;
    public Token specialToken = null;

    protected final void newLine() {
        //line++;
        //lineStart = index;
    }

    String getImage() {
        return new String(text, startIndex, index-startIndex);
    }

    String getImage(int startOffset, int endOffset) {
        return new String(text, startIndex+startOffset,
                          index-(startIndex+startOffset)+endOffset);
    }

    public final Token getNextToken() {
        return nextToken();
    }


    public Token nextToken() {
        Token token;

        try {
            token = internalNextToken();
        } catch (ArrayIndexOutOfBoundsException e) {
            token = new Token(EOF, "EOF", false);
        } catch (ParseException pe) {
            throw pe;
        } catch (RuntimeException re) {
            System.err.println("rt error at: "+index+", "+sourceInfo.getLine(index));
            throw re;
        } catch (Error er) {
            System.err.println("error at: "+index+", "+sourceInfo.getLine(index));
            throw er;
        }

        token.startPosition = startIndex;
        token.endPosition = index;
        token.specialToken = specialToken;
        specialToken = null;

        startIndex = index;
        //startLineStart = lineStart;
        //startLine = line;
        return token;
    }

    private final char peekChar() {
        return peekChar(0);
    }

    private final char peekChar(int offset) {
        //if ((index + offset) >= nChars) throw new ArrayIndexOutOfBoundsException();
        return text[index+offset];
    }

    private final char eatChar() {
        //if (index >= nChars) throw new ArrayIndexOutOfBoundsException();
        return text[index++];
    }

    private final void skipChars(int n) {
        index += n;
    }

    private final void skipChar() {
        index++;
    }


    public Token internalNextToken() {
        for (;;) {
            //System.out.println(""+index+": "+text[index]);
            switch(peekChar()) {
            // line terminators (\r is handled previously)
            case '\n':
                eatChar();
                startIndex++;
                //newLine();
                break;

            // white space
            case ' ':
            case '\t':
            case '\f':
                eatChar();
                startIndex++;
                break;

            // comments
            case '/':
            {
                char peekChar = peekChar(1);
                Comment comment;
                if (peekChar == '/') {
                    comment = getSingleLineComment();
                } else if (peekChar == '*') {
                    try {
                        if (peekChar(2) == '*') {
                            comment = getDocumentationComment();
                        } else {
                            comment = getTraditionalComment();
                        }
                    } catch (ArrayIndexOutOfBoundsException e) {
                        throw new ParseException(startIndex, sourceInfo,
                                                 "unclosed comment");
                    }   
                } else {
                    return getToken();
                }
                if (comment != null) {
                    comment.specialToken = specialToken;
                    comment.startPosition = startIndex;
                    comment.endPosition = index;
                    specialToken = comment;
                }

                startIndex = index;
                break;
            }

            // numeric literals
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
                return getNumericLiteral();

            case '.':
            {
                char peekChar = peekChar(1);
                if (peekChar >= '0' && peekChar <= '9') {
                    return getNumericLiteral();
                } else {
                    return getToken();
                }
            }

            // string literals
            case '\"':
                return getStringLiteral();

            case '\'':
                return getCharacterLiteral();

            default:

                // lump everything else together
                // separators, operators, boolean literals, null, keywords, identifiers
                return getToken();
            }
        }
    }

    public Token scanTillMatched(int kind) {
        int depth = 1;    // assume left brace was already eated
        for (;;) {
            switch(eatChar()) {
            // line terminators (\r is handled previously)
            // and white space
            case '\n':
            case ' ':
            case '\t':
            case '\f':
                continue;

            // comments
            case '/':
            {
                char peekChar = peekChar();
                if (peekChar == '/') {
                    while (eatChar() != '\n')
                        ;
                } else if (peekChar == '*') {
                        // skip the initial /*
                        skipChar();
                    // read until */ is found
                    while (true) {
                        if (eatChar() == '*' && peekChar() == '/')
                            break;
                    }
                    skipChar();
                }
                continue;
            }

            // string literals
            case '\"':
                skipStringLiteral();
                continue;

            case '\'':
                skipCharacterLiteral();
                continue;

            default:
                continue;

            // Now, the most intresting port - braces and nested depth
            case '{':
                if (kind == RBRACE)
                    depth++;
                continue;
            case '}':
                if (kind == RBRACE) {
                    depth--;
                    if (depth == 0) {
                        Token token = new Token(RBRACE, "}", false);
                        token.startPosition = index;
                        token.endPosition = index;
                        token.specialToken = null;
                        specialToken = null;
                        startIndex = index;
                        return token;
                    }
                }
                continue;
            }
        }
    }

    private Comment getSingleLineComment() {
        // skip the initial //
        skipChars(2); //index += 2;

        // read to end of line
        while (eatChar() != '\n') {}

        //newLine();

        Comment ret = new Comment(getImage());
        ret.kind = SINGLE_LINE_COMMENT;
        return ret;
    }

    private Comment getTraditionalComment() {
        // skip the initial /* or /** if documentation comment
        skipChars(2); //index += 2;

        // read until */ is found
        while (true) {
            char ch = eatChar();
            if (ch == '*' && peekChar() == '/') break;
            // newLine();
        }
        eatChar(); //index++;

        Comment ret = new Comment(getImage());
        ret.kind = MULTI_LINE_COMMENT;
        return ret;
    }

    private Comment getDocumentationComment() {
        if (peekChar(3) == '/') {
            skipChars(4);
            Comment ret = new Comment("");
            ret.kind = FORMAL_COMMENT;
            return ret;
        } else {
            skipChars(1);
            Comment ret = getTraditionalComment();
            ret.kind = FORMAL_COMMENT;
            return ret;
        }
    }

    private ParseException tokenError() {
        return tokenError("illegal character");
    }

    private ParseException tokenError(String message) {
        return new ParseException(index, sourceInfo, message);
    }

    private Token getNumericLiteral() {
        char ch = eatChar(); //text[index++];
        int radix = 10;
        int startOffset = 0;

        if (ch == '0') {
            char peekChar = peekChar(); //text[index];
            if (peekChar == 'x' || peekChar == 'X') {
                eatChar(); //index += 1;
                ch = eatChar(); //text[index++];
                radix = 16;
                startOffset = 2;
            } else if (peekChar >= '0' && peekChar <= '7') {
                radix = 8;
                startOffset = 1;
            } else {
                startOffset=0;
            }
        }

        // we know that it's either decimal or float literal
        // read through part that matches a decimal literal
        while (true) {
            if (Character.digit(ch, radix) == -1) break;
            ch = eatChar(); //text[index++];
        }

        // now decide if this was a decimal or a float literal
        switch(ch) {
        case 'l': case 'L': return new LongLiteral(getImage(startOffset, -1), radix);
        case 'f': case 'F':
        case 'd': case 'D':
        case 'e': case 'E': case '.':
            if (radix == 16) {
                throw tokenError();
            }
            radix = 10;
            skipChars(-1); //index--;
            return finishFloatLiteral();
        }
        skipChars(-1); //index--;
        return new IntegerLiteral(getImage(startOffset, 0), radix);
    }

    private Token finishFloatLiteral() {
        boolean seenDot = false;
        boolean seenExponent = false;
        loop: while (true) {
            switch(eatChar()) {
            case 'f': case 'F':
                return new FloatLiteral(getImage()); //0, -1));
            case 'd': case 'D':
                return new DoubleLiteral(getImage()); //0, -1));
            case 'e': case 'E':
                if (seenExponent) {
                    skipChars(-1); //index--;
                    break loop;
                }
                seenExponent = true;
                if (peekChar() == '+') skipChars(1); //index += 1;
                if (peekChar() == '-') skipChars(1); //index += 1;
                break;
            case '.':
                if (seenDot || seenExponent) {
                    skipChars(-1); //index--;
                    break loop;
                }
                seenDot = true;
                break;
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
                break;
            default:
                skipChars(-1); //index--;
                break loop;
            }
        }
        return new DoubleLiteral(getImage());
    }

    private Token getStringLiteral() {
        // skip the initial "
        skipChars(1); //index += 1;
        // build up the "real" value of the string in here
        //??? could optimize by doing this more lazily
        StringBuffer contents = new StringBuffer();

        while (true) {
            char ch = eatChar();
            switch (ch) {
            case '\"':
                return new StringLiteral(contents.toString());
            case '\\':
                ch = processCharacterEscape();
                break;
            case '\n':
                throw tokenError("newline not allowed in string literal");
            }
            contents.append(ch);
        }
    }

    private void skipStringLiteral() {
        // don't skip initial
        while (true) {
            char ch = eatChar();
            switch (ch) {
            case '\"':
                return;
            case '\\':
                ch = processCharacterEscape();
                break;
            case '\n':
                throw tokenError("newline not allowed in string literal");
            }
        }
    }

    private Token getCharacterLiteral() {
        // skip the initial '
        skipChars(1);
        return new CharacterLiteral(getCharacterLiteralInternal());
    }

    private void skipCharacterLiteral() {
        // don't skip initial
        getCharacterLiteralInternal();
    }

    private char getCharacterLiteralInternal() {
        char ch = eatChar();

        switch (ch) {
        case '\'':
            throw tokenError("empty character literal");
        case '\\':
            ch = processCharacterEscape();
            break;
        case '\n':
            throw tokenError("newline not allowed in character literal");
        }

        if (eatChar() != '\'') {
            throw tokenError("character literal too long");
        }
        return ch;
    }

    private char processCharacterEscape() {
        char ch = eatChar();
        switch (ch) {
        case 'b': return '\b';
        case 't': return '\t';
        case 'n': return '\n';
        case 'f': return '\f';
        case 'r': return '\r';
        case '\"': return '\"';
        case '\'': return '\'';
        case '\\': return '\\';
        }
        if (ch < '0' || ch > '7') {
            throw tokenError("octal escape must use values 0-7");
        }

        // this giberrish handles octal escapes
        char ch1 = peekChar();
        if (ch1 >= '0' && ch1 <= '7') {
            eatChar();
            char ch2 = peekChar(); //text[index+1];
            if (ch < '4' && (ch2 >= '0' && ch2 <= '7')) {
                eatChar(); //index++;
                //if (ch > '3') throw tokenError("illegal value for octal escape");
                return (char)((ch2-'0')+(ch1-'0')*8+(ch-'0')*64);
            } else {
                return (char)((ch1-'0') + (ch-'0')*8);
            }
        } else {
            return (char)(ch-'0');
        }
    }

    // use a regex-style table here for extensibility
    private Token getToken() {
        // first follow the FSM
        State state = beginState;

        while (!state.isEndState()) {
            state = state.forCharacter(eatChar());
            if (state == null) return getIdentifier();
        }
        skipChars(-1); //index -= 1;

        if (state.isKeyword() && Character.isJavaIdentifierPart(peekChar())) {
            while(Character.isJavaIdentifierPart(eatChar())) { }
            skipChars(-1); //index -= 1;
            return new Token(IDENTIFIER, getImage(), true);
        }

        return ((EndState)state).getToken();
    }

    private Token getIdentifier() {
        index = startIndex;

        if (!Character.isJavaIdentifierStart(eatChar())) {
            throw tokenError("identifier expected");
        }

        while(Character.isJavaIdentifierPart(eatChar())) { }
        skipChars(-1);
        return new Token(IDENTIFIER, getImage(), true);
    }

    //XXX awful hack
    private static State beginState = new ArrayState();

    static Map knownOperators = new HashMap();
    private static final int OFFSET = 100;

    public static int addOperator(String image) {
        return addOperator(-1, image);
    }

    public static int addOperator(int kindV, String image) {
        Integer kind = (Integer)knownOperators.get(image);
        if (kind != null) return kind.intValue();

        if (kindV == -1) kindV = knownOperators.size()+OFFSET;
        kind = new Integer(kindV);
        EndState endState = new EndState(kind.intValue(), image, isIdentifier(image));
        knownOperators.put(image, kind);

        beginState.addEndState(image, 0, endState);

        return kind.intValue();
    }

    public static final boolean isIdentifier(String image) {
        int n = image.length();
        if (n == 0) return false;
        if (!Character.isJavaIdentifierStart(image.charAt(0))) return false;
        for(int i=1; i<n; i++) {
            if (!Character.isJavaIdentifierPart(image.charAt(i))) return false;
        }
        return true;
    }


    private abstract static class State {
        public State forCharacter(char ch) {
            return null;
        }

        public boolean isEndState() {
            return false;
        }

        public boolean isKeyword() {
            return false;
        }

        public Object getData() {
            return null;
        }

        public abstract State addEndState(String image, int index, EndState endState);
    }

    private static class EndState extends State {
        private int kind;
        private String image;
        private boolean isIdentifier;

        public EndState(int kind, String image, boolean isIdentifier) {
            this.kind = kind;
            this.image = image;
            this.isIdentifier = isIdentifier;
        }

        public Token getToken() {
            return new Token(kind, image, isIdentifier);
        }

        public State forCharacter(char ch) {
            return null;
        }

        public boolean isEndState() {
            return true;
        }

        public boolean isKeyword() {
            return isIdentifier;
        }

        public State addEndState(String image, int index, EndState endState) {
            if (image.length() >= index) {
                return endState;
            }

            ArrayState newState = new ArrayState();
            newState.defaultState = this;
            return newState.addEndState(image, index, endState);
        }
    }

    private static class ArrayState extends State {
        private static final int MAX_CHAR = 128;
        private State[] states = new State[MAX_CHAR];
        State defaultState = null;

        public State addEndState(String image, int index, EndState endState) {
            if (index >= image.length()) {
                defaultState = endState;
                return this;
            }

            char ch = image.charAt(index);

            State newState = states[ch];
            if (newState == null) {
                if (image.length() >= index) {
                    newState = new ArrayState();
                } else {
                    newState = endState;
                }
            }
            states[ch] = newState.addEndState(image, index+1, endState);

            return this;
        }

        public State forCharacter(char ch) {
            if (ch >= 0 && ch < MAX_CHAR) {
                State state = states[ch];
                if (state != null) return state;
            }
            return defaultState;
        }
    }


    // test code
    {
  //int SINGLE_LINE_COMMENT = 9;
  //int FORMAL_COMMENT = 10;
  //int MULTI_LINE_COMMENT = 11;
  //int INTEGER_LITERAL = 13;
  //int DECIMAL_LITERAL = 14;
  //int HEX_LITERAL = 15;
  //int OCTAL_LITERAL = 16;
  //int FLOATING_POINT_LITERAL = 17;
  //int EXPONENT = 18;
  //int CHARACTER_LITERAL = 19;
  //int STRING_LITERAL = 20;

        addOperator(LPAREN, "(");
        addOperator(RPAREN, ")");
        addOperator(LBRACE, "{");
        addOperator(RBRACE, "}");
        addOperator(LBRACKET, "[");
        addOperator(RBRACKET, "]");
        addOperator(SEMICOLON, ";");
        addOperator(COMMA, ",");
        addOperator(DOT, ".");
        addOperator(ASSIGN, "=");
        addOperator(GT, ">");
        addOperator(LT, "<");
        addOperator(BANG, "!");
        addOperator(TILDE, "~");
        addOperator(HOOK, "?");
        addOperator(COLON, ":");
        addOperator(EQ, "==");
        addOperator(LE, "<=");
        addOperator(GE, ">=");
        addOperator(NE, "!=");
        addOperator(SC_OR, "||");
        addOperator(SC_AND, "&&");
        addOperator(INCR, "++");
        addOperator(DECR, "--");
        addOperator(PLUS, "+");
        addOperator(MINUS, "-");
        addOperator(STAR, "*");
        addOperator(SLASH, "/");
        addOperator(BIT_AND, "&");
        addOperator(BIT_OR, "|");
        addOperator(XOR, "^");
        addOperator(REM, "%");
        addOperator(LSHIFT, "<<");
        addOperator(RSIGNEDSHIFT, ">>");
        addOperator(RUNSIGNEDSHIFT, ">>>");
        addOperator(PLUSASSIGN, "+=");
        addOperator(MINUSASSIGN, "-=");
        addOperator(STARASSIGN, "*=");
        addOperator(SLASHASSIGN, "/=");
        addOperator(ANDASSIGN, "&=");
        addOperator(ORASSIGN, "|=");
        addOperator(XORASSIGN, "^=");
        addOperator(REMASSIGN, "%=");
        addOperator(LSHIFTASSIGN, "<<=");
        addOperator(RSIGNEDSHIFTASSIGN, ">>=");
        addOperator(RUNSIGNEDSHIFTASSIGN, ">>>=");
        addOperator(INSTANCEOF, "instanceof");
        addOperator(NEW, "new");

        addOperator(PUBLIC,"public");
        addOperator(PRIVATE,"private");
        addOperator(PROTECTED,"protected");
        addOperator(STATIC,"static");
        addOperator(FINAL,"final");
        addOperator(SYNCHRONIZED,"synchronized");
        addOperator(VOLATILE,"volatile");
        addOperator(TRANSIENT,"transient");
        addOperator(NATIVE,"native");
        addOperator(ABSTRACT,"abstract");
        addOperator(STRICT,"strictfp");
        addOperator(CONST,"const");

        addOperator(CLASS,"class");
        addOperator(INTERFACE,"interface");

        addOperator(THROWS,"throws");
        addOperator(TRY,"try");
        addOperator(CATCH,"catch");
        addOperator(FINALLY,"finally");

        addOperator(IF,"if");
        addOperator(WHILE,"while");
        addOperator(DO,"do");
        addOperator(FOR,"for");
        addOperator(SWITCH,"switch");
        addOperator(CASE,"case");
        addOperator(DEFAULT,"default");
        addOperator(BREAK,"break");
        addOperator(CONTINUE,"continue");
        addOperator(RETURN,"return");
        addOperator(THROW,"throw");
        addOperator(GOTO,"goto");

        //addOperator(BOOLEAN,"boolean");
        //addOperator(BYTE,"byte");
        //addOperator(CHAR,"char");
        //addOperator(SHORT,"short");
        //addOperator(INT,"int");
        //addOperator(LONG,"long");
        //addOperator(FLOAT,"float");
        //addOperator(DOUBLE,"double");
        //addOperator(VOID,"void");

        //addOperator(THIS,"this");
        //addOperator(SUPER,"super");
        //addOperator(CLASS,"class");
  //int AT = 69;
  //int IDENTIFIER = 70;
  //int LETTER = 71;
  //int DIGIT = 72;

    }

    private static final int FIRST_LINE_GUESS = 500;
    private static final int LINE_GUESS_MULTIPLE = 4;
    int nChars;
    
    private static final int BUF_SIZE = 4096;

    /**
     * Original @author <a href="mailto:nakamura@mamezou.com">NAKAMURA Tadashi</a>
     * Modified and integrated by Jim Hugunin
     */
    char[] readCharArray(File file) throws IOException {
        FileInputStream fstream = new FileInputStream(file);
        Reader reader;
        if (getOptions().encoding == null) {
            reader = new InputStreamReader(new BufferedInputStream(fstream));
        } else {
            reader = new InputStreamReader(new BufferedInputStream(fstream), 
                                           getOptions().encoding);
        }
        reader = new BufferedReader(reader);
        
        try {
            int length;
            final char[] buf = new char[BUF_SIZE];
            CharArrayWriter caw = new CharArrayWriter();
            
            while ((length = reader.read(buf, 0, BUF_SIZE)) != -1) {
                caw.write(buf, 0, length);
            }
            caw.write('\n');  // adding a newline to make things easier
            return caw.toCharArray();
        } finally {
            fstream.close();
        }
    }
    
    private char[] processUnicodeEscapes(char[] text) {
        CharArrayWriter writer = null;
        final int N = text.length;
        outer: for(int i=0; i<N; i++) {
            char ch = text[i];
            
            if (ch == '\\') {
                boolean oddSlashes = true;
                int startIndex = i;
                while (++i < N) {
                    ch = text[i];
                    if (ch == '\\') {
                        oddSlashes = !oddSlashes;
                    } else if (ch == 'u' && oddSlashes) {
                        // we now know we need to handle unicode escapes
                        if (writer == null) {
                            writer = new CharArrayWriter();
                            writer.write(text, 0, i-1);
                        }
                        // handle multiple u's
                        while (text[++i] == 'u') {}
                        i -= 1;

                        if (i+4 > N) {
                            throw new ParseException(i, null,
                            "invalid unicode escape");
                        }
                        int v = Character.digit(text[++i], 16);
                        v = v*16 + Character.digit(text[++i], 16);
                        v = v*16 + Character.digit(text[++i], 16);
                        v = v*16 + Character.digit(text[++i], 16);
                        ch = (char)v;
                        writer.write(ch);
                        continue outer;
                    } else {
                        break;
                    }
                }
                if (writer != null) writer.write(text, startIndex, i+1-startIndex);
            } else {
                if (writer != null) writer.write(ch);
            }
        }        
        if (writer != null) return writer.toCharArray();
        else return text;
    }
        
    public void setSourceInfo(SourceInfo si) throws IOException {
        char[] text = readCharArray(si.getFile());
        text = processUnicodeEscapes(text);
        si.setText(text);

        sourceInfo = si;
        this.text = text;
        this.startIndex = this.index = 0;
        this.nChars = text.length;
    }

    public void setOffset(int offset) {
        this.startIndex = this.index = offset;
    }

    public SourceInfo sourceInfo = null;


    public JavaTokenizer(JavaCompiler compiler) {super(compiler);}

/*
    public static void main(String[] args) throws IOException {
        long t0 = System.currentTimeMillis();
        JavaTokenizer tokenizer = new JavaTokenizer(new File(args[0]));

        Token tok;
        while ((tok = tokenizer.nextToken()) != null) {
            System.out.println(tok.startPosition+":"+tok.endPosition+" "+tok);
        }
        long t1 = System.currentTimeMillis();
        System.err.println(((t1-t0)/1000.0)+" seconds.");
    }
*/
}

