/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.AbstractCompilerPass;
import org.aspectj.compiler.base.WorldPass;
import org.aspectj.compiler.base.CompilationUnitPass;

import java.util.*;

import org.aspectj.compiler.crosscuts.ast.AspectDec;

/** apply is always true here */

public class PrivilegeFixer extends AbstractCompilerPass {
    private boolean apply;
    
    /** apply is used for separate compilation.  You need to make the access 
     * methods for everybody before you do any bytecode, but you don't need to 
     * change the call sites until you need to.  
     */
    
    public PrivilegeFixer(JavaCompiler compiler, boolean apply) {
        super(compiler);
        this.apply = apply;
    }

    public String getDisplayName() {
        return (apply ? "implement":"plan") + " backdoors";
    }

    public void transform(TypeDec td) {
        //XXX make this much less efficient in return for being simpler
        //XXXcheckTypeDecs(cu,apply);
        //System.out.println("BEFORE: " + td.unparse());
        AccessFixer fixer = new AccessFixer(getCompiler(), td, apply);
        fixer.fixAccesses(td); 
        //System.out.println("AFTER: " + td.unparse());
    }

    /** Scans compilation unit for privilaged aspects, and
     *  creates backdoor methods if apply==false, or
     *  replaces method calls/field access with backdoor
     *  methods, if apply == true
     */
    private void checkTypeDecs(CompilationUnit cu, boolean apply) {
        List typeDecs = cu.getDefinedTypes();
        final int N = typeDecs.size();

        for (int j = 0; j < N; j++) {
            Dec dec = (Dec)typeDecs.get(j);
            if (dec instanceof TypeDec) {
                TypeDec typeDec = (TypeDec)dec;
                if (typeDec instanceof AspectDec && 
                    typeDec.getModifiers().isPrivileged())
                {
                    //typeDec.bindNames();
                    AccessFixer fixer = new AccessFixer(getCompiler(), typeDec, apply);
                    fixer.fixAccesses(typeDec);
                }
                
                fixIntroducedDecs( typeDec, apply );
            }
        }
    }
    
    private void fixIntroducedDecs(TypeDec typeDec, boolean apply) {
        //System.out.println("fixing: " + typeDec);
        
        Decs decs = typeDec.getBody();
        
        for(int i=0; i < decs.size(); i++) {
            Dec dec = (Dec)decs.get(i);
            //System.out.println("  " + dec + " intro " + dec.getIntroducedFromType());
            if (dec.isIntroduced()) {
                new AccessFixer(getCompiler(), typeDec, apply).fixAccesses(dec);
            }
        }
    }     
}
