/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
 * @grammar modifiers? name {body};
 * @child Pcd pcd @after throws
 */

public abstract class AdviceDec extends CodeDec implements JpPlannerMaker {
    public String getKind() { return "advice"; }

    //private AspectDec aspectDec = null;
    public AspectDec getAspectDec() {
        return (AspectDec)getDeclaringType().getTypeDec();
    }

	/** 
	 * returns the after returning or after throwing formal.  
	 * Overridden in {@link AfterReturningAdviceDec} and {@link AfterThrowingAdviceDec}
	 * and {@link AroundAdviceDec}.
	 */
    public FormalDec getExtraFormal() { return null; }

    protected boolean isAfterAdvice() { return false; }

    protected boolean reverseSortOrder(AdviceDec a1, AdviceDec a2) {
        return a1.isAfterAdvice() || a2.isAfterAdvice();
    }

    public boolean dominates(AdviceDec otherAdvice) {
        AspectDec myAspectDec = getAspectDec();
        AspectDec otherAspectDec = otherAdvice.getAspectDec();

        if (myAspectDec == otherAspectDec) {
            if (reverseSortOrder(this, otherAdvice)) {
                return this.getBeginLine() > otherAdvice.getBeginLine();
            } else {
                return this.getBeginLine() < otherAdvice.getBeginLine();
            }
        } else {
            return myAspectDec.dominates(otherAspectDec);
        }
    }

    public JpPlanner makePlanner(PlanData planData) {
        if (planData.inAspect.isAbstract()) return JpPlanner.NO_PLANNER;

        JpPlanner planner = getPcd().makePlanner(planData);
        //System.out.println(this.toShortString() + ": " + planner);
        //??? we depend on the order that this will proced in dynamic tests
        planner = planData.inAspect.findPerClause().makeInnerPlanner(planData).and(planner);
        return new AdvicePlanner(planData.inAspect, this, planner);
    }

    public void wrapJoinPoint(JoinPoint joinPoint, AdvicePlan plan) {
        joinPoint.setStmts(this.wrapStmts(joinPoint.getStmts(), plan));
    }

    protected Stmts wrapStmts(Stmts stmts, AdvicePlan plan) {
        this.showError("didn't implement code to concretize this advice");
        return null;
    }

    protected final Formals makeFormals() {
        Formals newFormals = getAST().makeFormals();
        Formals oldFormals = getFormals();

        for (int i = 0; i < oldFormals.size(); i++) {
            oldFormals.get(i).clearParent();
            newFormals.add(oldFormals.get(i));
        }
        
        if (getExtraFormal() != null) {
			getExtraFormal().clearParent();
            newFormals.add(getExtraFormal());
        }        	

        return newFormals;
    }

    public final Type getExtraArgType() {
    	if (getExtraFormal() == null) return null;
    	return getExtraFormal().getType();
    }

    public MethodDec makeConcreteMethod() {
        if (methodDec == null) {
            methodDec = getMethodDec();
        }

        if (methodDec == null) return null;
        methodDec.setBody(makeMethodBody());
        finishMethodDec(methodDec);
        return methodDec;
    }
	
	
    /** returns the method body used for the java method corresponding to
     * this advice
     * 
     * this method body is my actual body in the AST, never a copy
     */
    protected CodeBody makeMethodBody() {
    	return body;
    }
	

    protected void finishMethodDec(MethodDec methodDec) {
        if (modifiers.isStrict()) methodDec.getModifiers().setStrict(true);
    }

    // make a short string based on how high this is in the hierarchy
    // this avoids some ridiculously long names, but feels hackish
    int getDepth(AspectDec ad) {
        Type superType = ad.getSuperClassType();
        if (superType.isAspect()) {
            return 1 + getDepth((AspectDec)superType.getTypeDec());
        } else {
            return 0;
        }
    }

    String makeUniqueAspectString() {
        int depth = getDepth(getAspectDec());
        if (depth == 0) return "";
        else return Integer.toString(depth, 36);
    }

    String makeUniqueAdviceString() {
        String kind = getAdviceKind();
        Decs body = (Decs)getParent();
        int position = 0;
        final int N = body.size();
        //System.out.println("searching for: " + this.toShortString());
        for (int i = 0; i < N; i++) {
            Dec dec = body.get(i);
            //System.out.println(dec.toShortString() + ": " + position);
            if (dec == this) break;
            if (!(dec instanceof AdviceDec)) continue;
            AdviceDec ad = (AdviceDec)dec;
            if (ad.getAdviceKind().equals(kind)) position++;
        }
        return kind + Integer.toString(position, 36);
    }


    String makeUniqueName() {
        return makeUniqueAdviceString() + "$ajc" + makeUniqueAspectString();
    }

    private MethodDec methodDec;
    public synchronized MethodDec getMethodDec() {
        if (methodDec != null) return methodDec;

        final AST ast = getAST();

        String name = makeUniqueName();
        Modifiers modifiers = ast.makeModifiers(getModifiers().getValue());
        modifiers.setPublic(true);
        modifiers.setFinal(true);

        Formals formals = makeFormals();
        addJoinPointFormals(formals);

        methodDec = ast.makeMethod(modifiers, getResultType(), name,
                                        formals, null);
        if (_throws != null) methodDec.setThrows((TypeDs)_throws.copy());

        methodDec.setAllEnclosingTypes(getAspectDec().getType());
        return methodDec;
    }


    public Expr makeCall(Expr parent, Exprs baseExprs) {
        Exprs args = (Exprs)baseExprs.copy();
        if (parent == null) {
           parent = getAST().makeTypeExpr(getAspectDec().getType());
        }
        CallExpr ret = getAST().makeCall(getMethodDec().getMethod(), parent, args);
        return ret;
    }




    //***********************************************************************
    // Section to determine what meta-joinpoint objects are needed

    private boolean joinPointFlagsSet = false;
    private boolean needsStaticEnclosingJoinPoint;
    private boolean needsStaticJoinPoint;
    private boolean needsDynamicJoinPoint;

    private void setJoinPointFlags() {
        if (!joinPointFlagsSet) {
            joinPointFlagsSet = true;
            DynamicJoinPointChecker jpChecker = new DynamicJoinPointChecker(getCompiler());
            jpChecker.checkAndFix(getBody());
            needsDynamicJoinPoint = jpChecker.needsDynamicJoinPoint();
            needsStaticJoinPoint = jpChecker.needsStaticJoinPoint();
            needsStaticEnclosingJoinPoint = jpChecker.needsStaticEnclosingJoinPoint();
        }
    }

    public boolean needsStaticEnclosingJoinPointFormal() {
        setJoinPointFlags();
        return needsStaticEnclosingJoinPoint;
    }

    public boolean needsStaticJoinPointFormal() {
        setJoinPointFlags();
        return needsStaticJoinPoint;
    }

    public boolean needsDynamicJoinPointFormal() {
        setJoinPointFlags();
        return needsDynamicJoinPoint;
    }

    public boolean needsCallSiteContext() {
        return needsDynamicJoinPointFormal() ||
                needsStaticJoinPointFormal() ||
                needsStaticEnclosingJoinPointFormal();
    }

    public void addJoinPointFormals(Formals formals) {
        if (needsStaticEnclosingJoinPointFormal()) formals.add(thisStaticEnclosingJoinPointFormal);
        if (needsStaticJoinPointFormal()) formals.add(thisStaticJoinPointFormal);
        if (needsDynamicJoinPointFormal()) formals.add(thisDynamicJoinPointFormal);
    }



    class DynamicJoinPointChecker extends Walker {
        boolean needsDynamic = false;
        boolean needsStatic = false;
        boolean needsStaticEnclosing = false;
        boolean needsFakeStatic = false;

        boolean makeFakeStatics = false;

        public boolean needsDynamicJoinPoint() { return needsDynamic; }
        public boolean needsStaticJoinPoint() { return needsStatic; }
        public boolean needsStaticEnclosingJoinPoint() { return needsStaticEnclosing; }

        DynamicJoinPointChecker(JavaCompiler compiler) {
            super(compiler);
        }

        public void checkAndFix(ASTObject body) {
            this.process(body);
            if (needsFakeStatic && !needsDynamic) {
                if (!this.getCompiler().getOptions().noMetaJoinPointOptimization) {
                    makeFakeStatics = true;
                    needsStatic = true;
                    this.process(body);
                } else {
                    needsDynamic = true;
                }
            }
        }

        boolean isStaticEnclosingJoinPoint(VarExpr varExpr) {
            return varExpr.getVarDec() == thisStaticEnclosingJoinPointFormal;
        }

        boolean isStaticJoinPoint(VarExpr varExpr) {
            return varExpr.getVarDec() == thisStaticJoinPointFormal;
        }

        boolean isDynamicJoinPoint(VarExpr varExpr) {
            return varExpr.getVarDec() == thisDynamicJoinPointFormal;
        }

        void checkNoBang(VarExpr varExpr) {
            if (varExpr.isLhs()) {
                varExpr.showError("not allowed to set thisJoinPoint");
            }
        }

        boolean isStaticPartCall(CallExpr callExpr) {
            Expr expr = callExpr.getExpr();
            if (expr != null && expr instanceof VarExpr &&
                        isDynamicJoinPoint( (VarExpr)expr ))
            {
                return callExpr.getMethod().getId().equals("getStaticPart");
            }
            return false;
        }

        boolean canTreatAsStatic(Method method) {
            String id = method.getId();
            return id.equals("toString") ||
                id.equals("toShortString") ||
                id.equals("toLongString") ||
                id.equals("getKind") ||
                id.equals("getSignature") ||
                id.equals("getSourceLocation") ||
                id.equals("getStaticPart");
        }


        boolean canTreatAsStatic(VarExpr varExpr) {
            ASTObject parent = varExpr.getParent();
            if (parent instanceof CallExpr) {
                Method calledMethod = ((CallExpr)parent).getMethod();
                return canTreatAsStatic(calledMethod);

                //??? should add a case here to catch
                //??? tjp.getEnclosingExecutionJoinPoint().STATIC_METHOD()
            } else if (parent instanceof BinopExpr) {
                BinopExpr binop = (BinopExpr)parent;
                if (binop.getType().isEquivalent(this.getTypeManager().getStringType())) {
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }


        boolean inBlockThatCantRun = false;

        boolean isFinalBoolean(Expr expr) {
            // this assumes that constant folding has already happened
            return (expr instanceof BooleanLiteralExpr);
        }

        boolean getFinalBooleanValue(Expr expr) {
            return ((BooleanLiteralExpr)expr).getBooleanValue();
        }

        protected Expr makeNullJoinPoint() {
            final AST ast = this.getAST();
            return ast.makeParen(ast.forceCast(this.getTypeManager().getJoinPointType(),
                                               ast.makeNull()));
        }

        public ASTObject process(ASTObject node) {
            // make a good effort not to generate non-static join point objects
            // when in an if block guarded by a false static final field
            if (node instanceof IfStmt) {
                IfStmt ifStmt = (IfStmt)node;

                if (!inBlockThatCantRun && isFinalBoolean(ifStmt.getTest())) {
                    ifStmt.setTest((Expr)process(ifStmt.getTest()));

                    inBlockThatCantRun = !getFinalBooleanValue(ifStmt.getTest());

                    ifStmt.setThen((Stmt)process(ifStmt.getThen()));

                    inBlockThatCantRun = !inBlockThatCantRun;

                    if (ifStmt.getElse() != null) {
                        ifStmt.setElse((Stmt)process(ifStmt.getElse()));
                    }

                    inBlockThatCantRun = false;
                    return ifStmt;
                } else {
                    return super.process(ifStmt);
                }
            } else if (node instanceof CallExpr) {
                CallExpr callExpr = (CallExpr)node;
                if (isStaticPartCall(callExpr)) {
                    needsFakeStatic = true;
                    if (makeFakeStatics) {
                        return AdviceDec.this.getAST().makeVar(thisStaticJoinPointFormal);
                    }
                }
            } else if (node instanceof VarExpr) {
                VarExpr varExpr = (VarExpr)node;

                if (isStaticEnclosingJoinPoint(varExpr)) {
                    needsStaticEnclosing = true;
                    checkNoBang(varExpr);
                } else if (isStaticJoinPoint(varExpr)) {
                    needsStatic = true;
                    checkNoBang(varExpr);
                } else if (isDynamicJoinPoint(varExpr)) {
                    checkNoBang(varExpr);
                    if (inBlockThatCantRun) {
                        //if (makeFakeStatics) {
                            return makeNullJoinPoint();
                        //}
                    } else if (canTreatAsStatic(varExpr)) {
                        needsFakeStatic = true;
                        if (makeFakeStatics) {
                            varExpr.setVarDec(thisStaticJoinPointFormal);
                            varExpr.setType(null);
                            varExpr.getType(); //XXX ugly idiom to reset the type
                        }
                    } else {
                        needsDynamic = true;
                    }
                }
                return node;
            }

            return super.process(node);
        }
    }

    protected FormalDec makeJoinPointFormal(String name) {
        return getAST().makeFormal(getTypeManager().getJoinPointType(), name);
    }

    protected FormalDec makeJoinPointStaticPartFormal(String name) {
        return getAST().makeFormal(getTypeManager().getJoinPointStaticPartType(), name);
    }

    protected FormalDec thisDynamicJoinPointFormal = null;
    protected FormalDec thisStaticJoinPointFormal = null;
    protected FormalDec thisStaticEnclosingJoinPointFormal = null;

    // ------------------------------
    // INTRO from FlowCheckerPass

    protected void setupFlowWalker(FlowCheckerPass w) {
        if (thisDynamicJoinPointFormal == null ||
            thisStaticJoinPointFormal == null ||
            thisStaticEnclosingJoinPointFormal == null ) {
            throw new RuntimeException("join point formals not created yet");
        }
        w.setVars(w.getVars().addAssigned(thisDynamicJoinPointFormal)
                  .addAssigned(thisStaticJoinPointFormal)
                  .addAssigned(thisStaticEnclosingJoinPointFormal));
        super.setupFlowWalker(w);
    }

    // INTRO from ASTFixerPass
    /**
     * Replace the advice with its corresponding MethodDec in the AspectDec
     */
    public ASTObject postFixAST(final ASTFixerPass fixer) {
        return makeConcreteMethod();
    }

    protected BlockScope makeBodyScope(Scope parent) {
        return new BlockScope(getCompiler(), parent);
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) {
        super.preScope(walker);

        if (thisDynamicJoinPointFormal == null ) {
            thisDynamicJoinPointFormal = makeJoinPointFormal("thisJoinPoint");
        }
        if (thisStaticJoinPointFormal == null ) {
            thisStaticJoinPointFormal = makeJoinPointStaticPartFormal("thisJoinPointStaticPart");
        }
        if (thisStaticEnclosingJoinPointFormal == null ) {
            thisStaticEnclosingJoinPointFormal = makeJoinPointStaticPartFormal("thisEnclosingJoinPointStaticPart");
        }

        walker.process(thisDynamicJoinPointFormal);
        walker.process(thisStaticJoinPointFormal);
        walker.process(thisStaticEnclosingJoinPointFormal);

        //XXX there must be a better way to ensure this happens
        //walker.process(getFormal());
    }

    public void checkSpec() {
        if (isStatic()) {
            getCompiler().warnVersion("0.8beta1", this, "no static modifier on advice");
        }
        for (int i=0; i<formals.size(); i++) {
            if (!formals.get(i).isBound) {
                formals.get(i).showError("not bound in this PCD");
            }
        }
        getBody().checkReturnType(getReturnType());

        //make sure this is done at the right time
        getMethodDec();
    }


    public String getId() {
        //XXX this name is probably not unique enough
        return getAdviceKind();
    }

    public TypeD getResultTypeD() {
        return getReturnType().makeTypeD();
    }

    public Type getReturnType() {
        return getTypeManager().voidType;
    }


    public String toShortString() {
        String modString = getModifiers().toShortString();
        if (modString.length() > 0) {
            modString += " ";
        }

        return modString + getAdviceKind() + getFormals().toShortString();
    }

    abstract protected String getAdviceKind();

    public void unparse(CodeWriter writer) {
        // do not unparse!
    }

    //BEGIN: Generated from @child and @property
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }

    public AdviceDec(SourceLocation location, Modifiers _modifiers, Formals _formals, TypeDs __throws, Pcd _pcd, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _body);
        setPcd(_pcd);
    }
    protected AdviceDec(SourceLocation source) {
        super(source);
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return formals;
        case 2: return _throws;
        case 3: return pcd;
        case 4: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "formals";
        case 2: return "throws";
        case 3: return "pcd";
        case 4: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setFormals((Formals)child); return;
        case 2: setThrows((TypeDs)child); return;
        case 3: setPcd((Pcd)child); return;
        case 4: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 5;
    }

    public String getDefaultDisplayName() {
        return "AdviceDec()";
    }

    //END: Generated from @child and @property
}
