/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import javax.swing.JPanel;

import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.base.JavaCompiler;

/**
 * @grammar after (formals) returning (formal)
 * @child FormalDec extraFormal @after formals
 */


public class AfterReturningAdviceDec extends AdviceDec {    
    protected String getAdviceKind() { return "afterReturning"; }

    protected boolean isAfterAdvice() { return true; }
    
    private static final String CONSTRUCTOR_EXECUTION_BINDING_CHANGE =
    	"return value will bind to null, use 'this' pointcut to bind executing object\n" +
    	"    behavior change to fix compiler bug in 1.0.3 and earlier";
    	
    private static final String VOID_RETURN_CHANGE =
    	"will not match this joinpoint, use Object return type in advice if match desired\n" +
    	"    behavior change to fix compiler bug in 1.0.3 and earlier";

    protected Stmts wrapStmts(Stmts stmts, AdvicePlan plan) {
        JoinPoint jp = plan.getJoinPoint();
        FormalDec formal = getExtraFormal();
        
        if (jp instanceof ExceptionHandlerJp) {
            showWarning("can't apply after returning advice to exception handler join points (compiler limitation)");
            jp.getSourceLocation().showWarning(
              "trying to apply above after returning advice to execution of this exception handler");
            return stmts;
        }
        if (jp.getResultType().isVoid()) {
            if (formal != null && !formal.getType().isObject()) {
                jp.showWarning(this, VOID_RETURN_CHANGE);
                return stmts;
            }
        }
        
        if (jp instanceof ConstructorExecutionJp) {
            if (formal != null && formal.getType().isObject()) {
                jp.showWarning(this, CONSTRUCTOR_EXECUTION_BINDING_CHANGE);
            }
        }

        ReturnFixerWalker walker = new ReturnFixerWalker(getCompiler(), plan);
        Stmts ret = walker.removeReturns(stmts);
        return ret;
    }
    
    /** 
     * 
     * generates code:
     * 
     *   stmts ->  ReturnType _thisResult = null;
     *             LABEL: if (true) { stmts; }
     *             doAdvicePlan(_thisResult);
     *             return _thisResult;
     * 
     *   within stmts:
     *     return x; -> _thisResult = x; break LABEL;
     */
    static class ReturnFixerWalker extends Walker {
        AdvicePlan plan;
	VarDec thisResultVar;
	String label;


        public ReturnFixerWalker(JavaCompiler compiler, AdvicePlan plan) {
            super(compiler);
            this.plan = plan;
        }
        
        public Stmts removeReturns(Stmts stmts) {
            final AST ast = getAST();
            Type resultType = plan.getJoinPoint().getResultType();
            
            Expr thisResultExpr;
            Stmt returnStmt;
            
            if (resultType.isVoid()) {
                thisResultVar = null;
                //??? Expr var = plan.getJoinPoint().makeVoidReturnExpr();
                thisResultExpr = ast.makeNull();
                returnStmt = ast.makeReturn();
            } else {
            	thisResultVar =
            	    ast.makeVarDec(resultType, "_thisResult", resultType.getNullExpr());
                thisResultExpr = ast.makeVar(thisResultVar);
                returnStmt = ast.makeReturn(ast.makeVar(thisResultVar));
            }
            
            label = getWorld().genLabel();
            
            Stmts ret = (Stmts)process(stmts);
            
            Stmt body = ast.makeLabeled(label, ast.makeIf(ast.makeLiteral(true),
                                                          ret.getSingleStmt()));
            
            ret = ast.makeStmts(
                body,
                plan.makeCall(thisResultExpr),
                returnStmt);
                
            if (!resultType.isVoid()) {
                ret.add(0, thisResultVar);
            }
            
            return ret;
        }

        boolean seenVoidReturn = false;
        boolean seenAnyReturns = false;
        public Stmt wrapReturn(ReturnStmt returnStmt) {
            final AST ast = this.getAST();
            
            Expr returnExpr = returnStmt.getExpr();
            if (returnExpr == null) {
                return ast.makeBreak(label);
            } else {
                return ast.makeBlock(
                    ast.makeSet(thisResultVar, returnExpr),
                    ast.makeBreak(label));
            }
        }
        
        public ASTObject process(ASTObject object) {
            // make sure we don't walk into inners
            if (object instanceof TypeDec) return object;
            if (object instanceof AnonymousMethodExpr) return object;
            
            if (object instanceof ReturnStmt) {
                return wrapReturn((ReturnStmt)object);
            }

            return super.process(object);
        }
    }    
    
    // ------------------------------
    // INTRO from FlowCheckerPass

    protected void setupFlowWalker(FlowCheckerPass w) {
        if (getExtraFormal() != null) 
            w.setVars(w.getVars().addAssigned(getExtraFormal()));
        super.setupFlowWalker(w);
    }    
    
    //BEGIN: Generated from @child and @property
    protected FormalDec extraFormal;
    public FormalDec getExtraFormal() { return extraFormal; }
    public void setExtraFormal(FormalDec _extraFormal) {
        if (_extraFormal != null) _extraFormal.setParent(this);
        extraFormal = _extraFormal;
    }
    
    public AfterReturningAdviceDec(SourceLocation location, Modifiers _modifiers, Formals _formals, FormalDec _extraFormal, TypeDs __throws, Pcd _pcd, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _pcd, _body);
        setExtraFormal(_extraFormal);
    }
    protected AfterReturningAdviceDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        AfterReturningAdviceDec ret = new AfterReturningAdviceDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (extraFormal != null) ret.setExtraFormal( (FormalDec)walker.process(extraFormal) );
        if (_throws != null) ret.setThrows( (TypeDs)walker.process(_throws) );
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        if (body != null) ret.setBody( (CodeBody)walker.process(body) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return formals;
        case 2: return extraFormal;
        case 3: return _throws;
        case 4: return pcd;
        case 5: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "formals";
        case 2: return "extraFormal";
        case 3: return "throws";
        case 4: return "pcd";
        case 5: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setFormals((Formals)child); return;
        case 2: setExtraFormal((FormalDec)child); return;
        case 3: setThrows((TypeDs)child); return;
        case 4: setPcd((Pcd)child); return;
        case 5: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 6;
    }
    
    public String getDefaultDisplayName() {
        return "AfterReturningAdviceDec()";
    }
    
    //END: Generated from @child and @property
}
