/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.*;

import java.util.*;

/**
  * @grammar ...
  * @child Exprs args
  * @child MethodDec methodDec
  *
  */
public class AnonymousMethodExpr extends Expr {
    protected Type discoverType() { return methodDec.getResultType(); }

    public void grabContext() {
        ExprMaker em = new ExprMaker(getCompiler());
        em.process(methodDec);
        Map inParams = em.getInParams();
        
        Exprs args = getArgs();
        Formals formals = methodDec.getFormals();
        
        // loop over inParams to create call args and formals together
        for (Iterator i = inParams.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry)i.next();
            VarDec varDec = (VarDec)entry.getKey();
            varDec.getModifiers().setFinal(true);  //??? assumes too much
            args.add(getAST().makeVar(varDec));
            
            FormalDec formalDec = (FormalDec)entry.getValue();
            formals.add(formalDec);
        }
    }        
    
    public ASTObject postMove(MovingWalker walker) {
        //XXX This architecture causes us to walk dec twice
        //XXX I'm not thrilled about that

        //if (!em.needsThis()) dec.getModifiers().setStatic(true);
        grabContext();
        return this;
    }
    
    //INTRO from CopyWalker
//    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
//
//        getExpr().setMethod(getDec().getMethod());
//        return this;
//    }

    public void fixAST(ASTFixerPass fixer) {
        // only walk args
        // methodDec will be walked after it's been moved in the tree
        args = (Exprs)fixer.process(args);
    }

    // INTRO from ASTFixerPass
    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (inStaticContext()) {
            methodDec.getModifiers().setStatic(true);
        } else {
            methodDec.getModifiers().setStatic(false);
        }
        
        TypeDec inTypeDec = getBytecodeTypeDec();
        if (methodDec.isStatic() && inTypeDec.isInner()) {
            inTypeDec = inTypeDec.getOutermostTypeDec();
        }

        Decs body = inTypeDec.getBody();
        methodDec.getModifiers().setPrivate(true);
        // we don't need to anonymize these since NameHygienePass will fix them for us
        body.add(getMethodDec());
        getMethodDec().setAllEnclosingTypes(inTypeDec.getType());
        
        Expr thisExpr;
        if (methodDec.isStatic()) {
            thisExpr = getAST().makeTypeExpr(inTypeDec.getType());
        } else {
            thisExpr = getAST().makeThis(inTypeDec.getType());
        }
        
        return getAST().makeCall(getMethodDec(), thisExpr, args);
    }
    
    //INTRO from MixinImplementationPass
    public void implementMixin(MixinImplementationPass fixer) {
        return;
    }    
    
    
    public void unparse(CodeWriter writer) {
        writer.write("DecInExpr[");
        writer.write(args);
        writer.write(" -> ");
        writer.write(methodDec);
        writer.write("]");
    }    
    
    //BEGIN: Generated from @child and @property
    protected Exprs args;
    public Exprs getArgs() { return args; }
    public void setArgs(Exprs _args) {
        if (_args != null) _args.setParent(this);
        args = _args;
    }
    
    protected MethodDec methodDec;
    public MethodDec getMethodDec() { return methodDec; }
    public void setMethodDec(MethodDec _methodDec) {
        if (_methodDec != null) _methodDec.setParent(this);
        methodDec = _methodDec;
    }
    
    public AnonymousMethodExpr(SourceLocation location, Exprs _args, MethodDec _methodDec) {
        super(location);
        setArgs(_args);
        setMethodDec(_methodDec);
    }
    protected AnonymousMethodExpr(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        AnonymousMethodExpr ret = new AnonymousMethodExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (args != null) ret.setArgs( (Exprs)walker.process(args) );
        if (methodDec != null) ret.setMethodDec( (MethodDec)walker.process(methodDec) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return args;
        case 1: return methodDec;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "args";
        case 1: return "methodDec";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setArgs((Exprs)child); return;
        case 1: setMethodDec((MethodDec)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "AnonymousMethodExpr()";
    }
    
    //END: Generated from @child and @property
}
