/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import java.util.*;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

/**
 * @grammar around {body; proceed(); body;};
 * @child TypeD resultTypeD @after modifiers
 */

public class AroundAdviceDec extends AdviceDec {
    Collection returnStmts = new ArrayList();
    Collection proceedExprs = new ArrayList();

    /** returns true if we should inline the advice at its static join point.  In the false
     * case, use closure objects
     */
    private boolean shouldInline() {
        return (!containsInners && !isTargetOfAnyAroundAdvice) && !getOptions().alwaysMakeAroundClosures;
    }

    private boolean containsInners = false;
    private boolean isTargetOfAnyAroundAdvice = false;

    public MethodDec getMethodDec() {
        MethodDec md = super.getMethodDec();
        md.setThrows(
                     getAST().makeTypeDs(getTypeManager().getThrowableType().makeTypeD()));
        return md;
    }

    protected String getAdviceKind() {
        return "around";
    }

    public List joinPoints1 = new ArrayList();
    public List joinPoints2 = new ArrayList();

    void implementPlans(List jps) {
        for (Iterator i = jps.iterator(); i.hasNext();) {
            JoinPoint joinPoint = (JoinPoint) i.next();
            joinPoint.implementPlans();
        }
    }

    private boolean ensuredIsWoven = false;
    private boolean ensuringIsWoven = false;

    private boolean ensureBodyIsWoven(AdvicePlan plan) {
        if (ensuredIsWoven)
            return true;

        if (ensuringIsWoven) {
            isTargetOfAnyAroundAdvice = true;
            return false;
        }
        ensuringIsWoven = true;

        implementPlans(joinPoints1);
        implementPlans(joinPoints2);

        ensuringIsWoven = false;
        ensuredIsWoven = true;
        return true;
    }

    /** returns a new wrapped advised Stmts object.
     *
     * @param stmts -- the statements of the join point static shadow
     * @param plan -- an object to give the parameters of the advice, among other things.
     */
    protected Stmts wrapStmts(Stmts stmts, AdvicePlan plan) {
        // if my body hasn't been woven yet, this will switch to non-inline mode
        if (shouldInline()) ensureBodyIsWoven(plan);

        // if there's an error in the advice plan, there'll be an error, so just stop
        if (!checkPlan(plan))
            return stmts;

        JoinPoint jp = plan.getJoinPoint();
        if (jp instanceof AdviceExecutionJp) {
            AdviceDec ad = (AdviceDec) jp.getTargetDec();
            if (ad instanceof AroundAdviceDec) {
                AroundAdviceDec aad = (AroundAdviceDec)ad;
                aad.isTargetOfAnyAroundAdvice = true;
                aad.fixProceedCalls();
            }
        }

        if (jp instanceof ExceptionHandlerJp) {
            showWarning("can't apply around advice to exception handler join points (compiler limitation)");
            jp.getSourceLocation().showWarning(
                                               "trying to apply above around advice to execution of this exception handler");
            return stmts;
        }

        if (shouldInline()) {
            return wrapStmtsInline(stmts, plan);
        } else {
            return wrapStmtsWithClosure(stmts, plan);
        }
    }

    private Stmts wrapStmtsWithClosure(Stmts stmts, AdvicePlan plan) {
        JoinPoint jp = plan.getJoinPoint();
        
        TypeDec parentTypeDec = stmts.getBytecodeTypeDec();
        
        if (parentTypeDec instanceof InterfaceDec) {
            showWarning("can't apply around advice which contains inner types to interface methods (compiler limitation)");
            jp.getTargetNode().showWarning("trying to apply to this location");
            return stmts;
        }

        // won't work for exception handlers (not that it was before) for these reasons:
        // 1. we will eventually need to allocate store space for the closed over vars
        // instead of copying.  Or something like that.
        // 2. we don't yet worry about breaks/continues/complicated control flow.
        // 3. we don't get access to fields in local types

        final AST ast = getAST();

        // let s = free vars of STMTS
        //

        // new Closure137(references to s);

        // add to target type:
        // private <synthetic> class Closure137 implements AroundClosure {
        //   fields for s;
        //  Closure137(formals for s) { initializations for s; }
        //  public Object run() { if(true) STMTS[return x/RETURN wrap(x),s->fields;
        //                        return/RETURN null]; RETURN null}
        // }

        Modifiers m = ast.makeModifiers(Modifiers.PUBLIC);

        FormalDec argsFormal = ast.makeFinalFormal(getTypeManager().getObjectType().getArrayType(),
                                                   "_args");

        ClosureWalker closureWalker = new ClosureWalker(getCompiler(), stmts,
                                                        makeProceedMap(plan), argsFormal);

        Constructor superConstructor =
            getTypeManager().getAroundClosureType().getConstructor(
                                                                   stmts,
                                                                   ast.makeExprs(),
                                                                   false);

        ConstructorDec constructorDec =
            ast.makeConstructor(
                                ast.makeModifiers(0),
                                closureWalker.getConstructorFormals(),
                                ast.makeTypeDs(),
                                ast.makeConstructorCall(true, ast.makeExprs(), superConstructor),
                                closureWalker.getFieldInitStmts());

        MethodDec methodDec =
            ast.makeMethod(
                           m,
                           getTypeManager().getObjectType(),
                           "run",
                           ast.makeFormals(argsFormal),
                           closureWalker.getRunBody());
        String closureName = genClosureName();

        ClassDec classDec =
            ast.makeClass(
                          ast.makeModifiers(Modifiers.PRIVATE),
                          closureName,
                          getTypeManager().getAroundClosureType().makeTypeD(),
                          ast.makeTypeDs());

        if (plan.getJoinPoint().isStaticContext()) {
            classDec.getModifiers().setStatic(true);
        }

        //  we add an explictly non-synthetic constructor to the body before
        // we ever create a type object.  Then, when the type object is created,
        // the type will have a non-default constructor, and so we won't eagerly create
        // a default-constructor.
        constructorDec.setExplicitlyNonSynthetic();
        classDec.addToBody(constructorDec);

        methodDec.setAllEnclosingTypes(classDec.getType());

        classDec.getBody().add(methodDec);

        for (Iterator i = closureWalker.getFields().iterator(); i.hasNext();) {
            FieldDec fd = (FieldDec) i.next();
            classDec.getBody().add(fd);
        }

        
        // see limitations for exception handlers
        //		while (parentTypeDec.isLocal()) {
        //			parentTypeDec = parentTypeDec.getEnclosingTypeDec();
        //		}

        //??? can we expose this better
        // set enclosing type dec so that various inner info will be "right"
        // also add this to the type graph
        parentTypeDec.getBody().add(classDec);
        classDec.setEnclosingTypeDec(parentTypeDec);
        classDec.getNameType().addToTypeGraph();
        classDec.getNameType().buildSignatures();

        NewInstanceExpr newExpr =
            ast.makeNew(classDec.getType(), closureWalker.getConstructorArgs());

        // from a non-static context, we create new inner types
        // when we do this after the InnerInfoPass, we need to set the inner flag
        // and add enclosing instance exprs
        //
        // the extra formal to the constructor is added automatically if we do the above
        // {@link:MemberClassMunger}. {@link:InnerInfoPass}
        if (!plan.getJoinPoint().isStaticContext()) {
            //System.err.println(parentTypeDec.getType() + ", " + closureName);
            classDec.setIsInner(true);
            newExpr.setEnclosingInstanceExpr(
                                             ast.makeQualifiedThis(parentTypeDec.getType()));
        }

        // must be defined
        getMethodDec();

        Exprs args = plan.makeCallExprs(newExpr);

        Expr ret = ast.makeCall(getMethodDec().getMethod(), plan.getInstance(), args);

        Stmts retStmts;

        if (plan.getJoinPoint().getResultType().isVoid()) {
            retStmts =  ast.makeStmts(ast.makeStmt(ret));
        } else {
            ret = plan.getJoinPoint().getResultType().fromObject(ret);
            retStmts = ast.makeStmts(ast.makeReturn(ret));
        }

        retStmts = wrapCatchThrowable(retStmts, new HashSet(jp.getPossibleCheckedExceptions()));

        Expr test = plan.getDynamicTest();

        if (test == null) return retStmts;

        return ast.makeStmts(ast.makeIf(test,
                                        ast.makeBlock(retStmts),
                                        ast.makeBlock(closureWalker.makeDispatchCall())));

    }


    private Stmts wrapCatchThrowable(Stmts s, Set checkedExcs) {
        final AST ast = getAST();
        checkedExcs = new HashSet(checkedExcs);
        CatchClauses clauses = new CatchClauses(ast.getSourceLocation());
        checkedExcs.add(getTypeManager().getErrorType());
        checkedExcs.add(getTypeManager().getRuntimeExceptionType());
        checkedExcs = Type.filterTopTypes(checkedExcs);

        for (Iterator i = checkedExcs.iterator(); i.hasNext();) {
            Type excType = (Type) i.next();
            FormalDec formal = ast.makeFinalFormal(excType, "ajc$exc");
            BlockStmt body = ast.makeBlock(ast.makeThrow(ast.makeVar(formal)));
            clauses.add(ast.makeCatch(formal, body));
        }

        clauses.add(
                    ast.makeCatch(
                                  ast.makeFinalFormal(getTypeManager().getThrowableType(), "ajc$ignore"),
                                  ast.makeBlock(ast.makeThrow(getTypeManager().getErrorType(), "can't throw"))));

        Stmt tryStmt =
            new TryCatchStmt(ast.getSourceLocation(), ast.makeBlock(s), clauses);
        return ast.makeStmts(tryStmt);
    }

    public FormalDec getExtraFormal() {
        if (closureFormal == null) {
            closureFormal = makeExtraFormal();
        }
        return closureFormal;
    }

    private FormalDec closureFormal = null;

    private FormalDec makeExtraFormal() {
        return getAST().makeFinalFormal(
                                        getTypeManager().getAroundClosureType(),
                                        "ajc$closure");
    }

    /**
     * Is there a ProceedExpr in this tree, but not in an enclosed method or type dec
     */
    private boolean containsProceed(ASTObject node) {
        return new Walker(getCompiler()) {
                private boolean hasProceed = false;

                public boolean hasProceed(ASTObject node1) {
                    process(node1);
                    return hasProceed;
                }

                public ASTObject process(ASTObject object) {
                    if (hasProceed) return object;

                    if (object instanceof ProceedExpr) {
                        hasProceed = true;
                        return object;
                    }

                    if (object instanceof TypeDec || object instanceof MethodDec) {
                        return object;
                    }

                    return super.process(object);
                }
            }.hasProceed(node);
    }


    // int around(int i): call(int m(int)) && args(i): {
    //    ... proceed(37);
    //

    // ==>

    //  m(10) --> A.around1(new Closure1())

    //  class Closure1() {
    //     Object run(Object[] args) {
    //        return (Object)m((int)args[0] /* wants x */);
    //     }

    //   aspect A {
    //      int around1(Closure c, int i) {
    //         ... c.run(new Object[] { (Integer) 37 }) ...
    //      }
    //    }


    /**
     * Make proceed map
     *
     * Returns a map from VarDec|"this" to Integers
     */
    private Map makeProceedMap(AdvicePlan plan) {
        JoinPoint jp = plan.getJoinPoint();

        Formals formals = getFormals();
        Map proceedMap = new HashMap();
        for (int i = 0, N=formals.size(); i < N; i++) {
            FormalDec formal = formals.get(i);
            Expr value = (Expr)plan.bindings.get(formal);

            if (value instanceof VarExpr) {
                VarExpr var = (VarExpr) value;
                proceedMap.put(var.getVarDec(), new Integer(i));
            } else if (value instanceof ThisExpr) { //XXX Vile encoding
                Type toType = jp.getBytecodeType();
                proceedMap.put("this", new Integer(i));
            } else if (value instanceof NewInstanceExpr) {
                //??? this case shows up because of casting primitives to Object
                //??? we're pretty sure that those cases will always have a VarExpr as the param
                VarExpr var = (VarExpr) ((NewInstanceExpr) value).getArgs().get(0);
                proceedMap.put(var.getVarDec(), new Integer(i));
            } else {
                //proceedMap.put(, new Integer(i));
                //getCompiler().internalError(this, "unexpected AST type: " + value);
            }
        }

        return proceedMap;
    }

    /**
     * !!! this must be in-sync with the matching done above in makeProceedMap
     */
    private boolean isProceedChangable(Expr expr) {
        return (expr instanceof VarExpr)
        	|| (expr instanceof ThisExpr)
        	|| (expr instanceof NewInstanceExpr);
    }



    class ClosureProceedFixer extends MovingWalker {
        Stmts stmts;

        public ClosureProceedFixer(JavaCompiler compiler, Stmts stmts) {
            super(compiler);
            this.stmts = stmts;
        }

        public ASTObject remapProceed(CodeDec enclosingCodeDec, ProceedExpr proceedExpr) {
            final AST ast = this.getAST();
            Exprs args = ast.makeExprs();
            Exprs proceedArgs = proceedExpr.getArgs();
            for (int i=0, len=proceedArgs.size(); i<len; i++) {
                Expr proceedArg = proceedArgs.get(i);
                args.add(proceedArg.getType().makeObject(proceedArg));
            }

            Expr expr = getReturnType().fromObject(
                                                   ast.makeCall(ast.makeVar(getExtraFormal()), "run",
                                                                ast.makeExprs(ast.makeArray(this.getTypeManager().getObjectType(), args))));
            return expr;
        }

        public CodeDec fixCodeDecExceptions(CodeDec codeDec) {
            //System.out.println("code: " + codeDec);
            codeDec = (CodeDec)super.process(codeDec);
            Stmts stmts = codeDec.getBody().getStmts();
            Set excTypes = codeDec.getPossibleCheckedExceptions();

            //??? might want to get possible exceptions from stmts
            if (stmts.size() > 0) {
            	codeDec.getBody().setStmts(wrapCatchThrowable(stmts, excTypes));
            }
            return codeDec;
        }

        public ASTObject process(ASTObject object) {
            ASTObject original = object;
            if (object instanceof CodeDec) {
                return fixCodeDecExceptions((CodeDec)object);
            }

            object = super.process(object);
            if (object instanceof ProceedExpr) {
                return remapProceed(original.getEnclosingCodeDec(), (ProceedExpr) object);
            } else {
                return object;
            }
        }
    }

    private boolean fixedProceedCalls = false;
    private void fixProceedCalls() {
        if (fixedProceedCalls) return;
        fixedProceedCalls = true;
        ClosureProceedFixer fixer =
            new ClosureProceedFixer(getCompiler(), getBody().getStmts());
        getBody().walk(fixer);
    }



    protected CodeBody makeMethodBody() {
        fixProceedCalls();
        return getBody();
    }


    abstract class DispatchWalker extends MovingWalker {
        private boolean seenReturn = false;

        private Map /*<VarDec,FormalDec>*/ varMap = new HashMap();

        protected Exprs methodArgs = this.getAST().makeExprs();
        private Exprs baseMethodArgs = this.getAST().makeExprs();
        protected Formals methodFormals = this.getAST().makeFormals();
        protected MethodDec dispatchMethodDec;

        protected Stmts newBodyStmts;

        private Map proceedMap;
        protected CodeDec enclosingCodeDec;
        
        CallExpr soleCallExpr = null;
        boolean hasMultipleCallSites = false;


        public final Expr makeDispatchCallExpr(Expr thisExpr, Exprs args) {
            final AST ast = this.getAST();
            MethodDec dec = makeDispatchMethod();
//            Expr thisExpr;
//            if (enclosingCodeDec.isStatic()) {
//                thisExpr = ast.makeTypeExpr(enclosingCodeDec.getBytecodeType());
//            } else {
//                thisExpr = ast.makeQualifiedThis(enclosingCodeDec.getBytecodeType());
//            }

            CallExpr ret = ast.makeCall(dec, thisExpr, args);
            if (soleCallExpr == null && !hasMultipleCallSites) {
                soleCallExpr = ret;
            } else {
                soleCallExpr = null;
                hasMultipleCallSites = true;
            }
            return ret;
        }


        public Stmts makeDispatchCall() {
            final AST ast = this.getAST();
            Expr thisExpr;
            if (enclosingCodeDec.isStatic()) {
                thisExpr = ast.makeTypeExpr(enclosingCodeDec.getBytecodeType());
            } else {
                thisExpr = ast.makeQualifiedThis(enclosingCodeDec.getBytecodeType());
            }
            Expr callExpr = makeDispatchCallExpr(thisExpr, baseMethodArgs);
            Type returnType = enclosingCodeDec.getResultType();
            if (returnType.isVoid()) {
                return ast.makeStmts(ast.makeStmt(callExpr)); //, ast.makeReturn());
            } else {
                return ast.makeStmts(ast.makeReturn(callExpr));
            }
        }


        MethodDec makeDispatchMethod() {
            if (dispatchMethodDec != null) return dispatchMethodDec;

            final AST ast = this.getAST();
            Stmts body = newBodyStmts;
            Type returnType = enclosingCodeDec.getResultType();
            if (returnType.isVoid()) {
                returnType = this.getTypeManager().getObjectType();
                if (body.size() == 0) {
                    body.add(ast.makeReturn(ast.makeNull()));
                } else if ( ( body.get(body.size()-1) instanceof ReturnStmt ) ||
                            ( body.get(body.size()-1) instanceof ThrowStmt )) {
                    // skip
                } else {
                    body = ast.makeStmts(ast.makeIf(ast.makeLiteral(true), ast.makeBlock(body)),
                                         ast.makeReturn(ast.makeNull()));
                }
            }



            dispatchMethodDec = ast.makeMethod(ast.makeModifiers(Modifiers.FINAL |
                                                                 (enclosingCodeDec.isStatic() ? Modifiers.STATIC : 0)),
                                               returnType,
                                               genDispatchName(enclosingCodeDec),
                                               methodFormals,
                                               ast.makeBlock(body));
            enclosingCodeDec.getBytecodeTypeDec().addToBody(dispatchMethodDec);
            //dispatchMethodDec.owner = enclosingCodeDec;
            return dispatchMethodDec;
        }

        //XXX this adds evidence to the hypothesis that QualifiedThisExpr is bad
        public Expr moveThisExpr(ThisExpr thisExpr, Type thisType) {
            if (thisExpr instanceof QualifiedThisExpr)
                return thisExpr;
            return this.getAST().makeQualifiedThis(thisExpr.getType());
        }

        protected Expr makeThisRef() {
            if (proceedMap.containsKey("this")) {
                return makeArgRef(enclosingCodeDec.getBytecodeType(), ((Integer)proceedMap.get("this")).intValue());
            } else {
                final AST ast = this.getAST();
                if (enclosingCodeDec.isStatic()) {
                    return ast.makeTypeExpr(enclosingCodeDec.getBytecodeType());
                } else {
                    return ast.makeQualifiedThis(enclosingCodeDec.getBytecodeType());
                }
            }
        }

        protected Expr handleFreeVar(VarExpr var) {
            VarDec dec = var.getVarDec();

            FormalDec formal = (FormalDec) varMap.get(dec);
            if (formal == null) {
                formal = genFormal(dec);
            }
            return this.getAST().makeVar(formal);
        }


        DispatchWalker(JavaCompiler compiler, Map proceedMap) {
            super(compiler);
            this.proceedMap = proceedMap;
        }

        abstract Expr makeArgRef(Type type, int index);

        abstract Expr makeFreeRef(VarDec dec);

        /** generates all sorts of stuff and puts them in the right places */
        private FormalDec genFormal(VarDec dec) {
            final AST ast = this.getAST();
            final Type type = dec.getType();
            final String name = dec.getName();
            FormalDec formalDec =
                ast.makeFormal(type, name, dec.isFinal());

            Expr e;
            if (proceedMap.containsKey(dec)) {
                e = makeArgRef(dec.getType(), ((Integer)proceedMap.get(dec)).intValue());
            } else {
                e = makeFreeRef(dec);
            }

            baseMethodArgs.add(ast.makeVar(dec));
            methodArgs.add(e);
            methodFormals.add(formalDec);
            varMap.put(dec, formalDec);
            return formalDec;

        }

        public void setup(Stmts stmts) {
            final AST ast = this.getAST();
            enclosingCodeDec = stmts.getEnclosingCodeDec();
            setupBodyStmts(stmts);
        }
        
        // declare error:
        //    call(void setupBodyStmts(Stmts)) && !withincode(void setup(Stmts)):
        //    "only called from setup";
        protected final void setupBodyStmts(Stmts stmts) {
            newBodyStmts = (Stmts) process(stmts);
            
            
            ReturnAndThrowFixer fixer = new ReturnAndThrowFixer(getCompiler());
            
            newBodyStmts = (Stmts)fixer.process(stmts);
            this.seenReturn = fixer.hasSeenReturn();
        }
    }
    
    class ReturnAndThrowFixer extends Walker {
        private boolean seenReturn = false;
        public ReturnAndThrowFixer(JavaCompiler compiler) {
            super(compiler);
        }
        
        public boolean hasSeenReturn() { 
            return seenReturn;
        }
        
        public final ASTObject process(ASTObject object) {
            if (object instanceof TypeDec) return object;
            if (object instanceof AnonymousMethodExpr) return object;
            
            if (object instanceof ReturnStmt) {
                ReturnStmt returnStmt = (ReturnStmt)object;
                seenReturn = true;
                if (returnStmt.getExpr() == null) {
                    returnStmt.setExpr(this.getAST().makeNull());
                }
            } else if (object instanceof ThrowStmt) {
                seenReturn = true;
            }
            return super.process(object);
        }
    }


    class ClosureWalker extends DispatchWalker {
        private Exprs constructorArgs = this.getAST().makeExprs();
        private Formals constructorFormals = this.getAST().makeFormals();

        private List /*<FieldDec>*/ fields = new ArrayList();
        private Stmts fieldInitStmts = this.getAST().makeStmts();

        private FormalDec argsFormal;

        Exprs getConstructorArgs() {
            return constructorArgs;
        }
        Formals getConstructorFormals() {
            return constructorFormals;
        }

        Stmts getFieldInitStmts() {
            return fieldInitStmts;
        }
        List getFields() {
            return fields;
        } // could just get it from varMap

        BlockStmt getRunBody() {
            final AST ast = this.getAST();
            MethodDec dec = makeDispatchMethod();

            Expr thisExpr = makeThisRef();

            Expr callExpr = ast.makeCall(dec, thisExpr, methodArgs);
            Type returnType = enclosingCodeDec.getResultType();
            if (returnType.isVoid()) {
                return ast.makeBlock(ast.makeStmt(callExpr), ast.makeReturn(ast.makeNull()));
            } else {
                return ast.makeBlock(ast.makeReturn(returnType.makeObject(callExpr)));
            }
        }

        Expr makeArgRef(Type type, int index) {
            final AST ast = this.getAST();
            return type.fromObject(ast.makeArrayRef(ast.makeVar(argsFormal), index));
        }

        Expr makeFreeRef(VarDec dec) {
            return this.getAST().makeGet(genField(dec));
        }

        ClosureWalker(JavaCompiler compiler, Stmts stmts, Map proceedMap, FormalDec argsFormal) {
            super(compiler, proceedMap);
            this.argsFormal = argsFormal;
            setup(stmts);
        }

        /** generates all sorts of stuff and puts them in the right places */
        private FieldDec genField(VarDec dec) {
            final AST ast = this.getAST();
            final Type type = dec.getType();
            final String name = dec.getName();
            FieldDec fieldDec =
                ast.makeField(ast.makeModifiers(Modifiers.PRIVATE), type, name);

            FormalDec formalDec = ast.makeFinalFormal(type, name);
            Stmt stmt = ast.makeStmt(ast.makeSet(fieldDec, ast.makeVar(formalDec)));

            constructorArgs.add(ast.makeVar(dec));
            fields.add(fieldDec);
            constructorFormals.add(formalDec);
            fieldInitStmts.add(stmt);
            return fieldDec;
        }

        /**
         * Does the following transformations
         *
         * <ul>
         * 	<li> replaces free variables with field references to fields of the closure,
         * 		for which the free values are passed in at closure creation time </li>
         * 	<li> replaces references that are supposed to result in values that will be
         *	  	provided by proceed with references to the Object[] args of the run method
         * 	 	which is passed in by the around method, and contains the arguments
         * 		to proceed.  </li>
         * </ul>
         */
    }


    class InlineWalker extends DispatchWalker {
        private FormalDec aroundThisFormal = null;
        private Exprs baseArgs;
        private Expr baseThis;

        Exprs aroundMethodArgs = this.getAST().makeExprs();
        private Formals aroundMethodFormals = this.getAST().makeFormals();

        private MethodDec aroundMethodDec;

        Expr makeArgRef(Type type, int index) {
            final AST ast = this.getAST();
            //XXX odd encoding here
            return ast.makeVar(getFormals().get(index)); //type.fromObject(ast.makeVar(getFormals().get(index)));
        }

        Expr makeFreeRef(VarDec dec) {
            final AST ast = this.getAST();

            FormalDec formal = ast.makeFinalFormal(dec.getType(), dec.getName());
            aroundMethodFormals.add(formal);
            aroundMethodArgs.add(ast.makeVar(dec));

            return ast.makeVar(formal);
        }

        InlineWalker(JavaCompiler compiler, Stmts stmts, Map proceedMap, Exprs baseArgs, Expr baseThis) {
            super(compiler, proceedMap);
            this.baseArgs = baseArgs;
            this.baseThis = baseThis;
            setup(stmts);
        }

        public void setup(Stmts stmts) {
            final AST ast = this.getAST();
            enclosingCodeDec = stmts.getEnclosingCodeDec();

            Formals formals = getMethodDec().getFormals();
            for (int i = 0, N=formals.size(); i<N; i++) {
                FormalDec baseFormal = formals.get(i);
                FormalDec formal = ast.makeFormal(baseFormal.getType(), baseFormal.getName(),
                                                  baseFormal.isFinal());
                aroundMethodFormals.add(formal);
                aroundMethodArgs.add(baseArgs.get(i));
            }

            aroundThisFormal = ast.makeFinalFormal(AroundAdviceDec.this.getBytecodeType(), "this_");
            aroundMethodFormals.add(aroundThisFormal);
            aroundMethodArgs.add(baseThis);
            
            setupBodyStmts(stmts);
        }


        public MethodDec makeInlineAroundMethod() {
            if (aroundMethodDec != null) return aroundMethodDec;

            final AST ast = this.getAST();

            InlineProceedFixer fixer = new InlineProceedFixer(this.getCompiler());
            Stmts stmts = (Stmts)fixer.process(getBody().getStmts());

            aroundMethodDec = ast.makeMethod(ast.makeModifiers(Modifiers.FINAL |
                                                               (enclosingCodeDec.isStatic() ? Modifiers.STATIC : 0)),
                                             getResultType(),
                                             genAroundName(enclosingCodeDec),
                                             aroundMethodFormals,
                                             ast.makeBlock(stmts));
            enclosingCodeDec.getBytecodeTypeDec().addToBody(aroundMethodDec);
            aroundMethodDec.setLexicalType(AroundAdviceDec.this.getLexicalType());
            //aroundMethodDec.owner = enclosingCodeDec;
            return aroundMethodDec;
        }
        
        public void inlineSoleCallExpr(JoinPoint jp) {
            //System.out.println(soleCallExpr + ", " + hasMultipleCallSites);
            if (soleCallExpr == null || hasMultipleCallSites) {
                if (hasMultipleCallSites) {
                    jp.showWarning(AroundAdviceDec.this, "couldn't inline sole call expr (multiple calls)");
                }
                return;
            }
            
            if (!soleCallExpr.tryToInline(jp.getBytecodeType())) {
                jp.showWarning(AroundAdviceDec.this, "couldn't inline sole call expr");
            }          
        }

        class InlineProceedFixer extends CopyWalker {
            public Expr moveThisExpr(ThisExpr thisExpr, Type thisType) {
                return this.getAST().makeVar(aroundThisFormal);
            }

            public InlineProceedFixer(JavaCompiler compiler) {
                super(compiler);
            }

            public ASTObject remapProceed(ProceedExpr proceedExpr) {
                final AST ast = this.getAST();
                Exprs args = ast.makeExprs();
                Exprs baseArgs = methodArgs;
                for (int i=0, N=baseArgs.size(); i<N; i++) {
                    args.add(remap(methodFormals.get(i).getType(), baseArgs.get(i), proceedExpr.getArgs()));
                }
                Expr thisExpr = remap(enclosingCodeDec.getDeclaringType(), makeThisRef(), proceedExpr.getArgs());
                Expr ret = makeDispatchCallExpr(thisExpr, args);
                if (getResultType().isObject()) {
                    ret = ret.getType().makeObject(ret);
                }
                return ret;
            }

            protected Expr handleFreeVar(VarExpr var) {
                VarDec dec = var.getVarDec();
                Formals formals = getMethodDec().getFormals();
                int index = formals.indexOf(dec);
                if (index != -1) {
                    return this.getAST().makeVar(aroundMethodFormals.get(index));
                } else {
                    this.getCompiler().internalError("unbound free: " + var);
                    return var;
                }
            }

            private Expr remap(Type expectedType, Expr expr, Exprs proceedArgs) {
                //System.out.println("remap: " + expr);
                if (expr instanceof VarExpr) {
                    VarDec dec = ((VarExpr)expr).getVarDec();
                    int index = getFormals().indexOf(dec);
                    if (index != -1) {
                        Expr proceedArg = proceedArgs.get(index);
                        //System.out.println(proceedArg + ", " + expectedType);
                    	return expectedType.fromObject(proceedArg);
                    }
                }

		//System.err.println("no remapping for: " + expr);

                return (Expr)CopyWalker.copy(expr);
            }

            public ASTObject process(ASTObject object) {
                ASTObject original = object;

                object = super.process(object);
                if (object instanceof ProceedExpr) {
                    return remapProceed((ProceedExpr) object);
                } else {
                    return object;
                }
            }
        }
    }


    //XXX remove static state
    static int closureNumber = 1;
    private String genClosureName() {
        return "Closure" + closureNumber++;
    }
    String genDispatchName(CodeDec codeDec) {
        return getAST().makeGeneratedName("dispatch" + closureNumber++ + "_" + codeDec.getName());
    }

    static int aroundNumber = 1;
    private String genAroundName(CodeDec codeDec) {
        return getAST().makeGeneratedName("around" + aroundNumber++ + "_" + codeDec.getName());
    }


    private Stmts wrapStmtsInline(Stmts stmts, AdvicePlan plan) {
        final AST ast = getAST();
        JoinPoint jp = plan.getJoinPoint();

        FormalDec argsFormal = ast.makeFinalFormal(getTypeManager().getObjectType().getArrayType(),
                                                   "_args");

        InlineWalker inlineWalker = new InlineWalker(getCompiler(), stmts,
                                                     makeProceedMap(plan), plan.makeCallExprs(ast.makeNull()),
                                                     plan.getInstance());


        Exprs args = inlineWalker.aroundMethodArgs;

        CallExpr callAroundMethod = ast.makeCall(inlineWalker.makeInlineAroundMethod().getMethod(),
                                jp.makeThisExprOrType(), args);
                                
        callAroundMethod.setSourceLocation(jp.getTargetNode().getSourceLocation());
        
        Expr ret = callAroundMethod;

        Stmts retStmts;

        if (plan.getJoinPoint().getResultType().isVoid()) {
            retStmts =  ast.makeStmts(ast.makeStmt(ret));
        } else {
            ret = plan.getJoinPoint().getResultType().fromObject(ret);
            retStmts = ast.makeStmts(ast.makeReturn(ret));
        }

        //retStmts = wrapCatchThrowable(retStmts, new HashSet(jp.getPossibleCheckedExceptions()));

        Expr test = plan.getDynamicTest();

        if (test != null) {
            retStmts = ast.makeStmts(ast.makeIf(test,
                                        ast.makeBlock(retStmts),
                                        ast.makeBlock(inlineWalker.makeDispatchCall())));
        }
        
        if (getOptions().XOcodeSize) {
            inlineWalker.inlineSoleCallExpr(jp);
            if (!callAroundMethod.tryToInline(jp.getBytecodeType())) {
                jp.showWarning(this, "couldn't inline sole call expr");
            }
        }
            
        return retStmts;
    }


    boolean checkType(JoinPoint jp, Expr expr, Type type, String kind) {
        // think about numbers more carefully
        if (expr.getType().isReferenceType()) {
            type = type.getRefType();
        }
        if (!expr.getType().isCoercableTo(type)) {
            jp.showError(
                         expr,
                         expr.getType().getString()
                         + " is not coercable to expected "
                         + kind
                         + " type "
                         + type.getString());
            return false;
        }
        return true;
    }
    
    boolean checkImmutableReference(VarDec varDec, Expr expr) {
        if (!varDec.isFinal()) {
            varDec.showError("cflow-exposed state must be final (compiler limitation)");
            return false;
        }
        
        if ( !(expr instanceof VarExpr)
              || ((VarExpr)expr).getVarDec() != varDec ) {
            expr.showError("must be a simple reference to " + varDec.toShortString() + 
            	" (compiler limitation)");
            return false;
        }
        
        return true;
    }


    public boolean checkPlan(AdvicePlan plan) {
        JoinPoint jp = plan.getJoinPoint();

        Type resultType = resultTypeD.getType();
        Type pointResultType = jp.getResultType();

        if (resultType.isEquivalent(getTypeManager().anyType)) {
            showError("returns * is not allowed (use returns Object instead)");
            return false;
        }

        if (!resultType.isEquivalent(getTypeManager().getObjectType())) {
            if (resultType.isEquivalent(getTypeManager().voidType)) {
                if (!pointResultType.isEquivalent(getTypeManager().voidType)) {
                    jp.showError(this, "doesn't return void");
                    return false;
                }
            } else {
                if (!resultType.isAssignableFrom(pointResultType)) {
                    jp.showError(this, "returns " + pointResultType.toShortString());
                    return false;
                }
            }
        }

        for (Iterator i = proceedExprs.iterator(); i.hasNext(); ) {
            ProceedExpr proceedExpr = (ProceedExpr)i.next();
            Exprs proceedArgs = proceedExpr.getArgs();
            
            for (int j = 0, N=formals.size(); j < N; j++) {
                FormalDec formal = formals.get(j);
                Expr value = (Expr)plan.bindings.get(formal);
                
                if (!isProceedChangable(value)) {
                    if (!checkImmutableReference(formal, proceedArgs.get(j))) {
                        return false;
                    }
                }
                        
                
                if (!checkType(jp, proceedArgs.get(j), value.getType(), "arg")) {
                    return false;
                }
            }
        }

        if (resultType.isVoid()) return true;

        for (Iterator i = returnStmts.iterator(); i.hasNext(); ) {
            ReturnStmt returnStmt = (ReturnStmt)i.next();
            if (!checkType(jp, returnStmt.getExpr(), pointResultType, "return")) {
                return false;
            }
        }

        return true;

        //super.checkPlan(jp);
    }

    public Type getReturnType() {
        Type resultType;
        if (resultTypeD == null) {
            resultType = getTypeManager().voidType;
        } else {
            resultType = resultTypeD.getType();
        }
        return resultType;
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) {
        if (!walker.walkBodies()) {
            // remap CallExprs("proceed") -> ProceedExprs
            new Walker(getCompiler()) {
            	private boolean inCodeOrType = false;
                public ASTObject process(ASTObject o) {
                	if (!inCodeOrType && (o instanceof TypeDec || o instanceof CodeDec)) {
                		inCodeOrType = true;
                		o.walk(this);
                		inCodeOrType = false;
                	} else {
                		o.walk(this);
                	}
                	
                    
                    if (o instanceof CallExpr) {
                        CallExpr callExpr = (CallExpr) o;
                        if (callExpr.getId().equals("proceed")) {
                            ProceedExpr ret = new ProceedExpr(
                                                              callExpr.getSourceLocation(),
                                                              callExpr.getArgs(),
                                                              AroundAdviceDec.this);
                            addProceedExpr(ret); 
                            return ret;
                        }
                    } else if (o instanceof TryStmt) {
                        TryStmt tryStmt = (TryStmt) o;
                        //XXX should check that tryStmt.getBody() contains a proceed
                        tryStmt.setIsOptional(true);
                        return o;
                    } else if (o instanceof TypeDec) {
                        containsInners = true;
                    } else if (!inCodeOrType && o instanceof ReturnStmt) {
                        returnStmts.add(o);
                    }
                    return o;
                }
            }
                .process(getBody());
        }

        super.preScope(walker);
    }
    
    public void checkSpec() {
        super.checkSpec();
        checkAllProceedExprs();
    }
    
    private void checkAllProceedExprs() {
        for (Iterator i = proceedExprs.iterator(); i.hasNext(); ) {
            boolean valid = checkProceedExpr( (ProceedExpr)i.next() );
            if (!valid) i.remove();
        }
    }
    
    
    private boolean checkProceedExpr(ProceedExpr expr) {
        Formals formals = getFormals();
        Exprs args = expr.getArgs();
        
        int nFormals = formals.size();
        int nArgs = args.size();
        if (nArgs != nFormals) {
            expr.showError("expected " + nFormals + " args, found " + nArgs);
            return false;
        }
        
        for (int i=0; i < nFormals; i++) {
            FormalDec formal = formals.get(i);
            Expr arg = args.get(i);
            //System.out.println(arg + ", " + formal);
            if (!arg.getType().isMethodConvertableTo(formal.getType())) {
                arg.showTypeError(arg.getType(), formal.getType());
                return false;
            }
        }
        return true;
    }
    
    private void addProceedExpr(ProceedExpr expr) {      
        proceedExprs.add(expr);
    }

    //BEGIN: Generated from @child and @property
    protected TypeD resultTypeD;
    public TypeD getResultTypeD() {
        return resultTypeD;
    }
    public void setResultTypeD(TypeD _resultTypeD) {
        if (_resultTypeD != null)
            _resultTypeD.setParent(this);
        resultTypeD = _resultTypeD;
    }

    public AroundAdviceDec(
                           SourceLocation location,
                           Modifiers _modifiers,
                           TypeD _resultTypeD,
                           Formals _formals,
                           TypeDs __throws,
                           Pcd _pcd,
                           CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _pcd, _body);
        setResultTypeD(_resultTypeD);
    }
    protected AroundAdviceDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        AroundAdviceDec ret = new AroundAdviceDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null)
            ret.setModifiers((Modifiers) walker.process(modifiers));
        if (resultTypeD != null)
            ret.setResultTypeD((TypeD) walker.process(resultTypeD));
        if (formals != null)
            ret.setFormals((Formals) walker.process(formals));
        if (_throws != null)
            ret.setThrows((TypeDs) walker.process(_throws));
        if (pcd != null)
            ret.setPcd((Pcd) walker.process(pcd));
        if (body != null)
            ret.setBody((CodeBody) walker.process(body));
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch (childIndex) {
        case 0 :
            return modifiers;
        case 1 :
            return resultTypeD;
        case 2 :
            return formals;
        case 3 :
            return _throws;
        case 4 :
            return pcd;
        case 5 :
            return body;
        default :
            return super.getChildAt(childIndex);
        }
    }
    public String getChildNameAt(int childIndex) {
        switch (childIndex) {
        case 0 :
            return "modifiers";
        case 1 :
            return "resultTypeD";
        case 2 :
            return "formals";
        case 3 :
            return "throws";
        case 4 :
            return "pcd";
        case 5 :
            return "body";
        default :
            return super.getChildNameAt(childIndex);
        }
    }
    public void setChildAt(int childIndex, ASTObject child) {
        switch (childIndex) {
        case 0 :
            setModifiers((Modifiers) child);
            return;
        case 1 :
            setResultTypeD((TypeD) child);
            return;
        case 2 :
            setFormals((Formals) child);
            return;
        case 3 :
            setThrows((TypeDs) child);
            return;
        case 4 :
            setPcd((Pcd) child);
            return;
        case 5 :
            setBody((CodeBody) child);
            return;
        default :
            super.setChildAt(childIndex, child);
            return;
        }
    }
    public int getChildCount() {
        return 6;
    }

    public String getDefaultDisplayName() {
        return "AroundAdviceDec()";
    }

    //END: Generated from @child and @property

}
