/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.util.FuzzyBoolean;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
  * @grammar ???
  * @child Modifiers modifiers
  * @child GenTypeName declaringTypeName
  */
public abstract class DecPattern extends ASTObject {
    public abstract String toShortString();

    protected String makeDeclaringTypeString() {
        if (getDeclaringTypeName() == null) return "";
        return getDeclaringTypeName().toShortString() + '.';
    }

    public abstract String getLookupId();

    private boolean shownWarning = false;

    public FuzzyBoolean matches(JoinPoint jp) {
        if (!(modifiers.matches(jp.getTargetSO().getModifiers()))) {
            return FuzzyBoolean.NO;
        }

        if (getDeclaringTypeName() != null) {
            Type targetType = jp.getTargetType();
            if (targetType == null) {
                for (Iterator i = jp.getTargetTypes().iterator(); i.hasNext(); ) {
                    targetType = (Type)i.next();
//                    System.out.println("matching " + targetType + " with " + 
//                                       getDeclaringTypeName());
                    if (!matchDeclaringTypeName(jp, targetType, jp.getTargetSO())) {
                        return FuzzyBoolean.NEVER;
                    }
                    //System.out.println("   MATCHED!");
                }
                return FuzzyBoolean.YES;
            } else {
                if(matchDeclaringTypeName(jp, jp.getTargetType(), jp.getTargetSO())) {
                    return FuzzyBoolean.YES;
                } else {
                    return FuzzyBoolean.NO;
                }                 
            }
        } else {
            return FuzzyBoolean.YES;
        }
    }


    protected boolean isStaticMatch(SemanticObject so) {
        return so.isStatic();
    }
    
    boolean matchDeclaringTypeName(JoinPoint joinPoint, Type inType, SemanticObject so) {        
        //System.out.println("  matching dec? " + inType + ": " + so.toShortString());
        
        //XXX This design is very brittle and depends on the details of SemanticObjects
        //XXX and the way that they are held in Types
        SemanticObject matchingSO = inType.findMatchingSemanticObject(so);
        if (matchingSO == null) return false;
        
        if (isStaticMatch(so)) {
            if (matchingSO != so) return false;
        }

        if (getDeclaringTypeName().matches(inType)) {
            return true;
        }

        //??? for fields, static methods and constructors this can probably be optimized
        //??? to only go up the class hierarchy ignoring interfaces
        for (Iterator i = inType.getDirectSuperTypes().iterator(); i.hasNext(); ) {
            if(matchDeclaringTypeName(joinPoint, (Type)i.next(), so)) return true;
        }

        return false;
    }    
    
    //BEGIN: Generated from @child and @property
    protected Modifiers modifiers;
    public Modifiers getModifiers() { return modifiers; }
    public void setModifiers(Modifiers _modifiers) {
        if (_modifiers != null) _modifiers.setParent(this);
        modifiers = _modifiers;
    }
    
    protected GenTypeName declaringTypeName;
    public GenTypeName getDeclaringTypeName() { return declaringTypeName; }
    public void setDeclaringTypeName(GenTypeName _declaringTypeName) {
        if (_declaringTypeName != null) _declaringTypeName.setParent(this);
        declaringTypeName = _declaringTypeName;
    }
    
    public DecPattern(SourceLocation location, Modifiers _modifiers, GenTypeName _declaringTypeName) {
        super(location);
        setModifiers(_modifiers);
        setDeclaringTypeName(_declaringTypeName);
    }
    protected DecPattern(SourceLocation source) {
        super(source);
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return declaringTypeName;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "declaringTypeName";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setDeclaringTypeName((GenTypeName)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "DecPattern()";
    }
    
    //END: Generated from @child and @property
}

