/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.*;
import org.aspectj.util.FuzzyBoolean;

import java.util.*;

/**
  * @grammar if(test)
  * @child Expr test
  */
public class IfPcd extends Pcd {
    public String toShortString() {
        return "if(" + test.unparse() + ")";
    }
    
    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
        int context = w.inMember(true);
        super.walkInnerInfo(w);
        w.restoreContext(context);
    }
    
    public void checkStatic() { showNonStaticError(); }
    
    public void checkSpec() {
        getTest().assertType(getTypeManager().booleanType);
        new Walker(getCompiler()) {
            public void postProcess(ASTObject object) {
                if (object instanceof VarExpr) {
                    checkVarExpr((VarExpr)object);
                }
            }
        }.process(getTest());
    }

    private AdviceDec getEnclosingAdviceDec(Pcd pcd) {
        ASTObject parent = pcd.getParent();
        if (parent == null) return null;
        if (parent instanceof AdviceDec) return (AdviceDec)parent;
        if (parent instanceof Pcd) return getEnclosingAdviceDec((Pcd)parent);
        return null;
    }


    //!!! the parents of a formal get changed when a corresponding method is
    //!!! made for advice.  this works around that other bad design.
    private void checkVarExpr(VarExpr var) {
    	if (var.isLhs()) {
    		var.showError("can not assign to join point bound variables");
    	}
    	
    	
        AdviceDec adviceDec = getEnclosingAdviceDec(this);
        if (adviceDec == null) return;

        VarDec dec = var.getVarDec();
        
        //!!! better error messages for after returning/throwing
        if (adviceDec instanceof AfterReturningAdviceDec) {
            if (((AfterReturningAdviceDec)adviceDec).getExtraFormal() == dec) {
                var.showError("can not use after returning formal in 'if' pcd");
            }
            return;
        }
        if (adviceDec instanceof AfterThrowingAdviceDec) {
            if (((AfterThrowingAdviceDec)adviceDec).getExtraFormal() == dec) {
                var.showError("can not use after throwing formal in 'if' pcd");
            }
            return;
        }
    }

    public void walkScope(ScopeWalker walker) {
        if (!walker.walkBodies()) return;
        
        walker.pushBlockScope();
        //fixTestExpr();
        //XXX make sure that thisJoinPoint, ... are 'in scope'
        final AST ast = getAST();
        walker.addVarDec(
          ast.makeFinalFormal(getTypeManager().getJoinPointType(), "thisJoinPoint"));
        walker.addVarDec(
          ast.makeFinalFormal(getTypeManager().getJoinPointStaticPartType(), "thisJoinPointStaticPart"));

        super.walkScope(walker);
        
        walker.popScope();
    }
    
    
    private boolean containsInners = false;
    public JpPlanner makePlanner(final PlanData planData) {
        return new JpPlanner() {
            protected FuzzyBoolean fastMatch(JoinPoint jp) {
                //??? optimize this better
                return FuzzyBoolean.MAYBE;
            }
            public JpPlan makePlan(JoinPoint jp) {
                //System.out.println("test: " + test.unparse());
                Expr planTest = makeTestExpr(jp, planData);
                
                if (containsInners) {
                    showError("can't use anonymous types in if expressions (compiler limitation)");
                }
                //System.out.println("ptest: " + planTest.unparse());
                if (planTest == null) return JpPlan.NO_PLAN;
                
                JpPlan plan = new JpPlan(jp);
                plan.ifTest = planTest;
                //remapExprs(plan);
                return plan;
            }
        };
    }
    
    /*
    void fixTestExpr() {
        final VarDec aspectInstanceDec =
            IfPcd.this.getAST().makeVarDec(getLexicalType(),
                                        "aspect$instance", null);
        System.out.println(test.unparse());
        test = (Expr)new Walker(getCompiler()) {
            public ASTObject process(ASTObject node) {
                node = super.process(node);
                if (node instanceof ThisExpr) {
                    return IfPcd.this.getAST().makeVar(aspectInstanceDec);
                } else if (node instanceof VarExpr) {
                    return fixVarExpr( (VarExpr)node );
                }
                return node;
            }
        }.process(test);
        System.out.println(">>> " + test.unparse());
    }
    
    Expr fixVarExpr(VarExpr var) {
        if (var.getId().equals("thisJoinPoint")) {
            return new JpExpr(var.getSourceLocation(), getTypeManager().getJoinPointType()) {
                public Expr resolve(JoinPoint jp) {
                    return jp.makeDynamicJoinPointVarExpr();
                }
            };
        } else if (var.getId().equals("thisJoinPointStaticPart")) {
            return new JpExpr(var.getSourceLocation(), getTypeManager().getJoinPointStaticPartType()) {
                public Expr resolve(JoinPoint jp) {
                    return jp.makeStaticJoinPointVarExpr();
                }
            };
        } else if (var.getId().equals("thisEnclosingJoinPointStaticPart")) {
            return new JpExpr(var.getSourceLocation(), getTypeManager().getJoinPointStaticPartType()) {
                public Expr resolve(JoinPoint jp) {
                    return jp.makeStaticEnclosingJoinPointVarExpr();
                }
            };
        } else {
            return var;
        }
    }
    */
        
    
    Expr makeTestExpr(JoinPoint jp, PlanData planData) {
        final VarDec aspectInstanceDec =
            IfPcd.this.getAST().makeVarDec(planData.getAspectDec().getType(),
                                        "aspect$instance", null);
        
        CopyWalker copyWalker = new CopyWalker(getCompiler()) {
            public ASTObject process(ASTObject node) {
                node = super.process(node);
                if (node instanceof ThisExpr) {
                    return IfPcd.this.getAST().makeVar(aspectInstanceDec);
                } else if (node instanceof TypeDec) {
                    containsInners = true;
                }
                return node;
            }
        };
        
        return (Expr)copyWalker.process(test);
        
        //XXX should simplify test and check to see if it is always false at this
        //XXX particular join point and return null if so
        //return (Expr)test.copy();
    }    
    
    //BEGIN: Generated from @child and @property
    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }
    
    public IfPcd(SourceLocation location, Expr _test) {
        super(location);
        setTest(_test);
    }
    protected IfPcd(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        IfPcd ret = new IfPcd(getSourceLocation());
        ret.preCopy(walker, this);
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return test;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "test";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTest((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "IfPcd()";
    }
    
    //END: Generated from @child and @property
}

