/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.crosscuts.*;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
 * @grammar (parentName.)?id(formals)
 * @child TypeD parentName
 * @property String id
 * @child FormalsPattern formals
 * @property PointcutSO pointcut
 */
public class NamePcd extends Pcd {
    /** 1.2 substitute for Collections.EMPTY_MAP */
    private static final Map EMPTY_MAP 
        = Collections.unmodifiableMap(new TreeMap());

    public String toShortString() {
        String ret = id + formals.toShortString();
        if (parentName != null) {
            ret = parentName.toShortString() + '.' + ret;
        }
        return ret;
    }
    
    public void checkStatic() {
        //XXX needs to include a pointer back to the static context it is being used in
        getPointcut().getPcd().checkStatic();
    }
    
    //XXX helper method will go away
    public PointcutDec getPointcutDec() {
        return getPointcut().getPointcutDec();
    }
    
    /**
     *  This is a different form of name binding than in this, target and args
     */
    public boolean allowsNameBinding() { return true; }

    public JpPlanner makePlanner(PlanData planData) {
        if (planData.isPlanning(this)) {
            showError("recursive pointcut reference not allowed");
            return JpPlanner.NO_PLANNER;
        }
        
        PointcutSO innerPointcut = concretizePointcut(planData.getNameBindingType());
        if (innerPointcut.isAbstract()) {
            planData.getPlanMaker().showError("references an abstract pointcut -- see below");
            innerPointcut.getCorrespondingDec().showError("can't use abstract pointcut where concrete is required");
        }
        
        final JpPlanner planner =
                innerPointcut.getPcd().makePlanner(planData.makeInner(this));
        final Map remapFormals = makeFormalsRemap(innerPointcut);
        
        //??? it would be nice to do the actual remapping of FormalDecs at Planner
        //??? creation time, not during the actual makePlan call
        return new WrappedJpPlanner(planner) {
            public JpPlan makePlan(JoinPoint jp) {
                JpPlan plan = super.makePlan(jp);
                if (plan == JpPlan.NEVER_PLAN ||
                    plan == JpPlan.NO_PLAN) {
                        return plan;
                }
                plan.moveBindings(remapFormals);
                return plan;
            }
        };
    }
    
    Map makeFormalsRemap(PointcutSO pointcut) {
        GenTypeNames params = getFormals().getNames();
        Formals decFormals = pointcut.getFormals();
        
        final int N = params.size();
        if (decFormals.size() != N) {
            getFormals().showError("arguments don't match pointcut signature: " +
                        pointcut.getFormals().toShortString());
            return NamePcd.EMPTY_MAP;
        }
        
        Map remap = new HashMap();
        for (int i = 0; i < N; i++) {
            FormalDec formal = decFormals.get(i);
            GenTypeName param = params.get(i);
            
            //???
            if (param.matchesInstance(formal.getType()).maybeFalse()) {
                param.showError("doesn't match " + formal.getType().getString());
                break;
            }
            
            if (param instanceof VarTypeName) {
                FormalDec paramFormal = ((VarTypeName)param).getFormalDec();
                remap.put(formal, paramFormal);
            }
        }
        return remap;
    }

    //INTRO from ScopeWalker
    public ASTObject postScope(ScopeWalker walker) {
        if (!walker.walkBodies()) return this;
        if (pointcut != null) return this;
        bindPointcut();
        return super.postScope(walker);
    }
    
    private PointcutSO findPointcut() {
        if (isStaticReference()) {
            PointcutSO ret = parentName.getType().getPointcut(id, this, true);
            if (ret.isAbstract()) {
                this.showError("can't reference abstract pointcut using static qualified name: " +
                    ret.toShortString());
            }
            return ret;
        } else {
            Type inType = getLexicalType();
            PointcutSO ret = inType.getPointcut(id, this, false);
            if (ret != null) return ret;
            
            // if my immediate type doesn't containt a matching pointcut, then
            // walk up through enclosing types to see if any of them contain
            // a matching declaration for this pointcut name
            while (ret == null) {                
                inType = inType.getDeclaringType();
                if (inType == null || inType == getTypeManager().TYPE_NOT_FOUND) {
                    showError("named pointcut not found: " + toShortString());
                    return ret;
                }
                ret = inType.getPointcut(id, this, false);
            }
            
            // if I'm in an enclosing aspect, then this is a static reference
            // make it explicitly one
            setParentName(inType.makeTypeD());
            if (ret.isAbstract()) {
                this.showError("can't reference abstract pointcut using static qualified name: " +
                ret.toShortString());
            }
            
            return ret;
        }
    }
    
    private boolean isStaticReference() { return parentName != null; }
        
    private void bindPointcut() {
        pointcut = findPointcut();
        ((AspectJCompiler)getCompiler()).getCorrespondences().
                addPointsTo(findNonPcd(getParent()), pointcut.getPointcutDec());
    }

    private ASTObject findNonPcd(ASTObject node) {
        if (!(node instanceof Pcd)) return node;
        return findNonPcd(node.getParent());
    }



    public PointcutSO concretizePointcut(Type inType) {
        // fast fail to avoid generate duplicate error messages
        if (pointcut == null) return null;
        
        if (inType == null || isStaticReference()) return pointcut;

        PointcutSO ret = pointcut;
        
        //XXX i don't like the structure of this loop
        //ClassDec inTypeDec = inAspect;
        while (inType != ret.getDeclaringType()) {
            if (!ret.getModifiers().isPrivate() && ret.isAccessible(inType.getTypeDec().getBody())) {
                ret = inType.getPointcut(id, inType.getTypeDec().getBody(), true);
                break;
            }
            inType = inType.getTypeDec().getSuperClassType();
        }

        return ret;
    }
    
    //BEGIN: Generated from @child and @property
    protected TypeD parentName;
    public TypeD getParentName() { return parentName; }
    public void setParentName(TypeD _parentName) {
        if (_parentName != null) _parentName.setParent(this);
        parentName = _parentName;
    }
    
    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }
    
    protected FormalsPattern formals;
    public FormalsPattern getFormals() { return formals; }
    public void setFormals(FormalsPattern _formals) {
        if (_formals != null) _formals.setParent(this);
        formals = _formals;
    }
    
    protected PointcutSO pointcut;
    public PointcutSO getPointcut() { return pointcut; }
    public void setPointcut(PointcutSO _pointcut) { pointcut = _pointcut; }
    
    public NamePcd(SourceLocation location, TypeD _parentName, String _id, FormalsPattern _formals, PointcutSO _pointcut) {
        super(location);
        setParentName(_parentName);
        setId(_id);
        setFormals(_formals);
        setPointcut(_pointcut);
    }
    protected NamePcd(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        NamePcd ret = new NamePcd(getSourceLocation());
        ret.preCopy(walker, this);
        if (parentName != null) ret.setParentName( (TypeD)walker.process(parentName) );
        ret.id = id;
        if (formals != null) ret.setFormals( (FormalsPattern)walker.process(formals) );
        ret.pointcut = pointcut;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return parentName;
        case 1: return formals;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "parentName";
        case 1: return "formals";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setParentName((TypeD)child); return;
        case 1: setFormals((FormalsPattern)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "NamePcd(id: "+id+", "+"pointcut: "+pointcut+")";
    }
    
    //END: Generated from @child and @property
}

