/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;

import java.util.*;

import org.aspectj.compiler.base.*;

/**
 * @grammar modifiers pointcut name(formals) returns resultTypeD: pointcut;
 * @child Modifiers modifiers
 * @property String id
 * @child TypeD resultTypeD
 * @child Formals formals
 * @child Pcd pcd
 */

public class PointcutDec extends Dec {
    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
        int context = w.inMember(true);
        super.walkInnerInfo(w);
        w.restoreContext(context);
    }
    
    // INTRO from ASTFixerPass
    public ASTObject postFixAST(final ASTFixerPass fixer) {
        return null;
    }
    
    public void checkSpec() {
        if (isStatic()) {
            getCompiler().warnVersion("1.0beta1", this,
                "pointcut declarations may not be static");
        }
        
        if (resultTypeD != null) {
            getCompiler().warnVersion("1.0beta1", this,
                "pointcut declarations don't use returns clause");
        }
        
        if (isAbstract()) {
        	if (isPrivate()) {
                showError("abstract private pointcut can never be implemented");
            }
        }
        
//        if (!getDeclaringType().isAspect() && !isStatic()) {
//            showError("pointcut declarations outside of aspects must be declared static (since 1.0alpha1)");
//        }
        if (!isAbstract() && !(pcd == null || pcd instanceof EmptyPcd)) {
            for (int i=0; i<formals.size(); i++) {
                if (!formals.get(i).isBound) {
                    //System.out.println(pointcut);
                    formals.get(i).showError("not bound in this PCD");
                }
            }
        }
    }
    
    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) {
        walker.pushScope(makeBlockScope(walker));
        //XXX want to add thisJoinPoint, ... to scope somehow
    }
    public ASTObject postScope(ScopeWalker walker) { walker.popScope(); return this; }
    
    public String toShortString() {
        return modifiers.toShortString() + " " +
                    getDeclaringType().toShortString() + "." + id + 
                    formals.toShortString();
    }    
    
    public String getKind() { return "pointcut"; }

    public SemanticObject makeCorrespondingSemanticObject() {
        return new PointcutSO(this);
    }

    public boolean isAbstract(PlanData planData) {
        if (pcd == null || pcd instanceof EmptyPcd) return true;
        return false; //pcd.isAbstract(planData);
    }
    
    public boolean conflictsWith(Dec otherDec) {
        //System.out.println(this + " conflictsWith? " + otherDec);        
        if (!getId().equals(otherDec.getId())) return false;
        if (!this.isAccessible(otherDec) && !otherDec.isAccessible(this)) return false;
        
        return true;
    }
    
    
    void showOverrideError(Dec other, String message) {
        String overrideKind = "override";
        showError(toShortString() + " cannot " + overrideKind + " " +
                other.toShortString() + "; " + message);
    }
    
    public boolean checkOverride(Type inType, Dec otherDec) {
        PointcutDec other = (PointcutDec)otherDec;
        if (other.getModifiers().isFinal()) {
            showOverrideError(other, "overridden pointcut is final");
        } else if (getModifiers().isWeakerThan(other.getModifiers())) {
            showOverrideError(other,
                "attempting to assign weaker access privileges; was " +
                        other.getModifiers().toShortString());
        } else if (!getFormals().matches(other.getFormals())) {
            showOverrideError(other,
                "signatures must match exactly; expected "
                + other.getFormals().toShortString());
        }
        return false;
    }
    
    /*
    public boolean canOverride(Dec otherDec) {        
        if (this == otherDec) return true;
        
        //XXX AAH
        if (getDeclaringType() == null) return true;
        
        if (!getDeclaringType().isStrictSubtypeOf(otherDec.getDeclaringType())) return false;
        
        PointcutDec otherPointcutDec = (PointcutDec)otherDec;
        
        if (otherDec.isMoreAccessibleThan(this)) return false;
        
        // overriding pointcuts must have the exact same signature
        return getFormals().matches(otherPointcutDec.getFormals());
    }    
    */    
    
    //BEGIN: Generated from @child and @property
    protected Modifiers modifiers;
    public Modifiers getModifiers() { return modifiers; }
    public void setModifiers(Modifiers _modifiers) {
        if (_modifiers != null) _modifiers.setParent(this);
        modifiers = _modifiers;
    }
    
    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }
    
    protected TypeD resultTypeD;
    public TypeD getResultTypeD() { return resultTypeD; }
    public void setResultTypeD(TypeD _resultTypeD) {
        if (_resultTypeD != null) _resultTypeD.setParent(this);
        resultTypeD = _resultTypeD;
    }
    
    protected Formals formals;
    public Formals getFormals() { return formals; }
    public void setFormals(Formals _formals) {
        if (_formals != null) _formals.setParent(this);
        formals = _formals;
    }
    
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    public PointcutDec(SourceLocation location, Modifiers _modifiers, String _id, TypeD _resultTypeD, Formals _formals, Pcd _pcd) {
        super(location);
        setModifiers(_modifiers);
        setId(_id);
        setResultTypeD(_resultTypeD);
        setFormals(_formals);
        setPcd(_pcd);
    }
    protected PointcutDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        PointcutDec ret = new PointcutDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        ret.id = id;
        if (resultTypeD != null) ret.setResultTypeD( (TypeD)walker.process(resultTypeD) );
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return resultTypeD;
        case 2: return formals;
        case 3: return pcd;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "resultTypeD";
        case 2: return "formals";
        case 3: return "pcd";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setResultTypeD((TypeD)child); return;
        case 2: setFormals((Formals)child); return;
        case 3: setPcd((Pcd)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 4;
    }
    
    public String getDefaultDisplayName() {
        return "PointcutDec(id: "+id+")";
    }
    
    //END: Generated from @child and @property
}
