/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AspectJCompiler;

import java.util.*;

/**
 * @grammar declare soft: getTypeName: pointcut;
 * @child GenTypeName genTypeName
 * @child Pcd pcd
 */

public class SoftThrowableDec extends Dec implements JpPlannerMaker {

    public Modifiers getModifiers() { return getAST().makeModifiers(0); }

    public String getId() { return "declare"; }
    public String toShortString() {
        return "declare soft: " + genTypeName.toShortString();
    }
    public String getKind() { return "declare soft"; }
    
    public void checkSpec() {
        pcd.checkStatic();
    }
    
    public JpPlanner makePlanner(PlanData planData) {
        return new WrappedJpPlanner(pcd.makePlanner(planData)) {
                public JpPlan makePlan(JoinPoint jp) {
                    JpPlan plan = super.makePlan(jp);
                    // these aren't in the static model so they're not affected by this
                    if (jp.getSourceLocation() == null) return JpPlan.NO_PLAN;
                    if (!plan.isPossible()) return plan;
                    
                    ASTObject target = jp.getTargetNode();
                    if (target instanceof PossibleSoftThrowable) {
                        ((PossibleSoftThrowable)target).setSoftThrowable();
                    }
                    
                    return new SoftThrowablePlan(getCompiler(), plan);
                }
                
                public boolean isStaticPlanner() { return true; }
            };
    }

    private class SoftThrowablePlan extends JpPlan {
        JpPlan innerPlan;
        
        public SoftThrowablePlan(JavaCompiler compiler, JpPlan innerPlan) {
            super(innerPlan.joinPoint);
            
            this.innerPlan = innerPlan;
        }
                
        public int getPreSortOrder() { return SOFT_THROWABLE; }
        
        public String toString() {
            return "makesoft" + super.toString();
        }
        public void wrapJoinPoint(JoinPoint joinPoint) {
            joinPoint.setStmts(wrapSoftThrowable(joinPoint.getStmts(), joinPoint));
        }        
    } 
    

    CatchClause makeCatchClause(Type excType, JoinPoint jp) {
        final AST ast = getAST();
        
        FormalDec formalDec = ast.makeFormal(excType, "exc");
        Expr softThrowable = ast.makeNew(
                getTypeManager().getType("org.aspectj.lang", "SoftException"),
                ast.makeVar(formalDec));
        Stmt softThrow = ast.makeBlock(ast.makeThrow(softThrowable));
        JpPlan plan = genTypeName.makePlan(jp, ast.makeVar(formalDec));        
        
        Expr test = plan.getDynamicTest();

        BlockStmt body;
	if (test == null) {
	    body = ast.makeBlock(softThrow);
	} else {
	    body = ast.makeBlock(ast.makeIf(test,
	    				ast.makeBlock(softThrow),
	    				ast.makeBlock(ast.makeThrow(ast.makeVar(formalDec)))));
	}        
        return ast.makeCatch(formalDec, body);
    } 
    
    Stmts wrapSoftThrowable(Stmts stmts, JoinPoint jp) {
        final AST ast = getAST();
        
        CatchClauses clauses = new CatchClauses(ast.getSourceLocation());
        
        Set excTypes = ExceptionFinder.getPossibleExceptions(stmts, true);
        // we are throwing away some information here that COULD lead to more optimized
        // but much more complicated code below
        // instead of filtering we could sort
        excTypes = Type.filterTopTypes(excTypes);
        
        for (Iterator i = excTypes.iterator(); i.hasNext(); ) {
            Type excType = (Type)i.next();
            if (genTypeName.matchesInstance(excType).alwaysFalse()) continue;
            clauses.add(makeCatchClause(excType, jp));
        }
        
        return ast.makeStmts(new TryCatchStmt(ast.getSourceLocation(), ast.makeBlock(stmts), clauses));
    }
    
    //BEGIN: Generated from @child and @property
    protected GenTypeName genTypeName;
    public GenTypeName getGenTypeName() { return genTypeName; }
    public void setGenTypeName(GenTypeName _genTypeName) {
        if (_genTypeName != null) _genTypeName.setParent(this);
        genTypeName = _genTypeName;
    }
    
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    public SoftThrowableDec(SourceLocation location, GenTypeName _genTypeName, Pcd _pcd) {
        super(location);
        setGenTypeName(_genTypeName);
        setPcd(_pcd);
    }
    protected SoftThrowableDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        SoftThrowableDec ret = new SoftThrowableDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (genTypeName != null) ret.setGenTypeName( (GenTypeName)walker.process(genTypeName) );
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return genTypeName;
        case 1: return pcd;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "genTypeName";
        case 1: return "pcd";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setGenTypeName((GenTypeName)child); return;
        case 1: setPcd((Pcd)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "SoftThrowableDec()";
    }
    
    //END: Generated from @child and @property
}



