/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;

import org.aspectj.compiler.crosscuts.AspectJCompiler;

import org.aspectj.util.PartialOrder;

import java.util.*;

/**
 */

public class AdvicePlan extends JpPlan {  
    public AspectDec aspectDec;
    public AdviceDec adviceDec;
    public JpPlan innerPlan;

    public AdviceDec getAdviceDec() { return adviceDec; }
    public AspectDec getAspectDec() { return aspectDec; }

    public AdvicePlan(JoinPoint jp, AspectDec aspectDec, AdviceDec adviceDec, JpPlan innerPlan) {
        super(jp);
        this.aspectDec = aspectDec;
        this.adviceDec = adviceDec;
        this.innerPlan = innerPlan;
        
        //???
        JpPlan other = innerPlan;
        test = simplify(other.test);
        ifTest = simplify(other.ifTest);
        instance = other.instance;
        bindings = other.bindings;
        dependencies = other.dependencies;
    }

    void checkExceptions(AdviceDec adviceDec, JoinPoint joinPoint) {
        TypeDs _throws = adviceDec.getThrows();
        if (_throws == null) return;

        for (int i = 0; i < _throws.size(); i++) {
            Type t = _throws.get(i).getType();
            if (!joinPoint.canThrow(t)) {
                joinPoint.showError(adviceDec, 
                                    "can't throw " + t.getPrettyString());
                break;
            }
        }
    }
        
    
    public void wrapJoinPoint(JoinPoint joinPoint) {
        checkExceptions(adviceDec, joinPoint);
        adviceDec.wrapJoinPoint(joinPoint, this);
    }


    /**
     * Sorting is based purely on the adviceDec.  See AdviceDec.dominates for
     * detailed rules.
     */
    public int compareTo(Object o) {
        int ret = super.compareTo(o);
        if (ret != 0) return ret;
        
        if (!(o instanceof AdvicePlan)) return 0;
        AdvicePlan otherPlan = (AdvicePlan)o;

        if (this.getAdviceDec().dominates(otherPlan.getAdviceDec())) {
            return -1;
        } else if (otherPlan.getAdviceDec().dominates(this.getAdviceDec())) {
            return +1;
        } else {
            return 0;
        }
    }
    
    public boolean matches(JpPlan other) {
        if (other instanceof AdvicePlan) {
            return getAdviceDec() == ((AdvicePlan)other).getAdviceDec();
        } else {
            return false;
        }
    }
    
    public int getPreSortOrder() {
        if (adviceDec instanceof AroundAdviceDec) return AROUND;
        if (adviceDec instanceof BeforeAdviceDec) return BEFORE;
        return AFTER;
    }
    
    /**
     * Uses alphabetical order of aspects as a falback comparision rule.
     * This is a controversial choice.
     */
    public int fallbackCompareTo(Object o) {
        if (!(o instanceof AdvicePlan)) return 0;
        AdvicePlan otherPlan = (AdvicePlan)o;
        
        //XXX keep things deterministic, whether or not that's the right thing to 
        //XXX do is subject to debate -- the sort code works hard to make this possible
        return getAspectDec().getId().compareTo(otherPlan.getAspectDec().getId());
    }


    public Exprs makeCallExprs(Expr extraExpr) {
        Expr instance = getInstance();

        Exprs exprs = getCallExprs(adviceDec.getFormals()); //.copy();
        Type type = adviceDec.getExtraArgType();
        if (type != null) {
            //System.out.println("extra type: " + type + ", " + extraExpr.getType());
            if (type.isEquivalent(getTypeManager().getObjectType())) {
                //System.out.println("old extra expr: " + extraExpr);
                extraExpr = extraExpr.getType().makeObject(extraExpr);
                //System.out.println("new extra expr: " + extraExpr);
            }

            if (!extraExpr.getType().isCoercableTo(type)) {
                //System.out.println(extraExpr.getTypeD() + " not coercableTo " + typeD);
                return null; 
            }
            extraExpr = getAST().makeCast(type, extraExpr);
            exprs.add(extraExpr);
        }
        if (adviceDec.needsStaticEnclosingJoinPointFormal()) {
            exprs.add(joinPoint.makeStaticEnclosingJoinPointVarExpr());
        }
        if (adviceDec.needsStaticJoinPointFormal()) {
            exprs.add(joinPoint.makeStaticJoinPointVarExpr());
        }
        if (adviceDec.needsDynamicJoinPointFormal()) {
//            System.out.println(exprs);
//            System.out.println(joinPoint);
            exprs.add(joinPoint.makeDynamicJoinPointVarExpr());
        }
        return exprs;
    }

    public Stmt makeCall(Expr extraExpr) {
        Exprs exprs = makeCallExprs(extraExpr);
        if (exprs == null) return getAST().makeEmptyStmt();
        Expr call = adviceDec.makeCall(instance, exprs);
        call.setSourceLocation(joinPoint.getTargetNode().getSourceLocation());
        Stmt callStmt = getAST().makeStmt(call);
        //System.out.println("plan: " + this + " test: " + hasDynamicTest());
        return wrapDynamicTest(callStmt);
    }


    public Expr getInstance() {
        if (instance == null) return null;

        return getAST().makeCast(getAspectDec().getType(), instance);
    }

    public String toString() {
        if (adviceDec == null || joinPoint == null) {
            return super.toString();
        } else {
            return "plan(" + joinPoint.toString() + ", " + adviceDec.toShortString() + ")";
        }
    }

}
