/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.crosscuts.*;

import java.util.*;


public abstract class ExprJp extends CodeBodyJp {
    public Expr expr;

    AnonymousMethodExpr syntheticCall = null;

    FormalDec targetFormal = null;
    FormalDec thisFormal = null;

    public ExprJp(Expr expr, JoinPoint enclosingJoinPoint) {
        super(enclosingJoinPoint);
        this.expr = expr;
    }

    public abstract Dec getTargetDec();

    public String toString() {
        return getKind().toShortString() + 
            "(" + getTargetDec().toShortString() + ")";
    }


    FormalDec getTargetFormal() {
        if (getTargetExprType() == null) return null;

        if (targetFormal == null) {
            targetFormal = getAST().makeFormal(getTargetExprType(), "target");
        }   
        return targetFormal;
    }

    FormalDec getThisFormal() {
        return null;
        /*
        if (getThisExprType() == null) return null;
        
        if (thisFormal == null) {
            thisFormal = getAST().makeFormal(getThisExprType(), "_this");
        }   
        return thisFormal;
        */
    }
    
    public abstract Type getTargetExprType();
    public abstract Type getTargetType();

    public Expr makeTargetExpr() {
        FormalDec formal = getTargetFormal();
        if (formal == null) return null;
        else return getAST().makeVar(formal);
    }

    public Expr makeTargetExprOrType() {
        Expr ret = makeTargetExpr();
        if (ret != null) return ret;
        return getAST().makeTypeExpr(getTargetType());
    }

    public Expr makeThisExpr() {
        return super.makeThisExpr();
//         FormalDec formal = getThisFormal();
//         if (formal == null) return null;
//         else return getAST().makeVar(formal);
    }

    public Type getResultType() {
        return expr.getType();
    }

    public ASTObject getSourceLocation() { return expr; }

    protected abstract AnonymousMethodExpr makeSyntheticCall();

    protected abstract ASTObject makeInnerCall(Expr targetExpr, Exprs argsExprs);

    protected AnonymousMethodExpr makeAnonMethodExpr(Formals formals, 
                                                     Expr targetExpr, Exprs argsExprs)
    {
        final AST ast = getAST();

        FormalDec targetFormal = getTargetFormal();
        if (targetFormal != null) {
            formals.add(0, targetFormal);
            argsExprs.add(0, targetExpr);
        }

        FormalDec thisFormal = getThisFormal();
        if (thisFormal != null) {
            formals.add(0, thisFormal);
            argsExprs.add(0, ast.makeThis(getThisExprType()));
        }

        ASTObject innerCall = makeInnerCall(makeTargetExprOrType(), makeArgsExprs());

        Modifiers modifiers = ast.makeModifiers(Modifiers.PRIVATE);
        if (getThisExprType() == null) modifiers.setStatic(true);

        String innerName =
            getAST().makeGeneratedName(getTargetDec().getName() + "$" + getKind());

        Type innerResultType = getResultType();
        Stmt stmt;
        if (innerResultType.isVoid()) {
            if (innerCall instanceof Stmt) stmt = (Stmt)innerCall;
            else stmt = ast.makeStmt((Expr)innerCall);
        } else {
            stmt = ast.makeReturn((Expr)innerCall);
        }
        
        
        SourceLocation loc = getSourceLocation().getSourceLocation();
        //System.out.println(innerCall + ", " + loc);
        innerCall.setSourceLocation(loc);
        stmt.setSourceLocation(loc);

        MethodDec innerMethod = ast.makeMethod(modifiers, innerResultType, innerName,
                formals, ast.makeBlock(stmt));
        //innerMethod.setSourceLocation(loc);
        innerMethod.setAllEnclosingTypes(getBytecodeType()); // let ast.makeCall work

        return new AnonymousMethodExpr(loc, argsExprs, innerMethod);
    }
    

    AnonymousMethodExpr getSyntheticCall() {
        if (syntheticCall == null) {
            syntheticCall = makeSyntheticCall();
            syntheticCall.grabContext();
            replaceExprWith(syntheticCall);
        }
        return syntheticCall;
    }
    
    protected void replaceExprWith(Expr newExpr) {    
        expr.replaceWith(newExpr);
    }
    
    public CodeDec getBaseCodeDec() {
        return getSyntheticCall().getMethodDec();
    }

    public Stmts getStmts() {
        return getBaseCodeDec().getBody().getStmts();
    }
    public void setStmts(Stmts stmts) {
        getBaseCodeDec().getBody().setStmts(stmts);
    }

    public void finishJoinPoint() {
        super.finishJoinPoint();
        getBaseCodeDec().computeMinimalThrows();
    }
}
