/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;

import org.aspectj.compiler.crosscuts.AspectJCompiler;

import org.aspectj.util.PartialOrder;

import java.util.*;

/**
 */

public class JpPlan extends CompilerObject implements PartialOrder.PartialComparable {
    // no sorting by kind:
    public static final int CFLOWENTRY_ROOT = 0;
    public static final int EACHOBJECT = 0;
    public static final int AROUND = 1;
    public static final int AFTER = 1;
    public static final int BEFORE = 1;
    public static final int CFLOWENTRY_NOROOT = 2;
    public static final int SOFT_THROWABLE  = 0;
    
    public JoinPoint joinPoint;
    
    
    /**
     * Maps the formals of the advice (or other jp planner) to expressions at this join point
     * that provide the appropriate values
     */
    public Map/*FormalDec, Expr*/ bindings = new HashMap();
    public Expr test;
    public Expr ifTest;
    public Expr instance;
    
    public static final JpPlan NO_PLAN = new JpPlan(null) {
        public boolean isPossible() { return false; }
        public JpPlan or(JpPlan other) {
            return other;
        }
        
        public JpPlan and(JpPlan other) {
            return NO_PLAN;
        }
        
        public JpPlan not() {
            return JpPlan.ANY_PLAN;
        }
        public String toString() { return "NO_PLAN"; }
    };
    
    
    public static final JpPlan NEVER_PLAN = new JpPlan(null) {
        public boolean isPossible() { return false; }
        public JpPlan or(JpPlan other) {
            return this;
        }
        
        public JpPlan and(JpPlan other) {
            return this;
        }
        
        public JpPlan not() {
            return this;
        }
        public String toString() { return "NEVER_PLAN"; }
    };
    
    public static final JpPlan ANY_PLAN = new JpPlan(null) {
        public JpPlan or(JpPlan other) {
            if (other == JpPlan.ANY_PLAN) return JpPlan.ANY_PLAN;
            else return other.or(JpPlan.ANY_PLAN);
        }
        
        public JpPlan and(JpPlan other) {
            return other;
        }
        
        public JpPlan not() {
            return NO_PLAN;
        }
        
        public String toString() { return "ANY_PLAN"; }
    };

    public JpPlan(JoinPoint jp) {
        super(jp == null ? null : jp.getCompiler());
        this.joinPoint = jp;
    }
    
    public boolean isPossible() {
        if (ifTest == null) {
            return test == null || !isAlwaysFalse(test);
        }
        
        if (test == null) {
            return ifTest == null || !isAlwaysFalse(ifTest);
        }
        
        return !(isAlwaysFalse(test) || isAlwaysFalse(ifTest));
    }
    
    public JoinPoint getJoinPoint() { return joinPoint; }
    public Expr getInstance() {
        return instance;
//        if (instance == null) return null;
//        return getAST().makeCast(getAspectDec().getType(), instance);
    }
    
    public void wrapJoinPoint(JoinPoint jp) {
        System.err.println("shouldn't be making this JpPlan concrete!");
    }

    // this field is used solely to track dependencies between cflow entries
    // there's got to be a cheaper way to do this
    protected List dependencies = new LinkedList();
    public boolean dependsOn(Object o) {
        //System.out.println(this + " dependsOn? " + o + " :: " + dependencies);
        return dependencies.contains(o);
    }
    
    public void addDependency(Object o) {
        dependencies.add(o);
    }
    
    public int getPreSortOrder() { return 0; }

    /**
     * Sorting is based purely on the adviceDec.  See AdviceDec.dominates for
     * detailed rules.
     */
    public int compareTo(Object o) {
        if (!(o instanceof JpPlan) || this == o) return 0;

        JpPlan otherPlan = (JpPlan)o;
        
        if (getPreSortOrder() < otherPlan.getPreSortOrder())  return -1;
        if (getPreSortOrder() > otherPlan.getPreSortOrder())  return +1;
        
        return 0;
    }
    
    public int fallbackCompareTo(Object o) { return 0; } 
    
    public boolean matches(JpPlan other) {
        //XXX knows that this is only used for CalleeSideCallPoint
        return this == other; //false;
//        if (getAdviceDec() != null && getAdviceDec().needsCallSiteContext()) return false;
//        
//        if (getAdviceDec() != other.getAdviceDec()) return false;
//        
//        if (hasDynamicTest() || other.hasDynamicTest()) return false;
//        
//        return true;
    }

    public JpPlan or(JpPlan other) {
        if (other == NEVER_PLAN) return other;
        
        if (!other.isPossible()) return this;
        
        JpPlan plan = new JpPlan(joinPoint);
        
        plan.test = orExprs(test, other.test);
        plan.ifTest = orExprs(ifTest, other.ifTest);
        plan.instance = mergeInstance(instance, other.instance);
        plan.bindings = mergeBindings(bindings, other.bindings);
        plan.dependencies.addAll(other.dependencies);

        return plan;
    }
    
    public JpPlan and(JpPlan other) {
        if (other == NEVER_PLAN) return other;

        if (!other.isPossible()) return NO_PLAN;
        
        JpPlan plan = new JpPlan(joinPoint);
        
        plan.test = andExprs(test, other.test);
        plan.ifTest = andExprs(ifTest, other.ifTest);
        plan.instance = mergeInstance(instance, other.instance);
        plan.bindings = mergeBindings(bindings, other.bindings);
        plan.dependencies.addAll(other.dependencies);

        return plan;
    }

    public JpPlan not() {        
        if (!isPossible()) return JpPlan.ANY_PLAN;
        
        //if (!hasDynamicTest()) return JpPlan.NO_PLAN;
        if (test == null && ifTest == null) return JpPlan.NO_PLAN;

        JpPlan newPlan = new JpPlan(joinPoint);
        newPlan.test = notExpr(this.test);
        newPlan.ifTest = notExpr(this.ifTest);
        newPlan.dependencies = dependencies;
        return newPlan;
    }
    
    public void setInstanceExpr(Expr i) {
        instance = i;
    }
    
    public void addExprTest(Expr t1) {
        test = andExprs(test, t1);
    }
    
    public void addIfTest(Expr t1) {
        ifTest = andExprs(ifTest, t1);
    }
    
    Expr andExprs(Expr e1, Expr e2) {
        if (e1 == null) return e2;
        if (e2 == null) return e1;
        return getAST().makeBinop("&&", e1, e2);
    }
    
    Expr orExprs(Expr e1, Expr e2) {
        if (e1 == null) return e2;
        if (e2 == null) return e1;
        return getAST().makeBinop("||", e1, e2);
    }
    
    Expr notExpr(Expr e) {
        //System.out.println(this + "!" + e);
        
        if (e == null) return null;
        return getAST().makeUnop("!", e);
    }
    
    //XXX this will produce too many errors
    void checkEquivalent(FormalDec fd, Expr e1, Expr e2) {
        if (!e1.equals(e2)) {
        	joinPoint.showError(fd, "incompatible bindings for this formal: both " +
        					        e1.unparse() + " and " + e2.unparse());
        }
    }
    
    Map mergeBindings(Map b1, Map b2) {
        Map ret = new HashMap(b1);
        for (Iterator i = b2.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry)i.next();
            Object key = entry.getKey();
            if (ret.containsKey(key)) {
               checkEquivalent((FormalDec)key, (Expr)ret.get(key), (Expr)entry.getValue());
            } else {
                ret.put(key, entry.getValue());
            }
        }
        return ret;
    }
    
    Expr mergeInstance(Expr i1, Expr i2) {
        //XXX should detect error when both are non-null
        if (i1 == null) return i2;
        else return i1;
    }
    
    //???
    public void bindInstance(Expr expr) {
        instance = expr;
    }
    
    public void bindExpr(FormalDec formalDec, Expr expr) {
        //System.out.println("binding: " + formalDec + " to " + expr);
        if (bindings.containsKey(formalDec)) {
            checkEquivalent(formalDec, expr, (Expr)bindings.get(formalDec));
        } else {
            bindings.put(formalDec, expr);
        }
    }


    /* returns trueStmt if no dynamic test is needed, otherwise returns a block containing
     * a one-armed-if with the dynamic test.
     */
    public Stmt wrapDynamicTest(Stmt trueStmt) {
        Expr dynamicTest = getDynamicTest();
        if (dynamicTest != null) {
            final AST ast = getAST();
            return ast.makeBlock(ast.makeIf(dynamicTest, trueStmt));
        } else {
            return trueStmt;
        }
    }

    private Expr makeDynamicTest(Expr expr, boolean keepTypes) {
        if (expr == null) return null;
        if (expr != null) {
            expr = (Expr)new ExprMover(getCompiler(), keepTypes).process(expr);
            expr = simplify(expr);
            if (isAlwaysTrue(expr)) expr = null;
        }
        return expr;        
    }

    
    private Expr getBasicDynamicTest() {
        return makeDynamicTest(test, false);
    }

    
    private Expr getIfDynamicTest() {
        return makeDynamicTest(ifTest, true);
    }
    
    
    public Expr getDynamicTest() {
        return andExprs(getBasicDynamicTest(), getIfDynamicTest());
    }
    
    boolean isAlwaysFalse(Expr expr) {
        return expr instanceof BooleanLiteralExpr && 
                !((BooleanLiteralExpr)expr).getBooleanValue();
    }
    boolean isAlwaysTrue(Expr expr) {
        return expr instanceof BooleanLiteralExpr && 
                ((BooleanLiteralExpr)expr).getBooleanValue();
    }
    
    Expr simplify(Expr expr) {
        if (expr == null) return null;
        return (Expr)new BooleanExprSimplifier(getCompiler()).process(expr);
    }
    
    private class BooleanExprSimplifier extends Walker {
        public BooleanExprSimplifier(JavaCompiler compiler) { super(compiler); }
        
        public ASTObject process(ASTObject node) {
            node.walk(this);
            
            if (node instanceof InstanceofExpr) {
                return ((InstanceofExpr)node).simplify();
            } else if (node instanceof AndAndOpExpr) {
                AndAndOpExpr e = (AndAndOpExpr)node;
                if (isAlwaysFalse(e.getRand1()) || isAlwaysFalse(e.getRand2())) {
                    return node.getAST().makeLiteral(false);
                }
                if (isAlwaysTrue(e.getRand1())) return e.getRand2();
                if (isAlwaysTrue(e.getRand2())) return e.getRand1();
                return node;            
            } else if (node instanceof OrOrOpExpr) {
                OrOrOpExpr e = (OrOrOpExpr)node;
                if (isAlwaysTrue(e.getRand1()) || isAlwaysTrue(e.getRand2())) {
                    return node.getAST().makeLiteral(true);
                }
                if (isAlwaysFalse(e.getRand1())) return e.getRand2();
                if (isAlwaysFalse(e.getRand2())) return e.getRand1();
                return node;
            } else if (node instanceof ParenExpr) {
                ParenExpr e = (ParenExpr)node;
                if (isAlwaysFalse(e.getExpr())) return e.getExpr();
                if (isAlwaysTrue(e.getExpr())) return e.getExpr();
                return node;
            } else if (node instanceof LogNotOpExpr) {
                LogNotOpExpr e = (LogNotOpExpr)node;
                if (isAlwaysTrue(e.getRand1())) return node.getAST().makeLiteral(false);
                if (isAlwaysFalse(e.getRand1())) return node.getAST().makeLiteral(true);
                return node;
            } else {
                return node;
            }
        }
    }
    
    /**
     * Moves bindings of expressions to formals to a different set of formals
     */
    public void moveBindings(final Map remap) {
        // before we move our bindings map, use the current one on our test exprs
        test = getBasicDynamicTest();
        ifTest = getIfDynamicTest();
        
        for (Iterator i = remap.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry)i.next();
            Object key = entry.getKey();
            if (bindings.containsKey(key)) {
                bindings.put(entry.getValue(), bindings.get(key));
                //??? remove old entry
            } else {
                //??? error
            }
        }
    }

    
    private class ExprMover extends Walker {
        private boolean keepTypes;   
        public ExprMover(JavaCompiler compiler, boolean keepTypes) {
            super(compiler);
            this.keepTypes = keepTypes;
        }
        
        public ASTObject process(ASTObject object) {
            object.walk(this);
            if (object instanceof VarExpr) {
                VarExpr var = (VarExpr)object;
                if (var.getId().equals("thisJoinPoint")) {
                    return joinPoint.makeDynamicJoinPointVarExpr();
                } else if (var.getId().equals("thisJoinPointStaticPart")) {
                    return joinPoint.makeStaticJoinPointVarExpr();
                } else if (var.getId().equals("aspect$instance")) {
                    //XXX handle keepTypes
                    if (instance != null) return instance.copy();
                }
                
                VarDec varDec = var.getVarDec();
                Expr expr = (Expr)bindings.get(varDec);
                if (expr != null) {
                    expr = (Expr)expr.copy();
                    if (keepTypes) {
                        return getAST().forceCast(varDec.getType(), expr);
                    } else {
                        return expr;
                    }
                } else {
                    return var;
                }
            }
            return object;
        }
    }    

    void showFormalNotBound(FormalDec formalDec) {
        formalDec.showError("formal not bound on " + joinPoint);
    }

    public Exprs getCallExprs(Formals formals) {
        final AST ast = getAST();
        
        //System.out.println("formals: " + formals.toShortString() + ", " + bindings);
        
        Exprs exprs = ast.makeExprs();
        for (int i=0; i<formals.size(); i++) {
            Expr expr = (Expr)bindings.get(formals.get(i));
            if (expr == null) {
                showFormalNotBound(formals.get(i));
                break;
            }
            exprs.add(makeLegalExpr(expr, formals.get(i).getType()));
            //??? is this cast actually ever needed
        }
        return exprs;
    }
    
    Expr makeLegalExpr(Expr expr, Type type) {
        Type exprType = expr.getType();
        if (type.isObject()) {
            return exprType.makeObject(expr);
        } else {
            return getAST().makeCast(type, expr);
        }
    }

    public String toString() {
        if (joinPoint != null) return "plan(" + joinPoint.toString() + ")";
        else return super.toString();
    }
}
