/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.util.FuzzyBoolean;
import java.util.*;

public abstract class JpPlanner {
    public boolean isStaticPlanner() { return false; }

//    protected static class FastMatch {
//        FuzzyBoolean value;
//        private FastMatch(FuzzyBoolean value) {
//            this.value = value;
//        }
        
        

//    protected abstract static class FastMatch {
//        public abstract FastMatch or(FastMatch other);
//        public abstract FastMatch and(FastMatch other);
//        public abstract FastMatch not();
//    }
//
//    protected static final FastMatch YES = new FastMatch() {
//            public FastMatch or(FastMatch other) { return this; }
//            public FastMatch and(FastMatch other) { return other; }
//            public FastMatch not() { return NO; }
//        };
//
//    protected static final FastMatch NO = new FastMatch() {
//            public FastMatch or(FastMatch other) { return other; }
//            public FastMatch and(FastMatch other) { return this; }
//            public FastMatch not() { return YES; }
//        };
//
//    protected static final FastMatch MAYBE = new FastMatch() {
//            public FastMatch or(FastMatch other) {
//                if (other == YES) return other;
//                else return this;
//            }
//            public FastMatch and(FastMatch other) {
//                if (other == NO) return other;
//                else return this;
//            }
//            public FastMatch not() { return this; }
//        };
//

    
    protected abstract FuzzyBoolean fastMatch(JoinPoint jp);

    public boolean alwaysMatches(JoinPoint jp) {
        return fastMatch(jp).alwaysTrue();
    }

    public boolean sometimesMatches(JoinPoint jp) {
        return fastMatch(jp).maybeTrue();
    }
                                                   
    public JpPlan makePlan(JoinPoint jp) {
        FuzzyBoolean m = fastMatch(jp);
        if (m.alwaysTrue()) {
            return JpPlan.ANY_PLAN;
        } else if (m.alwaysFalse()) {
            return JpPlan.NO_PLAN;
        } else {
            jp.getCompiler().internalError("bad fastMatch result from " + this);
            return JpPlan.NEVER_PLAN;
        }
    }
    
    public JpPlanner or(JpPlanner other) {
        return new OrJpPlanner(this, other);
    }
    public JpPlanner and(JpPlanner other) {
        return new AndJpPlanner(this, other);
    }
    public JpPlanner not() {
        List newList = new ArrayList();
        newList.add(this);
        return new AndJpPlanner(Collections.EMPTY_LIST, newList);
    }
    public void walk(Object data) {}
    
    
    public static final JpPlanner NO_PLANNER = new JpPlanner() {
        public FuzzyBoolean fastMatch(JoinPoint jp) {
            return FuzzyBoolean.NO;
        }
    };
}

class OrJpPlanner extends JpPlanner {
    List list;
    
    public OrJpPlanner(List list) {
        this.list = list;
    }
    
    public OrJpPlanner(JpPlanner p1, JpPlanner p2) {
        list = new ArrayList();
        list.add(p1); list.add(p2);
    }     
    
    public void walk(Object data) {
        for (Iterator i = list.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            p.walk(data);
        }
    }
    
    public FuzzyBoolean fastMatch(JoinPoint jp) {
        FuzzyBoolean ret = FuzzyBoolean.NO;
        for (Iterator i = list.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            ret = ret.or(p.fastMatch(jp));
        }
        return ret;
    }   
    
    public JpPlan makePlan(JoinPoint jp) {
        JpPlan plan = JpPlan.NO_PLAN;
        for (Iterator i = list.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            plan = p.makePlan(jp).or(plan);
        }
        return plan;
    }
    
    public JpPlanner or(JpPlanner other) {
        List newList = new ArrayList(list);
        if (other instanceof OrJpPlanner) {
            newList.addAll(((OrJpPlanner)other).list);
        } else {
            newList.add(other);
        }
        return new OrJpPlanner(newList);
    }
    
    public JpPlanner and(JpPlanner other) {
        List newList = new ArrayList(list.size());
        for (Iterator i = list.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            newList.add(p.and(other));
        }
        return new OrJpPlanner(newList);
    }
    
    public JpPlanner not() {
        List newList = new ArrayList(list.size());
        //XXX this wreaks the SumOfProducts game, but should still be valid
        for (Iterator i = list.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            newList.add(p.not());
        }
        return new AndJpPlanner(newList, Collections.EMPTY_LIST);
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();
        
        for (Iterator i = list.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            buf.append(p.toString());
            if (i.hasNext()) buf.append(" || ");
        }
        return buf.toString();
    }
}


class AndJpPlanner extends JpPlanner {
    List/*JpPlanner*/ _true;
    List/*JpPlanner*/ _false;
    
    public AndJpPlanner(List _true, List _false) {
        this._true = _true;
        this._false = _false;
    }
    
    public AndJpPlanner(JpPlanner p1, JpPlanner p2) {
        _true = new ArrayList();
        _true.add(p1); _true.add(p2);
        _false = new ArrayList();
    }
    
    public void walk(Object data) {
        for (Iterator i = _true.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            p.walk(data);
        }
        for (Iterator i = _false.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            p.walk(data);
        }
    }
    
    public FuzzyBoolean fastMatch(JoinPoint jp) {
        FuzzyBoolean ret = FuzzyBoolean.YES;
        for (Iterator i = _true.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            ret = ret.and(p.fastMatch(jp));
        }

        for (Iterator i = _false.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            ret = ret.and(p.fastMatch(jp).not());
        }
        return ret;
    }
    
    public JpPlan makePlan(JoinPoint jp) {
//        System.out.println("jp: " + jp + ", " + this);
        JpPlan plan = JpPlan.ANY_PLAN;
        for (Iterator i = _true.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            plan = plan.and(p.makePlan(jp));
        }        
        for (Iterator i = _false.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
//            System.out.println("plan: " + plan);
//            System.out.println("  p: " + p.makePlan(jp));
//            System.out.println("  !p: " + p.makePlan(jp).not());
            
            plan = plan.and(p.makePlan(jp).not());
        }
//        System.out.println("ret: " + plan);
        return plan;        
    }
    
    public JpPlanner or(JpPlanner other) {
        return new OrJpPlanner(this, other);
    }
    
    public JpPlanner and(JpPlanner other) {
        List newTrue = new ArrayList(_true);
        List newFalse = new ArrayList(_false);
        
        if (other instanceof AndJpPlanner) {
            AndJpPlanner otherAnd = (AndJpPlanner)other;
            newTrue.addAll(otherAnd._true);
            newFalse.addAll(otherAnd._false);
        } else {
            newTrue.add(other);
        }
        return new AndJpPlanner(newTrue, newFalse);
    }
    
    public JpPlanner not() {
        List newList = new ArrayList();
        for (Iterator i = _true.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            newList.add(p.not());
        }
        for (Iterator i = _false.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            newList.add(p);
        }
        return new OrJpPlanner(newList);
    }
    
    public String toString() {
        StringBuffer buf = new StringBuffer();
        
        for (Iterator i = _true.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            buf.append(p.toString());
            if (i.hasNext() || _false.size() > 0) buf.append(" && ");
        }        
        for (Iterator i = _false.iterator(); i.hasNext(); ) {
            JpPlanner p = (JpPlanner)i.next();
            buf.append("!");
            buf.append(p.toString());
            if (i.hasNext()) buf.append(" && ");
        }
        return buf.toString();
    }
}
