/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.base;

import org.aspectj.debugger.request.*;
import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;
import org.aspectj.tools.ide.SourceLine;
import java.io.*;
import java.util.*;

import org.aspectj.util.LineNumberTableMapper;

/**
 * Debugger.java
 *
 *
 * Created: Wed Aug 30 13:14:41 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class AJDebugger extends Debugger {

    public static AJDebugger INSTANCE = null;

    private static final String        workingdir = "ajworkingdir";
    private static final String        sep        = File.separator;
    private static final char          sepChar    = File.separatorChar;
    private static final int           NO_LINE    = -1;
    private static final int           NO_MAP     = -2;
//    private              SymbolManager sm         = SymbolManager.getSymbolManager();
    private              HashMap       noAJCHash  = new HashMap();

    private File workingdirFile = null;
    public File getWorkingdir() {
        return workingdirFile;
    }
    public void setWorkingdir(String newWorkingdir) {
          try {
              newWorkingdir = new File(newWorkingdir).getCanonicalPath();
          } catch (Throwable t) {}
          setWorkingdir(new File(newWorkingdir));
    }
    private void setWorkingdir(File file) {
        if (file == null) {
            error("The workingdir is null: " + file);
            return;
        }
        if (!file.exists()) {
            error("The workingdir doesn't exist: " + file);
            return;
        }
        if (!file.isDirectory()) {
            String msg = "The workingdir isn't a directory: " + file;
            File parent = file.getParentFile();
            msg += "\nTrying to set workingdir with " + parent;
            error(msg);
            setWorkingdir(parent);
            return;
        }
        this.workingdirFile = file;
    }


    public AJDebugger(DebuggerApp app) {
        this(app, true);
    }

    public AJDebugger(DebuggerApp app, boolean readStartupFiles) {
        super(app);
        if (readStartupFiles) {
            readStartupFiles();
        }
        setWorkingDirDefault();
        INSTANCE = this;
    }

    /****************************** Input Mapping *******************************/

    public ClassLineBreakpointRequestAction createClassLineEvent(String className,
                                                                 int line,
                                                                 boolean stop)
        throws DebuggerException {
        return createClassLineEvent(className, line, stop, null);
    }

    public ClassLineBreakpointRequestAction createClassLineEvent(String className,
                                                                 int line,
                                                                 boolean stop,
                                                                 String text)
        throws DebuggerException {
        SourceLine sl = classToOutput(className, line);
        ClassLineBreakpointRequestAction a = null;
        if (stop) {
            a = new StopAtRequest(this, className, sl.line);
            a.setText(text);
        } else {
            a = new ClearAtRequest(this, className, sl.line);
            a.setText(text);
        }
        return a;
    }

    public SourceLineBreakpointRequestAction
        createSourceLineEvent(String sourceName,
                              int line,
                              boolean stop)
        throws DebuggerException {
        return createSourceLineEvent(sourceName, line, stop, null);
    }

    public SourceLineBreakpointRequestAction
        createSourceLineEvent(String sourceName,
                              int line,
                              boolean stop,
                              String text)
        throws DebuggerException {
        SourceLine sl = fileToOutput(sourceName, line);
        SourceLineBreakpointRequestAction a = null;
        String justSource = removeFullWorkingDir(sl.filename);
        if (stop) {
            a = new StopOnRequest(this, justSource, sl.line);
            a.setText(text);
        } else {
            a = new ClearOnRequest(this, justSource, sl.line);
            a.setText(text);
        }
        return a;
    }

    /****************************** Output Mapping ******************************/

    private SourceLine mapToSourceLine(String fullPath, int lineNumber) {
//        System.err.println(">> AJDebugger.mapToSourceLine fullPath: " + fullPath + ", lineNumber: " + lineNumber);
        return new SourceLine(fullPath, lineNumber);
//        return null;
//        SourceLine sl = sm.mapToSourceLine(fullPath, lineNumber-1);
//        if (sl == null) {
//            String strippedPath = removeFullWorkingDir(fullPath);
//            if (outputFileExists(strippedPath)) {
//                return new SourceLine(strippedPath, NO_MAP);
//            }
//            sl = noSourceLine(addSourcePath(strippedPath), lineNumber);
//        }
//        return sl;
    }

    public SourceLine getSourceLineFromClass(String className, int lineNumber) {
        String fullOutputPath = getFullOutputPathFromClass(className);
        return mapToSourceLine(fullOutputPath, lineNumber);
    }

    public SourceLine getSourceLineFromSource(String relativePath, int lineNumber) {
        String fullOutputPath = getFullOutputPath(relativePath);
        return mapToSourceLine(fullOutputPath, lineNumber);
    }

    public SourceLine getSourceLine(BreakpointRequestAction ba) {
        SourceLine sl = null;
        if (ba instanceof ClassLineBreakpointRequestAction) {
            ClassLineBreakpointRequestAction ca = (ClassLineBreakpointRequestAction)ba;
            return getSourceLineFromClass(ca.getClassName(), ca.getLine());
        }
        if (ba instanceof BreakpointRequestAction) {
            sl = ((BreakpointRequestAction) ba).sourceLine();
        }
        if (sl == null) {
            //String sourceName = ba.getSourceName();
            String sourceName = ba.getRealSourceName();
            String pkgWithSep = getPackagePathWithSeparator(ba.getLocation());
            if ("".equals(pkgWithSep)) {
                //!!! Want to get JUST the source name
                int islash = sourceName.lastIndexOf(File.separatorChar);
                if (islash == -1) {
                    islash = sourceName.lastIndexOf('/');
                }
                if (islash == -1) {
                    islash = sourceName.lastIndexOf('\\');
                }
                if (islash != -1) {
                    sourceName = sourceName.substring(islash+1);
                }
                //!!! end
            }
            String fullOutputPath = getFullOutputPath(pkgWithSep + sourceName);
            int lineNumber = ba.getLine();
            sl = mapToSourceLine(fullOutputPath, lineNumber);
        }
        return sl;
    }

    /*************************** Other *************************/

//    public SymbolManager sm() {
//        return sm;
//    }

    public void use(String sourcePath) {
        super.use(sourcePath);
        setWorkingDirDefault();
    }

    protected void setWorkingDirDefault() {
        if (getWorkingdir() == null) {
            String defaultWorkingdir = getSourcePath() + sep + workingdir;
            app.outln("Using default workingdir: " + defaultWorkingdir);
            setWorkingdir(defaultWorkingdir);
        }
    }

    protected void go(String className,
                      String vmArgs,
                      String commandLine,
                      boolean isSuspended,
                      int debugTraceMode) {
        if (getWorkingdir() == null) {
            setWorkingDirDefault();
        }
        super.go(className, vmArgs, commandLine, isSuspended, debugTraceMode);
    }

    /************************* SourceLine Mapping ******************/

    private int line(SourceLine sl) {
        return sl.line+1;
    }

    private SourceLine fileToOutput(String relativePath, int sourceLine)
        throws NonMappingSourceLineException {
        String fullPath = getFullSourcePath(relativePath);
        return mapToOutputLine(fullPath, sourceLine);
    }

    private SourceLine classToOutput(String className, int sourceLine)
        throws NonMappingSourceLineException {
        String fullPath = getFullSourcePathFromAJCClass(className);
        if (fullPath == null) {
            return noSourceLine(getFullSourcePathFromClass(className), sourceLine);
        }
        return mapToOutputLine(fullPath, sourceLine);
    }

    public boolean isAjcFile(String fullPath) {
        String relativePath = getRelativePath(fullPath);
        String fullOutputPath = getFullOutputPath(relativePath);
        File outputFile = new File(fullOutputPath);
        if (outputFile != null &&
            outputFile.exists() &&
            !outputFile.isDirectory()) {
            return true;
        }
        return false;
    }

    private SourceLine mapToOutputLine(String fullPath, int sourceLine)
        throws NonMappingSourceLineException {
//        System.err.println(">> AJDebugger.mapToOutputLine fullPath: " + fullPath + ", sourceLine: " + sourceLine);
        return new SourceLine(fullPath, sourceLine);
//        return null;
//        SourceLine sl =  sm.mapToOutputLine(fullPath, sourceLine);
//        if (sl == null) {
//            if (outputFileExists(fullPath)) {
//                throw new NonMappingSourceLineException(fullPath, sourceLine);
//            }
//            sl = noSourceLine(fullPath, sourceLine);
//        } else {
//            sl = new SourceLine(sl.filename, sl.line+1);
//        }
//        return sl;
    }

    public SourceLine noSourceLine(String fullPath, int line) {
        return new SourceLine(fullPath, line);
    }

    public SourceLine sourceLine(SourceLine sl) {
        sl.filename = removeSourcePath(sl.filename);
        return sl;
    }

    public SourceLine nonMappingSourceLine(String sourceName, int lineNumber) {
        return new SourceLine("<non-mapping>:" + sourceName + ":" + lineNumber, -1);
    }

    public SourceLine emptySourceLine() {
        return new SourceLine("<no-source>", -1);
    }

    private boolean outputFileExists(String fullPath) {
        String relativePath = removeSourcePath(fullPath);
        File file = new File(getFullOutputPath(relativePath));
        return file.exists();
    }

    public String __removeSourcePath(String fullPath) {
        fullPath = new File(fullPath).getAbsolutePath();
        int iroot = fullPath.toLowerCase().indexOf(getSourcePath().toLowerCase());
        if (iroot != -1) {
            String relativePath = fullPath.substring(getSourcePath().length());
            while (relativePath.length() > 0 &&
                   !Character.isJavaIdentifierPart(relativePath.charAt(0))) {
                relativePath = relativePath.substring(1);
            }
            return relativePath;
        }
        return fullPath;
    }

    public String addSourcePath(String relativePath) {
        if (!(relativePath.startsWith(getSourcePath()))) {
            try {
                return new File(getSourcePath() + sep +
                                relativePath).getCanonicalPath();
            } catch (Exception ioe) {
            }
        }
        return relativePath;
    }

    public String removeWorkingDir(String relativePath) {
        String target = workingdir + sep;
        int iwd = relativePath.indexOf(target);
        if (iwd != -1) {
            return relativePath.substring(target.length());
        }
        return relativePath;
    }

    public String removeWorkingDirFromFullPath(String fullPath) {
        String target = workingdir + sep;
        int iwd = fullPath.indexOf(target);
        if (iwd == -1) return fullPath;
        return fullPath.substring(0, iwd) +
            fullPath.substring(iwd + target.length());
    }

    public String removeFilePath(String fullPath, File comparison) {
        int isep = fullPath.indexOf(sep);
        if (isep == -1) return fullPath;
        for (int i = 0; i < fullPath.length(); i++) {
            char c = fullPath.charAt(i);
            if (isSep(c)) {
                String filePath = fullPath.substring(0,i);
                File file = new File(fullPath.substring(0, i));
                if (file.equals(comparison)) {
                    return fullPath.substring(i+1);
                }
            }
        }
        return fullPath;
    }
    public String removeFilePath(String fullPath, String comparison) {
        return removeFilePath(fullPath, new File(comparison));
    }
    public String removeFullWorkingDir(String fullPath) {
        return removeFilePath(fullPath, getWorkingdir());
    }
    public String removeSourcePath(String fullPath) {
        return removeCurrentPath(_removeSourcePath(fullPath));
    }
    public String _removeSourcePath(String fullPath) {
        return removeFilePath(fullPath, getSourcePath());
    }
    public String removeCurrentPath(String fullPath) {
        return removeFilePath(fullPath, ".");
    }

    private boolean isSep(char c) {
        return c == sepChar || c == '/' || c == '\\';
    }

    private String getFullSourcePathFromClass(String className) {
        String relativePath = getRelativeFilePathFromClass(className);
        return getFullSourcePath(relativePath);
    }

    public String getFullSourcePath(String relativePath) {
        return getFullFilePath(relativePath, false);
    }

    public String getFullOutputPath(Location loc, String realSourceName) {
        String decString = loc.declaringType().name();
        String pkg = "";
        int idot = decString.lastIndexOf(".");
        if (idot != -1) {
            pkg = decString.substring(0, idot).replace('.', sepChar) + sep;
        }
        String relativePath = pkg + realSourceName;
        return getFullOutputPath(relativePath);
    }

    public String getPackagePathWithSeparator(Location loc) {
        if (loc == null) return "";
        String decString = loc.declaringType().name();
        String pkg = "";
        int idot = decString.lastIndexOf(".");
        if (idot != -1) {
            pkg = decString.substring(0, idot).replace('.', sepChar) + sep;
        }
        return pkg;
    }

    public String getFullOutputPath(String relativePath) {
        if (getWorkingdir() == null) {
            return ""; //XXX this could fuck things up
        }
        String wdName = getWorkingdir().getName();
        int iwd = relativePath.indexOf(wdName);
        if (iwd != -1) {
            relativePath = relativePath.substring(iwd + wdName.length());
        }
        char c = relativePath.charAt(0);
        if (c == '/' || c == '\\' || c == sepChar) {
            relativePath = relativePath.substring(1);
        }
        return getFullFilePath(relativePath, true);
    }

    public String getFullOutputPathFromClass(String className) {
        return getFullOutputPath(getRelativeFilePathFromClass(className));
    }

    public String getFullFilePath(String relativePath, boolean inWorkingDir) {
        String fullName = null;
        File file;
        if (inWorkingDir) {
            file = new File(getWorkingdir() + sep + relativePath);
        } else {
            file = new File(relativePath);
        }
        if (file != null && file.exists()) return file.getAbsolutePath();
        try {
            relativePath = getRelativePath(relativePath);
            if (relativePath.toLowerCase().startsWith(getSourcePath().toLowerCase())) {
                fullName = relativePath;
            } else {
                boolean oldway = false;
                if (oldway) {
                    fullName = getSourcePath() + sep;
                    if (inWorkingDir) {
                        fullName += workingdir + sep;
                    }
                    fullName += relativePath;
                } else {
                    if (inWorkingDir) {
                        fullName = getWorkingdir() + sep + relativePath;
                    } else {
                        fullName = getSourcePath() + sep + relativePath;
                    }
                }
            }
            if (false) {
                return new File(fullName).getCanonicalPath();
            } else {
                return new File(fullName).getAbsolutePath();
            }
        } catch (Exception ioe) {
            ioe.printStackTrace(System.err);
            System.err.println("relativePath=" + relativePath);
            System.err.println("fullName=" + fullName);
            return new File(fullName).getAbsolutePath();
        }
    }

    public String getRelativePath(String fullPath) {
        int isourcePath = fullPath.toLowerCase().indexOf(getSourcePath().toLowerCase());
        if (isourcePath == -1) return fullPath;
        return fullPath.substring(isourcePath+getSourcePath().length());
    }

    //XXX Hack
    public String getRelativeFilePathFromClass(String className) {
        String newClassName = className.replace('.', sepChar);
        newClassName += ".java";
        return newClassName;
    }

    public String getFullSourcePathFromAJCClass(String className) {
//        return "";
        if (noAJCHash.get(className) != null) {
            return null;
        }
        String ajFile = getFullOutputPathFromClass(className);
        return ajFile;
//        LineNumberTableMapper mapper = new LineNumberTableMapper(
//        System.err.println(">> AJDebugger.getFullSourcePathFromAJCClass ajFile: " + ajFile);
//        SourceLine sourceLine = null;
//        String filePath = null;
//        sm = SymbolManager.getSymbolManager();
//        if (new File(ajFile).exists()) {
//            for (int i = 0; i < 10; i++) {
//                if ( (sourceLine = sm.mapToSourceLine(ajFile, i)) != null) {
//                    filePath = sourceLine.filename;
//                    return filePath;
//                }
//            }
//        } else {
//            noAJCHash.put(className, className);
//            return null;
//        }
//        return filePath;
    }

    public String getClassNameFromFullSourcePath(String fullSourcePath, int line) {
//        System.err.println(">> AJDebugger.getClassNameFromFullSourcePath fullSourcePath: " + fullSourcePath + ", line:" + line);
        return "";
//        Declaration dec = sm.getDeclarationAtLine(fullSourcePath, line);
//        if (dec == null) {
//            return null;
//        }
//        String className = dec.getDeclaringType();
//        String packageName = dec.getPackageName();
//        if (packageName != null && !"".equals(packageName = packageName.trim())) {
//            className = packageName + "." + className;
//        }
//        return className;
    }

    /**************** Location Mapping ****************/

    private final int limit = 10;
    public String sourceName(ReferenceType refType) {
//        System.err.println(">> AJDebugger.sourceName refType: " + refType);
        return "";
//        String className = refType.name();
//        String fullOutputPath = getFullOutputPathFromClass(className);
//        String sourceName = "<not-available>";
//        try {
//            sourceName = refType.sourceName();
//        } catch (AbsentInformationException aie) {
//        }
//        if (new File(fullOutputPath).exists()) {
//            for (int line = 0; line < limit; line++) {
//                SourceLine sl = sm().mapToSourceLine(fullOutputPath, line);
//                if (sl != null) {
//                    sourceName =  removeSourcePath(sl.filename);
//                    break;
//                }
//            }
//        }
//        return sourceName;
    }

    public String sourceName(Location loc) throws AbsentInformationException {
        return sourceLine(loc).filename;
    }

    public int lineNumber(Location loc) {
        try {
            return sourceLine(loc).line;
        } catch (AbsentInformationException aie) {
            aie.printStackTrace();
        }
        return super.lineNumber(loc);
    }

    public boolean locationMaps(Location loc) {
        String sourceName = "";
        try {
            sourceName = loc.sourceName();
        } catch (AbsentInformationException aie) {
            return true;
        }
        return isMappingRelativePath(sourceName, loc);
    }
    public boolean isMappingRelativePath(String sourceName, Location loc) {
        String fullPath = getPackagePathWithSeparator(loc) + sourceName;
        return isMappingFullPath(fullPath, loc);
    }
    public boolean isMappingFullPath(String fullPath, Location loc) {
        String fullOutputPath = getFullOutputPath(fullPath);
        if (!new File(fullOutputPath).exists()) {
            return true;
        }
        int lineNumber = loc.lineNumber();
        SourceLine sl = null;
        try {
            sl = sourceLine(loc);
        } catch (AbsentInformationException aie) {
            return true;
        }
        if (sl == null || sl.line < 0) {
            return false;
        }
        return true;
    }

    public SourceLine sourceLine(String sourceName,
                                 String fullOutputPath,
                                 int lineNumber) {
        SourceLine sl = mapToSourceLine(fullOutputPath, lineNumber);
        if (sl != null) {
            return sourceLine(sl);
        } else if (new File(fullOutputPath).exists()) {
            return nonMappingSourceLine(sourceName, lineNumber);
        }
        return noSourceLine(sourceName, lineNumber);
    }

    public SourceLine sourceLine(Location loc)
        throws AbsentInformationException {
        String sourceName = "<not-available>";
        try {
            sourceName = loc.sourceName();
        } catch (AbsentInformationException aie) {
        }
        int lineNumber = loc.lineNumber();
        String fullSrcPath = getFullSourcePath(getPackagePathWithSeparator(loc) +
                                               sourceName);
        SourceLine sl = sourceLineOfMethodThatShouldMap(loc, fullSrcPath);
        if (sl != null) return sl;
        String fullOutputPath =
            getFullOutputPath(getPackagePathWithSeparator(loc) +
                              sourceName);
        return sourceLine(sourceName, fullOutputPath, lineNumber);
    }

    public SourceLine sourceLineOfMethodThatShouldMap(Location loc,
                                                       String fullSrcPath) {
        return new SourceLineFinder(fullSrcPath).findMethod(loc);
    }

    public SourceLine sourceLineOfMethodThatShouldMap(String className,
                                                      String methodName,
                                                      String fullSrcPath) {
        return new SourceLineFinder(fullSrcPath).
            findMethod(className, methodName);
    }

    public String method(Location loc) {
        return method(loc.method());
    }

    public void setOptions(Options options) {
        super.setOptions(options);
        if (options.isSet("workingdir")) {
            setWorkingdir(options.getOpt("workingdir"));
        }
    }

    public String method(Method method) {
        if (getOptions().isSet("nodemangle")) return method.name();
        return demangleMethodName(method.name());
    }

    public String name(Method method) {
        if (getOptions().isSet("nodemangle")) return method.name();
        return demangleMethodName(method.name());
    }

    public String name(ReferenceType refType) {
        return refType != null ? refType.name() : "null-name";
    }


    /*********************** Decide when to stop ***********************/

    // We don't want to stop at bad events

    void handleBreakpointEvent(BreakpointEvent event) {
        try {
            SourceLine sl = sourceLine(event.location());
            if (!locationMaps(event.location())) {
                // This shouldn't happen, but we'll return
                return;
            }
        } catch (Exception e) {
        }
        super.handleBreakpointEvent(event);
    }

    void handleStepEvent(StepEvent event) {
        try {
            if (!locationMaps(event.location())) {
                String name = ((StepRequest)event.request()).
                    getProperty("name")+"";
                if (name.equals("next")) {
                    nextCommand();
                    return;
                } else if (name.equals("step")) {
                    stepCommand();
                    return;
                } else if (name.equals("step up")) {
                    stepUpCommand();
                    return;
                } else if (name.equals("stepi")) {
                    stepiCommand();
                    return;
                } else {
                    error("Bad property!");
                }
            }
        } catch (Exception e) {
        }
        super.handleStepEvent(event);
    }

    /****************************** Format Helpers ******************************/

    final static String sigToken             = "$signature";
    final static String memberToken          = "$member";
    final static String staticMemberToken    = "$static$member";
    final static String adviceToken          = "$reception";
    final static String aroundToken          = "$around$reception";
    final static String ajcToken             = "$ajc";

    public String demangleMethodName(String methodName) {
        String translatedMethodName = "";
        int idollar = methodName.indexOf("$");
        if (idollar == -1) {
            return methodName;
        }
        int iajc = methodName.indexOf(ajcToken);
        if (iajc != -1) {
            translatedMethodName = "<advice #";
            int i = iajc + ajcToken.length();
            while (i < methodName.length() &&
                   Character.isDigit(methodName.charAt(i))) {
                translatedMethodName += methodName.charAt(i);
                i++;
            }
            translatedMethodName += ">";
        } else {
            translatedMethodName = methodName.substring(0, idollar);
        }

        return /*methodName + " // " + */ translatedMethodName;
    }

    public boolean isAdvice(String methodName) {
        return methodName.indexOf(ajcToken) != -1;
    }

    /****************************** Exceptions ******************************/

    public static class NonMappingSourceLineException extends DebuggerException {
        public NonMappingSourceLineException(String source, int line) {
            super("The following does not map to an executable output line: " +
                  source + "::" + line);
        }
    }

    public static class NonMappingOutputLineException extends DebuggerException {
        public NonMappingOutputLineException(String outputFile, int line) {
            super("The following does not map to an executable source line: " +
                  outputFile + "::" + line);
        }
    }

    /****************************** Formatting ******************************/

    interface Formatter {
        public String format(Object o);
    }

    public String iter(List list) {
        return iter(list,
                    new Formatter() {
                            public String format(Object o) {
                                return "  " + o;
                            }});
    }

    public String fields(List fields) {
        String result = "";
        Iterator iter = fields.iterator();
        while (iter.hasNext()) {
            Field field = (Field) iter.next();
            try {
                result += field.type().name() + " " + field.name() + "\n";
            } catch (ClassNotLoadedException cnle) {
            }
        }
        return result;
    }

    public String methods(List methods) {
        String result = "";
        Iterator iter = methods.iterator();
        while (iter.hasNext()) {
            Method method = (Method) iter.next();
            String methString= method + "";
            int idot = method.declaringType().name().length();
            methString = methString.substring(0, idot) + ' ' +
                         methString.substring(idot + 1);
            result += methString + "\n";
        }
        return result;
    }

    public String classes(List classes) {
        String result = "** classes list **";
        Iterator iter = classes.iterator();
        while (iter.hasNext()) {
            result += "\n" + ((ReferenceType) iter.next()).name();
        }
        return result;
    }

    public String clazz(ReferenceType refType) {
        String result = "";
        if (refType instanceof ArrayReference) {
            result = "Array: " + refType.name();
        } else {
            String name = refType.name();
            result = "Class: " + name + "\n";
            try {
                Class clazz = Class.forName(name);
                result += "extends: " + clazz.getSuperclass().getName() + "\n";
                Class[] interfaces = clazz.getInterfaces();
                for (int i = 0; i < interfaces.length; i++) {
                    result += "implements: " + interfaces[i].getName() + "\n";
                }
            } catch (Exception e) {
            }
        }
        return result;
    }

    public String threads(List threads) {
        String result = "";
        Iterator iter = threads.iterator();
        Hashtable threadHash = new Hashtable();
        String[][] cols = new String[threads.size()][3];
        int[] max = new int[3];
        int col = 0;
        while (iter.hasNext()) {
            ThreadReference threadRef = (ThreadReference) iter.next();
            ThreadGroupReference threadGroupRef = threadRef.threadGroup();
            cols[col] = new String[3];
            cols[col][0] = "(" + threadRef.type().name() + ")" +
                           toHex(threadRef.uniqueID());
            cols[col][1] = threadRef.name();
            cols[col][2] = status(threadRef);
            for (int i = 0; i < max.length; i++) {
                if (cols[col][i].length() > max[i]) {
                    max[i] = cols[col][i].length();
                }
            }
            Vector v = (Vector)threadHash.get(threadGroupRef);
            if (v == null) {
                v = new Vector();
            }
            v.add(cols[col]);
            threadHash.put(threadGroupRef, v);
        }
        Enumeration keys = threadHash.keys();
        while (keys.hasMoreElements()) {
            ThreadGroupReference threadGroupRef =
                (ThreadGroupReference) keys.nextElement();
            result += "Group " + threadGroupRef.name() + ":\n";
            Vector v = (Vector) threadHash.get(threadGroupRef);
            for (int i = 0; i < v.size(); i++) {
                String[] stringCol = (String[]) v.get(i);
                result += "   ";
                result += space(stringCol[0], max[0]) + " ";
                result += space(stringCol[1], max[1]) + " ";
                result += space(stringCol[2], max[2]) + "\n";
            }
        }
        return result;
    }

    public String space(String str, int length) {
        String result = str;
        int diff = length - str.length();
        for (int i = 0; i < diff; i++) {
            result += " ";
        }
        return result;
    }

    public static String status(ThreadReference threadRef) {
        switch(threadRef.status()) {
            case ThreadReference.THREAD_STATUS_UNKNOWN:     return "unknown";
            case ThreadReference.THREAD_STATUS_ZOMBIE:      return "zombie";
            case ThreadReference.THREAD_STATUS_RUNNING:     return "running";
            case ThreadReference.THREAD_STATUS_SLEEPING:    return "sleeping";
            case ThreadReference.THREAD_STATUS_MONITOR:     return "monitor";
            case ThreadReference.THREAD_STATUS_WAIT:        return "waiting";
            case ThreadReference.THREAD_STATUS_NOT_STARTED: return "not started";
        }
        return "?";
    }

    public List frames(ThreadReference threadRef)
                    throws IncompatibleThreadStateException {
        List frames = new Vector();
        try {
            Iterator iter = threadRef.frames().iterator();
            while (iter.hasNext()) {
                StackFrame frame = (StackFrame) iter.next();
                if (locationMaps(frame.location())) {
                    frames.add(frame);
                }
            }
        } catch (InvalidStackFrameException isfe) {
        }
        return frames;
    }

    public String frames(List frames) {
        String result = "";
        Iterator iter = frames.iterator();

        int i = 1;
        String tabs = "";
        int size = frames.size();
        while ((size /= 10) != 0) tabs += " ";
        while (iter.hasNext()) {
            result += "  [" + (i++) + "] " + tabs;
            try {
                Location loc = ((StackFrame) iter.next()).location();
                result += name(loc.declaringType()) + "."
                    +     name(loc.method()) + " ";
                try {
                    String sourceName = sourceName(loc);
                    int lineNumber = lineNumber(loc);
                    if (lineNumber == -2) {
                        result += "(" + sourceName + ")";
                    } else if (lineNumber != -1) {
                        result += "(" + sourceName + ":" + lineNumber + ")";
                    } else {
                        result += "[class " + loc.declaringType().name() + "]";
                    }
                } catch (AbsentInformationException aie) {
                }
            } catch (InvalidStackFrameException isfe) {
                result += "<thread resumed>";
            }
            result += "\n";
        }
        return result;
    }

    public String threadGroups(List threadGroups) {
        String result = "";
        Iterator iter = threadGroups.iterator();
        int i = 1;
        while (iter.hasNext()) {
            ThreadGroupReference ref = (ThreadGroupReference) iter.next();
            result += (i++) + ". (" + ref.type().name() + ")" +
                      toHex(ref.uniqueID()) + " " + ref.name() + "\n";
        }
        return result;
    }

    public String toHex(long l) {
        return "0x" + Long.toHexString(l);
    }

    class SourceLinesFormatter implements Formatter {
        final int tabs = 10;
        final boolean arrow;
        public SourceLinesFormatter(boolean arrow) {
            this.arrow = arrow;
        }
        public String format(Object o) {
            if (!(o instanceof SourceManager.SourceLine)) {
                return o + "";
            }
            SourceManager.SourceLine sl = (SourceManager.SourceLine) o;
            String lineString = sl.getLineString();
            int lineNumber = sl.getLineNumber();
            String result = "";
            if (lineNumber > 0) {
                result += lineNumber + " ";
                if      (lineNumber <       10) result += "      ";
                else if (lineNumber <      100) result += "     ";
                else if (lineNumber <     1000) result += "    ";
                else if (lineNumber <    10000) result += "   ";
                else if (lineNumber <   100000) result += "  ";
                else                            result += "           ";
                result += (isStopped(sl) && arrow) ? "=>" : "  ";
                result += lineString;
            }
            return result;
        }
    }

    public String sourceLines(List sourceLines) {
        return sourceLines(sourceLines, false);
    }

    public String sourceLines(List sourceLines, boolean arrow) {
        return iter(sourceLines, new SourceLinesFormatter(arrow));
    }

    public String sourceLine(SourceManager.SourceLine sl) {
        Vector v = new Vector(1);
        v.add(sl);
        return sourceLines(v);
    }

    public String locals(List locals) {
        if (locals == null || locals.size() == 0) {
            return "No local variables";
        }
        String result = "";
        Iterator iter = locals.iterator();
        Vector args = new Vector();
        Vector vars = new Vector();
        while (iter.hasNext()) {
            LocalVariable local = (LocalVariable) iter.next();
            if (local.isArgument()) {
                args.add(local);
            } else {
                vars.add(local);
            }
        }
        result += "Method arguments:";
        for (int i = 0; i < args.size(); i++) {
            result += "\n" + local((LocalVariable) args.get(i));
        }
        result += (args.size() < 1 ? "\n" : "") + "Local variables:";
        for (int i = 0; i < vars.size(); i++) {
            result += "\n" + local((LocalVariable) vars.get(i));
        }
        return result;
    }

    public String local(LocalVariable local) {
        String result = "  " + local.name();
        try {
            StackFrame frame = getDefaultFrame();
            result += " = " + frame.getValue(local);
        } catch (NoVMException nvme) {
        } catch (Exception e) {
        }
        return result;
    }

    public String iter(List list, Formatter f) {
        String str = "";
        boolean newline = false;;
        Iterator iter = list.iterator();
        while (iter.hasNext()) {
            str += (newline ? "\n" : "") + f.format(iter.next());
            newline = true;
        }
        return str;
    }

    public String format(ExceptionEvent e) {
        String str = e.exception().referenceType().name();
        Location loc = e.catchLocation();
        str += " (";
        str += loc == null
            ? "uncaught" :
            "to be caught at: " + format(loc, e.thread());
        str += ")";
        str += "\n" + format((LocatableEvent) e);
        ObjectReference exception = e.exception();
        try {
            ThreadReference thread = threads.getThread();
            List methods = exception.referenceType().
                methodsByName("printStackTrace");
            Iterator iter = methods.iterator();
            while (iter.hasNext()) {
                Method method = (Method) iter.next();
                List argumentTypeNames = method.argumentTypeNames();
                if (argumentTypeNames.size() == 0) {
                    List arguments = new Vector();
                    int options = ObjectReference.INVOKE_SINGLE_THREADED;
                    exception.invokeMethod(thread, method, arguments, options);
                    break;
                }
            }
        } catch (DebuggerException de) {
          } catch (InvalidTypeException ite) {
          } catch (ClassNotLoadedException cnle) {
          } catch (IncompatibleThreadStateException itse) {
          } catch (InvocationException ie) {
        }
        return str;
    }

    public String format(LocatableEvent e) {
        return format(e.location(), e.thread());
    }

    public String format(Location loc, ThreadReference thread) {
        int lineNumber = lineNumber(loc);
        String method = method(loc);
        long codeIndex = loc.codeIndex();
        String str =
            "thread=\"" + thread.name() + "\", "                 +
            loc.declaringType().name()  + "." + method + "()"    +
            (lineNumber >  0 ? ", line=" + lineNumber : "")      +
            (codeIndex  >= 0 ? ", bci="  + codeIndex : "");
        try {
            SourceLine sl = sourceLine(loc);
            if (sl != null && sl.line > 0) {
                String smsl =
                    sourceLine(getSourceManager().getSourceLine(sl.filename,
                                                                sl.line - 1));
                if (!smsl.startsWith("null")) {
                    str += "\n" + smsl;
                }
            }
        } catch (AbsentInformationException aie) {
        }
        return str;
    }

    public String strip(String sourceName) {
        int islash = sourceName.lastIndexOf(File.separator);
        if (islash == -1) islash = sourceName.lastIndexOf('/');
        if (islash == -1) islash = sourceName.lastIndexOf('\\');
        if (islash != -1) return sourceName.substring(islash+1);
        return sourceName;
    }

    public String methodName(LocatableEvent e) {
        return method(e.location());
    }

    public static String d() { return "AJDebugger"; }
    public String toString() { return d(); }

    public String toLongString() {
        Object[] objects = new Object[] {
            "Sourcepath", getSourcePath(),
            "Workingdir", getWorkingdir(),
            "Running class", runningClass,
            "Status", status,
            "Old class", oldClass,
            "Old VM args", oldVmArgs,
            "Old command line", oldCommandLine,
        };
        return format(map(objects));
    }

    private Map map(Object[] objects) {
        Map map = new HashMap();
        for (int i = 0; i < objects.length; i += 2) {
            map.put(objects[i], objects[i+1]);
        }
        return map;
    }

    private String format(Map map) {
        Iterator iter;
        iter = map.keySet().iterator();
        List keys = new Vector();
        List vals = new Vector();
        while (iter.hasNext()) {
            Object keyObj = iter.next();
            String key = keyObj + "";
            String val = map.get(keyObj) + "";;
            keys.add(key);
            vals.add(val);
        }
        int longest = -1;
        iter = keys.iterator();
        while (iter.hasNext()) {
            int length = (iter.next() + "").length();
            if (length > longest) {
                longest = length;
            }
        }
        String str =
            "-----------------------------------" +
            "-----------------------------------\n";
        for (int i = 0; i < keys.size(); i++) {
            String key = keys.get(i)+"";
            String val = vals.get(i)+"";
            while (key.length() < longest) {
                key = key + " ";
            }
            str += key + "  : " + val + "\n";
        }
        str +=
            "-----------------------------------" +
            "-----------------------------------\n";
        return str;
    }
}
