/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.base;

//import org.aspectj.debugger.DebuggerBuild;
import org.aspectj.debugger.request.*;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import com.sun.jdi.connect.*;
import com.sun.jdi.request.*;
import java.io.*;
import java.util.*;
import java.text.*;

/**
 * Debugger.java
 * The main debugger application.
 *
 * Created: Mon Aug 21 11:29:37 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class Debugger implements CommandReceiver, ErrorLoggable {

    public static boolean demoMode = true;

    protected String          currentStoppedSource    = "";
    protected String          date                    = "00/11/14"; //TODO Change
    protected String          version                 = "ajdb, version " + date;
    protected String          prompt                  = ">";
    protected String          runningClass;
    protected String          status;
    protected String          oldClass                = "";
    protected String          oldVmArgs               = "";
    protected String          oldCommandLine          = "";

    private List            requestActions;
    private List            monitorList;
    private List            excludes;
    private List            importPaths;
    private List            filePaths;

    protected Threads         threads;
    protected DebuggerApp     app;
    
    private Parser          parser;
    private SourceManager   sourceManager;
    private Evaluator       evaluator;
    private EventSpinner    spinner;
    private Options         options                 = new Options();

    private int             currentStoppedLine      = -1;
    private int             monitors                = 1;
    private boolean         isRunning               = false;
    private boolean         isStandAlone            = true; //TODO: Will be false
    private boolean         isStopped               = false;

    public static boolean debug = false; //true
    public static void db(Object o) {
        if (debug) {
            Thread.yield();
            System.err.println("<DEBUGGER>: " + o);
        }
    }

    /**
     * Start an empty debugger, mainly for testing...
     */
    public Debugger() {
        this(new DebuggerAdapter());
    }

    /**
     * Use the application passed in to start and start the debugger.
     *
     * @param app The DebuggerApp that will drive the application.
     */
    public Debugger(DebuggerApp app) {
        super();
        me();
        this.app       = app;
        app.outln("Initializing ajdb...");
        new Checker(this).check();
        spinner        = new EventSpinner(this);
        threads        = new Threads(this);
        parser         = new Parser(this);
        sourceManager  = new SourceManager(this);
        evaluator      = new Evaluator(this);
        requestActions = new Vector();
        monitorList    = new Vector();
        excludes       = new Vector();
        importPaths    = new Vector();
        filePaths      = new Vector();
        addImportPath("");
    }

    /**
     * Returns a VM for the arguments.
     *
     * @param vmArgs      arguments sent to the VM.
     * @param className   name of the class to invoke.
     * @param commandLine commandline arguments as a String.
     * @return            a new VM.
     */
    private void getVM(final String vmArgs,
                       final String className,
                       final String commandLine) {
        if (vm != null) {
            vm = null;
        }
        if (Modes.isJBuilder() || Modes.isForte()) {
            vm = safeConnect(vmArgs, className, commandLine);
        } else {
            vm = findVM(vmArgs, className, commandLine);
        }
    }

    /**
     * Safely connects the VM and handles any errors
     * that arise.
     *
     * @param vmArgs      arguments sent to the VM.
     * @param className   name of the class to invoke.
     * @param commandLine commandline arguments as a String.
     * @return            a new VM.
     */
    private VirtualMachine safeConnect(String vmArgs,
                                       String className,
                                       String commandLine) {
        try {
            return connectCommand(vmArgs, className, commandLine);
        } catch (DebuggerException de) {
            app.handleDebuggerException(de);
        } catch (Throwable t) {
            app.handleInternalException(t);
        }
        return null;
    }

    private void saveClassName(String oldClassName) {
        this.oldClass = oldClassName;
    }

    private static boolean empty(String str) {
        return str == null || str.length() == 0 || str.trim().equals("");
    }
    
    /**
     * Starts up the debugger.
     *
     * @param className      name of the class to load, this can
     *                       be null.
     * @param vmArgs         arguments going directly to the VM.
     * @param commandLine    command line arguments as a String.
     * @param isSuspended    true is we should wait at main(String[]).
     * @param debugTraceMode true is we should use tracing mode.
     */
    protected void go(String className,
                      String vmArgs,
                      String commandLine,
                      boolean isSuspended,
                      int debugTraceMode) {

        // get rid of the old Threads
        (threads = new Threads(this)).flush();

        // Save the previous run...
        if (empty(className))     className = oldClass;
        if (className != null)    oldClass  = className;
        if (empty(vmArgs))        vmArgs    = oldVmArgs;
        if (vmArgs != null)       oldVmArgs = vmArgs;
        if (empty(commandLine))   commandLine = oldCommandLine;
        if (commandLine != null)  oldCommandLine = commandLine;
        if (Modes.isJBuilder() || Modes.isForte() ||
            (className != null && !className.trim().equals(""))) {
            getVM(vmArgs, className, commandLine);
        }

        String error =
            "The class '" + className + "', args '" + vmArgs + "'" +
            "\n, and/or command line '" + commandLine + "' are invalid";

        // try to start the VM
        try {
            startUpVM(vm, error, className, debugTraceMode, isSuspended);
        } catch (ClassNotFoundException cnfe) {
            handleClassNotFound(cnfe);
        } catch (NoClassDefFoundError ncdfe) {
            handleClassNotFound(ncdfe);
        }

        // tell the listeners we've started
        fireStartEvent();
        return;
    }

    private void handleClassNotFound(Throwable cnfe) {
        cnfe.printStackTrace();
    }

    private final static String NO_ERROR = "default.error";
    private final static String NO_CLASS = "default.class";
    private void startUpVM (VirtualMachine newVm)
        throws ClassNotFoundException,
               NoClassDefFoundError  {
        startUpVM(newVm, NO_ERROR, NO_CLASS,
                  VirtualMachine.TRACE_NONE, false);
    }

    /**
     * Starts up the passed in VM.
     *
     * @param newVm          the new VM to use and start up.
     * @param error          error message to print if problems
     *                       arise -- this can be null.
     * @param debugTraceMode trace mode to use.
     * @param isSuspended    true to stop at main(String[]).
     */
    private void startUpVM(VirtualMachine newVm,
                           String error,
                           String className,
                           int debugTraceMode,
                           boolean isSuspended)
        throws ClassNotFoundException,
               NoClassDefFoundError {
        this.vm = newVm;        
        if (vm == null) {
            if (!Modes.isJBuilder() && !Modes.isForte()) {
                error("The VM is null.\n" + error);
            }
            return;
        }        
        Process process = vm.process();
        redirect(process.getErrorStream(), "error-dumper");
        redirect(process.getInputStream(), "input-dumper");
        vm.setDebugTraceMode(debugTraceMode);
        spinner.start();
        wakeUpRequests();
        if (isSuspended) {
            try {
                stopInCommand(className, "main(java.lang.String[])");
            } catch (NoVMException nvme) {
                app.handleInternalException(nvme);
                return;
            } catch (DebuggerException de) {
                app.handleInternalException(de);
                return;
            }
        }

        EventRequestManager em = vm.eventRequestManager();
        ClassPrepareRequest cpr = em.createClassPrepareRequest();
        cpr.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        cpr.enable();

        ClassUnloadRequest cur = em.createClassUnloadRequest();
        cur.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        cur.enable();

        ThreadStartRequest tsr = em.createThreadStartRequest();
        tsr.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        tsr.enable();

        ThreadDeathRequest tdr = em.createThreadDeathRequest();
        tdr.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        tdr.enable();

        ExceptionRequest er = em.createExceptionRequest(null, false, true);
        er.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        er.enable();

        vm.resume();
    }

    public void setOptions(Options options) {
        this.options = options;
        if (options.isSet("sourcepath")) {
            use(options.getOpt("sourcepath"));
        }
    }

    public void log(Object o) {
        if (!getOptions().isSet("nologging")) {
            app.log(o);
        }
    }

    public Options getOptions() {
        return options;
    }

    /**
     * Resumes all the threads until the wait count
     * is back to zero.
     */
    public void resumeUntilZero() throws DebuggerException {
        resumeUntilZero(getVM().allThreads(), 1);
        getVM().resume();
    }

    /**
     * Resumes all the threads.
     */
    public void resume() throws DebuggerException {
        resume(getVM().allThreads());
    }

    /**
     * Resumes the list of threads until their
     * wait count is down to 0.
     *
     * @param threads List of threads to resume.
     */
    public void resumeUntilZero(List threads) throws DebuggerException {
        resumeUntilZero(threads, 0);
    }

    /**
     * Resumes all the threads until their wait count is
     * <code>num</code>.
     *
     * @param threads Threads to resume.
     * @param num     bottom limit on the wait count.
     */
    public void resumeUntilZero(List threads, int num) throws DebuggerException {
        Iterator iter = threads.iterator();
        while (iter.hasNext()) {
            ThreadReference threadRef = (ThreadReference) iter.next();
            resumeUntilZero(threadRef, num);
        }

    }

    /**
     * Resumes the list of threads.
     *
     * @param threads Threads to resume.
     */
    public void resume(List threads) {
        Iterator iter = threads.iterator();
        while (iter.hasNext()) {
            resume((ThreadReference) iter.next());
        }
    }

    /**
     * Resumes single thread until wait count is 0.
     *
     * @param threadRef Thread to resume to 0.
     */
    public void resumeUntilZero(ThreadReference threadRef) {
        resumeUntilZero(threadRef, 0);
    }

    /**
     * Resumes single thread to <code>num</code>.
     *
     * @param threadRef Thread to resume.
     * @param num       bottom limit wait count.
     */
    public void resumeUntilZero(ThreadReference threadRef, int num) {
        int n = threadRef.suspendCount();
        while ((n--) > num) {
            threadRef.resume();
        }
    }

    /**
     * Resumes single thread.
     */
    public void resume(ThreadReference threadRef) {
        threadRef.resume();
    }

    public Evaluator getEvaluator() {
        return evaluator;
    }

    /**
     * Gets the value for <code>o</code> in the default frame.
     *
     * @param o Object whose Value we're after.
     * @return  Value of <code>o</code>.
     */
    public Value getValue(Object o)
        throws NoVMException, DebuggerException {
        return getValue(o, getDefaultFrame());
    }

    /**
     * Gets the value for <code>o</code> in <code>frame</code>
     *
     * @param o     Object whose Value we're after.
     * @param frame StackFrame in which to for the value.
     * @return      Value of <code>o</code>.
     */
    public Value getValue(Object o, StackFrame frame)
        throws NoVMException, DebuggerException {
        return getEvaluator().getValue(o, frame);
    }

    /**
     * Returns a RefernceType for the given <code>className</code>
     * in the current VM.  This will be null if there is
     * no running VM, or the className is shit.
     *
     * @param className name of the class.
     * @return          ReferenceType for <code>className</code> in
     *                  the running VM.
     */
    public ReferenceType getReferenceTypeFromToken(String className)
        throws NoVMException {
        ReferenceType refType = null;
        Iterator iter = getVM().allClasses().iterator();
        while (iter.hasNext()) {
            refType = (ReferenceType) iter.next();
            if (refType.name().equals(className)) {
                return refType;
            }
        }
        return null;
    }

    /**
     * Returns a ThreadReference for the given <code>threadName</code>
     * in the current VM.  This will be null if there is
     * no running VM, or the threadName is shit.  If the passed
     * in value is the empty string the default thread is returned.
     *
     * @param threadName name of the thread
     * @return           ThreadReference for <code>className</code> in
     *                   the running VM.
     */
    public ThreadReference getThread(String threadName)
        throws DebuggerException {
        if (threadName.equals("")) {
            return getDefaultThread();
        }
        Iterator iter = vm().allThreads().iterator();
        while (iter.hasNext()) {
            ThreadReference threadRef = (ThreadReference) iter.next();
            if (threadRef.name().equals(threadName)) {
                return threadRef;
            } else {
                try {
                    long ourID = Long.parseLong(threadName);
                    long refID = threadRef.uniqueID();
                    if (ourID == refID) {
                        return threadRef;
                    }
                } catch (NumberFormatException e) {
                }
            }
        }
        throw new ThreadNotFoundException(threadName);
    }

    public void outln(Object o) {
        app.outln(o);
    }

    public DebuggerApp app() {
        return app;
    }

    public String getRunningClass() {
        return runningClass;
    }

    /**
     * Returns a ReferenceType for <code>className</code>.
     *
     * @param className name of the class.
     * @return          ReferenceType for <code>className</code>, the
     *                  returned value will not be null, but
     *                  an exception may be thrown if things don't
     *                  work out all that well.
     */
    public ReferenceType classByName(String className)
        throws AmbiguousClassException,
               InvalidClassException {
        try {
            Iterator iter = importPaths.iterator();
            while (iter.hasNext()) {
                String fullName = iter.next() + "";
                if (!fullName.equals("")) {
                    fullName += ".";
                }
                fullName += className;
                List classes = getVM().classesByName(fullName);
                if (classes.size() > 1) {
                    throw new AmbiguousClassException(className, classes);
                } else if (classes.size() == 1) {
                    ReferenceType refType = (ReferenceType) classes.get(0);
                    return (ReferenceType) refType;
                }
            }
        } catch (NoVMException nvme) {
        }
        throw new InvalidClassException(className);
    }

    public void addImportPath(String importPath) {
        importPaths.add(importPath);
        filePaths.add(fileFromPath(importPath));
    }

    private String fileFromPath(String importPath) {
        return importPath.replace('.', File.separatorChar);
    }

    public boolean removeImportPath(String importPath) {
        filePaths.remove(fileFromPath(importPath));
        return importPaths.remove(importPath);
    }

    public boolean addRequest(Request ra) {
        if (!requestActions.contains(ra)) {
            requestActions.add(ra);
            fireRequestSetEvent(new RequestEvent(ra, true));
            return true;
        }
        return false;
    }

    public EventRequest toggleBreakpoint(String source, int line) {
        BreakpointRequestAction ba = null;
        try {
            ba = createSourceLineEvent(source, line, true);
        } catch (DebuggerException de) {
            return null;
        }
        if (ba == null) return null;
        try {
            if (!requestActions.contains(ba)) {
                return stopOnCommand(source, line);
            } else {
                return clearOnCommand(source, line);
            }
        } catch (DebuggerException de) {}
        return null;
    }

    public RequestAction removeRequest(Request ra) {
        int i = 0;
        RequestAction req = null;
        while ((i = requestActions.indexOf(ra)) > -1) {
            if (i != -1) {
                req = (RequestAction)requestActions.remove(i);
                SourceManager.SourceLine sl = null;
                if (req != null) {
                    req.removeFromDebugger(this);
                    sl = req.getSourceLine();
                    fireRequestClearEvent(new RequestEvent(req, true));
                }
                if (sl != null) {
                    sl.clear();
                }
                EventRequest er = req.getRequest();
                if (er != null && vm() != null) {
                    vm().eventRequestManager().deleteEventRequest(er);
                }
            }
        }
        return req;
    }

    public List getRequests() {
        return requestActions;
    }

    private void wakeUpRequests() {
        Iterator iter = getRequests().iterator();
        try {
            while (iter.hasNext()) {
                RequestAction request = (RequestAction)iter.next();
                if (request.isEnabled()) request.go();
            }
        } catch (DebuggerException de) {
        }
    }

    public String help() {
        String s = "** command list **\n";
        s += "run [class [args]]        -- start execution of application's main class\n";
        s += "\n";
        s += "threads [threadgroup]     -- list threads\n";
        s += "thread <thread id>        -- set default thread\n";
        s += "suspend [thread id(s)]    -- suspend threads (default: all)\n";
        s += "resume [thread id(s)]     -- resume threads (default: all)\n";
        s += "where [thread id] | all   -- dump a thread's stack\n";
        s += "wherei [thread id] | all  -- dump a thread's stack, with pc info\n";
        s += "up [n frames]             -- move up a thread's stack\n";
        s += "down [n frames]           -- move down a thread's stack\n";
        s += "kill <thread> <expr>      -- kill a thread with the given exception object\n";
        s += "interrupt <thread>        -- interrupt a thread\n";
        s += "\n";
        s += "print <expr>              -- print value of expression\n";
        s += "dump <expr>               -- print all object information\n";
        s += "eval <expr>               -- evaluate expression (same as print)\n";
        s += "set <lvalue> = <expr>     -- assign new value to field/variable/array element\n";
        s += "locals                    -- print all local variables in current stack frame\n";
        s += "\n";
        s += "classes                   -- list currently known classes\n";
        s += "class <class id>          -- show details of named class\n";
        s += "methods <class id>        -- list a class's methods\n";
        s += "fields <class id>         -- list a class's fields\n";
        s += "\n";
        s += "threadgroups              -- list threadgroups\n";
        s += "threadgroup <name>        -- set current threadgroup\n";
        s += "\n";
        s += "stop in <class id>.<method>[(argument_type,...)]\n";
        s += "                          -- set a breakpoint in a method\n";
        s += "stop at <class id>:<line> -- set a breakpoint at a line\n";
        s += "clear   <class id>.<method>[(argument_type,...)]\n";
        s += "                          -- clear a breakpoint in a method\n";
        s += "clear <class id>:<line>   -- clear a breakpoint at a line\n";
        s += "clear                     -- list breakpoints\n";
        s += "catch <class id>          -- break when specified exception thrown\n";
        s += "ignore <class id>         -- cancel 'catch'  for the specified exception\n";
        s += "watch [access|all] <class id>.<field name>\n";
        s += "                          -- watch access/modifications to a field\n";
        s += "unwatch [access|all] <class id>.<field name>\n";
        s += "                          -- discontinue watching access/modifications to a field\n";
        s += "trace methods [thread]    -- trace method entry and exit\n";
        s += "untrace methods [thread]  -- stop tracing method entry and exit\n";
        s += "step                      -- execute current line\n";
        s += "step up                   -- execute until the current method returns to its caller\n";
        s += "stepi                     -- execute current instruction\n";
        s += "next                      -- step one line (step OVER calls)\n";
        s += "cont                      -- continue execution from breakpoint\n";
        s += "\n";
        s += "list [line number|method] -- print source code\n";
        s += "use (or sourcepath) [source file path]\n";
        s += "                          -- display or change the source path\n";
        s += "exclude [class id ... | \"none\"]\n";
        s += "                          -- do not report step or method events for specified classes\n";
        s += "classpath                 -- print classpath info from target VM\n";
        s += "\n";
        s += "monitor <command>         -- execute command each time the program stops\n";
        s += "monitor                   -- list monitors\n";
        s += "unmonitor <monitor#>      -- delete a monitor\n";
        s += "read <filename>           -- read and execute a command file\n";
        s += "\n";
        s += "lock <expr>               -- print lock info for an object\n";
        s += "threadlocks [thread id]   -- print lock info for a thread\n";
        s += "\n";
        s += "disablegc <expr>          -- prevent garbage collection of an object\n";
        s += "enablegc <expr>           -- permit garbage collection of an object\n";
        s += "\n";
        s += "!!                        -- repeat last command\n";
        s += "<n> <command>             -- repeat command n times\n";
        s += "help (or ?)               -- list commands\n";
        s += "version                   -- print version information\n";
        s += "exit (or quit)            -- exit debugger\n";
        s += "workingdir <dir>          -- set the user's workingdir to <dir>\n";
        s += "\n";
        s += "<class id>: full class name with package qualifiers or a\n";
        s += "pattern with a leading or trailing wildcard ('*').\n";
        s += "<thread id>: thread number as reported in the 'threads' command\n";
        s += "<expr>: a Java(tm) Programming Language expression.\n";
        s += "Most common syntax is supported.\n";
        s += "\n";
        s += "Startup commands can be placed in either \"ajdb.ini\" or \".ajdbrc\"\n";
        s += "in user.home or user.dir\n";
        return s;
    }

    /****************************** Empty Mapping ******************************/

    public String sourceName(Location loc) throws AbsentInformationException {
        return loc.sourceName();
    }

    public int lineNumber(Location loc) {
        return loc.lineNumber();
    }

    public long codeIndex(Location loc) {
        return loc.codeIndex();
    }

    public String method(Location loc) {
        return loc.method() + "";
    }

    public String name(Method method) {
        return method.name();
    }

    /****************************** Misc. ******************************/

    public void loadMainClass(Options options) {
        //getVM(options.getJavaArgs(), options.getFullCommandLine());
        getVM(options.getJavaArgs(), options.getClassName(), options.getCommandLine());
    }

    public List getRequest(RequestComparator rc) {
        Vector v = new Vector();
        Iterator iter = requestActions.iterator();
        while (iter.hasNext()) {
            Object o = iter.next();
            if (rc.wantsRequest(o)) {
                v.add(o);
            }
        }
        return v;
    }

    interface RequestComparator {
        public boolean wantsRequest(Object o);
    }

    public List getBreakpointRequests() {
        return getRequest(new RequestComparator() {
                public boolean wantsRequest(Object o) {
                    return
                        (o instanceof StopOnRequest) ||
                        (o instanceof StopAtRequest) ||
                        (o instanceof StopInRequest);
            }});
    }

    public List getAccessWatchpointRequests() {
        return getRequest(new RequestComparator() {
                public boolean wantsRequest(Object o) {
                    return (o instanceof WatchAccessRequest);
            }});
    }

    public List getModificationWatchpointRequests() {
        return getRequest(new RequestComparator() {
                public boolean wantsRequest(Object o) {
                    return (o instanceof WatchAllRequest);
            }});
    }

    public List getExceptionRequests() {
        return getRequest(new RequestComparator() {
                public boolean wantsRequest(Object o) {
                    return (o instanceof CatchRequest);
            }});
    }

    public void addMonitorRequest(MonitorRequest request) {
        if (!monitorList.contains(request)) {
            monitorList.add(request);
            addStopListener(request);
        }
    }

    public boolean removeMonitorRequest(MonitorRequest request) {
        removeStopListener(request);
        return monitorList.remove(request);
    }

    public MonitorRequest removeMonitorRequest(int number) {
        MonitorRequest request = null;
        Iterator iter = monitorList.iterator();
        while (iter.hasNext()) {
            request = (MonitorRequest) iter.next();
            if (request.getNumber() == number) {
                monitorList.remove(request);
                break;
            }
            request = null;
        }
        if (request != null) {
            removeStopListener(request);
        }
        return request;
    }

    public List getMonitorRequests() {
        return monitorList;
    }

    public List getImportPaths() {
        return importPaths;
    }

    public List getFilePaths() {
        return filePaths;
    }

    public void addExclude(String className) {
        if (!excludes.contains(className)) {
            excludes.add(className);
        }
    }

    public void addExcludesTo(StepRequest request) {
        Iterator iter = excludes.iterator();
        while (iter.hasNext()) {
            request.addClassExclusionFilter(iter.next() + "");
        }
    }

    public void addExcludesTo(MethodEntryRequest request) {
        Iterator iter = excludes.iterator();
        while (iter.hasNext()) {
            request.addClassExclusionFilter(iter.next() + "");
        }
    }

    public void addExcludesTo(MethodExitRequest request) {
        Iterator iter = excludes.iterator();
        while (iter.hasNext()) {
            request.addClassExclusionFilter(iter.next() + "");
        }
    }

    public void clearExcludes() {
        excludes.clear();
    }

    public void shutDown() {
        spinner.shutDown();
        try {
            // This may stop some of the memory exceptions being
            // thrown under windoze
            try {
                getVM().resume();
            } catch (Throwable t) {
            }
            try {
                getVM().exit(0);
            } catch (VMDisconnectedException vmde) {
            }
            try {
                getVM().dispose();
            } catch (VMDisconnectedException vmde) {
            }
        } catch (NoVMException nvme) {
        } finally {
            vm = null;
        }
    }

    public void exit() {
        exit(true);
    }

    public void exit(boolean exit) {
        shutDown();
        if (exit && app.wantsToExit()) {
            System.exit(0);
        }
    }

    public void quit() {
        //shutDown();
        exit();
        //System.exit(0);
    }

    public String version() {
        return version + (vm() != null
                            ? "\n" + vm().description()
                            : "");
    }

    public void setStatus(Object o) {
        status = o + "";
    }

    public String getStatus() {
        return status;
    }

    public ThreadReference getDefaultThread() throws NoVMException {
        return threads.getThread();
    }

    public ThreadGroupReference getDefaultThreadGroup() throws NoVMException {
        return threads.getThreadGroup();
    }

    public StackFrame getDefaultFrame() throws NoVMException {
        return threads.getStackFrame();
    }

    public ObjectReference getDefaultThisObject() throws NoVMException {
        return getDefaultFrame().thisObject();
    }

    public int getFrame() {
        return threads.getFrame();
    }

    public int getCurrentStoppedLine() {
        return currentStoppedLine;
    }

    public String getCurrentStoppedSource() {
        return currentStoppedSource;
    }

    public boolean isStopped(SourceManager.SourceLine sl) {
        return
            sl.getSourceName().equals(getCurrentStoppedSource()) &&
            sl.getLineNumber() == getCurrentStoppedLine();
    }

    protected boolean isSame(ThreadReference threadRef, String threadName) {
        if (threadRef.name().equals(threadName)) {
            return true;
        } else {
            try {
                long ourID = Long.parseLong(threadName);
                long refID = threadRef.uniqueID();
                if (ourID == refID) {
                    return true;
                }
            } catch (NumberFormatException e) {
            }
        }
        return false;
    }

    public ThreadReference getStoppedThread() throws NoVMException {
        return threads.getStoppedThread();
    }

    public String getPrompt() {
        return prompt;
    }

    private void setPrompt() {
        try {
            setPrompt(getDefaultThread(), getFrame());
        } catch (NoVMException nvme) {
        }
    }

    private void setPrompt(ThreadReference threadRef, int frame) {
        this.prompt = threadRef.name() + "[" + (frame+1) + "]";
        fireResetPrompt();
    }

    private void resetPrompt(boolean isRunning) {
        this.isRunning = isRunning;
        this.prompt = ">";
    }

    public boolean isRunning() {
        return isRunning && (vm != null);
    }

    public void setRunning(boolean isRunning) {
        this.isRunning = isRunning;
    }

    public boolean isAtBreakpoint() {
        try {
            return getDefaultThread().isAtBreakpoint();
        } catch (Exception e) {
        }
        return false;
    }

    public boolean isExtra() {
        return getOptions().isSet("extra");
    }

    public boolean isStopped() {
        return isStopped;
    }

    private void stop() {
        isStopped = true;
    }

    private void unstop() {
        isStopped = false;
    }

    public ThreadReference setDefaultThread(ThreadReference threadRef) {
        return threads.setThread(threadRef);
    }

    public ThreadGroupReference setDefaultThreadGroup(ThreadGroupReference threadGroupRef) {
        return threads.setThreadGroup(threadGroupRef);
    }

    public int setDefaultFrame(int frame) {
        return threads.setFrame(frame);
    }

    public int setFrame(int frame) {
        return threads.setFrame(frame);
    }

    public StackFrame up(int frames) {
        setFrame(getFrame() + frames);
        try {
            return getDefaultFrame();
        } catch (NoVMException e) {
        }
        return null;
    }

    public StackFrame down(int frames) {
        setFrame(getFrame() - frames);
        try {
            return getDefaultFrame();
        } catch (NoVMException e) {
        }
        return null;
    }

    public void use(String sourcePath) {
        sourceManager.setSourcePath(sourcePath);
        fireSourcePathChanged();
    }

    public SourceManager getSourceManager() {
        return sourceManager;
    }

    public String getSourcePath() {
        return sourceManager.getSourcePath();
    }

    public void resumeAllThreads() throws NoVMException {
        Iterator iter = getVM().allThreads().iterator();
        while (iter.hasNext()) {
            ((ThreadReference) iter.next()).resume();
        }
    }

    private ThreadReference setStoppedThread(ThreadReference threadRef) {
        return threads.setStoppedThread(threadRef);
    }

    private void setDefaults(LocatableEvent le) {
        stop();
        ThreadReference threadRef = le.thread();
        setDefaultThread(threadRef);
        setStoppedThread(threadRef);
        setDefaultThreadGroup(threadRef.threadGroup());
        setDefaultFrame(0);
        setStopped(le);
    }

    private void setStopped(LocatableEvent le) {
        currentStoppedLine = lineNumber(le.location());
        try {
            currentStoppedSource = sourceName(le.location());
        } catch (AbsentInformationException aie) {
        }
    }

    public void fatal(String str) {
        System.err.println("FATAL ERROR:");
        System.err.println(str);
        System.exit(1);
    }

    public void error(String str) {
        System.err.println(str);
    }

    public synchronized VirtualMachine getVM() throws NoVMException {
        if (vm == null) {
            throw new NoVMException();
        } else {
            return vm;
        }
    }

    public synchronized VirtualMachine vm() {
        return vm;
    }

    private VirtualMachine findVM(Options options) {
        //return findVM(options.getJavaArgs(), options.getCommandLine());
        return findVM(options.getJavaArgs(), options.getClassName(), options.getCommandLine());
    }

    private VirtualMachine findVM(String vmArgs, String className, String commandLine) {
        return findLaunchingVM(vmArgs, className, commandLine);
    }

    private VirtualMachine findLaunchingVM(String vmArgs, String className, String commandLine) {
        return findLaunchingVM(vmArgs, className, commandLine, true);
    }

    private VirtualMachine findLaunchingVM(String vmArgs, String className, String commandLine, boolean recurse) {
        Debug.debug("------------- findLaunchingVM -----------");
        Debug.debug("commandLine=" + commandLine);
        Debug.debug("vmArgs=" + vmArgs);
        if (vmArgs == null) vmArgs = "";
        if (commandLine == null) commandLine = "";
        final LaunchingConnector con = (LaunchingConnector) findConnector("");
        final Map args = con.defaultArguments();
        //((Connector.Argument) args.get("main")).setValue(commandLine);
        ((Connector.Argument) args.get("main")).setValue((className + " " + commandLine).trim());
        ((Connector.Argument) args.get("options")).setValue(vmArgs);
        Debug.debug("args=" + args);
        Debug.debug("con=" + con);
        Debug.debug("--------------------------------------------------");
        vm = null;
        try {
            Debug.debug("before: vm=" + vm);
            //if (true) return vm;
            if (false) throw new Exception("foo");
            vm = con.launch(args);
            Debug.debug("after: vm=" + vm);
        } catch (VMDisconnectedException vmde) {
            app.handleInternalException(vmde);
        } catch (VMStartException vmse) {
            if (!recurse) {
                app.handleInternalException(vmse);
            }
        } catch (IllegalConnectorArgumentsException icae) {
            if (!recurse) {
                app.handleInternalException(icae);
            }
        } catch (Exception e) {
            if (!recurse) {
                app.handleInternalException(e);
            }
        }
        Debug.debug("Done vm=" + vm);

        if (vm == null) {
            if (recurse) {
                Debug.debug("Recursing...");
                String newCommandLine = findClassName(commandLine);
                if (newCommandLine == null) {
                    app.handleInternalException
                        (new Exception("No class specified on the command line: " + commandLine));
                } else {
                    app.outln("Cannot support the arguments: " + args);
                    return findLaunchingVM(vmArgs, className, newCommandLine, false);
                }
            } else {
                app.handleInternalException(new Exception("VM is null"));
            }
            return vm;
        }

//          Process process = vm.process();
//          if (process != null) {
//              redirect(process.getErrorStream(), "error-dumper");
//              redirect(process.getInputStream(), "input-dumper");
//          }
        return vm;
    }

    private Connector findConnector(String name) {
        Debug.debug(this, "Looking for connector " + name);
        Iterator iter = Bootstrap.virtualMachineManager().allConnectors().iterator();
        while (iter.hasNext()) {
            Connector con = (Connector) iter.next();
            if (con.name().equals(name)) {
                return con;
            }
        }
        return Bootstrap.virtualMachineManager().defaultConnector();
    }

    private String findClassName(String commandLine) {
        StringTokenizer tok = new StringTokenizer(commandLine);
        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            if (!tok.hasMoreTokens() && !token.startsWith("-")) {
                return token;
            }
        }
        return null;
    }

    private void redirect(final InputStream in, final String name) {
        new Thread(name) {
                public void run() {
                    dump(in);
                }
            }.start();
    }

    private void dump(InputStream stream) {
        PrintStream out = app.getOutputStream();
        BufferedReader in = new BufferedReader(new InputStreamReader(stream));
        String line;
        try {
            while ((line = in.readLine()) != null) {
                out.println(line);
            }
        } catch (IOException ioe) {
            //TODO
            ioe.printStackTrace(System.err);
        }
    }

    protected void readStartupFiles() {
        try {
            readCommand("ajdb.ini");
            readCommand(".ajdbrc");
        } catch (NoVMException nvme) {
        } catch (DebuggerException de) {
        }
    }


     /****************************** Execution *****************************/

     public Object execute(String line) {
         Object o = null;
         db("execute line=" + line);
         try {
             o = parser.parse(line);
         } catch (Exception e) {
             handle(e);
         }
        return o;
    }

    public void handle(Exception t) {
        if (t instanceof ParseException) {
            app.handleParseException((ParseException)t);
        } else if (t instanceof NoVMException) {
            app.handleNoVMException((NoVMException)t);
        } else if (t instanceof DebuggerException) {
            app.handleDebuggerException((DebuggerException)t);
        } else if (t instanceof VMDisconnectedException) {
            app.handleVMDisconnectedException((VMDisconnectedException)t);
        } else {
            app.handleInternalException(t);
        }
    }

    public Object run(String mainClass, String[] args)
        throws NoVMException,
               DebuggerException {
        String str = "";
        int N = args.length;
        for (int i = 0; i < N; i++) {
            str += args[i] + (i<N-1 ? " " : "");
        }
        return runCommand(mainClass, str);
    }

    public Object run(String mainClass)
        throws NoVMException,
               DebuggerException {
        return run(mainClass, new String[0]);
    }

    public Object executeCommand(String line) {
        return execute(line);
    }

    /************************** Creating breakpoints **********************/

    public ClassLineBreakpointRequestAction createClassLineEvent(String className,
                                                                 int line,
                                                                 boolean stop)
        throws DebuggerException {
        return createClassLineEvent(className, line, stop, null);
    }
    
    public ClassLineBreakpointRequestAction
        createClassLineEvent(String className,
                             int line,
                             boolean stop,
                             String text) 
        throws DebuggerException {
        ClassLineBreakpointRequestAction a = null;
        if (stop) {
            a = new StopAtRequest(this, className, line);
            a.setText(text);
        } else {
            a = new ClearAtRequest(this, className, line);
            a.setText(text);
        }
        return a;
    }

    public SourceLineBreakpointRequestAction
        createSourceLineEvent(String sourceName,
                              int line,
                              boolean stop)
        throws DebuggerException {
        return createSourceLineEvent(sourceName, line, stop, null);
    }

    public SourceLineBreakpointRequestAction
        createSourceLineEvent(String sourceName,
                              int line,
                              boolean stop,
                              String text)
        throws DebuggerException {
        SourceLineBreakpointRequestAction a = null;        
        if (stop) {
            a = new StopOnRequest(this, sourceName, line);
            a.setText(text);
        } else {
            a = new ClearOnRequest(this, sourceName, line);
            a.setText(text);
        }
        return a;
    }

    /****************************** Commands ******************************/

    public Object runCommand() {
        try {
            return runCommand(getOptions());
        } catch (DebuggerException de) {
        }
        return null;
    }

    public Object runCommand(Options opts)
        throws NoVMException, DebuggerException {
        String className = opts.getClassName();
        String javaArgs = opts.getJavaArgs();
        String commandLine = opts.getCommandLine();
        return runCommand(className, javaArgs, commandLine,
                          false, VirtualMachine.TRACE_NONE);
    }

    public Object runCommand(String className, String exArgs)
        throws NoVMException, DebuggerException {
        return runCommand(className, "", exArgs, false,
                          VirtualMachine.TRACE_NONE);
    }
    
    public Object runCommand(String className, String vmArgs, String exArgs,
                             boolean isSuspended, int debugTraceMode)
        throws NoVMException, DebuggerException {
        if (!getOptions().isSet("extra") &&
            isRunning() &&
            !app.canRestart()) {
            app.outln("VM already running. " +
                      "Use 'cont' to continue after events.");
            return null;
        }
        unstop();
        this.runningClass = className;
        if (!Modes.isJBuilder() && !Modes.isForte()) {
            app.outln("run " + className);
        } else {
            app.outln("connect " + className);
        }
        go(className, vmArgs, exArgs, isSuspended, debugTraceMode);
        return null;
    }

    public EventRequest clearAtCommand(String className, int line)
        throws NoVMException, DebuggerException {
        EventRequest er = (EventRequest)
            (createClassLineEvent(className, line, false).go());
        app.clearAtCommand(className, line, er);
        return er;
    }

    public EventRequest clearInCommand(String className, String methodProto)
        throws NoVMException, DebuggerException {
        EventRequest er =  (EventRequest)
            (new ClearInRequest(this, className, methodProto).go());
        app.clearInCommand(className, methodProto, er);
        return er;
    }

    public EventRequest clearOnCommand(String sourceName, int line)
        throws NoVMException, DebuggerException {
        EventRequest er = (EventRequest)
            (createSourceLineEvent(sourceName, line, false).go());
        app.clearOnCommand(sourceName, line, er);
        return er;
    }

    public List clearCommand()
        throws NoVMException, DebuggerException {
        List breakpoints = (List) (new ClearListRequest(this).go());
        app.clearCommand(breakpoints);
        return breakpoints;
    }

    public List clearAllCommand()
        throws NoVMException, DebuggerException {
        List breakpoints = (List) (new ClearAllRequest(this).go());
        app.clearAllCommand(breakpoints);
        return breakpoints;
    }

    public EventRequest stopAtCommand(String className, int line)
        throws NoVMException, DebuggerException {
        return stopAtCommand(className, line, null);
    }

    public EventRequest stopAtCommand(String className, int line, String text)
        throws NoVMException, DebuggerException {
        EventRequest er = (EventRequest)
            (createClassLineEvent(className, line, true, text).go());
        app.stopAtCommand(className, line, er);
        return er;
    }

    public EventRequest stopInCommand(String className, String methodProto)
        throws NoVMException, DebuggerException {
        return stopInCommand(className, methodProto, null);
    }

    public EventRequest stopInCommand(String className, String methodProto, String text)
        throws NoVMException, DebuggerException {
        EventRequest er =  (EventRequest)
            (new StopInRequest(this, className, methodProto).setText(text).go());
        app.stopInCommand(className, methodProto, er);
        return er;
    }

    public EventRequest stopOnCommand(String sourceName, int line) 
        throws NoVMException, DebuggerException {
        return stopOnCommand(sourceName, line, null);
    }

    public EventRequest stopOnCommand(String sourceName, int line, String text)
        throws NoVMException, DebuggerException {            
        EventRequest er =  (EventRequest)
            (createSourceLineEvent(sourceName, line, true, text).go());
        app.stopOnCommand(sourceName, line, er);
        return er;
    }

    public List stopCommand()
        throws NoVMException, DebuggerException {
        List breakpoints = (List) (new StopListRequest(this).go());
        app.stopCommand(breakpoints);
        return breakpoints;
    }

    public Object contCommand()
        throws NoVMException, DebuggerException {
        new ContRequest(this).go();
        fireStartEvent();
        app.contCommand();
        return null;
    }

    public List threadsCommand(String threadGroupName)
        throws NoVMException, DebuggerException {
        List threads = (List) (new ThreadsRequest(this, threadGroupName).go());
        app.threadsCommand(threadGroupName, threads);
        return threads;
    }

    public List threadGroupsCommand()
        throws NoVMException, DebuggerException {
        List threadGroups = (List) (new ThreadGroupsRequest(this).go());
        app.threadGroupsCommand(threadGroups);
        return threadGroups;
    }

    public List fieldsCommand(String className)
        throws NoVMException, DebuggerException {
        List fields = (List) (new FieldsRequest(this, className).go());
        app.fieldsCommand(className, fields);
        return fields;
    }

    public List methodsCommand(String className)
        throws NoVMException, DebuggerException {
        List methods = (List) (new MethodsRequest(this, className).go());
        app.methodsCommand(className, methods);
        return methods;
    }

    public List classesCommand()
        throws NoVMException, DebuggerException {
        List classes = (List) (new ClassesRequest(this).go());
        app.classesCommand(classes);
        return classes;
    }

    public ReferenceType classCommand(String className)
        throws NoVMException, DebuggerException {
        ReferenceType refType = (ReferenceType)
            (new ClassRequest(this, className).go());
        app.classCommand(className, refType);
        return refType;
    }

    public ThreadReference threadCommand(String threadName)
        throws NoVMException, DebuggerException {
        ThreadReference threadRef =
            setDefaultThread((ThreadReference)
                             (new ThreadRequest(this, threadName).go()));
        app.threadCommand(threadName, threadRef);
        return threadRef;
    }

    public ThreadGroupReference threadGroupCommand(String threadGroupName)
        throws NoVMException, DebuggerException {
        ThreadGroupReference threadGroupRef =
            setDefaultThreadGroup((ThreadGroupReference)
                                  (new ThreadGroupRequest
                                      (this, threadGroupName).go()));
        app.threadGroupCommand(threadGroupName, threadGroupRef);
        return threadGroupRef;
    }

    public List suspendCommand()
        throws NoVMException, DebuggerException {
        List threads = (List) suspendCommand(new Vector());
        fireStopEvent();
        return threads;
    }

    public List suspendCommand(List threadNames)
        throws NoVMException, DebuggerException {
        List threads = (List) (new SuspendRequest(this, threadNames).go());
        fireStopEvent();
        app.suspendCommand(threadNames, threads);
        return threads;
    }

    public List resumeCommand(List threadNames)
        throws NoVMException, DebuggerException {
        List threads = (List) (new ResumeRequest(this, threadNames).go());
        fireStartEvent();
        app.resumeCommand(threadNames, threads);
        return threads;
    }

    public List resumeCommand() 
        throws NoVMException, DebuggerException {
        List threads = resumeCommand(new Vector());
        fireStartEvent();
        return threads;
    }

    public List whereCommand(String threadName)
        throws NoVMException, DebuggerException {
        List frames = (List) (new WhereRequest(this, threadName).go());
        app.whereCommand(threadName, frames);
        return frames;
    }

    public List localsCommand()
        throws NoVMException, DebuggerException {
        List locals = (List) (new LocalsRequest(this).go());
        app.localsCommand(locals);
        return locals;
    }

    public ThreadReference interruptCommand(String threadName)
        throws NoVMException, DebuggerException {
        ThreadReference threadRef = (ThreadReference)
            (new InterruptRequest(this, threadName).go());
        fireStopEvent();
        app.interruptCommand(threadName, threadRef);
        return threadRef;
    }

    public StackFrame upCommand(int frames)
        throws NoVMException, DebuggerException {
        StackFrame frame = (StackFrame) (new UpRequest(this, frames).go());
        app.upCommand(frames, frame);
        return frame;

    }

    public StackFrame downCommand(int frames)
        throws NoVMException, DebuggerException {
        StackFrame frame = (StackFrame) (new DownRequest(this, frames).go());
        app.downCommand(frames, frame);
        return frame;
    }

    public StepRequest stepCommand()
        throws NoVMException, DebuggerException {
        unstop();
        StepRequest request = (StepRequest) (new StepIntoRequest(this).go());
        fireStartEvent();
        app.stepCommand(request);
        return request;
    }

    public StepRequest stepUpCommand()
        throws NoVMException, DebuggerException {
        unstop();
        StepRequest request = (StepRequest) (new StepUpRequest(this).go());
        fireStartEvent();
        app.stepUpCommand(request);
        return request;
    }

    public StepRequest stepiCommand()
        throws NoVMException, DebuggerException {
        unstop();
        StepRequest request = (StepRequest) (new StepIRequest(this).go());
        fireStartEvent();
        app.stepiCommand(request);
        return request;
    }

    public StepRequest nextCommand()
        throws NoVMException, DebuggerException {
        unstop();
        StepRequest request = (StepRequest) (new NextRequest(this).go());
        fireStartEvent();
        app.nextCommand(request);
        return request;
    }

    public Value printCommand(Object valueRep)
        throws NoVMException, DebuggerException {
        Value value = (Value) (new PrintRequest(this, valueRep).go());
        app.printCommand(valueRep, value);
        return value;
    }

    public Value dumpCommand(Object valueRep)
        throws NoVMException, DebuggerException {
        Value value = (Value) (new DumpRequest(this, valueRep).go());
        app.dumpCommand(valueRep, value);
        return value;
    }

    public Value evalCommand(Object valueRep)
        throws NoVMException, DebuggerException {
        Value value = (Value) (new EvalRequest(this, valueRep).go());
        app.evalCommand(valueRep, value);
        return value;
    }

    public Value setCommand(Object lvalue, Object rvalue)
        throws NoVMException, DebuggerException {
        Value oldValue = (Value) (new DumpRequest(this, lvalue).go());
        Value newValue = (Value) (new SetRequest(this, lvalue, rvalue).go());
        app.setCommand(lvalue, rvalue, oldValue, newValue);
        return newValue;
    }

    public ClasspathRequest.Package classpathCommand()
        throws NoVMException, DebuggerException {
        ClasspathRequest.Package pkg = (ClasspathRequest.Package)
            (new ClasspathRequest(this).go());
        app.classpathCommand(pkg.baseDirectory, pkg.paths);
        return pkg;
    }

    public LockInformation lockCommand(Object valueRep)
        throws NoVMException, DebuggerException {
        LockInformation lockInfo = (LockInformation)
            (new LockRequest(this, valueRep).go());
        app.lockCommand(valueRep, lockInfo);
        return lockInfo;
    }

    public ThreadLockInformation threadlocksCommand(String threadName)
        throws NoVMException, DebuggerException {
        ThreadLockInformation threadLockInfo =
            (ThreadLockInformation)
            (new ThreadLocksRequest(this, threadName).go());
        app.threadlocksCommand(threadName, threadLockInfo);
        return threadLockInfo;
    }

    public WatchpointRequest watchAccessCommand(String className,
                                                String fieldName)
        throws NoVMException, DebuggerException {
        return watchAccessCommand(className, fieldName, null);
    }

    public WatchpointRequest watchAccessCommand(String className,
                                                String fieldName,
                                                String text)
        throws NoVMException, DebuggerException {
        WatchpointRequest request =
            (WatchpointRequest) (new WatchAccessRequest(this, className, fieldName).setText(text).go());
        app.watchAccessCommand(className, fieldName, request);
        return request;
    }

    public WatchpointRequest watchAllCommand(String className,
                                             String fieldName)
        throws NoVMException, DebuggerException {
        return watchAllCommand(className, fieldName, null);
    }

    public WatchpointRequest watchAllCommand(String className,
                                             String fieldName,
                                             String text)
        throws NoVMException, DebuggerException {
        WatchpointRequest request =
            (WatchpointRequest)
            (new WatchAllRequest(this, className, fieldName).setText(text).go());
        app.watchAllCommand(className, fieldName, request);
        return request;
    }

    public WatchpointRequest unwatchAccessCommand(String className, String fieldName)
        throws NoVMException, DebuggerException {
        WatchpointRequest request =
            (WatchpointRequest)
            (new UnwatchAccessRequest(this, className, fieldName).go());
        app.unwatchAccessCommand(className, fieldName, request);
        return request;
    }

    public WatchpointRequest unwatchAllCommand(String className, String fieldName)
        throws NoVMException, DebuggerException {
        WatchpointRequest request =
            (WatchpointRequest)
            (new UnwatchAllRequest(this, className, fieldName).go());
        app.unwatchAllCommand(className, fieldName, request);
        return request;
    }

    public ExceptionRequest catchCommand(String className)
        throws NoVMException, DebuggerException {
        return catchCommand(className, null);
    }

    public ExceptionRequest catchCommand(String className, String text)
        throws NoVMException, DebuggerException {        
        ExceptionRequest request = (ExceptionRequest)
            (new CatchRequest(this, className).setText(text).go());
        app.catchCommand(className, request);
        return request;
    }

    public ExceptionRequest ignoreCommand(String className)
        throws NoVMException, DebuggerException {
        ExceptionRequest request = (ExceptionRequest)
            (new IgnoreRequest(this, className).go());
        app.ignoreCommand(className, request);
        return request;
    }

    public TraceMethodsRequest.EntryExitPair
        traceMethodsCommand(String threadName)
        throws NoVMException, DebuggerException {
        TraceMethodsRequest.EntryExitPair pair =
            (TraceMethodsRequest.EntryExitPair)
            (new TraceMethodsRequest(this, threadName).go());
        app.traceMethodsCommand(threadName, pair);
        return pair;
    }

    public UntraceMethodsRequest.EntryExitPair
        untraceMethodsCommand(String threadName)
        throws NoVMException, DebuggerException {
        UntraceMethodsRequest.EntryExitPair pair =
            (UntraceMethodsRequest.EntryExitPair)
            (new UntraceMethodsRequest(this, threadName).go());
        app.untraceMethodsCommand(threadName, pair);
        return pair;
    }

    public List excludeCommand(List classNames)
        throws NoVMException, DebuggerException {
        List classes = (List) (new ExcludeRequest(this, classNames).go());
        app.excludeCommand(classNames, classes);
        return classes;
    }

    public Object quitCommand()
        throws NoVMException, DebuggerException {
        fireStopEvent();
        app.quitCommand();
        return (new QuitRequest(this).go());
    }

    public Object helpCommand()
        throws NoVMException, DebuggerException {
        String helpString = (String) (new HelpRequest(this).go());
        app.helpCommand(helpString);
        return helpString;
    }

    public Object versionCommand()
        throws NoVMException, DebuggerException {
        Object version = (new VersionRequest(this).go());
        app.versionCommand(version);
        return version;
    }

    public ThreadReference killCommand(String threadName,
                                       String valueRep)
        throws NoVMException, DebuggerException {
        ThreadReference threadRef = (ThreadReference)
            (new KillRequest(this, threadName, valueRep).go());
        app.killCommand(threadName, valueRep, threadRef);
        return threadRef;
    }

    public String useCommand(String sourcePath)
        throws NoVMException, DebuggerException {
        String newSourcePath = (String)
            (new UseRequest(this, sourcePath).go());
        app.useCommand(sourcePath, newSourcePath);
        return newSourcePath;
    }

    public String workingdirCommand(String workingdirPath)
        throws NoVMException, DebuggerException {
        String newWorkingdirPath = (String)
            (new WorkingdirRequest(this, workingdirPath).go());
        app.workingdirCommand(workingdirPath, newWorkingdirPath);
        return newWorkingdirPath;
    }    

    public List listCommand()
        throws NoVMException, DebuggerException {
        ListRequest.SourceAndLines sal =
            (ListRequest.SourceAndLines) (new ListRequest(this).go());
        app.listCommand(sal.getSourceName(), sal.getLines());
        return sal.getLines();
    }

    public List listCommand(String sourceName)
        throws NoVMException, DebuggerException {
        ListRequest.SourceAndLines sal =
            (ListRequest.SourceAndLines)
            (new ListRequest(this, sourceName).go());
        app.listCommand(sal.getSourceName(), sal.getLines());
        return sal.getLines();
    }

    public SourceManager.SourceLine listCommand(String sourceName, int lineNumber)
        throws NoVMException, DebuggerException {
        ListRequest.SourceAndLines sal =
            (ListRequest.SourceAndLines)
            (new ListRequest(this, sourceName, lineNumber).go());
        SourceManager.SourceLine sl = (SourceManager.SourceLine)
            sal.getLines().get(0);
        app.listCommand(sal.getSourceName(), sl.lineNumber, sl);
        return sl;
    }

    public List listCommand(String sourceName, int startLine, int endLine)
        throws NoVMException, DebuggerException {
        ListRequest.SourceAndLines sal =
            (ListRequest.SourceAndLines)
            (new ListRequest(this, sourceName, startLine, endLine).go());
        app.listCommand(sal.getSourceName(), startLine, endLine, sal.getLines());
        return sal.getLines();
    }

    public String monitorCommand(String command)
        throws NoVMException, DebuggerException {
        String result = (String) (new MonitorRequest(this, command,
                                                     (monitors++)).go());
        app.monitorCommand(command, result);
        return result;
    }

    public List monitorCommand()
        throws NoVMException, DebuggerException {
        List monitors = (List) (new MonitorListRequest(this).go());
        app.monitorCommand(monitors);
        return monitors;
    }

    public MonitorRequest unmonitorCommand(int number)
        throws NoVMException, DebuggerException {
        MonitorRequest request = (MonitorRequest)
            (new UnmonitorRequest(this, number).go());
        app.unmonitorCommand(number, request);
        return request;
    }

    public List readCommand(String sourceName)
        throws NoVMException, DebuggerException {
        return readCommand(sourceName, false);
    }

    public List readCommand(String sourceName, boolean silent)
        throws NoVMException, DebuggerException {
        List commands = (List) (new ReadRequest(this, sourceName, silent).go());
        app.readCommand(sourceName, commands);
        return commands;
    }

    public VirtualMachine connectCommand()
        throws NoVMException, DebuggerException {
        return connectCommand("", "", "");
    }
    
    public VirtualMachine connectCommand(String vmArgs,
                                         String className,
                                         String commandLine)
        throws NoVMException, DebuggerException {
        try {
            startUpVM((VirtualMachine)
                      (new ConnectRequest(this, vmArgs, className,
                                          commandLine).go()));
        } catch (ClassNotFoundException cnfe) {
            handleClassNotFound(cnfe);
            return null;
        } catch (NoClassDefFoundError ncdfe) {
            handleClassNotFound(ncdfe);
            return null;
        }
        app.connectCommand(vm);
        return vm;
    }

    //TODO: Implement the 'app' commands

    public Object bangBangCommand(String lastCommand)
        throws NoVMException, DebuggerException {
        Object result = (new BangBangRequest(this, lastCommand).go());
        return result;
    }

    public Object importCommand(String importPath)
        throws NoVMException, DebuggerException {
        Object path = (new ImportRequest(this, importPath).go());
        return path;
    }

    public Object deportCommand(String importPath)
        throws NoVMException, DebuggerException {
        Object path = (new DeportRequest(this, importPath).go());
        return path;
    }

    public List viewCommand(String source)
        throws NoVMException, DebuggerException {
        ListRequest.SourceAndLines sal =
            (ListRequest.SourceAndLines) (new ViewRequest(this, source).go());
        app.listCommand(sal.getSourceName(), sal.getLines());
        return sal.getLines();
    }

    public String tostringCommand()
        throws NoVMException, DebuggerException {
        String tostring = (String) (new TostringRequest(this).go());
        app.tostringCommand(tostring);
        return tostring;
    }

    public File pwdCommand()
        throws NoVMException, DebuggerException {
        File pwd = (File)(new PwdRequest(this).go());
        app.pwdCommand(pwd);
        return pwd;
    }

    public List lsCommand(String fileName)
        throws NoVMException, DebuggerException {
        List files = (List)(new LsRequest(this, fileName).go());
        app.lsCommand(fileName, files);
        return files;
    }

    // end-commands


    protected VirtualMachine vm;
    public boolean handle(Event event) {
        if (event == null) return false;
        if (event instanceof AccessWatchpointEvent) {
            handleAccessWatchpointEvent((AccessWatchpointEvent) event);
            return true;
        } else if (event instanceof BreakpointEvent) {
            handleBreakpointEvent((BreakpointEvent) event);
            return false; ///true;
        } else if (event instanceof ClassPrepareEvent) {
            handleClassPrepareEvent((ClassPrepareEvent) event);
            return true;
        } else if (event instanceof ClassUnloadEvent) {
            handleClassUnloadEvent((ClassUnloadEvent) event);
            return true;
        } else if (event instanceof ExceptionEvent) {
            handleExceptionEvent((ExceptionEvent) event);
            return true;
        } else if (event instanceof MethodEntryEvent) {
            handleMethodEntryEvent((MethodEntryEvent) event);
            return true;
        } else if (event instanceof MethodExitEvent) {
            handleMethodExitEvent((MethodExitEvent) event);
            return true;
        } else if (event instanceof ModificationWatchpointEvent) {
            handleModificationWatchpointEvent((ModificationWatchpointEvent) event);
            return true;
        } else if (event instanceof StepEvent) {
            handleStepEvent((StepEvent) event);
            return false; //true;
        } else if (event instanceof ThreadDeathEvent) {
            handleThreadDeathEvent((ThreadDeathEvent) event);
            return true;
        } else if (event instanceof ThreadStartEvent) {
            handleThreadStartEvent((ThreadStartEvent) event);
            return true;
        } else if (event instanceof VMDeathEvent) {
            handleVMDeathEvent((VMDeathEvent) event);
            return false;
        } else if (event instanceof VMDisconnectEvent) {
            handleVMDisconnectEvent((VMDisconnectEvent) event);
            return false;
        } else if (event instanceof VMStartEvent) {
            handleVMStartEvent((VMStartEvent) event);
            return false;
        }
        return false;
    }

    public List locals() {
        try {
            StackFrame frame = getDefaultFrame();
            if (frame != null) {
                return frame.visibleVariables();
            }
        } catch (InvalidStackFrameException e0) {
        } catch (AbsentInformationException e1) {
            error(Messages.NO_LOCALS_MSG);
        } catch (NoVMException e2) {
        }
        return new Vector();
    }


    /* ------------------------------------------------------------
     * Events
     * ------------------------------------------------------------
     */

    /* VM listners */
    void handleVMStartEvent(VMStartEvent event) {
        resetPrompt(true);
        for (int i = 0; i < vmListeners.size(); i++) {
            ((VMListener) vmListeners.get(i)).
                vmStartEvent((VMStartEvent) event);
        }
    }

    void handleVMDeathEvent(VMDeathEvent event) {
        resetPrompt(false);
        for (int i = 0; i < vmListeners.size(); i++) {
            ((VMListener) vmListeners.get(i)).
                vmDeathEvent((VMDeathEvent) event);
        }
    }

    void handleVMDisconnectEvent(VMDisconnectEvent event) {
        resetPrompt(false);
        for (int i = 0; i < vmListeners.size(); i++) {
            ((VMListener) vmListeners.get(i)).
                vmDisconnectEvent((VMDisconnectEvent) event);
        }
    }


    /* Thread listeners */
    void handleThreadStartEvent(ThreadStartEvent event) {
        for (int i = 0; i < threadListeners.size(); i++) {
            ((ThreadListener) threadListeners.get(i)).
                threadStartEvent((ThreadStartEvent) event);
        }
    }

    void handleThreadDeathEvent(ThreadDeathEvent event) {
        for (int i = 0; i < threadListeners.size(); i++) {
            ((ThreadListener) threadListeners.get(i)).
                threadDeathEvent((ThreadDeathEvent) event);
        }
    }


    /* Class listeners */
    void handleClassPrepareEvent(ClassPrepareEvent event) {
        for (int i = 0; i < classListeners.size(); i++) {
            ((ClassListener) classListeners.get(i)).
                classPrepareEvent((ClassPrepareEvent) event);
        }
    }

    void handleClassUnloadEvent(ClassUnloadEvent event) {
        for (int i = 0; i < classListeners.size(); i++) {
            ((ClassListener) classListeners.get(i)).
                classUnloadEvent((ClassUnloadEvent) event);
        }
    }


    /* Stop listeners */
    void handleBreakpointEvent(BreakpointEvent event) {
        Thread.yield();
        setDefaults((LocatableEvent) event);
        for (int i = 0; i < stopListeners.size(); i++) {
            ((StopListener) stopListeners.get(i)).
                breakpointEvent((BreakpointEvent) event);
        }
        setPrompt();
        fireStopEvent();
    }

    void handleAccessWatchpointEvent(AccessWatchpointEvent event) {
        Thread.yield();
        setDefaults((LocatableEvent) event);
        for (int i = 0; i < stopListeners.size(); i++) {
            ((StopListener) stopListeners.get(i)).
                accessWatchpointEvent((AccessWatchpointEvent) event);
        }
        setPrompt();
        fireStopEvent();
    }

    void handleModificationWatchpointEvent(ModificationWatchpointEvent event) {
        Thread.yield();
        setDefaults((LocatableEvent) event);
        for (int i = 0; i < stopListeners.size(); i++) {
            ((StopListener) stopListeners.get(i)).
                modificationWatchpointEvent((ModificationWatchpointEvent) event);
        }
        setPrompt();
        fireStopEvent();
    }

    void handleStepEvent(StepEvent event) {
        Thread.yield();
        setDefaults((LocatableEvent) event);
        for (int i = 0; i < stopListeners.size(); i++) {
            ((StopListener) stopListeners.get(i)).
                stepEvent((StepEvent) event);
        }
        setPrompt();
        fireStopEvent();
    }

    void handleExceptionEvent(ExceptionEvent event) {
        setDefaults((LocatableEvent) event);
        for (int i = 0; i < stopListeners.size(); i++) {
            ((StopListener) stopListeners.get(i)).
                exceptionEvent((ExceptionEvent) event);
        }
        if (event.location() == null) {
            try {
                suspendCommand();
            } catch (DebuggerException de) {
            }
        }
        setPrompt();
        fireStopEvent();
    }

    /* Method listeners */
    void handleMethodEntryEvent(MethodEntryEvent event) {
        for (int i = 0; i < methodListeners.size(); i++) {
            ((MethodListener) methodListeners.get(i)).
                methodEntryEvent((MethodEntryEvent) event);
        }
    }

    void handleMethodExitEvent(MethodExitEvent event) {
        for (int i = 0; i < methodListeners.size(); i++) {
            ((MethodListener) methodListeners.get(i)).
                methodExitEvent((MethodExitEvent) event);
        }
    }

    /* Debugger listeners */
    public void fireRequestSetEvent(RequestEvent re) {
        for (int i = 0; i < debuggerListeners.size(); i++) {
            ((DebuggerListener) debuggerListeners.get(i)).
                requestSetEvent(re);
        }
    }

    public void fireRequestClearEvent(RequestEvent re) {
        for (int i = 0; i < debuggerListeners.size(); i++) {
            ((DebuggerListener) debuggerListeners.get(i)).
                requestClearEvent(re);
        }
    }

    public void fireRequestDeferredEvent(RequestEvent re) {
        for (int i = 0; i < debuggerListeners.size(); i++) {
            ((DebuggerListener) debuggerListeners.get(i)).
                requestDeferredEvent(re);
        }
    }

    public void fireRequestFailedEvent(RequestEvent re) {
        for (int i = 0; i < debuggerListeners.size(); i++) {
            ((DebuggerListener) debuggerListeners.get(i)).
                requestFailedEvent(re);
        }
    }

    /* Prompt listeners */
    public void fireResetPrompt() {
        for (int i = 0; i < promptListeners.size(); i++) {
            ((PromptListener) promptListeners.get(i)).
                resetPrompt();
        }
    }

    /* Source listeners */
    public void fireSourcePathChanged() {
        for (int i = 0; i < sourcePathListeners.size(); i++) {
            ((SourcePathListener) sourcePathListeners.get(i)).
                sourcePathChanged();
        }
    }

    /* Progress listeners */
    public void fireStartEvent() {
    if (vm() == null) return;
        for (int i = 0; i < progressListeners.size(); i++) {
            ((ProgressListener) progressListeners.get(i)).
                startEvent();
        }
    }

    public void fireStopEvent() {
        if (vm() == null) return;
        for (int i = 0; i < progressListeners.size(); i++) {
            ((ProgressListener) progressListeners.get(i)).
                stopEvent();
        }
    }       

    /****************************** Support ******************************/

    static class AppVector extends Vector {
        public boolean add(Object o) {
            if (o instanceof DebuggerApp) {
                add(0, o);
                return true;
            } else {
                return super.add(o);
            }
        }
    }

    protected Vector classListeners         = new AppVector();
    protected Vector stopListeners          = new AppVector();
    protected Vector methodListeners        = new AppVector();
    protected Vector threadListeners        = new AppVector();
    protected Vector vmListeners            = new AppVector();
    protected Vector debuggerListeners      = new AppVector();
    protected Vector promptListeners        = new AppVector();
    protected Vector sourcePathListeners    = new AppVector();
    protected Vector progressListeners      = new AppVector();

    /* Add listeners */
    public void addClassListener(ClassListener listener) {
        if (!classListeners.contains(listener)) {
            classListeners.add(listener);
        }
    }

    public void addStopListener(StopListener listener) {
        if (!stopListeners.contains(listener)) {
            stopListeners.add(listener);
        }
    }

    public void addMethodListener(MethodListener listener) {
        if (!methodListeners.contains(listener)) {
            methodListeners.add(listener);
        }
    }

    public void addThreadListener(ThreadListener listener) {
        if (!threadListeners.contains(listener)) {
            threadListeners.add(listener);
        }
    }

    public void addVMListener(VMListener listener) {
        if (!vmListeners.contains(listener)) {
            vmListeners.add(listener);
        }
    }

    public void addDebuggerListener(DebuggerListener listener) {
        if (!debuggerListeners.contains(listener)) {
            debuggerListeners.add(listener);
        }
    }

    public void addPromptListener(PromptListener listener) {
        if (!promptListeners.contains(listener)) {
            promptListeners.add(listener);
        }
    }

    public void addSourcePathListener(SourcePathListener listener) {
        if (!sourcePathListeners.contains(listener)) {
            sourcePathListeners.add(listener);
        }
    }

    public void addProgressListener(ProgressListener listener) {
        if (!progressListeners.contains(listener)) {
            progressListeners.add(listener);
        }
    }

    /* Remove listeners */
    public boolean removeClassListener(ClassListener listener) {
        return classListeners.remove(listener);
    }

    public boolean removeStopListener(StopListener listener) {
        return stopListeners.remove(listener);
    }

    public boolean removeMethodListener(MethodListener listener) {
        return methodListeners.remove(listener);
    }

    public boolean removeThreadListener(ThreadListener listener) {
        return threadListeners.remove(listener);
    }

    public boolean removeVMListener(VMListener listener) {
        return vmListeners.remove(listener);
    }

    public boolean removeDebuggerListener(DebuggerListener listener) {
        return debuggerListeners.remove(listener);
    }

    public boolean removePromptListener(PromptListener listener) {
        return promptListeners.remove(listener);
    }

    public boolean removeSourcePathListener(SourcePathListener listener) {
        return sourcePathListeners.remove(listener);
    }

    public boolean removeProgressListener(ProgressListener listener) {
        return progressListeners.remove(listener);
    }    

    private final boolean shipping = false;
    private final void me() {
        if (shipping) return;
        String user = System.getProperty("user.name");
        if (user != null && user.equals("palm") &&
            System.getProperty("no.logger") != null) {
            System.out.println("Setting options for me...");
            getOptions().set("debugging-for-jeff");
            getOptions().set("nodemangle");
            getOptions().set("extra");
        }
    }

    public String toLongString() { return toString(); }

    // ------------------- Exception stuff -----------------------
    public void exception(String msg) {
        exception(msg, null);
    }
    public void exception(Throwable t) {
        exception(null, t);
    }
    public void exception(String msg, Throwable t) {
        error((msg != null ? msg : "") + (t != null ? ":" + t.getMessage() : ""));
    }
    public void classNotPreparedException(ClassNotPreparedException e,
                                          String className) {
        exception("The class " + className + " was not prepared", e);
    }
    public void objectCollectedException(ObjectCollectedException e,
                                         String object) {
        exception("The object " + object + " was collected", e);
    }
    public void invalidLineNumberException(InvalidLineNumberException e,
                                           String className, int line) {
        exception("The line " + line + " is invalid in class " + className, e);
    }
}

class Threads {

    private ThreadReference threadRef;
    private ThreadGroupReference threadGroupRef;
    private int frame;
    private Debugger debugger;
    private ThreadReference stoppedThreadRef;

    Threads(Debugger debugger) {
        this.debugger = debugger;
    }

    void flush() {
        threadRef        = null;
        threadGroupRef   = null;
        frame            = 0;
        stoppedThreadRef = null;
    }

    ThreadReference setThread(ThreadReference threadRef) {
        return (this.threadRef = threadRef);
    }

    ThreadGroupReference setThreadGroup(ThreadGroupReference threadGroupRef) {
        return (this.threadGroupRef = threadGroupRef);
    }

    int setFrame(int frame) {
        return (this.frame = frame);
    }

    ThreadReference setStoppedThread(ThreadReference stoppedThreadRef) {
        return (this.stoppedThreadRef = stoppedThreadRef);
    }

    ThreadReference getThread() throws NoVMException {
        if (threadRef == null) {
            Iterator iter = debugger.getVM().allThreads().iterator();
            while (iter.hasNext()) {
                return (threadRef = (ThreadReference) iter.next());
            }
        }
        return threadRef;
    }

    ThreadGroupReference getThreadGroup() throws NoVMException {
        if (threadGroupRef == null) {
            Iterator iter = debugger.getVM().topLevelThreadGroups().iterator();
            while (iter.hasNext()) {
                return (threadGroupRef = (ThreadGroupReference) iter.next());
            }
        }
        return threadGroupRef;
    }

    int getFrame() {
        return frame;
    }

    StackFrame getStackFrame() throws NoVMException {
        StackFrame stackFrame = null;
        try {
            stackFrame = getThread().frame(frame);
        } catch (IncompatibleThreadStateException itse) {
        }
        return stackFrame;
    }

    ThreadReference getStoppedThread() throws NoVMException {
        if (stoppedThreadRef == null) {
            Iterator iter = debugger.getVM().allThreads().iterator();
            while (iter.hasNext()) {
                ThreadReference threadRef = (ThreadReference) iter.next();
                if (threadRef.isAtBreakpoint()) {
                    return (stoppedThreadRef = threadRef);
                }
            }
        }
        return stoppedThreadRef;
    }
}
