/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.debugger.base;

import java.io.*;
import java.util.*;
import java.text.*;

public class ErrorLogger extends AbstractErrorLogger {

    public ErrorLogger(Debugger debugger) {
        super(debugger);
    }

    public ErrorLogger(Adapter adapter) {
        super(adapter);
    }

    public ErrorLogger() {
        super(new Adapter());
    }

    public static class Adapter implements ErrorLoggable {
        public Options getOptions() {
            return new Options();
        }
        public void fatal(String str) {
            System.err.println(str);
            System.exit(-1);
        }
    }

    public static void main(String[] args) {
        AbstractErrorLogger logger = new AbstractErrorLogger(
            new ErrorLoggable() {
                public Options getOptions() {
                    return new Options();
                }
                public void fatal(String str) {
                    System.err.println(str);
                }
            }
        );
        try {
            String s = null; s.toString();
        } catch (Exception e) {
            logger.log(e);
        }
    }

    public static void error(Object o) {
        AbstractErrorLogger logger = new AbstractErrorLogger(
            new ErrorLoggable() {
                public Options getOptions() {
                    return new Options();
                }
                public void fatal(String str) {
                    System.err.println(str);
                }
            }
        );
        if (o instanceof String) {
            logger.log((String) o);
        } else if (o instanceof Throwable) {
            logger.log((Throwable) o);
        } else {
            logger.log(o + "");
        }
    }
}

interface ErrorLoggable {
    public Options getOptions();
    public void fatal(String str);
}

class AbstractErrorLogger {

    private ErrorLoggable loggable;
    private String errorFileName = ".ajdbErrors";

    public AbstractErrorLogger(ErrorLoggable loggable) {
        this.loggable = loggable;
        Options options = loggable.getOptions();
        Object userErrorFileName = options.get("errorfile");
        if (userErrorFileName != null) {
            errorFileName = userErrorFileName + "";
        }
    }

    public void log(Throwable t) {
        log("No message", t);
    }

    public void log(String msg) {
        log(msg, new Throwable(msg));
    }

    public void log(String msg, Throwable t) {
        try {
            PrintWriter printWriter = new PrintWriter(getOutputStream());
            printWriter.println(heading());
            printWriter.println(msg);
            t.printStackTrace(printWriter);
            printWriter.close();
        } catch (IOException ioe) {
            loggable.fatal("Could not open error file for writing.");
        }
    }


    private String heading() {
        String date = DateFormat.getDateTimeInstance(DateFormat.FULL,
                                                     DateFormat.FULL).format(new Date());
        String username = System.getProperty("user.name");
        StringBuffer buf = new StringBuffer();
        buf.append("# -----------------------------------------------------------\n");
        buf.append("# Date: " + date + "\n");
        buf.append("# User: " + username + "\n");
        buf.append("# -----------------------------------------------------------\n");
        return buf.toString();
    }

    public OutputStream getOutputStream() throws IOException {
        File file = new File(getErrorFileName());
        boolean append = false;
        if (file != null && file.exists()) {
            append = true;
        }
        FileOutputStream out = new FileOutputStream(getErrorFileName(), append);
        return out;
    }

    public String getErrorFileName() {
        return errorFileName;
    }

    public void setErrorFile(String errorFileName) {
        this.errorFileName = errorFileName;
    }
}