/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.base;

import com.sun.jdi.*;
import java.math.*;

// For parsing expressions
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.parser.*;

/**
 * Evaluator.java
 *
 *
 * Created: Fri Sep 08 17:07:25 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class Evaluator  {

    private Debugger debugger;

    public final static String[] keywords = {
        "abstract",
        "boolean",
        "break",
        "byte",
        "catch",
        "char",
        "class",
        "continue",
        "do",
        "double",
        "else",
        "extends",
        "false",
        "final",
        "finally",
        "float",
        "for",
        "if",
        "implements",
        "import",
        "instanceof",
        "int",
        "long",
        "new",
        "null",
        "package",
        "private",
        "protected",
        "public",
        "return",
        "short",
        "static",
        "super",
        "switch",
        "case",
        "default",
        "synchronized",
        "this",
        "throw",
        "throws",
        "try",
        "true",
        "void",
        "while",
    };

    public Evaluator(Debugger debugger) {
        this.debugger = debugger;
    }

    private AJDebugger dbg() {
        return (AJDebugger) debugger;
    }

    private VirtualMachine vm() throws NoVMException {
        return dbg().getVM();
    }

    public Value getValue(Object o, StackFrame frame)
        throws NoVMException, DebuggerException {
        String valueString = (o + "").trim();
        if (isIdent(valueString)) {
            LocalVariable local = getLocalVariable(o, frame);
            if (local != null) {
                return frame.getValue(local);
            }
        }
        return getPrimitiveValue(valueString);
    }

    public Value getPrimitiveValue(String valueString)
        throws NoVMException, DebuggerException {
        if (valueString.startsWith("\"")) {
            if (valueString.length() > 1) {
                String s;
                if (valueString.endsWith("\"")) {
                    s = valueString.substring(1, valueString.length() - 1);
                } else {
                    s = valueString.substring(1);
                }
                return vm().mirrorOf(s);
            }
        } else if (valueString.startsWith("'")) {
            if (valueString.length() > 1) {
                char c = valueString.charAt(1);
                return vm().mirrorOf(c);
            }
        } else if (valueString.equals("true")) {
            return vm().mirrorOf(true);
        } else if (valueString.equals("false")) {
            return vm().mirrorOf(false);
        } else {
            try {
                return vm().mirrorOf(Integer.parseInt(valueString));
            } catch (NumberFormatException e) {
            }
            try {
                return vm().mirrorOf(Long.parseLong(valueString));
            } catch (NumberFormatException e) {
            }
            try {
                return vm().mirrorOf(Float.parseFloat(valueString));
            } catch (NumberFormatException e) {
            }
            try {
                return vm().mirrorOf(Double.parseDouble(valueString));
            } catch (NumberFormatException e) {
            }
        }
        throw new UnableToParseException(valueString);
    }

    public LocalVariable getLocalVariable(Object o, StackFrame frame)
        throws NoVMException, DebuggerException {
        String valueString = o + "";
        if (isIdent(valueString)) {
            try {
                LocalVariable local = frame.visibleVariableByName(valueString);
                return local;
            } catch (AbsentInformationException aie) {
            } catch (InvalidStackFrameException isfe) {
            }
        }
        throw new UnableToParseException(valueString);
    }

    private boolean isNumber(String s) {
        try {
            new BigDecimal(s);
            return true;
        } catch (NumberFormatException e) {
        }
        return false;
    }

    private boolean isIdent(String s) {
        if (s == null || s.length() == 0 || s.equals("")) {
            return true;
        }
        if (isJavaKeyword(s)) {
            return false;
        }
        if (!(Character.isJavaIdentifierStart(s.charAt(0)))) {
            return false;
        }
        for (int i = 1; i < s.length(); i++) {
            if (!Character.isJavaIdentifierPart(s.charAt(i))) {
                return false;
            }
        }
        return true;
    }

    private boolean isJavaKeyword(String s) {
        for (int i = 0; i < keywords.length; i++) {
            if (s.equals(keywords[i])) {
                return true;
            }
        }
        return false;
    }

    class UnableToParseException extends DebuggerException {
        public UnableToParseException(String expr) {
            super("The following expression cannot be currently parsed: " + expr);
        }
    }
}
