/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.base;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import com.sun.jdi.request.*;
import java.util.*;

/**
 * EventSpinner.java
 *
 * The event list.
 *
 * Created: Mon Aug 21 11:43:00 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class EventSpinner implements Runnable {

    private Thread runner;
    private EventQueue eventQueue;
    private Debugger debugger;
    private boolean isRunning;
    private boolean isDone;

    /**
     * Create a new EventSpinner out of the object that should be notified.
     */
    public EventSpinner(Debugger debugger) {
        this.debugger = debugger;
    }

    public void start() {
        Debug.debug(this, "Starting the event queue");
        try {
            eventQueue = debugger.getVM().eventQueue();
            isRunning = true;
            (runner = new Thread(this)).start();
        } catch (NoVMException e) {
        }
    }

    public void run() {
        EventIterator iter = null;
        boolean disconnect = false;
        try {
            while (isRunning) {
                EventSet eventSet = eventQueue.remove();
                handle(eventSet);
            }
        } catch (InterruptedException ie) {
        } catch (VMDisconnectedException vmde) {
            disconnect = true;
            vmde.getMessage(); // wait a bit
        } catch (Exception e) {
            debugger.app().handleInternalException(e);
        } finally {
            try {
                EventSet eventSet = null;
                while ((eventSet = eventQueue.remove(10)) != null) {
                    handle(eventSet);
                }
            } catch (InterruptedException ie) {
            } catch (VMDisconnectedException vmde) {
                disconnect = true;
            }
        }

        try {
            isDone = true;
            notifyAll();
        } catch (IllegalMonitorStateException e) {
        }
    }

    private void handle(EventSet set) throws VMDisconnectedException {
        if (set == null) {
            return;
        }
        try {
            EventIterator iter = set.eventIterator();
            boolean resume = false;
            printThreads(true);
            while (iter.hasNext()) {
                Event event = iter.nextEvent();
                resume |= debugger.handle(event);
            }
            if (resume) {
                debugger.getVM().resume();
            }
            printThreads(false);
        } catch (NoVMException nvme) {
        }
    }

    boolean printThreads = false;
    void printThreads(boolean before) {
        try {
            if (!printThreads) return;
            Iterator threadIter = debugger.getVM().allThreads().iterator();
            System.out.print((before ? "BEFORE" : "AFTER") + ": ");
            while (threadIter.hasNext()) {
                ThreadReference threadRef = (ThreadReference) threadIter.next();
                System.out.print("(" + threadRef.uniqueID() + "):" + threadRef.suspendCount() + ", ");
            }
            System.out.println();
        } catch (NoVMException nvme) {
        }
    }

    public void shutDown() {
        isRunning = false;
        if (runner != null) runner.interrupt();
        while (!isDone) {
            try {wait();} catch (InterruptedException ie) {}
                          catch (IllegalMonitorStateException imse) { return; }
        }
    }

    class DummyVMDisconnectEvent implements VMDisconnectEvent {
        public String toString() {
            return "Dummy VMDisconnectEvent: vm=" + debugger.vm();
        }

        public VirtualMachine virtualMachine() {
            return debugger.vm();
        }

        public EventRequest request() {
            return null;
        }
    }
}
