/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.base;

import org.aspectj.debugger.request.*;
import java.util.*;


/**
 * Parser.java
 *
 *
 * Created: Mon Aug 28 13:26:40 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class Parser { //implements Errors {

    private CommandReceiver     debugger;
    private StringTokenizer     tok;
    private Request             lastRequest;
    private String              command;
    private String              lastCommand = "";

    private final static String[] commands = {
        "break",
        "catch",
        "class",
        "classes",
        "classpath",
        "clear",
        "connect",
        "cont",
        "deport",
        "dir",
        "down",
        "dump",
        "eval",
        "exclude",
        "exit",
        "fields",
        "help",
        "ignore",
        "import",
        "interrupt",
        "kill",
        "list",
        "locals",
        "lock",
        "ls",
        "methods",
        "monitor",
        "next",
        "object",
        "print",
        "pwd",
        "quit",
        "read",
        "resume",
        "run",
        "set",
        "step",
        "stepi",
        "stop",
        "suspend",
        "thread",
        "threadgroup",
        "threadgroups",
        "threadlocks",
        "threads",
        "tostring",
        "trace",
        "unmonitor",
        "untrace",
        "unwatch",
        "up",
        "use",
        "version",
        "view",
        "watch",
        "where",
        "workingdir",
        "?",
    };

    private static Map aliases = new HashMap();
    static {
        String[] strings = {
            "break",        "b",
            "cont",         "c",
            "down",         "d",
            "help",         "h",
            "list",         "l",
            "monitor",      "m",
            "next",         "n",
            "quit",         "q",
            "read",         "r",
            "step",         "s",
            "stepi",        "si",
            "step up",      "su",
            "up",           "u",
            "where",        "w",
        };
        for (int i = 0; i < strings.length; i += 2) {
            aliases.put(strings[i+1], strings[i]);
        }
    }
    private String alias(String line) {
        Object alias = aliases.get(line);
        return alias != null ? alias + "" : line;
    }

    private static String[] noVmCommands = {
        "catch",
        "clear",
        "dir",
        "exit",
        "help",
        "ignore",
        "import",
        "list",
        "ls",
        "pwd",
        "quit",
        "read",
        "run",
        "sourcepath",
        "stop",
        "unwatch",
        "use",
        "version",
        "watch",
        "?",
    };

    public final static String THREAD                 = "Thread number not specified.";
    public final static String KILL                   = "Usage: kill <thread id> <throwable>";
    public final static String PRINT                  = "No objects specified.";
    public final static String DUMP                   = PRINT;
    public final static String EVAL                   = PRINT;
    public final static String SET                    = "Usage: set <lvalue> = <expr>";
    public final static String CLASS                  = "No class specified.";
    public final static String METHODS                = CLASS;
    public final static String FIELDS                 = CLASS;
    public final static String THREADGROUP            = "Threadgroup name not specified.";
    public final static String CATCH                  = "Threadgroup name not specified.";
    public final static String WATCH                  = "watch [access|all] ...";
    public final static String UNWATCH                = "unwatch [access|all] ...";
    public final static String UNMONITOR              = "usage: unmonitor <monitor#>";
    public final static String READ                   = "usage: read <command-filename>";
    public final static String UNKOWN_COMMAND         = "Unkown command, type 'help' for more help";
    public final static String CLASS_METHOD           = "<class>.<method>";
    public final static String SOURCE_LINE            = "<source>:<line>";
    public final static String SOURCE_LINE_NO_NUMBER  = "<source>:<line> (<line was not a number)";
    public final static String CLASS_LINE             = "<class>:<line>";
    public final static String CLASS_LINE_NO_NUMBER   = "<class>:<line> (<line was not a number)";
    public final static String END_OF_LINE            = "Unexpected end of line";
    public final static String EXPECTING_NUMBER       = "Expecting a number";
    public final static String IMPORT                 = "Usage: import <path>";
    public final static String DEPORT                 = "Usage: deport <path>";

    private CommandCompletor completor = new CommandCompletor(commands);

    public Parser(CommandReceiver debugger) {
        this.debugger = debugger;
    }

    public Object parse(String line) throws ParseException,
                                            NoVMException,
                                            DebuggerException {
        Object result;
        command = alias(line);
        tok = new StringTokenizer(line);

        if (!tok.hasMoreTokens()) {
            //return new BangBangRequest(lastRequest);
            return null;
        }

        String first = tok.nextToken().trim();
        first = completor.getCommand(first);

        checkRunningVM(first);

        if (first.endsWith("!")) {
            //Debugger.demoMode = false;
            first = first.substring(0, first.length()-1);
        }

        // threads <threadgroup>?
        if (first.equals("threads")) {
            String threadGroup = question();
            result = debugger.threadsCommand(threadGroup);

        } // thread <thread>
        else if (first.equals("thread")) {
            String thread = next(THREAD);
            result = debugger.threadCommand(thread);

        } // suspend <thread>*
        else if (first.equals("suspend")) {
            List threads = listStar();
            result = debugger.suspendCommand(threads);

        } // resume <thread>*
        else if (first.equals("resume")) {
            List threads = listStar();
            result = debugger.resumeCommand(threads);

        } // where <thread>?
        else if (first.equals("where")) {
            String thread = question();
            result = debugger.whereCommand(thread);

        } // up <#>
        else if (first.equals("up")) {
            int num = integerStar();
            if (num == -1) num = 1;
            result = debugger.upCommand(num);

        } // down <#>
        else if (first.equals("down")) {
            int num = integerStar();
            if (num == -1) num = 1;
            result = debugger.downCommand(num);

        } // kill <thread> <expr>
        else if (first.equals("kill")) {
            String thread = next(KILL);
            String expr = rest(KILL);
            result = debugger.killCommand(thread, expr);

        } // interrupt <thread>
        else if (first.equals("interrupt")) {
            String thread = next();
            result = debugger.interruptCommand(thread);

        } // print <expr>
        else if (first.equals("print")) {
            String expr = rest(PRINT);
            result = debugger.printCommand(expr);

        } // dump <expr>
        else if (first.equals("dump")) {
            String expr = rest(DUMP);
            result = debugger.dumpCommand(expr);

        } // eval <expr>
        else if (first.equals("eval")) {
            String expr = rest(EVAL);
            result = debugger.evalCommand(expr);

        } // set <lvalue> = <expr>
        else if (first.equals("set")) {
            String lvalue = ident(SET);
            String eq     = next(SET);
            String expr   = "";
            if (eq.indexOf("=") == -1) {
                expr = eq;
            } else {
                expr = rest(SET);
            }
            result = debugger.setCommand(lvalue, expr);

        } // locals
        else if (first.equals("locals")) {
            result = debugger.localsCommand();

        } // classes
        else if (first.equals("classes")) {
            result = debugger.classesCommand();

        } // class <classID>
        else if (first.equals("class")) {
            String classID = next(CLASS);
            result = debugger.classCommand(classID);

        } // methods <classID>
        else if (first.equals("methods")) {
            String classID = next(METHODS);
            result = debugger.methodsCommand(classID);

        } // fields <classID>
        else if (first.equals("fields")) {
            String classID = next(FIELDS);
            result = debugger.fieldsCommand(classID);

        } // threadgroups
        else if (first.equals("threadgroups")) {
            result = debugger.threadGroupsCommand();

        } // threadgroup <threadgroup>
        else if (first.equals("threadgroup")) {
            String threadgroup = next(THREADGROUP);
            result = debugger.threadGroupCommand(threadgroup);

        } // stop ...
        else if (first.equals("stop") ||
                 first.equals("break")) {
            result =  stop(restStar());

        } // clear ...
        else if (first.equals("clear")) {
            result = clear(restStar());

        } // catch <classID>
        else if (first.equals("catch")) {
            String classID = next(CATCH);
            result = debugger.catchCommand(classID);

        } // ignore <classID>
        else if (first.equals("ignore")) {
            String classID = next();
            result = debugger.ignoreCommand(classID);

        } // watch ...
        else if (first.equals("watch")) {
            String accessOrAll = next(WATCH);
            ClassMethod cm = classMethod(rest(WATCH));
            if (accessOrAll.equals("access")) {
                result = debugger.watchAccessCommand(cm.className, cm.methodName);
            } else if (accessOrAll.equals("all")) {
                result = debugger.watchAllCommand(cm.className, cm.methodName);
            } else {
                throw new SyntaxException(WATCH);
            }

        } // unwatch ...
        else if (first.equals("unwatch")) {
            String accessOrAll = next(UNWATCH);
            ClassMethod cm = classMethod(rest(UNWATCH));
            if (accessOrAll.equals("access")) {
                result = debugger.unwatchAccessCommand(cm.className, cm.methodName);
            } else if (accessOrAll.equals("all")) {
                result = debugger.unwatchAllCommand(cm.className, cm.methodName);
            } else {
                throw new SyntaxException(UNWATCH);
            }

        } // trace methods <thread>
        else if (first.equals("trace")) {
            eat("methods");
            String thread = next();
            result = debugger.traceMethodsCommand(thread);

        } // untrace methods <thread>
        else if (first.equals("untrace")) {
            eat("methods");
            String thread = next();
            result = debugger.untraceMethodsCommand(thread);

        } // step [up]
        else if (first.equals("step")) {
            if (star().equals("up")) {
                result = debugger.stepUpCommand();
            } else {
                result = debugger.stepCommand();
            }

        } // stepi
        else if (first.equals("stepi")) {
            result = debugger.stepiCommand();

        } // next
        else if (first.equals("next")) {
            result = debugger.nextCommand();

        } // cont
        else if (first.equals("cont")) {
            result = debugger.contCommand();

        } // exclude <classes>*
        else if (first.equals("exclude")) {
            List classes = listStar();
            result = debugger.excludeCommand(classes);

        } // classpath
        else if (first.equals("classpath")) {
            result = debugger.classpathCommand();

        } // lock <expr>
        else if (first.equals("lock")) {
            String expr = rest();
            result = debugger.lockCommand(expr);

        } // threadlocks <thread>
        else if (first.equals("threadlocks")) {
            String thread = next();
            result = debugger.threadlocksCommand(thread);

        } // !!
        else if (first.equals("!!")) {
            System.err.println("lastCommand=" + lastCommand);
            result = debugger.bangBangCommand(lastCommand);

        } // help
        else if (first.equals("help") || first.equals("?")) {
            result = debugger.helpCommand();

        } // version
        else if (first.equals("version")) {
            result = debugger.versionCommand();

        } // [quit|exit]
        else if (first.equals("exit") || first.equals("quit")) {
            result = debugger.quitCommand();

        } // run [<class> <arg>*]
        else if (first.equals("run")) {
            Options opts = debugger.getOptions();
            opts.fill(restStar());
            result = debugger.runCommand(opts);

        } // connect
        else if (first.equals("connect")) {
            result = debugger.connectCommand();
            
        } // use <sourcepath>*
        else if (first.equals("use")) {
            String sourcePath = restStar();
            result = debugger.useCommand(sourcePath);

        } // workingdir <workingdirpath>*
        else if (first.equals("workingdir")) {
            String workingdirPath = restStar();
            result = debugger.workingdirCommand(workingdirPath);
        } // list <source>:<line>
        else if (first.equals("list")) {
            String sourceName = star();
            int startLine = -1;
            int endLine = -1;
            if (sourceName.equals("")) {
                mandateRunningVM("list");
                result = debugger.listCommand();
            } else {
                if (isNumber(sourceName)) {
                    startLine = Integer.parseInt(sourceName);
                } else {
                    startLine = integerStar();
                }
                endLine = integerStar();
                if (startLine == -1) {
                    result = debugger.listCommand(sourceName);
                } else if (endLine == -1) {
                    result = debugger.listCommand(sourceName, startLine);
                } else {
                    result = debugger.listCommand(sourceName, startLine, endLine);
                }
            }

        } // monitor <command>
        else if (first.equals("monitor")) {
            String command = restStar();
            if (command.equals("")) {
                result = debugger.monitorCommand();
            } else {
                command = completor.getCommand(command);
                result = debugger.monitorCommand(command);
            }

        } // unmonitor <#>
        else if (first.equals("unmonitor")) {
            int number = integer(rest(UNMONITOR), UNMONITOR);
            result = debugger.unmonitorCommand(number);

        } // read <filename>
        else if (first.equals("read")) {
            String fileName = rest(READ);
            result = debugger.readCommand(fileName);

        } // import <path>
        else if (first.equals("import")) {
            String path = rest(IMPORT);
            result = debugger.importCommand(path);

        } // deport <path>
        else if (first.equals("deport")) {
            String path = rest(DEPORT);
            result = debugger.deportCommand(path);

        } // view <source>
        else if (first.equals("view")) {
            String source = rest();
            result = debugger.viewCommand(source);
        } // tostring
        else if (first.equals("tostring")) {
            result = debugger.tostringCommand();
        }  // pwd
        else if (first.equals("pwd")) {
            result = debugger.pwdCommand();
        } // ls
        else if (first.equals("ls") || first.equals("dir")) {
            result = debugger.lsCommand(rest());
        }

        // Default        
        else {
            throw new ParseException(UNKOWN_COMMAND);
        }

        lastCommand = command;

        return result;
    }

    private boolean equals(String first, String str) {
        return first.equalsIgnoreCase(str);
    }

    private void mandateRunningVM(String command) throws ParseException {
        checkRunningVM(command, true);
    }
    private void checkRunningVM(String first) throws ParseException {
        checkRunningVM(first, false);
    }
    private void checkRunningVM(String first, boolean force) throws ParseException {
        if (((Debugger) debugger).isRunning()) {
            return;
        }
        if (!force) {
            if (debugger.getOptions().isSet("extra")) {
                return;
            }
            int N = noVmCommands.length;
            for (int i = 0; i < N; i++) {
                if (noVmCommands[i].equals(first)) {
                    return;
                }
            }
        }
        String msg = "Command '" + first
                   + "' is not valid until the VM is started "
                   + "with the 'run' command";
        throw new ParseException(msg);
    }

    private boolean isNumber(String str) {
        try {
            Integer.parseInt(str);
            return true;
        } catch (NumberFormatException e) {
        }
        return false;
    }

    private String star() {
        if (tok.hasMoreTokens()) {
            return tok.nextToken().trim();
        }
        return "";
    }

    private String restStar() {
        if (tok.hasMoreTokens()) {
            return tok.nextToken("").trim();
        }
        return "";
    }

    private String eat(String str) throws ParseException {
        if (tok.hasMoreTokens()) {
            return next();
        }
        throw new ParseException("Expecting '" + str + "'; found an end-of-line");
    }

    private Object stop(String theRest)
        throws ParseException, NoVMException, DebuggerException {
        return parseBreakpoint(theRest, true);
    }

    private Object clear(String theRest)
        throws ParseException, NoVMException, DebuggerException {
        return parseBreakpoint(theRest, false);
    }

    private Object parseBreakpoint(String theRest, boolean isStop)
        throws ParseException, NoVMException, DebuggerException {
        try {
            String rest = theRest;
            rest = rest.trim();
            if (rest.equals("") || rest.length() == 0) {
                return isStop ? debugger.stopCommand() : debugger.clearCommand();
            } else if (rest.startsWith("at ") ||
                       rest.startsWith("in ") ||
                       rest.startsWith("on ")) {
                rest = rest.substring(3).trim();
                theRest = rest;
            }

            String classOrSource = "";
            int index = 0;
            boolean isMethod = true;
            for (int i = rest.length()-1; i >= 0; i--) {
                if (isOneOf(rest.charAt(i), ":; ")) {
                    index = i;
                    isMethod = false;
                    break;
                }
                if (rest.charAt(i) == '(') {
                    index = i;
                    isMethod = true;
                    break;
                }
            }
            if (index > 0) {
                classOrSource = rest.substring(0, index).trim();
            } else {
                classOrSource = rest;
            }
            if (index < rest.length()) {
                rest = rest.substring(index).trim();
            } else {
                rest = "";
            }
            isMethod |= rest.length() == 0 || rest.startsWith("(");
            if (isMethod) {
                int ilastDot = classOrSource.length() - 1;
                while (ilastDot > 0) {
                    if (classOrSource.charAt(ilastDot) == '.') {
                        break;
                    }
                    ilastDot--;
                }
                String className   = theRest.substring(0, ilastDot);
                String methodProto = theRest.substring(ilastDot+1);
                return isStop
                        ? debugger.stopInCommand (className, methodProto)
                        : debugger.clearInCommand(className, methodProto);
            } else {
                if (isOneOf(rest.charAt(0), ":;,")) {
                    rest = rest.substring(1);
                }
                if (classOrSource.endsWith(".java")) {
                    int integer = integer(rest.trim());
                    return isStop
                            ? debugger.stopOnCommand (classOrSource, integer)
                            : debugger.clearOnCommand(classOrSource, integer);
                } else {
                    int integer = integer(rest);
                    return isStop
                            ? debugger.stopAtCommand (classOrSource, integer)
                            : debugger.clearAtCommand(classOrSource, integer);
                }
            }
        } catch (IndexOutOfBoundsException ioobe) {
        }
        throw new ParseException("Unexpected command");
    }


    private boolean isOneOf(char c, String s) {
        for (int i = 0; i < s.length(); i++) {
            if (c == s.charAt(i)) {
                return true;
            }
        }
        return false;
    }

    static class ClassMethod {
        String className;
        String methodName;
        public ClassMethod(String className, String methodName) {
            this.className = className;
            this.methodName = methodName;
        }
    }

    private ClassMethod classMethod(String rest) throws ParseException {
        try {
            return _classMethod(rest);
        } catch (Exception e) {
        }
        throw new SyntaxException(CLASS_METHOD);
    }

    private ClassMethod _classMethod(String rest) throws ParseException {
        int iparen = rest.indexOf("(");
        String className = "";
        String methodName = "";
        int idot = -1;
        if (iparen == -1) {
            idot = rest.lastIndexOf(".");
        } else {
            idot = iparen;
            while (idot > 0) {
                if (rest.charAt(idot) == '.') {
                    break;
                }
                idot--;
            }
        }
        if (idot == -1) {
            throw new SyntaxException(CLASS_METHOD);
        }
        className = rest.substring(0, idot);
        methodName = rest.substring(idot + 1);
        return new ClassMethod(className, methodName);
    }

    static class SourceLine {
        String sourceName;
        int line;
        public SourceLine(String sourceName, int line) {
            this.sourceName = sourceName;
            this.line = line;
        }
    }

    private SourceLine sourceLine(String rest) throws ParseException {
        StringTokenizer colonTok = new StringTokenizer(rest, ":");
        int icolon = rest.lastIndexOf(":");
        if (icolon == -1)  {
            throw new SyntaxException(SOURCE_LINE);
        }
        try {
            String sourceName = rest.substring(0, icolon);
            String lineName   = rest.substring(icolon+1);
            int    line       = Integer.parseInt(lineName);
            return new SourceLine(sourceName, line);
        } catch (NoSuchElementException e) {
            throw new SyntaxException(SOURCE_LINE);
        } catch (NumberFormatException nfe) {
            throw new SyntaxException(SOURCE_LINE_NO_NUMBER);
        }
    }

    static class ClassLine {
        String className;
        int line;
        public ClassLine(String className, int line) {
            this.className = className;
            this.line = line;
        }
    }

    private ClassLine classLine(String rest) throws ParseException {
        StringTokenizer colonTok = new StringTokenizer(rest, ":");
        int icolon = rest.lastIndexOf(":");
        if (icolon == -1)  {
            throw new SyntaxException(SOURCE_LINE);
        }
        try {
            String className = rest.substring(0, icolon);
            String lineName  = rest.substring(icolon+1);
            int    line      = Integer.parseInt(lineName);
            return new ClassLine(className, line);
        } catch (NoSuchElementException e) {
            throw new SyntaxException(CLASS_LINE);
        } catch (NumberFormatException nfe) {
            throw new SyntaxException(CLASS_LINE_NO_NUMBER);
        }
    }

    private String dump() {
        if (tok.hasMoreTokens()) {
            String str = tok.nextToken("").trim();
            tok = new StringTokenizer(str);
            return str;
        }
        return "";
    }

    private String nextStar() {
        if (tok.hasMoreTokens()) {
            return tok.nextToken().trim();
        }
        return "";
    }

    private String next(String s) throws ParseException {
        if (tok.hasMoreTokens()) {
            return tok.nextToken().trim();
        }
        throw new ParseException(s);
    }

    private String next() throws ParseException {
        return next(END_OF_LINE);
    }

    private String question() {
        if (tok.hasMoreTokens()) {
            return tok.nextToken("").trim();
        }
        return "";
    }

    private String rest() throws ParseException {
        return rest(END_OF_LINE);
    }

    private String rest(String s) throws ParseException {
        if (tok.hasMoreTokens()) {
            return tok.nextToken("").trim();
        }
        throw new ParseException(s);
    }

    private List listStar() {
        List list = new Vector();
        while (tok.hasMoreTokens()) {
            list.add(tok.nextToken().trim());
        }
        return list;
    }

    private int integer() throws ParseException {
        return integer(rest());
    }

    private int integer(String num) throws ParseException {
        return integer(num, null);
    }

    private int integer(String rest, String msg) throws ParseException {
        if (rest != null) {
            try {
                return Integer.parseInt(rest.trim());
            } catch (NumberFormatException e) {
            }
        }
        throw new ParseException((msg == null ? "Expecting a number" : msg));
    }

    private int integerStar() throws ParseException {
        String next = nextStar();
        if (next.equals("")) {
            return -1;
        }
        try {
            return Integer.parseInt(next.trim());
        } catch (NumberFormatException e) {
        }
        throw new ParseException(EXPECTING_NUMBER);
    }

    private boolean ws(String s) {
        for (int i = 0; i < s.length(); i++) {
            if (Character.isWhitespace(s.charAt(i))) {
                return true;
            }
        }
        return false;
    }

    private String ident(String s) throws ParseException {
        try {
            return ident();
        } catch (ParseException e) {}
        throw new ParseException(s);
    }

    private String ident() throws ParseException {
        String token = next();
        if (!(Character.isJavaIdentifierStart(token.charAt(0)))) {
            throw new InvalidIdentException(token);
        }
        for (int i = 1; i < token.length(); i++) {
            char c = token.charAt(i);
            if (!(Character.isJavaIdentifierPart(c)) &&
                c != '.' &&
                c != '[' &&
                c != ']') {
                throw new InvalidIdentException(token);
            }
        }
        return token;
    }

    public String command() {
        return command;
    }

    class CommandCompletor {
        String[] commands;

        public CommandCompletor(String[] commands) {
            this.commands = commands;
        }

        public String getCommand(String start) throws ParseException {
            return getCommand(start, true);
        }

        public String getCommand(String start, boolean throwException) throws ParseException {
            Object match = aliases.get(start);
            if (match != null) {
                return match + "";
            }

            Vector matches = new Vector();
            for (int i = 0; i < commands.length; i++) {
                if (commands[i].equals(start)) {
                    return start;
                } else if (commands[i].startsWith(start)) {
                    matches.add(commands[i]);
                }
            }
            if (matches.size() == 0) {
                if (start.endsWith("!")) {
                    String str = getCommand(start.substring(0, start.length()-1), false);
                    if (str != null) return str+"!";
                    else throwException = true;
                }
                if (throwException)
                    throw new ParseException("No command begins with the prefix '" + start + "'");
                else return null;
            } else if (matches.size() > 1) {
                if (throwException) {
                    String msg = "The prefix '" + start + "' has the possible completions:\n";
                    Iterator iter = matches.iterator();
                    while (iter.hasNext()) {
                        msg += "\t" + iter.next() + "\n";
                    }
                    throw new ParseException(msg);
                }
                else return null;
            }
            return matches.get(0) + "";
        }
    }
}
