/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.base;

import org.aspectj.debugger.request.*;
import java.io.*;
import java.util.*;

/**
 * SourceManager.java
 *
 *
 * Created: Tue Aug 29 11:38:31 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class SourceManager  {

    private Debugger debugger;
    private String sourcePath;

    public SourceManager(Debugger debugger) {
        this(debugger, ".");
    }

    public SourceManager(Debugger debugger, String sourcePath) {
        this.debugger = debugger;
        try {
            this.sourcePath = new File(sourcePath).getCanonicalPath();
        } catch (IOException ioe) {
        } //FILE
//            this.sourcePath = new File(sourcePath).getAbsolutePath();
    }

    public void setSourcePath(String sourcePath) {
        try {
            File file = new File(sourcePath);
            if (file.exists()) {
                this.sourcePath = file.getCanonicalPath();
                //FILE
//                this.sourcePath = file.getAbsolutePath();
                return;
            }
        } catch (IOException ioe) {
        }
        //!!! Do we still want to assign it
        this.sourcePath = sourcePath;
        if (debugger.getOptions().isSet("extra")) {
            debugger.error("The path " + sourcePath + " does not exist");
            showPossiblePaths(sourcePath);
        }
    }

    private void showPossiblePaths(String sourcePath) {
        Iterator iter = getPossiblePaths(sourcePath).iterator();
        if (!iter.hasNext()) {
            debugger.error("There are no possible paths");
            return;
        }
        String error = "Possible paths are:\n";
        while (iter.hasNext()) {
            error += "\n  " + iter.next();
        }
        debugger.error(error);
    }

    private List getPossiblePaths(String sourcePath) {
        List completions = new Vector();
        char c;
        int isep1 = sourcePath.lastIndexOf(File.separatorChar);
        int isep2 = sourcePath.lastIndexOf('/');
        int isep3 = sourcePath.lastIndexOf('\\');
        int ilastSep = -1;
        if        ((isep1 >= isep2) && (isep1 >= isep3)) {
            ilastSep = isep1;
        } else if ((isep2 >= isep1) && (isep2 >= isep3)) {
            ilastSep = isep2;
        } else if ((isep3 >= isep2) && (isep3 >= isep2)) {
            ilastSep = isep3;
        } else {
            ilastSep = isep1;
        }
        if (ilastSep == -1) {
            return completions;
        }
        File file = new File(sourcePath.substring(0, ilastSep+1) + ".");
        File parent = file.getParentFile();
        if (!parent.exists()) {
            return completions;
        }
        File[] files = parent.listFiles(new FileFilter() {
            public boolean accept(File _file) {
                return _file.isDirectory();
            }
        });
        for (int i = 0; i < files.length; i++) {
            completions.add(files[i]);
        }
        return completions;
    }

    public String getSourcePath() {
        return sourcePath;
    }

    public List getSourceLines(String sourceName) {
        Vector list = new Vector();
        LineNumberReader in = getReader(sourceName);
        if (in == null) {
            return list;
        }
        try {
            String lineString = "";
            int lineNumber = -1;
            while ((lineNumber = in.getLineNumber()) != -1 &&
                   (lineString = in.readLine())      != null) {
                list.add(new SourceLine(sourceName, lineString, lineNumber));
            }
        } catch (IOException ioe) {
        }
        try {
            in.close();
        } catch (IOException ioe) {
        }
        return list;
    }

    public String getLine(String sourceName, int lineNumber) {
        return getSourceLine(sourceName, lineNumber).lineString;
    }

    public List getSourceLines(String sourceName, int startLine, int endLine) throws DebuggerException {
        Vector lines = new Vector();
        lines.add(emptySourceLine());
        LineNumberReader in = getReader(sourceName);
        if (in == null) {
            return lines;
        }
        if (startLine > endLine) {
            throw new DebuggerException("Start line must be greater than the end line.");
        }
        while (in.getLineNumber() < startLine) {
            try {
                in.readLine();
            } catch (IOException ioe) {
                return lines;
            }
        }
        lines = new Vector();
        int num = -1;
        String lineStr = "";
        try {
            while ((lineStr = in.readLine()) != null &&
                   (num = in.getLineNumber()) <= endLine) {
                lines.add(new SourceLine(sourceName, lineStr, num-1));
            }
        } catch (IOException ioe) {
        }
        return lines;
    }

    public SourceLine getSourceLine(String sourceName, int lineNumber) {
        SourceLine sl = emptySourceLine();
        LineNumberReader in = getReader(sourceName);
        if (in == null) {
            return null;
        }
        while (in.getLineNumber() < lineNumber) {
            try {
                in.readLine();
            } catch (IOException ioe) {
                return null;
            }
        }
        String line = "";
        try {
            line = in.readLine();
            sl = new SourceLine(sourceName, line, lineNumber);
            in.close();
        } catch (IOException ioe) {
        }
        return sl;
    }

    private LineNumberReader getReader(String relativeFileName) {
        LineNumberReader in = null;
        try {
            if (relativeFileName != null) {
                in = new LineNumberReader(new FileReader(resolve(relativeFileName)));
            }
        } catch (FileNotFoundException fnfe) {
        }
        return in;
    }

    public File getFile(String relativeFileName) {
        String fullName = resolve(relativeFileName);
        File file = new File(fullName);
        return file;
    }

    private String resolve(String fileName) {
        return resolve(fileName, debugger.getFilePaths());
    }

    private String resolve(String fileName, List filePaths) {
        Iterator iter = filePaths.iterator();
        while (iter.hasNext()) {
            String filePath = iter.next() + "";
            String fullName = fileName;
            if (!fullName.startsWith(getSourcePath())) {
                fullName = sourcePath + File.separator +
                    (filePath.equals("") ? "" : File.separator) + fileName;
            }
            try {
                File file = new File(fullName);
                if (file.exists()) {
                    //FILE
                    return file.getCanonicalPath();
//                    return file.getAbsolutePath();
                }
            } catch (IOException ioe) {
            }
        }
        return fileName;
    }

    private SourceLine emptySourceLine() {
        return new SourceLine("<not available>", "", -1);
    }

    public interface LineUpdatable {
        public void update(int line);
    }

    public static class SourceLine {
        public String sourceName = "";
        public String lineString = "";
        public int lineNumber = -1;
        public final static String tab = "    ";
        private static char c;
        private boolean hasBreakpoint = false;
        private boolean isBreakpoint  = false;
        private BreakpointRequestAction request;
        private boolean isExecutable = false;
        private String className     = null;
        public final static SourceLine emptyLine =
            new SourceLine("<not-available>", "", 0);

        public final static int     NONE       = 0;
        public final static int     UNVERIFIED = 1;
        public final static int     VERIFIED   = 2;

        private             int     status     = NONE;

        private LineUpdatable updater;
        public void setUpdater(LineUpdatable updater) {
            this.updater = updater;
        }
        private void update() {
            if (updater != null) updater.update(getLineNumber());
        }

        public SourceLine(String sourceName, String lineString, int lineNumber) {
            this.sourceName = sourceName;
            this.lineNumber = lineNumber + 1;
            this.lineString = expandTabs(lineString);
        }

        public String toString() {
            return sourceName + ":" + lineNumber + ": " + lineString;
        }

        public String getSourceName() {
            return sourceName;
        }

        public String getLineString() {
            return lineString;
        }

        public int getLineNumber() {
            return lineNumber;
        }

        public void setStatus(int status) {
            this.status = status;
            update();
        }

        public int getStatus() {
            return status;
        }

        public String getClassName() {
            return className;
        }

        public void setClassName(String className) {
            this.className = className;
        }

        public void setRequest(BreakpointRequestAction request) {
            this.request = request;
            request.setSourceLine(this);
        }

        public boolean hasBreakpoint() {
            return status != NONE;
        }

        public void clear() {
            setStatus(NONE);
        }

        public void unverify(BreakpointRequestAction request) {
            setRequest(request);
            setStatus(UNVERIFIED);
        }

        public void verify(BreakpointRequestAction request) {
            setRequest(request);
            setStatus(VERIFIED);
        }

        public void ex(String className) {
            setExecutable(true, className);
        }

        public void unex() {
            setExecutable(false);
        }

        public void setExecutable(boolean isExecutable) {
            setExecutable(isExecutable, null);
        }

        public void setExecutable(boolean isExecutable, String className) {
            this.isExecutable = isExecutable;
            setClassName(className);
        }

        public boolean isExecutable() {
            return isExecutable;
        }

        public BreakpointRequestAction getRequest() {
            return this.request;
        }

        private String expandTabs(String line) {
            char[] chars = line.toCharArray();
            StringBuffer buf = new StringBuffer();
            for (int i = 0; i < chars.length; i++) {
                buf.append(((c = chars[i]) == '\t') ? tab : c+"");
            }
            return buf.toString();
        }
    }
}
