/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.ide.IDEComponentDirector;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

interface K {
    final static int QUIT               = KeyEvent.VK_Q;
    final static int STEPUP             = KeyEvent.VK_U;
    final static int STEPI              = KeyEvent.VK_I;
    final static int STEP               = KeyEvent.VK_S;
    final static int NEXT               = KeyEvent.VK_N;
    final static int CONT               = KeyEvent.VK_O;
    final static int RUN                = KeyEvent.VK_R;
    final static int BREAKPOINT         = KeyEvent.VK_B;
    final static int RESUME             = KeyEvent.VK_E;
    final static int SUSPEND            = KeyEvent.VK_X;
    final static int WHERE              = KeyEvent.VK_W;
    final static int USE                = KeyEvent.VK_K;
    final static int CLEAR              = KeyEvent.VK_L;
    final static int DUMP               = KeyEvent.VK_M;
    final static int EVAL               = KeyEvent.VK_V;
    final static int FIELDS             = KeyEvent.VK_F;
    final static int LOCALS             = KeyEvent.VK_Q;
    final static int PRINT              = KeyEvent.VK_P;
    final static int HELP               = KeyEvent.VK_F1;
    final static int ABOUT              = KeyEvent.VK_F2;
    final static int DISCONNECT         = KeyEvent.VK_D;
    final static int CATCH              = KeyEvent.VK_C;
    final static int EXECUTE            = KeyEvent.VK_1;
    final static int WATCH              = KeyEvent.VK_H;
}

interface S {
    final static String QUIT            = "Quit";
    final static String STEPUP          = "Step up";
    final static String STEPI           = "I-Step";
    final static String STEP            = "Step";
    final static String NEXT            = "Next";
    final static String CONT            = "Continue";
    final static String RUN             = "Run";
    final static String BREAKPOINT      = "Break";
    final static String RESUME          = "Resume";
    final static String SUSPEND         = "Suspend";
    final static String WHERE           = "Where";
    final static String USE             = "Use";
    final static String CLEAR           = "Clear";
    final static String DUMP            = "Dump";
    final static String EVAL            = "Evaluate";
    final static String FIELDS          = "Fields";
    final static String LOCALS          = "Locals";
    final static String PRINT           = "Print";
    final static String HELP            = "Help";
    final static String ABOUT           = "About";
    final static String DISCONNECT      = "Disconnect";
    final static String CATCH           = "Catch";
    final static String EXECUTE         = "Execute";
    final static String WATCH           = "Watch";
}

interface C {
    final static String QUIT            = "quit";
    final static String STEPUP          = "step up";
    final static String STEPI           = "stepi";
    final static String STEP            = "step";
    final static String NEXT            = "next";
    final static String CONT            = "cont";
    final static String RUN             = "run";
    final static String BREAKPOINT      = "stop";
    final static String RESUME          = "resume";
    final static String SUSPEND         = "suspend";
    final static String WHERE           = "where";
    final static String USE             = "use";
    final static String CLEAR           = "clear";
    final static String DUMP            = "dump";
    final static String EVAL            = "eval";
    final static String FIELDS          = "fields";
    final static String LOCALS          = "locals";
    final static String PRINT           = "print";
    final static String HELP            = "help";
    final static String ABOUT           = "about";
    final static String DISCONNECT      = "disconnect";
    final static String CATCH           = "catch";
    final static String WATCH           = "watch";
}

interface AJActionsConsts {
    public final static int ALL         = -1;
    public final static int NONE        =  0;
    public final static String LABEL    = "label";
    public final static String ICON     = "icon";
}

abstract class AbstractActionStruct
    extends AbstractAction
    implements AJActionsConsts {

    private boolean hasIcon;
    private GUIDebugger guiDebugger;

    AbstractActionStruct(GUIDebugger guiDebugger, String s, Icon i) {
        super(s, i);
        this.guiDebugger = guiDebugger;
        putValue(LABEL, s);
        putValue(ICON,  i);
        hasIcon = !i.equals(AJIcons.NO_ICON);
    }

    AbstractActionStruct(GUIDebugger guiDebugger) {
        super();
        this.guiDebugger = guiDebugger;
    }

    public boolean hasIcon() {
        return hasIcon;
    }

    protected GUIDebugger guid() {
        return guiDebugger;
    }

    protected AJDebugger ajdb() {
        return guid().getDebugger();
    }

    abstract int key();
    abstract int mask();
}

abstract class AllMaskAbstractActionStruct extends AbstractActionStruct {
    AllMaskAbstractActionStruct(GUIDebugger guid) {
        super(guid);
    }
    int mask() { return ALL; }
}

abstract class NoMaskAbstractActionStruct extends AbstractActionStruct {
    NoMaskAbstractActionStruct(GUIDebugger guid, String s, Icon i) {
        super(guid, s, i);
    }
    int mask() { return NONE; }
}

abstract class CtrlAbstractActionStruct extends AbstractActionStruct {
    CtrlAbstractActionStruct(GUIDebugger guid, String s, Icon i) {
        super(guid, s, i);
    }
    int mask() {
        return Event.CTRL_MASK;
    }
}

class BreakpointAction extends CtrlAbstractActionStruct {
    public BreakpointAction(GUIDebugger guid) {
        super(guid, S.BREAKPOINT, AJIcons.getIcon(AJIcons.STOP_ICON));
    }
    protected BreakpointAction(GUIDebugger guid, String name) {
        super(guid, name, AJIcons.getIcon(AJIcons.STOP_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showSetBreakpointDialog();
    }
    int key()   {
        return K.BREAKPOINT;
    }
}

//  class JBuilderLineBreakpointAction extends BreakpointAction {
//      public JBuilderLineBreakpointAction(GUIDebugger guid) {
//          super(guid, "Add Breakpoint");
        
//      }
//      public void actionPerformed(ActionEvent e) {
//          ((IDEComponentDirector)guid().getGui()).showBreakpointDialog();
//      }
//  }

class IDELineBreakpointAction extends BreakpointAction {
    public IDELineBreakpointAction(GUIDebugger guid) {
        super(guid, "Add Breakpoint");
        
    }
    public void actionPerformed(ActionEvent e) {
        ((IDEComponentDirector)guid().getGui()).showBreakpointDialog();
    }
}


class ClearAction extends CtrlAbstractActionStruct {
    public ClearAction(GUIDebugger guid) {
        super(guid, S.CLEAR, AJIcons.getIcon(AJIcons.CLEAR_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showClearBreakpointDialog();
    }
    int key()   {
        return K.CLEAR;
    }
}

class RunAction extends CtrlAbstractActionStruct {
    public RunAction(GUIDebugger guid) {
        super(guid, S.RUN, AJIcons.getIcon(AJIcons.RUN_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showRunDialog();
    }
    int key() {
        return K.RUN;
    }
}

class NoArgsRunAction extends RunAction {
    public NoArgsRunAction(GUIDebugger guid) {
        super(guid);
    }
    public void actionPerformed(ActionEvent e) {
        if (ajdb().vm() == null) {
            ajdb().execute(C.RUN);
        } else {
            ajdb().execute(C.CONT);
        }
    }
}

class JBuilderRunAction extends NoArgsRunAction {
    public JBuilderRunAction(GUIDebugger guid) {
        super(guid);
    }
}

class ForteRunAction extends NoArgsRunAction {
    public ForteRunAction(GUIDebugger guid) {
        super(guid);
    }
}

class CatchAction extends CtrlAbstractActionStruct {
    public CatchAction(GUIDebugger guid) {
        super(guid, S.CATCH, AJIcons.getIcon(AJIcons.CATCH_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showCatchDialog();
    }
    int key() {
        return K.CATCH;
    }
}

class WatchAction extends CtrlAbstractActionStruct {
    public WatchAction(GUIDebugger guid) {
        super(guid, S.WATCH, AJIcons.getIcon(AJIcons.WATCH_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showWatchDialog();
    }
    int key() {
        return K.WATCH;
    }
}

class HelpAction extends NoMaskAbstractActionStruct {
    public HelpAction(GUIDebugger guid) {
        super(guid, S.HELP, AJIcons.getIcon(AJIcons.HELP_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.HELP);
    }
    int key() {
        return K.HELP;
    }
}

class AboutAction extends NoMaskAbstractActionStruct {
    public AboutAction(GUIDebugger guid) {
        super(guid, S.ABOUT, AJIcons.getIcon(AJIcons.ABOUT_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showAboutDialog();
    }
    int key() {
        return K.ABOUT;
    }
}

abstract class NoIconCtrlAbstractActionStruct
    extends CtrlAbstractActionStruct {
    public NoIconCtrlAbstractActionStruct(GUIDebugger guid, String s) {
        super(guid, s, AJIcons.NO_ICON);
    }
}

class DumpAction extends CtrlAbstractActionStruct {
    public DumpAction(GUIDebugger guid) {
        super(guid, S.DUMP, AJIcons.getIcon(AJIcons.DUMP_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showDumpDialog();
    }
    int key() {
        return K.DUMP;
    }
}

class EvalAction extends CtrlAbstractActionStruct {
    public EvalAction(GUIDebugger guid) {
        super(guid, S.EVAL, AJIcons.getIcon(AJIcons.EVALUATE_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showEvalDialog();
    }
    int key() {
        return K.EVAL;
    }
}

class FieldsAction extends CtrlAbstractActionStruct {
    public FieldsAction(GUIDebugger guid) {
        super(guid, S.FIELDS, AJIcons.getIcon(AJIcons.FIELDS_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showFieldsDialog();
    }
    int key() {
        return K.FIELDS;
    }
}

class DisconnectAction extends CtrlAbstractActionStruct {
    public DisconnectAction(GUIDebugger guid) {
        super(guid, S.DISCONNECT, AJIcons.getIcon(AJIcons.DISCONNECT_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.DISCONNECT);
    }
    int key() {
        return K.DISCONNECT;
    }
}

class LocalsAction extends CtrlAbstractActionStruct {
    public LocalsAction(GUIDebugger guid) {
        super(guid, S.LOCALS, AJIcons.getIcon(AJIcons.LOCALS_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.LOCALS);
    }
    int key() {
        return K.LOCALS;
    }
}

class PrintAction extends CtrlAbstractActionStruct {
    public PrintAction(GUIDebugger guid) {
        super(guid, S.PRINT, AJIcons.getIcon(AJIcons.PRINT_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showPrintDialog();
    }
    int key() {
        return K.PRINT;
    }
}



class UseAction extends CtrlAbstractActionStruct {
    public UseAction(GUIDebugger guid) {
        super(guid, S.USE, AJIcons.getIcon(AJIcons.USE_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showUseDialog();
    }
    int key() {
        return K.USE;
    }
}

class ExecuteAction extends CtrlAbstractActionStruct {
    public ExecuteAction(GUIDebugger guid) {
        super(guid, S.EXECUTE, AJIcons.getIcon(AJIcons.BLANK_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        Dialogs.showExecuteDialog();
    }
    int key() {
        return K.EXECUTE;
    }
}

class ContAction extends CtrlAbstractActionStruct {
    public ContAction(GUIDebugger guid) {
        super(guid, S.CONT, AJIcons.getIcon(AJIcons.CONT_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.CONT);
    }
    int key() {
        return K.CONT;
    }
}

class NextAction extends CtrlAbstractActionStruct {
    public NextAction(GUIDebugger guid) {
        super(guid, S.NEXT, AJIcons.getIcon(AJIcons.NEXT_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.NEXT);
    }
    int key() {
        return K.NEXT;
    }
}

class StepAction extends CtrlAbstractActionStruct {
    public StepAction(GUIDebugger guid) {
        super(guid, S.STEP, AJIcons.getIcon(AJIcons.STEP_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.STEP);
    }
    int key() {
        return K.STEP;
    }
}

class StepIAction extends CtrlAbstractActionStruct {
    public StepIAction(GUIDebugger guid) {
        super(guid, S.STEPI, AJIcons.getIcon(AJIcons.STEPI_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.STEPI);
    }
    int key() {
        return K.STEPI;
    }
}

class StepUpAction extends CtrlAbstractActionStruct {
    public StepUpAction(GUIDebugger guid) {
        super(guid, S.STEPUP, AJIcons.getIcon(AJIcons.STEPUP_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.STEPUP);
    }
    int key() {
        return K.STEPUP;
    }
}

class QuitAction extends CtrlAbstractActionStruct {
    public QuitAction(GUIDebugger guid) {
        super(guid, S.QUIT, AJIcons.getIcon(AJIcons.QUIT_ICON));
    }
    public void actionPerformed(ActionEvent e) {
        ajdb().execute(C.QUIT);
    }
    int key() {
        return K.QUIT;
    }
}

class JBuilderQuitAction extends QuitAction {
    public JBuilderQuitAction(GUIDebugger guid) {
        super(guid);
    }
    //XXX
    public void _actionPerformed(ActionEvent e) {
        guid().getGui().setVisible(false);
        guid().getGui().dispose();
    }
}

interface ThreadGetter {
    public String getThread();
}

abstract class CtrlThreadGetterNeedingActionStruct
    extends CtrlAbstractActionStruct {

    protected ThreadGetter getter;
    protected String cmd;

    public CtrlThreadGetterNeedingActionStruct(GUIDebugger guid, String label, Icon icon,
                                               String cmd, ThreadGetter getter)
    {
        super(guid, label, icon);
        this.getter = getter;
        this.cmd    = cmd;
    }

    public void actionPerformed(ActionEvent e) {
        String thread = getter.getThread();
        if (thread == null) {
            return;
        }
        String command = cmd + " " + thread;
        ajdb().execute(command);
    }

}

class WhereAction extends CtrlThreadGetterNeedingActionStruct {
    public WhereAction(GUIDebugger guid, ThreadGetter getter) {
        super(guid, S.WHERE, AJIcons.getIcon(AJIcons.WHERE_ICON),
              C.WHERE, getter);
    }
    int key() {
        return K.WHERE;
    }
}

class SuspendAction extends CtrlThreadGetterNeedingActionStruct {
    public SuspendAction(GUIDebugger guid, ThreadGetter getter) {
        super(guid, S.SUSPEND, AJIcons.getIcon(AJIcons.SUSPEND_ICON),
              C.SUSPEND, getter);
    }
    int key() {
        return K.SUSPEND;
    }
}

class ResumeAction extends CtrlThreadGetterNeedingActionStruct {
    public ResumeAction(GUIDebugger guid, ThreadGetter getter) {
        super(guid, S.RESUME, AJIcons.getIcon(AJIcons.RESUME_ICON),
              C.RESUME, getter);
    }
    int key() {
        return K.RESUME;
    }
}

class ForteResumeAction extends ResumeAction {
    public ForteResumeAction(GUIDebugger guid, ThreadGetter getter) {
        super(guid, getter);
    }
    public void actionPerformed(ActionEvent e) {
        if (ajdb().vm() == null) {
            ajdb().execute(C.RUN);
        } else {
            super.actionPerformed(e);
        }
    }
}
