/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.debugger.gui;

import org.aspectj.tools.ide.*;
import com.sun.jdi.*;
import java.util.*;
import javax.swing.tree.*;

public class AJCClassNode extends ClassNode {

    private String name;
    private Declaration classDec;
    private boolean isAspect;

    public AJCClassNode(String name) {
        this(null, name);
    }

    public AJCClassNode(Declaration classDec) {
        super(null);
        this.classDec = classDec;
        init(null, classDec.getSignature());
    }

    public AJCClassNode(ReferenceType refType) {
        this(refType, refType.name());
    }

    private AJCClassNode(ReferenceType refType, String name) {
        super(refType);
        init(refType, name);
    }

    private void init(ReferenceType refType, String name) {
        this.name = name;
        addKids();
    }

    protected String name() {
        return name;
    }

    protected String packageName() {
        return classDec.getPackageName();
    }

    protected boolean callsAddKids() {
        return true;
    }

    public int getType() {
        if (isAspect) return AJIcons.ASPECT_ICON;
        if (kind().equals("class")) return AJIcons.CLASS_ICON;
        return AJIcons.INTERFACE_ICON;
    }

    public void addMethods() {
        if (name == null) return;
        if (classDec == null) {
            String fileName = AJLineMapper.getSourceFilePathFromAJCClass(name);
            if (fileName == null) return;
            Declaration[] decs = SymbolManager.getSymbolManager().getDeclarations(fileName);
            classDec = classDec(decs);
        }
        if (classDec == null) return;
        addMethods(classDec);
        addAdvices(classDec);
    }

    public static AJCClassNode[] makeNodes(String fileName) {
        Declaration[] decs = SymbolManager.getSymbolManager().getDeclarations(fileName);
        if (decs == null) return new AJCClassNode[0];
        List classNodes = new Vector();
        for (int i = 0; i < decs.length; i++) {
            Declaration dec = decs[i];
            if (dec != null) classNodes.add(new AJCClassNode(dec));
        }
        int N = classNodes.size();
        AJCClassNode[] nodes = new AJCClassNode[N];
        int i = 0;
        for (Iterator iter = classNodes.iterator(); iter.hasNext() && i < N; i++) {
            nodes[i] = (AJCClassNode) iter.next();
        }
        return nodes;
    }

    public String getName() {
        return name;
    }

    private void addAdvices(Declaration classDec) {
        addAdvices(classDec.getDeclarations());
    }

    private void addAdvices(Declaration[] decs) {
        if (decs == null) return;
        for (int i = 0; i < decs.length; i++) {
            Declaration dec = decs[i];
            if (dec != null && dec.getKind().equals("advice")) {
                addAdvice(dec);
            }
        }
    }

    private void addMethods(Declaration classDec) {
        addMethods(classDec.getDeclarations());
    }

    private void addMethods(Declaration[] decs) {
        if (decs == null) return;
        for (int i = 0; i < decs.length; i++) {
            Declaration dec = decs[i];
            if (dec != null && isMethod(dec)) {
                addMethod(dec);
            }
        }
    }

    private boolean isMethod(Declaration dec) {
        return dec.getKind().equals("method") || dec.getKind().equals("constructor");
    }

    private void addAdvice(Declaration dec) {
        AdviceNode adviceNode = new AdviceNode(dec);
        add(adviceNode);
        isAspect = true;
        addMethods(adviceNode, dec);
    }

    private void addMethod(Declaration dec) {
        if (dec == null) return;
        MethodNode methodNode = new AdvisableMethodNode(dec);
        add(methodNode);
        addAdvice(methodNode, dec);
    }

    private void addMethods(AdviceNode adviceNode, Declaration dec) {
        Declaration[] pointsTo = dec.getPointsTo();
        if (pointsTo == null) return;
        for (int i = 0; i < pointsTo.length; i++) {
            Declaration methodDec = pointsTo[i];
            if (methodDec != null) {
                MethodNode methodNode = new AdvisableMethodNode(methodDec);
                adviceNode.add(methodNode);
            }
        }
    }

    private void addAdvice(MethodNode methodNode, Declaration dec) {
        Declaration[] pointedToBy = dec.getPointedToBy();
        if (pointedToBy == null) return;
        for (int i = 0; i < pointedToBy.length; i++) {
            Declaration advice = pointedToBy[i];
            if (advice != null) {
                AdviceNode adviceNode = new AdviceNode(advice);
                methodNode.add(adviceNode);
            }
        }
    }

    private Declaration classDec(Declaration[] decs) {
        if (decs == null) return null;
        List q = new Vector();
        for (int i = 0; i < decs.length; i++) {
            q.add(decs[i]);
        }
        while (!q.isEmpty()) {
            Declaration dec = (Declaration) q.remove(0);
            if (dec != null && isType(dec)) {
                String signature = dec.getSignature();
                String packageName = dec.getPackageName();
                String fullSignature;
                if (packageName != null && !"".equals(packageName)) {
                    fullSignature = packageName + "." + signature;
                } else {
                    fullSignature = signature;
                }
                fullSignature = fullSignature.trim();
                //if (dec.getSignature().equals(name)) return dec;
                if (fullSignature.equals(name)) return dec;
                Declaration[] newDecs = dec.getDeclarations();
                if (newDecs != null) {
                    for (int i = 0; i < newDecs.length; i++) {
                        q.add(newDecs[i]);
                    }
                }
            }
        } while (!q.isEmpty());
        return null;
    }

    private boolean isType(Declaration dec) {
        return dec.getKind().equals("class") || dec.getKind().equals("interface");
    }

    protected String kind() {
        if (isAspect) return "aspect";
        if (classDec == null) {
            return "class";
        }
        return classDec.getKind();
    }

    public String getRelativePath() {
        return classDec.getFilename();
    }

    public int getLine() {
        return classDec.getBeginLine();
    }

    public List getAdviceStrings() {
        if (!isAspect) return null;
        List adviceStrings = new Vector();
        if (children == null) return adviceStrings;
        Iterator iter = children.iterator();
        while (iter.hasNext()) {
            TreeNode treeNode = (TreeNode) iter.next();
            if (treeNode instanceof AdviceNode) {
                adviceStrings.add(((AdviceNode) treeNode).getBreakpoint());
            }
        }
        return adviceStrings;
    }
}
