/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.debugger.gui;

import com.sun.jdi.*;
import java.util.*;

public class AJClassNode extends AJTreeNode {

    private ReferenceType   refType;

    public AJClassNode(ReferenceType refType) {
        super(AJIcons.CLASS_ICON);
        this.refType    = refType;
        setUserObject(refType);
        if (!callsAddKids()) addKids();
    }

    private byte numAddKids = 0;
    protected void addKids() {
        if (numAddKids++ > 1) throw new RuntimeException("Only call addKids() once");
        addSuperTypes();
        addStaticFields();
        addMethods();
    }

    protected boolean callsAddKids() {
        return false;
    }

    private void addSuperTypes() {
        if      (refType instanceof ClassType)     addSuperTypes((ClassType)     refType);
        else if (refType instanceof InterfaceType) addSuperTypes((InterfaceType) refType);
        else if (refType instanceof ArrayType)     addSuperTypes((ArrayType)     refType);
    }

    private void addSuperTypes(ClassType classType) {
        if (classType.name().equals("java.lang.Object")) {
            return;
        }
        ClassType superClass = classType.superclass();
        if (superClass != null) {
            add(newClassNode(superClass));
        }
        addSuperTypes(classType.allInterfaces());
    }

    protected AJClassNode newClassNode(ReferenceType refType) {
        return new AJClassNode(refType);
    }

    private void addSuperTypes(InterfaceType interfaceType) {
        addSuperTypes(interfaceType.superinterfaces());
    }

    private void addSuperTypes(List interfaces) {
        Iterator iter = interfaces.iterator();
        while (iter.hasNext()) {
            InterfaceType interfaceType = (InterfaceType) iter.next();
            add(newClassNode(interfaceType));
        }
    }

    private void addSuperTypes(ArrayType arrayType) {
    }

    private void addStaticFields() {
        if (refType == null) return;
        Iterator iter = refType.allFields().iterator();
        while (iter.hasNext()) {
            Field field = (Field) iter.next();
            if (field.isStatic() && field.declaringType().equals(refType)) {
                AJValueNode valueNode = AJValueNodeFactory.make(refType, field);

                //??? We don't want any compiler-generated fields
                if (valueNode.getName().indexOf("$") == -1) {
                    add(valueNode);
                }
            }
        }
    }

    /**
     * Nodes in the class tree should implement this method.
     */
    protected void addMethods() {
    }

    public int getType() {
        if (refType == null) {
//              throw new RuntimeException
//                  ("This should have been overriden " + this + ":" + getClass());
        } else if (refType instanceof ClassType) {
            ClassType type = (ClassType) refType;
            if (type.isStatic()) {
                if (type.isPublic())         return AJIcons.CLASS_STATIC_PUBLIC_ICON;
                if (type.isPackagePrivate()) return AJIcons.CLASS_STATIC_PACKAGE_ICON;
                if (type.isProtected())      return AJIcons.CLASS_STATIC_PROTECTED_ICON;
                if (type.isPrivate())        return AJIcons.CLASS_STATIC_PRIVATE_ICON;
            } else {
                if (type.isPublic())         return AJIcons.CLASS_PUBLIC_ICON;
                if (type.isPackagePrivate()) return AJIcons.CLASS_PACKAGE_ICON;
                if (type.isProtected())      return AJIcons.CLASS_PROTECTED_ICON;
                if (type.isPrivate())        return AJIcons.CLASS_PRIVATE_ICON;
            }
            return AJIcons.CLASS_ICON;
        } else if (refType instanceof InterfaceType) {
            return AJIcons.INTERFACE_ICON;
        }
        //XXX
        //throw new RuntimeException("Bad type for class node " + refType);
        return AJIcons.CLASS_ICON;
    }

    public String toString() {
        return ((ReferenceType)getUserObject()).name();
    }

}
