/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import java.io.*;
import java.util.*;
import java.util.List;
import org.aspectj.tools.ide.*;
import com.sun.jdi.*;

public class AJLineMapper {

    /** Put classes that were not compiled with AJC here. */
    private static HashMap noAJCHash = new HashMap();

    public final static String NO_SOURCE_STRING             = "<not-available>";
    public final static String NON_MAPPING_SOURCE_STRING    = "<non-mapping-source>";
    public final static String UNKNOWN_SOURCE_STRING        = "<unknown>";
    public final static int NO_SOURCE_LINE                  = -1;
    public final static int NON_MAPPING_SOURCE_LINE         = -2;

    public static SymbolManager symbolManager = new SymbolManager();

    /**
     * TODO: move to org.aspectj.tools.SourceLine
     *
     * @param   filePath    the full path to the preprocessed source file
     * @param   lineNumber  line number in the preprocessed source file
     * @return  true     when the method has a corresponding signature in the source code
     * @return  false    otherwise
     */
    public boolean hasBody() {
        return true;
    }

    public static SymbolManager sm() {
        return symbolManager;
    }

    public static int getCorrespondingOutputLineNo(String classID, int line) {
        return AJLineMapper.getCorrespondingOutputLine(classID, line).line;
    }

    public static SourceLine getCorrespondingOutputLine(String classID, int line) {
        String filePath = getSourceFilePathFromAJCClass(classID);
        return getCorrespondingOutputLineFromFullSource(filePath, line);
    }

    public static SourceLine getCorrespondingOutputLineFromSource(String source,
                                                                  int line) {
        String filePath = getFullFileNameFromFile(source, false);
        return getCorrespondingOutputLineFromFullSource(filePath, line);
    }

    /**
     * Takes (class, line) and returns output line.  If <code>class</code> was
     * not compiled with AJC, it will return <code>line</code>.  If this line
     * does not map to the output file, this method will return the greatest
     * line number less that <code>line</code> that does.
     */
    static SourceLine getCorrespondingOutputLineFromFullSource(String filePath,
                                                                      int line) {
        //
        // The filePath is null when a non-AJC-compiled class was passed in.
        //
        if (filePath == null) {
            return AJLineMapper.unknownFileSourceLine(line);
        }

        //
        // If filePath is not null, then we can map it to the ajworkingdir.
        // We go backwards until a valid line number is found.  We can't go
        // forward because we have no bound in that direction.
        //
        int lineNo = -1;
        String newFile = UNKNOWN_SOURCE_STRING;
        SourceLine sl = null;
        line++;
        while (--line >= 0 && lineNo == -1) {
            sl = symbolManager.mapToOutputLine(filePath, line);
//            //palm.util.Util.debug("filePath=" + filePath + ", line=" + line + ", sl=" + sl);
            if (sl != null) {
                lineNo = sl.line;
                newFile = sl.filename;
            }
        }

        if (lineNo == -1) {
            return nonMappingSourceLine();
        } else {
                //palm.util.Util.debug("Returning " + sl);
            return new SourceLine(newFile, lineNo + 1);
        }
    }

    /**
      * TODO: move to org.aspectj.tools.SymbolManager
      *
      * @param      methodName  method name without type or parameter list
      * @return     method name with ajc-specific name mangling removed, unchanged if it's a Java name
      */
    public static String translateGeneratedMethodName(String methodName) {
        String sigToken             = "$signature";
        String memberToken          = "$member";
        String staticMemberToken    = "$static$member";
        String adviceToken          = "$reception";
        String aroundToken          = "$around$reception";
        int sigTokenStart           = methodName.indexOf(sigToken);
        int memberTokenStart        = methodName.indexOf(memberToken);
        int staticMemberTokenStart  = methodName.indexOf(staticMemberToken);
        int adviceTokenStart        = methodName.indexOf(adviceToken);
        int aroundTokenStart        = methodName.indexOf(aroundToken);
        int indexNameStart          = methodName.lastIndexOf("$");
        if (sigTokenStart > -1 && indexNameStart > -1) {
            return methodName.substring(methodName.indexOf("$")+1, sigTokenStart);
        }
        else if (staticMemberTokenStart > -1 && indexNameStart > -1) {
            return methodName.substring(methodName.indexOf("$")+1, staticMemberTokenStart);
        }
        else if (memberTokenStart > -1 && indexNameStart > -1) {
            return methodName.substring(methodName.indexOf("$")+1, memberTokenStart);
        }
        else if (adviceTokenStart > -1 && indexNameStart > -1) {
            String adviceNumber =  methodName.substring(adviceTokenStart + adviceToken.length());
            return "<advice #" + adviceNumber +">";
        }
        else if (aroundTokenStart > -1 && indexNameStart > -1) {
            String adviceNumber =  methodName.substring(aroundTokenStart + aroundToken.length());
            return "<advice #" + adviceNumber +">";
        }
        else {
            return methodName;
        }
    }

    public static String getCorrespondingFile(Location loc) {
        String filename = "";
        String qualifiedName = null;
        try {
            String locName = loc.declaringType().name();
            String sourceName = loc.declaringType().sourceName();
            String strippedSourceName = sourceName.substring(0, sourceName.indexOf(".java"));

            String packageName = locName.substring(0, locName.indexOf(strippedSourceName));
            qualifiedName = packageName + strippedSourceName;
            filename = getFileNameFromQualified(qualifiedName, true);

        } catch (AbsentInformationException aie) {

        } //catch (AJLineMapperException
        return filename;
    }

    /**
     * Given a class, this will return the source file in which it was
     * declared.  If the class wasn't compiled with AJC, it will return null;
     */
    public static String getSourceFilePathFromAJCClass(String classID) {

        // First, look in the outlaw hash to see if this classID was already
        // denied
        if (noAJCHash.get(classID) != null) {
            return null;
        }

        // Second, see if this class was compiled with AJC.  If it was it will
        // have a file in the ajworkingdir, and that file will point at the
        // source file in which the class we're looking for is.
        String ajFile = getFileNameFromQualified(classID, true);

        SourceLine sourceLine = null;
        String filePath = null;
        if (new File(ajFile).exists()) {
            for (int i = 0; i < 1000 && sourceLine == null; i++) {
                if ( (sourceLine = symbolManager.mapToSourceLine(ajFile, i)) != null) {
                    filePath = sourceLine.filename;
                }
            }
        } else {
            // Store the outlaw for the next time
            noAJCHash.put(classID, classID);
            return null;
        }
        return filePath;
    }

    /**
     * ajworkingdir -> src
     */
    public static SourceLine getCorrespondingLine(Location loc) {
        SourceLine ajline = null;
        String qualifiedName = null;
        try {
            qualifiedName = getQualifiedNameForSource(loc);
            int locLineNumber = loc.lineNumber();
            ajline = getCorrespondingLine(qualifiedName, locLineNumber);

            //HERE//palm.util.Util.debug("AJLine: " + ajline + " for " + qualifiedName + ":" + locLineNumber);

            // If ajline doesn't map, we'll try to map it via the dec parser
            if (ajline != null &&
                ajline.line == AJLineMapper.NON_MAPPING_SOURCE_LINE) {
                String classMethodString = loc.method() + "";
                classMethodString = classMethodString.substring
                                        (0, classMethodString.indexOf(")")+1);
                int i2dot = classMethodString.indexOf("(");
                int numDots = 0;
                while (i2dot > -1) {
                    if (classMethodString.charAt(i2dot) == '.') {
                        if ((++numDots) == 2) {
                            break;
                        }
                    }
                    i2dot--;
                }
                if (i2dot != -1) {
                    classMethodString = classMethodString.substring(i2dot + 1);
                }
                String ajFile = AJLineMapper.getSourceFilePathFromAJCClass
                                                (loc.declaringType().name());
                if (ajFile != null) {
                    int line = AJDecParser.getMethodLineNumber
                                                    (ajFile, classMethodString);
                    ajline = new SourceLine(ajFile, line);
                } else {
                    ajline = new SourceLine(loc.sourceName(), loc.lineNumber());
                }
            }
        } catch (AbsentInformationException aie) {
        }
        return (ajline == null) ? emptySourceLine() : ajline;
    }

    /**
     * Returns whether this is an unkown source file.
     */
    public static boolean isUnknownSource(String source) {
        return source.equals(AJLineMapper.UNKNOWN_SOURCE_STRING);
    }

    static String getQualifiedNameForSource(Location loc)
                                    throws AbsentInformationException {
        ReferenceType refType = loc.declaringType();
        if (refType == null) {
            throw new AbsentInformationException();
        }
        String locName = refType.name();
        String sourceName = loc.sourceName();
        String strippedSourceName =
        sourceName.substring(0, sourceName.indexOf(".java"));
        int lastDot = locName.lastIndexOf('.');
        String packageName = "";
        if (lastDot != -1) {
            packageName = locName.substring(0, lastDot+1);
        }
        String qualifiedName = packageName + strippedSourceName;
        return qualifiedName;
    }

    private static SourceLine emptySourceLine() {
        return new SourceLine(NO_SOURCE_STRING, NO_SOURCE_LINE);
    }

    private static SourceLine nonMappingSourceLine() {
        return new SourceLine(NON_MAPPING_SOURCE_STRING, NON_MAPPING_SOURCE_LINE);
    }

    private static SourceLine unknownFileSourceLine(int line) {
        return new SourceLine(UNKNOWN_SOURCE_STRING, line);
    }

    public static SourceLine getCorrespondLineFromFullSource(String filePath,
                                                             int line) {
        SourceLine newLine = null;
        if (new File(filePath).exists()) {
            int offsetLine = -1;
            try {
                offsetLine = line - 1; // HACK!
                newLine = symbolManager.mapToSourceLine
                        (new File(filePath).getCanonicalPath(), offsetLine);
                        //FILE-HERE
//                        (new File(filePath).getAbsolutePath(), offsetLine);
            } catch (IOException ioe) {
            }
            if (newLine == null) {
                //
                // The file exists, but there is there corresponding mapping
                //
                return nonMappingSourceLine();
            }
            //
            // Good mapping
            //
            return newLine;
        }
        //
        // No workingdir file
        //
        return new SourceLine(AJLineMapper.UNKNOWN_SOURCE_STRING, line);

    }

    /**
     * ajworkingdir -> src
     */
    public static SourceLine getCorrespondingLine(String qualifiedName, int line) {
        String filePath = getFileNameFromQualified(qualifiedName, true);
        return getCorrespondLineFromFullSource(filePath, line);
    }

    /**
     * ajworkingdir -> src
     */
    public static SourceLine getCorrespondingLineFromSource(String sourceName,
                                                            int line) {
        String filePath = AJLineMapper.insertWorkingDir(sourceName);
        return getCorrespondLineFromFullSource(filePath, line);
    }

    static String getFileNameFromQualified(String qualifiedName, boolean inAJWorkingDir) { //Util.debug("getFileNameFromQualified: " + qualifiedName);

        //PALM: Inner classes would be located in the parent file
        String qualifiedNameNoDollars = AJLineMapper.stripDollarsFrom(qualifiedName);

        //PALM: String offset = Env.getSourcePath();
        String offset = ComponentRepository.getSourcePane().getSourcePath() + "";
        if (inAJWorkingDir) {
            offset += "/ajworkingdir";
        }
        String sourcePath = new File(offset).getAbsolutePath();

        //String filename = qualifiedName.replace('.', File.separatorChar) + ".java"; // HACK: coded dirsepchar
        String filename = qualifiedNameNoDollars.replace('.', File.separatorChar) + ".java";
        return sourcePath + File.separator + filename;
    }

    //PALM: Added this method to add the file offset and ajworking to a file name
    static String getFullFileNameFromFile(String fullName, boolean inAJWorkingDir) {
        //PALM: String offset = Env.getSourcePath();
        String offset = ComponentRepository.getSourcePane().getSourcePath() + "";
        if (inAJWorkingDir) {
            offset += "/ajworkingdir";
        }
        String sourcePath = new File(offset).getAbsolutePath();
        if (fullName.indexOf(offset) == -1) {
            return new File(sourcePath + File.separator + fullName).getAbsolutePath();
        } else {
            return new File(fullName).getAbsolutePath();
        }
    }

    //PALM: Added this method to add the file offset and ajworking to a file name
    static String getFullFileNameFromLocation(Location loc, boolean inAJWorkingDir)
            throws AbsentInformationException {
        //PALM: String offset = Env.getSourcePath();
        String offset = ComponentRepository.getSourcePane().getSourcePath() + "";
        if (inAJWorkingDir) {
            offset += "/ajworkingdir";
        }
        String fullName = loc.declaringType().name();
        String srcName = loc.sourceName();
        String pkg = "";
        int iDot = fullName.lastIndexOf(".");
        if (iDot != -1) {
            pkg = fullName.substring(0, iDot);
            pkg = pkg.replace('.', File.separatorChar);
        }
        String srcPath = pkg + File.separator + srcName;

        String sourcePath = new File(offset).getAbsolutePath();
        if (srcPath.indexOf(offset) == -1) {
            return new File(sourcePath + File.separator + srcPath).getAbsolutePath();
        } else {
            return new File(srcPath).getAbsolutePath();
        }
    }

    static String removeRootAndAJWorkingDir(String fullPath) {
        String offset =
            ComponentRepository.getSourcePane().getSourcePath() + File.separator +
            "ajworkingdir" + File.separator;
        int index = offset.length();
        if (!fullPath.startsWith(offset)) {
            return fullPath;
        }
        String returnString = fullPath.substring(index);
        return returnString;
    }

    static String removeRoot(String fullPath) {
        String offset = "";
        int index;
//        try {
            String path = ComponentRepository.getSourcePane().getSourcePath() + "";
            path += File.separator + "ajworkingdir" + File.separator;
            File file = new File(path);
//            offset = file.getCanonicalPath();
            offset = file.getAbsolutePath();
//        } catch (Exception e) {
//            return fullPath;
//        }
        if (fullPath.startsWith(offset)) {
            index = offset.length();
            return fullPath.substring(index+1);
        }
        offset = "";
//        try {
            /*File*/ file = new File(ComponentRepository.getSourcePane().getSourcePath() + "");
//            offset = file.getCanonicalPath();
            offset = file.getAbsolutePath();
//        } catch (Exception e) {
//            return fullPath;
//        }
        if (fullPath.startsWith(offset)) {
            index = offset.length();
            return fullPath.substring(index+1);
        }
        return fullPath;
    }

    static String stripDollarsFrom(String str) {
        int index = str.indexOf("$");
        if (index != -1) {
            return str.substring(0, index);
        }
        return str;
    }

    static String insertWorkingDir(String fullPath) {
        String path = ComponentRepository.getSourcePane().getSourcePath() + "";
        //
        // If fullPath already has a ajwd, return fullpath
        //
        if (fullPath.indexOf("ajworkingdir") != -1) {
            return fullPath;
        }
        //
        // If the root doesn't exist use getFullFileNextFromFile
        //
        int ipath = fullPath.indexOf(path);
        if (ipath == -1) {
            return AJLineMapper.getFullFileNameFromFile(fullPath, true);
        }
        //
        // Add the path and ajwd to fullPath
        //
        String rest = fullPath.substring(ipath + path.length() + 1);
        return  path           + File.separator +
                "ajworkingdir" + File.separator +
                rest;
    }
}