/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.request.*;

import java.awt.*;
import java.awt.event.*;
import java.util.List;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import com.sun.jdi.event.*;

public class AJMenuBar
    extends JMenuBar
    implements AJActionsConsts {

    private JMenu       menu;
    private JMenuItem   item;
    private GUIDebugger guiDebugger;


    public AJMenuBar(GUIDebugger guiDebugger) {
        super();
        this.guiDebugger = guiDebugger;
        mainMenu();
        breakMenu();
        threadMenu();
        objectMenu();
        monitorMenu();
        aspects();
        helpMenu();
    }

    void aspects() {
        // Aspects can advise this
    }

    void mainMenu() {
        menu("Main",KeyEvent.VK_M);
        item(new RunAction(guiDebugger),
             KeyEvent.VK_R);
        item(new UseAction(guiDebugger),
             KeyEvent.VK_U);
        item(new DisconnectAction(guiDebugger),
             KeyEvent.VK_D);
        sep();
        item(new QuitAction(guiDebugger),
             KeyEvent.VK_Q);
        add(menu);
    }

    void breakMenu() {
        menu("Breakpoints", KeyEvent.VK_B);
        item(new BreakpointAction(guiDebugger), KeyEvent.VK_B);
        item(new CatchAction(guiDebugger), KeyEvent.VK_A);
        item(new ClearAction(guiDebugger), KeyEvent.VK_C);
        sep();
        item(new ContAction(guiDebugger), KeyEvent.VK_O);
        item(new NextAction(guiDebugger), KeyEvent.VK_N);
        item(new StepAction(guiDebugger), KeyEvent.VK_S);
        item(new StepUpAction(guiDebugger), KeyEvent.VK_U);
        item(new StepIAction(guiDebugger), KeyEvent.VK_I);
        add(menu);
    }

    void threadMenu() {
        menu("Threads",
             KeyEvent.VK_T);
        item(new SuspendAction(guiDebugger, threadGetterToSuspend()),
             KeyEvent.VK_S);
        item(new ResumeAction(guiDebugger, threadGetterToResume()),
             KeyEvent.VK_R);
        item(new WhereAction(guiDebugger, threadGetterToWhere()),
             KeyEvent.VK_W);
        add(menu);
    }

    void objectMenu() {
        menu("Objects", KeyEvent.VK_T);
        item(new DumpAction(guiDebugger),      KeyEvent.VK_D);
        item(new EvalAction(guiDebugger),      KeyEvent.VK_E);
        item(new FieldsAction(guiDebugger),    KeyEvent.VK_F);
        item(new LocalsAction(guiDebugger),    KeyEvent.VK_L);
        item(new PrintAction(guiDebugger),     KeyEvent.VK_P);
        add(menu);
    }

    void monitorMenu() {
        menu("Monitor", KeyEvent.VK_N);
        check("where");
        check("locals");
        check("list");
        add(menu);
    }

    void check(final String command) {
        boolean present = false;
        final Debugger debugger = guiDebugger.getDebugger();
        Iterator iter = debugger.getMonitorRequests().iterator();
        while (iter.hasNext()) {
            MonitorRequest request = (MonitorRequest) iter.next();
            if (request.getCommand().equals(command)) {
                present = true;
            }
        }
        final JCheckBoxMenuItem item =
            new JCheckBoxMenuItem(command, present);
        item.setMnemonic(mnemonic(command));
        item.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (item.isSelected()) {
                    try {
                        debugger.monitorCommand(command);
                    } catch (Exception ee) {
                        item.setSelected(false);
                    }
                } else {
                    unmonitorCommand(command);
                }
            }
            void unmonitorCommand(String _command) {
                Iterator _iter = debugger.getMonitorRequests().iterator();
                while (_iter.hasNext()) {
                    MonitorRequest request = (MonitorRequest)_iter.next();
                    if (request.getCommand().equals(_command)) {
                        try {
                            debugger.unmonitorCommand(request.getNumber());
                        } catch (Exception ee) {
                            item.setSelected(true);
                        }
                        return;
                    }
                }
            }
        });
        menu.add(item);
    }

    int mnemonic(String command) {
        int N = command.length();
        int M = menu.getItemCount();
        for (int i = 0; i < N; i++) {
            int c = (int) command.charAt(i);
            boolean there = false;
            for (int j = 0; j < M; j++) {
                if (menu.getItem(j).getMnemonic() == c) {
                    there = true;
                }
            }
            if (!there) {
                return c;
            }
        }
        return -1;
    }

    void helpMenu() {
        menu("Help",            KeyEvent.VK_T);
        add(Box.createHorizontalGlue());
        item(new HelpAction(guiDebugger),  KeyEvent.VK_F1);
        item(new AboutAction(guiDebugger), KeyEvent.VK_F2);
        add(menu);
    }

    ThreadGetter threadGetterToSuspend() {
        return new ThreadGetter() {
            public String getThread() {
                return Dialogs.showSuspendDialog();
            }
        };
    }

    ThreadGetter threadGetterToResume() {
        return new ThreadGetter() {
            public String getThread() {
                return Dialogs.showResumeDialog();
            }
        };
    }

    ThreadGetter threadGetterToWhere() {
        return new ThreadGetter() {
            public String getThread() {
                return Dialogs.showWhereDialog();
            }
        };
    }


    void sep() {
        menu.addSeparator();
    }


    void menu(String title, int mnemonic) {
        menu = new JMenu(title);
        menu.setMnemonic(mnemonic);
    }

    final static String jdk = "1.2";

    void item(AbstractActionStruct a, int key) {
        String title = a.getValue(LABEL) + "";
        int mnemonic = key;
        KeyStroke keyStroke = KeyStroke.getKeyStroke(a.key(), a.mask());
        if (jdk.equals("1.3")) {
            //item = new JMenuItem(a);
        } else {
            item = new JMenuItem((String) a.getValue(LABEL),
                                 (Icon)   a.getValue(ICON));
            item.addActionListener(a);
        }
        item.setText(title);
        item.setMnemonic(mnemonic);
        item.setAccelerator(keyStroke);
        menu.add(item);

    }

    public static void main(String[] args) {
        JFrame f = new JFrame("Menu Test");
        f.setJMenuBar(new AJMenuBar(null));
        f.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.out.println("Exiting...");
                System.exit(0);
        }});
        f.pack();
        f.setVisible(true);
    }

    public String toString() { return d(); }
    public static String d() { return "Menu Bar"; }
}
