/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import java.util.*;
import com.sun.jdi.*;
import org.aspectj.tools.ide.*;
import java.io.File;

public class AJStackFrameFormatter {

    public final static String INVALID      = "Invalid Frame";
    public final static String reception    = "$reception";
    public final static String execution    = "$execution";
    public final static String around       = "$around";

    public static MethodAndSource format(StackFrame frame) {
        String methodStr = "";
        String sourceStr = "";
        int lineNumber   = -1;
        String path      = "";

        try {
            if (frame == null) {
                return new MethodAndSource
                    ("<BAD FRAME>", AJLineMapper.UNKNOWN_SOURCE_STRING,
                     -1, path);
            }

            Location loc = frame.location();
            Method meth = loc.method();
            String decName = meth.declaringType().name();
            String methName = ComponentRepository.getAJDebugger().method(loc);

            // Get the path from the declaring type
            int idot = decName.lastIndexOf('.');
            if (idot != -1) {
                path = decName.substring(0, idot).replace('.', File.separatorChar);
            } else {
                path = "";
            }

            methodStr = decName + "." + methName;
            sourceStr = "";

            int locLine = ComponentRepository.getAJDebugger().lineNumber(loc);
            lineNumber = locLine;
            if (meth instanceof Method && ((Method)meth).isNative()) {
                sourceStr += "native method";
            } else if (locLine == AJLineMapper.NON_MAPPING_SOURCE_LINE) {
                try {
                    sourceStr =
                        ComponentRepository.getAJDebugger().
                        sourceName(loc);
                } catch (AbsentInformationException e) {
                    sourceStr += "<unknown>";
                }
            } else if (locLine != AJLineMapper.NO_SOURCE_LINE) {
                try {
                    sourceStr =
                        ComponentRepository.getAJDebugger().
                        sourceName(loc);
                } catch (AbsentInformationException e) {
                    sourceStr += "<unknown>";
                }
                lineNumber =
                    ComponentRepository.getAJDebugger().
                    lineNumber(loc);
            } else {
                sourceStr += "<unknown>";
            }
        } catch (InvalidStackFrameException e) {
            methodStr = "<THREAD RESUMED>";
            sourceStr = AJLineMapper.UNKNOWN_SOURCE_STRING;
        }
        return new MethodAndSource(methodStr, sourceStr, lineNumber, path);
    }

    public static class MethodAndSource {
        String method;
        String sourceName;
        int lineNumber = -1;
        String path;
        String shortSourceName;

        MethodAndSource(String method, String sourceName,
                        int lineNumber, String path) {
            this.method     = method;
            this.sourceName = sourceName;
            this.shortSourceName = shortSourceName(sourceName);
            this.lineNumber = lineNumber;
            this.path       = path;
        }

        public String getProto() {
            return getSourceName() + ":" + getLineNumber();
        }

        public String getSourceName() {
            return sourceName;
        }
        
        public int getLineNumber() {
            return lineNumber;
        }

        public boolean isValid() {
            return (lineNumber > -1);
        }

        public String getFullPath() {
            String pfx = "";
            if (!path.equals("") && !sourceName.startsWith(path)) {
                pfx = path + java.io.File.separatorChar;
            }
            return pfx + sourceName;
        }

        public String toString() {
            return method + " (" + shortSourceName +
                   (isValid() ? ":" + lineNumber : "") + ")";
        }

        private String shortSourceName(String sourceName) {
            int ilastSlash          = sourceName.lastIndexOf('\\');
            int ilastBackSlash      = sourceName.lastIndexOf('/');
            int ilastFileSeparator  = sourceName.lastIndexOf(File.separatorChar);
            int ilastSep = ilastSlash;
            if (ilastBackSlash > ilastSep) {
                ilastSep = ilastBackSlash;
            }
            if (ilastFileSeparator > ilastSep) {
                ilastSep = ilastFileSeparator;
            }
            if (ilastSep != -1) {
                return sourceName.substring(ilastSep+1);
            }
            return sourceName;
        }
    }

    static String formatMethod(Method method) {
        String decName = method.declaringType().name();
        String methName = method.name();
        String methString = method + "";
        String ajFile = AJLineMapper.getSourceFilePathFromAJCClass(decName);
        if (ajFile == null) {
            return methName;
        }

        // Need to strip the package off the method
        String classMethodString = methString;
        int i2dot = methString.indexOf("(");
        int numDots = 0;
        while (i2dot > -1) {
            if (methString.charAt(i2dot) == '.') {
                if ((++numDots) == 2) {
                    break;
                }
            }
            i2dot--;
        }
        if (i2dot != -1) {
            classMethodString = methString.substring(i2dot + 1);
        }
        int line = AJDecParser.getMethodLineNumber(ajFile, classMethodString);

        // Find the base name of this method
        int iDollarOne = methName.indexOf("$");
        String baseName = null;
        if (iDollarOne != -1) {
            baseName = methName.substring(0, iDollarOne);
        } else {
            baseName = methName;
        }

        String ext = "";
        String name = baseName;

        //
        // If this method is advice, we'll just use AJTranslator
        // Otherwise, there are only two mapping lines, and these labels are:
        //  - call:      Line that maps to a method dec. in the source.
        //  - body:      Line that does map to a method dec. in the source.
        //  - int. call: Line that doesn't map to a method dec. in the source.
        //
        Declaration dec = AJLineMapper.sm().getDeclarationAtLine(ajFile, line);
        if (AJTranslator.isAdvice(methName)) {
            name = AJTranslator.translateWorkingDirMethodName(methName);
        //XXX Work around because 214 in spacewar/Ship.java was mapping to
        //XXX a method call, when 212 was the real method line
        } else if (AJTranslator.isReception(methName)) {
            name += " [reception]";
        } else if (dec != null && dec.getKind().equals("method")) {
            name += " [call]";
        } else if (!AJTranslator.isNonMappingMethod(method)) {
            name += " [body]";
        } else {
            name += " [intermediate call]";
        }
        return name;
    }

    public static String format(StackFrame frame,
                                int index,
                                boolean showPC) {
        return format(frame, index, showPC, null);
    }

    public static String format(AJStackFrameNode node,
                                boolean showPC) {
        return format(node.getStackFrame(),
                      node.index()+1,
                      showPC,
                      node.getOriginalLocation());
    }

    private static String format(StackFrame frame,
                                 int index,
                                 boolean showPC,
                                 Location originalLocation) {
        String frameString = "";
        Location loc = null;
        int lineNum = -1;
        String sourceName = "<not available>";
        try {
            loc = frame.location();
            lineNum = loc.lineNumber();
            sourceName = loc.sourceName();
        } catch (InvalidStackFrameException e) {
            return INVALID + ": " + e.getMessage();
        }catch (AbsentInformationException e) {
            //return INVALID + ": " + e.getMessage();
        }

        Method meth = loc.method();
        frameString += "  [" + index + "] ";
        frameString += meth;
        frameString += " (";

        if (meth instanceof Method && ((Method)meth).isNative()) {
            frameString += "native method";
        } else {
            if (lineNum == 2 ) { // HACK: !!!
                frameString += "receiving";
            }
            else {
                frameString += new File(sourceName).getName();
            }
            if (lineNum != -1) {
                frameString += ':';
                frameString += lineNum; //newLineNum;
            }
        }
        frameString += ')';
        if (showPC) { //XX
            long pc = loc.codeIndex();
            if (pc != -1) {
                frameString += ", pc = " + pc;
            }
        }

        return frameString;
    }
}
