/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.request.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import java.util.List;
import java.util.Map;
import java.util.Iterator;
import com.sun.jdi.*;

public class AJStackFrameNode extends AJTreeNode {

    /**
     * The underlying StackFrame user object.
     */
    protected StackFrame frame = null;

    /**
     * The original location of this frame.
     */
    protected Location originalLocation = null;

    /**
     * The representation of the method and source.
     */
    protected AJStackFrameFormatter.MethodAndSource ms = null;

    private String str = "<unititialized frame>";

    /**
     * Constructs a new node with underlying user object <code>frame<\code>.
     * NEW VERSION: We don't need the index.
     * @param   frame   the underlying user object.
     */
    public AJStackFrameNode(StackFrame frame) {
        super(AJIcons.STACKFRAME_ICON);
        setUserObject(frame);
        //setFrame(frame); //XXX Must come before setting original location
        this.frame = frame;
        originalLocation = frame.location(); //TODO: Check for the exception
        ms = AJStackFrameFormatter.format(getStackFrame());
        str = ms + "";
        thread = frame.thread();
    }

    //XXX
    ThreadReference thread = null;

    // TODO: Deal with exception
    public int index() {
        int index = -1;
        try {
            index =
                ComponentRepository.getAJDebugger().
                frames(frame.thread()).indexOf(frame);
            //index = frame.thread().frames().indexOf(frame);
        } catch (Exception e) {}
        return index;
    }

    public void rightMouseButton(MouseEvent e) {
        JPopupMenu menu = null;
        if (ms.isValid()) {
            menu = new ShowMenu();
        } else {
            menu = new NoShowMenu();
        }
        menu.show(ComponentRepository.
                  getThreadGroupTreePane().getTree(),
                  e.getX(),
                  e.getY());
    }

    class ShowMenu extends JPopupMenu {
        ShowMenu() {
            //String s = AJStackFrameNode.this.ms.sourceName + ":" +
            String s = AJStackFrameNode.this.ms.getFullPath() + ":" +
                       AJStackFrameNode.this.ms.lineNumber;
            JMenuItem item;
            item = new JMenuItem("goto " + s);
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        ComponentRepository.getSourcePane().
                            showSourceForFileAndLine
                            (AJStackFrameNode.this.ms.getFullPath(),
                             AJStackFrameNode.this.ms.lineNumber);
                    }
                });
            add(item);
            Iterator iter =
                ComponentRepository.getAJDebugger().
                getBreakpointRequests().iterator();
            String cmd = " " + AJStackFrameNode.this.ms.getProto();
            boolean found = false;
            while (iter.hasNext()) {
                BreakpointRequestAction req =
                    (BreakpointRequestAction)iter.next();
                if (req.equals(AJStackFrameNode.this.ms)) {
                    found = true;
                    break;
                }
            }
            final String msg = (found ? "clear" : "stop on") + cmd;
            item = new JMenuItem(msg);
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        ComponentRepository.execute(msg);
                    }
                });
            add(item);
        }
    }

    static class NoShowMenu extends JPopupMenu {
        NoShowMenu() {
            add(new JMenuItem("Source is unavailable"));
        }
    }


    // ------------------------------------------------------------------------
    // Access to the user object.
    //
    // ASPECT: Do null-checking on these methods and other checking.
    // ------------------------------------------------------------------------

    /**
     * Gets the Value of a LocalVariable in this frame. The variable must be
     * valid for this frame's method and visible according to the rules
     * described in visibleVariables().
     *
     * @param   variable    the LocalVariable to be accessed
     * @return  the Value of the instance field.
     */
    public Value getValue(LocalVariable variable) {
        return frame.getValue(variable);
    }

    /**
     * Returns the values of multiple local variables in this frame. Each
     * variable must be valid for this frame's method and visible according to
     * the rules described in visibleVariables().
     *
     * @param   variables   a list of LocalVariable objects to be accessed
     * @return  a map associating each LocalVariable with its Value
     */
    public Map getValues(java.util.List variables) {
        return frame.getValues(variables);
    }

    /**
     * Returns the Location of the current instruction in the frame. The method
     * for which this frame was created can also be accessed through the
     * returned location. For the top frame in the stack, this location
     * identifies the next instruction to be executed. For all other frames,
     * this location identifies the instruction that caused the next frame's
     * method to be invoked. If the frame represents a native method invocation,
     * the returned location indicates the class and method, but the code index
     * will not be valid.
     *
     * @return  the Location of the current instruction.
     */
    public Location location() {
        return frame.location();
    }

    /**
     * Returns the value of 'this' for the current frame. The ObjectReference
     * for 'this' is only available for non-native instance methods.
     *
     * @return  an ObjectReference, or null if the frame represents a native or
     *          static method.
     */
    public ObjectReference thisObject() {
        return frame.thisObject();
    }

    /**
     * Returns the thread under which this frame's method is running.
     *
     * @return  a ThreadReference which mirrors the frame's thread.
     */
    public ThreadReference thread() {
        return frame.thread();
    }

    /**
     * Finds a LocalVariable that matches the given name and is visible at the
     * current frame location. See visibleVariables() for more information on
     * visibility.
     *
     * @param   name    the variable name to find.
     * @return  the matching LocalVariable, or null if there is no visible
     *          variable with the given name.
     * @throws  AbsentInformationException  if there is no line number
     *                                      information for this method.
     */
    public LocalVariable visibleVariableByName(String name)
            throws AbsentInformationException {
        return frame.visibleVariableByName(name);
    }

    /**
     * Returns a list containing each LocalVariable that can be accessed from
     * this frame's location.
     * <br>
     * Visibility is based on the code index of the current instruction of this
     * StackFrame. Each variable has a range of byte code indices in which it
     * is accessible. If this stack frame's method matches this variable's
     * method and if the code index of this StackFrame is within the variable's
     * byte code range, the variable is visible.
     * <br>
     * A variable's byte code range is at least as large as the scope of that
     * variable, but can continue beyond the end of the scope under certain
     * circumstances:
     * - the compiler/VM does not immediately reuse the variable's slot.
     * - the compiler/VM is implemented to report the extended range that would
     *   result from the item above.
     * The advantage of an extended range is that variables from recently
     * exited scopes may remain available for examination (this is especially
     * useful for loop indices). If, as a result of the extensions above, the
     * current frame location is contained within the range of multiple local
     * variables of the same name, the variable with the highest-starting range
     * is chosen for the returned list.
     *
     * @return  the list of LocalVariable objects currently visible.
     * @throws  AbsentInformationException  if there is no line number
     *                                      information for this method.
     */
    public List visibleVariables()
            throws AbsentInformationException {
        return frame.visibleVariables();
    }

    // ------------------------------------------------------------------------
    // Other misc. methods
    // ------------------------------------------------------------------------

    /**
     * Access to the user object StackFrame.
     *
     * @return  the StackFrame <code>frame<\code>.
     */
    public StackFrame getStackFrame() {
        return frame;
    }

    /**
     * Returns the line number of the location of this frame.
     *
     * @return  the line number of the location of this frame.
     */
    public int lineNumber() {
        return this.location().lineNumber();
    }

    /**
     * Returns the source file name for this frame.
     *
     * @return  the source file name for this frame.
     * @throws  AbsentInformationException  if there is no line number
     *                                      information for this method.
     */
    public String sourceName()
            throws AbsentInformationException {
        return this.location().sourceName();
    }

    /**
     * Gets the method, if any, containing this Location.
     *
     * @return  the Method if this location is in a method, otherwise null.
     */
    public Method method() {
        return this.location().method();
    }

    /**
     * Returns true because this <b>is<\b> a stack frame.
     *
     * @return  true.
     */
    public boolean isStackFrame() {
        return true;
    }

    /**
     * Checks to be sure that this frame is still valid.
     */
    public boolean isValid() {
        try {
            frame.location();
        } catch (InvalidStackFrameException ie) {
            return false;
        }
        return true;
    }

    public void showKids() {
        //if (getChildCount() > 0) return;
        try {
            removeAllChildren();
            showThisObject();
            showLocalVariables();
        } catch (InvalidStackFrameException isfe) {
        }
    }

    public void hideKids() {
        this.removeAllChildren();
    }

    /**
     * Shows the local variables in this frame.
     */
    private void showLocalVariables() {
        try {
            Iterator iter = visibleVariables().iterator();
            while (iter.hasNext()) {
                LocalVariable local = (LocalVariable) iter.next();
                StackFrame frame = getStackFrame();
                AJValueNode valueNode = AJValueNodeFactory.make(local, frame);
                //System.out.println("       * " + valueNode);
                this.add(valueNode);
            }
        } catch (AbsentInformationException aie) {
        } catch (NativeMethodException nme) {
        }
    }

    /**
     * Shows the ObjectReference <code>this<\code> in this frame.
     */
    private void showThisObject() {
        ObjectReference object = frame.thisObject();

        // The object will be null if the frame is static
        if (object == null) {
            return;
        }
        AJValueNode valueNode = AJValueNodeFactory.make(object, "this");
        this.add(valueNode);
    }

    /**
     * Overrides AJDBTreeNode.toString to return the <TODO>.
     */
    boolean sdb = false;
    public String toString() {
        String t = "thread=";
        try {
            t += System.identityHashCode(frame.thread()) + "";
        } catch (Exception e) {
            t += "***" + System.identityHashCode(thread);
        }
        t += ": ";

        return (sdb ? dd() + t : "") +
            str;
    }

    /**
     * First checks to see if this node it valid, if it is sets the icon to
     * the INVALID icon.
     */
    public int getType() {
        if (isValid()) {
            return super.getType();
        }

        // Tell the parent that this and all frames are crap
        DefaultMutableTreeNode parent = (DefaultMutableTreeNode)getParent();
        if (parent != null) {
            parent.removeAllChildren();
        }        
        return AJIcons.INVALID_ICON;
    }

    /**
     * Returns the original location of this frame.
     */
    public Location getOriginalLocation() {
        return this.originalLocation;
    }

    /**
     * Allows someone to reset the frame node, this should be done <b>only</b>
     * if the original location is the same as the frames location.
     * TODO: Either debug here or in the caller <i>or both</i>.
     * TODO: Exceptions.
     */
    public void setFrame(StackFrame frame) {
        //TODO: Testing to see that this frame is compatible with the
        // original location
        this.frame = frame;
        setUserObject(frame);
        originalLocation = frame.location(); //TODO: Check for the exception
        ms = AJStackFrameFormatter.format(getStackFrame());
        str = ms + "";
        thread = frame.thread();
        //
        // Possibly adding kids as soon as we get a frame will work
        //
        showKids();
    }
}
