/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import java.util.*;
import javax.swing.tree.*;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.ThreadGroupReference;

/**
 * An AJTreeNode with a ThreadGroupReference used object.  This node can
 * contain other AJThreadGroupNode instances or AJThreadNode
 * instances.  Access is provided to the underlying ThreadGropuReference, so
 * it is not adviced to access the actual reference directly.
 */
public class AJThreadGroupNode extends AJTreeNode {

    /**
     * The underlying ThreadGroupReference.
     */
    ThreadGroupReference threadGroup =  null;

    /**
     * Sets the ThreadGroupReference.
     *
     * @param   threadGroup the underlying ThreadGroupReference.
     */
    public AJThreadGroupNode(ThreadGroupReference threadGroup, DefaultTreeModel model) {
        super(AJIcons.THREADGROUP_ICON, model);
        this.threadGroup = threadGroup;
        this.setUserObject(threadGroup);
    }

    // ------------------------------------------------------------------------
    // Interface to the underlying ThreadGroupReference
    // ------------------------------------------------------------------------

    /**
     * Returns the name of this thread group.
     *
     * @return  the string containing the thread group name.
     */
    public String name() {
        return threadGroup.name();
    }

    /**
     * Returns the parent of this thread group.
     *
     * @return  a ThreadGroupReference mirroring the parent of
     *          this thread group in the target VM, or null if this
     *          is a top-level thread group.
     */
    public ThreadGroupReference parent() {
        return threadGroup.parent();
    }

    /**
     * Resumes all threads in this thread group. Each thread in
     * this group and in all of its subgroups will be resumed as
     * described in ThreadReference.resume().
     */
    public void resume() {
        threadGroup.resume();
    }

    /**
     * Suspends all threads in this thread group. Each thread in this
     * group and in all of its subgroups will be suspended as described
     * in ThreadReference.suspend(). This is not guaranteed to be an atomic
     * operation; if the target VM is not interrupted at the time this method
     * is called, it is possible that new threads will be created between
     * the time that threads are enumerated and all of them have been
     * suspended.
     */
    public void suspend() {
        threadGroup.suspend();
    }

    /**
     * Returns a List containing each ThreadGroupReference in this thread
     * group. Only the thread groups in this immediate thread group (and
     * not its subgroups) are returned.
     *
     * @return  a List of ThreadGroupReference objects mirroring the threads
     *          from this thread group in the target VM.
     */
    public List threadGroups() {
        return threadGroup.threadGroups();
    }

    /**
     * Returns a List containing each ThreadReference in this thread group.
     * Only the threads in this immediate thread group (and not its
     * subgroups) are returned.
     *
     * @return  a List of ThreadReference objects mirroring the threads
     *          from this thread group in the target VM.
     */
    public List threads() {
        return threadGroup.threads();
    }

    // ------------------------------------------------------------------------
    // Misc. methods
    // ------------------------------------------------------------------------

    /**
     * Returns the underlying ThreadReference.
     *
     * @return  The underlying ThreadReference.
     */
    public ThreadGroupReference getThreadGroup() {
        return threadGroup;
    }

    /**
     * Overrides isThreadGroup of AJTreeNode because this <b>is</b> a
     * thread group.
     *
     * @return  true.
     */
    public boolean isThreadGroup() {
        return true;
    }

    public String toString() {
        return threadGroup.name() + " (" + threadGroup.uniqueID() + ")";
    }

    public boolean isOKToExpand() {
        return true;
    }
}
