/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import java.awt.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.tree.*;
import com.sun.jdi.*;

//TODO: For debugging
import javax.swing.ToolTipManager;

/**
 * An AJTree that is used to display AJTheadGroupTreeNode instances
 * at its top level.  These thread groups may either contain AJThreadTreeNode
 * instances or other AJThreadGroupTreeNode instances.  See the
 * AJThreadTreeNode notes for information on what they can contain.
 */
public class AJThreadGroupTree extends AJTree
        implements  DropTargetListener,
                    DragSourceListener,
                    DragGestureListener,
                    java.io.Serializable
{

    public /*private*/ Map threadNodes = new HashMap();

    /**
     * DND stuff.
     */
    DropTarget dropTarget = null;
    DragSource dragSource = null;

    /**
     * Create a new tree with a model, renderer, and root for that of
     * thread groups.
     */
    public AJThreadGroupTree() {
        super();
        AJRootNode          root   = root();
        AJTreeModel         model  = new AJTreeModel(root);
        AJTreeCellRenderer  rend   = new AJTreeCellRenderer();
        setRoot(root);
        setModel(model);
        setCellRenderer(rend);
        addMouseListener(new AJThreadGroupTreeMouseListener(this));
        addTreeWillExpandListener(new NoCollapsingRootExpansionListener(this));

        // DND stuff
        dropTarget = new DropTarget(this, this);
        dragSource = new DragSource();
        dragSource.createDefaultDragGestureRecognizer
            (this, DnDConstants.ACTION_MOVE, this);
    }

    /**
     * Adds a new ThreadReference <code>ref<\code> to the tree by first
     * constructing a List <code>list<\code> of the path, and then adding the
     * list to the root of the tree.
     *
     * @param   ref the ThreadReference that will by the user object of
     *          the target AJThreadTreeNode.
     * @return  whether <code>ref<\code> was inserted.
     */
    public boolean add(ThreadReference ref) {
        boolean result = false;
        try {
            if (threadNodes.containsKey(ref)) {
                return false;
            }
            MutableTreeNode treeNode = addToRoot(ref);
            result = true;
        } catch (Exception e) { Util.ex(e); }
        return result;
    }

    private MutableTreeNode addToRoot(ThreadReference ref) {
        List list = findParent(ref);
        boolean expand = getRoot().getChildCount() == 0;
        AJTreeNode newNode = getRoot().add(list);
        if (expand) expandRow(0);
        threadNodes.put(ref, newNode);
        return newNode;
    }

    public void expandToFrame(StackFrame frame) {
        TreePath path = pathToFrame(frame);
        expand(path);
    }

    private void expand(TreePath path) {
        if (path != null) {
            Object o = path.getLastPathComponent();
            TreePath newPath = null;
            if (o instanceof AJThreadNode) {
                AJThreadNode threadNode = (AJThreadNode)o;
                threadNode.showKids();
                int i = threadNode.getChildCount();
                if (i > 0) {
                    newPath = path.pathByAddingChild((AJTreeNode)
                                                     threadNode.
                                                     getChildAt(i-1));
                }
            }
            System.out.println("expand path:" + path);
            scrollPathToVisible(newPath != null ? newPath : path);
            setSelectionPath(path);

        }
    }

    AJThreadNode threadNode(long id) {
        Enumeration enum =
            ((DefaultMutableTreeNode)getModel().getRoot()).
            breadthFirstEnumeration();
        while (enum.hasMoreElements()) {
            Object next = enum.nextElement();
            if (next instanceof AJThreadNode) {
                AJThreadNode threadNode = (AJThreadNode)next;
                if (threadNode.getId() == id) {
                    return threadNode;
                }
            }
        }
        return null;
    }



    private TreePath pathToFrame(StackFrame frame) {
        TreePath path = null;
        try {
            ThreadReference ref = frame.thread();
            long id = ref.uniqueID();
            AJThreadNode node = threadNode(id);
            if (node != null) {
                TreeNode[] nodes = node.getPath();
//                  if (node.getChildCount() > 0) {
//                      AJTreeNode frameNode = (AJTreeNode)node.getChildAt(0);
//                      path = new TreePath(nodes).pathByAddingChild(frameNode);
//                  } else {
                    path = new TreePath(nodes);
//                  }
            }
        } catch (NullPointerException npe) {
        }
        return path;
    }

    public void expandToLocals(StackFrame frame) {
        TreePath path = pathToFrame(frame);
        if (path == null) {
            return;
        }
        AJTreeNode node = (AJTreeNode) path.getLastPathComponent();
        if (node.getChildCount() > 0) {
            path = path.pathByAddingChild(node.getChildAt(0));
        }
        expand(path);
    }

    public void expandToValue(Value value, String name) {
        Enumeration nodes = ((AJRootNode) getModel().getRoot()).breadthFirstEnumeration();
        if (nodes == null) {
            return;
        }
        AJTreeNode node = null;
        while (nodes.hasMoreElements()) {
            node = (AJTreeNode) nodes.nextElement();
            if (!(node instanceof AJValueNode)) {
                continue;
            }
            AJValueNode vnode = (AJValueNode) node;
            Value val = vnode.getValue();
            if (val != null) {
                if (val.equals(value) && vnode.getName().equals(name)) {
                    break;
                }
            }
        }
        if (node == null) {
            return;
        }
        TreePath path = new TreePath(node.getPath());
        if (node.getChildCount() > 0) {
            path = path.pathByAddingChild(node.getChildAt(0));
        }
        expand(path);
    }

    /**
     * Removes the ThreadReference <code>ref<\code> from the tree using the
     * same reasoning used in adding it.
     *
     * @param   ref the ThreadReference to remove.
     * @return  whether <code>ref<\code> was removed.
     */
    public boolean remove(ThreadReference ref) {
        if (!threadNodes.containsKey(ref)) {
            return false;
        }
        List list = findParent(ref);
        AJTreeNode removedNode = this.getRoot().remove(list);
        threadNodes.remove(removedNode);
        modernize();
        return true;
    }

    /**
     * Validates all the threads in the tree.  Calls validate on all the thread
     * nodes.
     */
    public void validateThreadNodes() {
        try {
            Iterator iter = threadNodes.values().iterator();
            while (iter.hasNext()) {
                AJThreadNode node = (AJThreadNode) iter.next();
                AJTreeModel model = (AJTreeModel) getModel();
                node.validate(model);
                model.nodeStructureChanged(node);
            }
            reloadAll();
        } catch (Exception e) {
            if (!Debugger.demoMode) {
                System.err.println("VALIDATE ----------------------------");
                System.err.flush();
                e.printStackTrace(System.err);
            }
        }
    }

    public void invalidateAllThreads () {
        try {
            Iterator iter = threadNodes.values().iterator();
            while (iter.hasNext()) {
                AJThreadNode node = (AJThreadNode) iter.next();
                AJTreeModel model = (AJTreeModel) getModel();
                node.invalidate();
                model.nodeStructureChanged(node);
            }
            reloadAll();
        } catch (Exception e) {
            if (!Debugger.demoMode) {
                System.err.println("VALIDATE ----------------------------");
                System.err.flush();
                e.printStackTrace(System.err);
            }
        }
    }

    public void doModernize() {
        try {
            validateThreadNodes();
        } catch (ThreadDeath td) {
        } catch (VMDisconnectedException vmde) {
        }
    }

    /**
     * Returns a java.util.List containing a path of AJTreeNode instances
     * with <code>ref<\code> at index 0.  For example, if <code>ref<\code> was
     * in a thread group B, and B was in a thread group A, the list returned
     * would look like: <code>[ref,B,A]<\code>.
     *
     * @param   ref the ThreadReference that will be the leaf in the returned
     *              list.
     * @return  a java.util.List containg the path to <code>ref<\code>.
     */
    private List findParent(ThreadReference ref) {
        ArrayList list = new ArrayList();
        int i = 0;
        list.add(i++, new AJThreadNode(ref, (DefaultTreeModel)getModel()));
        ThreadGroupReference parent = ref.threadGroup();
        while (parent != null) {
            list.add(i++,
                     new AJThreadGroupNode(parent,
                                           (DefaultTreeModel)getModel()));
            parent = parent.parent();
        }
        return list;
    }

    /**
     * Clears the tree using the same procedure used in constructing it.
     */
    public void clear() {
        AJRootNode          root   = root();
        AJTreeModel         model  = new AJTreeModel(root);
        AJTreeCellRenderer  rend   = new AJTreeCellRenderer();
        setRoot(root);
        setModel(model);
        setCellRenderer(rend);
    }

    private AJRootNode root() {
        return new AJRootNode("Threads", AJIcons.THREADS_ICON);
    }

    // ------------------------------------------------------------------------
    // DND Stuff
    // ------------------------------------------------------------------------

    //
    // ASPECT: Ensuring drags from valid sources and drops on correct targets
    //

    /*
     * DropTargetListener
     */
    public void dragEnter(DropTargetDragEvent e) {}
    public void dragExit(DropTargetEvent e) {}
    public void dragOver(DropTargetDragEvent e) {
        Point point = e.getLocation();
        int x = (int) point.getX();
        int y = (int) point.getY();
        int row = getRowForLocation(x, y);
        TreePath path = getPathForLocation(x, y);
        if (row == -1) {
            return;
        }
        AJTreeNode node = (AJTreeNode) path.getLastPathComponent();
        //TODO: Make this more robust
        if (!node.isValueNode()) {
            e.rejectDrag();
        } else {
            e.acceptDrag(DnDConstants.ACTION_MOVE);
        }
    }
    public void dropActionChanged(DropTargetDragEvent e) {}
    public void drop(DropTargetDropEvent e) {
        e.acceptDrop(DnDConstants.ACTION_MOVE);
        Point point = e.getLocation();
        int x = (int) point.getX();
        int y = (int) point.getY();
        int row = getRowForLocation(x, y);
        TreePath path = getPathForLocation(x, y);
        if (row == -1) {
            return;
        }
        AJTreeNode node = (AJTreeNode) path.getLastPathComponent();

        try {
            Value newValue =
                ((AJValueWrapper) e.getTransferable().
                    getTransferData
                        (AJDataFlavor.FLAVORS[AJDataFlavor.AJ_VALUE_NODE])).
                            unwrap();
            AJValueNode valueNode = (AJValueNode) node;
            valueNode.setValue(newValue);
            ((AJTreeModel) getModel()).reload(valueNode);
            e.getDropTargetContext().dropComplete(true);
        } catch (NullPointerException npe) {
            AJUtil.warn(npe);
        } catch (Exception ee) { AJUtil.ex(ee); }
    }

    /*
     * DragSourceListener
     */
    public void dragDropEnd(DragSourceDropEvent e) {}
    public void dragEnter(DragSourceDragEvent e) {}
    public void dragExit(DragSourceEvent e) {}
    public void dragOver(DragSourceDragEvent e) {}
    public void dropActionChanged(DragSourceDragEvent e) {}

    /*
     * DragGestureListener
     */
    public void dragGestureRecognized(DragGestureEvent e) {
        Point point = e.getDragOrigin();
        int x = (int) point.getX();
        int y = (int) point.getY();
        int row = getRowForLocation(x, y);
        TreePath path = getPathForLocation(x, y);
        if (row == -1) {
            return;
        }
        AJTreeNode node = (AJTreeNode) path.getLastPathComponent();
        AJValueWrapper trans = null;
        if (node.isValueNode()) {
            trans = new AJValueWrapper( ((AJValueNode) node).getValue() );
        }
        dragSource.startDrag(e, DragSource.DefaultMoveDrop, trans, this);
    }
}
