/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import java.awt.Color;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.util.*;
import java.util.List;
import com.sun.jdi.*;

/**
 * An AJTreeNode that has a ThreadReference as its user object.  These nodes
 * may contain instances of AJStackFrameNode.
 *
 * ASPECT:  1. Suspend, resume, block...
 *          2. Popup menu
 */
public class AJThreadNode extends AJTreeNode {

    /**
     * The underlying ThreadReference user object.
     */
    ThreadReference thread = null;

    /**
     * The long id of this thread.
     */
    private long id = -1;

    /**
     * Construct a new node with user object <code>thread<\code>.
     *
     * @param   thread  new user object.
     */
    public AJThreadNode(ThreadReference thread, DefaultTreeModel model) {
        super(AJIcons.THREAD_ICON, model);
        this.thread = thread;
        this.setUserObject(thread);
        this.id = thread.uniqueID();
    }

    /**
     * Returns the id.
     */
    public long getId() {
        return id;
    }

    public void showKids() {
        try {
            showFrames();
        } catch (InvalidStackFrameException isfe) {
        }
    }

    public void hideKids() {
        hideFrames();
    }

    public boolean isOKToExpand() {
        if (!thread.isSuspended() && !thread.isAtBreakpoint()) {
            return false;
        }
        return super.isOKToExpand();
    }

    /**
     * Called when we need to check all frames.
     * Any frames that have been around since this thread resumed will be
     * invalid.  In that case, we need to try to replace that frame.  If we
     * cannot replace it we must remove it.  We then have to have all the new
     * valid frames show themselves.
     */
    public void validate() {
        showFrames();
    }

    public void invalidate() {
        hideFrames();
    }

    public void validate(AJTreeModel model) {
        showFrames(model);
    }

    private boolean locationsAreEqual(Location l1, Location l2) {
        try {
            String type1 = l1.declaringType().toString();
            String type2 = l2.declaringType().toString();
            String meth1 = l1.method().toString();
            String meth2 = l2.method().toString();
            boolean result = ( type1.equals(type2) && meth1.equals(meth2) );
            return result;
        } catch (Exception e) {}
        return false;
    }

    private boolean framesAreEqual(StackFrame f1, StackFrame f2) {
        try {
            String m1 = f1.location().method().toString();
            String m2 = f2.location().method().toString();
            return m1.equals(m2);
        } catch (Exception e) {}
        return false;
    }

    /**
     * Checks to see if all the frames are valid.
     */
    private boolean framesAreValid() {
        Enumeration enum = this.children();
        while (enum.hasMoreElements()) {
            if (!isValidFrame((AJStackFrameNode) enum.nextElement())) {
                return false;
            }
        }
        return true;
    }

    /**
     * Checks if the passed in AJStackFrameNode has a valid underlying frame.
     */
    private boolean isValidFrame(AJStackFrameNode node) {
        try {
            node.location();
        } catch (InvalidStackFrameException e) {
            return false;
        }
        return true;
    }

    private void showFrames(DefaultTreeModel model) {
        try {
            Iterator iter = frames().iterator();
            removeAllChildren();
            int i = 0;
            while (iter.hasNext()) {
                AJStackFrameNode node = new AJStackFrameNode((StackFrame)iter.next());
                //model.insertNodeInto(node, this, i++);
                add(node);
                node.showKids();
                //model.nodeStructureChanged(node);
            }
        } catch (IncompatibleThreadStateException itse) {
        }
        model.reload(this);
    }
 
    private void showFrames() {
        showFrames(getModel());
//          boolean showMain = false && name().equals("main");
//          if (showMain) {
//              System.out.println("showFrames... " + thread.isSuspended());
//          }
//          try {
//              Iterator iter = frames().iterator();

//              removeAllChildren();
//              int i = 0;
//              int count = getChildCount();
//              while (iter.hasNext()) {
//                  StackFrame frame = (StackFrame) iter.next();
//                  AJStackFrameNode node = new AJStackFrameNode(frame);
//                  add(node);
//                  node.showKids();
//              }
//          } catch (IncompatibleThreadStateException itse) {
//          }
    }

    /**
     * Removes all frames.  Currently just removes all the children.
     */
    private void hideFrames() {
        try {
            this.removeAllChildren();
        } catch (Exception e) {}
    }

    public void rightMouseButton(MouseEvent e) {
        JPopupMenu menu = null;
        if (thread.isAtBreakpoint()) {
            menu = new ContPopupMenu();
        } else if (thread.isSuspended()) {
            menu = new ResumePopupMenu();
        } else {
            menu = new SuspendPopupMenu();
        }
        menu.show(ComponentRepository.getThreadGroupTreePane().getTree(), e.getX(), e.getY());
    }

    abstract class TwoActionPopupMenu extends JPopupMenu {
        TwoActionPopupMenu() {
            final String command = getCommand();
            char[] chars = command.toCharArray();
            chars[0] = Character.toUpperCase(chars[0]);
            String upperCommand = new String(chars);
            final String id = AJThreadNode.this.getThread().uniqueID() + "";
            JMenuItem item1 = new JMenuItem(upperCommand + " thread #" + id);
            item1.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    ComponentRepository.getCommandLine().executeCommand(command + " " + id);
                    ComponentRepository.getThreadGroupTreePane().modernizeTree();
            }});
            JMenuItem item2 = new JMenuItem(upperCommand + " all");
            item2.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    ComponentRepository.getCommandLine().executeCommand(command);
                    ComponentRepository.getThreadGroupTreePane().modernizeTree();
            }});

            this.add(item1);
            this.add(item2);
        }
        abstract String getCommand();
    }

    private class ResumePopupMenu extends TwoActionPopupMenu {
        String getCommand() { return "resume"; }
    }

    private class SuspendPopupMenu extends TwoActionPopupMenu {
        String getCommand() { return "suspend"; }
    }

    private class ContPopupMenu extends JPopupMenu {
        ContPopupMenu() {
            JMenuItem item = new JMenuItem("Continue");
            item.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    ComponentRepository.getCommandLine().executeCommand("cont");
            }});
            this.add(item);
        }
    }

    // ------------------------------------------------------------------------
    // Access to the underlying ThreadReference.
    // This is just a selected number of the access methods, they may
    // turn out to be useless, but we'll see.
    //
    // ASPECT: Possibly a good place to check for null with aspects
    // ------------------------------------------------------------------------

    /**
     * Returns a List containing each StackFrame in the thread's current
     * call stack. The thread must be suspended (normally through an
     * interruption to the VM) to get this information, and it is only
     * valid until the thread is resumed again.
     *
     * @return  a List of StackFrame with the current frame first followed
     *          by each caller's frame.
     * @throws  IncompatibleThreadStateException - if the thread is not
     *          suspended in the target VM
     *          ObjectCollectedException - if this object has been garbage
     *          collected.
     */
    public List frames() throws IncompatibleThreadStateException {
        return ComponentRepository.getAJDebugger().frames(thread);
//        return thread.frames();
    }

    /**
     * Interrupts the thread.
     */
    public void interrupt() {
        thread.interrupt();
    }

    /**
     * Returns the name of this thread.
     *
     * @return  the string containing the thread name.
     */
    public String name() {
        return thread.name();
    }

    /**
     * Resumes this thread. If this thread was not previously suspended
     * through suspend() or through VirtualMachine.suspend(), invoking this
     * method has no effect. Otherwise, the count of pending suspends on this
     * thread is decremented. If it is decremented to 0, the thread will
     * continue to execute.
     */
    public void resume() {
        thread.resume();
    }

    /**
     * Returns the thread's status. If the thread is not suspended the
     * thread's current status is returned. If the thread is suspended,
     * the thread's status before the suspension is returned (or
     * THREAD_STATUS_UNKNOWN if this information is not available.
     * isSuspended() can be used to determine if the thread has been suspended.
     *
     * @return  one of THREAD_STATUS_UNKNOWN, THREAD_STATUS_ZOMBIE,
     *          THREAD_STATUS_RUNNING, THREAD_STATUS_SLEEPING,
     *          THREAD_STATUS_MONITOR, THREAD_STATUS_WAIT,
     *          THREAD_STATUS_NOT_STARTED.
     */
    public int status() {
        return thread.status();
    }

    /**
     * Suspends this thread. The thread can be resumed through resume() or
     * resumed with other threads through VirtualMachine.resume().
     * Unlike java.lang.Thread#suspend, suspends of both the virtual machine
     * and individual threads are counted. Before a thread will run again, it
     * must be resumed (through resume() or resume()) the same number of times
     * it has been suspended.
     * Suspending single threads with this method has the same dangers as
     * java.lang.Thread.suspend(). If the suspended thread holds a monitor
     * needed by another running thread, deadlock is possible in the target VM
     * (at least until the suspended thread is resumed again). The suspended
     * thread is guaranteed to remain suspended until resumed through one of
     * the JDI resume methods mentioned above; the application in the target VM
     * cannot resume the suspended thread through java.lang.Thread#resume.
     */
    public void suspend() {
        thread.suspend();
    }

    // ------------------------------------------------------------------------
    // Other misc. methods
    // ------------------------------------------------------------------------

    /**
     * Returns the underlying ThreadReference.
     *
     * @return  the underlying ThreadReference.
     */
    public ThreadReference getThread() {
        return thread;
    }

    /**
     * Sets the underlying thread.
     */
    public void setThread(ThreadReference thread) {
        this.thread = thread;
        setUserObject(thread);
    }

    /**
     * True, because this <b>is<\b> a thread.
     *
     * @return  <code>true<\code>.
     */
    public boolean isThread() {
        return true;
    }

    /**
     * Overrides AJTreeNode.getType to see if this guy is suspended,
     * resumed, or at a breakpoint.
     *
     * @return
     */
    public int getType() {
        if (thread.isAtBreakpoint()) {
            return AJIcons.THREAD_BREAKPOINT_ICON;
        } else if (thread.isSuspended()) {
            return AJIcons.THREAD_SUSPENDED_ICON;
        } else {
            return AJIcons.THREAD_RUNNING_ICON;
        }
    }

    /**
     * Returns the <code>name<\code> and <code>uniqueID<\code>.
     *
     * @return the <code>name<\code> and <code>uniqueID<\code>.
     */
    boolean tdb = false;
    public String toString() {
        String t = "";
        try {
            t = "[ ";
            Object[] frames = frames().toArray();
            for (int i = 0; i < frames.length; i++) {
                t += System.identityHashCode((StackFrame) frames[i]) + " ";
            }
            t += " ]:";
            t += "bp=" + thread.isAtBreakpoint() + ": ";
            t += "su=" + thread.isSuspended() + ": ";
        } catch (Exception e) {
        }
        return (tdb ? dd() + t : "") +
            AJThreadFormatter.format(thread);
    }

    private String other() {
        if (thread.isAtBreakpoint()) return "at breakpoint";
        else if (thread.isSuspended()) return "suspended";
        else return "running";
    }

    /**
     * Maps a threads's <code>int<\code> status to a <code>String<\code>.
     *
     * @return  a String rep of the thread's status.
     */
    private String statusString() {
        switch(status()) {
            case ThreadReference.THREAD_STATUS_MONITOR:
                return "monitor";
            case ThreadReference.THREAD_STATUS_NOT_STARTED:
                return "not started";
            case ThreadReference.THREAD_STATUS_RUNNING:
                return "running";
            case ThreadReference.THREAD_STATUS_SLEEPING:
                return "sleeping";
            case ThreadReference.THREAD_STATUS_UNKNOWN:
                return "unkown";
            case ThreadReference.THREAD_STATUS_WAIT:
                return "waiting";
            case ThreadReference.THREAD_STATUS_ZOMBIE:
                return "zombie";
        }
        return "?";
    }
}

class AJThreadFormatter {
    public static String format(ThreadReference thread) {
        if (thread == null) {
            return "<bad thread>";
        }
        return thread.name() + " (" + thread.uniqueID() + ")"; // + ":" + thread.suspendCount();
    }
}
