/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.ide.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import com.sun.jdi.*;
import com.sun.jdi.event.*;


public class AJToolBar
    extends JToolBar
    implements AJActionsConsts,
               ThreadListener,
               VMListener {

    JComboBox           comboBox        = comboBox();
    JToolBar            toolBar         = new JToolBar();
    final static String ALL_THREADS     = "All threads";
    final static int    FONT_SIZE       = 10;
    final static int    SC_FONT_SIZE    =  8;
    final static double WIDTH_FACTOR    = .7;
    final static double HEIGHT_FACTOR   = .4;
    private AJDebugger debugger;

    public AJToolBar(GUIDebugger guid) {

        this.debugger = guid.getDebugger();
        debugger.addThreadListener(this);
        debugger.addVMListener(this);

        String mode = Modes.getMode();
        if (mode.equals(Modes.GUI)) {
            createGuiToolbar(guid);
//        } 
//        else if (mode.equals(Modes.JBUILDER) 
//                   || mode.equals(Modes.JBUILDER3)
//                   || mode.equals(Modes.JBUILDER4)) {
//            createJBuilderToolbar((JBuilderGUIDebugger)guid);
        } else {
            createForteToolbar(guid);
        }
    }

    private void createForteToolbar(GUIDebugger guid) {
        badd(new ForteRunAction(guid));
        badd(new ContAction(guid));
        //!!! attach
        badd(new QuitAction(guid));
        badd(new SuspendAction(guid, threadGetter()));
        badd(new ForteResumeAction(guid, threadGetter()));
        badd(new StepAction(guid));
        badd(new StepIAction(guid));
        badd(new NextAction(guid));
        badd(new StepUpAction(guid));
        //!!! add/remove
        //badd(new BreakpointAction(guid));
        badd(new IDELineBreakpointAction(guid), "Add Breakpoint");
        badd(new WatchAction(guid));
    }

//    private void createJBuilderToolbar(JBuilderGUIDebugger guid) {
//        badd(new JBuilderQuitAction(guid), "Reset Program");
//        badd(new JBuilderRunAction(guid), "Restart Program", "Cont", guid);
//        badd(new SuspendAction(guid, threadGetter()), "Pause Program");
//        jbSeparator();
//        //smart step: only in professional edition
//        badd(new NextAction(guid), "Step Over");
//        badd(new StepAction(guid), "Step Into");
//        badd(new StepUpAction(guid), "Step Out");
//        badd(new IDELineBreakpointAction(guid), "Add Breakpoint");
//        badd(new WatchAction(guid), "Add Watch");
//        badd(new WhereAction(guid, threadGetter()), "Show Current Frame");
//        add(new CommandLinePanel(guid));
//    }

    private void jbSeparator() {
        addSeparator();
    }

     private void createGuiToolbar(GUIDebugger guid) {

         // Main
         badd(new QuitAction(guid));
         badd(new UseAction(guid));
         badd(new DisconnectAction(guid));
         badd(new RunAction(guid));
         addSeparator();

         // Breakpoints
         badd(new BreakpointAction(guid));
         badd(new CatchAction(guid));
         badd(new ClearAction(guid));
         badd(new ContAction(guid));
         badd(new NextAction(guid));
         badd(new StepAction(guid));
         badd(new StepUpAction(guid));
         badd(new StepIAction(guid));
         addSeparator();

        // Threads
        badd(new SuspendAction(guid, threadGetter()));
        badd(new ResumeAction(guid, threadGetter()));
        badd(new WhereAction(guid, threadGetter()));
        comboBox = comboBox();
        add(comboBox);
        addSeparator();

        // Objects
        badd(new DumpAction(guid));
        badd(new EvalAction(guid));
        badd(new FieldsAction(guid));
        badd(new LocalsAction(guid));
        badd(new PrintAction(guid));

        add(Box.createHorizontalGlue());
    }

    public AJDebugger getDebugger() {
        return debugger;
    }

    /****************************** ThreadListener ******************************/

    public void threadDeathEvent(ThreadDeathEvent e) {
        removeThread(e.thread());
    }

    public void threadStartEvent(ThreadStartEvent e) {
        addThread(e.thread());
    }

    /****************************** VMListener ******************************/

    public void vmDeathEvent(VMDeathEvent e) {
        clearThreads();
    }

    public void vmDisconnectEvent(VMDisconnectEvent e) {
        clearThreads();
    }

    public void vmStartEvent(VMStartEvent e) {
        clearThreads();
        try {
            Iterator iter = getDebugger().vm().allThreads().iterator();
            while (iter.hasNext()) {
                addThread((ThreadReference) iter.next());
            }
        } catch (Exception ee) {}
    }

    JComboBox comboBox() {
        if (comboBox == null) {
            comboBox = new JComboBox();
        }
        comboBox.addItem(new ThreadPair(ALL_THREADS, -1));
        comboBox.setPreferredSize(new Dimension(100, 30));
        return comboBox;
    }

    ThreadGetter threadGetter() {
        return new ThreadGetter() {
            public String getThread() {
                AJToolBar.ThreadPair pair =
                    (AJToolBar.ThreadPair) comboBox.getSelectedItem();
                if (pair.id != -1) {
                    return pair.id + "";
                }
                return "";
            }
        };
    }

    private void addToT(JComponent c) {
        toolBar.add(c);
    }

    private void addT(String title) {
        toolBar.setToolTipText(title);
        toolBar.setName(title);
        add(toolBar);
        toolBar = new JToolBar();
    }

    private void badd(AbstractActionStruct a) {
        badd(a, null);
    }


    private void badd(AbstractActionStruct a, String toolTipText) {
        badd(a, toolTipText, null, null);
    }

    private void badd(AbstractActionStruct a, String toolTipText, final String toolTipText2,
                      final GUIDebugger guid) {
        Icon icon = (Icon) a.getValue(ICON);
        JButton b;
        if (guid == null) {
            b = new JButton(icon);
        } else {
            b = new JButton(icon) {
                    public String getToolTipText() {
                        return guid.getDebugger().isRunning() ? toolTipText2 : super.getToolTipText();
                    }
                };
        }
        String name = (String) a.getValue(LABEL);
        JLabel label = new JLabel(name);
        String shortCutStr = KeyEvent.getKeyModifiersText(a.mask()) + "-" +
                             KeyEvent.getKeyText(a.key());
        JLabel shortCutLabel = new JLabel(shortCutStr);
        Font f = label.getFont();
        label.setFont(new Font(f.getName(), f.getStyle(), SC_FONT_SIZE));
        shortCutLabel.setFont(new Font(f.getName(), f.getStyle(), FONT_SIZE));
        b.addActionListener(a);
        if (toolTipText == null) {
            b.setToolTipText(name + " [" + shortCutStr + "]");
        } else {
            b.setToolTipText(toolTipText);
        }
        forward(b);
        JPanel p = new JPanel();
        p.setLayout(new BorderLayout());
        p.add(b, BorderLayout.CENTER);
        forward(p);
        changeDim(p);
        add(p);
    }

    void forward(JComponent c) {
        c.setNextFocusableComponent(ComponentRepository.getCommandLine());
    }

    void changeDim(JComponent b) {
        Dimension d = b.getPreferredSize();
        double w = d.getWidth();
        double h = d.getHeight();
        double wp = WIDTH_FACTOR * w;
        double hp = HEIGHT_FACTOR * h;
        Dimension newD = new Dimension((int) wp,(int) h);
        b.setMaximumSize(newD);
        b.setMinimumSize(newD);
        b.setPreferredSize(newD);
    }

    private Action action(final String str, int icon) {
        return new AbstractAction(str, AJIcons.getIcon(icon)) {
            public void actionPerformed(ActionEvent e) {
                execute(str);
            }
        };
    }

    void execute(String str) {
        CommandLine cl = ComponentRepository.getCommandLine();
        if (cl != null) {
            cl.executeCommand(str);
        }
    }


    void clearThreads() {
        if (comboBox != null) {
            comboBox.removeAllItems();
        }
        comboBox = comboBox();
    }

    void addThread(ThreadReference threadRef) {
        addThread(threadRef.name(), threadRef.uniqueID());
    }

    void addThread(String name, long id) {
        if (comboBox == null) return;
        ThreadPair newPair = new ThreadPair(name, id);
        int i = 1;
        int size = comboBox.getModel().getSize();
        long pid = -1;
        ThreadPair pair = null;
        while (i < size &&
               (pid = ((ThreadPair) comboBox.getModel().getElementAt(i)).id) < id) {
            i++;
        }
        if (pid != id ) {
            comboBox.insertItemAt(newPair, i);
        }
    }

    void removeThread(ThreadReference threadRef) {
        removeThread(threadRef.name(), threadRef.uniqueID());
    }

    void removeThread(String name, long id) {
        if (comboBox == null) return;
        for (int i = 0; i < comboBox.getModel().getSize(); i++) {
            ThreadPair pair = (ThreadPair) comboBox.getModel().getElementAt(i);
            if (pair.name.equals(name) && pair.id == id) {
                comboBox.removeItemAt(i);
            }
        }
    }

    static class ThreadPair {
        String name;
        long id;
        ThreadPair(String name, long id) {
            this.name = name;
            this.id   = id;
        }
        public String toString() {
            return name + (id == -1 ? "" : " (" + id + ")");
        }
        public boolean equals(Object o) {
            if (!(o instanceof ThreadPair)) {
                return super.equals(o);
            }
            return ( ((ThreadPair) o).name.equals(name) &&
                     ((ThreadPair) o).id == id );
        }
    }

    public static String d() { return "Toolbar"; }
    public String toString() { return d(); }
}
