/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import com.sun.jdi.*;
import java.util.*;
import org.aspectj.tools.ide.*;

public class AJTranslator {

    final static String AJC                 = "$ajc";
    final static String BEFORE_AJC          = "before"  + AJC;
    final static String AFTER_AJC           = "after"   + AJC;
    final static String AROUND_AJC          = "around"  + AJC;

    final static String EXECUTION           = "$execution";
    final static String RECEPTION           = "$reception";
    final static String CALL                = "$call";
    final static String ASPECT_OF           = "aspectOf";
    final static String HAS_ASPECT          = "hasAspect";

    final static String AROUND              = "$around";
    final static String AROUND_EXECUTION    = AROUND + EXECUTION;
    final static String AROUND_RECEPTION    = AROUND + RECEPTION;
    final static String AROUND_CALL         = AROUND + CALL;

    final static String[] SPECIAL_STRINGS = {
        EXECUTION,
        RECEPTION,
        CALL,
        ASPECT_OF,
        HAS_ASPECT,
    };

    final static String[] ADVICE_STRINGS = {
        BEFORE_AJC,
        AFTER_AJC,
        AROUND_AJC
    };

    /**
     * Takes a FULLY qualified method name and returns the method name that
     * really maps from an output file to a sourcefile.
     */
    public static String getMappingMethodName(String method) {
        String clazz = AJUtil.getTypeFromMethod(method);

        // If this class was compiled with AJC, it will jave a ajwd file
        String file = AJLineMapper.getSourceFilePathFromAJCClass(clazz);
        if (file == null) {
            return method;
        }

        // Get the line of the method
        int line = AJDecParser.getMethodLineNumber(file, method);
        if (line == -1) {
            return method;
        }

        // See if it is even advised
        Declaration dec = AJLineMapper.symbolManager.getDeclarationAtLine(file, line);
        if (dec == null                  ||
            dec.getPointedToBy() == null ||
            dec.getPointedToBy().length == 0) {
            return method;
        }

        // If this method has advice, it's name will *not* map, so find all
        // methods that start with <basename>$<clazz>, and see if these methods
        // map back to source
        String baseName = AJUtil.stripParens(method);
        //ReferenceType refType = Env.getReferenceTypeFromToken(clazz);
        ReferenceType refType = null;
        try {
          refType = ComponentRepository.getAJDebugger().getReferenceTypeFromToken(clazz);
        } catch (DebuggerException de) {
        }
        if (refType == null) {
            return method;
        }
        Iterator iter = refType.methods().iterator();
        String uScoreClass = clazz.replace('.', '_');
        String comp = AJUtil.stripTypeFromMethod(baseName) + "$" + uScoreClass;
        while (iter.hasNext()) {
            Method meth = (Method) iter.next();
            if (!meth.name().startsWith(comp) || isNonMappingMethod(meth)) {
                continue;
            }
            return meth + "";
        }
        return method;
    }

    public static String translateFullWorkingDirMethodName(String meth) {
        String type     = AJUtil.getTypeFromMethod(meth);
        String name     = translateWorkingDirMethodName(
                            AJUtil.getNameFromMethod(meth));
        String params   = AJUtil.getParamsFromMethod(meth);
        return type + "." + name + "(" + params + ")";
    }

    /**
     * @param method    a fully-qualified method with type, name, and params
     */
    public static boolean isNonMappingMethod(Method method) {
        Location loc = method.location();
        if (loc == null) {
            return false;
        }
        int line = AJLineMapper.getCorrespondingLine(loc).line;
        if (line == AJLineMapper.NON_MAPPING_SOURCE_LINE) {
            return true;
        }
        return false;
    }

    public static boolean isNonMappingMethod(String clazz, int oldLine) {
        String ajFile = AJLineMapper.getSourceFilePathFromAJCClass(clazz);
        if (ajFile == null) {
            return false;
        }
        int line = AJLineMapper.getCorrespondingLine(clazz, oldLine).line;
        if (line == AJLineMapper.NON_MAPPING_SOURCE_LINE) {
            return true;
        }
        return false;
    }

    public static String getBaseName(String method) {
        int end = method.indexOf("(");
        if (end == -1) {
            end = method.length();
        }
        int i = end - 1;
        int idollar = -1;
        while (i >= 0) {
            char c = method.charAt(i);
            if (c == '.') {
                break;
            } else if (c == '$') {
                idollar = i;
            }
            i--;
        }
        return (idollar != -1) ? method.substring(0, idollar) : method;
    }

    public static String getBaseNameWithParams(String method) {
        int ip = method.indexOf("(");
        String params = "";
        String nameAndMaybeType = method;
        if (ip != -1) {
            params = method.substring(ip);
            nameAndMaybeType = method.substring(0, ip);
        }
        return getBaseName(nameAndMaybeType) + params;
    }

    public static String translateWorkingDirMethodName_(String methodName) {
        int firstDollar = methodName.indexOf('$');

        if (firstDollar == -1) return methodName;

        String baseName = methodName.substring(0, firstDollar);

        int iajc;
        int iaroundExecution;
        int iaroundReception;
        int iaroundCall;
        int iexecution;
        int ireception;
        int icall;

        if ((iajc = methodName.indexOf(AJC)) != -1) {
            String numStr = methodName.substring(iajc + AJC.length());
            return baseName + "<" + "advice-#" + numStr + ">";
        } else if ((iaroundExecution = methodName.indexOf(AROUND_EXECUTION)) != -1) {
            return baseName + "<" + "around-execution>";
        } else if ((iaroundReception = methodName.indexOf(AROUND_RECEPTION)) != -1) {
            return baseName + "<" + "around-reception>";
        } else if ((iaroundCall = methodName.indexOf(AROUND_CALL)) != -1) {
            String numStr = methodName.substring(iaroundCall + AROUND_CALL.length());
            return baseName + "<" + "around-call-#" + numStr + ">";
        } else if ((iexecution = methodName.indexOf(EXECUTION)) != -1) {
            return baseName + "<" + "execution>";
        } else if ((ireception = methodName.indexOf(RECEPTION)) != -1) {
            return baseName + "<" + "reception>";
        } else if ((icall  = methodName.indexOf(CALL))!= -1) {
            String numStr = methodName.substring(icall + CALL.length());
            return baseName + "<" + "call-#" + numStr + ">";
        }
        //return baseName;
        return methodName;
    }

    public static String translateWorkingDirMethod(String clazz, String methodName) {
        if (isAdvice(methodName)) {
            return translateWorkingDirMethodName(methodName);
        }
        return getMappingMethodName(clazz + "." + methodName);
    }

    public static String translateWorkingDirMethodName(String methodName) {
        int firstDollar = methodName.indexOf('$');
        if (firstDollar == -1) {
            return methodName;
        }

        String baseName = methodName.substring(0, firstDollar);
        String noDollars = methodName.substring(firstDollar + 1).replace('$', '-');
        return baseName + "<" + noDollars + ">";
    }

    public static boolean isAdvice(String methodName) {
        if (methodName.indexOf(AJC) != -1) {
            return true;
        }
        return false;
    }

    public static boolean isReception(String methodName) {
        if (methodName.indexOf(RECEPTION) != -1) {
            return true;
        }
        return false;
    }

    public static boolean isSpecialMethod(String methodName) {
        for (int i = 0; i < SPECIAL_STRINGS.length; i++) {
            if (methodName.indexOf(SPECIAL_STRINGS[i]) != -1) {
                return true;
            }
        }
        return false;
    }

    public static String translateSourceMethodNameWithParams(String methodName) {
        String params = AJUtil.getParamsFromMethod(methodName);
        String rest   = AJUtil.stripParens(methodName);
        String result = translateSourceMethodName(rest);
//        if (result.indexOf("-advice-#") == -1) {
            return result + "(" + params + ")";
//        }
//        return result;

    }

    /**
     * @param   methodName  A method with or w/o params, but <b>NO TYPE</b>.
     */
    public static String translateSourceMethodName(String methodName) {

        // Keep <init>, <clinit>, and any name without <'s
        int firstLT = methodName.indexOf("<");
        if (firstLT < 1) {
            return methodName;
        }

        String baseName = methodName.substring(0, firstLT);
        String dollars = methodName.substring(firstLT + 1, methodName.length() - 1).replace('-', '$');
        return baseName + "$" + dollars;
    }

    public static String translateSourceMethodName_(String methodName) {
        if (methodName.equals("<init>") ||
            methodName.equals("<clinit>")) {
            return methodName;
        } else if (methodName.startsWith("<") && methodName.endsWith(">")) {
            //
            // Advice
            //
            int iAdvice = methodName.indexOf("-advice-#");
            if (iAdvice == -1) {
                return methodName;
            }
            String adviceType = methodName.substring(1, iAdvice);
            int iNum = iAdvice + 9;
            String numStr = methodName.substring(iNum, methodName.length()-1);
            return adviceType + "$ajc" + numStr;

        } else {
            //
            // Other stuff
            //

        }
        return methodName;
    }

    public static String translateWorkingDirMethodNameWithParams(String methodName) {
        String params = AJUtil.getParamsFromMethod(methodName);
        String rest   = AJUtil.stripParens(methodName);
        String result = translateWorkingDirMethodName(rest);
//        if (result.indexOf("-advice-#") == -1) {
            return result + "(" + params + ")";
//        }
//        return result;
    }

    /**
     * Returns the <b>real</b> name of this method.
     * @param   <code>classID</code>    The full name of the class.
     * @param   <code>meth</code>       The name of the method with params.
     */
    public static String getTranslatedMethodName(String classID, String meth) {
        return "";
//        String filePath = AJLineMapper.getSourceFilePathFromAJCClass(classID);
//        if (filePath == null) {
//            return meth;
//        }
//
//        //
//        // Search all the declarations util we find a method with name 'meth'
//        //
//        int line = AJDecParser.getMethodLineNumber(filePath, classID + "." + meth);
//        Util.msg("LINE", line+"", java.awt.Color.darkGray);
//        SourceLine sl = getCorrespondingOutputLine(classID, line);
//        if (sl == null) {
//            return meth;
//        }
//        int outLine = sl.line;
//        try {
//            ReferenceType refType = Env.getReferenceTypeFromToken(classID);
//            Iterator iter = refType.methods().iterator();
//            while (iter.hasNext()) {
//                Method method = (Method) iter.next();
//                Location loc = method.location();
//                int locLine = loc.lineNumber();
//                if (locLine == outLine) {
//                    Util.debug("GOT ONE: " + loc);
//                    return method.name();
//                }
//            }
//        } catch (VMNotConnectedException e) {
//            return meth;
//        }
//        return meth;
    }
}
