/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import java.awt.Component;
import java.awt.event.MouseEvent;
import java.util.List;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;

/**
 * The class from which all tree nodes in the debugger are derived.  It
 * provides some basic functions, such as accessing the type of the node.  It
 * also provides the functions <code>add<\code> and <code>remove<\code>.  These
 * both take <code>List<\code>s... read more below.
 *
 * The user object of these nodes should be of type com.sun.jdi.ReferenceType.
 * Comparison between these nodes is done on the user object.
 */
public class AJTreeNode extends DefaultMutableTreeNode {

    /**
     * The model to use to insert nodes into dynamic trees.
     */
    protected DefaultTreeModel model;

    /**
     * The type of the node's icon.
     */
    protected int type = -1;

    /**
     * A default string that is displayedi f the user object is null.
     */
    protected String string = "";

    /**
     * Whether this node can always expand.
     */
    protected boolean alwaysExpand = false;

    /**
     * Constructs a new node with icon type <code>type<\code>.  The type
     * <code>type<\code> <b>must<\b> come from AJIcons.
     *
     * @param   type     a type according to AJIcons.
     * @param   model    the model to use
     * @see              #AJIcons
     */
    public AJTreeNode(int type, DefaultTreeModel model) {
        super();
        setAllowsChildren(true);
        setType(type);
        setModel(model);
    }    

    /**
     * Constructs a new node with icon type <code>type<\code>.  The type
     * <code>type<\code> <b>must<\b> come from AJIcons.
     *
     * @param   type     a type according to AJIcons.
     * @see              #AJIcons
     */
    public AJTreeNode(int type) {
        this(type, null);
    }

    /**
     * 
     */
    public AJTreeNode() {
        this(AJIcons.DEFAULT_ICON);
    }

    /**
     * A synchronized add that uses the tree model if it's not <code>null</code>.
     */
    public void add(AJTreeNode treeNode) {
        synchronized (this) {
            if (model == null) {
                super.add(treeNode);
            } else {
                model.insertNodeInto(treeNode, this, getChildCount());
            }
        }
    }

    /**
     * Returns the model of this node -- could be <code>null</code>.
     *
     * @return the tree model for this node.
     */
    public DefaultTreeModel getModel() {
        return model;
    }

    /**
     * Sets the tree model for this node.
     *
     * @param model the new model.
     */
    public void setModel(DefaultTreeModel model) {
        this.model = model;
    }
    

    /**
     * Should be implemented by a tree to figure out if it's valid.
     */
    public boolean isValid() {
        return true;
    }

    /**
     * Returns the type corresponding to that found in AJIcons.
     *
     * @return the type corresponding to that found in AJTreeIconConstancts.
     * @see     #AJIcons
     */
    public int getType() {
        return type;
    }

    /**
     * @param alwaysExpand whether this node always expands or not.
     */
    public void setAlwaysExpand(boolean alwaysExpand) {
        this.alwaysExpand = alwaysExpand;
    }

    /**
     * Sets the type according to a type found in AJIcons.
     *
     * @param   type    the type according to a type found in
     *                  AJIcons
     * @see     #AJIcons
     */
    public void setType(int type) {
        this.type = type;
        //string = AJTreeIcons.mapTypeToString(type);
    }

    /**
     * Inserts the element at the beginning of the list.
     *
     * @param   list    a java.util.List with the element to be inserted
     *                  at the beginning.
     * @return  the AJTreeNode added, or null if it wasn't added.
     */
    protected AJTreeNode add(List list) {
        return add(list, list.size() - 1);
    }

    /**
     * Private version of add(List), so it doesn't destroy the List.
     *
     * @param   list    a java.util.List with the element to be inserted
     *                  at the beginning.
     * @return  the AJTreeNode added, or null if it wasn't added.
     */
    private AJTreeNode add(List list, int index) {
        if (index < 0) {
            return (AJTreeNode) list.get(0);
        }
        Enumeration enum = children();
        AJTreeNode add = (AJTreeNode) list.get(index);
        while (enum.hasMoreElements()) {
            AJTreeNode next = (AJTreeNode) enum.nextElement();
            if (next.equals(add)) {
                return next.add(list, index - 1);
            }
        }
        add(add);
        return add.add(list, index - 1);
    }

    /**
     * Removes the element at the beginning of the list.
     *
     * @param   list    A java.util.List containing instances of AJTreeNode
     *                  with the element to be removed at the beginning.
     */
    protected AJTreeNode remove(List list) {
        return this.remove(list, list.size() - 1);
    }

    /**
     * Private version of remove(List), so it doesn't destroy the List.
     *
     * @param   list    A java.util.List containing instances of AJTreeNode
     *                  with the element to be removed at the beginning.
     */
    private AJTreeNode remove(List list, int index) {
        if (index < 0) {
            //return true;
            return (AJTreeNode) list.get(0);
        }
        Enumeration enum = this.children();
        AJTreeNode remove = (AJTreeNode) list.get(index);
        while (enum.hasMoreElements()) {
            AJTreeNode next = (AJTreeNode) enum.nextElement();
            if (next.equals(remove)) {
                if (index == 0) {
                    this.remove(next);
                    return next;
                    //return true;
                } else {
                    return next.remove(list, index - 1);
                }
            }
        }
        //this.remove(remove);
        return null;    //TODO: Shouldn't do anything
        //return remove;
        //return true;
    }

    /**
     * Calls <code>equals<\code> on the underlying user object.
     *
     * @param   node    another AJTreeNode
     */
    public boolean equals(AJTreeNode node) {
        return this.getUserObject().equals(node.getUserObject());
    }

    /**
     * Shows the kids of this node.
     * TODO: Maybe make this class and this method abstract.
     */
    public void showKids() {}

    /**
     * Hides the kids of this node.
     * TODO: Maybe make this class and this method abstract
     */
    public void hideKids() {}

    /**
     * A AJTreeWillExpandListener asks this node before it expands it.
     */
    public boolean isOKToExpand() {
        if (alwaysExpand) return true;
        return ComponentRepository.getAJDebugger().isStopped();
    }

    /**
     * An AJTreeWillExpandListener asks this node before it expands it.
     */
    public boolean isOKToCollapse() {
        //return ComponentRepository.getAJDebugger().isAtBreakpoint();
        return true;
    }

    /**
     * Primitive nodes can be expanded.
     */
    public boolean isEditable() {
        return false;
    }

    /**
     * Should be implemented to validate your kids.  Called by
     * org.aspectj.tools.debug.AJTree.validateThreads to all the current
     * thread nodes.
     */
    public void validate() {}

    public void mouseEvent(MouseEvent e) {
        if      (SwingUtilities.isLeftMouseButton(e))   leftMouseButton(e);
        else if (SwingUtilities.isMiddleMouseButton(e)) middleMouseButton(e);
        else if (SwingUtilities.isRightMouseButton(e))  rightMouseButton(e);
    }

    /**
     * Override when someone hits mouse button 1.
     */
    public void leftMouseButton(MouseEvent e) {}

    /**
     * Override when someone hits mouse button 2.
     */
    public void middleMouseButton(MouseEvent e) {}

    /**
     * Override when someone hits mouse button 3.
     */
    public void rightMouseButton(MouseEvent e) {}

    /**
     * Returns the menu -- if any -- that will appear when a right mouse button
     * is pressed on this node.
     */
    public JPopupMenu getPopupMenu() { return null; }

    /**
     * Override to get tool tip text.
     */
    public String getToolTipText() { return ""; }

    public boolean isThread()           { return false; }
    public boolean isThreadGroup()      { return false; }
    public boolean isMethod()           { return false; }
    public boolean isField()            { return false; }
    public boolean isStackFrame()       { return false; }
    public boolean isRoot()             { return false; }
    public boolean isLocalVariable()    { return false; }
    public boolean isObject()           { return false; }
    public boolean isPrimitive()        { return false; }
    public boolean isArray()            { return false; }
    public boolean isBoolean()          { return false; }
    public boolean isByte()             { return false; }
    public boolean isChar()             { return false; }
    public boolean isDouble()           { return false; }
    public boolean isFloat()            { return false; }
    public boolean isInteger()          { return false; }
    public boolean isLong()             { return false; }
    public boolean isShort()            { return false; }
    public boolean isValueNode()        { return false; }
    public boolean isString()           { return false; }


    /**
     * Returns the user object's <code>toString<\code> unless that object is
     * null, then it returns <code>string<\code>.
     *
     * @return  string rep of this node's user object if it's not null.
     */
    public String toString() {
        return getUserObject() == null ? string : "" + getUserObject();
    }

    public Enumeration children() {
        try {
            return super.children();
        } catch (Exception e) {
            AJUtil.warn("Caught " + e.getMessage());
            e.printStackTrace(System.err);
        }
        return new Enumeration() {
                public boolean hasMoreElements() {
                    return false;
                }
                public Object nextElement() {
                    throw new NoSuchElementException();
                }
            };
    }

    private boolean dd = true;
    protected String dd() {
        return (dd ? System.identityHashCode(this)
                /*System.identityHashCode(getUserObject())*/ + ": " : "");
    }
}
