/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import org.aspectj.debugger.base.*;
import org.aspectj.debugger.request.*;
import org.aspectj.tools.ide.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.List;
import javax.swing.*;

/**
 * A class encapsulating the abstract behavior of a source file viewer.  This class'
 * abstract methods are teh basic behavior required by a graphical debugger.
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */
public abstract class AbstractSourcePane
    extends AJPanel
    implements DebuggerListener,
               VMListener,
               StopListener {

    protected ComponentDirector director;

    public AbstractSourcePane(GUIDebugger guid) {
        super(guid);
        debugger().addDebuggerListener(this);
        debugger().addVMListener(this);
        debugger().addStopListener(this);
        this.director = guid.getGui();
    }

    /**
     * Displays the source file with relative path <code>relativePath</code>.
     *
     * @param relativePath The relative path of the source we want to view.
     */
    public abstract boolean showSource(String relativePath);

    /**
     * This method will display the current source file at line number <code>line</code>
     * and display a breakpoint token if <code>isAtBreakpoint</code> is <code>true</code>.
     *
     * @param line              Line number from 1 to N to display.
     * @param isAtBreakpoint    Whether the program is broken at line <code>line</code>.
     */
    public abstract void showLineForCurrentModel(int line, boolean isAtBreakpoint);

    /**
     * Returns the name of the current source file.
     *
     * @return  The name of the currrent source file.
     */
    public abstract String getSourceName();

    /**
     * This method is invoked by the debugger when a breakpoint is set at line number
     * <code>line</code> in file <code>filename</code>.  The implementing source pane
     * should take the necessary steps when this occurs.
     *
     * @param filename  The file at which the program broke.
     * @param line      The line number at which the program broke.
     * @param ba        The <code>org.aspectj.debugger.base.BreakpointRequestAction</code>
     *                  that describes the breakpoint.
     * @see             #org.aspectj.debugger.base.BreakpointRequestAction
     */
    public abstract void requestSet(String filename, int line,
                                    BreakpointRequestAction ba);

    /**
     * This method is invoked by the debugger when a breakpoint is cleared at line number
     * <code>line</code> in file <code>filename</code>.  The implementing source pane
     * should take the necessary steps when this occurs.
     *
     * @param filename  The file at which the program broke.
     * @param line      The line number at which the program broke.
     * @param ba        The <code>org.aspectj.debugger.base.BreakpointRequestAction</code>
     *                  that describes the breakpoint.
     * @see             #org.aspectj.debugger.base.BreakpointRequestAction
     */
    public abstract void requestClear(String filename, int line,
                                      BreakpointRequestAction ba);

    /**
     * This method is invoked by the debugger when a breakpoint is deferred at line number
     * <code>line</code> in file <code>filename</code>.  The implementing source pane
     * should take the necessary steps when this occurs.
     *
     * @param filename  The file at which the program broke.
     * @param line      The line number at which the program broke.
     * @param ba        The <code>org.aspectj.debugger.base.BreakpointRequestAction</code>
     *                  that describes the breakpoint.
     * @see             #org.aspectj.debugger.base.BreakpointRequestAction
     */
    public abstract void requestDeferred(String filename, int line,
                                         BreakpointRequestAction ba);


    // ---------------------------------- Access methods ---------------------------------

    /**
     * Returns whether the file exists or not.
     *
     * @param  relativePath The relative path name of the file.
     * @return              Whether the file exists.
     */
    public /*abstract*/ boolean fileExists(String relativePath) {
        String fullPath = debugger().getFullSourcePath(relativePath);
        File file = new File(fullPath);
        if (file != null && file.exists()) {
            return true;
        }
        return false;
    }    

    public String getSourcePath() {
        return debugger().getSourceManager().getSourcePath();
    }

    public SourceManager getSourceManager() {
        return debugger.getSourceManager();
    }

    public void showSourceForFrame(StackFrame frame, boolean isAtBreakpoint) {
        try {
            showSourceForLocation(frame.location(), isAtBreakpoint);
        } catch (InvalidStackFrameException isfe) {
        }
    }

    public void showSourceForLocation(Location loc, boolean isAtBreakpoint) {
        SourceLine sl = null;
        try {
            sl = debugger().sourceLine(loc);
        } catch (AbsentInformationException aie) {
            return;
        }
        if (sl == null) {
            return;
        }
        if (sl.line < 0) {
            return;
        }
        if (fileExists(sl.filename)) {
            if (showSource(sl.filename)) {
                showLineForCurrentModel(sl.line, isAtBreakpoint);
            }
        }
        director.makeTopLevel();
    }

    public void showSourceForLocation(Location loc) {
        showSourceForLocation(loc, false);
    }

    public void showSourceForFileAndLine(String relativePath, int line) {
        if (line < 0) {
            return;
        }
        if (!(new File(relativePath).exists()) &&
            !(new File(AJLineMapper.removeRoot(relativePath)).exists())) {
            return;
        }
        if (fileExists(relativePath)) {
            if (showSource(relativePath)) {
                showLineForCurrentModel(line);
            }
        }
        director.makeTopLevel();
    }

    protected void showLineForCurrentModel(int line) {
        showLineForCurrentModel(line, false);
    }

    public File getPath() {
        return new File(getSourceName());
    }

    /****************************** DebuggerListener ******************************/

    public void requestSetEvent(RequestEvent re) {
        Request request = re.getRequest();
        if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction ba = (BreakpointRequestAction) request;
            SourceLine sl = debugger().getSourceLine(ba);
            requestSet(sl.filename, sl.line, ba);
        }
    }
    public void requestClearEvent(RequestEvent re) {
        Request request = re.getRequest();
        if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction ba = (BreakpointRequestAction) request;
            SourceLine sl = debugger().getSourceLine(ba);
            requestClear(sl.filename, sl.line, ba);
        }
    }
    public void requestDeferredEvent(RequestEvent re) {
        Request request = re.getRequest();
        if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction ba = (BreakpointRequestAction) request;
            SourceLine sl = debugger().getSourceLine(ba);
            requestDeferred(sl.filename, sl.line, ba);
        }
    }
    public void requestFailedEvent(RequestEvent re) {
    }

    /****************************** VMListener ******************************/

    public void vmStartEvent(VMStartEvent e) {
    }
    public void vmDeathEvent(VMDeathEvent e) {
    }
    public void vmDisconnectEvent(VMDisconnectEvent e) {
    }

    /****************************** StopListener ******************************/
    public void accessWatchpointEvent(AccessWatchpointEvent e) {
        showSourceForLocation(e.location());
    }
    public void breakpointEvent(BreakpointEvent e) {
        showSourceForLocation(e.location());
    }
    public void exceptionEvent(ExceptionEvent e) {
        showSourceForLocation(e.location());
    }
    public void modificationWatchpointEvent(ModificationWatchpointEvent e) {
        showSourceForLocation(e.location());
    }
    public void stepEvent(StepEvent e) {
        showSourceForLocation(e.location());
    }

    public static String d() { return "Source Pane"; }
    public String toString() { return d(); }


}
