/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import com.sun.jdi.request.*;
import org.aspectj.debugger.base.*;
import org.aspectj.debugger.request.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;

public class BreakpointTreePane
    extends AJPanel
    implements DebuggerListener
{

    private BreakpointTree tree = null;
    private JScrollPane scrollPane = null;

    public BreakpointTreePane(Frame frame, GUIDebugger guid) {
        super(guid);
        debugger().addDebuggerListener(this);
        setLayout(new BorderLayout());
        tree = new BreakpointTree(frame);
        scrollPane = new JScrollPane(tree);
        add(scrollPane, BorderLayout.CENTER);

    }

    public BreakpointTreePane() {
        this(null, null);
    }

    BreakpointTreeModel getModel() {
        return (BreakpointTreeModel) tree.getModel();
    }

    void modernizeTree() {
        tree.modernize();
    }

    void add(Object o, boolean verified, Location loc) {
        ((BreakpointTreeModel) tree.getModel()).add(o, verified, loc);
        tree.modernize();
    }

    void remove(Object o) {
        ((BreakpointTreeModel) tree.getModel()).remove(o);
        tree.modernize();
    }

    /************************* DebuggerListener *************************/

    public void requestSetEvent(RequestEvent re) {
        Request r = re.getRequest();
        if (!(r instanceof BreakpointRequestAction)) {
            return;
        }
        BreakpointRequestAction ra = (BreakpointRequestAction) r;
        add(ra, true, ra.getLocation());
    }

    public void requestClearEvent(RequestEvent re) {
        Request r = re.getRequest();
        if (!(r instanceof BreakpointRequestAction)) {
            return;
        }
        BreakpointRequestAction ra = (BreakpointRequestAction) r;
        remove(ra);
    }

    public void requestDeferredEvent(RequestEvent re) {
        Request r = re.getRequest();
        if (!(r instanceof BreakpointRequestAction)) {
            return;
        }
        BreakpointRequestAction ra = (BreakpointRequestAction) r;
        add(ra, false, ra.getLocation());
    }

    public void requestFailedEvent(RequestEvent re) {
        Request r = re.getRequest();
        if (!(r instanceof BreakpointRequestAction)) {
            return;
        }
        BreakpointRequestAction ra = (BreakpointRequestAction) r;
        remove(ra);
    }

    public static String d() { return "Breakpoint Tree Pane"; }
    public String toString() { return d(); }
}

class BreakpointTree extends JTree {
    private Frame frame = null;

    public BreakpointTree(Frame frame) {
        super(new BreakpointTreeModel());
        this.frame = frame;
        setCellRenderer(new BreakpointTreeCellRenderer());
    	addMouseListener(new BreakpointMouseListener());
        addTreeSelectionListener(new BreakpointTreeSelectionListener());
        addTreeWillExpandListener(new NoCollapsingRootExpansionListener(this));
    }

    public BreakpointTree() {
        this(null);
    }

    private class BreakpointMouseListener extends MouseAdapter {
        public void mouseClicked(MouseEvent e) {
            int row = getRowForLocation(e.getX(), e.getY());
            if(row != -1 && e.getClickCount() == 1) {
                TreePath[] paths = BreakpointTree.this.getSelectionPaths();
                if (paths == null) {
                    paths = new TreePath[0];
                    paths[0] = BreakpointTree.this.getPathForLocation(e.getX(), e.getY());
                }
                for (int i = 0; i < paths.length; i++) {
                    BreakpointTreeNode node =
                        (BreakpointTreeNode) paths[i].getLastPathComponent();
                    if (SwingUtilities.isLeftMouseButton(e)) {
                        node.leftMouseButton(BreakpointTree.this, e);
                    } else if (SwingUtilities.isMiddleMouseButton(e)) {
                        node.middleMouseButton(BreakpointTree.this, e);
                    } else if (SwingUtilities.isRightMouseButton(e)) {
                        node.rightMouseButton(BreakpointTree.this, e);
                    }
                }
            }
        }
    }

    private class BreakpointTreeSelectionListener
        implements TreeSelectionListener {
        public void valueChanged(TreeSelectionEvent e) {
            TreePath path = e.getPath();
            BreakpointTreeNode node =
                (BreakpointTreeNode) path.getLastPathComponent();
            node.selection(BreakpointTree.this);
        }
    }

    public Frame getFrame() {
        return frame;
    }

    void remove(Object o) {
        ((BreakpointTreeModel)getModel()).remove(o);
        modernize();
    }

    void modernize() {
        TreeUI newUI =  (TreeUI) UIManager.getUI(this);
        if (ui != null) {
            try {
                ui.uninstallUI(this);
            } catch (Throwable t) {
                // who cares...
            }
        }
        ComponentUI oldUI = ui;
        ui = newUI;
        if (ui != null) {
            ui.installUI(this);
        }
        firePropertyChange("UI", oldUI, newUI);
        revalidate();
        repaint();
    }
}

class BreakpointTreeModel extends DefaultTreeModel {
    public BreakpointTreeModel() {
        super(new BreakpointRootTreeNode());
    }

    void add(Object o, boolean verified, Location loc) {
        BreakpointTreeNode node = null;
        DefaultMutableTreeNode root = (DefaultMutableTreeNode) getRoot();
        Enumeration enum = root.children();
        while(enum.hasMoreElements()) {
            node = (BreakpointTreeNode) enum.nextElement();
            if (node.toString().equals(o+"")) {
                if (node.isVerified()) {
                } else if (verified) {
                    root.remove(node);
                    node = new VerifiedBreakpointTreeNode(node.getUserObject(), loc);
                    root.add(node);
                } else { // Now our dick's really caught in the fan!!!
                }
                return;
            }
        }

        if (verified) {
            node = new VerifiedBreakpointTreeNode(o, loc);
        } else {
            node = new UnverifiedBreakpointTreeNode(o, loc);
        }
        root.add(node);
    }

    void remove(Object o) {
        BreakpointTreeNode node = null;
        DefaultMutableTreeNode root = (DefaultMutableTreeNode) getRoot();
        Enumeration enum = root.children();
        while(enum.hasMoreElements()) {
            node = (BreakpointTreeNode) enum.nextElement();
            if (node.toString().equals(o+"")) {
                root.remove(node);
            }
        }
    }
}

class BreakpointTreeCellRenderer extends DefaultTreeCellRenderer {
    public Component getTreeCellRendererComponent(JTree tree,
                                                  Object value,
                                                  boolean selected,
                                                  boolean expanded,
                                                  boolean leaf,
                                                  int row,
                                                  boolean hasFocus)
    {
        super.getTreeCellRendererComponent(tree, value, selected, expanded,
                                           leaf, row, hasFocus);
        BreakpointTreeNode node = (BreakpointTreeNode) value;
        setIcon(node.getIcon());
        setText(node + "");
        return this;
    }
}

abstract class BreakpointTreeNode extends DefaultMutableTreeNode {
    Location loc = null;
    public BreakpointTreeNode(Object o) {
        super(o);
    }
    public BreakpointTreeNode(Object o, Location loc) {
        this(o);
        this.loc = loc;
    }
    abstract Icon getIcon();
    void leftMouseButton(JTree tree, MouseEvent e)      {}
    void selection(JTree tree)                          {}
    boolean isVerified() {
        return false;
    }

    void middleMouseButton(JTree tree, MouseEvent e) {
        showClearMenu(tree, e);
    }
    void rightMouseButton(JTree tree, MouseEvent e) {
        showClearMenu(tree, e);
    }

    protected void showClearMenu(JTree tree, MouseEvent e) {
        int x = e.getX();
        int y = e.getY();
        new BreakpointClearMenu((BreakpointTree) tree, this).show(tree, x, y);
    }
}

class BreakpointRootTreeNode extends BreakpointTreeNode {
    public BreakpointRootTreeNode() {
        super("Breakpoints");
    }
    Icon getIcon() {
        return AJIcons.getIcon(AJIcons.BREAKPOINTS_ICON);
    }
    void leftMouseButton(JTree tree, MouseEvent e) {
    }
    void middleMouseButton(JTree tree, MouseEvent e) {
        showSetMenu(tree, e);
    }
    void rightMouseButton(JTree tree, MouseEvent e) {
        showSetMenu(tree, e);
    }
    private void showSetMenu(JTree tree, MouseEvent e) {
        //new BreakpointDialog(ComponentRepository.getComponentDirector()).show();
        Dialogs.showBreakpointDialog();
    }


}

class VerifiedBreakpointTreeNode extends BreakpointTreeNode {
    public VerifiedBreakpointTreeNode(Object o, Location loc) {
        super(o, loc);
    }
    Icon getIcon() {
        return AJIcons.getIcon(AJIcons.VERIFIED_BREAKPOINT_ICON);
    }
    boolean isVerified() {
        return true;
    }
    void selection(JTree tree) {
        if (loc != null) {
            ComponentRepository.getSourcePane().showSourceForLocation(loc);
        }
    }
    void leftMouseButton(JTree tree, MouseEvent e) {
        if (loc != null) {
            ComponentRepository.getSourcePane().showSourceForLocation(loc);
        }
    }
}

class UnverifiedBreakpointTreeNode extends BreakpointTreeNode {
    public UnverifiedBreakpointTreeNode(Object o, Location loc) {
        super(o, loc);
    }
    Icon getIcon() {
        return AJIcons.getIcon(AJIcons.UNVERIFIED_BREAKPOINT_ICON);
    }
    void leftMouseButton(JTree tree, MouseEvent e) {
        BreakpointRequestAction breakpoint = (BreakpointRequestAction) getUserObject();
        String source = breakpoint.getSourceName();
        int line = breakpoint.getLine();
    }
}

class BreakpointClearMenu extends JPopupMenu {
    private BreakpointTree tree = null;
    private BreakpointTreeNode node = null;
    BreakpointClearMenu(final BreakpointTree     tree,
                        final BreakpointTreeNode node) {
        super();
        this.tree = tree;
        this.node = node;
        String msg = node.getUserObject() + "";
        String bp = "breakpoint";
        int ibreakpoint = msg.indexOf(bp);
        if (ibreakpoint != -1) {
            msg = msg.substring(ibreakpoint + bp.length());
        }
        final String _msg = "clear " + msg;
        JMenuItem clearItem = new JMenuItem(_msg);

        final RequestAction ra = (RequestAction) node.getUserObject();
        
        clearItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                //!!!Don't want to remove by text matching!!!
                //ComponentRepository.getGUIDebugger().executeCommand(_msg);
                ComponentRepository.getGUIDebugger().getDebugger().removeRequest(ra);
                tree.remove(_msg);
            }});
        add(clearItem);
    }
}
