/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
*
* This file is part of the debugger and core tools for the AspectJ(tm)
* programming language; see http://aspectj.org
*
* The contents of this file are subject to the Mozilla Public License
* Version 1.1 (the "License"); you may not use this file except in
* compliance with the License. You may obtain a copy of the License at
* either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is AspectJ.
*
* The Initial Developer of the Original Code is Xerox Corporation. Portions
* created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
* All Rights Reserved.
*/
package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.tree.*;

public class ClassNodeMouseListener extends MouseAdapter {

    protected JTree tree;
    protected GUIDebugger guiDebugger;

    public ClassNodeMouseListener(GUIDebugger guiDebugger, JTree tree) {
        this.guiDebugger = guiDebugger;
        this.tree = tree;
    }

    public void mouseClicked(MouseEvent e) {
        int x = e.getX();
        int y = e.getY();
        int selRow = tree.getRowForLocation(x, y);
        TreePath path = tree.getPathForLocation(x, y);
        if (path == null) return;
        if (SwingUtilities.isRightMouseButton(e)) {

            AJTreeNode treeNode = (AJTreeNode) path.getLastPathComponent();
            
            String breakString          = null;
            List adviceStrings          = null;
            List advisedMethodStrings   = null;
            List classMethodStrings     = null;
            List classAdviceStrings     = null;
            String name                 = null;
            boolean isAdvice = true;

            if (treeNode instanceof MethodNode) {
                isAdvice = false;
            }

            if (treeNode instanceof Breakable) {
                Breakable breakable = (Breakable) treeNode;
                breakString = breakable.getBreakpoint();
                final String cmd = "stop in " + breakString;
                JPopupMenu menu = new JPopupMenu();
                JMenuItem item = new JMenuItem(cmd);
                item.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent _) {
                        ComponentRepository.getAJDebugger().execute(cmd);
                    }
                });
                menu.add(item);
                menu.show(tree, x, y);
            }

            if (treeNode instanceof Advisedable) {
                Advisedable advisedable = (Advisedable) treeNode;
                adviceStrings = advisedable.getAdviceStrings();
            }

            if (treeNode instanceof Advisable) {
                Advisable advisable = (Advisable) treeNode;
                advisedMethodStrings = advisable.getAdvisedMethodStrings();
            }

            if (treeNode instanceof ClassNode) {
                ClassNode classNode = (ClassNode) treeNode;
                classMethodStrings = classNode.getMethodStrings();
//                name = classNode.name();
                name = classNode.fullName();
            }
            if (treeNode instanceof AJCClassNode) {
                AJCClassNode ajcClassNode = (AJCClassNode) treeNode;
                classAdviceStrings = ajcClassNode.getAdviceStrings();
            }

            show(x, y, isAdvice, breakString, adviceStrings, advisedMethodStrings,
                 classMethodStrings, classAdviceStrings, name);
        } else if (SwingUtilities.isLeftMouseButton(e) && e.getClickCount() == 2) {
        }
    }

    final static String stop = "stop in ";
    private void show(int x, int y,
                      final boolean isAdvice,
                      final String breakString,
                      final List adviceStrings,
                      final List advisedMethodStrings,
                      final List classMethodStrings,
                      final List classAdviceStrings,
                      final String name) {
        JPopupMenu menu = new JPopupMenu();
        if (breakString != null && !breakString.equals("")) {
            final String adviceOrMethod = isAdvice ? "advice" : "method";
            final String cmd = stop + breakString;
            final String title = cmd;
            menu.add(new AbstractAction(title) {
                public void actionPerformed(ActionEvent _) {
                    ComponentRepository.getAJDebugger().execute(cmd);
            }});
        }
        add(menu, "stop in all advice", adviceStrings);
        add(menu, "stop in all advised methods", advisedMethodStrings);
        add(menu, "stop in all declared methods", classMethodStrings);
        add(menu, "stop in all declared advice", classAdviceStrings);
        if (name != null && !name.equals("null"))  {
            add(menu, "run class " + name, "run " + name);
        }
        menu.show(tree, x, y);
    }

    private void add(JPopupMenu menu, final String title, final String command) {
        List list = new Vector();
        list.add(command);
        add(menu, title, list, "");
    }
    private void add(JPopupMenu menu, final String title, final List strings) {
        add(menu, title, strings, stop);
    }
    private void add(JPopupMenu menu, final String title,
                     final List strings, final String prefix) {
        if (strings == null || strings.size() == 0) return;
        menu.add(new AbstractAction(title) {
            public void actionPerformed(ActionEvent _) {
                for (int i = 0; i < strings.size(); i++) {
                    Object command = strings.get(i);
                    if (command != null && !(command+"").equals("null"))
                        guiDebugger.getDebugger().execute(prefix + command);
                }
            }
        });
    }

    protected JTree tree() {
        return tree;
    }

}
