/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import java.io.*;
import java.util.*;
import java.util.List;

import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;

import com.sun.jdi.*;
import com.sun.jdi.event.*;

import org.aspectj.tools.ide.*;

public class ClassTreePane extends JPanel
                           implements VMListener,
                                      ClassListener
{

    private JTree tree;
    private DefaultTreeModel treeModel;
    private RootNode root;
    private GUIDebugger guiDebugger;

    public ClassTreePane(GUIDebugger guiDebugger) {

    	super(new BorderLayout());

        this.guiDebugger = guiDebugger;
        guiDebugger.getDebugger().addVMListener(this);
        guiDebugger.getDebugger().addClassListener(this);

    	root = new RootNode();
    	treeModel = new DefaultTreeModel(root);

        tree = new AJTree(treeModel) {
                public void clear() {}
                public int getVisibleRowCount() { return getRowCount(); }
            };
        tree.getSelectionModel().setSelectionMode
            (TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
        tree.setCellRenderer(new AJTreeCellRenderer());
        tree.addTreeSelectionListener(new ClassNodeSelectionListener(guiDebugger));
        tree.addMouseListener(new ClassNodeMouseListener(guiDebugger, tree));
        tree.addTreeWillExpandListener(new NoCollapsingRootExpansionListener(tree));
        //ToolTipManager.sharedInstance().registerComponent(tree);

    	JScrollPane treeView = new JScrollPane(tree);
    	add(treeView);
    }

    public JTree getTree() {
        return tree;
    }

    public DefaultTreeModel getTreeModel() {
        return treeModel;
    }

    final static String stop = "stop in ";
    void showMenu(int x, int y,
                  final String cmd,
                  final Vector advice,
                  final Vector advisedMethods,
                  final Vector methods,
                  final boolean isAdvice) {
        JPopupMenu menu = new JPopupMenu();
        if (cmd != null && !cmd.equals("")) {
            final String adviceOrMethod = isAdvice ? "advice" : "method";
            menu.add(new AbstractAction("Break in " + adviceOrMethod) {
                public void actionPerformed(ActionEvent e) {
                    ComponentRepository.getCommandLine().executeCommand(stop + cmd);
            }});
        }
        if (advice != null && advice.size() > 0) {
            menu.add(new AbstractAction("Break in all advice") {
                public void actionPerformed(ActionEvent e) {
                    for (int i = 0; i < advice.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + advice.get(i));
                    }
                }});
        }
        if (methods != null && methods.size() > 0) {
            menu.add(new AbstractAction("Break in all methods") {
                public void actionPerformed(ActionEvent e) {
                    for (int i = 0; i < methods.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + methods.get(i));
                    }
                }});
        }
        if (advisedMethods != null && advisedMethods.size() > 0) {
            menu.add(new AbstractAction("Break in all advised methods") {
                public void actionPerformed(ActionEvent e) {
                    for (int i = 0; i < advisedMethods.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + advisedMethods.get(i));
                    }
                }});
        }
        if (methods != null && methods.size() > 0 &&
            advice  != null && advice.size()  > 0) {
            menu.add(new AbstractAction("Break in all advice & methods") {
                public void actionPerformed(ActionEvent e) {
                    for (int i = 0; i < methods.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + methods.get(i));
                    }
                    for (int i = 0; i < advice.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + advice.get(i));
                    }
                }});
        }
        if (methods        != null && methods.size()        > 0 &&
            advice         != null && advice.size()         > 0 &&
            advisedMethods != null && advisedMethods.size() > 0) {
            menu.add(new AbstractAction
                ("Break in all advice, advised methods, and methods") {
                public void actionPerformed(ActionEvent e) {
                    for (int i = 0; i < methods.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + methods.get(i));
                    }
                    for (int i = 0; i < advice.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + advice.get(i));
                    }
                    for (int i = 0; i < methods.size(); i++) {
                        ComponentRepository.getCommandLine().executeCommand
                            (stop + advisedMethods.get(i));
                    }
                }});
        }
        menu.show(tree, x, y);
    }

    void showMultipleMenu(int x, int y, final Vector cmds) {
        JPopupMenu menu = new JPopupMenu();
        menu.add(new AbstractAction("Break in " + cmds.size() + " points") {
            public void actionPerformed(ActionEvent e) {
                for (int i = 0; i < cmds.size(); i++) {
                    ComponentRepository.getCommandLine().executeCommand
                        (stop + cmds.get(i));
                }
        }});
        menu.show(tree, x, y);
    }

    public void loadAJCClasses(String newSourcePath) {
        new AJCClassLoader().load(newSourcePath);
    }

    private static class AJCClassLoader {
        public void load(String sourcePath) {
            if (sourcePath == null || "".equals(sourcePath)) return;
            SymbolManager sm = SymbolManager.getSymbolManager();
            File file = new File(sourcePath);
            if (file == null || !file.exists()) return;
            if (file.isDirectory()) loadAJCClasses(file);
            else loadAJCClasses(file.getParentFile());
        }

        private void loadAJCClasses(File dir) {
            if (dir == null || !dir.exists() || !dir.isDirectory()) return;
            File[] javaFiles = dir.listFiles(javaFilter());
            for (int i = 0; i < javaFiles.length; i++) {
                loadAJCClass(javaFiles[i]);
            }
            File[] dirs = dir.listFiles(dirFilter());
            for (int i = 0; i < dirs.length; i++) {
                loadAJCClasses(dirs[i]);
            }
        }

        private void loadAJCClass(File file) {
//            String outputFile = file.getAbsolutePath();
//            System.out.println("outputFile=" + outputFile);
        }

        private static FileFilter dirFilter() {
            return new FileFilter() {
                public boolean accept(File file) {
                    return file.isDirectory();
                }
            };
        }


        private static FileFilter javaFilter() {
            return new FileFilter() {
                public boolean accept(File file) {
                    return file.getName().indexOf("ajworkingdir") == -1 &&
                           file.getName().endsWith(".java");
                }
            };
        }
    }


    protected void reset() {
        root = new RootNode();
    	treeModel = new DefaultTreeModel(root);
    	tree.setModel(treeModel);
    }

    public void addClass(ReferenceType refType) {
        boolean expand = root.getChildCount() == 0;
        root.addClass(refType);
        if (expand) tree.expandRow(0);
    }

    public void removeClass(String className) {
        root.removeClass(className);
    }

    /* VMListener */
    public void vmDeathEvent(VMDeathEvent e) {
        reset();
    }
    public void vmDisconnectEvent(VMDisconnectEvent e) {
        reset();
    }
    public void vmStartEvent(VMStartEvent e) {
        reset();
    }

    /* ClassListener */
    public void classPrepareEvent(ClassPrepareEvent e) {
        addClass(e.referenceType());
    }
    public void classUnloadEvent(ClassUnloadEvent e) {
        removeClass(e.className());
    }   
    
    MutableTreeNode insertNode(String name,
                               MutableTreeNode parent,
                               MutableTreeNode newChild) {
        int N = parent.getChildCount();
        for (int i = 0; i < N; i++) {
            MutableTreeNode child = (MutableTreeNode) parent.getChildAt(i);
            //XXX Hack until
            String childString = child.toString();
            int ispace = childString.indexOf(" ");
            if (ispace != -1) {
                childString = childString.substring(0, ispace);
            }
            int cmp = name.compareTo(childString);
            if (cmp == 0) {
                return child;
            } else if (cmp < 0) {
                treeModel.insertNodeInto(newChild, parent, i);
                return newChild;
            }
        }
        treeModel.insertNodeInto(newChild, parent, N);
        if (parent == root) {
            //System.out.println("expanding");
            tree.expandPath(new TreePath(((DefaultMutableTreeNode)newChild).getPath()));
        }
        return newChild;
    }

    class RootNode extends AJTreeNode {
        private int classCount = 0;

        public RootNode() {
            super(AJIcons.CLASSES_ICON);
            setUserObject("Classes");
        }

        public boolean isOKToExpand() {
            return true;
        }

        public ClassNode addClass(ReferenceType refType) {
            return addClass(refType.name(), refType);
        }

        public ClassNode addClass(String name) {
            return addClass(name, null);
        }

        public ClassNode removeClass(ReferenceType refType) {
            return null;
        }

        public ClassNode removeClass(String className) {
            return null;
        }

        protected ClassNode addClass(String name, ReferenceType refType) {
            incClassCount();
            int idot = name.indexOf(".");
            if (idot == -1) {
                ClassNode classNode = ClassNodeFactory.classNode(name, refType);
                classNode.setAlwaysExpand(true);
                ClassTreePane.this.insertNode(name, this, classNode);
                return classNode;
            } else {
                String head = name.substring(0, idot);
                String tail = name.substring(idot+1);
                return ((RootNode)ClassTreePane.this.insertNode
                        (head, this, new PackageNode(head))).addClass(tail, refType);
            }
        }

        public int getClassCount() {
            return classCount;
        }

        public void incClassCount() {
            changeClassCount(1);
        }

        public void decClassCount() {
            changeClassCount(-1);
        }

        private void changeClassCount(int delta) {
            classCount += delta;
            try {
                getTreeModel().nodeChanged(this);
            } catch (ArrayIndexOutOfBoundsException aioobe) {
            }
        }

    }

    class PackageNode extends RootNode {
        private String name;

        public PackageNode(String name) {
            super();
            setType(AJIcons.PACKAGE_ICON);
            setUserObject(this.name = name);
        }
    }

    public String toString() { return d(); }
    public static String d() {
        return "Class Tree Pane";
    }
}
