/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.util.gui.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import com.sun.jdi.*;
import com.sun.jdi.event.*;
import com.sun.jdi.event.Event;

public class ComponentDirector extends AJCenteredJFrame  {

    PrintStream out = System.out;

    static final int THREAD_GROUP_TREE_PANE_HEIGHT      = 550;
    static final int STACK_TABLE_PANE_HEIGHT            = 150;

    static final int RIGHT_UPPER_HEIGHT                 = 380;
    static final int OUTPUT_TEXT_AREA_HEIGHT            = 150;
    static final int ENV_TEXT_AREA_HEIGHT               = 150;
    static final int COMMAND_LINE_HEIGHT                = 20;
    static final int RIGHT_LOWER_HEIGHT                 = OUTPUT_TEXT_AREA_HEIGHT   +
                                                          ENV_TEXT_AREA_HEIGHT      +
                                                          COMMAND_LINE_HEIGHT;

    static final int SOURCE_TREE_PANE_WIDTH             = 200;
    static final int TREE_TABBED_PANE_WIDTH             = 200;
    static final int SOURCE_PANE_WIDTH                  = 400;

    static final int LEFT_PANEL_WIDTH                   = 400;
    static final int RIGHT_PANEL_WIDTH                  = SOURCE_TREE_PANE_WIDTH    +
                                                          SOURCE_PANE_WIDTH;

    static final int COMMAND_LINE_WIDTH                 = RIGHT_PANEL_WIDTH;

    static final int DIVIDER_SIZE                       = 5;

    static final Dimension THREAD_GROUP_TREE_PANE_DIM
        = new Dimension(LEFT_PANEL_WIDTH, THREAD_GROUP_TREE_PANE_HEIGHT);
    static final Dimension STACK_TABLE_PANE_DIM
        = new Dimension(LEFT_PANEL_WIDTH, STACK_TABLE_PANE_HEIGHT);
    static final Dimension SOURCE_TREE_PANE_DIM
        = new Dimension(SOURCE_TREE_PANE_WIDTH, RIGHT_UPPER_HEIGHT);
    static final Dimension TREE_TABBED_PANE_DIM
        = new Dimension(TREE_TABBED_PANE_WIDTH, RIGHT_UPPER_HEIGHT);
    static final Dimension SOURCE_PANE_DIM
        = new Dimension(SOURCE_PANE_WIDTH, RIGHT_UPPER_HEIGHT);
    static final Dimension OUTPUT_TEXT_AREA_DIM
        = new Dimension(RIGHT_PANEL_WIDTH, OUTPUT_TEXT_AREA_HEIGHT);
    static final Dimension ENV_TEXT_AREA_DIM
        = new Dimension(RIGHT_PANEL_WIDTH, ENV_TEXT_AREA_HEIGHT);
    static final Dimension COMMAND_LINE_DIM
        = new Dimension(COMMAND_LINE_WIDTH, COMMAND_LINE_HEIGHT);
    static final Dimension COMMAND_OUTPUT_PANEL_DIM
        = new Dimension(RIGHT_PANEL_WIDTH, RIGHT_LOWER_HEIGHT);


    protected ComponentDirector() {
        super();
    }

    protected ComponentDirector(String title, String mode) {
        super(title);
        Modes.setMode(mode);
    }

    public ComponentDirector(String[] args) {
        this(args, null);
    }

        protected OutputTextArea outputTextArea = null;
        protected EnvTextArea envTextArea = null;
        public AbstractTextArea getOutputTextArea() { return outputTextArea; }
        public AbstractTextArea getEnvTextArea() { return envTextArea; }

    public ComponentDirector(String[] args, WindowListener windowListener) {
        super("AspectJ Debugger");
        Modes.setMode(Modes.GUI);
        ComponentRepository.init(this, args);

        AJMenuBar menuBar = ComponentRepository.getMenuBar();
        setJMenuBar(menuBar);

        StackTablePane stackTablePane = ComponentRepository.getStackTablePane();
        stackTablePane.setPreferredSize(STACK_TABLE_PANE_DIM);

        JPanel commandLine = new CommandLinePanel(true);
        commandLine.setPreferredSize(COMMAND_LINE_DIM);

        OutputTextArea outputTextArea = ComponentRepository.getOutputTextArea();
        outputTextArea.setPreferredSize(OUTPUT_TEXT_AREA_DIM);

        EnvTextArea envTextArea = ComponentRepository.getEnvTextArea();
        envTextArea.setPreferredSize(ENV_TEXT_AREA_DIM);

        //SourceTreePane sourceTreePane = ComponentRepository.getSourceTreePane();
        BreakpointTreePane breakpointTreePane = ComponentRepository.getBreakpointTreePane();
        //ClassTreePane classTreePane = ComponentRepository.getClassTreePane();
        VariablesTreePane variablesTreePane = ComponentRepository.getVariablesTreePane();
    	JTabbedPane treeTabbedPane = new JTabbedPane(JTabbedPane.BOTTOM);
    	//treeTabbedPane.addTab("Sources",    null, sourceTreePane);
        //treeTabbedPane.addTab("Classes",    null, classTreePane);
    	treeTabbedPane.addTab("Breaks",     null, breakpointTreePane);
        treeTabbedPane.addTab("Variables",  null, variablesTreePane);
        treeTabbedPane.setPreferredSize(TREE_TABBED_PANE_DIM);

        AbstractSourcePane sourcePane = ComponentRepository.getSourcePane();
        sourcePane.setPreferredSize(SOURCE_PANE_DIM);

        AJThreadGroupTreePane threadGroupTreePane = ComponentRepository.getThreadGroupTreePane();
        threadGroupTreePane.setPreferredSize(THREAD_GROUP_TREE_PANE_DIM);

        AJToolBar toolbar = ComponentRepository.getToolBar();

        JSplitPane splitMain            = null;
        JSplitPane splitLeft            = null;
        JSplitPane splitRight           = null;
        JSplitPane splitUpperRight      = null;
        JSplitPane splitLowerRight      = null;
        JPanel commandOutputPanel       = null;

        splitUpperRight = makeSplit(JSplitPane.HORIZONTAL_SPLIT,
                                    treeTabbedPane,
                                    sourcePane);

        commandOutputPanel = new JPanel();
        commandOutputPanel.setLayout(new BorderLayout());
        commandOutputPanel.add(envTextArea, BorderLayout.CENTER);
        commandOutputPanel.add(commandLine, BorderLayout.SOUTH);
        commandOutputPanel.setPreferredSize(COMMAND_OUTPUT_PANEL_DIM);


        splitRight = makeSplit(JSplitPane.VERTICAL_SPLIT,
                               splitUpperRight,
                               commandOutputPanel);
        splitLeft = makeSplit(JSplitPane.VERTICAL_SPLIT,
                              threadGroupTreePane,
                              stackTablePane);
        splitMain = makeSplit(JSplitPane.HORIZONTAL_SPLIT,
                              splitLeft,
                              splitRight);

        getContentPane().add(splitMain, BorderLayout.CENTER);
        JPanel bottomFiller = new JPanel();
        bottomFiller.setLayout(new BoxLayout(bottomFiller, BoxLayout.X_AXIS));
        bottomFiller.add(Box.createHorizontalGlue());
        bottomFiller.add(toolbar);
        getContentPane().add(toolbar, BorderLayout.SOUTH);

        if (windowListener != null) {
            addWindowListener(windowListener);
        } else {
            addWindowListener(getDefaultWindowListener());
        }
        pack();
        setVisible(wantsToBeVisible());
    }

    /** Override this to provide your own window closing stuff. */
        protected WindowListener getDefaultWindowListener() {
            return new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        //XXX
                        try {
                            new org.aspectj.debugger.request.QuitRequest(getDebugger()).go();
                        } catch (Throwable t) {
                            getDebugger().quit();
                        }
                        System.exit(0);
                    }
                };
        }

    protected boolean wantsToBeVisible() { return true; }

    private JSplitPane makeSplit(int placement, Component one, Component two) {
        JSplitPane sp = new JSplitPane(placement, one, two);
        sp.setDividerSize(DIVIDER_SIZE);
        return sp;
    }


        public GUIDebugger getGUIDebugger() {
            return ComponentRepository.getGUIDebugger();
        }

        public Debugger getDebugger() {
            return getGUIDebugger().getDebugger();
        }


    public void go() {
        getGUIDebugger().go();
    }

    public void setPrintStream(PrintStream out) {
        this.out = out;
    }

        public void restoreStreams() {
            if (outputTextArea != null) outputTextArea.shutDown();
            if (envTextArea != null) envTextArea.shutDown();
        }

    public void shutDown() {
        restoreStreams();
    }

    public PrintStream getPrintStream() {
        return out;
    }

    public String toString() { return d(); }
        public static String d() { return "Component Director"; }

    /**
     * Implement this to show the thread pane.
     * It's called before the super.whereCommand().
     */
     public /*abstract*/ void showThreadPane() {}

        public void makeTopLevel() {
            requestFocus();
        }
}
