/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.util.gui.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import java.io.*;

class FileFrame extends CenteredJDialog {

    JLabel label;
    FilePane filePane;
    JTextField text;
    JButton setButton;
    JScrollPane scroller;
    JPanel panel;

    public FileFrame() {
        this(null);
    }

    public FileFrame(String root) {
        super(null, "FileFrame", false);

        panel = new JPanel();
        panel.setLayout(new BorderLayout());

        label = new JLabel("Please set the source path");

        filePane = new FilePane(root);
        filePane.setPreferredSize(new Dimension(200, 100));
        scroller = new JScrollPane(filePane);

        text = new JTextField(30);
        setButton = new JButton("Set");
        setButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent _) {
                set();
        }});
        JPanel bottomPanel = new JPanel();
        bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.X_AXIS));
        bottomPanel.add(text);
        bottomPanel.add(setButton);

        panel.add(label,        BorderLayout.NORTH);
        panel.add(scroller,     BorderLayout.CENTER);
        panel.add(bottomPanel,  BorderLayout.SOUTH);

        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        getContentPane().add(panel);
    }

    void set() {
        TreePath path = filePane.tree().getSelectionPath();
        if (path == null) {
            msg("You must select a directory");
        }
        File file = (File) path.getLastPathComponent();
        while (file != null && !file.isDirectory()) {
            file = file.getParentFile();
        }
        if (file == null) {
            System.out.println("Something funny");
            return;
        }
        text.setText(file + "");
    }

    void msg(String s) {
        System.out.println(s);
    }
}

class FilePane extends JPanel {

    private FileTree    tree        = null;
    private JButton     setButton   = null;

    public FilePane() {
        this(null);
    }

    public FilePane(String root) {
        tree = new FileTree(root);
        setBackground(Color.white);
        add(tree, BorderLayout.CENTER);
    }

    FileTree tree() {
        return tree;
    }
}

public class FileTree extends JTree {
    public FileTree(String root) {
        this(new FileModel(root));
    }

    public FileTree(FileModel model) {
        super(model);
        setSelectionModel(new FileTreeSelecionModel());
        this.setCellRenderer(new FileTreeCellRenderer());
    }

    class FileTreeSelecionModel extends DefaultTreeSelectionModel {
        public FileTreeSelecionModel() {
            super();
            setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        }
    }

    public static void main(String[] args) {
        JDialog frame = new FileFrame();
        frame.show();
    }
}

class FileTreeCellRenderer extends DefaultTreeCellRenderer {
    public Component getTreeCellRendererComponent(JTree tree,
                                                  Object value,
                                                  boolean sel,
                                                  boolean expanded,
                                                  boolean leaf,
                                                  int row,
                                                  boolean hasFocus)
    {
        super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf,
                                           row, hasFocus);
        setText(((File) value).getName());
        return this;
    }
}

class FileModel implements TreeModel {

    private Vector  listeners   = new Vector();
    private String  root        = null;

    public FileModel() {
        this(System.getProperty("user.home"));
    }

    public FileModel(String root) {
        if (root == null) {
            root = System.getProperty("user.home");
        }
        this.root = root;
    }

    public Object getChild(Object parent, int index) {
        File dir = (File) parent;
        String[] names = dir.list();
        return new File(dir, names[index]);
    }

    public int getChildCount(Object parent) {
        File pops = (File) parent;
        if (pops.isDirectory()) {
            return pops.list().length;
        }
        return 0;
    }

    public int getIndexOfChild(Object parent, Object child) {
        File dir  = (File) parent;
        File file = (File) child;
        if (!dir.isDirectory()) {
            return 0;
        }
        String[] names = dir.list();
        for (int i = 0; i < names.length; i++) {
            if (names[i].equals(file.getName())) {
                return i;
            }
        }
        return -1;
    }

    public Object getRoot() {
        return new File(root);
    }

    public boolean isLeaf(Object node) {
        return !((File) node).isDirectory();
    }

    public void valueForPathChanged(TreePath path, Object newValue) {}


    //
    // Event stuff
    //

    public void addTreeModelListener(TreeModelListener listener) {
        if ((listener != null) && (!listeners.contains(listener))) {
            listeners.addElement(listener);
        }
    }

    public void removeTreeModelListener( TreeModelListener listener ) {
        if ( listener != null ) {
            listeners.removeElement( listener );
        }
    }

    public void fireTreeNodesChanged( TreeModelEvent e ) {
        traverse(e, new TreeModelListenerDoer() {
            public void doit(TreeModelListener listener, TreeModelEvent _e) {
                listener.treeNodesChanged(_e);
        }});
    }

    public void fireTreeNodesInserted( TreeModelEvent e ) {
        traverse(e, new TreeModelListenerDoer() {
            public void doit(TreeModelListener listener, TreeModelEvent _e) {
                listener.treeNodesInserted(_e);
        }});
    }

    public void fireTreeNodesRemoved( TreeModelEvent e ) {
        traverse(e, new TreeModelListenerDoer() {
            public void doit(TreeModelListener listener, TreeModelEvent _e) {
                listener.treeNodesRemoved(_e);
        }});
    }

    public void fireTreeStructureChanged( TreeModelEvent e ) {
        traverse(e, new TreeModelListenerDoer() {
            public void doit(TreeModelListener listener, TreeModelEvent _e) {
                listener.treeStructureChanged(_e);
        }});
    }

    private void traverse(TreeModelEvent e,
                          TreeModelListenerDoer doer) {
      Enumeration enum = listeners.elements();
      while (enum.hasMoreElements()) {
         doer.doit((TreeModelListener)enum.nextElement(), e);
      }
    }

    interface TreeModelListenerDoer {
        public void doit(TreeModelListener listener, TreeModelEvent e);
    }
}
