/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.request.*;
import org.aspectj.debugger.tty.*;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;
import java.awt.event.*;
import java.util.*;

/**
 * GUIDebugger.java
 *
 *
 * Created: Tue Sep 05 14:38:25 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class GUIDebugger extends CommandLineDebugger {

    private final static String[] emptyArgs = new String[0];

    protected GUIDebugger() {
        this(emptyArgs);
    }

    protected GUIDebugger(String[] args) {
        super(args, false);
    }

    public GUIDebugger(ComponentDirector gui, String[] args) {
        super(args, false);
        setGui(gui);
    }

    public boolean isGUI() {
        return true;
    }

    private ComponentDirector gui;
    public ComponentDirector getGui() {
        return gui;
    }
    public void setGui(ComponentDirector gui) {
        this.gui = gui;
    }

    public String getClassNameToRun() {
        return null;
    }

    public String getClasspath() {
        return null;
    }

    public String getVMParameters() {
        return null;
    }


    /************************** Over-ridden commands **************************/

    public void viewCommand(String sourceName, List sourceLines)
        throws NoVMException, DebuggerException {
        ComponentRepository.getSourcePane().showSource(sourceName);
    }

    public void whereCommand(String threadName, List frames)
        throws NoVMException, DebuggerException {
        if (frames == null || frames.size() == 0) return;
        StackFrame frame = (StackFrame) frames.get(0);
        if (frame == null) return;
        expand(frame);
        ComponentRepository.getSourcePane().
            showSourceForFrame(frame, false);
        ThreadReference threadRef = getDebugger().getThread(threadName);
        ComponentRepository.getStackTablePane().
            showForThread(threadRef);
    }

    public void localsCommand(List locals)
        throws NoVMException, DebuggerException {
        super.localsCommand(locals);
        expandToLocals();
    }

    public void printCommand(Object valueRep, Value value)
        throws NoVMException, DebuggerException {
        super.printCommand(valueRep, value);
        expandToValue(value, valueRep);
    }

    public void dumpCommand(Object valueRep, Value value)
        throws NoVMException, DebuggerException {
        super.dumpCommand(valueRep, value);
        expandToValue(value, valueRep);
    }

    public void evalCommand(Object valueRep, Value value)
        throws NoVMException, DebuggerException {
        super.evalCommand(valueRep, value);
        expandToValue(value, valueRep);
    }

    public void listCommand(List sourceLines)
        throws NoVMException, DebuggerException {
        if (sourceLines.size() == 0) {
            return;
        }
        SourceManager.SourceLine slFirst =
            (SourceManager.SourceLine)sourceLines.get(0);
        SourceManager.SourceLine slLast  =
            (SourceManager.SourceLine)sourceLines.get(sourceLines.size()-1);
        String sourceName = slFirst.getSourceName();
        int first = slFirst.getLineNumber();
        int last  = slLast.getLineNumber();
        if (first > 0 && last > 0) {
            int middle = first;
            ComponentRepository.getSourcePane().
                showSourceForFileAndLine(sourceName, middle);
        } else {
            ComponentRepository.getSourcePane().
                showSource(sourceName);
        }
    }

    public void listCommand(String sourceName, List sourceLines)
        throws NoVMException, DebuggerException {
        listCommand(sourceLines);
    }

    public void listCommand(String sourceName,
                            int lineNumber,
                            SourceManager.SourceLine sl)
        throws NoVMException, DebuggerException {
        List sourceLines = new Vector();
        sourceLines.add(sl);
        listCommand(sourceLines);
    }

    public void listCommand(String sourceName,
                            int startLine,
                            int endLine,
                            List sourceLines)
        throws NoVMException, DebuggerException {
        listCommand(sourceLines);
    }

    public void useCommand(String sourcePath, String newSourcePath)
        throws NoVMException, DebuggerException {
        super.useCommand(sourcePath, newSourcePath);
//        ComponentRepository.setRoot(newSourcePath);
//        ComponentRepository.getClassTreePane().loadAJCClasses(newSourcePath);
    }

    public void connectCommand(VirtualMachine vm)
        throws NoVMException, DebuggerException {
        if (vm != null) {
            result("Connected to " + vm.name());
        }
    }

    private void expand(LocatableEvent e) {
        try {
            expand(e.thread().frame(0));
        } catch (IncompatibleThreadStateException itse) {
        }
    }

    private void expand(StackFrame frame) {
        ComponentRepository.getThreadGroupTreePane().expandToFrame(frame);
        //XXX Something's forcing us to do this twice, figure it out later...
        //ComponentRepository.getThreadGroupTreePane().expandToFrame(frame);
    }

    private void expandToLocals() {
        try {
            ComponentRepository.getThreadGroupTreePane().
                expandToLocals(getDebugger().getDefaultFrame());
        } catch (NoVMException nvme) {
        }
    }

    private void expandToValue(Value value, Object valueRep) {
        ComponentRepository.getThreadGroupTreePane().
            expandToValue(value, valueRep + "");
    }

    public boolean wantsToExit() {
        return false;
    }

    public void exit() {
        gui.shutDown();
        debugger.shutDown();
        super.exit();
    }

    public boolean canRestart() {
        return true;
    }

    public static String d() { return "GUI Debugger"; }
    public String toString() { return d(); }

    public VirtualMachine connect(String vmArgs,
                                  String className,
                                  String commandLine) {
        return ConnectFrame.vm(this, vmArgs, className, commandLine);
    }
}
