/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.debugger.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import org.aspectj.util.gui.*;

/**
 * Prompts user for a line breakpoint as in JBuilder.
 *
 * @see org.aspectj.util.gui.CenteredJDialog
 * @author Jeff Palm
 */
public abstract class LineBreakpointDialog extends CenteredJDialog {

    private GUIDebugger guid;
    protected GUIDebugger guid() {
        return guid;
    }
    protected Frame gui() {
        return guid != null ? guid.getGui() : null;
    }

    private final static int TEXT_WIDTH = 30;
    private final static int LINE_WIDTH = 10;
    private final static String CLASS_NAME_STR = "Class name:";
    private final static String FILE_NAME_STR = "File name:";
    private final static String LINE_NUMBER_STR = "Line number:";
    private final static String METHOD_STR = "Method:";

    private JTextField classNameField = new JTextField(TEXT_WIDTH);
    private JTextField fileNameField = new JTextField(TEXT_WIDTH);
    private JTextField lineNumberField = new JTextField(LINE_WIDTH);
    {
        lineNumberField.setFocusAccelerator('n');
    }
    private JTextField methodField = new JTextField(TEXT_WIDTH);
    {
        methodField.setFocusAccelerator('m');
    }
    private JLabel lineNumberLabel = jlabel(LINE_NUMBER_STR);
    {
        lineNumberLabel.setLabelFor(lineNumberField);
        lineNumberLabel.setDisplayedMnemonic('n');
    }
    private JLabel methodLabel = jlabel(METHOD_STR);
    {
        methodLabel.setLabelFor(methodField);
        methodLabel.setDisplayedMnemonic('m');
    }
    private JRadioButton classNameRadioButton =
        new JRadioButton(new ClassNameSelectAction());
    {
        classNameRadioButton.setSelected(true);
        classNameRadioButton.setMnemonic('a');
    }
    private JRadioButton fileNameRadioButton =
        new JRadioButton(new FileNameSelectAction());
    {
        fileNameRadioButton.setMnemonic('F');
    }
    private ButtonGroup buttonGroup = new ButtonGroup();
    {
        
        buttonGroup.add(classNameRadioButton);
        buttonGroup.add(fileNameRadioButton);
    }
    private JButton classNameButton = new JButton(new FindClassNameAction());
    private JButton fileNameButton = new JButton(new FindFileNameAction());
    private JButton okButton = new JButton(new OKAction());
    private JButton cancelButton = new JButton(new CancelAction());    

    private final class ClassNameSelectAction extends AbstractAction {
        public ClassNameSelectAction() {
            super(CLASS_NAME_STR);
        }
        public void actionPerformed(ActionEvent e) {
            fileNameField.setEnabled(false);
            fileNameButton.setEnabled(false);
            classNameField.setEnabled(true);
            classNameButton.setEnabled(true);
            methodField.setEnabled(true);
        }
    }

    private final class FileNameSelectAction extends AbstractAction {
        public FileNameSelectAction() {
            super(FILE_NAME_STR);
        }
        public void actionPerformed(ActionEvent e) {
            classNameField.setEnabled(false);
            classNameButton.setEnabled(false);
            methodField.setEnabled(false);
            fileNameField.setEnabled(true);
            fileNameButton.setEnabled(true);
        }
    }

    private final class FindClassNameAction extends AbstractAction {
        public FindClassNameAction() {
            super("...");
        }
        public void actionPerformed(ActionEvent e) {
            classNameField.setText(findClassName());
        }
    }

    private final class FindFileNameAction extends AbstractAction {
        public FindFileNameAction() {
            super("...");
        }
        public void actionPerformed(ActionEvent e) {
            fileNameField.setText(findFileName());
        }
    }

    private final class OKAction extends AbstractAction {
        public OKAction() {
            super("OK");
        }
        public void actionPerformed(ActionEvent e) {
            ok();
        }
    }

    private final class CancelAction extends AbstractAction {
        public CancelAction() {
            super("Cancel");
        }
        public void actionPerformed(ActionEvent e) {
            cancel();
        }
    }    

    public abstract String findClassName();
    public abstract String findFileName();
    
    public LineBreakpointDialog(GUIDebugger guid) {
        super(guid != null ? guid.getGui() : null,
              "Add Line Breakpoint", true);
        this.guid = guid;
        if (guid == null) {
            addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent we) {
                        System.exit(0);
                    }
                });
        }
        
        JPanel lineBreakpointPanel = new JPanel();
        lineBreakpointPanel.setBorder(BorderFactory.createTitledBorder
                                      ("Line Breakpoint"));
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.EAST;
        c.weightx = 0.5;
        c.weighty = 0.5;
        c.anchor = GridBagConstraints.WEST;
        GridBagLayout gb = new GridBagLayout();
        gb.setConstraints(lineBreakpointPanel, c);
        lineBreakpointPanel.setLayout(gb);
        int gridy = 0;
        
        c.gridx = 0;
        c.gridy = gridy;
        gb.setConstraints(classNameRadioButton, c);
        lineBreakpointPanel.add(classNameRadioButton);
        c.gridx = 1;
        c.gridy = gridy;
        gb.setConstraints(classNameField, c);
        lineBreakpointPanel.add(classNameField);
        c.gridx = 2;
        c.gridy = gridy++;
        gb.setConstraints(classNameButton, c);
        lineBreakpointPanel.add(classNameButton);        

        c.gridx = 0;
        c.gridy = gridy;
        gb.setConstraints(fileNameRadioButton, c);
        lineBreakpointPanel.add(fileNameRadioButton);
        c.gridx = 1;
        c.gridy = gridy;
        gb.setConstraints(fileNameField, c);
        lineBreakpointPanel.add(fileNameField);
        c.gridx = 2;
        c.gridy = gridy++;
        gb.setConstraints(fileNameButton, c);
        lineBreakpointPanel.add(fileNameButton);

        c.gridx = 0;
        c.gridy = gridy;
        gb.setConstraints(lineNumberLabel, c);
        lineBreakpointPanel.add(lineNumberLabel);
        c.gridx = 1;
        c.gridy = gridy++;
        gb.setConstraints(lineNumberField, c);
        lineBreakpointPanel.add(lineNumberField);        

        JPanel bottomPanel = new JPanel();
        bottomPanel.setLayout(new FlowLayout(FlowLayout.RIGHT));
        bottomPanel.add(okButton);
        bottomPanel.add(cancelButton);
        
        getContentPane().setLayout(new BoxLayout(getContentPane(),
                                                 BoxLayout.Y_AXIS));
        getContentPane().add(lineBreakpointPanel);
        getContentPane().add(bottomPanel);
    }

    private void ok() {
        try {
            if (ok_()) {
                dispose();
            }
        } catch (Throwable t) {
            AJUtil.error(t);
        }
    }

    private boolean ok_() throws Throwable {
        int lineNumber = Integer.parseInt(lineNumberField.getText().trim());
        if (classNameRadioButton.isSelected()) {
            String className = classNameField.getText();
            if (className == null) className = "";
            className = className.trim();
            if (className.equals("")) {
                throw new Error("Class name cannot be empty");
            }
            if (guid != null) {
                guid.getDebugger().stopAtCommand(className,
                                                 lineNumber);
                return true;
            } else {
                System.out.println("className=" + className +
                                   " line=" + lineNumber);
            }
        }
        if (fileNameRadioButton.isSelected()) {
            String fileName = fileNameField.getText();
            if (fileName == null) fileName = "";
            fileName = fileName.trim();
            if (fileName.equals("")) {
                throw new Error("File name cannot be empty");
            }
            if (guid != null) {
                guid.getDebugger().stopOnCommand(fileName, lineNumber);
                return true;
            } else {
                System.out.println("fileName=" + fileName +
                                   " line=" + lineNumber);
            }
        }
        return false;
    }

    private void cancel() {
        dispose();
    }

    private JLabel jlabel(String str) {
        JLabel label = new JLabel(str);
        label.setForeground(Color.black);
        return label;
    }
}
