/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
*
* This file is part of the debugger and core tools for the AspectJ(tm)
* programming language; see http://aspectj.org
*
* The contents of this file are subject to the Mozilla Public License
* Version 1.1 (the "License"); you may not use this file except in
* compliance with the License. You may obtain a copy of the License at
* either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is AspectJ.
*
* The Initial Developer of the Original Code is Xerox Corporation. Portions
* created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
* All Rights Reserved.
*/
package org.aspectj.debugger.gui;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import org.aspectj.debugger.base.*;
import org.aspectj.debugger.request.*;
import org.aspectj.tools.ide.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.List;
import javax.swing.*;

/**
 * SourcePane.java
 *
 *
 * Created: Tue Sep 05 14:08:06 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class SourcePane
    extends AbstractSourcePane  implements ClassListener
{

    private       HashMap       models      = new HashMap();
    private final Font          font        = new Font("monospaced", Font.PLAIN, 10);
    private       JScrollPane   scrollPane;
    private       JList         jlist;
    private       TitlePanel    titlePanel  = new TitlePanel();

    public SourcePane(GUIDebugger guid) {
        super(guid);
        jlist = new JList(new DefaultListModel());
        jlist.setCellRenderer(new SourcePaneCellRenderer());
        jlist.addMouseListener(new SourcePaneMouseListener());
        jlist.setFont(font);
        add(scrollPane = new JScrollPane(jlist));
        add(titlePanel, BorderLayout.SOUTH);
        guid.getDebugger().addClassListener(this);
    }

    /************************ AbstractSourcePane ***************************/

    public boolean showSource(String relativePath) {
        SourceListModel model = getModel(relativePath);
        if (model == null) return false;
        setModel(model);
        titlePanel.setTitle(relativePath);
        return true;
    }

    public void showLineForCurrentModel(int line, boolean isAtBreakpoint) {
        if (line > 0 && line <= jlist.getModel().getSize()) {
            jlist.ensureIndexIsVisible(line-1);
            jlist.ensureIndexIsVisible(line-1-4);
            jlist.ensureIndexIsVisible(line-1+4);
            jlist.setSelectedIndex(line-1);
        }
    }

    public String getSourceName() {
        return ((SourceListModel) jlist.getModel()).getSourceName();
    }

    private void setModel(SourceListModel model) {
        if (model != null) {
            markLines(model);
            jlist.setModel(model);
        } else {
            //TODO
            System.out.println("Model is null");
        }
    }

    private SourceListModel getModel(ReferenceType refType) {
        SourceListModel model = null;
        if ((model = (SourceListModel) models.get(refType)) != null) {
            return model;
        }
        try {
            String sourceName = debugger().sourceName(refType); //refType.sourceName();
            if (false) throw new AbsentInformationException();
            if ((model = getModel(sourceName)) != null) {
                models.put(refType, model);
                return model;
            }
            String strippedSourceName = debugger().strip(sourceName);
            if ((model = getModel(strippedSourceName)) != null) {
                models.put(sourceName, model);
                models.put(refType, model);
                return model;
            }
        } catch (AbsentInformationException aie) {
        }
        return null;
    }

    private SourceListModel getModel(Request request) {
        if (request instanceof ClassMethodBreakpointRequestAction) {
            ClassMethodBreakpointRequestAction ba =
                (ClassMethodBreakpointRequestAction) request;
            return getModel(ba.getSourceName());
        } else if (request instanceof ClassBreakpointRequestAction) {
            ClassBreakpointRequestAction ba =
                (ClassBreakpointRequestAction) request;
            String className = ba.getClassName();
            SourceListModel model = (SourceListModel) models.get(className);
            if (model != null) {
                return model;
            }
            try {
                Iterator iter = debugger().getVM().allClasses().iterator();
                while (iter.hasNext()) {
                    ReferenceType refType = (ReferenceType) iter.next();
                    if (className.equals(refType.name()) &&
                        (model = getModel(refType)) != null) {
                        return model;
                    }
                }
            } catch (NoVMException nvme) {
            }
        } else if (request instanceof BreakpointRequestAction) {
            SourceLine sl = debugger.getSourceLine((BreakpointRequestAction) request);
            return getModel(sl.filename);
        }
        return null;
    }

    private SourceListModel getModel(String relativePath) {
        if (relativePath == null || relativePath.equals("<not-available>")) {
            return null;
        }
        String oldRelativePath = relativePath;
        relativePath = guid().getDebugger().removeSourcePath(relativePath);
        SourceListModel model = (SourceListModel) models.get(relativePath);
        if (model != null) {
            return model;
        }
        List lines = debugger.getSourceManager().getSourceLines(relativePath);
        if (lines != null) {
            model = new SourceListModel(lines, relativePath);
            models.put(relativePath,    model);
            models.put(oldRelativePath, model);
            String fullPath = AJLineMapper.getFullFileNameFromFile(relativePath, false).toLowerCase();
            models.put(fullPath, model);
        } else {
        }
        return model;
    }

    private void markLines(SourceListModel model) {
        try {
            Iterator iter = debugger().getVM().allClasses().iterator();
            while (iter.hasNext()) {
                ReferenceType refType = (ReferenceType) iter.next();
                try {
//                    String sourceName = refType.sourceName();
                    String sourceName = debugger().sourceName(refType);
                    if (sourceName.equals(model.getSourceName())) {
                        model.markLines(refType);
                    }
                } catch (Exception e) { //AbsentInformationException aie) {

                }
            }
        } catch (NoVMException nvme) {
        } catch (VMDisconnectedException vmde) {
        }
    }

    /************************** DebuggerListener **************************/

    public void requestSet(String filename, int line, BreakpointRequestAction ba) {
        SourceListModel model = getModel(filename);
        if (model != null) {
            model.verifyBreakpoint(line, ba);
        }
    }

    public void requestClear(String filename, int line, BreakpointRequestAction ba) {
        SourceListModel model = getModel(ba);
        if (model != null) {
            model.clearBreakpoint(line);
        }
    }

    public void requestDeferred(String filename, int line, BreakpointRequestAction ba) {
        SourceListModel model = getModel(ba);
        if (model != null) {
            model.unverifyBreakpoint(line, ba);
        }
    }

    /****************************** VMListener ******************************/

    public void vmStartEvent(VMStartEvent e) {
        try {
            Iterator iter = debugger.getVM().allClasses().iterator();
            while (iter.hasNext()) {
                ReferenceType refType = (ReferenceType) iter.next();
                SourceListModel model = getModel(refType);
                if (model != null) {
                    model.markLines(refType);
                }
            }
        } catch (NoVMException nvme) {
        }
    }

    /**************************** Inner Classes ****************************/

   class TitlePanel extends JPanel {
        private JLabel label = new JLabel();
        public TitlePanel() {
            super();
            setLayout(new FlowLayout());
            add(new JLabel("Source: "));
            add(label);
        }
        public void setTitle(String str) {
            set(str);
        }
        public void loading(String str) {
            set("Loading..." + str);
        }
        private void set(String str) {
            String relativePath = str;
            int i = relativePath.length() - 1;
            for (; i >= 0; i--) {
                char c = relativePath.charAt(i);
                if (c == File.pathSeparatorChar ||
                    c == '\\' ||
                    c == '/') break;
            }
            if (i > -1 && i < relativePath.length()-1) {
                relativePath = relativePath.substring(i+1);
            }
            label.setText(relativePath);
        }
    }

    class SourcePaneMouseListener extends MouseAdapter {
        public void mousePressed(MouseEvent e) {
            popup(e);
        }

        private void popup(MouseEvent e) {
            JList jlist = (JList) e.getSource();
            int x = e.getX();
            int y = e.getY();
            SourceManager.SourceLine sl = (SourceManager.SourceLine) jlist.getSelectedValue();
            if (SwingUtilities.isMiddleMouseButton(e) ||
                SwingUtilities.isRightMouseButton(e)) {
                show(jlist, sl, x, y);
            }
        }

        private void show(JList jlist, SourceManager.SourceLine sl, int x, int y) {
            if (sl == null) { System.out.println("sl = null");
                return;
            }
            BreakpointRequestAction req = sl.getRequest();
            if (req != null) {
                System.out.println("request=" + req.getProto());
            }
            final String sourceName = sl.getSourceName();
            final int line = jlist.getSelectedIndex() + 1;
            JPopupMenu popup = new JPopupMenu();
            JMenuItem item;
            switch (sl.getStatus()) {
                case SourceManager.SourceLine.NONE:
                    popup.add(item("stop at " + sourceName + ":" + line,
                                   "stop at line " + line));
                    break;
                case SourceManager.SourceLine.VERIFIED:
                case SourceManager.SourceLine.UNVERIFIED:
                    popup.add(item("clear " + req.getProto(),
                                   "clear " + req.getProto()));
            }
            popup.show(jlist, x + popup.getWidth()/2,
                              y + popup.getHeight()/2);
        }

        private JMenuItem item(final String setString, String titleString) {
            JMenuItem item = new JMenuItem(titleString);
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        debugger.execute(setString);
                    }
                });
            return item;
        }
    }

    class SourcePaneCellRenderer extends DefaultListCellRenderer {

        private final Font font = new Font("monospaced", Font.PLAIN, 10);
        private final Color BLUE = Color.blue.darker();
        private final Color RED  = Color.red.darker();

        public Component getListCellRendererComponent(JList list,
                                                      Object value,
                                                      int index,
                                                      boolean isSelected,
                                                      boolean cellHasFocus) {
            super.getListCellRendererComponent(list, value, index,
                                               isSelected, cellHasFocus);
            SourceManager.SourceLine sl = (SourceManager.SourceLine) value;
            if (sl.getStatus() == SourceManager.SourceLine.NONE && sl.isExecutable()) {
                setIcon(AJIcons.getIcon(AJIcons.EXECUTABLE_ICON));
            } else {
                switch (sl.getStatus()) {
                case SourceManager.SourceLine.NONE:
                    setIcon(AJIcons.getIcon(AJIcons.BLANK_ICON));
                    break;
                case SourceManager.SourceLine.UNVERIFIED:
                    setIcon(AJIcons.getIcon(AJIcons.UNVERIFIED_BREAKPOINT_ICON));
                    break;
                case SourceManager.SourceLine.VERIFIED:
                    setIcon(AJIcons.getIcon(AJIcons.VERIFIED_BREAKPOINT_ICON));
                }
            }
            String str = shrink(SourcePane.this.guid().getDebugger().sourceLine(sl));
            setText(str);
//            if (debugger.isStopped(sl)) {
//                setBackground(RED);
//            } else {
//                setBackground(DEFAULT);
//            }
            return this;
        }

        public Font getFont() {
            return font;
        }

        private String shrink(String str) {
            return str.trim();
        }
    }

    class SourceListModel
        extends AbstractListModel
        implements SourceManager.LineUpdatable {

        private List lines;
        private String sourceName;

        public SourceListModel(List lines, String sourceName) {
            this.lines = lines;
            this.sourceName = sourceName;
            setUpdater(lines);
        }

        public void update(int line) {
            fireContentsChanged(this, line, line);
        }

        public int getSize() {
            return lines.size();
        }

        public void setExecutable(int line, String className) {
            line(line).ex(className);
            update(line);
        }

        public void verifyBreakpoint(int line, BreakpointRequestAction request) {
            line(line).verify(request);
            update(line);
        }

        public void unverifyBreakpoint(int line, BreakpointRequestAction request) {
            line(line).unverify(request);
            update(line);
        }

        public void clearBreakpoint(int line) {
            line(line).clear();
            update(line);
        }

        SourceManager.SourceLine line(int line) {
            return (SourceManager.SourceLine) getElementAt(line-1);
        }

        public String getSourceName() {
            return sourceName;
        }

        public Object getElementAt(int index) {
            SourceManager.SourceLine line = SourceManager.SourceLine.emptyLine;
            try {
                line = (SourceManager.SourceLine) lines.get(index);
            } catch (ArrayIndexOutOfBoundsException aioobe) {
            }
            return line;
        }

        public boolean equals(Object other) {
            if (!(other instanceof SourceListModel)) {
                return super.equals(other);
            }
            SourceListModel model = (SourceListModel) other;
            return getSourceName().equals(model.getSourceName());
        }

        /**
         * FIXES: removed
         */
        public void markLines(ReferenceType refType) {
//            String className = refType.name();
//            try {
//                Iterator iter = refType.allLineLocations().iterator();
//                while (iter.hasNext()) {
//                    SourceLine sl = debugger().sourceLine((Location) iter.next());
//                    if (sl.line > 0) {
//                        setExecutable(sl.line, className);
//                    }
//                }
//            } catch (AbsentInformationException aie) {
//            } catch (ClassNotPreparedException cnpe) {
//            }
        }

        private void setUpdater(List lines) {
            Iterator iter = lines.iterator();
            while (iter.hasNext()) {
                ((SourceManager.SourceLine) iter.next()).setUpdater(this);
            }
        }

        public String toString() {
            return sourceName + ":" + (lines != null ? lines.size() : -1);
        }
    }

    /* ClassListener */
    public void classPrepareEvent(ClassPrepareEvent e) {
        SourceListModel model = getModel(e.referenceType());
        if (model != null) {
            model.markLines(e.referenceType());
        }
    }
    public void classUnloadEvent(ClassUnloadEvent e) {
        //!!! unmarklines?
    }

}
