/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import com.sun.jdi.*;

public class SourceTreePane extends JPanel {

    private AbstractSourcePane sourcePane;
    private JTree tree;
    protected GUIDebugger guiDebugger;

    public SourceTreePane(GUIDebugger guiDebugger, AbstractSourcePane sourcePane) {
    	super(new BorderLayout());
        this.sourcePane = sourcePane;
        this.guiDebugger = guiDebugger;
        tree = new JTree();
        setRoot(sourcePane.getSourcePath());
        tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
    	tree.addMouseListener(new SourceTreePaneMouseListener(tree));
        tree.addTreeSelectionListener(new ClassNodeSelectionListener(guiDebugger));
        tree.setCellRenderer(new AJTreeCellRenderer());
    	JScrollPane treeView = new JScrollPane(tree);
    	add(treeView);
    }

    public void setRoot(String root) {
//        tree.setModel(new SourceTreeModel(root));
        tree.setModel(new DefaultTreeModel(new RootNode(new File(root))));
    }

    private void showRunMenu(int x, int y, final String pathString) {
        showMenu(x, y, new AbstractAction("Debug class " + pathString) {
            public void actionPerformed(ActionEvent e) {
                ComponentRepository.getGUIDebugger().executeCommand("run " + pathString);
        }});
    }

    private void showUseMenu(int x, int y) {
        showMenu(x, y, new AbstractAction("Change the source path") {
            public void actionPerformed(ActionEvent e) {
                Dialogs.showUseDialog();
        }});
    }

    private void showMenu(int x, int y, AbstractAction a) {
        JPopupMenu menu = new JPopupMenu();
        menu.add(a);
        menu.show(tree, x, y);
    }

    void showSourceFile(String relativePath) {
      sourcePane.showSource(relativePath);
    }

    class RootNode extends FileNode {
        public RootNode(File file) {
            super(file);
        }

        public String toString() {
            return file + "";
        }

        public void rightMouseButton(MouseEvent e) {
            showUseMenu(e.getX(), e.getY());
        }
    }

    class FileNode extends AJTreeNode {

        protected File file;

        public FileNode(File file) {
            super();
            setUserObject(this.file = file);
            if (file != null) {
                if (file.isDirectory()) {
                    addFiles();
                } else if (!guiDebugger.getDebugger().getOptions().isSet("noajcclasses")){
                    addClasses();
                }
            }
        }

        public int getType() {
            if (file.isDirectory()) return AJIcons.FOLDER_ICON;
            return AJIcons.DOCUMENT_ICON;
        }

        private void addFiles() {
            File[] dirs = file.listFiles(new FilenameFilter() {
                    public boolean accept(File dir, String name) {
                        return new File(dir, name).isDirectory();
                    }});
            File[] javaFiles = file.listFiles(new FilenameFilter() {
                    public boolean accept(File dir, String name) {
                        return name.endsWith(".java");
                    }});
            File[] files;
            if (javaFiles == null || dirs == null) {
                files = new File[0];
            } else {
                files = new File[dirs.length + javaFiles.length];
                System.arraycopy(dirs, 0, files, 0, dirs.length);
                System.arraycopy(javaFiles, 0, files, dirs.length, javaFiles.length);
            }
            addFiles(files);
        }

        private void addFiles(File[] files) {
            for (int i = 0; i < files.length; i++) {
                add(new FileNode(files[i]));
            }
        }

        private void addClasses() {
            try {
            String fullName;
            try {
                fullName = file.getCanonicalPath();
            } catch (IOException e) {
                fullName = file.getAbsolutePath();
            }
            if (fullName == null) return;
            AJCClassNode[] nodes = AJCClassNode.makeNodes(fullName);
            for (int i = 0; i < nodes.length; i++) {
                add(nodes[i]);
            }
            } catch (Throwable t) {
                new ErrorLogger(ComponentRepository.getAJDebugger()).log(t);
            }
        }

        public void leftMouseButton(MouseEvent e) {
            String fullName;
            try {
                fullName = file.getCanonicalPath();
            } catch (IOException ee) {
                fullName = file.getAbsolutePath();
            }
            String relativePath =
                ComponentRepository.getAJDebugger().removeSourcePath(fullName);
            showSourceFile(relativePath);
        }

        public void rightMouseButton(MouseEvent e) {
            String name = file.getName();
            int ijava = name.indexOf(".java");
            if (ijava != -1) {
                name = name.substring(0, ijava);
            }
            String title = "run class " + name;
            final String runCommand = "run " + name;
            //!!! Do it later
//            showMenu(e.getX(), e.getY(), new AbstractAction(title) {
//                public void actionPerformed(ActionEvent ae) {
//                    sourcePane.debugger.execute(runCommand);
//                }
//            });

        }

        public String toString() {
            return file.getName();
        }
    }

    class SourceTreeModel implements TreeModel {

        private String root;

        public SourceTreeModel(String root) {
            this.root = root;
        }

        public void addTreeModelListener(TreeModelListener l) {
        }
        public void removeTreeModelListener(TreeModelListener l) {
        }
        public void valueForPathChanged(TreePath path, Object newValue) {
        }

        public Object getChild(Object parent, int index) {
            if (parent instanceof File) return getChildren((File) parent)[index];
            return null;
        }

        public int getChildCount(Object parent) {
            if (parent instanceof File) {
                return getChildren((File) parent).length;
            }
            return -1;
        }

        public int getIndexOfChild(Object parent, Object child) {
            if (parent instanceof File) {
                return getIndexOfChild(getChildren((File) parent), child);
            }
            return -1;
        }

        private int getIndexOfChild(File[] kids, Object child) {
            for (int i = 0; i < kids.length; i++) {
                if (kids[i].equals(child)) {
                    return i;
                }
            }
            return -1;
        }

        public Object getRoot() {
            return sourcePane.getSourceManager().getFile(root);
        }

        public boolean isLeaf(Object node) {
            return !((File) node).isDirectory();

        }

        private ClassNode[] getClassNodes(File file) {
            try {
                return AJCClassNode.makeNodes(file.getCanonicalPath());
            } catch (IOException ioe) {
            }
            return new ClassNode[0];
        }

        private File[] getChildren(File file) {
            File[] dirs = file.listFiles(new FilenameFilter() {
                    public boolean accept(File dir, String name) {
                        return new File(dir, name).isDirectory();
                    }});
            File[] javaFiles = file.listFiles(new FilenameFilter() {
                    public boolean accept(File dir, String name) {
                        return name.endsWith(".java");
                    }});
            File[] files;
            if (javaFiles == null || dirs == null) {
                files = new File[0];
            } else {
                files = new File[dirs.length + javaFiles.length];
                System.arraycopy(dirs, 0, files, 0, dirs.length);
                System.arraycopy(javaFiles, 0, files, dirs.length, javaFiles.length);
            }
            return files;
        }
    }

    class SourceTreePaneMouseListener extends ClassNodeMouseListener {

        public SourceTreePaneMouseListener(JTree tree) {
            super(SourceTreePane.this.guiDebugger, tree);
        }

        public void mouseClicked(MouseEvent e) {

            int x = e.getX();
            int y = e.getY();
            int selRow = tree().getRowForLocation(x, y);
            TreePath selPath = tree().getPathForLocation(x, y);
            if (selPath == null) return;
            AJTreeNode node = (AJTreeNode) selPath.getLastPathComponent();
            if (node instanceof FileNode) {
                node.mouseEvent(e);
                return;
            }
            super.mouseClicked(e);
        }
    }

//    class SourceTreePaneTreeCellRenderer extends DefaultTreeCellRenderer {
//        public Component getTreeCellRendererComponent(JTree tree,
//                                                      Object value,
//                                                      boolean sel,
//                                                      boolean expanded,
//                                                      boolean leaf,
//                                                      int row,
//                                                      boolean hasFocus) {
//            super.getTreeCellRendererComponent(tree, value, sel, expanded,
//                                               leaf, row, hasFocus);
//            File file = (File) value;
//            if (row != 0) {
//                setText(file.getName());
//            }
//            if (file.isDirectory()) {
//                setIcon(AJIcons.getIcon(AJIcons.FOLDER_ICON));
//            } else {
//                setIcon(AJIcons.getIcon(AJIcons.DOCUMENT_ICON));
//            }
//            return this;
//        }
//    }

    public static String d() {
        return "Source Tree Pane";
    }

    public String toString() { return d(); }
}
