/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.gui;

import org.aspectj.debugger.base.*;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import javax.swing.plaf.basic.*;
import java.util.*;
import java.util.List;

public class StackTablePane
    extends AJPanel
    implements StopListener {

    private JTable table = null;
    private JScrollPane scrollPane = null;

    public StackTablePane(GUIDebugger debugger) {
        super(debugger);
        debugger.getDebugger().addStopListener(this);
        setLayout(new BorderLayout());
        ThreadReference threadRef = null;
        try {
            threadRef = guid().getDebugger().getDefaultThread();
        } catch (DebuggerException de) {
        }
        table = new StackTable(new StackTableModel(threadRef));
        table.setDefaultRenderer(Integer.class, new IntegerTableCellRenderer());
        scrollPane = new JScrollPane(table);
        add(scrollPane, BorderLayout.CENTER);
    }

    protected JTable getTable() {
        return table;
    }

    /* StopListener */
    public void accessWatchpointEvent(AccessWatchpointEvent e) {
        showForThread(e.thread());
    }
    public void breakpointEvent(BreakpointEvent e) {
        showForThread(e.thread());
    }
    public void exceptionEvent(ExceptionEvent e) {
        showForThread(e.thread());
    }
    public void modificationWatchpointEvent(ModificationWatchpointEvent e) {
        showForThread(e.thread());
    }
    public void stepEvent(StepEvent e) {
        showForThread(e.thread());
    }

    public void showForThread(ThreadReference threadRef) {
        table.setModel(new StackTableModel(threadRef));
        try {
            table.setRowSelectionInterval(0, 0);
        } catch (IllegalArgumentException iae) {
        } catch (IndexOutOfBoundsException ioobe) {
        }
    }

    public static String d() { return "Stack Table Pane"; }
    public String toString() { return d(); }
}

class StackTable extends JTable {
    public StackTable(TableModel model) {
        super(model);
        setUI(new StackTableUI());
    }

    public void setModel(TableModel model) {
        super.setModel(model);
        TableColumn col = getColumnModel().getColumn(0);
        int charWidth = 10;
        col.setWidth(4 * charWidth);
        col.setMinWidth(3 * charWidth);
        col.setMaxWidth(5 * charWidth);
    }

    class StackTableUI extends BasicTableUI {
        protected MouseInputListener createMouseInputListener() {
            return new BasicTableUI.MouseInputHandler() {
                public void mouseClicked(MouseEvent e) {
                    int row = StackTable.this.getSelectedRow();
                    int col = StackTable.this.getSelectedColumn();
                    int x = e.getX();
                    int y = e.getY();
                    AJStackFrameFormatter.MethodAndSource
                        ms = (AJStackFrameFormatter.MethodAndSource)
                                StackTable.this.getModel().getValueAt
                                                                (row, 1);
                        if (!ms.isValid()) {
                            return;
                        }
                        String s = ms.getFullPath();
                        final int    l = ms.lineNumber;
                    if (SwingUtilities.isRightMouseButton(e) ||
                        SwingUtilities.isMiddleMouseButton(e)) {
                        JPopupMenu popup = new JPopupMenu();
                        final String source = AJLineMapper.removeRoot(s);
                        JMenuItem item = new JMenuItem("Set breakpoint");
                        popup.add(item);
                        final String cmd = "stop on " + source + ":" + l;

                        //XXX Can't get the popup right, now
                        item.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent _) {
                                //XXX Total hack
//                                ComponentRepository.getCommandLine().executeCommand(cmd);
                                Integer ing = new Integer(l);
                            }
                        });
                        popup.show(StackTable.this,
                                   x + popup.getWidth()/2,
                                   y + popup.getHeight()/2);


                    } else if (e.getClickCount() == 2) {
                        ComponentRepository.getSourcePane().showSourceForFileAndLine(s, l);
                    }
                }
            };
        }
    }
}

class StackTableModel extends DefaultTableModel {

    private final String[] colNames
        = {"#", "Location"};
    //private final Class[] colClasses = {Integer.class, String.class};
    private final Class[] colClasses
        = {Integer.class, AJStackFrameFormatter.MethodAndSource.class};

    public StackTableModel(ThreadReference threadRef) {
        if (threadRef != null) {
            colNames[1] += ": " + AJThreadFormatter.format(threadRef);
        }
        setColumnIdentifiers(colNames);
        if (threadRef != null) {
            try {
//                List stack = threadRef.frames();
                List stack = ComponentRepository.getAJDebugger().frames(threadRef);
                for (int i = 0; i < stack.size(); i++) {
                    Object[] os = { new Integer(i+1),
                                    AJStackFrameFormatter.format((StackFrame) stack.get(i)) };
                    addRow(os);
                }
            } catch (IncompatibleThreadStateException e) {
            }
        }
    }

    public Class getColumnClass(int col) {
        return colClasses[col];
    }

    public boolean isCellEditable(int row, int col) {
        return false;
    }

    String format(StackFrame frame) {
        if (frame == null) {
            return "<BAD FRAME>";
        }
        Location loc = frame.location();
        Method meth = loc.method();
        String str = meth.declaringType().name() + "." + meth.name() + " (";
        if (meth instanceof Method && ((Method)meth).isNative()) {
            str += "native method";
        } else if (loc.lineNumber() > -1) {
            try {
                str += loc.sourceName();
            } catch (AbsentInformationException e) {
                str += "<unknown>";
            }
            str += ":";
            str += loc.lineNumber();
        } else {
            str += "<unknown>";
        }
        str += ")";
        return str;
    }

    public String toString() {
        return "Stack Table Pane";
    }
}

class IntegerTableCellRenderer extends DefaultTableCellRenderer {
    public Component getTableCellRendererComponent(JTable table,
                                                   Object value,
                                                   boolean isSelected,
                                                   boolean hasFocus,
                                                   int row,
                                                   int col)
    {
        super.getTableCellRendererComponent
            (table, value, isSelected, hasFocus, row, col);

        if (value != null) {
            setHorizontalAlignment(SwingConstants.CENTER);
            setText("[" + value + "]");
        }
        return this;
    }
}
