/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.ide;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.gui.*;
import org.aspectj.debugger.request.*;

import java.io.File;

/**
 * FullPathSourceShower.java
 *
 *
 * Created: Wed Sep 20 11:00:21 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public abstract class FullPathSourceShower implements SourceShower {

    /**
     * Displays the source file <code>fullPath</code> in a window.
     * This will only get called is fullPath exists.
     *
     * @param fullPath The fully-qualified file name to display
     */
    public abstract void showSourceForFullPath(String fullPath);

    /**
     * Shows the line #<code>line</code> for the file currently being displayed.
     * This method will <b>only</b> follow a call to <code>showSourceForFullPath</code>.
     *
     * @param line The line to display
     * @param isAtBreakpoint <code>true</code> is this line is at a breakpoint
     * @see #showSourceForFullPath
     */
    public abstract void showLineForCurrentModel(int line, boolean isAtBreakpoint);

    /**
     * Should return the name of the currently-displayed source.
     *
     * @return The name of the currently-display source.
     */
    public abstract String getSourceName();

    /**
     * Fufills the <code>showSource</code> method of the <code>SourceShower<code>
     * interface, but turn sthe <code>relativePath</code> into a full path as needed
     * by <code>showSourceForFullPath</code>.
     *
     * @param relativePath The relative (to the source path set in the debugger)
     *                     path of the file to display
     * @see                #showSourceForFullPath
     */
    public boolean showSource(String relativePath) {
        String fullPath = ComponentRepository.getAJDebugger().getFullSourcePath(relativePath);
        File file = new File(fullPath);
        if (file == null || file.isDirectory() || !file.exists()) {
            return false;
        }
        showSourceForFullPath(fullPath);
        return true;
    }

    /**
     * See the description in org.aspectj.debugger.gui.AbstractSourcePane.
     */
    public /*abstract*/ void requestSet(String filename, int line,
                                        BreakpointRequestAction ba) {}

    /**
     * See the description in org.aspectj.debugger.gui.AbstractSourcePane.
     */
    public /*abstract*/ void requestClear(String filename, int line,
                                          BreakpointRequestAction ba) {}

    /**
     * See the description in org.aspectj.debugger.gui.AbstractSourcePane.
     */
    public /*abstract*/ void requestDeferred(String filename, int line,
                                             BreakpointRequestAction ba) {}


    /**
     * A dummy prototype of the behavior of this class.
     */
    public final static SourceShower proto = new FullPathSourceShower() {
        public void showSourceForFullPath(String fullPath) {
            System.out.println("fullPath:" + fullPath);
        }
        public void showLineForCurrentModel(int line, boolean isAtBreakpoint) {
            System.out.println("line:" + line + " isAtBreakpoint:" + isAtBreakpoint);
        }
        public String getSourceName() {
            return "who knows?";
        }
        public void requestSet(String filename,
                               int line,
                               BreakpointRequestAction ba) {}
        public void requestClear(String filename,
                                 int line,
                                 BreakpointRequestAction ba) {}
        public void requestDeferred(String filename,
                                    int line,
                                    BreakpointRequestAction ba) {}
    };
}
