/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.debugger.ide;

import java.io.File;

import org.aspectj.debugger.base.Options;
import org.aspectj.debugger.base.Debugger;
import org.aspectj.debugger.base.AJDebugger;
import org.aspectj.debugger.base.Modes;
import org.aspectj.debugger.gui.ComponentDirector;

/**
 * The main class to use to launch a debugger in an IDE.
 * To use correctly call the 'launch' method with the proper args.
 *
 * @see #launch
 **/
public class IDEDebugger {

    /** The main app. */
    public static ComponentDirector director;

    /** The name of the main class to launch. */
    private String mainClass;

    /** The source path to use in the debugger. */
    private String sourcePath;

    /** The classpath to use in the debugger. */
    private String classPath;

    /** The workingdir to use in the debugger. */
    private String workingdir;

    /**
     * The mode we're in, this has to be one of:
     *   - Modes.FORTE
     *   - Modes.JBUILDER
     *   - Modes.JBUILDER3
     *   - Modes.JBUILDER4
     *   - Modes.GUI
     *
     * @see org.aspectj.debugger.base.Modes
     * @see org.aspectj.debugger.base.Modes.FORTE
     * @see org.aspectj.debugger.base.Modes.JBUILDER
     * @see org.aspectj.debugger.base.Modes.JBUILDER3
     * @see org.aspectj.debugger.base.Modes.JBUILDER4
     * @see org.aspectj.debugger.base.Modes.GUI
     */
    private String mode;

    /** Args to pass to the running debugger. */
    private String[] args;

    /**
     * Launch a debugger.  Will only allow one.
     *
     * @param mainClass    The class whose 'main' method will be debugged.
     * @param sourcePath   The source path to use in the debugger.
     * @param classPath    The classpath to use in the debugger.
     * @param mode         The mode we're in, see the docs for the 'String' mode.
     * @param args         VM args to give the running application in the debugger.
     * @param shower       The 'SourceShower' to give the debugger.
     * @param ide          The 'IDEInterface' to give the component director.
     * @see                org.aspectj.debugger.base.Modes
     * @see                org.aspectj.debugger.ide.SourceShower
     **/
    public static void launch(String       mainClass,
                              String       sourcePath,
                              String       classPath,
                              String       workingdir,
                              String       mode,
                              String[]     args,
                              SourceShower shower,
                              IDEInterface ide) {
        new IDEDebugger(mainClass, sourcePath, classPath, workingdir, mode, args, shower, ide).go();
    }

    //XXX temp
    public static void launch(String       mainClass,
                              String       sourcePath,
                              String       classPath,
                              String       mode,
                              String[]     args,
                              SourceShower shower,
                              IDEInterface ide) {
        new IDEDebugger(mainClass, sourcePath, classPath, null, mode, args, shower, ide).go();
    }

    private IDEDebugger(String mainClass,
                        String sourcePath,
                        String classPath,
                        String workingdir,
                        String mode,
                        String[] args,
                        SourceShower sourceShower,
                        IDEInterface ide) {
        this.mainClass  = mainClass;
        this.sourcePath = sourcePath;
        this.classPath  = classPath;
        this.workingdir  = workingdir;
        this.mode = mode;
        this.args = args;
        if (ide == null) {
            ide = new IDEInterfaceImpl();
        }
//        if (mode.equals(Modes.FORTE)) {
//            ide.setComponentDirector(director = new ForteDebuggerFrame(args, sourceShower, ide));
//        } else if (mode.equals(Modes.JBUILDER4) ||
//                   mode.equals(Modes.JBUILDER3)) {
//            ide.setComponentDirector(director = new JBuilder4DebuggerFrame(args, sourceShower, ide));
//        } else {
//            System.err.println("Invalid mode: " + mode);
//        }
    }

    private void go() {
        director.go();
        Debugger debugger = director.getGUIDebugger().getDebugger();
        Options options = debugger.getOptions();
        if (check(classPath))  options.setClassPath(classPath);
        if (check(sourcePath)) options.setSourcePath(sourcePath);
        if (check(mainClass))  options.setMainClass(mainClass);
        debugger.setOptions(options);
        //XXX This sucks!!!
        debugger.execute("use " + debugger.getSourcePath());
        if (check(workingdir)) {
            debugger.execute("workingdir " + workingdir);
        }
//          File workingdir = ((AJDebugger)debugger).getWorkingdir();
//          if (workingdir != null) {
//              debugger.execute("workingdir " + workingdir.getAbsolutePath());
//          }
        if (director instanceof IDEComponentDirector) {
            ((IDEComponentDirector)director).finishUp();
        }
        try {
            if (check(mainClass)) debugger.run(mainClass, args);
        } catch (Throwable t) {
            org.aspectj.debugger.base.ErrorLogger.error(t);
        }
    }

    private boolean check(String str) {
        return str != null && str.length() > 0;
    }
}
