/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.debugger.ide;

import org.aspectj.debugger.gui.ComponentDirector;
import java.util.List;
import javax.swing.JComponent;

/**
 * A general-purpose interface to the AJDE stuff.
 * When needing functionality between the debugger and IDEs, expand
 * this interface and all implementing classes.
 *
 * @author Jeff Palm
 */
public interface IDEInterface {

    /**
     * Does whatever is necessary to exit, i.e. close output windows, etc...
     */
    public void exit();

    /**
     * Gives the IDEInterface a chance to init itself
     * with the IDEComponentDirector.
     *
     * @param director the IDEComponentDirector to which this
     *                 IDEInterface is bound.
     */
    public void init(IDEComponentDirector director);

    /**
     * Called from the IDEComponentDirector when it is setting breakpoints
     * from the IDE.  This should serve as a 'down' operation.
     */
    public void settingBreakpointsFromIDE();

    /**
     * Called from the IDEComponentDirector when it is done setting breakpoints
     * from the IDE source.  This should serve as an 'up' operation.
     */
    public void doneSettingBreakpointsFromIDE();

    /**
     * Returns the current classpath.
     *
     * @return The current classpath.
     */
    public String getClasspath();

    /**
     * Returns the current VM parameters.
     *
     * @return The current VM parameters.
     */
    public String getVMParameters();

    /**
     * Returns the class name to run right away.
     *
     * @return The class name to run right away.
     */
    public String getClassNameToRun();

    /**
     * Gets a class name from the user some way.
     *
     * @return The class name from the user.
     */
    public String findClassName();

    /**
     * Gets a file name from the user some way.
     *
     * @return The file name from the user.
     */
    public String findFileName();

    /**
     * Retrieves breakpoints to set initially.
     *
     * @return A java.util.List of breakpoints to set initially of
     *         type SourceLineBreakable.
     * @see    SourceLineBreakable.
     */
    public List getInitialBreakpoints();

    /**
     * Sets the ComponentDirector for use by this IDEInterface.
     *
     * @param director The ComponentDirector to use.
     */
    public void setComponentDirector(ComponentDirector director);

    /**
     * An interface allowing us to pass around
     * source-line breakpoint information.
     */
    public interface SourceLineBreakable {

        /**
         * Returns the full source name of this breakpoint.
         *
         * @return The full source name of this breakpoint.
         */
        public abstract String getSourceName();

        /**
         * Returns the line number of this breakpoint.
         *
         * @return The line number of this breakpoint.
         */
        public abstract int getLine();
    }

    /**
     * A simple implementation of SourceLinBreakable that
     * has a toString() method.
     */
    public abstract class AbstractSourceLineBreakpoint
        implements SourceLineBreakable
    {

        /**
         * @see SourceLineBreakable#getSourceName
         */
        public abstract String getSourceName();

        /**
         * @see SourceLineBreakable#getLine
         */
        public abstract int getLine();

        /**
         * Returns a String of the form source ':' line.
         *
         * @return A String of the form source ':' line.
         */
        public String toString() {
            return getSourceName() + ":" + getLine();
        }
    }
}
