/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.debugger.ide;

import org.aspectj.debugger.base.*;
import org.aspectj.debugger.gui.*;
import org.aspectj.debugger.request.*;
import org.aspectj.tools.ide.SourceLine;
import java.util.List;
import javax.swing.JComponent;

// For log
import javax.swing.*;

public abstract class IDEInterfaceHelper implements IDEInterface {

    /* ----------------------------------------------------------------------
     *  Abstract methods
     * ----------------------------------------------------------------------
     */

    public abstract String getClasspath();
    public abstract String getVMParameters();
    public abstract String getClassNameToRun();
    public abstract String findClassName();
    public abstract String findFileName();
    public abstract List getInitialBreakpoints();

    protected abstract void markLine(String className, List argumentTypeNames);
    protected abstract void markLineWithClass(String className, int line);
    protected abstract void markLine(String sourceName, int line);
    protected abstract void unmarkLine(String sourceName, int line);

    public /*static*/ void toggle(String sourceName, int lineNumber) {}


    public static IDEInterfaceHelper singletonInstance = null;
    //public abstract void toggle(String sourceName, int lineNumber);


    protected abstract void handle(Throwable t);

    /**
     * @todo    this should return a boolean or throw an exception on failure
     */
    public void toggleBreakpoint(String sourcePath, int lineNumber) {
        this.ajdb().toggleBreakpoint(sourcePath, lineNumber);
    }


    /* ----------------------------------------------------------------------
     *  IDEInterface methods
     * ----------------------------------------------------------------------
     */

    /**
     * Does nothing by default.
     *
     * @see org.aspectj.debugger.ide.IDEInterface#exit
     */
    public void exit() {}

    /**
     * Does nothing by default.
     *
     * @see org.aspectj.debugger.ide.IDEInterface#init
     */
    public void init(IDEComponentDirector director) {}

    /**
     * @see org.aspectj.debugger.ide.IDEInterface#settingBreakpointsFromIDE
     */
    public void settingBreakpointsFromIDE() {
        downSettingFromDebugger();
    }

    /**
     * @see org.aspectj.debugger.ide.IDEInterface#doneSettingBreakpointsFromIDE
     */
    public void doneSettingBreakpointsFromIDE() {
        upSettingFromDebugger();
    }


    /* ----------------------------------------------------------------------
     *  Utility members
     * ----------------------------------------------------------------------
     */

    protected ComponentDirector director;
    public void setComponentDirector(ComponentDirector director) {
        this.director = director;
        init(director);
    }


    /* ----------------------------------------------------------------------
     *  Utility methods
     * ----------------------------------------------------------------------
     */

    public ComponentDirector getComponentDirector() {
        return director;
    }

    public AJDebugger ajdb() {
        GUIDebugger guid = guid();
        return guid != null ? (AJDebugger)guid.getDebugger() : null;
    }

    public GUIDebugger guid() {
        return director != null ? director.getGUIDebugger() : null;
    }

    public void init(ComponentDirector _director) {
        director = _director;
        director.getDebugger().addDebuggerListener(new RequestListener());
    }

    final class RequestListener implements DebuggerListener {
        public void requestSetEvent(RequestEvent re) {
            debug("requestSetEvent:" + re);
            set(re);
        }
        public void requestClearEvent(RequestEvent re) {
            debug("requestClearEvent:" + re);
            clear(re);
        }
        public void requestDeferredEvent(RequestEvent re) {
            debug("requestDeferredEvent:" + re);
            set(re);
        }
        public void requestFailedEvent(RequestEvent re) {
            debug("requestFailedEvent:" + re);
            clear(re);
        }
        private void set(RequestEvent re) {
            if (!downSettingFromDebugger()) return;
            markLine(re);
            upSettingFromDebugger();
        }
        private void clear(RequestEvent re) {
            if (!downSettingFromDebugger()) return;
            unmarkLine(re);
            upSettingFromDebugger();
        }
    }

    // Mutex stuff
    private boolean settingFromIDE = false;
    protected boolean downSettingFromIDE() {
        return settingFromIDE ? false : (settingFromIDE = true);
    }
    protected void upSettingFromIDE() {
        settingFromIDE = false;
    }
    private boolean settingFromDebugger = false;
    protected boolean downSettingFromDebugger() {
        return settingFromDebugger ? false : (settingFromDebugger = true);
    }
    protected void upSettingFromDebugger() {
        settingFromDebugger = false;
    }

    protected final boolean breakpoint
        (IDEInterface.AbstractSourceLineBreakpoint breakpoint,
         boolean stopping) {
        return breakpoint(breakpoint.getSourceName(),
                          breakpoint.getLine(),
                          stopping);
    }

    protected final boolean breakpoint(String sourceName,
                                       int line,
                                       boolean stopping) {
        try {
            if (!downSettingFromDebugger()) {
                return false;
            }
            if (stopping) {
                debug("stopping " + sourceName + ":" + line);
                director.getDebugger().stopOnCommand(sourceName, line);
            } else {
                debug("clearing " + sourceName + ":" + line);
                director.getDebugger().clearOnCommand(sourceName, line);
            }
        } catch (Exception t) {
            handle(t);
        } finally {
            upSettingFromDebugger();
        }
        return true;
    }

    private void unmarkLine(RequestEvent re) {
        Request request = re.getRequest();

        debug("request:" + request);

        if (request == null) {
            return;
        }

        String sourceName;
        int line;

        if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction bra =
                (BreakpointRequestAction)request;
            sourceName = bra.getRealSourceName();
            line = bra.getRealLine();
            if (ajdb().isAjcFile(sourceName)) {
                SourceLine sourceLine = ajdb().getSourceLine(bra);
                sourceName = ajdb().getFullSourcePath(bra.getRealSourceName());
                debug("*** unmarking.real:" + sourceName + ":" + line);
                debug("*** but sourceLine:" + sourceLine);
                debug("*** bra:" + bra + ",bra.class:" + bra.getClass());
                downSettingFromIDE();
                //unmarkLine(sourceName, line);
                unmarkLine(sourceLine.filename, sourceLine.line);
                upSettingFromIDE();

                return;
            }
        }
    }

    public final void atomicMarkLine(String sourceFilename,
                                     int lineNumber) {
        downSettingFromIDE();
        markLine(sourceFilename, lineNumber);
        upSettingFromIDE();
    }

    private final void markLine(RequestEvent re) {
        Request request = re.getRequest();

        debug("request:" + request);

        if (request == null) {
            return;
        }

        String sourceName;
        int line;

        if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction bra = (BreakpointRequestAction)request;
            sourceName = bra.getRealSourceName();
            debug("sourceName:" + sourceName);
            line = bra.getRealLine();
            if (ajdb().isAjcFile(sourceName)) {
                SourceLine sourceLine = ajdb().getSourceLine(bra);
                sourceName = ajdb().getFullSourcePath(bra.getRealSourceName());
                debug("*** marking sourceLine:" + sourceLine + " " +
                      sourceName + ":" + line + "," +
                      bra.getLine() + ",class:" +
                      bra.getClass().getName());
                //markLine(sourceName, line);
                downSettingFromIDE();
                markLine(sourceLine.filename, sourceLine.line);
                upSettingFromIDE();
                return;
            }
        }

        if (request instanceof SourceLineBreakpointRequestAction) {
            SourceLineBreakpointRequestAction ra =
                (SourceLineBreakpointRequestAction)request;
            SourceLine sl = ajdb().getSourceLine(ra);
            downSettingFromIDE();
            markLine(sl.filename, sl.line);
            upSettingFromIDE();
        } else if (request instanceof ClassLineBreakpointRequestAction) {
            ClassLineBreakpointRequestAction ra =
                (ClassLineBreakpointRequestAction)request;
            String className = ra.getClassName();
            line = ra.getLine();
            downSettingFromIDE();
            markLineWithClass(className, line);
            upSettingFromIDE();
        } else if (request instanceof ClassMethodBreakpointRequestAction) {
            ClassMethodBreakpointRequestAction ra =
                (ClassMethodBreakpointRequestAction)request;
            String methodProto = ra.getMethodProto();
            List argNames = ra.getArgumentTypeNames();
            String className = ra.getClassName();
            downSettingFromIDE();
            markLine(className, argNames);
            upSettingFromIDE();
        } else if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction ra =
                (BreakpointRequestAction)request;
            SourceLine sl = ajdb().getSourceLine(ra);
            downSettingFromIDE();
            markLine(sl.filename, sl.line);
            upSettingFromIDE();
        }

        sourceName = "asdf";
        line = -123;
        if (request instanceof BreakpointRequestAction) {
            BreakpointRequestAction bra =
                (BreakpointRequestAction)request;
            sourceName = bra.getSourceName();
            line = bra.getLine();
        }
        debug("sourceName:" + sourceName + " line:" + line);
    }

    public final static Log log = new Log();
    public static void debug(Object msg) {
        log.println(msg);
    }
    public static class Log {
        private JTextArea area = new JTextArea(30,100);
        private boolean ok = false;
        public void up() {
            if (!ok) return;
            println("up");
            area.setForeground(java.awt.Color.black);
        }
        public void down() {
            if (!ok) return;
            println("down");
            area.setForeground(java.awt.Color.red.darker());
        }
        {
            JFrame frame = new JFrame("IDE Logger");
            if ((ok =
                 System.getProperty("user.name").equals("palm") &&
                 System.getProperty("no.logger") != null)) {
                frame.getContentPane().add(new JScrollPane(area));
                frame.pack();
                frame.setVisible(true);
            }
        }
        public void println() {
            println("");
        }
        public void println(Object msg) {
            if (ok) area.append(msg + "\n");
        }
    }
}
