/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.request;

import org.aspectj.debugger.base.*;
import org.aspectj.tools.ide.*;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import java.util.*;

/**
 * ClassMethodBreakpointRequestAction.java
 *
 *
 * Created: Wed Aug 23 11:36:26 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public abstract class ClassMethodBreakpointRequestAction extends ClassBreakpointRequestAction {

    private String methodProto;

    ClassMethodBreakpointRequestAction(Debugger debugger,
                                       String className,
                                       String methodProto) {
        super(debugger, className);
        this.methodProto = methodProto;
    }

    public String getMethodProto() {
        return methodProto;
    }

    public String getErrorMessage() {
        return "No method " + getMethodName() + " in " + getClassName();
    }

    public String getProto() {
        return getClassName() + "." + getMethodProto();
    }



    public boolean equals(Object o) {
        if (!(o instanceof ClassMethodBreakpointRequestAction)) {
            return super.equals(o);
        }
        ClassMethodBreakpointRequestAction ra = (ClassMethodBreakpointRequestAction)o;
        return (ra.getClassName().equals(getClassName())   &&
                ra.getMethodName().equals(getMethodName()) &&
                ra.argumentTypeNames().equals(argumentTypeNames()));
    }

    Location findLocation()
        throws NoVMException,
               MultipleLocationsException,
               UnableToSetRequestException {
        Location location = null;
        Iterator iter = vm().classesByName(className).iterator();
        while (iter.hasNext()) {
            ReferenceType refType = (ReferenceType) iter.next();
            if (refType.name().equals(className)) {
                List methods = refType.methodsByName(getMethodName());
                Iterator methodsIter = methods.iterator();
                List argumentTypeNames = argumentTypeNames();
                boolean justName = true;
                for (int i = 0; i < methodProto.length(); i++) {
                    if (!Character.isJavaIdentifierPart(methodProto.charAt(i))) {
                        justName = false;
                        break;
                    }
                }
                if (justName) {
                    if (methods.size() == 1) {
                        location = ((Method) methods.get(0)).location();
                        return location;
                    } else if (methods.size() == 0) {
                        throw new UnableToSetRequestException(this);
                    } else {
                        throw new MultipleLocationsException(className + "." + methodProto, methods);
                    }
                }
                while (methodsIter.hasNext()) {
                    Method method = (Method) methodsIter.next();
                    if (argumentTypeNames().equals(method.argumentTypeNames())) {
                        location = method.location();
                        return location;
                    }
                }
                throw new UnableToSetRequestException(this);
            }
        }
        return location;
    }

    String getMethodName() {
        int iparen = methodProto.indexOf("(");
        if (iparen != -1) {
            return methodProto.substring(0, iparen);
        }
        return methodProto;
    }

    public List getArgumentTypeNames() {
        return argumentTypeNames();
    }

    List argumentTypeNames() {
        List argNames = new ArrayList();
        int iparen = methodProto.indexOf("(");
        if (iparen == -1) {
            return argNames;
        }
        StringTokenizer tok = new StringTokenizer(methodProto, "(,;:)");
        try {
            String meth = tok.nextToken();
            while (tok.hasMoreTokens()) {
                String token = tok.nextToken();
                argNames.add(token);
            }
        } catch (NoSuchElementException e) {
        }
        return argNames;
    }

    //!!! Mapping
    public int getLine() {
        SourceLine sl = sourceLine();
        if (sl == null || sl.line < 0) {
            return super.getLine();
        }
        return sl.line;
    }

    //!!! Mapping
    public String getSourceName() {
        SourceLine sl = sourceLine();
        if (sl == null) {
            return super.getSourceName();
        }
        return sl.filename;
    }


    public SourceLine sourceLine() {
        String fullSrcPath = ajdbg().getFullSourcePathFromAJCClass(className);
        Location loc = getLocation();
        if (loc == null) {
            return ajdbg().sourceLineOfMethodThatShouldMap(getClassName(),
                                                           getMethodName(),
                                                           fullSrcPath);
        } else {
            return ajdbg().sourceLineOfMethodThatShouldMap(loc, fullSrcPath);
        }
    }
}
