/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved. 
 */

package org.aspectj.debugger.request;

import org.aspectj.debugger.base.*;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import java.util.*;

/**
 * EventRequestWrapper.java
 *
 *
 * Created: Tue Aug 22 10:48:34 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public abstract class EventRequestWrapper {

    /** The debugger from which we can get information. */
    protected Debugger debugger;

    public EventRequestWrapper(Debugger debugger) {
        this.debugger = debugger;
    }

    /**
     * When a request is made it will be compared against all loaded classes.
     * If <code>refType</code> doesn't resolve then this method returns null.
     * If no loaded classes resolve with this request it is put on a queue and
     * compared to classes as they are loaded.  Once resolved the wrapper is pulled
     * from the queue and the request is set.
     */
    public abstract EventRequest resolve(ReferenceType refType);

    protected VirtualMachine vm() throws NoVMException {
        return debugger.getVM();
    }
}

class ClassLineBreakpointRequestWrapper extends EventRequestWrapper {

    /** The String rep of the className on which this breakpoint should occur. */
    private String className;

    /** The line at which this breakpoint should occur. */
    private int line;

    ClassLineBreakpointRequestWrapper(Debugger debugger, String className, int line) {
        super(debugger);
        this.className = className;
        this.line  = line;
    }

    public EventRequest resolve(ReferenceType refType) {
        BreakpointRequest request = null;
        try {
            if (refType.name().equals(className)) {
                EventRequestManager em = vm().eventRequestManager();
                Location location = findLocation();
                if (location == null) {
                    return null;
                }
                request = em.createBreakpointRequest(location);
                request.enable();
                request.setSuspendPolicy(EventRequest.SUSPEND_NONE); //TODO: Change this to SUSPEND_ALL
            }
        } catch (NoVMException e) {
        }
        return request;
    }

    public String toString() {
        return "breakpoint " + className + ":" + line;
    }

    private Location findLocation() throws NoVMException {
        Location location = null;
        Iterator iter = vm().classesByName(className).iterator();
        while (iter.hasNext()) {
            ReferenceType refType = (ReferenceType) iter.next();
            if (refType.name().equals(className)) {
                try {
                    Iterator linesIter = refType.locationsOfLine(line).iterator();
                    while (linesIter.hasNext()) {
                        return (Location) linesIter.next();
                    }
                } catch (AbsentInformationException e) {
                    return null;
                }
            }
        }
        return location;
    }
}
