/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.request;

import org.aspectj.debugger.base.*;

import com.sun.jdi.*;
import java.util.*;

/**
 * ListRequest.java
 *
 *
 * Created: Tue Aug 29 12:53:17 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public class ListRequest extends Request {

    private String sourceName      = null;
    private int    startLineNumber = -1;
    private int    endLineNumber   = -1;

    public ListRequest(Debugger debugger) {
        this(debugger, null, -1, -1);
    }

    public ListRequest(Debugger debugger, String sourceName) {
        this(debugger, sourceName, -1, -1);
    }

    public ListRequest(Debugger debugger, String sourceName, int startLineNumber) {
        this(debugger, sourceName, startLineNumber, -1);
    }

    public ListRequest(Debugger debugger, String sourceName, int startLineNumber, int endLineNumber) {
        super(debugger);
        this.sourceName      = sourceName;
        this.startLineNumber = startLineNumber;
        this.endLineNumber   = endLineNumber;
    }

    private SourceAndLines empty(String sourceName) {
        Vector v = new Vector();
        v.add(ajdbg().emptySourceLine());
        return new SourceAndLines(sourceName, v);
    }

    public static class SourceAndLines {
        private String sourceName;
        public String getSourceName() { return sourceName; }
        private List lines;
        public List getLines() { return lines; }
        public SourceAndLines(String sourceName, List lines) {
            this.sourceName = sourceName;
            this.lines = lines;
        }
        public String toString() {
            return sourceName + ":" + lines;
        }
    }

    public Object go() throws NoVMException, DebuggerException {
        SourceManager sm = debugger.getSourceManager();
        if (sourceName == null) {
            try {
                StackFrame frame = dbg().getDefaultFrame();
                if (frame == null) {
                    return empty(sn());
                }
                int line = ajdbg().lineNumber(frame.location());
                int start = line;
                int end = line;
                if (line > 0) {
                    for (int i = 0; i < 4; i++) {
                        end++;
                        if (start > 1) {
                            start--;
                        }
                    }
                    sourceName = ajdbg().sourceName(frame.location());
                    if (sourceName == null) {
                        return empty(sn());
                    }
                }
                startLineNumber = start;
                endLineNumber = end + 2;
            } catch (AbsentInformationException aie) {
                return empty(sn());
            }
        }
        if (startLineNumber == -1) {
            return sal(sm.getSourceLines(sourceName));
        } else if (endLineNumber == -1) {
            return sal(sm.getSourceLine(sourceName, startLineNumber-1));
        } else {
            return sal(sm.getSourceLines(sourceName, startLineNumber-1, endLineNumber));
        }
    }

    private String sn() {
        return sourceName == null ? "<not found>" : sourceName;
    }

    private SourceAndLines sal(SourceManager.SourceLine line) {
        List lines = new Vector();
        lines.add(line);
        return sal(lines);
    }

    private SourceAndLines sal(List lines) {
        return new SourceAndLines(sourceName, lines);
    }
}
