/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.request;

import org.aspectj.debugger.base.*;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;
import java.util.*;

/**
 * RequestActionSupport.java
 *
 *
 * Created: Mon Aug 28 09:45:12 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public abstract class RequestActionSupport extends Request implements VMListener, StopListener {

    protected boolean      isSet = false;
    protected EventRequest request;
    protected String text;
    public RequestActionSupport setText(String text) {
        this.text = text;
        return this;
    }
    
    private boolean isDeferred = false;
    private SourceManager.SourceLine sourceLine;

    /* Make sure we disable ourselves when we remove ourselves. */
    protected boolean isEnabled = true;
    public boolean isEnabled() { return isEnabled; }

    public RequestActionSupport(Debugger debugger, String text) {
        super(debugger);
        if (/*false && */text != null && !text.trim().equals("")) debugger.addStopListener(this);
        setText(text);
    }

    public RequestActionSupport(Debugger debugger) {
        this(debugger, null);
    }

    abstract void resolveAll();

    public boolean isSetting() {
        return true;
    }

    public boolean isBreakpoint() {
        return false;
    }

    public boolean isSet() {
        return isSet;
    }

    public EventRequest setRequest(EventRequest request) {
        return (this.request = request);
    }

    protected EventRequest set() {
        add();
        try {
            resolveAll();
        } catch (VMDisconnectedException e) {
        }
        return getRequest();
    }

    protected EventRequest unset() {
        //List requests = debugger.getRequests();
        RequestAction request = remove(); //System.out.println(">>> request:" + request);
        EventRequest req = null;
        if (request != null) {
            try {
                req = request.getRequest(); //System.out.println(">>> req.request:" + req);
                if (req != null) {
                    vm().eventRequestManager().deleteEventRequest(req);
                }
            } catch (NoVMException e) {
            }
        } else {
            try {
                req = getRequest(); req = getRequest(); //System.out.println(">>> getRequest:" + req);
                if (req != null) {
                    vm().eventRequestManager().deleteEventRequest(req);
                    
                }
            } catch (NoVMException e) {
            }
        }
        return req; //getRequest();
    }

    public static String f(BreakpointRequest req) {
        String str = req + "";
        str += ", suspendPolicy=" + req.suspendPolicy();
        str += ", enabled=" + req.isEnabled();
        str += ", location=" + req.location();
        return str;
    }

    public EventRequest getRequest() {
        return request;
    }

    public boolean isDeferred() {
        return isDeferred;
    }

    protected void setDeferred(boolean isDeferred) {
        this.isDeferred = isDeferred;
    }

    protected VirtualMachine vm() throws NoVMException {
        return debugger.getVM();
    }

    protected void addListeners() {
        debugger.addVMListener(this);
    }

    protected void add() {
        addListeners();
        debugger.addRequest(this);
        isEnabled = true;
    }

    protected void removeListeners() {
        debugger.removeVMListener(this);
    }

    protected RequestAction remove() {
        removeListeners();
        RequestAction req = debugger.removeRequest(this);
        isEnabled = false;
        if (req != null) {
            dbg().fireRequestClearEvent(new RequestEvent(this, true));
        }
        return req;
    }

    public SourceManager.SourceLine getSourceLine() {
        return this.sourceLine;
    }

    public void setSourceLine(SourceManager.SourceLine sourceLine) {
        this.sourceLine = sourceLine;
    }

    protected boolean success() {
        if (isSetting()) {
            dbg().setStatus("Set " + this);
            dbg().fireRequestSetEvent(new RequestEvent(this, true));
        } else {
            dbg().setStatus("Clear " + this);
            dbg().fireRequestClearEvent(new RequestEvent(this, true));
        }
        return (isSet = true);
    }

    protected boolean defer() {
        setDeferred(true);
        dbg().fireRequestDeferredEvent(new RequestEvent(this, false));
        return (isSet = false);
    }

    protected boolean failure(boolean verbose) {
        dbg().setStatus("Deferring " + this);
        if (verbose) {
            dbg().fireRequestFailedEvent(new RequestEvent(this, false));
        }
        return (isSet = false);
    }

    private void reset() {
        isSet   = false;
        request = null;
    }

    /* VMListener */
    public void vmStartEvent(VMStartEvent e) {
        if (!isSet()) {
            resolveAll();
        }
    }
    public void vmDeathEvent(VMDeathEvent e) {}
    public void vmDisconnectEvent(VMDisconnectEvent e) {}

    /* Stop Listener */
    public void breakpointEvent(BreakpointEvent e) { }
    public void exceptionEvent(ExceptionEvent e) { }
    public void accessWatchpointEvent(AccessWatchpointEvent e) { }
    public void modificationWatchpointEvent(ModificationWatchpointEvent e) { }
    public void stepEvent(StepEvent e) { }

    protected String format(String str) { return str; }
    protected void print(Event e) {
        if (isEnabled() && text != null && e != null && e.request().equals(request)) {
            debugger.app().outln(format(text));
        }
    }
}
