/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */

package org.aspectj.debugger.request;

import org.aspectj.debugger.base.*;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;
import org.aspectj.tools.ide.SourceLine;
import java.io.*;
import java.util.*;

/**
 * SourceLineBreakpointRequestAction.java
 *
 *
 * Created: Tue Aug 29 10:50:47 2000
 *
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */

public abstract class SourceLineBreakpointRequestAction
    extends BreakpointRequestAction {

    protected String sourceName;
    protected int line;

    public SourceLineBreakpointRequestAction(Debugger debugger,
                                             String sourceName, int line) {
        super(debugger);
        System.err.println(">> source line request: " + sourceName + " : " + line);
        this.sourceName = ensureDriveIsUpperCase(sourceName);
        this.line = line;
    }

    EventRequest resolve(ReferenceType refType) throws  MultipleLocationsException,
                                                        UnableToSetRequestException {
        BreakpointRequest request = null;
        try {
            String refTypeSourceName = refType.sourceName();
            String strippedName = ajdbg().strip(getSourceName());
            if (refTypeSourceName.equals(strippedName)) {
                EventRequestManager em = vm().eventRequestManager();
                Location location = findLocation(refType);
                if (location == null) {
                    return null;
                }
                request = request(location);
            }
        } catch (NoVMException e) {
        } catch (AbsentInformationException aie) {
        }
        return request;
    }

    /**
     * Exceptions were being thrown if c:\... was used as a file path instead of
     * C:\...  This method makes sure sourceName has an upper case drive letter.
     */
    private String ensureDriveIsUpperCase(String sourceName) {
        try {
            if (Character.isLetter(sourceName.charAt(0)) &&
                sourceName.charAt(1) == ':' &&
                isFileSeparator(sourceName.charAt(2))) {
                return Character.toUpperCase(sourceName.charAt(0)) +
                       sourceName.substring(1);
            }
        } catch (ArrayIndexOutOfBoundsException aoobe) {
        }
        return sourceName;
    }

    private boolean isFileSeparator(char c) {
        return c == File.separatorChar || c == '/' || c == '\\';
    }

    private Location findLocation(ReferenceType refType)
        throws MultipleLocationsException,
               UnableToSetRequestException {
        Location location = null;
        try {
            List lines = refType.locationsOfLine(line);
            if (lines == null || lines.size() == 0) {

                //TODO
                //Need a better way to find if a class matches the
                //source desired.

                //throw new UnableToSetRequestException(this);
                return null;
            }
            location = (Location) lines.get(0);
        } catch (AbsentInformationException aie) {
            ajdbg().exception("Unable to set " + this + " : No code at line " +
                              line + " in " + refType.name(), aie);
        } catch (ClassNotPreparedException cpe) {
            ajdbg().classNotPreparedException(cpe, refType.name());
        } catch (ObjectCollectedException oce) {
            ajdbg().objectCollectedException(oce, refType.name());
        } catch (InvalidLineNumberException ilne){
            ajdbg().invalidLineNumberException(ilne, refType.name(), line);
        }
        return location;
    }

    public int getLine() {
        return line;
    }

    public String getSourceName() {
        return sourceName;
    }

    public int getRealLine() {
        SourceLine sl = sourceLine();
        if (sl == null || sl.line < 0) {
            return super.getRealLine();
        }
        return sl.line;
    }

    public String getRealSourceName() {
        SourceLine sl = sourceLine();
        if (sl == null) {
            return super.getRealSourceName();
        }
        return sl.filename;
    }

    public SourceLine sourceLine() {
        return ajdbg().getSourceLineFromSource(getSourceName(), getLine());
    }

    public String getProto() {
        SourceLine sl = ajdbg().getSourceLineFromSource(getSourceName(), getLine());
        String newSource = "<source-unkown>";
        if (sl != null) {
            newSource = ajdbg().strip(ajdbg().removeFullWorkingDir(sl.filename));
        } else {
            System.out.println("sl is NULL: " + sourceName + ":" + getLine());
        }
        return newSource +  ":" + sl.line;
    }

    public boolean equals(Object other) {
        if (other == null || !(other instanceof SourceLineBreakpointRequestAction)) {
            return super.equals(other);
        }
        SourceLineBreakpointRequestAction ba = (SourceLineBreakpointRequestAction) other;
        return (getSourceName().equals(ba.getSourceName()) &&
                getLine() == ba.getLine());
//          return (getSourceName() != null && ba.getSourceName() != null &&
//                  getSourceName().equals(ba.getSourceName()) &&
//                  getLine() == ba.getLine());
    }

    public String getErrorMessage() {
        SourceLine sl = ajdbg().getSourceLineFromSource(getSourceName(), getLine());
        String newSourceName = getSourceName();
        int    newLine       = getLine();
        if (sl != null) {
            newSourceName = ajdbg().removeSourcePath(sl.filename);
            newLine       = sl.line;
        }
        return "No code at line " + newLine + " in file " + newSourceName;
    }
}
