/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.tools.ajc;

import org.aspectj.compiler.Version;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AspectJCompiler;
import org.aspectj.util.*;

import java.io.*;
import java.util.*;
import java.text.DateFormat;

/**
 * The main method of this class is the entrypoint to the weaver.
 *
 * This parses all command line arguments and passes the results on
 * to Weaver class to do the actual work.
 *
 */
public class Main {
    static final String usage =
        "Usage: ajc <options> <source files>\n" +
        "\n" +
        "where <options> includes:\n"+
//        "  -threads <n>          How many concurrent threads to use for compilation\n"+
//        "                        <n> defaults to 0 -- multi-threading disabled\n"+
//        "                        set <n> to 2 for reasonable multi-threaded compilation\n"+
        "  -verbose              Output messages about what ajc is doing\n"+
        "  -version              Print the version of ajc               \n"+
        "  -nocomments           Don't generate any comments into the woven code\n"+
        "  -emacssym             Generate symbols used by AJDE for Emacs\n" +
        //"  -nosymbols            Don't generate .ajsym or .ajsline files\n"+
        "  -usejavac             Use javac to generate .class files\n"+
        "  -preprocess           Generate regular Java code into <workingdir>\n"+
        "                        Don't try to generate any .class files\n"+
        "  -workingdir <dir>     Only relevant with -usejavac or -preprocess modes\n"+
        "                        Specify where to place intermediate .java files\n"+
        "                        <dir> defaults to ./ajworkingdir\n"+

        "  -O                    Optimize; may hinder debugging or enlarge class files\n"+
        //!!!"  -deprecation          Output source locations where deprecated APIs are used\n"+
        "  -d <directory>        Specify where to place generated .class files\n"+
        "                        <directory> defaults to the current working dir\n"+

        "  -classpath <path>     Specify where to find user class files\n"+
        "  -bootclasspath <path> Override location of bootstrap class files\n"+
        "  -extdirs <dirs>       Override location of installed extensions\n"+
        "  -argfile <file>       the file is a line-delimited list of arguments\n"+
        "                        these arguments are inserted into the argument list\n"+
        "  -encoding <encoding>  Specify character encoding used by source files\n"+
        "  -source 1.4           Support assertions as defined in JLS-1.4\n"+
        "  -lenient              Be extra-lenient in interpreting the java specification\n"+
        "  -strict               Be extra-strict in interpreting the java specification\n"+
        "  -porting              Make the use of some features from pre-1.0 versions\n" +
        "                        of AspectJ be warnings to ease porting old code\n" +
        "\n"+
        "If an argument is of the form @<filename>, the file will be interpreted as\n"+
        "a line delimited set of arguments to insert into the argument list.";

    protected List filenames;
    //protected File currentWorkingDir = null;
    protected JavaCompiler compiler;
    //protected boolean isIncremental = false;

    /** The ErrorHandler to pass to the JavaCompiler */
    protected ErrorHandler errorHandler;

    public static void main(String[] args) throws java.io.IOException {
        new Main().realMain(args);
    }

    /**
     * Constructs the compiler using <code>errorHandler</code>.
     * @param errorHandler The errorHandler to pass to the JavaCompiler. If
     *                     this is null, the JavaCompiler will use the default
     *                     ErrorHandler.
     */
    public Main(ErrorHandler errorHandler) {
        this.errorHandler = errorHandler;
    }

    /**
     * Construct the compiler using the default ErrorHandler.
     */
    public Main() {
        this(null);
    }

    public void checkJavaVersion() {
        try {
            Class c = Class.forName("java.util.Map");
        } catch (ClassNotFoundException cnfe) {
            fail("Java 2 or later required to run ajc (jdk1.2 or 1.3); not java " +
                 System.getProperty("java.version"));
        }

        try {
            Class c = Class.forName("java.lang.CharSequence");
            //!!! insert jdk1.4 warning here
        } catch (ClassNotFoundException cnfe) {
        }
    }

    public void realMain(String[] args) {
        int exitCode = -1;
        while(true) {
            compiler = null;
            exitCode = compile(args);
            if (!shouldRecompile()) break;
        }
        exit(exitCode);
    }

    public Options getOptions() {
        return getCompiler().getOptions();
    }

    public int compile(String[] args) {
        try {
            checkJavaVersion();
            return run(args);
        } catch (CompilerErrors errs) {
            return -1;
        } catch (InternalCompilerError compilerError) {
            handleInternalError(compilerError.uncaughtThrowable, compilerError);
            //System.err.println("showing where: "+compilerError.where);
            return -2;
        } catch (ExitRequestException exitException) {
            return exitException.getValue();
        } catch (Throwable e) {
            handleInternalError(e, null);
            return -2;
        }
    }

    /**
     * Returns an <code>AspectJCompiler</code> constructed from <code>errorHandler</code>
     * if it's non-null.
     * @return An <code>AspectJCompiler</code> using <code>errorHandler</code> if it's non-null.
     */
    protected JavaCompiler getCompiler() {
        if (compiler != null) return compiler;
        if (compiler == null) {
            if (errorHandler == null) {
                compiler = new AspectJCompiler();
            } else {
                compiler = new AspectJCompiler(errorHandler);
            }
        }
        return compiler;
    }

    public void clearState() {
        if (compiler != null) {
            compiler.clearState();
        }
    }


    public int run(String[] args) {
        filenames = new LinkedList();
        parseCommandLine(args);
        getCompiler().compile(filenames);
        return 0;
    }


    public boolean shouldRecompile() {
        if (!getCompiler().getOptions().incremental) return false;

        System.out.print("ajc> ");
        System.out.flush();
        byte[] buffer = new byte[2048];
        int n = 0;
        try {
            n = System.in.read(buffer);
            if (n <= 0) throw new java.io.IOException();
        } catch (Throwable e) {
            System.out.println("bye!");
            exit(0);
        }
        if (buffer[0] >= 'a' && buffer[0] <= 'z') return false;
        return true;
    }

    private class AjcConfigParser extends ConfigParser {
        protected void parseOption(String arg, LinkedList args) {
            if (arg.equals("-d")) {
                getOptions().outputDir = makeFile(removeStringArg(args));
            } else if (arg.equals("-workingdir")) {
                getOptions().workingDir = makeFile(removeStringArg(args));
            } else if (arg.equals("-version")) {
                printVersion();
                internalExit(0);
            } else if (arg.equals("-showversion")) {
                printVersion();
            } else if (arg.startsWith("-")) {
                String key = arg.substring(1);
                Object value = Boolean.TRUE;
                if (getOptions().isStringOption(key)) {
                    value = removeStringArg(args);
                } else if (getOptions().isIntegerOption(key)) {
                    value = Integer.valueOf(removeStringArg(args));
                }
                if (!getOptions().set(key, value)) {
                    showWarning("invalid flag: " + key);
                    displayHelpAndExit(null);
                }
            } else {
                showWarning(arg+" is an invalid option or argument.");
                displayHelpAndExit(null);
            }
        }

        protected void showError(String message) {
            showWarning(message);
            internalExit(1);
        }
    }

    protected void parseCommandLine(String[] args) {
        if (args.length == 0) displayHelpAndExit(null);
        
        AjcConfigParser configParser = new AjcConfigParser();
        configParser.parseCommandLine(args);
        filenames = configParser.getFiles();

        if (filenames.size() == 0) {
            displayHelpAndExit("Nothing to compile.");
        }
    }

    protected void printVersion() {
        DateFormat formatter = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.SHORT);
        TimeZone tz = TimeZone.getTimeZone("PST");
        formatter.setTimeZone(tz);
        Date buildDate = new Date(getCompiler().getBuildTime());
        String dateString = formatter.format(buildDate) + " " + tz.getDisplayName(false, TimeZone.SHORT);
        System.err.print("ajc version " + getCompiler().getVersion() + 
                         " (built " + dateString + ")");
        System.err.println(" running on java " + System.getProperty("java.version"));
    }

    /* This section of code handles errors that occur during compilation */
    static final String internalErrorMessage =
        "Please copy the following text into an email message and send it,\n" +
        "along with any additional information you can add to:            \n" +
        "                                                                 \n" +
        "       support@aspectj.org                            \n" +
        "                                                                 \n";

    static final String outOfMemoryMessage =
        "The compiler ran out of memory while trying to compile your system.\n" +
        "By default, only 64MB are available to ajc.  You can increase\n" +
        "this amount by editing the ajc launch script to pass -Xmx128M or\n" +
        "some larger amount of memory to the JVM used to run ajc.\n" +
        "More information about this issue can be found in the FAQ at\n" +
        "http://aspectj.org/faq by searching for \"OutOfMemoryError\".\n";

    static final String generalCodeRequest =
        "If possible, please include a small fragment of code that can \n"+
        "reproduce the error, provided that code isn't confidential.  \n";


    public void handleInternalError(Throwable uncaughtThrowable,
                                    InternalCompilerError compilerError)
    {
        if (uncaughtThrowable instanceof OutOfMemoryError) {
            System.err.println(outOfMemoryMessage);
            return;
        }

        System.err.println("An internal error occured in AspectJ-"+getCompiler().getVersion());
        System.err.println(internalErrorMessage);
        System.err.println(uncaughtThrowable.toString());
        uncaughtThrowable.printStackTrace();
        System.err.println();

        if (compilerError != null) {
            compilerError.showWhere(new PrintWriter(System.err));
        }
        System.err.println(generalCodeRequest);
    }


    protected void displayHelpAndExit(String message) {
        if (message != null) System.err.println(message);
        System.err.println();
        System.err.println(usage);
        internalExit(0);
    }

    protected void fail(Exception e) {
        fail(e.toString());
    }

    protected void fail(String message) {
        System.err.println(message);
        internalExit(1);
    }

    void internalExit(int value) {
        System.out.flush();
        System.err.flush();
        throw new ExitRequestException(value);
    }

    protected void exit(int value) {
        System.out.flush();
        System.err.flush();
        System.exit(value);
    }
}
