/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.util;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import org.aspectj.tools.ide.SymbolManager;
import org.aspectj.tools.ide.Declaration;
import org.aspectj.tools.ide.SourceLine;

/**
 * A class to map <code>java.lang.Throwable</code> stack traces.
 * <p>
 * The methods offered are:
 * <tr>
 *   <p><th><code>org.aspectj.util.ExceptionMapper#printStackTrace(Throwable)</code></th>
 *   <p><th><code>org.aspectj.util.ExceptionMapper#printStackTrace(Throwable,java.io.PrintStream)</code></th>
 *   <p><th><code>org.aspectj.util.ExceptionMapper#printStackTrace(Throwable,java.io.PrintWriter)</code></th> 
 * </tr>
 * <p>
 * These methods mimic those found in <code>java.lang.Throwable</code>:
 * <tr>
 *   <p><th><code>java.lang.Throwable#printStackTrace()</code></th>
 *   <p><th><code>java.lang.Throwable#printStackTrace(java.io.PrintStream)</code></th>
 *   <p><th><code>java.lang.Throwable#printStackTrace(java.io.PrintWriter)</code></th> 
 * </tr>
 *
 * @see    java.lang.Throwable
 * @author <a href="mailto:palm@parc.xerox.com"Jeffrey Palm</a>
 */
public class ExceptionMapper {

    /** The <code>org.aspectj.tools.ide.SymbolManager</code> with which to map. */
    private SymbolManager sm;

    /** The <code>java.io.File</code> to use as the source path. */
    private File sourcepath;

    /** The <code>java.io.File</code> to use as the workingdir. */
    private File workingdir;

    /**
     * Constructs a new mapper from the <code>sourcepath</code>.
     * This ctor is private because as it stands SymbolManager sucks and
     * we have to take steps to ensure that <code>sourcepath.getAbsolutPath()</code>
     * returns something we can use with <code>SymbolManager</code>.
     *
     * @param sourcepath the java.io.File to use as <code>sourcepath</code>.
     */
    private ExceptionMapper(File sourcepath) {
        this.sourcepath = sourcepath;
        workingdir = new File(sourcepath, "ajworkingdir");
        sm = SymbolManager.getSymbolManager();
    }

    /**
     * Constructs a new mapper by taking the (possibly relative) <code>sourcepath</code>.
     *
     * @param sourcepath (possibly relative) path to source.
     */
    public ExceptionMapper(String sourcepath) {
        this(validate(sourcepath));
    }

    /**
     * Constructs a new mapper using the current directory for the source path.
     */
    public ExceptionMapper() {
        this(".");
    }

    /**
     * Maps the stack trace of <code>t</code>.
     *
     * @param t the <code>java.lang.Throwable</code> whose stack trace is mapped.
     * @see     java.lang.Throwable#printStackTrace()
     */
    public void printStackTrace(Throwable t) {
        System.err.println(map(t));
    }

    /**
     * Maps the stack trace of <code>t</code> onto <code>out</code>.
     *
     * @param t    the <code>java.lang.Throwable</code> whose stack trace is mapped.
     * @param out  the <code>java.io.PrintStream</code> to which <code>t</code>'s stack
     *             will be put.
     * @see        java.lang.Throwable#printStackTrace(java.io.PrintStream)
     */
    public void printStackTrace(Throwable t,PrintStream out) {
        out.println(map(t));
    }

    /* Maps the stack trace of <code>t</code> onto <code>out</code>.
     *
     * @param t    the <code>java.lang.Throwable</code> whose stack trace is mapped.
     * @param out  the <code>java.io.PrintWriter</code> to which <code>t</code>'s stack
     *             will be put.
     * @see        java.lang.Throwable#printStackTrace(java.io.PrintWriter)
     */
    public void printStackTrace(Throwable t, PrintWriter out) {
        out.println(map(t));
    }

    private static File validate(String fileName) {
        String path;
        try {
            path = new File(fileName).getCanonicalPath();
        } catch (IOException ioe) {
            path = new File(fileName).getAbsolutePath();
        }
        while (!Character.isLetterOrDigit(path.charAt(path.length()-1))) {
            path = path.substring(0, path.length()-1);
        }
        return new File(path);
    }    

    private String map(Throwable t) {
        return map(collect(t));
    }
    
    private String collect(Throwable t) {
        StringWriter sout = new StringWriter();
        PrintWriter out = new PrintWriter(sout);
        t.printStackTrace(out);
        return sout + "";
    }
    
    private final static String newline = System.getProperty("line.separator");
    private String map(String backtrace) {
        String result = "";
        StringTokenizer tok = new StringTokenizer(backtrace, newline, false);
        result += tok.nextToken() + newline;
        while (tok.hasMoreTokens()) {
            result += frame(tok.nextToken()) + newline;
        }
        return result;
    }

    private String frame(String frame) {
        int i, iclass, ilastdot, ilparen, icolon, irparen;
        char[] chars = frame.toCharArray();
        for (i        = 0;       chars[i]        != 'a'; i       ++) {}
        for (iclass   = i;       chars[iclass++] != ' ';           ) {}
        for (ilparen  = iclass;  chars[ilparen]  != '('; ilparen ++) {}
        for (ilastdot = ilparen; chars[ilastdot] != '.'; ilastdot--) {}
        for (icolon   = ilparen; chars[icolon]   != ':'; icolon  ++) {}
        for (irparen  = icolon;  chars[irparen]  != ')'; irparen ++) {}
        return map(new String(chars, iclass,     ilastdot-iclass),
                   new String(chars, ilastdot+1, ilparen-ilastdot-1),
                   new String(chars, ilparen+1,  icolon-ilparen-1),
                   new String(chars, icolon+1,   irparen-icolon-1));
    }

    private String map(String s0, String s1, String s2, String num) {
        return map(s0, s1, s2, Integer.parseInt(num));
    }

    private String map(String className, String method, String source, int line) {
        SourceLine sl = mapToSourceLine(className, line);
        return sl != null
            ? frame(className, method, strip(sl.filename), sl.line)
            : frame(className, method, source,             line);
    }

    private String strip(String filename) {
        int isep = filename.lastIndexOf(File.separator);
        return isep != -1 ? filename.substring(isep+1) : filename;
    }
    
    private String frame(String className, String method, String source, int line) {
        return "\tat " + className + "." + method + "(" + source + ":" + line + ")";
    }

    private Map classToFile = new HashMap();
    private File fileForClass(String className) {
        return (File) classToFile.get(className);
    }
    private File resolve(String className) {
        File result = fileForClass(className);
        if (result != null) return result;
        String resolvedName = className.replace('.', '/');
        int idollar = resolvedName.indexOf('$');
        if (idollar != -1) {
            resolvedName = resolvedName.substring(0, idollar);
        }
        result = new File(workingdir + File.separator + resolvedName + ".java");
        classToFile.put(className, result);
        return result;
    }

    private SourceLine mapToSourceLine(String className, int line) {
        return sm.mapToSourceLine(resolve(className).getAbsolutePath(), line-1);
    }
}
