/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.util;

import java.io.File;
import java.util.*;

/** This class is responsible for parsing the information stored in
    the SourceFile and LineNumberTable attributes of classfiles
    produced by AspectJ.  Said names/numbers are currently produced by
    {@link org.aspectj.compiler.base.bcg.ClassfileBuilder#resolve()}.
    AspectJ produces classfiles that have multiple source files.  The
    source file attribute of such files looks like:

<pre>
Attribute := Filename SecondaryFiles
SecondaryFiles := Empty | ';' Pathname '(' Integer 'k' ')' SecondaryFiles
Filename := any sequence of characters not including ';' or '/'
Pathname := Filename | /-separated-package-name '/' Filename
</pre>

    <p> The <em>meaning</em> of such a string is an ordered set of
    triples of packages, filenames, and integer multiples of a
    thousand.  The first filename listed implicitly is associated with
    '0k'.  So, the <em>meaning</em> of the string </p>

<pre>
Foo.java;org/aspectj/Goo Bear.java(3k);strange\filename:d(x(7k)
</pre>

    <p> is the sorted association </p>

<pre>
0k  Foo's package    File Foo.java 
3k  pkg org.aspectj  File Goo Bear.java
7k  default pkg      File strange\filename:d(x
</pre>

    <p> which means that line numbers in the interval (0k, 3k) are
    from Foo.java, those in the interval (3k, 7k) are from the next
    file, and those in the interval (7k, +inf) are from the third. </p>

    <p> In addition, the line number in the association forms an
    <em>offset</em>.  So the line number 3137 is in the second file,
    and in fact is line 137 of the second file.  </p>

    <p> A mapper is created by passing in a string (the source file
    attribute), a string (the package name of the classfile), and a
    File object (the root from which to find filenames) to the
    constructor.  Then the mapper can be passed line numbers
    to determine what file the line number is in, and what is the
    actual line number in that file.  </p>

*/

public class LineNumberTableMapper {

    private static final char PACKAGE_SEPARATOR = '/';
    private static final char FILE_SEPARATOR = ';';

    /** Create a new LineNumberTableMapper given the string of a
        SourceFile attribute.

        @param attrib The SourceFile attribute string
        @param pkg This class's package, with '/' separators (NOT
        '.').  For the default package, pass in null.
        @param root The package root for this class */

    public LineNumberTableMapper(String attrib, String pkg, File root) {
        int len = attrib.length();
        int i = attrib.indexOf(FILE_SEPARATOR);
        if (i == -1) {
            files = new File[] { buildFile(root, 
                                           (pkg == null) 
                                           ? attrib 
                                           : pkg + PACKAGE_SEPARATOR + attrib) };
            offsets = new int[] { 0 };
        } else {
            List strings = new Vector();
            List ints = new Vector();
            strings.add(attrib.substring(0, i));
            ints.add(new Integer(0));
            while (i != -1) {
                i++;
                int start = i;
                i = attrib.indexOf(FILE_SEPARATOR, start);
                int end = (i == -1) ? len : i;
                end--; // now pointing at ')'
                end--; // now pointing at 'k'
                end--; // now pointing at digit
                int accumulator = 0;
                int power = 1;
                for (char c = attrib.charAt(end); c != '('; end--, c = attrib.charAt(end)) {
                    accumulator += Character.digit(c, 10) * power;
                    power *= 10;
                }
                strings.add(attrib.substring(start, end));
                ints.add(new Integer(accumulator * 1000));
            }
            int size = strings.size();
            files = new File[size];
            offsets = new int[size];
            for (int j = 0; j < size; j++) {
                files[j] = buildFile(root, (String) strings.get(j));
                offsets[j] = ((Integer) ints.get(j)).intValue();
            }
        }
    }

    /** Builds a new file by adding a string relative path to a File
        root.  This should be the same thing as {@link File(File,
        String)}, but there the concatenation is based on the system
        path separator.  Here, we know exactly that path is '/'
        separated.

        @param root the root file to build upon
        @param path the '/'-separated child path to add
        @return a new file 
    */
    private static File buildFile(File root, String path) {
        for (int i = 0, j = path.indexOf(PACKAGE_SEPARATOR);
             true; 
             i = j + 1, j = path.indexOf(PACKAGE_SEPARATOR, i)) {
            if (j == -1) {
                return new File(root, path.substring(i, path.length()));
            } else {
                root = new File(root, path.substring(i, j));
            }
        }
    }

    /** Returns the File corresponding to the region of a source
        line.
	
        @param n the source line in question, from the LineNumberTable
        attribute

        @return a file representing the source file for that line */

    public File getCorrespondingFile(int n) {
        int end = files.length - 1;
        for (int i = 1; i <= end; i++) {
            if (n < offsets[i]) return files[i - 1];
        }
        return files[end];
    }
    
    /** Returns the actual line number corresponding to a line number
        in the line number table attribute.
	
        @param n the source line in question, from the LineNumberTable
        attribute.

        @return the line number in its true source file. 
    */
    public int getCorrespondingLineNumber(int n) {
        int end = offsets.length - 1;
        for (int i = 1; i <= end; i++) {
            if (n < offsets[i]) return n - offsets[i - 1];
        }
        return n - offsets[end];
    }

    /** Returns the fictitious line number corresponding to a line
        number of a particular file.  Users of this method should
        first check that this classfile actually has code
        corresponding to the source file + line number pair.  This
        procedure will signal an exception if it has no code from the
        source file, but will return an erronious result if it has
        code from the source file, but no code from the passed in line
        number.

        @param f the source file containing the line number
        @param n the line number in question
        
        @return a line number suitable for setting a breakpoint
        @throws IllegalArgumentException if this classfile contains no
        code from the passed in source file. 

    */
    public int getCorrespondingLineNumber(File f, int n) {
        for (int i = 0, len = files.length; i < len; i++) {
            if (f.equals(files[i])) {
                return offsets[i] + n;
            } 
        }
        throw new IllegalArgumentException("No code from " 
                                           + f 
                                           + " in this classfile");
    }

    // ------------------------------
    // fields

    private File[] files;
    private int[] offsets;

    // ------------------------------
    // debugging

    public String toString() {
        String s = "";
        for (int i = 0; i < files.length - 1; i++) {
            s += offsets[i] + " " + files[i] + ":";
        }
        s += offsets[files.length - 1] + " " 
            + files[files.length - 1];
        return s;
    }

    public static void main(String[] args) {
        System.err.println(new LineNumberTableMapper(args[0], "", new File(".")));
    }
}
